@@ -42,7 +42,8 @@
 #          ifdef WIN32_THREADS
 	      GC_API CRITICAL_SECTION GC_allocate_ml;
 #          else
-#             ifdef IRIX_THREADS
+#             if defined(IRIX_THREADS) || defined(LINUX_THREADS) \
+		 || defined(IRIX_JDK_THREADS)
 #		ifdef UNDEFINED
 		    pthread_mutex_t GC_allocate_ml = PTHREAD_MUTEX_INITIALIZER;
 #		endif
@@ -61,24 +62,32 @@
 GC_FAR struct _GC_arrays GC_arrays /* = { 0 } */;
 
 
-bool GC_debugging_started = FALSE;
+GC_bool GC_debugging_started = FALSE;
 	/* defined here so we don't have to load debug_malloc.o */
 
 void (*GC_check_heap)() = (void (*)())0;
 
+void (*GC_start_call_back)() = (void (*)())0;
+
 ptr_t GC_stackbottom = 0;
 
-bool GC_dont_gc = 0;
+GC_bool GC_dont_gc = 0;
 
-bool GC_quiet = 0;
+GC_bool GC_quiet = 0;
+
+#ifdef FIND_LEAK
+  int GC_find_leak = 1;
+#else
+  int GC_find_leak = 0;
+#endif
 
 /*ARGSUSED*/
-void * GC_default_oom_fn GC_PROTO((size_t bytes_requested))
+GC_PTR GC_default_oom_fn GC_PROTO((size_t bytes_requested))
 {
     return(0);
 }
 
-void * (*GC_oom_fn) GC_PROTO((size_t bytes_requested)) = GC_default_oom_fn;
+GC_PTR (*GC_oom_fn) GC_PROTO((size_t bytes_requested)) = GC_default_oom_fn;
 
 extern signed_word GC_mem_found;
 
@@ -153,12 +162,11 @@ extern signed_word GC_mem_found;
 	/* If we can fit the same number of larger objects in a block,	*/
 	/* do so.							*/ 
 	{
+	    size_t number_of_objs = BODY_SZ/word_sz;
+	    word_sz = BODY_SZ/number_of_objs;
 #	    ifdef ALIGN_DOUBLE
-#	        define INCR 2
-#	    else
-#		define INCR 1
+		word_sz &= ~1;
 #	    endif
-	    while (BODY_SZ/word_sz == BODY_SZ/(word_sz + INCR)) word_sz += INCR;
 	}
     	byte_sz = WORDS_TO_BYTES(word_sz);
 #	ifdef ADD_BYTE_AT_END
@@ -180,7 +188,11 @@ extern signed_word GC_mem_found;
  * sections of the stack whenever we get control.
  */
 word GC_stack_last_cleared = 0;	/* GC_no when we last did this */
-# define CLEAR_SIZE 213
+# ifdef THREADS
+#   define CLEAR_SIZE 2048
+# else
+#   define CLEAR_SIZE 213
+# endif
 # define DEGRADE_RATE 50
 
 word GC_min_sp;		/* Coolest stack pointer value from which we've */
@@ -229,8 +241,6 @@ word limit;
 }
 #endif
 
-extern ptr_t GC_approx_sp();	/* in mark_rts.c */
-
 /* Clear some of the inaccessible part of the stack.  Returns its	*/
 /* argument, so it can be used in a tail call position, hence clearing  */
 /* another frame.							*/
@@ -239,7 +249,7 @@ ptr_t arg;
 {
     register word sp = (word)GC_approx_sp();  /* Hotter than actual sp */
 #   ifdef THREADS
-        word dummy[CLEAR_SIZE];;
+        word dummy[CLEAR_SIZE];
 #   else
     	register word limit;
 #   endif
@@ -328,8 +338,7 @@ ptr_t arg;
     /* Make sure r points to the beginning of the object */
 	r &= ~(WORDS_TO_BYTES(1) - 1);
         {
-	    register int offset =
-	        	(char *)r - (char *)(HBLKPTR(r)) - HDR_BYTES;
+	    register int offset = (char *)r - (char *)(HBLKPTR(r));
 	    register signed_word sz = candidate_hdr -> hb_sz;
 	    
 #	    ifdef ALL_INTERIOR_POINTERS
@@ -390,7 +399,7 @@ size_t GC_get_bytes_since_gc GC_PROTO(())
     return ((size_t) WORDS_TO_BYTES(GC_words_allocd));
 }
 
-bool GC_is_initialized = FALSE;
+GC_bool GC_is_initialized = FALSE;
 
 void GC_init()
 {
@@ -408,9 +417,7 @@ void GC_init()
     extern void GC_init_win32();
 #endif
 
-#if defined(SOLARIS_THREADS) || defined(IRIX_THREADS)
-    extern void GC_thr_init();
-#endif
+extern void GC_setpagesize();
 
 void GC_init_inner()
 {
@@ -419,18 +426,30 @@ void GC_init_inner()
 #   endif
     
     if (GC_is_initialized) return;
+    GC_setpagesize();
+    GC_exclude_static_roots(beginGC_arrays, end_gc_area);
+#   ifdef PRINTSTATS
+	if ((ptr_t)endGC_arrays != (ptr_t)(&GC_obj_kinds)) {
+	    GC_printf0("Reordering linker, didn't exclude obj_kinds\n");
+	}
+#   endif
 #   ifdef MSWIN32
  	GC_init_win32();
 #   endif
+#   if defined(LINUX) && (defined(POWERPC) || defined(ALPHA) || defined(SPARC))
+	GC_init_linux_data_start();
+#   endif
 #   ifdef SOLARIS_THREADS
 	GC_thr_init();
 	/* We need dirty bits in order to find live stack sections.	*/
         GC_dirty_init();
 #   endif
-#   ifdef IRIX_THREADS
-	GC_thr_init();
+#   if defined(IRIX_THREADS) || defined(LINUX_THREADS) \
+       || defined(IRIX_JDK_THREADS)
+        GC_thr_init();
 #   endif
-#   if !defined(THREADS) || defined(SOLARIS_THREADS) || defined(WIN32_THREADS) || defined(IRIX_THREADS)
+#   if !defined(THREADS) || defined(SOLARIS_THREADS) || defined(WIN32_THREADS) \
+       || defined(IRIX_THREADS) || defined(LINUX_THREADS)
       if (GC_stackbottom == 0) {
 	GC_stackbottom = GC_get_stack_base();
       }
@@ -457,9 +476,11 @@ void GC_init_inner()
         if ((word)(&dummy) > (word)GC_stackbottom) {
           GC_err_printf0(
           	"STACK_GROWS_DOWN is defd, but stack appears to grow up\n");
-          GC_err_printf2("sp = 0x%lx, GC_stackbottom = 0x%lx\n",
-          	   	 (unsigned long) (&dummy),
-          	   	 (unsigned long) GC_stackbottom);
+#	  ifndef UTS4  /* Compiler bug workaround */
+            GC_err_printf2("sp = 0x%lx, GC_stackbottom = 0x%lx\n",
+          	   	   (unsigned long) (&dummy),
+          	   	   (unsigned long) GC_stackbottom);
+#	  endif
           ABORT("stack direction 3\n");
         }
 #     else
@@ -488,10 +509,10 @@ void GC_init_inner()
     }
     
     /* Add initial guess of root sets.  Do this first, since sbrk(0)	*/
-    /* mightbe used.							*/
+    /* might be used.							*/
       GC_register_data_segments();
     GC_init_headers();
-        GC_bl_init();
+    GC_bl_init();
     GC_mark_init();
     if (!GC_expand_hp_inner((word)MINHINCR)) {
         GC_err_printf0("Can't start up: not enough memory\n");
@@ -543,15 +564,17 @@ void GC_init_inner()
 
 void GC_enable_incremental GC_PROTO(())
 {
+# if !defined(SMALL_CONFIG)
+  if (!GC_find_leak) {
     DCL_LOCK_STATE;
     
-# ifndef FIND_LEAK
     DISABLE_SIGNALS();
     LOCK();
     if (GC_incremental) goto out;
+    GC_setpagesize();
 #   ifdef MSWIN32
       {
-        extern bool GC_is_win32s();
+        extern GC_bool GC_is_win32s();
 
 	/* VirtualProtect is not functional under win32s.	*/
 	if (GC_is_win32s()) goto out;
@@ -580,6 +603,7 @@ void GC_enable_incremental GC_PROTO(())
 out:
     UNLOCK();
     ENABLE_SIGNALS();
+  }
 # endif
 }
 
@@ -745,6 +769,39 @@ char * msg;
 }
 #endif
 
+#ifdef NEED_CALLINFO
+
+void GC_print_callers (info)
+struct callinfo info[NFRAMES];
+{
+    register int i;
+    
+#   if NFRAMES == 1
+      GC_err_printf0("\tCaller at allocation:\n");
+#   else
+      GC_err_printf0("\tCall chain at allocation:\n");
+#   endif
+    for (i = 0; i < NFRAMES; i++) {
+     	if (info[i].ci_pc == 0) break;
+#	if NARGS > 0
+	{
+	  int j;
+
+     	  GC_err_printf0("\t\targs: ");
+     	  for (j = 0; j < NARGS; j++) {
+     	    if (j != 0) GC_err_printf0(", ");
+     	    GC_err_printf2("%d (0x%X)", ~(info[i].ci_arg[j]),
+     	    				~(info[i].ci_arg[j]));
+     	  }
+	  GC_err_printf0("\n");
+	}
+# 	endif
+     	GC_err_printf1("\t\t##PC##= 0x%X\n", info[i].ci_pc);
+    }
+}
+
+#endif /* SAVE_CALL_CHAIN */
+
 # ifdef SRC_M3
 void GC_enable()
 {
