@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 1991-1995 by Xerox Corporation.  All rights reserved.
+ * Copyright (c) 1996-1997 by Silicon Graphics.  All rights reserved.
  *
  * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
  * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
@@ -15,21 +16,43 @@
    and the work-around for SIG_FILL on gnu-win32.  -fjh. */
 
 # include "gc_priv.h"
-# ifdef LINUX
-    /* For Linux 2.1.x, `struct sigcontext_struct' has been renamed as  */
-    /* `struct sigcontext'; hence the following #define.		*/
-#   define sigcontext_struct sigcontext
-    /* Ugly hack to get struct sigcontext_struct definition.  Required	*/
-    /* for some early 1.3.X releases.  Will hopefully go away soon.	*/
-    /* in some later Linux releases, asm/sigcontext.h may have to	*/
-    /* be included instead.						*/
-#   define __KERNEL__
-#   include <asm/signal.h>
-#   undef __KERNEL__
+
+# if defined(LINUX) && !defined(POWERPC)
+#   include <linux/version.h>
+#   if (LINUX_VERSION_CODE <= 0x10400)
+      /* Ugly hack to get struct sigcontext_struct definition.  Required      */
+      /* for some early 1.3.X releases.  Will hopefully go away soon. */
+      /* in some later Linux releases, asm/sigcontext.h may have to   */
+      /* be included instead.                                         */
+#     define __KERNEL__
+#     include <asm/signal.h>
+#     undef __KERNEL__
+#   else
+      /* Kernels prior to 2.1.1 defined struct sigcontext_struct instead of */
+      /* struct sigcontext.  libc6 (glibc2) uses "struct sigcontext" in     */
+      /* prototypes, so we have to include the top-level sigcontext.h to    */
+      /* make sure the former gets defined to be the latter if appropriate. */
+#     include <features.h>
+#     if 2 <= __GLIBC__
+#       if 2 == __GLIBC__ && 0 == __GLIBC_MINOR__
+	  /* glibc 2.1 no longer has sigcontext.h.  But signal.h	*/
+	  /* has the right declaration for glibc 2.1.			*/
+#         include <sigcontext.h>
+#       endif /* 0 == __GLIBC_MINOR__ */
+#     else /* not 2 <= __GLIBC__ */
+        /* libc5 doesn't have <sigcontext.h>: go directly with the kernel   */
+        /* one.  Check LINUX_VERSION_CODE to see which we should reference. */
+#       include <asm/sigcontext.h>
+#     endif /* 2 <= __GLIBC__ */
+#   endif
 # endif
 # if !defined(OS2) && !defined(PCR) && !defined(AMIGA) && !defined(MACOS)
 #   include <sys/types.h>
+#   if !defined(MSWIN32) && !defined(SUNOS4)
+#   	include <unistd.h>
+#   endif
 # endif
+
 # include <stdio.h>
 # include <signal.h>
 
@@ -52,6 +75,10 @@
 #   define NEED_FIND_LIMIT
 # endif
 
+# if defined(LINUX) && (defined(POWERPC) || defined(SPARC) || defined(ALPHA))
+#   define NEED_FIND_LIMIT
+# endif
+
 #ifdef NEED_FIND_LIMIT
 #   include <setjmp.h>
 #endif
@@ -79,11 +106,13 @@
 
 #ifdef IRIX5
 # include <sys/uio.h>
+# include <malloc.h>   /* for locking */
+#endif
+#ifdef USE_MMAP
 # include <sys/types.h>
 # include <sys/mman.h>
 # include <sys/stat.h>
 # include <fcntl.h>
-# include <malloc.h>	/* for locking */
 #endif
 
 #ifdef SUNOS5SIGS
@@ -98,7 +127,7 @@
 #ifdef DJGPP
   /* Apparently necessary for djgpp 2.01.  May casuse problems with	*/
   /* other versions.							*/
-  typedef caddr_t long unsigned int;
+  typedef long unsigned int caddr_t;
 #endif
 
 #ifdef PCR
@@ -107,17 +136,34 @@
 # include "mm/PCR_MM.h"
 #endif
 
-#if defined(PROT_EXEC) && !defined(NO_EXECUTE_PERMISSION)
+#if !defined(NO_EXECUTE_PERMISSION)
 # define OPT_PROT_EXEC PROT_EXEC
 #else
 # define OPT_PROT_EXEC 0
 #endif
 
+#if defined(LINUX) && (defined(POWERPC) || defined(SPARC) || defined(ALPHA))
+  /* The I386 case can be handled without a search.  The Alpha case	*/
+  /* used to be handled differently as well, but the rules changed	*/
+  /* for recent Linux versions.  This seems to be the easiest way to	*/
+  /* cover all versions.						*/
+  ptr_t GC_data_start;
+
+  extern char * GC_copyright[];  /* Any data symbol would do. */
+
+  void GC_init_linux_data_start()
+  {
+    extern ptr_t GC_find_limit();
+
+    GC_data_start = GC_find_limit((ptr_t)GC_copyright, FALSE);
+  }
+#endif
+
 # ifdef OS2
 
 # include <stddef.h>
 
-# ifndef __IBMC__ /* e.g. EMX */
+# if !defined(__IBMC__) && !defined(__WATCOMC__) /* e.g. EMX */
 
 struct exe_hdr {
     unsigned short      magic_number;
@@ -217,7 +263,7 @@ void GC_enable_signals(void)
 # else
 
 #  if !defined(PCR) && !defined(AMIGA) && !defined(MSWIN32) \
-      && !defined(MACOS) && !defined(DJGPP)
+      && !defined(MACOS) && !defined(DJGPP) && !defined(DOS4GW)
 
 #   if defined(sigmask) && !defined(UTS4)
 	/* Use the traditional BSD interface */
@@ -236,7 +282,7 @@ void GC_enable_signals(void)
 #	define SIGSETMASK(old, new) sigprocmask(SIG_SETMASK, &(new), &(old))
 #   endif
 
-static bool mask_initialized = FALSE;
+static GC_bool mask_initialized = FALSE;
 
 static SIGSET_T new_mask;
 
@@ -296,28 +342,47 @@ void GC_enable_signals()
 
 # endif /*!OS/2 */
 
-/*
- * Find the base of the stack.
+/* Ivan Demakov: simplest way (to me) */
+#ifdef DOS4GW
+  void GC_disable_signals() { }
+  void GC_enable_signals() { }
+#endif
+
+/* Find the page size */
+word GC_page_size;
+
+# ifdef MSWIN32
+  void GC_setpagesize()
+  {
+    SYSTEM_INFO sysinfo;
+    
+    GetSystemInfo(&sysinfo);
+    GC_page_size = sysinfo.dwPageSize;
+  }
+
+# else
+#   if defined(MPROTECT_VDB) || defined(PROC_VDB) || defined(USE_MMAP) \
+       || defined(USE_MUNMAP)
+	void GC_setpagesize()
+	{
+	    GC_page_size = GETPAGESIZE();
+	}
+#   else
+	/* It's acceptable to fake it. */
+	void GC_setpagesize()
+	{
+	    GC_page_size = HBLKSIZE;
+	}
+#   endif
+# endif
+
+/* 
+ * Find the base of the stack. 
  * Used only in single-threaded environment.
  * With threads, GC_mark_roots needs to know how to do this.
  * Called with allocator lock held.
  */
-# ifdef MSWIN32
-
-/* Get the page size.	*/
-word GC_page_size = 0;
-
-word GC_getpagesize()
-{
-    SYSTEM_INFO sysinfo;
-    
-    if (GC_page_size == 0) {
-        GetSystemInfo(&sysinfo);
-        GC_page_size = sysinfo.dwPageSize;
-    }
-    return(GC_page_size);
-}
-
+# ifdef MSWIN32 
 # define is_writable(prot) ((prot) == PAGE_READWRITE \
 			    || (prot) == PAGE_WRITECOPY \
 			    || (prot) == PAGE_EXECUTE_READWRITE \
@@ -347,7 +412,7 @@ ptr_t GC_get_stack_base()
 {
     int dummy;
     ptr_t sp = (ptr_t)(&dummy);
-    ptr_t trunc_sp = (ptr_t)((word)sp & ~(GC_getpagesize() - 1));
+    ptr_t trunc_sp = (ptr_t)((word)sp & ~(GC_page_size - 1));
     word size = GC_get_writable_length(trunc_sp, 0);
    
     return(trunc_sp + size);
@@ -374,6 +439,24 @@ ptr_t GC_get_stack_base()
 
 # ifdef AMIGA
 
+ptr_t GC_get_stack_base()
+{
+    struct Process *proc = (struct Process*)SysBase->ThisTask;
+ 
+    /* Reference: Amiga Guru Book Pages: 42,567,574 */
+    if (proc->pr_Task.tc_Node.ln_Type==NT_PROCESS
+        && proc->pr_CLI != NULL) {
+	/* first ULONG is StackSize */
+	/*longPtr = proc->pr_ReturnAddr;
+	size = longPtr[0];*/
+
+	return (char *)proc->pr_ReturnAddr + sizeof(ULONG);
+    } else {
+	return (char *)proc->pr_Task.tc_SPUpper;
+    }
+}
+
+#if 0 /* old version */
 ptr_t GC_get_stack_base()
 {
     extern struct WBStartup *_WBenchMsg;
@@ -398,10 +481,9 @@ ptr_t GC_get_stack_base()
     }
     return (ptr_t)(__base + GC_max(size, __stack));
 }
+#endif /* 0 */
 
-# else
-
-
+# else /* !AMIGA, !OS2, ... */
 
 # ifdef NEED_FIND_LIMIT
   /* Some tools to implement HEURISTIC2	*/
@@ -421,19 +503,22 @@ ptr_t GC_get_stack_base()
 	typedef void (*handler)();
 #   endif
 
-#   ifdef SUNOS5SIGS
-	static struct sigaction oldact;
+#   if defined(SUNOS5SIGS) || defined(IRIX5) || defined(OSF1)
+	static struct sigaction old_segv_act;
+#	if defined(_sigargs) /* !Irix6.x */
+	    static struct sigaction old_bus_act;
+#	endif
 #   else
         static handler old_segv_handler, old_bus_handler;
 #   endif
     
     void GC_setup_temporary_fault_handler()
     {
-#	ifdef SUNOS5SIGS
+#	if defined(SUNOS5SIGS) || defined(IRIX5) || defined(OSF1)
 	  struct sigaction	act;
 
 	  act.sa_handler	= GC_fault_handler;
-          act.sa_flags          = SA_RESTART | SA_SIGINFO | SA_NODEFER;
+          act.sa_flags          = SA_RESTART | SA_NODEFER;
           /* The presence of SA_NODEFER represents yet another gross    */
           /* hack.  Under Solaris 2.3, siglongjmp doesn't appear to     */
           /* interact correctly with -lthread.  We hide the confusion   */
@@ -441,7 +526,20 @@ ptr_t GC_get_stack_base()
           /* signal mask.                                               */
 
 	  (void) sigemptyset(&act.sa_mask);
-	  (void) sigaction(SIGSEGV, &act, &oldact);
+#	  ifdef IRIX_THREADS
+		/* Older versions have a bug related to retrieving and	*/
+		/* and setting a handler at the same time.		*/
+	        (void) sigaction(SIGSEGV, 0, &old_segv_act);
+	        (void) sigaction(SIGSEGV, &act, 0);
+#	  else
+	        (void) sigaction(SIGSEGV, &act, &old_segv_act);
+#		ifdef _sigargs	/* Irix 5.x, not 6.x */
+		    /* Under 5.x, we may get SIGBUS.			*/
+		    /* Pthreads doesn't exist under 5.x, so we don't	*/
+		    /* have to worry in the threads case.		*/
+		    (void) sigaction(SIGBUS, &act, &old_bus_act);
+#		endif
+#	  endif	/* IRIX_THREADS */
 #	else
     	  old_segv_handler = signal(SIGSEGV, GC_fault_handler);
 #	  ifdef SIGBUS
@@ -452,8 +550,11 @@ ptr_t GC_get_stack_base()
     
     void GC_reset_fault_handler()
     {
-#       ifdef SUNOS5SIGS
-	  (void) sigaction(SIGSEGV, &oldact, 0);
+#       if defined(SUNOS5SIGS) || defined(IRIX5) || defined(OSF1)
+	  (void) sigaction(SIGSEGV, &old_segv_act, 0);
+#	  ifdef _sigargs	/* Irix 5.x, not 6.x */
+	      (void) sigaction(SIGBUS, &old_bus_act, 0);
+#	  endif
 #       else
   	  (void) signal(SIGSEGV, old_segv_handler);
 #	  ifdef SIGBUS
@@ -466,7 +567,7 @@ ptr_t GC_get_stack_base()
     /* the smallest location q s.t. [q,p] is addressible (!up).	*/
     ptr_t GC_find_limit(p, up)
     ptr_t p;
-    bool up;
+    GC_bool up;
     {
         static VOLATILE ptr_t result;
     		/* Needs to be static, since otherwise it may not be	*/
@@ -537,6 +638,9 @@ ptr_t GC_get_stack_base()
 #	    endif
 
 #	endif /* HEURISTIC2 */
+#	ifdef STACK_GROWS_DOWN
+	    if (result == 0) result = (ptr_t)(signed_word)(-sizeof(ptr_t));
+#	endif
     	return(result);
 #   endif /* STACKBOTTOM */
 }
@@ -653,9 +757,9 @@ void GC_register_data_segments()
   /* all real work is done by GC_register_dynamic_libraries.  Under	*/
   /* win32s, we cannot find the data segments associated with dll's.	*/
   /* We rgister the main data segment here.				*/
-  bool GC_win32s = FALSE;	/* We're running under win32s.	*/
+  GC_bool GC_win32s = FALSE;	/* We're running under win32s.	*/
   
-  bool GC_is_win32s()
+  GC_bool GC_is_win32s()
   {
       DWORD v = GetVersion();
       
@@ -682,9 +786,9 @@ void GC_register_data_segments()
     
     GetSystemInfo(&sysinfo);
     limit = sysinfo.lpMinimumApplicationAddress;
-    p = (ptr_t)((word)start & ~(GC_getpagesize() - 1));
+    p = (ptr_t)((word)start & ~(GC_page_size - 1));
     for (;;) {
-    	q = (LPVOID)(p - GC_getpagesize());
+    	q = (LPVOID)(p - GC_page_size);
     	if ((ptr_t)q > (ptr_t)p /* underflow */ || q < limit) break;
     	result = VirtualQuery(q, &buf, sizeof(buf));
     	if (result != sizeof(buf) || buf.AllocationBase == 0) break;
@@ -695,7 +799,7 @@ void GC_register_data_segments()
   
   /* Is p the start of either the malloc heap, or of one of our */
   /* heap sections?						*/
-  bool GC_is_heap_base (ptr_t p)
+  GC_bool GC_is_heap_base (ptr_t p)
   {
      
      register unsigned i;
@@ -764,6 +868,72 @@ void GC_register_data_segments()
 # else
 # ifdef AMIGA
 
+   void GC_register_data_segments()
+   {
+     struct Process	*proc;
+     struct CommandLineInterface *cli;
+     BPTR myseglist;
+     ULONG *data;
+ 
+     int	num;
+
+
+#    ifdef __GNUC__
+        ULONG dataSegSize;
+        GC_bool found_segment = FALSE;
+	extern char __data_size[];
+
+	dataSegSize=__data_size+8;
+	/* Can`t find the Location of __data_size, because
+           it`s possible that is it, inside the segment. */
+
+#     endif
+
+	proc= (struct Process*)SysBase->ThisTask;
+
+	/* Reference: Amiga Guru Book Pages: 538ff,565,573
+		     and XOper.asm */
+	if (proc->pr_Task.tc_Node.ln_Type==NT_PROCESS) {
+	  if (proc->pr_CLI == NULL) {
+	    myseglist = proc->pr_SegList;
+	  } else {
+	    /* ProcLoaded	'Loaded as a command: '*/
+	    cli = BADDR(proc->pr_CLI);
+	    myseglist = cli->cli_Module;
+	  }
+	} else {
+	  ABORT("Not a Process.");
+ 	}
+
+	if (myseglist == NULL) {
+	    ABORT("Arrrgh.. can't find segments, aborting");
+ 	}
+
+	/* xoper hunks Shell Process */
+
+	num=0;
+        for (data = (ULONG *)BADDR(myseglist); data != NULL;
+             data = (ULONG *)BADDR(data[0])) {
+	  if (((ULONG) GC_register_data_segments < (ULONG) &data[1]) ||
+	      ((ULONG) GC_register_data_segments > (ULONG) &data[1] + data[-1])) {
+#             ifdef __GNUC__
+		if (dataSegSize == data[-1]) {
+		  found_segment = TRUE;
+		}
+# 	      endif
+	      GC_add_roots_inner((char *)&data[1],
+				 ((char *)&data[1]) + data[-1], FALSE);
+          }
+          ++num;
+        } /* for */
+# 	ifdef __GNUC__
+	   if (!found_segment) {
+	     ABORT("Can`t find correct Segments.\nSolution: Use an newer version of ixemul.library");
+	   }
+# 	endif
+  }
+
+#if 0 /* old version */
   void GC_register_data_segments()
   {
     extern struct WBStartup *_WBenchMsg;
@@ -805,6 +975,7 @@ void GC_register_data_segments()
          }
     }
   }
+#endif /* old version */
 
 
 # else
@@ -828,7 +999,9 @@ int * etext_addr;
     if (setjmp(GC_jmp_buf) == 0) {
     	/* Try writing to the address.	*/
     	*result = *result;
+        GC_reset_fault_handler();
     } else {
+        GC_reset_fault_handler();
     	/* We got here via a longjmp.  The address is not readable.	*/
     	/* This is known to happen under Solaris 2.4 + gcc, which place	*/
     	/* string constants in the text segment, but after etext.	*/
@@ -836,7 +1009,6 @@ int * etext_addr;
     	/* text and data segments, so plan A bought us something.	*/
     	result = (char *)GC_find_limit((ptr_t)(DATAEND) - MIN_PAGE_SIZE, FALSE);
     }
-    GC_reset_fault_handler();
     return((char *)result);
 }
 # endif
@@ -844,7 +1016,8 @@ int * etext_addr;
 
 void GC_register_data_segments()
 {
-#   if !defined(PCR) && !defined(SRC_M3) && !defined(NEXT) && !defined(MACOS)
+#   if !defined(PCR) && !defined(SRC_M3) && !defined(NEXT) && !defined(MACOS) \
+       && !defined(MACOSX)
 #     if defined(REDIRECT_MALLOC) && defined(SOLARIS_THREADS)
 	/* As of Solaris 2.3, the Solaris threads implementation	*/
 	/* allocates the data structure for the initial thread with	*/
@@ -858,7 +1031,7 @@ void GC_register_data_segments()
 	GC_add_roots_inner(DATASTART, (char *)(DATAEND), FALSE);
 #     endif
 #   endif
-#   if !defined(PCR) && defined(NEXT)
+#   if !defined(PCR) && (defined(NEXT) || defined(MACOSX))
       GC_add_roots_inner(DATASTART, (char *) get_end(), FALSE);
 #   endif
 #   if defined(MACOS)
@@ -872,9 +1045,19 @@ void GC_register_data_segments()
 #     if defined(__MWERKS__)
 #       if !__POWERPC__
 	  extern void* GC_MacGetDataStart(void);
+	  /* MATTHEW: Function to handle Far Globals (CW Pro 3) */
+#         if __option(far_data)
+	  extern void* GC_MacGetDataEnd(void);
+#         endif
 	  /* globals begin above stack and end at a5. */
 	  GC_add_roots_inner((ptr_t)GC_MacGetDataStart(),
           		     (ptr_t)LMGetCurrentA5(), FALSE);
+	  /* MATTHEW: Handle Far Globals */          		     
+#         if __option(far_data)
+      /* Far globals follow he QD globals: */
+	  GC_add_roots_inner((ptr_t)LMGetCurrentA5(),
+          		     (ptr_t)GC_MacGetDataEnd(), FALSE);
+#         endif
 #       else
 	  extern char __data_start__[], __data_end__[];
 	  GC_add_roots_inner((ptr_t)&__data_start__,
@@ -898,9 +1081,11 @@ void GC_register_data_segments()
  */
  
 # if !defined(OS2) && !defined(PCR) && !defined(AMIGA) \
-	&& !defined(MSWIN32) && !defined(MACOS)
+	&& !defined(MSWIN32) && !defined(MACOS) && !defined(DOS4GW)
 
-extern caddr_t sbrk();
+# ifdef SUNOS4
+    extern caddr_t sbrk();
+# endif
 # ifdef __STDC__
 #   define SBRK_ARG_T ptrdiff_t
 # else
@@ -914,80 +1099,87 @@ extern caddr_t sbrk();
 ptr_t GC_unix_get_mem(bytes)
 word bytes;
 {
-    caddr_t cur_brk = sbrk(0);
+    caddr_t cur_brk = (caddr_t)sbrk(0);
     caddr_t result;
-    SBRK_ARG_T lsbs = (word)cur_brk & (HBLKSIZE-1);
+    SBRK_ARG_T lsbs = (word)cur_brk & (GC_page_size-1);
     static caddr_t my_brk_val = 0;
     
     if ((SBRK_ARG_T)bytes < 0) return(0); /* too big */
     if (lsbs != 0) {
-        if(sbrk(HBLKSIZE - lsbs) == (caddr_t)(-1)) return(0);
+        if((caddr_t)(sbrk(GC_page_size - lsbs)) == (caddr_t)(-1)) return(0);
     }
     if (cur_brk == my_brk_val) {
     	/* Use the extra block we allocated last time. */
         result = (ptr_t)sbrk((SBRK_ARG_T)bytes);
         if (result == (caddr_t)(-1)) return(0);
-        result -= HBLKSIZE;
+        result -= GC_page_size;
     } else {
-        result = (ptr_t)sbrk(HBLKSIZE + (SBRK_ARG_T)bytes);
+        result = (ptr_t)sbrk(GC_page_size + (SBRK_ARG_T)bytes);
         if (result == (caddr_t)(-1)) return(0);
     }
-    my_brk_val = result + bytes + HBLKSIZE;	/* Always HBLKSIZE aligned */
+    my_brk_val = result + bytes + GC_page_size;	/* Always page aligned */
     return((ptr_t)result);
 }
 
 #else  /* Not RS6000 */
 
 #if defined(USE_MMAP)
-/* Tested only under IRIX5 */
+/* Tested only under IRIX5 and Solaris 2 */
+
+#ifdef USE_MMAP_FIXED
+#   define GC_MMAP_FLAGS MAP_FIXED | MAP_PRIVATE
+	/* Seems to yield better performance on Solaris 2, but can	*/
+	/* be unreliable if something is already mapped at the address.	*/
+#else
+#   define GC_MMAP_FLAGS MAP_PRIVATE
+#endif
 
 ptr_t GC_unix_get_mem(bytes)
 word bytes;
 {
-    static bool initialized = FALSE;
+    static GC_bool initialized = FALSE;
     static int fd;
     void *result;
     static ptr_t last_addr = HEAP_START;
-    static size_t page_size;
 
     if (!initialized) {
 	fd = open("/dev/zero", O_RDONLY);
-	page_size = getpagesize();
 	initialized = TRUE;
     }
+    if (bytes & (GC_page_size -1)) ABORT("Bad GET_MEM arg");
     result = mmap(last_addr, bytes, PROT_READ | PROT_WRITE | OPT_PROT_EXEC,
-		  MAP_PRIVATE | MAP_FIXED, fd, 0/* offset */);
+		  GC_MMAP_FLAGS, fd, 0/* offset */);
     if (result == MAP_FAILED) return(0);
-    last_addr = (ptr_t)result + bytes + page_size - 1;
-    last_addr = (ptr_t)((word)last_addr & ~(page_size - 1));
+    last_addr = (ptr_t)result + bytes + GC_page_size - 1;
+    last_addr = (ptr_t)((word)last_addr & ~(GC_page_size - 1));
     return((ptr_t)result);
 }
 
-#else /* Not RS6000, USE_MMAP */
+#else /* Not RS6000, not USE_MMAP */
 ptr_t GC_unix_get_mem(bytes)
 word bytes;
 {
-  caddr_t result;
+  ptr_t result;
 # ifdef IRIX5
     /* Bare sbrk isn't thread safe.  Play by malloc rules.	*/
     /* The equivalent may be needed on other systems as well. 	*/
     __LOCK_MALLOC();
 # endif
   {
-    caddr_t cur_brk = sbrk(0);
-    SBRK_ARG_T lsbs = (word)cur_brk & (HBLKSIZE-1);
+    ptr_t cur_brk = (ptr_t)sbrk(0);
+    SBRK_ARG_T lsbs = (word)cur_brk & (GC_page_size-1);
     
     if ((SBRK_ARG_T)bytes < 0) return(0); /* too big */
     if (lsbs != 0) {
-        if(sbrk(HBLKSIZE - lsbs) == (caddr_t)(-1)) return(0);
+        if((ptr_t)sbrk(GC_page_size - lsbs) == (ptr_t)(-1)) return(0);
     }
-    result = sbrk((SBRK_ARG_T)bytes);
-    if (result == (caddr_t)(-1)) result = 0;
+    result = (ptr_t)sbrk((SBRK_ARG_T)bytes);
+    if (result == (ptr_t)(-1)) result = 0;
   }
 # ifdef IRIX5
     __UNLOCK_MALLOC();
 # endif
-  return((ptr_t)result);
+  return(result);
 }
 
 #endif /* Not USE_MMAP */
@@ -1040,8 +1232,108 @@ word bytes;
     return(result);			  
 }
 
+void GC_win32_free_heap ()
+{
+    if (GC_win32s) {
+ 	while (GC_n_heap_bases > 0) {
+ 	    GlobalFree (GC_heap_bases[--GC_n_heap_bases]);
+ 	    GC_heap_bases[GC_n_heap_bases] = 0;
+ 	}
+    }
+}
+
+
 # endif
 
+#ifdef USE_MUNMAP
+
+/* For now, this only works on some Unix-like systems.  If you 	*/
+/* have something else, don't define USE_MUNMAP.		*/
+/* We assume ANSI C to support this feature.			*/
+#include <unistd.h>
+#include <sys/mman.h>
+#include <sys/stat.h>
+#include <sys/types.h>
+#include <fcntl.h>
+
+/* Compute a page aligned starting address for the unmap 	*/
+/* operation on a block of size bytes starting at start.	*/
+/* Return 0 if the block is too small to make this feasible.	*/
+ptr_t GC_unmap_start(ptr_t start, word bytes)
+{
+    ptr_t result = start;
+    /* Round start to next page boundary.       */
+        result += GC_page_size - 1;
+        result = (ptr_t)((word)result & ~(GC_page_size - 1));
+    if (result + GC_page_size > start + bytes) return 0;
+    return result;
+}
+
+/* Compute end address for an unmap operation on the indicated	*/
+/* block.							*/
+ptr_t GC_unmap_end(ptr_t start, word bytes)
+{
+    ptr_t end_addr = start + bytes;
+    end_addr = (ptr_t)((word)end_addr & ~(GC_page_size - 1));
+    return end_addr;
+}
+
+/* We assume that GC_remap is called on exactly the same range	*/
+/* as a previous call to GC_unmap.  It is safe to consistently	*/
+/* round the endpoints in both places.				*/
+void GC_unmap(ptr_t start, word bytes)
+{
+    ptr_t start_addr = GC_unmap_start(start, bytes);
+    ptr_t end_addr = GC_unmap_end(start, bytes);
+    word len = end_addr - start_addr;
+    if (0 == start_addr) return;
+    if (munmap(start_addr, len) != 0) ABORT("munmap failed");
+    GC_unmapped_bytes += len;
+}
+
+
+void GC_remap(ptr_t start, word bytes)
+{
+    static int zero_descr = -1;
+    ptr_t start_addr = GC_unmap_start(start, bytes);
+    ptr_t end_addr = GC_unmap_end(start, bytes);
+    word len = end_addr - start_addr;
+    ptr_t result;
+
+    if (-1 == zero_descr) zero_descr = open("/dev/zero", O_RDWR);
+    if (0 == start_addr) return;
+    result = mmap(start_addr, len, PROT_READ | PROT_WRITE | OPT_PROT_EXEC,
+	          MAP_FIXED | MAP_PRIVATE, zero_descr, 0);
+    if (result != start_addr) {
+	ABORT("mmap remapping failed");
+    }
+    GC_unmapped_bytes -= len;
+}
+
+/* Two adjacent blocks have already been unmapped and are about to	*/
+/* be merged.  Unmap the whole block.  This typically requires		*/
+/* that we unmap a small section in the middle that was not previously	*/
+/* unmapped due to alignment constraints.				*/
+void GC_unmap_gap(ptr_t start1, word bytes1, ptr_t start2, word bytes2)
+{
+    ptr_t start1_addr = GC_unmap_start(start1, bytes1);
+    ptr_t end1_addr = GC_unmap_end(start1, bytes1);
+    ptr_t start2_addr = GC_unmap_start(start2, bytes2);
+    ptr_t end2_addr = GC_unmap_end(start2, bytes2);
+    ptr_t start_addr = end1_addr;
+    ptr_t end_addr = start2_addr;
+    word len;
+    GC_ASSERT(start1 + bytes1 == start2);
+    if (0 == start1_addr) start_addr = GC_unmap_start(start1, bytes1 + bytes2);
+    if (0 == start2_addr) end_addr = GC_unmap_end(start1, bytes1 + bytes2);
+    if (0 == start_addr) return;
+    len = end_addr - start_addr;
+    if (len != 0 && munmap(start_addr, len) != 0) ABORT("munmap failed");
+    GC_unmapped_bytes += len;
+}
+
+#endif /* USE_MUNMAP */
+
 /* Routine for pushing any additional roots.  In THREADS 	*/
 /* environment, this is also responsible for marking from 	*/
 /* thread stacks.  In the SRC_M3 case, it also handles		*/
@@ -1157,7 +1449,9 @@ void GC_default_push_other_roots()
 
 # endif /* SRC_M3 */
 
-# if defined(SOLARIS_THREADS) || defined(WIN32_THREADS) || defined(IRIX_THREADS)
+# if defined(SOLARIS_THREADS) || defined(WIN32_THREADS) \
+     || defined(IRIX_THREADS) || defined(LINUX_THREADS) \
+     || defined(IRIX_PCR_THREADS)
 
 extern void GC_push_all_stacks();
 
@@ -1193,7 +1487,7 @@ void (*GC_push_other_roots)() = GC_default_push_other_roots;
  *		or write only to the stack.
  */
  
-bool GC_dirty_maintained = FALSE;
+GC_bool GC_dirty_maintained = FALSE;
 
 # ifdef DEFAULT_VDB
 
@@ -1219,7 +1513,7 @@ void GC_read_dirty()
 /* of the pages overlapping h are dirty.  This routine may err on the	*/
 /* side of labelling pages as dirty (and this implementation does).	*/
 /*ARGSUSED*/
-bool GC_page_was_dirty(h)
+GC_bool GC_page_was_dirty(h)
 struct hblk *h;
 {
     return(TRUE);
@@ -1234,7 +1528,7 @@ struct hblk *h;
  
 /* Could any valid GC heap pointer ever have been written to this page?	*/
 /*ARGSUSED*/
-bool GC_page_was_ever_dirty(h)
+GC_bool GC_page_was_ever_dirty(h)
 struct hblk *h;
 {
     return(TRUE);
@@ -1314,30 +1608,6 @@ struct hblk *h;
 	  
 # endif
 
-VOLATILE page_hash_table GC_dirty_pages;
-				/* Pages dirtied since last GC_read_dirty. */
-
-word GC_page_size;
-
-bool GC_just_outside_heap(addr)
-word addr;
-{
-    register unsigned i;
-    register word start;
-    register word end;
-    word mask = GC_page_size-1;
-    
-    for (i = 0; i < GC_n_heap_sects; i++) {
-    	start = (word) GC_heap_sects[i].hs_start;
-    	end = start + (word)GC_heap_sects[i].hs_bytes;
-        if (addr < ((start + mask) & ~mask) && addr >= (start & ~mask)
-            || addr >= (end & ~mask) && addr < ((end + mask) & ~mask)) {
-            return(TRUE);
-        }
-    }
-    return(FALSE);
-}
-
 #if defined(SUNOS4) || defined(FREEBSD)
     typedef void (* SIG_PF)();
 #endif
@@ -1357,7 +1627,30 @@ word addr;
     typedef void (* REAL_SIG_PF)(int, struct siginfo *, void *);
 #endif
 #if defined(LINUX)
-    typedef void (* REAL_SIG_PF)(int, struct sigcontext_struct);
+#   include <linux/version.h>
+#   if (LINUX_VERSION_CODE >= 0x20100) && !defined(M68K) || defined(ALPHA)
+      typedef struct sigcontext s_c;
+#   else
+      typedef struct sigcontext_struct s_c;
+#   endif
+#   if defined(ALPHA) || defined(M68K)
+      typedef void (* REAL_SIG_PF)(int, int, s_c *);
+#   else
+      typedef void (* REAL_SIG_PF)(int, s_c);
+#   endif
+#   ifdef ALPHA
+    /* Retrieve fault address from sigcontext structure by decoding	*/
+    /* instruction.							*/
+    char * get_fault_addr(s_c *sc) {
+        unsigned instr;
+	word faultaddr;
+
+	instr = *((unsigned *)(sc->sc_pc));
+	faultaddr = sc->sc_regs[(instr >> 16) & 0x1f];
+	faultaddr += (word) (((int)instr << 16) >> 16);
+	return (char *)faultaddr;
+    }
+#   endif /* !ALPHA */
 # endif
 
 SIG_PF GC_old_bus_handler;
@@ -1392,7 +1685,11 @@ SIG_PF GC_old_segv_handler;	/* Also old MSWIN32 ACCESS_VIOLATION filter */
 #   endif
 # endif
 # if defined(LINUX)
-    void GC_write_fault_handler(int sig, struct sigcontext_struct sc)
+#   if defined(ALPHA) || defined(M68K)
+      void GC_write_fault_handler(int sig, int code, s_c * sc)
+#   else
+      void GC_write_fault_handler(int sig, s_c sc)
+#   endif
 #   define SIG_OK (sig == SIGSEGV)
 #   define CODE_OK TRUE
 	/* Empirically c.trapno == 14, but is that useful?      */
@@ -1426,7 +1723,37 @@ SIG_PF GC_old_segv_handler;	/* Also old MSWIN32 ACCESS_VIOLATION filter */
 #     ifdef I386
 	char * addr = (char *) (sc.cr2);
 #     else
-        char * addr = /* As of 1.3.90 there seemed to be no way to do this. */;
+#	if defined(M68K)
+          char * addr = NULL;
+
+	  struct sigcontext *scp = (struct sigcontext *)(&sc);
+
+	  int format = (scp->sc_formatvec >> 12) & 0xf;
+	  unsigned long *framedata = (unsigned long *)(scp + 1); 
+	  unsigned long ea;
+
+	  if (format == 0xa || format == 0xb) {
+	  	/* 68020/030 */
+	  	ea = framedata[2];
+	  } else if (format == 7) {
+	  	/* 68040 */
+	  	ea = framedata[3];
+	  } else if (format == 4) {
+	  	/* 68060 */
+	  	ea = framedata[0];
+	  	if (framedata[1] & 0x08000000) {
+	  		/* correct addr on misaligned access */
+	  		ea = (ea+4095)&(~4095);
+	  	}
+	  }	
+	  addr = (char *)ea;
+#	else
+#	  ifdef ALPHA
+            char * addr = get_fault_addr(sc);
+#	  else
+		--> architecture not supported
+#	  endif
+#	endif
 #     endif
 #   endif
 #   if defined(MSWIN32)
@@ -1438,7 +1765,7 @@ SIG_PF GC_old_segv_handler;	/* Also old MSWIN32 ACCESS_VIOLATION filter */
     if (SIG_OK && CODE_OK) {
         register struct hblk * h =
         		(struct hblk *)((word)addr & ~(GC_page_size-1));
-        bool in_allocd_block;
+        GC_bool in_allocd_block;
         
 #	ifdef SUNOS5SIGS
 	    /* Address is only within the correct physical page.	*/
@@ -1451,7 +1778,8 @@ SIG_PF GC_old_segv_handler;	/* Also old MSWIN32 ACCESS_VIOLATION filter */
 #	else
 	    in_allocd_block = (HDR(addr) != 0);
 #	endif
-        if (!in_allocd_block && !GC_just_outside_heap((word)addr)) {
+        if (!in_allocd_block) {
+	    /* Heap blocks now begin and end on page boundaries */
             SIG_PF old_handler;
             
             if (sig == SIGSEGV) {
@@ -1461,6 +1789,7 @@ SIG_PF GC_old_segv_handler;	/* Also old MSWIN32 ACCESS_VIOLATION filter */
             }
             if (old_handler == SIG_DFL) {
 #		ifndef MSWIN32
+		    GC_err_printf1("Segfault at 0x%lx\n", addr);
                     ABORT("Unexpected bus error or segmentation fault");
 #		else
 		    return(EXCEPTION_CONTINUE_SEARCH);
@@ -1475,7 +1804,11 @@ SIG_PF GC_old_segv_handler;	/* Also old MSWIN32 ACCESS_VIOLATION filter */
 		    return;
 #		endif
 #		if defined (LINUX)
-		    (*(REAL_SIG_PF)old_handler) (sig, sc);
+#		    if defined(ALPHA) || defined(M68K)
+		        (*(REAL_SIG_PF)old_handler) (sig, code, sc);
+#		    else 
+		        (*(REAL_SIG_PF)old_handler) (sig, sc);
+#		    endif
 		    return;
 #		endif
 #		if defined (IRIX5) || defined(OSF1)
@@ -1493,7 +1826,7 @@ SIG_PF GC_old_segv_handler;	/* Also old MSWIN32 ACCESS_VIOLATION filter */
             set_pht_entry_from_index(GC_dirty_pages, index);
         }
         UNPROTECT(h, GC_page_size);
-#	if defined(IRIX5) || defined(OSF1) || defined(LINUX)
+#	if defined(OSF1) || defined(LINUX)
 	    /* These reset the signal handler each time by default. */
 	    signal(SIGSEGV, (SIG_PF) GC_write_fault_handler);
 #	endif
@@ -1508,6 +1841,7 @@ SIG_PF GC_old_segv_handler;	/* Also old MSWIN32 ACCESS_VIOLATION filter */
 #ifdef MSWIN32
     return EXCEPTION_CONTINUE_SEARCH;
 #else
+    GC_err_printf1("Segfault at 0x%lx\n", addr);
     ABORT("Unexpected bus error or segmentation fault");
 #endif
 }
@@ -1521,7 +1855,7 @@ struct hblk *h;
 {
     register struct hblk * h_trunc;
     register unsigned i;
-    register bool found_clean;
+    register GC_bool found_clean;
     
     if (!GC_dirty_maintained) return;
     h_trunc = (struct hblk *)((word)h & ~(GC_page_size-1));
@@ -1539,34 +1873,23 @@ struct hblk *h;
     }
 }
 
-#if defined(SVR4)
-#include <unistd.h>
-int
-GC_getpagesize()
-{
-    return sysconf(_SC_PAGESIZE);
-}
-#else
-# ifdef MSWIN32
-   /* GC_getpagesize() defined above */
-# else
-#  define GC_getpagesize() getpagesize()
-# endif
-#endif
-				 
 void GC_dirty_init()
 {
-#if defined(SUNOS5SIGS)
+#if defined(SUNOS5SIGS) || defined(IRIX5) /* || defined(OSF1) */
     struct sigaction	act, oldact;
-    act.sa_sigaction	= GC_write_fault_handler;
-    act.sa_flags	= SA_RESTART | SA_SIGINFO;
+#   ifdef IRIX5
+    	act.sa_flags	= SA_RESTART;
+        act.sa_handler  = GC_write_fault_handler;
+#   else
+    	act.sa_flags	= SA_RESTART | SA_SIGINFO;
+        act.sa_sigaction = GC_write_fault_handler;
+#   endif
     (void)sigemptyset(&act.sa_mask); 
 #endif
 #   ifdef PRINTSTATS
 	GC_printf0("Inititalizing mprotect virtual dirty bit implementation\n");
 #   endif
     GC_dirty_maintained = TRUE;
-    GC_page_size = GC_getpagesize();
     if (GC_page_size % HBLKSIZE != 0) {
         GC_err_printf0("Page size not multiple of HBLKSIZE\n");
         ABORT("Page size not multiple of HBLKSIZE");
@@ -1583,7 +1906,7 @@ void GC_dirty_init()
 #	endif
       }
 #   endif
-#   if defined(IRIX5) || defined(OSF1) || defined(SUNOS4) || defined(LINUX)
+#   if defined(OSF1) || defined(SUNOS4) || defined(LINUX)
       GC_old_segv_handler = signal(SIGSEGV, (SIG_PF)GC_write_fault_handler);
       if (GC_old_segv_handler == SIG_IGN) {
         GC_err_printf0("Previously ignored segmentation violation!?");
@@ -1595,13 +1918,24 @@ void GC_dirty_init()
 #	endif
       }
 #   endif
-#   if defined(SUNOS5SIGS)
-      sigaction(SIGSEGV, &act, &oldact);
-      if (oldact.sa_flags & SA_SIGINFO) {
+#   if defined(SUNOS5SIGS) || defined(IRIX5)
+#     if defined(IRIX_THREADS) || defined(IRIX_PCR_THREADS)
+      	sigaction(SIGSEGV, 0, &oldact);
+      	sigaction(SIGSEGV, &act, 0);
+#     else
+      	sigaction(SIGSEGV, &act, &oldact);
+#     endif
+#     if defined(_sigargs)
+	/* This is Irix 5.x, not 6.x.  Irix 5.x does not have	*/
+	/* sa_sigaction.					*/
+	GC_old_segv_handler = oldact.sa_handler;
+#     else /* Irix 6.x or SUNOS5SIGS */
+        if (oldact.sa_flags & SA_SIGINFO) {
           GC_old_segv_handler = (SIG_PF)(oldact.sa_sigaction);
-      } else {
+        } else {
           GC_old_segv_handler = oldact.sa_handler;
-      }
+        }
+#     endif
       if (GC_old_segv_handler == SIG_IGN) {
 	     GC_err_printf0("Previously ignored segmentation violation!?");
 	     GC_old_segv_handler = SIG_DFL;
@@ -1628,18 +1962,13 @@ void GC_dirty_init()
 
 void GC_protect_heap()
 {
-    word ps = GC_page_size;
-    word pmask = (ps-1);
     ptr_t start;
-    word offset;
     word len;
     unsigned i;
     
     for (i = 0; i < GC_n_heap_sects; i++) {
-        offset = (word)(GC_heap_sects[i].hs_start) & pmask;
-        start = GC_heap_sects[i].hs_start - offset;
-        len = GC_heap_sects[i].hs_bytes + offset;
-        len += ps-1; len &= ~pmask;
+        start = GC_heap_sects[i].hs_start;
+        len = GC_heap_sects[i].hs_bytes;
         PROTECT(start, len);
     }
 }
@@ -1654,7 +1983,7 @@ void GC_read_dirty()
     GC_protect_heap();
 }
 
-bool GC_page_was_dirty(h)
+GC_bool GC_page_was_dirty(h)
 struct hblk * h;
 {
     register word index = PHT_HASH(h);
@@ -1712,27 +2041,28 @@ word len;
 /* Replacement for UNIX system call.	 */
 /* Other calls that write to the heap	 */
 /* should be handled similarly.		 */
-# ifndef LINT
-  int read(fd, buf, nbyte)
+# if defined(__STDC__) && !defined(SUNOS4)
+#   include <unistd.h>
+    ssize_t read(int fd, void *buf, size_t nbyte)
 # else
-  int GC_read(fd, buf, nbyte)
+#   ifndef LINT
+      int read(fd, buf, nbyte)
+#   else
+      int GC_read(fd, buf, nbyte)
+#   endif
+    int fd;
+    char *buf;
+    int nbyte;
 # endif
-int fd;
-char *buf;
-int nbyte;
 {
     int result;
     
     GC_begin_syscall();
     GC_unprotect_range(buf, (word)nbyte);
 #   ifdef IRIX5
-	/* Indirect system call exists, but is undocumented, and	*/
-	/* always seems to return EINVAL.  There seems to be no		*/
-	/* general way to wrap system calls, since the system call	*/
-	/* convention appears to require an immediate argument for	*/
-	/* the system call number, and building the required code	*/
-	/* in the data segment also seems dangerous.  We can fake it	*/
-	/* for read; anything else is up to the client.			*/
+	/* Indirect system call may not always be easily available.	*/
+	/* We could call _read, but that would interfere with the	*/
+	/* libpthread interception of read.				*/
 	{
 	    struct iovec iov;
 
@@ -1750,7 +2080,7 @@ int nbyte;
 #endif
 
 /*ARGSUSED*/
-bool GC_page_was_ever_dirty(h)
+GC_bool GC_page_was_ever_dirty(h)
 struct hblk *h;
 {
     return(TRUE);
@@ -1779,6 +2109,7 @@ word n;
  * address space), but it avoids intercepting system calls.
  */
 
+#include <errno.h>
 #include <sys/types.h>
 #include <sys/signal.h>
 #include <sys/fault.h>
@@ -1791,8 +2122,6 @@ word n;
 word GC_proc_buf_size = INITIAL_BUF_SZ;
 char *GC_proc_buf;
 
-page_hash_table GC_written_pages = { 0 };	/* Pages ever dirtied	*/
-
 #ifdef SOLARIS_THREADS
 /* We don't have exact sp values for threads.  So we count on	*/
 /* occasionally declaring stack pages to be fresh.  Thus we 	*/
@@ -1843,6 +2172,7 @@ void GC_dirty_init()
     	ABORT("/proc open failed");
     }
     GC_proc_fd = syscall(SYS_ioctl, fd, PIOCOPENPD, 0);
+    close(fd);
     if (GC_proc_fd < 0) {
     	ABORT("/proc ioctl failed");
     }
@@ -1903,6 +2233,7 @@ int dummy;
                 WARN("Insufficient space for /proc read\n", 0);
                 /* Punt:	*/
         	memset(GC_grungy_pages, 0xff, sizeof (page_hash_table));
+		memset(GC_written_pages, 0xff, sizeof(page_hash_table));
 #		ifdef SOLARIS_THREADS
 		    BZERO(GC_fresh_pages,
 		    	  MAX_FRESH_PAGES * sizeof (struct hblk *)); 
@@ -1960,11 +2291,11 @@ int dummy;
 
 #undef READ
 
-bool GC_page_was_dirty(h)
+GC_bool GC_page_was_dirty(h)
 struct hblk *h;
 {
     register word index = PHT_HASH(h);
-    register bool result;
+    register GC_bool result;
     
     result = get_pht_entry_from_index(GC_grungy_pages, index);
 #   ifdef SOLARIS_THREADS
@@ -1978,11 +2309,11 @@ struct hblk *h;
     return(result);
 }
 
-bool GC_page_was_ever_dirty(h)
+GC_bool GC_page_was_ever_dirty(h)
 struct hblk *h;
 {
     register word index = PHT_HASH(h);
-    register bool result;
+    register GC_bool result;
     
     result = get_pht_entry_from_index(GC_written_pages, index);
 #   ifdef SOLARIS_THREADS
@@ -2058,7 +2389,7 @@ void GC_read_dirty()
     }
 }
 
-bool GC_page_was_dirty(h)
+GC_bool GC_page_was_dirty(h)
 struct hblk *h;
 {
     if((ptr_t)h < GC_vd_base || (ptr_t)h >= GC_vd_base + NPAGES*HBLKSIZE) {
@@ -2081,22 +2412,36 @@ struct hblk *h;
  * Call stack save code for debugging.
  * Should probably be in mach_dep.c, but that requires reorganization.
  */
-#if defined(SPARC)
+#if defined(SPARC) && !defined(LINUX)
 #   if defined(SUNOS4)
 #     include <machine/frame.h>
 #   else
 #     if defined (DRSNX)
 #	include <sys/sparc/frame.h>
 #     else
-#       include <sys/frame.h>
+#        if defined(OPENBSD)
+#          include <frame.h>
+#        else
+#          include <sys/frame.h>
+#        endif
 #     endif
 #   endif
 #   if NARGS > 6
 	--> We only know how to to get the first 6 arguments
 #   endif
 
+#ifdef SAVE_CALL_CHAIN
 /* Fill in the pc and argument information for up to NFRAMES of my	*/
 /* callers.  Ignore my frame and my callers frame.			*/
+
+#ifdef OPENBSD
+#  define FR_SAVFP fr_fp
+#  define FR_SAVPC fr_pc
+#else
+#  define FR_SAVFP fr_savfp
+#  define FR_SAVPC fr_savpc
+#endif
+
 void GC_save_callers (info) 
 struct callinfo info[NFRAMES];
 {
@@ -2107,11 +2452,11 @@ struct callinfo info[NFRAMES];
 
   frame = (struct frame *) GC_save_regs_in_stack ();
   
-  for (fp = frame -> fr_savfp; fp != 0 && nframes < NFRAMES;
-       fp = fp -> fr_savfp, nframes++) {
+  for (fp = frame -> FR_SAVFP; fp != 0 && nframes < NFRAMES;
+       fp = fp -> FR_SAVFP, nframes++) {
       register int i;
       
-      info[nframes].ci_pc = fp->fr_savpc;
+      info[nframes].ci_pc = fp->FR_SAVPC;
       for (i = 0; i < NARGS; i++) {
 	info[nframes].ci_arg[i] = ~(fp->fr_arg[i]);
       }
@@ -2119,28 +2464,8 @@ struct callinfo info[NFRAMES];
   if (nframes < NFRAMES) info[nframes].ci_pc = 0;
 }
 
+#endif /* SAVE_CALL_CHAIN */
 #endif /* SPARC */
 
-#ifdef SAVE_CALL_CHAIN
-
-void GC_print_callers (info)
-struct callinfo info[NFRAMES];
-{
-    register int i,j;
-    
-    GC_err_printf0("\tCall chain at allocation:\n");
-    for (i = 0; i < NFRAMES; i++) {
-     	if (info[i].ci_pc == 0) break;
-     	GC_err_printf0("\t\targs: ");
-     	for (j = 0; j < NARGS; j++) {
-     	    if (j != 0) GC_err_printf0(", ");
-     	    GC_err_printf2("%d (0x%X)", ~(info[i].ci_arg[j]),
-     	    				~(info[i].ci_arg[j]));
-     	}
-     	GC_err_printf1("\n\t\t##PC##= 0x%X\n", info[i].ci_pc);
-    }
-}
-
-#endif /* SAVE_CALL_CHAIN */
 
 
