@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1996-1997 The University of Melbourne.
+% Copyright (C) 1996-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -14,8 +14,8 @@
 
 :- interface.
 
-:- import_module hlds_data, prog_data, llds, tree.
-:- import_module list, std_util, io.
+:- import_module hlds_data, prog_data, tree, builtin_ops.
+:- import_module char, list, std_util, io.
 
 :- type byte_tree	==	tree(list(byte_code)).
 
@@ -53,8 +53,8 @@
 					list(pair(byte_var, byte_dir)))
 			;	complex_deconstruct(byte_var, byte_cons_id,
 					list(pair(byte_var, byte_dir)))
-			;	place_arg(reg_type, int, byte_var)
-			;	pickup_arg(reg_type, int, byte_var)
+			;	place_arg(byte_reg_type, int, byte_var)
+			;	pickup_arg(byte_reg_type, int, byte_var)
 			;	call(byte_module_id, byte_pred_id,
 					arity, byte_proc_id)
 			;	higher_order_call(byte_var, arity, arity,
@@ -71,6 +71,11 @@
 			;	not_supported
 			.
 
+	% Currently we only support integer registers.
+	% This might one day be extended to support separate
+	% floating-point registers.
+:- type byte_reg_type	--->	r.	% general-purpose (integer) register.
+
 :- type byte_cons_id	--->	cons(byte_module_id, string,
 					arity, byte_cons_tag)
 			;	int_const(int)
@@ -80,17 +85,19 @@
 					arity, byte_proc_id)
 			;	code_addr_const(byte_module_id, byte_pred_id,
 					arity, byte_proc_id)
-			;	base_type_info_const(byte_module_id, string,
+			;	type_ctor_info_const(byte_module_id, string,
 					int)
+			;	base_typeclass_info_const(byte_module_id,
+					class_id, string)
 			;	char_const(char)
 			.
 
 :- type byte_var_info	--->	var_info(string, type).
 
 :- type byte_cons_tag	--->	no_tag
-			;	simple_tag(tag_bits)
-			;	complicated_tag(tag_bits, int)
-			;	complicated_constant_tag(tag_bits, int)
+			;	unshared_tag(tag_bits)
+			;	shared_remote_tag(tag_bits, int)
+			;	shared_local_tag(tag_bits, int)
 			;	enum_tag(int)
 			.
 
@@ -104,7 +111,7 @@
 			;	to_none
 			.
 
-:- type byte_module_id	==	string.
+:- type byte_module_id	==	module_name.
 :- type byte_pred_id	==	string.
 :- type byte_proc_id	==	int.
 :- type byte_label_id	==	int.
@@ -122,8 +129,8 @@
 
 :- implementation.
 
-:- import_module hlds_pred, llds_out.
-:- import_module char, library, int, string, require.
+:- import_module bytecode_data, hlds_pred, prog_out, llds_out.
+:- import_module library, int, string, require.
 
 :- pred bytecode__version(int::out) is det.
 
@@ -485,21 +492,17 @@ debug_determinism(Detism) -->
 
 %---------------------------------------------------------------------------%
 
-:- pred output_reg(reg_type, int, io__state, io__state).
+:- pred output_reg(byte_reg_type, int, io__state, io__state).
 :- mode output_reg(in, in, di, uo) is det.
 
 output_reg(r, N) -->
 	output_byte(N).
-output_reg(f, _) -->
-	{ error("we do not handle floating point registers yet") }.
 
-:- pred debug_reg(reg_type, int, io__state, io__state).
+:- pred debug_reg(byte_reg_type, int, io__state, io__state).
 :- mode debug_reg(in, in, di, uo) is det.
 
 debug_reg(r, N) -->
 	debug_int(N).
-debug_reg(f, _) -->
-	{ error("we do not handle floating point registers yet") }.
 
 %---------------------------------------------------------------------------%
 
@@ -633,13 +636,14 @@ debug_var_dirs([Var - Dir | VarDirs]) -->
 :- mode output_module_id(in, di, uo) is det.
 
 output_module_id(ModuleId) -->
-	output_string(ModuleId).
+	{ prog_out__sym_name_to_string(ModuleId, ModuleIdString) },
+	output_string(ModuleIdString).
 
 :- pred debug_module_id(byte_module_id, io__state, io__state).
 :- mode debug_module_id(in, di, uo) is det.
 
 debug_module_id(ModuleId) -->
-	debug_string(ModuleId).
+	debug_sym_name(ModuleId).
 
 %---------------------------------------------------------------------------%
 
@@ -661,15 +665,13 @@ debug_pred_id(PredId) -->
 :- mode output_proc_id(in, di, uo) is det.
 
 output_proc_id(ProcId) -->
-	{ ModeId is ProcId mod 10000 },
-	output_byte(ModeId).
+	output_byte(ProcId).
 
 :- pred debug_proc_id(byte_proc_id, io__state, io__state).
 :- mode debug_proc_id(in, di, uo) is det.
 
 debug_proc_id(ProcId) -->
-	{ ModeId is ProcId mod 10000 },
-	debug_int(ModeId).
+	debug_int(ProcId).
 
 %---------------------------------------------------------------------------%
 
@@ -692,7 +694,7 @@ debug_label_id(LabelId) -->
 
 output_cons_id(cons(ModuleId, Functor, Arity, Tag)) -->
 	output_byte(0),
-	output_string(ModuleId),
+	output_module_id(ModuleId),
 	output_string(Functor),
 	output_short(Arity),
 	output_tag(Tag).
@@ -717,7 +719,7 @@ output_cons_id(code_addr_const(ModuleId, PredId, Arity, ProcId)) -->
 	output_pred_id(PredId),
 	output_length(Arity),
 	output_proc_id(ProcId).
-output_cons_id(base_type_info_const(ModuleId, TypeName, TypeArity)) -->
+output_cons_id(type_ctor_info_const(ModuleId, TypeName, TypeArity)) -->
 	output_byte(6),
 	output_module_id(ModuleId),
 	output_string(TypeName),
@@ -727,12 +729,17 @@ output_cons_id(char_const(Char)) -->
 	{ char__to_int(Char, Byte) },
 	output_byte(Byte).
 
+	% XXX
+output_cons_id(base_typeclass_info_const(_, _, _)) -->
+	{ error("Sorry, bytecode for typeclass not yet implemented") },
+	output_byte(8).
+
 :- pred debug_cons_id(byte_cons_id, io__state, io__state).
 :- mode debug_cons_id(in, di, uo) is det.
 
 debug_cons_id(cons(ModuleId, Functor, Arity, Tag)) -->
 	debug_string("functor"),
-	debug_string(ModuleId),
+	debug_sym_name(ModuleId),
 	debug_string(Functor),
 	debug_int(Arity),
 	debug_tag(Tag).
@@ -757,11 +764,20 @@ debug_cons_id(code_addr_const(ModuleId, PredId, Arity, ProcId)) -->
 	debug_pred_id(PredId),
 	debug_length(Arity),
 	debug_proc_id(ProcId).
-debug_cons_id(base_type_info_const(ModuleId, TypeName, TypeArity)) -->
-	debug_string("base_type_info_const"),
+debug_cons_id(type_ctor_info_const(ModuleId, TypeName, TypeArity)) -->
+	debug_string("type_ctor_info_const"),
 	debug_module_id(ModuleId),
 	debug_string(TypeName),
 	debug_int(TypeArity).
+debug_cons_id(base_typeclass_info_const(ModuleId, 
+		class_id(ClassName, ClassArity), Instance)) -->
+	debug_string("base_typeclass_info_const"),
+	debug_module_id(ModuleId),
+	debug_string("class_id"),
+	debug_sym_name(ClassName),
+	debug_string("/"),
+	debug_int(ClassArity),
+	debug_string(Instance).
 debug_cons_id(char_const(Char)) -->
 	debug_string("char_const"),
 	{ string__from_char_list([Char], String) },
@@ -772,14 +788,14 @@ debug_cons_id(char_const(Char)) -->
 :- pred output_tag(byte_cons_tag, io__state, io__state).
 :- mode output_tag(in, di, uo) is det.
 
-output_tag(simple_tag(Primary)) -->
+output_tag(unshared_tag(Primary)) -->
 	output_byte(0),
 	output_byte(Primary).
-output_tag(complicated_tag(Primary, Secondary)) -->
+output_tag(shared_remote_tag(Primary, Secondary)) -->
 	output_byte(1),
 	output_byte(Primary),
 	output_int(Secondary).
-output_tag(complicated_constant_tag(Primary, Secondary)) -->
+output_tag(shared_local_tag(Primary, Secondary)) -->
 	output_byte(2),
 	output_byte(Primary),
 	output_int(Secondary).
@@ -792,15 +808,15 @@ output_tag(no_tag) -->
 :- pred debug_tag(byte_cons_tag, io__state, io__state).
 :- mode debug_tag(in, di, uo) is det.
 
-debug_tag(simple_tag(Primary)) -->
-	debug_string("simple_tag"),
+debug_tag(unshared_tag(Primary)) -->
+	debug_string("unshared_tag"),
 	debug_int(Primary).
-debug_tag(complicated_tag(Primary, Secondary)) -->
-	debug_string("complicated_tag"),
+debug_tag(shared_remote_tag(Primary, Secondary)) -->
+	debug_string("shared_remote_tag"),
 	debug_int(Primary),
 	debug_int(Secondary).
-debug_tag(complicated_constant_tag(Primary, Secondary)) -->
-	debug_string("complicated_constant_tag"),
+debug_tag(shared_local_tag(Primary, Secondary)) -->
+	debug_string("shared_local_tag"),
 	debug_int(Primary),
 	debug_int(Secondary).
 debug_tag(enum_tag(Enum)) -->
@@ -1064,14 +1080,6 @@ unop_debug((not),		"not").
 
 %---------------------------------------------------------------------------%
 
-:- pred output_string(string, io__state, io__state).
-:- mode output_string(in, di, uo) is det.
-
-output_string(Val) -->
-	io__write_bytes(Val),
-	io__write_byte(0).
-
-
 /*
 **	debug_cstring prints a string quoted in the manner of C.
 */
@@ -1086,165 +1094,6 @@ debug_cstring(Str) -->
 	% the string as a bytecode argument. This is not very elegant.
 	io__write_char('"'),
 	io__write_char(' ').
-	
-
-:- pred output_byte(int, io__state, io__state).
-:- mode output_byte(in, di, uo) is det.
-
-output_byte(Val) -->
-	( { Val < 256 } ->
-		io__write_byte(Val)
-	;
-		{ error("byte does not fit in eight bits") }
-	).
-
-/*
-** Spit out a `short' in a portable format.
-** This format is: big-endian, 16-bit, 2's-complement.
-**
-** NOTE: We -assume- the machine architecture uses 2's-complement.
-*/
-:- pred output_short(int, io__state, io__state).
-:- mode output_short(in, di, uo) is det.
-
-output_short(Val) -->
-	{ Val1 is Val >> 8 },
-	{ Val2 is Val mod 256 },
-	( { Val1 < 256 } ->
-		io__write_byte(Val1),
-		io__write_byte(Val2)
-	;
-		{ error("small integer does not fit in sixteen bits") }
-	).
-
-/*
-** Spit out an `int' in a portable `highest common denominator' format.
-** This format is: big-endian, 64-bit, 2's-complement int.
-**
-** NOTE: We -assume- the machine architecture uses 2's-complement.
-*/
-:- pred output_int(int, io__state, io__state).
-:- mode output_int(in, di, uo) is det.
-
-output_int(IntVal) -->
-	{ int__bits_per_int(IntBits) },
-	( { IntBits > bytecode_int_bits } ->
-		{ error("size of int is larger than size of bytecode integer.")}
-	;
-		{ ZeroPadBytes is (bytecode_int_bits - IntBits) // 
-			bits_per_byte },
-		output_padding_zeros(ZeroPadBytes),
-		{ FirstByteToDump is bytecode_int_bytes - ZeroPadBytes - 1 },
-		output_int_bytes(FirstByteToDump, IntVal)
-	).
-
-:- func bytecode_int_bits = int.
-:- mode bytecode_int_bits = out is det.
-
-bytecode_int_bits = bits_per_byte * bytecode_int_bytes.
-
-:- func bytecode_int_bytes = int.
-:- mode bytecode_int_bytes = out is det.
-
-bytecode_int_bytes = 8.
-
-:- func bits_per_byte = int.
-:- mode bits_per_byte = out is det.
-
-bits_per_byte = 8.
-
-:- pred output_padding_zeros(int, io__state, io__state).
-:- mode output_padding_zeros(in, di, uo) is det.
-
-output_padding_zeros(NumBytes) -->
-	( { NumBytes > 0 } ->
-		io__write_byte(0),
-		{ NumBytes1 is NumBytes - 1 },
-		output_padding_zeros(NumBytes1)
-	;
-		{ true }
-	).
-
-:- pred output_int_bytes(int, int, io__state, io__state).
-:- mode output_int_bytes(in, in, di, uo) is det.
-
-output_int_bytes(ByteNum, IntVal) -->
-	( { ByteNum >= 0 } ->
-		{ BitShifts is ByteNum * bits_per_byte },
-		{ Byte is (IntVal >> BitShifts) mod (1 << bits_per_byte) },
-		{ ByteNum1 is ByteNum - 1 },
-		io__write_byte(Byte),
-		output_int_bytes(ByteNum1, IntVal)
-	;
-		{ true }
-	).
-
-/*
-** Spit out a `float' in a portable `highest common denominator format.
-** This format is: big-endian, 64-bit, IEEE-754 floating point value.
-**
-** NOTE: We -assume- the machine architecture uses IEEE-754.
-*/
-:- pred output_float(float, io__state, io__state).
-:- mode output_float(in, di, uo) is det.
-
-output_float(Val) -->
-	{ float_to_float64_bytes(Val, B0, B1, B2, B3, B4, B5, B6, B7) },
-	output_byte(B0),
-	output_byte(B1),
-	output_byte(B2),
-	output_byte(B3),
-	output_byte(B4),
-	output_byte(B5),
-	output_byte(B6),
-	output_byte(B7).
-
-/*
-** Convert a `float' to the representation used in the bytecode.
-** That is, a sequence of eight bytes.
-*/
-:- pred float_to_float64_bytes(float::in, 
-		int::out, int::out, int::out, int::out, 
-		int::out, int::out, int::out, int::out) is det.
-:- pragma(c_code, 
-	float_to_float64_bytes(FloatVal::in, B0::out, B1::out, B2::out, B3::out,
-		B4::out, B5::out, B6::out, B7::out),
-	"
-
-	{
-		Float64		float64;
-		unsigned char	*raw_mem_p;
-
-		float64 = (Float64) FloatVal;
-		raw_mem_p = (unsigned char*) &float64;
-
-		#if defined(MR_BIG_ENDIAN)
-			B0 = raw_mem_p[0];
-			B1 = raw_mem_p[1];
-			B2 = raw_mem_p[2];
-			B3 = raw_mem_p[3];
-			B4 = raw_mem_p[4];
-			B5 = raw_mem_p[5];
-			B6 = raw_mem_p[6];
-			B7 = raw_mem_p[7];
-		#elif defined(MR_LITTLE_ENDIAN)
-			B7 = raw_mem_p[0];
-			B6 = raw_mem_p[1];
-			B5 = raw_mem_p[2];
-			B4 = raw_mem_p[3];
-			B3 = raw_mem_p[4];
-			B2 = raw_mem_p[5];
-			B1 = raw_mem_p[6];
-			B0 = raw_mem_p[7];
-		#else
-			#error	Weird-endian architecture
-		#endif
-	}
-	
-	"
-).
-
-%---------------------------------------------------------------------------%
 
 :- pred debug_string(string, io__state, io__state).
 :- mode debug_string(in, di, uo) is det.
@@ -1267,4 +1116,16 @@ debug_float(Val) -->
 	io__write_float(Val),
 	io__write_char(' ').
 
+:- pred debug_sym_name(sym_name, io__state, io__state).
+:- mode debug_sym_name(in, di, uo) is det.
+
+debug_sym_name(unqualified(Val)) -->
+	io__write_string(Val),
+	io__write_char(' ').
+debug_sym_name(qualified(Module, Val)) -->
+	debug_sym_name(Module),
+	io__write_char(':'),
+	io__write_string(Val),
+	io__write_char(' ').
+
 %---------------------------------------------------------------------------%
