@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1996-1997 The University of Melbourne.
+% Copyright (C) 1996-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -26,7 +26,7 @@
 
 :- import_module hlds_pred, hlds_goal, hlds_data, prog_data, llds, arg_info.
 :- import_module passes_aux, call_gen, mode_util, code_util, goal_util.
-:- import_module globals, tree.
+:- import_module globals, tree, varset, term.
 
 :- import_module bool, int, string, list, assoc_list, set, map, varset.
 :- import_module std_util, require, term.
@@ -89,7 +89,7 @@ bytecode_gen__proc(ProcId, PredInfo, ModuleInfo, Code) :-
 
 	proc_info_goal(ProcInfo, Goal),
 	proc_info_vartypes(ProcInfo, VarTypes),
-	proc_info_variables(ProcInfo, VarSet),
+	proc_info_varset(ProcInfo, VarSet),
 	proc_info_interface_determinism(ProcInfo, Detism),
 	determinism_to_code_model(Detism, CodeModel),
 
@@ -147,7 +147,8 @@ bytecode_gen__proc(ProcId, PredInfo, ModuleInfo, Code) :-
 bytecode_gen__goal(GoalExpr - GoalInfo, ByteInfo0, ByteInfo, Code) :-
 	bytecode_gen__goal_expr(GoalExpr, GoalInfo, ByteInfo0, ByteInfo,
 		GoalCode),
-	goal_info_get_context(GoalInfo, term__context(_, Line)),
+	goal_info_get_context(GoalInfo, Context),
+	term__context_line(Context, Line),
 	Code = tree(node([context(Line)]), GoalCode).
 
 :- pred bytecode_gen__goal_expr(hlds_goal_expr::in, hlds_goal_info::in,
@@ -155,11 +156,20 @@ bytecode_gen__goal(GoalExpr - GoalInfo, ByteInfo0, ByteInfo, Code) :-
 
 bytecode_gen__goal_expr(GoalExpr, GoalInfo, ByteInfo0, ByteInfo, Code) :-
 	(
-		GoalExpr = higher_order_call(PredVar, ArgVars, ArgTypes,
-			ArgModes, Detism, _IsPredOrFunc),
-		bytecode_gen__higher_order_call(PredVar, ArgVars,
-			ArgTypes, ArgModes, Detism, ByteInfo0, Code),
-		ByteInfo = ByteInfo0
+		GoalExpr = generic_call(GenericCallType, 
+			ArgVars, ArgModes, Detism),
+		( GenericCallType = higher_order(PredVar, _, _) ->
+			bytecode_gen__higher_order_call(PredVar, ArgVars,
+				ArgModes, Detism, ByteInfo0, Code),
+			ByteInfo = ByteInfo0
+		;
+			% XXX
+			functor(GenericCallType, GenericCallFunctor, _),
+			string__append_list([
+				"sorry: bytecode not yet implemented for ",
+				GenericCallFunctor, " calls"], Msg),
+			error(Msg)
+		)
 	;
 		GoalExpr = call(PredId, ProcId, ArgVars, BuiltinState, _, _),
 		( BuiltinState = not_builtin ->
@@ -183,7 +193,7 @@ bytecode_gen__goal_expr(GoalExpr, GoalInfo, ByteInfo0, ByteInfo, Code) :-
 		EndofCode = node([endof_negation, label(EndLabel)]),
 		Code = tree(EnterCode, tree(SomeCode, EndofCode))
 	;
-		GoalExpr = some(_, Goal),
+		GoalExpr = some(_, _, Goal),
 		bytecode_gen__goal(Goal, ByteInfo0, ByteInfo1, SomeCode),
 		bytecode_gen__get_next_temp(ByteInfo1, Temp, ByteInfo),
 		EnterCode = node([enter_commit(Temp)]),
@@ -192,6 +202,9 @@ bytecode_gen__goal_expr(GoalExpr, GoalInfo, ByteInfo0, ByteInfo, Code) :-
 	;
 		GoalExpr = conj(GoalList),
 		bytecode_gen__conj(GoalList, ByteInfo0, ByteInfo, Code)
+	;
+		GoalExpr = par_conj(_GoalList, _SM),
+		error("sorry, bytecode_gen of parallel conj not implemented")
 	;
 		GoalExpr = disj(GoalList, _),
 		( GoalList = [] ->
@@ -236,15 +249,19 @@ bytecode_gen__goal_expr(GoalExpr, GoalInfo, ByteInfo0, ByteInfo, Code) :-
 			tree(ElseCode,
 			     EndofIfCode))))))
 	;
-		GoalExpr = pragma_c_code(_, _, _, _, _, _, _, _),
+		GoalExpr = pragma_c_code(_, _, _, _, _, _, _),
 		Code = node([not_supported]),
 		ByteInfo = ByteInfo0
+	;
+		GoalExpr = bi_implication(_, _),
+		% these should have been expanded out by now
+		error("bytecode_gen__goal_expr: unexpected bi_implication")
 	).
 
 %---------------------------------------------------------------------------%
 
-:- pred bytecode_gen__gen_places(list(pair(var, arg_loc))::in, byte_info::in,
-	byte_tree::out) is det.
+:- pred bytecode_gen__gen_places(list(pair(prog_var, arg_loc))::in,
+		byte_info::in, byte_tree::out) is det.
 
 bytecode_gen__gen_places([], _, empty).
 bytecode_gen__gen_places([Var - Loc | OutputArgs], ByteInfo, Code) :-
@@ -252,8 +269,8 @@ bytecode_gen__gen_places([Var - Loc | OutputArgs], ByteInfo, Code) :-
 	bytecode_gen__map_var(ByteInfo, Var, ByteVar),
 	Code = tree(node([place_arg(r, Loc, ByteVar)]), OtherCode).
 
-:- pred bytecode_gen__gen_pickups(list(pair(var, arg_loc))::in, byte_info::in,
-	byte_tree::out) is det.
+:- pred bytecode_gen__gen_pickups(list(pair(prog_var, arg_loc))::in,
+		byte_info::in, byte_tree::out) is det.
 
 bytecode_gen__gen_pickups([], _, empty).
 bytecode_gen__gen_pickups([Var - Loc | OutputArgs], ByteInfo, Code) :-
@@ -265,17 +282,15 @@ bytecode_gen__gen_pickups([Var - Loc | OutputArgs], ByteInfo, Code) :-
 
 	% Generate bytecode for a higher order call.
 
-:- pred bytecode_gen__higher_order_call(var::in, list(var)::in, list(type)::in,
+:- pred bytecode_gen__higher_order_call(prog_var::in, list(prog_var)::in,
 	list(mode)::in, determinism::in, byte_info::in, byte_tree::out) is det.
 
-bytecode_gen__higher_order_call(PredVar, ArgVars, ArgTypes, ArgModes, Detism,
+bytecode_gen__higher_order_call(PredVar, ArgVars, ArgModes, Detism,
 		ByteInfo, Code) :-
 	determinism_to_code_model(Detism, CodeModel),
 	bytecode_gen__get_module_info(ByteInfo, ModuleInfo),
-	module_info_globals(ModuleInfo, Globals),
-	globals__get_args_method(Globals, ArgsMethod),
-	make_arg_infos(ArgsMethod, ArgTypes, ArgModes, CodeModel, ModuleInfo,
-		ArgInfo),
+	list__map(bytecode_gen__get_var_type(ByteInfo), ArgVars, ArgTypes),
+	make_arg_infos(ArgTypes, ArgModes, CodeModel, ModuleInfo, ArgInfo),
 	assoc_list__from_corresponding_lists(ArgVars, ArgInfo, ArgVarsInfos),
 
 	call_gen__partition_args(ArgVarsInfos, InVars, OutVars),
@@ -300,7 +315,7 @@ bytecode_gen__higher_order_call(PredVar, ArgVars, ArgTypes, ArgModes, Detism,
 
 	% Generate bytecode for an ordinary call.
 
-:- pred bytecode_gen__call(pred_id::in, proc_id::in, list(var)::in,
+:- pred bytecode_gen__call(pred_id::in, proc_id::in, list(prog_var)::in,
 	determinism::in, byte_info::in, byte_tree::out) is det.
 
 bytecode_gen__call(PredId, ProcId, ArgVars, Detism, ByteInfo, Code) :-
@@ -328,7 +343,7 @@ bytecode_gen__call(PredId, ProcId, ArgVars, Detism, ByteInfo, Code) :-
 
 	% Generate bytecode for a call to a builtin.
 
-:- pred bytecode_gen__builtin(pred_id::in, proc_id::in, list(var)::in,
+:- pred bytecode_gen__builtin(pred_id::in, proc_id::in, list(prog_var)::in,
 	byte_info::in, byte_tree::out) is det.
 
 bytecode_gen__builtin(PredId, ProcId, Args, ByteInfo, Code) :-
@@ -370,7 +385,7 @@ bytecode_gen__map_test(ByteInfo, Rval, Code) :-
 		error("builtin test is not in a recognized form")
 	).
 
-:- pred bytecode_gen__map_assign(byte_info::in, var::in, rval::in,
+:- pred bytecode_gen__map_assign(byte_info::in, prog_var::in, rval::in,
 	byte_tree::out) is det.
 
 bytecode_gen__map_assign(ByteInfo, Var, Rval, Code) :-
@@ -409,11 +424,11 @@ bytecode_gen__map_arg(ByteInfo, Rval, ByteArg) :-
 
 	% Generate bytecode for a unification.
 
-:- pred bytecode_gen__unify(unification::in, var::in, unify_rhs::in,
+:- pred bytecode_gen__unify(unification::in, prog_var::in, unify_rhs::in,
 	byte_info::in, byte_tree::out) is det.
 
-bytecode_gen__unify(construct(Var, ConsId, Args, UniModes), _, _, ByteInfo,
-		Code) :-
+bytecode_gen__unify(construct(Var, ConsId, Args, UniModes, _, _, _),
+		_, _, ByteInfo, Code) :-
 	bytecode_gen__map_var(ByteInfo, Var, ByteVar),
 	bytecode_gen__map_vars(ByteInfo, Args, ByteArgs),
 	bytecode_gen__map_cons_id(ByteInfo, Var, ConsId, ByteConsId),
@@ -459,10 +474,10 @@ bytecode_gen__unify(simple_test(Var1, Var2), _, _, ByteInfo, Code) :-
 	bytecode_gen__map_var(ByteInfo, Var1, ByteVar1),
 	bytecode_gen__map_var(ByteInfo, Var2, ByteVar2),
 	Code = node([test(ByteVar1, ByteVar2)]).
-bytecode_gen__unify(complicated_unify(_, _), _Var, _RHS, _ByteInfo, _Code) :-
+bytecode_gen__unify(complicated_unify(_,_,_), _Var, _RHS, _ByteInfo, _Code) :-
 	error("complicated unifications should have been handled by polymorphism.m").
 
-:- pred bytecode_gen__map_uni_modes(list(uni_mode)::in, list(var)::in,
+:- pred bytecode_gen__map_uni_modes(list(uni_mode)::in, list(prog_var)::in,
 	byte_info::in, list(byte_dir)::out) is det.
 
 bytecode_gen__map_uni_modes([], [], _, []).
@@ -551,7 +566,7 @@ bytecode_gen__disj([Disjunct | Disjuncts], ByteInfo0, EndLabel,
 
 	% Generate bytecode for each arm of a switch.
 
-:- pred bytecode_gen__switch(list(case)::in, var::in, byte_info::in,
+:- pred bytecode_gen__switch(list(case)::in, prog_var::in, byte_info::in,
 	int::in, byte_info::out, byte_tree::out) is det.
 
 bytecode_gen__switch([], _, ByteInfo, _, ByteInfo, empty).
@@ -568,7 +583,7 @@ bytecode_gen__switch([case(ConsId, Goal) | Cases], Var, ByteInfo0, EndLabel,
 
 %---------------------------------------------------------------------------%
 
-:- pred bytecode_gen__map_cons_id(byte_info::in, var::in, cons_id::in,
+:- pred bytecode_gen__map_cons_id(byte_info::in, prog_var::in, cons_id::in,
 	byte_cons_id::out) is det.
 
 bytecode_gen__map_cons_id(ByteInfo, Var, ConsId, ByteConsId) :-
@@ -604,10 +619,18 @@ bytecode_gen__map_cons_id(ByteInfo, Var, ConsId, ByteConsId) :-
 		ConsId = float_const(FloatVal),
 		ByteConsId = float_const(FloatVal)
 	;
-		ConsId = pred_const(PredId, ProcId),
-		predicate_id(ModuleInfo, PredId, ModuleName, PredName, Arity),
-		proc_id_to_int(ProcId, ProcInt),
-		ByteConsId = pred_const(ModuleName, PredName, Arity, ProcInt)
+		ConsId = pred_const(PredId, ProcId, EvalMethod),
+		( EvalMethod = normal ->
+			predicate_id(ModuleInfo, PredId,
+				ModuleName, PredName, Arity),
+			proc_id_to_int(ProcId, ProcInt),
+			ByteConsId = pred_const(ModuleName,
+				PredName, Arity, ProcInt)
+		;
+			% XXX
+			error(
+	"sorry: bytecode not yet implemented for Aditi lambda expressions")
+		)
 	;
 		ConsId = code_addr_const(PredId, ProcId),
 		predicate_id(ModuleInfo, PredId, ModuleName, PredName, Arity),
@@ -615,36 +638,48 @@ bytecode_gen__map_cons_id(ByteInfo, Var, ConsId, ByteConsId) :-
 		ByteConsId = code_addr_const(ModuleName, PredName, Arity,
 			ProcInt)
 	;
-		ConsId = base_type_info_const(ModuleName, TypeName, TypeArity),
-		ByteConsId = base_type_info_const(ModuleName, TypeName,
+		ConsId = type_ctor_info_const(ModuleName, TypeName, TypeArity),
+		ByteConsId = type_ctor_info_const(ModuleName, TypeName,
 			TypeArity)
+	;
+		ConsId = base_typeclass_info_const(ModuleName, ClassId,
+			_, Instance),
+		ByteConsId = base_typeclass_info_const(ModuleName, ClassId,
+			Instance)
+	;
+		ConsId = tabling_pointer_const(_, _),
+		error("bytecode cannot implement tabling")
 	).
 
 :- pred bytecode_gen__map_cons_tag(cons_tag::in, byte_cons_tag::out) is det.
 
 bytecode_gen__map_cons_tag(no_tag, no_tag).
-bytecode_gen__map_cons_tag(simple_tag(Primary), simple_tag(Primary)).
-bytecode_gen__map_cons_tag(complicated_tag(Primary, Secondary),
-	complicated_tag(Primary, Secondary)).
-bytecode_gen__map_cons_tag(complicated_constant_tag(Primary, Secondary),
-	complicated_constant_tag(Primary, Secondary)).
+bytecode_gen__map_cons_tag(unshared_tag(Primary), unshared_tag(Primary)).
+bytecode_gen__map_cons_tag(shared_remote_tag(Primary, Secondary),
+	shared_remote_tag(Primary, Secondary)).
+bytecode_gen__map_cons_tag(shared_local_tag(Primary, Secondary),
+	shared_local_tag(Primary, Secondary)).
 bytecode_gen__map_cons_tag(string_constant(_), _) :-
 	error("string_constant cons tag for non-string_constant cons id").
 bytecode_gen__map_cons_tag(int_constant(IntVal), enum_tag(IntVal)).
 bytecode_gen__map_cons_tag(float_constant(_), _) :-
 	error("float_constant cons tag for non-float_constant cons id").
-bytecode_gen__map_cons_tag(pred_closure_tag(_, _), _) :-
+bytecode_gen__map_cons_tag(pred_closure_tag(_, _, _), _) :-
 	error("pred_closure_tag cons tag for non-pred_const cons id").
 bytecode_gen__map_cons_tag(code_addr_constant(_, _), _) :-
 	error("code_addr_constant cons tag for non-address_const cons id").
-bytecode_gen__map_cons_tag(base_type_info_constant(_, _, _), _) :-
-	error("base_type_info_constant cons tag for non-base_type_info_constant cons id").
+bytecode_gen__map_cons_tag(type_ctor_info_constant(_, _, _), _) :-
+	error("type_ctor_info_constant cons tag for non-type_ctor_info_constant cons id").
+bytecode_gen__map_cons_tag(base_typeclass_info_constant(_, _, _), _) :-
+	error("base_typeclass_info_constant cons tag for non-base_typeclass_info_constant cons id").
+bytecode_gen__map_cons_tag(tabling_pointer_constant(_, _), _) :-
+	error("tabling_pointer_constant cons tag for non-tabling_pointer_constant cons id").
 
 %---------------------------------------------------------------------------%
 
-:- pred bytecode_gen__create_varmap(list(var)::in, varset::in,
-	map(var, type)::in, int::in, map(var, byte_var)::in,
-	map(var, byte_var)::out, list(byte_var_info)::out) is det.
+:- pred bytecode_gen__create_varmap(list(prog_var)::in, prog_varset::in,
+	map(prog_var, type)::in, int::in, map(prog_var, byte_var)::in,
+	map(prog_var, byte_var)::out, list(byte_var_info)::out) is det.
 
 bytecode_gen__create_varmap([], _, _, _, VarMap, VarMap, []).
 bytecode_gen__create_varmap([Var | VarList], VarSet, VarTypes, N0,
@@ -661,16 +696,16 @@ bytecode_gen__create_varmap([Var | VarList], VarSet, VarTypes, N0,
 
 :- type byte_info
 	--->	byte_info(
-			map(var, byte_var),
-			map(var, type),
+			map(prog_var, byte_var),
+			map(prog_var, type),
 			module_info,
 			int,		% next label number to use
 			int		% next temp number to use
 		).
 
-:- pred bytecode_gen__init_byte_info(module_info::in, map(var, byte_var)::in,
-	map(var, type)::in, byte_info::out) is det.
-
+:- pred bytecode_gen__init_byte_info(module_info::in,
+	map(prog_var, byte_var)::in, map(prog_var, type)::in,
+	byte_info::out) is det.  
 bytecode_gen__init_byte_info(ModuleInfo, VarMap, VarTypes, ByteInfo) :-
 	ByteInfo = byte_info(VarMap, VarTypes, ModuleInfo, 0, 0).
 
@@ -679,25 +714,27 @@ bytecode_gen__init_byte_info(ModuleInfo, VarMap, VarTypes, ByteInfo) :-
 bytecode_gen__get_module_info(byte_info(_, _, ModuleInfo, _, _), ModuleInfo).
 
 :- pred bytecode_gen__map_vars(byte_info::in,
-	list(var)::in, list(byte_var)::out) is det.
+	list(prog_var)::in, list(byte_var)::out) is det.
 
 bytecode_gen__map_vars(byte_info(VarMap, _, _, _, _), Vars, ByteVars) :-
 	bytecode_gen__map_vars_2(VarMap, Vars, ByteVars).
 
-:- pred bytecode_gen__map_vars_2(map(var, byte_var)::in,
-	list(var)::in, list(byte_var)::out) is det.
+:- pred bytecode_gen__map_vars_2(map(prog_var, byte_var)::in,
+	list(prog_var)::in, list(byte_var)::out) is det.
 
 bytecode_gen__map_vars_2(_VarMap, [], []).
 bytecode_gen__map_vars_2(VarMap, [Var | Vars], [ByteVar | ByteVars]) :-
 	map__lookup(VarMap, Var, ByteVar),
 	bytecode_gen__map_vars_2(VarMap, Vars, ByteVars).
 
-:- pred bytecode_gen__map_var(byte_info::in, var::in, byte_var::out) is det.
+:- pred bytecode_gen__map_var(byte_info::in, prog_var::in,
+	byte_var::out) is det.
 
 bytecode_gen__map_var(byte_info(VarMap, _, _, _, _), Var, ByteVar) :-
 	map__lookup(VarMap, Var, ByteVar).
 
-:- pred bytecode_gen__get_var_type(byte_info::in, var::in, (type)::out) is det.
+:- pred bytecode_gen__get_var_type(byte_info::in, prog_var::in,
+	(type)::out) is det.
 
 bytecode_gen__get_var_type(byte_info(_, VarTypes, _, _, _), Var, Type) :-
 	map__lookup(VarTypes, Var, Type).
