@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -15,8 +15,8 @@
 
 :- interface.
 
-:- import_module code_info, hlds_goal.
-:- import_module varset.
+:- import_module code_info, hlds_module, hlds_goal, prog_data.
+:- import_module bool.
 
 	% code_aux__contains_only_builtins(G) is true if G is a leaf procedure,
 	% i.e. control does not leave G to call another procedure, even if
@@ -45,15 +45,15 @@
 :- pred code_aux__contains_simple_recursive_call(hlds_goal, code_info, bool).
 :- mode code_aux__contains_simple_recursive_call(in, in, out) is semidet.
 
-:- pred code_aux__explain_stack_slots(stack_slots, varset, string).
+:- pred code_aux__explain_stack_slots(stack_slots, prog_varset, string).
 :- mode code_aux__explain_stack_slots(in, in, out) is det.
 
 %---------------------------------------------------------------------------%
 
 :- implementation.
 
-:- import_module hlds_module, llds, llds_out, type_util.
-:- import_module bool, string, set, term, std_util, assoc_list, require.
+:- import_module hlds_pred, llds, llds_out, varset, type_util, term_util.
+:- import_module string, set, std_util, assoc_list, require.
 :- import_module list, map.
 
 code_aux__contains_only_builtins(Goal - _GoalInfo) :-
@@ -70,7 +70,7 @@ code_aux__contains_only_builtins_2(switch(_Var, _Category, Cases, _)) :-
 	code_aux__contains_only_builtins_cases(Cases).
 code_aux__contains_only_builtins_2(not(Goal)) :-
 	code_aux__contains_only_builtins(Goal).
-code_aux__contains_only_builtins_2(some(_Vars, Goal)) :-
+code_aux__contains_only_builtins_2(some(_Vars, _, Goal)) :-
 	code_aux__contains_only_builtins(Goal).
 code_aux__contains_only_builtins_2(if_then_else(_Vars, Cond, Then, Else, _)) :-
 	code_aux__contains_only_builtins(Cond),
@@ -84,7 +84,7 @@ code_aux__contains_only_builtins_2(unify(_, _, _, Uni, _)) :-
 	;
 		Uni = simple_test(_, _)
 	;
-		Uni = construct(_, _, _, _)
+		Uni = construct(_, _, _, _, _, _, _)
 	;
 		Uni = deconstruct(_, _, _, _, _)
 	).
@@ -133,20 +133,24 @@ code_aux__goal_cannot_loop_2(ModuleInfo, switch(_Var, _Category, Cases, _)) :-
 	).
 code_aux__goal_cannot_loop_2(ModuleInfo, not(Goal)) :-
 	code_aux__goal_cannot_loop(ModuleInfo, Goal).
-code_aux__goal_cannot_loop_2(ModuleInfo, some(_Vars, Goal)) :-
+code_aux__goal_cannot_loop_2(ModuleInfo, some(_Vars, _, Goal)) :-
 	code_aux__goal_cannot_loop(ModuleInfo, Goal).
 code_aux__goal_cannot_loop_2(ModuleInfo,
 		if_then_else(_Vars, Cond, Then, Else, _)) :-
 	code_aux__goal_cannot_loop(ModuleInfo, Cond),
 	code_aux__goal_cannot_loop(ModuleInfo, Then),
 	code_aux__goal_cannot_loop(ModuleInfo, Else).
+code_aux__goal_cannot_loop_2(ModuleInfo, call(PredId, ProcId, _, _, _, _)) :-
+	module_info_pred_proc_info(ModuleInfo, PredId, ProcId, _, ProcInfo),
+	proc_info_get_maybe_termination_info(ProcInfo, MaybeTermInfo),
+	MaybeTermInfo = yes(cannot_loop).
 code_aux__goal_cannot_loop_2(_, unify(_, _, _, Uni, _)) :-
 	(
 		Uni = assign(_, _)
 	;
 		Uni = simple_test(_, _)
 	;
-		Uni = construct(_, _, _, _)
+		Uni = construct(_, _, _, _, _, _, _)
 	;
 		Uni = deconstruct(_, _, _, _, _)
 	).
@@ -165,12 +169,12 @@ code_aux__goal_is_flat_2(conj(Goals)) :-
 	code_aux__goal_is_flat_list(Goals).
 code_aux__goal_is_flat_2(not(Goal)) :-
 	code_aux__goal_is_flat(Goal).
-code_aux__goal_is_flat_2(some(_Vars, Goal)) :-
+code_aux__goal_is_flat_2(some(_Vars, _, Goal)) :-
 	code_aux__goal_is_flat(Goal).
-code_aux__goal_is_flat_2(higher_order_call(_, _, _, _, _, _)).
+code_aux__goal_is_flat_2(generic_call(_, _, _, _)).
 code_aux__goal_is_flat_2(call(_, _, _, _, _, _)).
 code_aux__goal_is_flat_2(unify(_, _, _, _, _)).
-code_aux__goal_is_flat_2(pragma_c_code(_, _, _, _, _, _, _, _)).
+code_aux__goal_is_flat_2(pragma_c_code(_, _, _, _, _, _, _)).
 
 %-----------------------------------------------------------------------------%
 
@@ -229,8 +233,8 @@ code_aux__explain_stack_slots(StackSlots, VarSet, Explanation) :-
 	string__append("\nStack slot assignments (if any):\n", Explanation1,
 		Explanation).
 
-:- pred code_aux__explain_stack_slots_2(assoc_list(var, lval), varset, string,
-				string).
+:- pred code_aux__explain_stack_slots_2(assoc_list(prog_var, lval), prog_varset,
+		string, string).
 :- mode code_aux__explain_stack_slots_2(in, in, in, out) is det.
 
 code_aux__explain_stack_slots_2([], _, String, String).
