@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -24,6 +24,11 @@
 %		`don't care' mode are ignored, and that assignments
 %		are cached.
 
+% XXX Currently the compiler generates code which does unnecessary
+% shuffling, in code_exprn__place_exprn.  The stuff in
+% code_exprn__fix_clearcode does not seem to be working.
+% See the XXX comment below.
+
 %------------------------------------------------------------------------------%
 %------------------------------------------------------------------------------%
 
@@ -31,7 +36,8 @@
 
 :- interface.
 
-:- import_module hlds_goal, llds, list, varset, assoc_list, options.
+:- import_module prog_data, hlds_goal, llds, options.
+:- import_module map, set, list, assoc_list.
 
 :- type exprn_info.
 
@@ -49,8 +55,8 @@
 %		the table of options; this is used to decide what expressions
 %		are considered constants.
 
-:- pred code_exprn__init_state(assoc_list(var, rval), varset, stack_slots,
-	follow_vars, option_table, exprn_info).
+:- pred code_exprn__init_state(assoc_list(prog_var, rval), prog_varset,
+		stack_slots, follow_vars, option_table, exprn_info).
 :- mode code_exprn__init_state(in, in, in, in, in, out) is det.
 
 %	code_exprn__reinit_state(VarLocs, ExprnInfo0, ExprnInfo)
@@ -62,7 +68,8 @@
 %		of variables and lvalues. The new state places the given
 %		variables at their corresponding locations.
 
-:- pred code_exprn__reinit_state(assoc_list(var, rval), exprn_info, exprn_info).
+:- pred code_exprn__reinit_state(assoc_list(prog_var, rval),
+		exprn_info, exprn_info).
 :- mode code_exprn__reinit_state(in, in, out) is det.
 
 %	code_exprn__clobber_regs(CriticalVars, ExprnInfo0, ExprnInfo)
@@ -72,17 +79,18 @@
 %		registers, and are not stored on the stack, then this
 %		predicate will abort.
 
-:- pred code_exprn__clobber_regs(list(var), exprn_info, exprn_info).
+:- pred code_exprn__clobber_regs(list(prog_var), exprn_info, exprn_info).
 :- mode code_exprn__clobber_regs(in, in, out) is det.
 
 %	code_exprn__set_var_location(Var, Lval, ExprnInfo0, ExprnInfo)
 %		Modifies ExprnInfo0 to produce ExprnInfo in which
 %		Var is *magically* stored in Lval.
 
-:- pred code_exprn__set_var_location(var, lval, exprn_info, exprn_info).
+:- pred code_exprn__set_var_location(prog_var, lval, exprn_info, exprn_info).
 :- mode code_exprn__set_var_location(in, in, in, out) is det.
 
-:- pred code_exprn__maybe_set_var_location(var, lval, exprn_info, exprn_info).
+:- pred code_exprn__maybe_set_var_location(prog_var, lval,
+		exprn_info, exprn_info).
 :- mode code_exprn__maybe_set_var_location(in, in, in, out) is det.
 
 :- pred code_exprn__lval_in_use(lval, exprn_info, exprn_info).
@@ -94,7 +102,7 @@
 %		any cached expressions which still need those resources
 %		will inherit them appropriately).
 
-:- pred code_exprn__var_becomes_dead(var, exprn_info, exprn_info).
+:- pred code_exprn__var_becomes_dead(prog_var, exprn_info, exprn_info).
 :- mode code_exprn__var_becomes_dead(in, in, out) is det.
 
 %	code_exprn__cache_exprn(Var, Rval, ExprnInfo0, ExprnInfo)
@@ -102,14 +110,15 @@
 %		which indicates that when a value of Var is needed,
 %		code to evaluate Rval should be produced.
 
-:- pred code_exprn__cache_exprn(var, rval, exprn_info, exprn_info).
+:- pred code_exprn__cache_exprn(prog_var, rval, exprn_info, exprn_info).
 :- mode code_exprn__cache_exprn(in, in, in, out) is det.
 
 %	code_exprn__place_var(Var, Lval, Code, ExprnInfo0, ExprnInfo)
 %		Produces Code and a modified version of ExprnInfo0,
 %		ExprnInfo which places the value of Var in Lval.
 
-:- pred code_exprn__place_var(var, lval, code_tree, exprn_info, exprn_info).
+:- pred code_exprn__place_var(prog_var, lval, code_tree,
+		exprn_info, exprn_info).
 :- mode code_exprn__place_var(in, in, out, in, out) is det.
 
 %	code_exprn__place_vars(StoreMap, Code, ExprnInfo0, ExprnInfo)
@@ -117,7 +126,7 @@
 %		ExprnInfo which places the value of each variable
 %		mentioned in the store map into the corresponding location.
 
-:- pred code_exprn__place_vars(assoc_list(var, lval), code_tree,
+:- pred code_exprn__place_vars(assoc_list(prog_var, lval), code_tree,
 	exprn_info, exprn_info).
 :- mode code_exprn__place_vars(in, out, in, out) is det.
 
@@ -125,14 +134,15 @@
 %		Produces a code fragment Code to evaluate Var and
 %		provide it as Rval (which may be a const, etc, or an lval).
 
-:- pred code_exprn__produce_var(var, rval, code_tree, exprn_info, exprn_info).
+:- pred code_exprn__produce_var(prog_var, rval, code_tree,
+		exprn_info, exprn_info).
 :- mode code_exprn__produce_var(in, out, out, in, out) is det.
 
 %	code_exprn__produce_var_in_reg(Var, Rval, Code, ExprnInfo0, ExprnInfo)
 %		Produces a code fragment Code to evaluate Var and
 %		provide it as an Rval of the form lval(reg(_)).
 
-:- pred code_exprn__produce_var_in_reg(var, rval, code_tree,
+:- pred code_exprn__produce_var_in_reg(prog_var, rval, code_tree,
 	exprn_info, exprn_info).
 :- mode code_exprn__produce_var_in_reg(in, out, out, in, out) is det.
 
@@ -142,7 +152,7 @@
 %		provide it as an Rval of the form lval(reg(_)),
 %		lval(stackvar(_)), or lval(framevar(_)).
 
-:- pred code_exprn__produce_var_in_reg_or_stack(var, rval, code_tree,
+:- pred code_exprn__produce_var_in_reg_or_stack(prog_var, rval, code_tree,
 	exprn_info, exprn_info).
 :- mode code_exprn__produce_var_in_reg_or_stack(in, out, out, in, out) is det.
 
@@ -209,7 +219,7 @@
 %		Returns a map from each variable that occurs in ExprnInfo to
 %		the set of locations (really rvals) in which it may be found.
 
-:- pred code_exprn__get_varlocs(exprn_info, map(var, set(rval))).
+:- pred code_exprn__get_varlocs(exprn_info, map(prog_var, set(rval))).
 :- mode code_exprn__get_varlocs(in, out) is det.
 
 %	code_exprn__get_stack_slots(StackSlots)
@@ -233,22 +243,28 @@
 :- pred code_exprn__set_follow_vars(follow_vars, exprn_info, exprn_info).
 :- mode code_exprn__set_follow_vars(in, in, out) is det.
 
+%	code_exprn__max_reg_in_use(MaxReg)
+%		Returns the number of the highest numbered rN register in use.
+
+:- pred code_exprn__max_reg_in_use(exprn_info, int).
+:- mode code_exprn__max_reg_in_use(in, out) is det.
+
 %------------------------------------------------------------------------------%
 %------------------------------------------------------------------------------%
 
 :- implementation.
 
-:- import_module exprn_aux, tree.
-:- import_module bool, map, bag, set, require, int, term, string, std_util.
+:- import_module code_util, exprn_aux, tree, varset, term.
+:- import_module bool, bag, require, int, string, std_util.
 
 :- type var_stat	--->	evaled(set(rval))
 			;	cached(rval).
 
-:- type var_map	==	map(var, var_stat).
+:- type var_map	==	map(prog_var, var_stat).
 
 :- type exprn_info	--->
 		exprn_info(
-			varset,		% all the variables and their names
+			prog_varset,	% all the variables and their names
 			var_map,	% what each variable stands for
 			bag(lval),	% the 'in use' markers for regs
 			set(lval),	% extra markers for acquired regs
@@ -279,7 +295,7 @@ code_exprn__reinit_state(VarLocs, ExprnInfo0, ExprnInfo) :-
 	ExprnInfo = exprn_info(Varset, Vars, Regs, Acqu,
 		StackSlots, FollowVars, ExprnOpts).
 
-:- pred code_exprn__init_state_2(assoc_list(var, rval), var_map, var_map,
+:- pred code_exprn__init_state_2(assoc_list(prog_var, rval), var_map, var_map,
 	bag(lval), bag(lval)).
 :- mode code_exprn__init_state_2(in, in, out, in, out) is det.
 
@@ -312,8 +328,8 @@ code_exprn__get_varlocs(ExprnInfo, Locations) :-
 	map__init(Locations0),
 	code_exprn__repackage_locations(VarList, Locations0, Locations).
 
-:- pred code_exprn__repackage_locations(assoc_list(var, var_stat),
-			map(var, set(rval)), map(var, set(rval))).
+:- pred code_exprn__repackage_locations(assoc_list(prog_var, var_stat),
+			map(prog_var, set(rval)), map(prog_var, set(rval))).
 :- mode code_exprn__repackage_locations(in, in, out) is det.
 
 code_exprn__repackage_locations([], Loc, Loc).
@@ -341,8 +357,8 @@ code_exprn__clobber_regs(CriticalVars) -->
 	{ set__init(Acqu) },
 	code_exprn__set_acquired(Acqu).
 
-:- pred code_exprn__clobber_regs_2(assoc_list(var, var_stat), list(var),
-						var_map, var_map, var_map).
+:- pred code_exprn__clobber_regs_2(assoc_list(prog_var, var_stat),
+		list(prog_var), var_map, var_map, var_map).
 :- mode code_exprn__clobber_regs_2(in, in, in, in, out) is det.
 
 code_exprn__clobber_regs_2([], _Critical, _OldVars, Vars, Vars).
@@ -398,7 +414,7 @@ code_exprn__rval_depends_on_reg(var(Var), Vars) :-
 		code_exprn__filter_out_reg_depending(Rvals0, Vars, Rvals),
 		set__empty(Rvals)
 	).
-code_exprn__rval_depends_on_reg(create(_Tag, Rvals, _Unique, _LabNum), Vars) :-
+code_exprn__rval_depends_on_reg(create(_, Rvals, _, _, _, _), Vars) :-
 	code_exprn__args_depend_on_reg(Rvals, Vars).
 code_exprn__rval_depends_on_reg(mkword(_Tag, Rval), Vars) :-
 	code_exprn__rval_depends_on_reg(Rval, Vars).
@@ -494,33 +510,21 @@ code_exprn__maybe_set_var_location(Var, Lval) -->
 %------------------------------------------------------------------------------%
 
 code_exprn__lval_in_use(Lval) -->
-	code_exprn__lval_in_use_except(Lval, no).
-
-:- pred code_exprn__lval_in_use_except(lval, maybe(var),
-	exprn_info, exprn_info).
-:- mode code_exprn__lval_in_use_except(in, in, in, out) is semidet.
-
-code_exprn__lval_in_use_except(Lval, MaybeVar) -->
 	code_exprn__get_vars(Vars),
-	{ map__to_assoc_list(Vars, VarStatList) },
-	{ code_exprn__lval_in_use_2(Lval, MaybeVar, VarStatList) }.
-
-	% The auxiliary predicate is needed to allow Mercury to recognize
-	% that the subcomputation has no outputs without using existential
-	% quantification, whose syntax SICStus doesn't like.
-
-:- pred code_exprn__lval_in_use_2(lval, maybe(var), assoc_list(var, var_stat)).
-:- mode code_exprn__lval_in_use_2(in, in, in) is semidet.
-
-code_exprn__lval_in_use_2(Lval, MaybeVar, VarStatList) :-
-	list__member(VarStat, VarStatList),
-	VarStat = Var - Stat,
+	code_exprn__get_regs(Regs),
 	(
-		MaybeVar = no
+		{ bag__contains(Regs, Lval) }
 	;
-		MaybeVar = yes(ExceptionVar),
-		\+ Var = ExceptionVar
-	),
+		{ map__to_assoc_list(Vars, VarStatList) },
+		{ code_exprn__lval_in_use_by_vars(Lval, VarStatList) }
+	).
+
+:- pred code_exprn__lval_in_use_by_vars(lval, assoc_list(prog_var, var_stat)).
+:- mode code_exprn__lval_in_use_by_vars(in, in) is semidet.
+
+code_exprn__lval_in_use_by_vars(Lval, VarStatList) :-
+	list__member(VarStat, VarStatList),
+	VarStat = _Var - Stat,
 	(
 		Stat = cached(Rval),
 		exprn_aux__rval_contains_lval(Rval, Lval)
@@ -540,8 +544,8 @@ code_exprn__clear_lval_of_synonyms(Lval) -->
 	{ map__to_assoc_list(Vars, VarStatList) },
 	code_exprn__clear_lval_of_synonyms_1(VarStatList, Lval).
 
-:- pred code_exprn__clear_lval_of_synonyms_1(assoc_list(var, var_stat), lval,
-	exprn_info, exprn_info).
+:- pred code_exprn__clear_lval_of_synonyms_1(assoc_list(prog_var, var_stat),
+		lval, exprn_info, exprn_info).
 :- mode code_exprn__clear_lval_of_synonyms_1(in, in, in, out) is det.
 
 code_exprn__clear_lval_of_synonyms_1([], _) --> [].
@@ -612,7 +616,7 @@ code_exprn__add_rval_list_reg_dependencies([R | Rs]) -->
 code_exprn__add_rval_reg_dependencies(lval(Lval)) -->
 	code_exprn__add_lval_reg_dependencies(Lval).
 code_exprn__add_rval_reg_dependencies(var(_Var)) --> [].
-code_exprn__add_rval_reg_dependencies(create(_, Rvals, _, _)) -->
+code_exprn__add_rval_reg_dependencies(create(_, Rvals, _, _, _, _)) -->
 	code_exprn__add_arg_reg_dependencies(Rvals).
 code_exprn__add_rval_reg_dependencies(mkword(_Tag, Rval)) -->
 	code_exprn__add_rval_reg_dependencies(Rval).
@@ -685,7 +689,7 @@ code_exprn__rem_rval_list_reg_dependencies([R | Rs]) -->
 code_exprn__rem_rval_reg_dependencies(lval(Lval)) -->
 	code_exprn__rem_lval_reg_dependencies(Lval).
 code_exprn__rem_rval_reg_dependencies(var(_Var)) --> [].
-code_exprn__rem_rval_reg_dependencies(create(_, Rvals, _, _)) -->
+code_exprn__rem_rval_reg_dependencies(create(_, Rvals, _, _, _, _)) -->
 	code_exprn__rem_arg_reg_dependencies(Rvals).
 code_exprn__rem_rval_reg_dependencies(mkword(_Tag, Rval)) -->
 	code_exprn__rem_rval_reg_dependencies(Rval).
@@ -764,7 +768,8 @@ code_exprn__var_becomes_dead(Var) -->
 
 %------------------------------------------------------------------------------%
 
-:- pred code_exprn__update_dependent_vars(var, rval, exprn_info, exprn_info).
+:- pred code_exprn__update_dependent_vars(prog_var, rval,
+		exprn_info, exprn_info).
 :- mode code_exprn__update_dependent_vars(in, in, in, out) is det.
 
 code_exprn__update_dependent_vars(Var, Rval) -->
@@ -774,8 +779,9 @@ code_exprn__update_dependent_vars(Var, Rval) -->
 	{ map__from_assoc_list(VarList, Vars) },
 	code_exprn__set_vars(Vars).
 
-:- pred code_exprn__update_dependent_vars_2(assoc_list(var, var_stat),
-		var, rval, assoc_list(var, var_stat), exprn_info, exprn_info).
+:- pred code_exprn__update_dependent_vars_2(assoc_list(prog_var, var_stat),
+		prog_var, rval, assoc_list(prog_var, var_stat),
+		exprn_info, exprn_info).
 :- mode code_exprn__update_dependent_vars_2(in, in, in, out, in, out) is det.
 
 code_exprn__update_dependent_vars_2([], _Var, _Rval, []) --> [].
@@ -811,8 +817,8 @@ code_exprn__update_dependent_vars_2([V - Stat0 | Rest0], Var, Rval,
 	),
 	code_exprn__update_dependent_vars_2(Rest0, Var, Rval, Rest).
 
-:- pred code_exprn__update_dependent_vars_3(list(rval), var, rval, list(rval),
-					exprn_info, exprn_info).
+:- pred code_exprn__update_dependent_vars_3(list(rval), prog_var, rval,
+		list(rval), exprn_info, exprn_info).
 :- mode code_exprn__update_dependent_vars_3(in, in, in, out, in, out) is det.
 
 code_exprn__update_dependent_vars_3([], _Var, _Rval, []) --> [].
@@ -911,11 +917,16 @@ code_exprn__expr_is_constant(mkword(Tag, Expr0), Vars, ExprnOpts,
 		mkword(Tag, Expr)) :-
 	code_exprn__expr_is_constant(Expr0, Vars, ExprnOpts, Expr).
 
-code_exprn__expr_is_constant(create(Tag, Args0, Unique, Label), Vars, ExprnOpts,
-		create(Tag, Args, Unique, Label)) :-
-	ExprnOpts = nlg_asm_sgt_ubf(_, _, StaticGroundTerms, _),
-	StaticGroundTerms = yes,
-	code_exprn__args_are_constant(Args0, Vars, ExprnOpts, Args).
+code_exprn__expr_is_constant(create(Tag, Args0, ArgTypes, StatDyn, Label, Msg),
+		Vars, ExprnOpts, NewRval) :-
+	( StatDyn = must_be_static ->
+		NewRval = create(Tag, Args0, ArgTypes, StatDyn, Label, Msg)
+	;
+		ExprnOpts = nlg_asm_sgt_ubf(_, _, StaticGroundTerms, _),
+		StaticGroundTerms = yes,
+		code_exprn__args_are_constant(Args0, Vars, ExprnOpts, Args),
+		NewRval = create(Tag, Args, ArgTypes, StatDyn, Label, Msg)
+	).
 
 code_exprn__expr_is_constant(var(Var), Vars, ExprnOpts, Rval) :-
 	map__search(Vars, Var, Stat),
@@ -997,7 +1008,7 @@ code_exprn__place_var(Var, Lval, Code) -->
 		code_exprn__place_evaled(Rvals, Var, Lval, Code)
 	).
 
-:- pred code_exprn__place_cached(rval, var, lval, code_tree,
+:- pred code_exprn__place_cached(rval, prog_var, lval, code_tree,
 						exprn_info, exprn_info).
 :- mode code_exprn__place_cached(in, in, in, out, in, out) is det.
 
@@ -1035,7 +1046,7 @@ code_exprn__place_cached(Rval0, Var, Lval, Code) -->
 			_, Code)
 	).
 
-:- pred code_exprn__place_evaled(set(rval), var, lval, code_tree,
+:- pred code_exprn__place_evaled(set(rval), prog_var, lval, code_tree,
 	exprn_info, exprn_info).
 :- mode code_exprn__place_evaled(in, in, in, out, in, out) is det.
 
@@ -1121,7 +1132,7 @@ code_exprn__place_arg(Rval0, MaybeLval, Lval, Code) -->
 			Code)
 	).
 
-:- pred code_exprn__place_exprn(maybe(lval), maybe(var), rval, bool, bool,
+:- pred code_exprn__place_exprn(maybe(lval), maybe(prog_var), rval, bool, bool,
 	lval, code_tree, exprn_info, exprn_info).
 :- mode code_exprn__place_exprn(in, in, in, in, in, out, out, in, out) is det.
 
@@ -1163,7 +1174,7 @@ code_exprn__place_exprn(MaybeLval, MaybeVar, Rval0, StandAlone, IsConst,
 	code_exprn__add_lval_reg_dependencies(Lval),
 	code_exprn__maybe_get_var_name(MaybeVar, VarName),
 	( { IsConst = yes } ->
-		{ string__append("Assigning from ", VarName, Comment) },
+		{ string__append("Assigning from const ", VarName, Comment) },
 		{ ExprnCode = node([assign(Lval, Rval2) - Comment]) }
 	;
 		( { StandAlone = yes } ->
@@ -1190,6 +1201,15 @@ code_exprn__place_exprn(MaybeLval, MaybeVar, Rval0, StandAlone, IsConst,
 	exprn_info, exprn_info).
 :- mode code_exprn__maybe_fix_clearcode(in, in, out, in, out) is det.
 
+	% XXX
+	% This predicate is supposed to optimize away unnecessary
+	% register shuffles.  It doesn't do much currently,
+	% because the target register is always marked in use
+	% in code_expr__place_exprn via the call to
+	% add_lval_reg_dependencies above.
+	% (The fact that there is no matching call to
+	% rem_lval_reg_dependencies is perhaps the problem?)
+
 code_exprn__maybe_fix_clearcode(ClearCode, ExprnCode, Code) -->
 	(
 		{ ClearCode = node([assign(reg(RegT, RegN), lval(Lval)) - _]) },
@@ -1234,7 +1254,7 @@ code_exprn__find_real_creates([MaybeRval | MaybeRvals], SubCreates) -->
 
 code_exprn__rval_is_real_create(Rval) -->
 	(
-		{ Rval = create(_, _, _, _) },
+		{ Rval = create(_, _, _, _, _, _) },
 		code_exprn__get_vars(Vars0),
 		code_exprn__get_options(ExprnOpts),
 		{ \+ code_exprn__expr_is_constant(Rval, Vars0, ExprnOpts, _) }
@@ -1260,8 +1280,11 @@ code_exprn__rval_is_real_create(Rval) -->
 code_exprn__construct_code(Lval, VarName, Rval0, Code) -->
 	{ exprn_aux__simplify_rval(Rval0, Rval) },
 	(
-		{ Rval = create(Tag, Rvals, _Unique, _Label) }
+		{ Rval = create(Tag, Rvals, ArgTypes, _StatDyn, _Label, Msg) }
 	->
+		{ require(lambda([] is semidet,
+			(llds__all_args_are_word_size(ArgTypes, yes))),
+		"trying to construct heap cell with non-word-size arg(s)") },
 		{ list__length(Rvals, Arity) },
 		(
 			{ Arity = 0 }
@@ -1273,8 +1296,8 @@ code_exprn__construct_code(Lval, VarName, Rval0, Code) -->
 		;
 			( { Lval = field(_, _, _) } ->
 				code_exprn__acquire_reg(r, Reg),
-				code_exprn__construct_cell(Reg,
-					VarName, Tag, Arity, Rvals, Code0),
+				code_exprn__construct_cell(Reg, VarName,
+					Tag, Arity, Rvals, Msg, Code0),
 				{ string__append(VarName, " placement",
 					Comment) },
 				{ Code1 = node([
@@ -1283,8 +1306,8 @@ code_exprn__construct_code(Lval, VarName, Rval0, Code) -->
 				{ Code = tree(Code0, Code1) },
 				code_exprn__release_reg(Reg)
 			;
-				code_exprn__construct_cell(Lval,
-					VarName, Tag, Arity, Rvals, Code)
+				code_exprn__construct_cell(Lval, VarName,
+					Tag, Arity, Rvals, Msg, Code)
 			)
 		)
 	;
@@ -1293,13 +1316,14 @@ code_exprn__construct_code(Lval, VarName, Rval0, Code) -->
 	).
 
 :- pred code_exprn__construct_cell(lval, string, tag, int, list(maybe(rval)),
-	code_tree, exprn_info, exprn_info).
-:- mode code_exprn__construct_cell(in, in, in, in, in, out, in, out) is det.
+	string, code_tree, exprn_info, exprn_info).
+:- mode code_exprn__construct_cell(in, in, in, in, in, in, out, in, out) is det.
 
-code_exprn__construct_cell(Lval, VarName, Tag, Arity, Rvals, Code) -->
+code_exprn__construct_cell(Lval, VarName, Tag, Arity, Rvals, TypeMsg, Code) -->
 	{ string__append("Allocating heap for ", VarName, Comment) },
 	{ Code0 = node([
-		incr_hp(Lval, yes(Tag), const(int_const(Arity))) - Comment
+		incr_hp(Lval, yes(Tag), const(int_const(Arity)), TypeMsg)
+			- Comment
 	]) },
 	code_exprn__construct_args(Rvals, Tag, Lval, 0, Targets, Code1),
 	code_exprn__free_arg_dependenciess(Targets),
@@ -1314,7 +1338,7 @@ code_exprn__construct_args([R | Rs], Tag, Lval, N0, Targets, Code) -->
 	(
 		{ R = yes(Rval) }
 	->
-		{ Target0 = field(Tag, lval(Lval), const(int_const(N0))) },
+		{ Target0 = field(yes(Tag), lval(Lval), const(int_const(N0))) },
 		{ MaybeTarget = yes(Target0) },
 		code_exprn__place_arg(Rval, yes(Target0), _, Code0)
 	;
@@ -1342,8 +1366,8 @@ code_exprn__free_arg_dependenciess([Target | Targets]) -->
 
 %------------------------------------------------------------------------------%
 
-:- pred code_exprn__produce_vars(list(var), assoc_list(var, rval), code_tree,
-	exprn_info, exprn_info).
+:- pred code_exprn__produce_vars(list(prog_var), assoc_list(prog_var, rval),
+		code_tree, exprn_info, exprn_info).
 :- mode code_exprn__produce_vars(in, out, out, in, out) is det.
 
 code_exprn__produce_vars([], [], empty) --> [].
@@ -1365,7 +1389,7 @@ code_exprn__produce_var(Var, Rval, Code) -->
 			;
 				RvalX = unop(_, _)
 			;
-				RvalX = create(_, _, _, _)
+				RvalX = create(_, _, _, _, _, _)
 			;
 				RvalX = mkword(_, _)
 			}
@@ -1439,7 +1463,8 @@ code_exprn__select_reg_or_stack_rval([Rval0 | Rvals0], Rval) :-
 
 %------------------------------------------------------------------------------%
 
-:- pred code_exprn__select_preferred_lval(var, lval, exprn_info, exprn_info).
+:- pred code_exprn__select_preferred_lval(prog_var, lval,
+		exprn_info, exprn_info).
 :- mode code_exprn__select_preferred_lval(in, out, in, out) is det.
 
 code_exprn__select_preferred_lval(Var, Lval) -->
@@ -1459,7 +1484,8 @@ code_exprn__select_preferred_lval(Var, Lval) -->
 		code_exprn__get_spare_reg(r, Lval)
 	).
 
-:- pred code_exprn__select_preferred_reg(var, lval, exprn_info, exprn_info).
+:- pred code_exprn__select_preferred_reg(prog_var, lval,
+		exprn_info, exprn_info).
 :- mode code_exprn__select_preferred_reg(in, out, in, out) is det.
 
 code_exprn__select_preferred_reg(Var, Lval) -->
@@ -1568,8 +1594,8 @@ code_exprn__clear_lval_return_shuffle(Lval, MaybeShuffle, Code) -->
 		{ Code = empty }
 	).
 
-:- pred code_exprn__relocate_lval(assoc_list(var, var_stat), lval, lval,
-	assoc_list(var, var_stat), exprn_info, exprn_info).
+:- pred code_exprn__relocate_lval(assoc_list(prog_var, var_stat), lval, lval,
+	assoc_list(prog_var, var_stat), exprn_info, exprn_info).
 :- mode code_exprn__relocate_lval(in, in, in, out, in, out) is det.
 
 code_exprn__relocate_lval([], _OldVal, _NewVal, []) --> [].
@@ -1617,7 +1643,7 @@ code_exprn__relocate_lval_2([R0 | Rs0], OldVal, NewVal, [R | Rs]) -->
 
 %------------------------------------------------------------------------------%
 
-:- pred code_exprn__get_var_status(var, var_stat, exprn_info, exprn_info).
+:- pred code_exprn__get_var_status(prog_var, var_stat, exprn_info, exprn_info).
 :- mode code_exprn__get_var_status(in, out, in, out) is det.
 
 code_exprn__get_var_status(Var, Stat) -->
@@ -1635,7 +1661,7 @@ code_exprn__get_var_status(Var, Stat) -->
 		{ error(Msg) }
 	).
 
-:- pred code_exprn__maybe_set_evaled(maybe(var), list(rval),
+:- pred code_exprn__maybe_set_evaled(maybe(prog_var), list(rval),
 	exprn_info, exprn_info).
 :- mode code_exprn__maybe_set_evaled(in, in, in, out) is det.
 
@@ -1647,7 +1673,7 @@ code_exprn__maybe_set_evaled(yes(Var), RvalList) -->
 	{ map__set(Vars0, Var, Stat, Vars) },
 	code_exprn__set_vars(Vars).
 
-:- pred code_exprn__maybe_add_evaled(maybe(var), rval, exprn_info, exprn_info).
+:- pred code_exprn__maybe_add_evaled(maybe(prog_var), rval, exprn_info, exprn_info).
 :- mode code_exprn__maybe_add_evaled(in, in, in, out) is det.
 
 code_exprn__maybe_add_evaled(no, _) --> [].
@@ -1748,7 +1774,7 @@ code_exprn__unlock_reg(Reg) -->
 
 %------------------------------------------------------------------------------%
 
-:- pred code_exprn__maybe_get_var_name(maybe(var), string,
+:- pred code_exprn__maybe_get_var_name(maybe(prog_var), string,
 	exprn_info, exprn_info).
 :- mode code_exprn__maybe_get_var_name(in, out, in, out) is det.
 
@@ -1756,7 +1782,7 @@ code_exprn__maybe_get_var_name(no, "unknown variable") --> [].
 code_exprn__maybe_get_var_name(yes(Var), Name) -->
 	code_exprn__get_var_name(Var, Name).
 
-:- pred code_exprn__get_var_name(var, string, exprn_info, exprn_info).
+:- pred code_exprn__get_var_name(prog_var, string, exprn_info, exprn_info).
 :- mode code_exprn__get_var_name(in, out, in, out) is det.
 
 code_exprn__get_var_name(Var, Name) -->
@@ -1765,10 +1791,59 @@ code_exprn__get_var_name(Var, Name) -->
 
 %------------------------------------------------------------------------------%
 
-:- pred code_exprn__get_varset(varset, exprn_info, exprn_info).
+code_exprn__max_reg_in_use(ExprnInfo, Max) :-
+	code_exprn__get_vars(Vars, ExprnInfo, _),
+	map__to_assoc_list(Vars, VarStats),
+	assoc_list__values(VarStats, Stats),
+	code_exprn__max_reg_in_use_vars(Stats, 0, Max1),
+	code_exprn__get_regs(InUseRegs, ExprnInfo, _),
+	bag__to_list_without_duplicates(InUseRegs, IRegs),
+	code_exprn__max_reg_in_use_lvals(IRegs, Max1, Max2),
+	code_exprn__get_acquired(Acquired, ExprnInfo, _),
+	set__to_sorted_list(Acquired, ARegs),
+	code_exprn__max_reg_in_use_lvals(ARegs, Max2, Max).
+
+:- pred code_exprn__max_reg_in_use_vars(list(var_stat), int, int).
+:- mode code_exprn__max_reg_in_use_vars(in, in, out) is det.
+
+code_exprn__max_reg_in_use_vars([], Max, Max).
+code_exprn__max_reg_in_use_vars([Stat | Stats], Max0, Max) :-
+	(
+		Stat = evaled(RvalSet),
+		set__to_sorted_list(RvalSet, Rvals),
+		code_exprn__max_reg_in_use_rvals(Rvals, Max0, Max1)
+	;
+		Stat = cached(Rval),
+		code_exprn__max_reg_in_use_rvals([Rval], Max0, Max1)
+	),
+	code_exprn__max_reg_in_use_vars(Stats, Max1, Max).
+
+:- pred code_exprn__max_reg_in_use_rvals(list(rval), int, int).
+:- mode code_exprn__max_reg_in_use_rvals(in, in, out) is det.
+
+code_exprn__max_reg_in_use_rvals(Rvals, Max0, Max) :-
+	list__map(code_util__lvals_in_rval, Rvals, LvalLists),
+	list__condense(LvalLists, Lvals),
+	code_exprn__max_reg_in_use_lvals(Lvals, Max0, Max).
+
+:- pred code_exprn__max_reg_in_use_lvals(list(lval), int, int).
+:- mode code_exprn__max_reg_in_use_lvals(in, in, out) is det.
+
+code_exprn__max_reg_in_use_lvals(Lvals, Max0, Max) :-
+	list__filter_map(code_exprn__lval_is_r_reg, Lvals, RegNumbers),
+	list__foldl(int__max, RegNumbers, Max0, Max).
+
+:- pred code_exprn__lval_is_r_reg(lval, int).
+:- mode code_exprn__lval_is_r_reg(in, out) is semidet.
+
+code_exprn__lval_is_r_reg(reg(r, N), N).
+
+%------------------------------------------------------------------------------%
+
+:- pred code_exprn__get_varset(prog_varset, exprn_info, exprn_info).
 :- mode code_exprn__get_varset(out, in, out) is det.
 
-:- pred code_exprn__set_varset(varset, exprn_info, exprn_info).
+:- pred code_exprn__set_varset(prog_varset, exprn_info, exprn_info).
 :- mode code_exprn__set_varset(in, in, out) is det.
 
 :- pred code_exprn__get_vars(var_map, exprn_info, exprn_info).
