@@ -1,24 +1,24 @@
-%---------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+%-----------------------------------------------------------------------------%
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
-%---------------------------------------------------------------------------%
-%---------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
 %
 % file: code_util.m.
 %
 % various utilities routines for code generation and recognition
 % of builtins.
 %
-%---------------------------------------------------------------------------%
-%---------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
 
 :- module code_util.
 
 :- interface.
 
-:- import_module hlds_module, hlds_pred, hlds_goal, hlds_data, llds.
-:- import_module list.
+:- import_module hlds_module, hlds_pred, hlds_goal, hlds_data, prog_data, llds.
+:- import_module list, assoc_list, set, std_util.
 
 	% Create a code address which holds the address of the specified
 	% procedure.
@@ -57,6 +57,12 @@
 :- pred code_util__make_uni_label(module_info, type_id, proc_id, proc_label).
 :- mode code_util__make_uni_label(in, in, in, out) is det.
 
+:- pred code_util__extract_proc_label_from_code_addr(code_addr, proc_label).
+:- mode code_util__extract_proc_label_from_code_addr(in, out) is det.
+
+:- pred code_util__extract_proc_label_from_label(label, proc_label).
+:- mode code_util__extract_proc_label_from_label(in, out) is det.
+
 :- pred code_util__arg_loc_to_register(arg_loc, lval).
 :- mode code_util__arg_loc_to_register(in, out) is det.
 
@@ -110,8 +116,8 @@
 	% Each to be assigned rval is guaranteed to be either in a form
 	% acceptable for a test rval, or in the form of a variable.
 
-:- pred code_util__translate_builtin(string, string, proc_id, list(var),
-	maybe(rval), maybe(pair(var, rval))).
+:- pred code_util__translate_builtin(module_name, string, proc_id,
+	list(prog_var), maybe(rval), maybe(pair(prog_var, rval))).
 :- mode code_util__translate_builtin(in, in, in, in, out, out) is semidet.
 
 	% Find out how a function symbol (constructor) is represented
@@ -147,11 +153,29 @@
 	int, int).
 :- mode code_util__count_recursive_calls(in, in, in, out, out) is det.
 
+	% Return the set of locations occupied by output arguments.
+
+:- pred code_util__output_args(assoc_list(prog_var, arg_info), set(lval)).
+:- mode code_util__output_args(in, out) is det.
+
+	% These predicates return the set of lvals referenced in an rval
+	% and an lval respectively. Lvals referenced indirectly through
+	% lvals of the form var(_) are not counted.
+
+:- pred code_util__lvals_in_rval(rval, list(lval)).
+:- mode code_util__lvals_in_rval(in, out) is det.
+
+:- pred code_util__lvals_in_lval(lval, list(lval)).
+:- mode code_util__lvals_in_lval(in, out) is det.
+
 %---------------------------------------------------------------------------%
 
 :- implementation.
-:- import_module prog_data, type_util, special_pred.
-:- import_module bool, char, int, string, map, term, varset, require, std_util.
+
+:- import_module prog_data, builtin_ops, type_util, special_pred.
+
+:- import_module bool, char, int, string, set, map, term, varset.
+:- import_module require, std_util, assoc_list.
 
 %---------------------------------------------------------------------------%
 
@@ -181,13 +205,7 @@ code_util__make_local_entry_label(ModuleInfo, PredId, ProcId, Immed, Label) :-
 	module_info_preds(ModuleInfo, Preds),
 	map__lookup(Preds, PredId, PredInfo),
 	(
-		(
-			pred_info_is_exported(PredInfo)
-		;
-			pred_info_is_pseudo_exported(PredInfo),
-			% only the (in, in) mode of a unification is exported
-			hlds_pred__in_in_unification_proc_id(ProcId)
-		)
+		procedure_is_exported(PredInfo, ProcId)
 	->
 		(
 			Immed = no,
@@ -249,7 +267,7 @@ code_util__make_proc_label(ModuleInfo, PredId, ProcId, ProcLabel) :-
 	(
 		code_util__compiler_generated(PredInfo)
 	->
-		pred_info_arg_types(PredInfo, _TypeVarSet, ArgTypes),
+		pred_info_arg_types(PredInfo, ArgTypes),
 		(
 			special_pred_get_type(PredName, ArgTypes, Type),
 			type_to_type_id(Type, TypeId, _),
@@ -308,6 +326,29 @@ code_util__make_uni_label(ModuleInfo, TypeId, UniModeNum, ProcLabel) :-
 		error("code_util__make_uni_label: unqualified type_id")
 	).
 
+code_util__extract_proc_label_from_code_addr(CodeAddr, ProcLabel) :-
+	( code_util__proc_label_from_code_addr(CodeAddr, ProcLabelPrime) ->
+		ProcLabel = ProcLabelPrime
+	;
+		error("code_util__extract_label_from_code_addr failed")
+	).
+
+:- pred code_util__proc_label_from_code_addr(code_addr::in,
+	proc_label::out) is semidet.
+
+code_util__proc_label_from_code_addr(CodeAddr, ProcLabel) :-
+	(
+		CodeAddr = label(Label),
+		code_util__extract_proc_label_from_label(Label, ProcLabel)
+	;
+		CodeAddr = imported(ProcLabel)
+	).
+
+code_util__extract_proc_label_from_label(local(ProcLabel, _), ProcLabel).
+code_util__extract_proc_label_from_label(c_local(ProcLabel), ProcLabel).
+code_util__extract_proc_label_from_label(local(ProcLabel), ProcLabel).
+code_util__extract_proc_label_from_label(exported(ProcLabel), ProcLabel).
+
 %-----------------------------------------------------------------------------%
 
 code_util__arg_loc_to_register(ArgLoc, reg(r, ArgLoc)).
@@ -319,9 +360,8 @@ code_util__predinfo_is_builtin(PredInfo) :-
 	pred_info_name(PredInfo, PredName),
 %	code_util__translate_builtin(ModuleName, PredName, _, _, _, _).
 	pred_info_arity(PredInfo, Arity),
-	( code_util__inline_builtin(ModuleName, PredName, 0, Arity)
-	; code_util__inline_builtin(ModuleName, PredName, 10000, Arity)
-	).
+	ProcId = 0,
+	code_util__inline_builtin(ModuleName, PredName, ProcId, Arity).
 
 code_util__builtin_state(ModuleInfo, PredId0, ProcId, BuiltinState) :-
 	predicate_module(ModuleInfo, PredId0, ModuleName),
@@ -334,112 +374,130 @@ code_util__builtin_state(ModuleInfo, PredId0, ProcId, BuiltinState) :-
 		BuiltinState = not_builtin
 	).
 
-:- pred code_util__inline_builtin(string, string, int, int).
+:- pred code_util__inline_builtin(module_name, string, int, int).
 :- mode code_util__inline_builtin(in, in, in, in) is semidet.
 
-code_util__inline_builtin(ModuleName, PredName, ProcId, Arity) :-
+code_util__inline_builtin(FullyQualifiedModule, PredName, ProcId, Arity) :-
 	Arity =< 3,
 	varset__init(VarSet),
 	varset__new_vars(VarSet, Arity, Args, _),
-	code_util__translate_builtin_2(ModuleName, PredName, ProcId, Args, _, _).
+	% --- not yet:
+	% FullyQualifiedModule = qualified(unqualified("std"), ModuleName),
+	FullyQualifiedModule = unqualified(ModuleName),
+	code_util__translate_builtin_2(ModuleName, PredName, ProcId, Args,
+		_, _).
 
-code_util__translate_builtin(Module, PredName, ProcId, Args, BinOp, AsgOp) :-
+code_util__translate_builtin(FullyQualifiedModule, PredName, ProcId, Args,
+		BinOp, AsgOp) :-
 	proc_id_to_int(ProcId, ProcInt),
-	code_util__translate_builtin_2(Module, PredName, ProcInt, Args,
+	% -- not yet:
+	% FullyQualifiedModule = qualified(unqualified("std"), ModuleName),
+	FullyQualifiedModule = unqualified(ModuleName),
+	code_util__translate_builtin_2(ModuleName, PredName, ProcInt, Args,
 		BinOp, AsgOp).
 
-:- pred code_util__translate_builtin_2(string, string, int, list(var),
-	maybe(rval), maybe(pair(var, rval))).
+:- pred code_util__translate_builtin_2(string, string, int, list(prog_var),
+	maybe(rval), maybe(pair(prog_var, rval))).
 :- mode code_util__translate_builtin_2(in, in, in, in, out, out) is semidet.
 
-code_util__translate_builtin_2("mercury_builtin", "builtin_int_gt", 0, [X, Y],
+% WARNING: any changes here will probably require similar changes
+% in ml_code_gen:ml_translate_builtin_2 and vice versa.
+
+code_util__translate_builtin_2("private_builtin", "unsafe_type_cast", 0,
+		[X, Y], no, yes(Y - var(X))).
+code_util__translate_builtin_2("builtin", "unsafe_promise_unique", 0,
+		[X, Y], no, yes(Y - var(X))).
+
+code_util__translate_builtin_2("private_builtin", "builtin_int_gt", 0, [X, Y],
 	yes(binop((>), var(X), var(Y))), no).
-code_util__translate_builtin_2("mercury_builtin", "builtin_int_lt", 0, [X, Y],
+code_util__translate_builtin_2("private_builtin", "builtin_int_lt", 0, [X, Y],
 	yes(binop((<), var(X), var(Y))), no).
 
-code_util__translate_builtin_2("int", "builtin_plus", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_plus", 0, [X, Y, Z],
 	no, yes(Z - binop((+), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_plus", 10001, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_plus", 1, [X, Y, Z],
 	no, yes(X - binop((-), var(Z), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_plus", 10002, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_plus", 2, [X, Y, Z],
 	no, yes(Y - binop((-), var(Z), var(X)))).
-code_util__translate_builtin_2("int", "+", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "+", 0, [X, Y, Z],
 	no, yes(Z - binop((+), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "+", 10001, [X, Y, Z],
+code_util__translate_builtin_2("int", "+", 1, [X, Y, Z],
 	no, yes(X - binop((-), var(Z), var(Y)))).
-code_util__translate_builtin_2("int", "+", 10002, [X, Y, Z],
+code_util__translate_builtin_2("int", "+", 2, [X, Y, Z],
 	no, yes(Y - binop((-), var(Z), var(X)))).
-code_util__translate_builtin_2("int", "builtin_minus", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_minus", 0, [X, Y, Z],
 	no, yes(Z - binop((-), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_minus", 10001, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_minus", 1, [X, Y, Z],
 	no, yes(X - binop((+), var(Y), var(Z)))).
-code_util__translate_builtin_2("int", "builtin_minus", 10002, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_minus", 2, [X, Y, Z],
 	no, yes(Y - binop((-), var(X), var(Z)))).
-code_util__translate_builtin_2("int", "-", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "-", 0, [X, Y, Z],
 	no, yes(Z - binop((-), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "-", 10001, [X, Y, Z],
+code_util__translate_builtin_2("int", "-", 1, [X, Y, Z],
 	no, yes(X - binop((+), var(Y), var(Z)))).
-code_util__translate_builtin_2("int", "-", 10002, [X, Y, Z],
+code_util__translate_builtin_2("int", "-", 2, [X, Y, Z],
 	no, yes(Y - binop((-), var(X), var(Z)))).
-code_util__translate_builtin_2("int", "builtin_times", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_times", 0, [X, Y, Z],
 	no, yes(Z - binop((*), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_times", 10001, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_times", 1, [X, Y, Z],
 	no, yes(X - binop((/), var(Z), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_times", 10002, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_times", 2, [X, Y, Z],
 	no, yes(Y - binop((/), var(Z), var(X)))).
-code_util__translate_builtin_2("int", "*", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "*", 0, [X, Y, Z],
 	no, yes(Z - binop((*), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "*", 10001, [X, Y, Z],
+code_util__translate_builtin_2("int", "*", 1, [X, Y, Z],
 	no, yes(X - binop((/), var(Z), var(Y)))).
-code_util__translate_builtin_2("int", "*", 10002, [X, Y, Z],
+code_util__translate_builtin_2("int", "*", 2, [X, Y, Z],
 	no, yes(Y - binop((/), var(Z), var(X)))).
-code_util__translate_builtin_2("int", "builtin_div", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_div", 0, [X, Y, Z],
 	no, yes(Z - binop((/), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_div", 10001, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_div", 1, [X, Y, Z],
 	no, yes(X - binop((*), var(Y), var(Z)))).
-code_util__translate_builtin_2("int", "builtin_div", 10002, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_div", 2, [X, Y, Z],
 	no, yes(Y - binop((/), var(X), var(Z)))).
-code_util__translate_builtin_2("int", "//", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "//", 0, [X, Y, Z],
 	no, yes(Z - binop((/), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "//", 10001, [X, Y, Z],
+code_util__translate_builtin_2("int", "//", 1, [X, Y, Z],
 	no, yes(X - binop((*), var(Y), var(Z)))).
-code_util__translate_builtin_2("int", "//", 10002, [X, Y, Z],
+code_util__translate_builtin_2("int", "//", 2, [X, Y, Z],
 	no, yes(Y - binop((/), var(X), var(Z)))).
-code_util__translate_builtin_2("int", "builtin_mod", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_mod", 0, [X, Y, Z],
 	no, yes(Z - binop((mod), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "rem", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "rem", 0, [X, Y, Z],
 	no, yes(Z - binop((mod), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_left_shift", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_left_shift", 0, [X, Y, Z],
 	no, yes(Z - binop((<<), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "<<", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "unchecked_left_shift", 0, [X, Y, Z],
 	no, yes(Z - binop((<<), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_right_shift", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_right_shift", 0, [X, Y, Z],
 	no, yes(Z - binop((>>), var(X), var(Y)))).
-code_util__translate_builtin_2("int", ">>", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "unchecked_right_shift", 0, [X, Y, Z],
 	no, yes(Z - binop((>>), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_bit_and", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_bit_and", 0, [X, Y, Z],
 	no, yes(Z - binop((&), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "/\\", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "/\\", 0, [X, Y, Z],
 	no, yes(Z - binop((&), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_bit_or", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_bit_or", 0, [X, Y, Z],
 	no, yes(Z - binop(('|'), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "\\/", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "\\/", 0, [X, Y, Z],
 	no, yes(Z - binop(('|'), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_bit_xor", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_bit_xor", 0, [X, Y, Z],
 	no, yes(Z - binop((^), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "^", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "^", 0, [X, Y, Z],
 	no, yes(Z - binop((^), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_unary_plus", 10000, [X, Y],
+code_util__translate_builtin_2("int", "xor", 0, [X, Y, Z],
+	no, yes(Z - binop((^), var(X), var(Y)))).
+code_util__translate_builtin_2("int", "builtin_unary_plus", 0, [X, Y],
 	no, yes(Y - var(X))).
-code_util__translate_builtin_2("int", "+", 10000, [X, Y],
+code_util__translate_builtin_2("int", "+", 0, [X, Y],
 	no, yes(Y - var(X))).
-code_util__translate_builtin_2("int", "builtin_unary_minus", 10000, [X, Y],
+code_util__translate_builtin_2("int", "builtin_unary_minus", 0, [X, Y],
 	no, yes(Y - binop((-), const(int_const(0)), var(X)))).
-code_util__translate_builtin_2("int", "-", 10000, [X, Y],
+code_util__translate_builtin_2("int", "-", 0, [X, Y],
 	no, yes(Y - binop((-), const(int_const(0)), var(X)))).
-code_util__translate_builtin_2("int", "builtin_bit_neg", 10000, [X, Y],
+code_util__translate_builtin_2("int", "builtin_bit_neg", 0, [X, Y],
 	no, yes(Y - unop(bitwise_complement, var(X)))).
-code_util__translate_builtin_2("int", "\\", 10000, [X, Y],
+code_util__translate_builtin_2("int", "\\", 0, [X, Y],
 	no, yes(Y - unop(bitwise_complement, var(X)))).
 code_util__translate_builtin_2("int", ">", 0, [X, Y],
 	yes(binop((>), var(X), var(Y))), no).
@@ -450,57 +508,57 @@ code_util__translate_builtin_2("int", ">=", 0, [X, Y],
 code_util__translate_builtin_2("int", "=<", 0, [X, Y],
 	yes(binop((<=), var(X), var(Y))), no).
 
-code_util__translate_builtin_2("float", "builtin_float_plus", 10000, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_plus", 0, [X, Y, Z],
 	no, yes(Z - binop(float_plus, var(X), var(Y)))).
-code_util__translate_builtin_2("float", "builtin_float_plus", 10001, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_plus", 1, [X, Y, Z],
 	no, yes(X - binop(float_minus, var(Z), var(Y)))).
-code_util__translate_builtin_2("float", "builtin_float_plus", 10002, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_plus", 2, [X, Y, Z],
 	no, yes(Y - binop(float_minus, var(Z), var(X)))).
-code_util__translate_builtin_2("float", "+", 10000, [X, Y, Z],
+code_util__translate_builtin_2("float", "+", 0, [X, Y, Z],
 	no, yes(Z - binop(float_plus, var(X), var(Y)))).
-code_util__translate_builtin_2("float", "+", 10001, [X, Y, Z],
+code_util__translate_builtin_2("float", "+", 1, [X, Y, Z],
 	no, yes(X - binop(float_minus, var(Z), var(Y)))).
-code_util__translate_builtin_2("float", "+", 10002, [X, Y, Z],
+code_util__translate_builtin_2("float", "+", 2, [X, Y, Z],
 	no, yes(Y - binop(float_minus, var(Z), var(X)))).
-code_util__translate_builtin_2("float", "builtin_float_minus", 10000, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_minus", 0, [X, Y, Z],
 	no, yes(Z - binop(float_minus, var(X), var(Y)))).
-code_util__translate_builtin_2("float", "builtin_float_minus", 10001, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_minus", 1, [X, Y, Z],
 	no, yes(X - binop(float_plus, var(Y), var(Z)))).
-code_util__translate_builtin_2("float", "builtin_float_minus", 10002, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_minus", 2, [X, Y, Z],
 	no, yes(Y - binop(float_minus, var(X), var(Z)))).
-code_util__translate_builtin_2("float", "-", 10000, [X, Y, Z],
+code_util__translate_builtin_2("float", "-", 0, [X, Y, Z],
 	no, yes(Z - binop(float_minus, var(X), var(Y)))).
-code_util__translate_builtin_2("float", "-", 10001, [X, Y, Z],
+code_util__translate_builtin_2("float", "-", 1, [X, Y, Z],
 	no, yes(X - binop(float_plus, var(Y), var(Z)))).
-code_util__translate_builtin_2("float", "-", 10002, [X, Y, Z],
+code_util__translate_builtin_2("float", "-", 2, [X, Y, Z],
 	no, yes(Y - binop(float_minus, var(X), var(Z)))).
-code_util__translate_builtin_2("float", "builtin_float_times", 10000, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_times", 0, [X, Y, Z],
 	no, yes(Z - binop(float_times, var(X), var(Y)))).
-code_util__translate_builtin_2("float", "builtin_float_times", 10001, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_times", 1, [X, Y, Z],
 	no, yes(X - binop(float_divide, var(Z), var(Y)))).
-code_util__translate_builtin_2("float", "builtin_float_times", 10002, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_times", 2, [X, Y, Z],
 	no, yes(Y - binop(float_divide, var(Z), var(X)))).
-code_util__translate_builtin_2("float", "*", 10000, [X, Y, Z],
+code_util__translate_builtin_2("float", "*", 0, [X, Y, Z],
 	no, yes(Z - binop(float_times, var(X), var(Y)))).
-code_util__translate_builtin_2("float", "*", 10001, [X, Y, Z],
+code_util__translate_builtin_2("float", "*", 1, [X, Y, Z],
 	no, yes(X - binop(float_divide, var(Z), var(Y)))).
-code_util__translate_builtin_2("float", "*", 10002, [X, Y, Z],
+code_util__translate_builtin_2("float", "*", 2, [X, Y, Z],
 	no, yes(Y - binop(float_divide, var(Z), var(X)))).
-code_util__translate_builtin_2("float", "builtin_float_divide", 10000, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_divide", 0, [X, Y, Z],
 	no, yes(Z - binop(float_divide, var(X), var(Y)))).
-code_util__translate_builtin_2("float", "builtin_float_divide", 10001, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_divide", 1, [X, Y, Z],
 	no, yes(X - binop(float_times, var(Y), var(Z)))).
-code_util__translate_builtin_2("float", "builtin_float_divide", 10002, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_divide", 2, [X, Y, Z],
 	no, yes(Y - binop(float_divide, var(X), var(Z)))).
-code_util__translate_builtin_2("float", "/", 10000, [X, Y, Z],
+code_util__translate_builtin_2("float", "/", 0, [X, Y, Z],
 	no, yes(Z - binop(float_divide, var(X), var(Y)))).
-code_util__translate_builtin_2("float", "/", 10001, [X, Y, Z],
+code_util__translate_builtin_2("float", "/", 1, [X, Y, Z],
 	no, yes(X - binop(float_times, var(Y), var(Z)))).
-code_util__translate_builtin_2("float", "/", 10002, [X, Y, Z],
+code_util__translate_builtin_2("float", "/", 2, [X, Y, Z],
 	no, yes(Y - binop(float_divide, var(X), var(Z)))).
-code_util__translate_builtin_2("float", "+", 10000, [X, Y],
+code_util__translate_builtin_2("float", "+", 0, [X, Y],
 	no, yes(Y - var(X))).
-code_util__translate_builtin_2("float", "-", 10000, [X, Y],
+code_util__translate_builtin_2("float", "-", 0, [X, Y],
 	no, yes(Y - binop(float_minus, const(float_const(0.0)), var(X)))).
 code_util__translate_builtin_2("float", "builtin_float_gt", 0, [X, Y],
 	yes(binop(float_gt, var(X), var(Y))), no).
@@ -540,12 +598,13 @@ code_util__goal_may_allocate_heap(Goal - _GoalInfo) :-
 :- pred code_util__goal_may_allocate_heap_2(hlds_goal_expr).
 :- mode code_util__goal_may_allocate_heap_2(in) is semidet.
 
-code_util__goal_may_allocate_heap_2(higher_order_call(_, _, _, _, _, _)).
+code_util__goal_may_allocate_heap_2(generic_call(_, _, _, _)).
 code_util__goal_may_allocate_heap_2(call(_, _, _, Builtin, _, _)) :-
 	Builtin \= inline_builtin.
-code_util__goal_may_allocate_heap_2(unify(_, _, _, construct(_,_,Args,_), _)) :-
+code_util__goal_may_allocate_heap_2(
+		unify(_, _, _, construct(_,_,Args,_,_,_,_), _)) :-
 	Args = [_|_].
-code_util__goal_may_allocate_heap_2(some(_Vars, Goal)) :-
+code_util__goal_may_allocate_heap_2(some(_Vars, _, Goal)) :-
 	code_util__goal_may_allocate_heap(Goal).
 code_util__goal_may_allocate_heap_2(not(Goal)) :-
 	code_util__goal_may_allocate_heap(Goal).
@@ -641,9 +700,13 @@ code_util__cons_id_to_tag(int_const(X), _, _, int_constant(X)).
 code_util__cons_id_to_tag(float_const(X), _, _, float_constant(X)).
 code_util__cons_id_to_tag(string_const(X), _, _, string_constant(X)).
 code_util__cons_id_to_tag(code_addr_const(P,M), _, _, code_addr_constant(P,M)).
-code_util__cons_id_to_tag(pred_const(P,M), _, _, pred_closure_tag(P,M)).
-code_util__cons_id_to_tag(base_type_info_const(M,T,A), _, _,
-		base_type_info_constant(M,T,A)).
+code_util__cons_id_to_tag(pred_const(P,M,E), _, _, pred_closure_tag(P,M,E)).
+code_util__cons_id_to_tag(type_ctor_info_const(M,T,A), _, _,
+		type_ctor_info_constant(M,T,A)).
+code_util__cons_id_to_tag(base_typeclass_info_const(M,C,_,N), _, _,
+		base_typeclass_info_constant(M,C,N)).
+code_util__cons_id_to_tag(tabling_pointer_const(PredId,ProcId), _, _,
+		tabling_pointer_constant(PredId,ProcId)).
 code_util__cons_id_to_tag(cons(Name, Arity), Type, ModuleInfo, Tag) :-
 	(
 			% handle the `character' type specially
@@ -690,7 +753,7 @@ code_util__cannot_stack_flush(GoalExpr - _) :-
 :- mode code_util__cannot_stack_flush_2(in) is semidet.
 
 code_util__cannot_stack_flush_2(unify(_, _, _, Unify, _)) :-
-	Unify \= complicated_unify(_, _).
+	Unify \= complicated_unify(_, _, _).
 code_util__cannot_stack_flush_2(call(_, _, _, BuiltinState, _, _)) :-
 	BuiltinState = inline_builtin.
 code_util__cannot_stack_flush_2(conj(Goals)) :-
@@ -742,7 +805,7 @@ code_util__cannot_fail_before_stack_flush_conj([Goal | Goals]) :-
 			GoalExpr = call(_, _, _, BuiltinState, _, _),
 			BuiltinState \= inline_builtin
 		;
-			GoalExpr = higher_order_call(_, _, _, _, _, _)
+			GoalExpr = generic_call(_, _, _, _)
 		)
 	->
 		true
@@ -766,12 +829,13 @@ code_util__count_recursive_calls(Goal - _, PredId, ProcId, Min, Max) :-
 
 code_util__count_recursive_calls_2(not(Goal), PredId, ProcId, Min, Max) :-
 	code_util__count_recursive_calls(Goal, PredId, ProcId, Min, Max).
-code_util__count_recursive_calls_2(some(_, Goal), PredId, ProcId, Min, Max) :-
+code_util__count_recursive_calls_2(some(_, _, Goal),
+		PredId, ProcId, Min, Max) :-
 	code_util__count_recursive_calls(Goal, PredId, ProcId, Min, Max).
 code_util__count_recursive_calls_2(unify(_, _, _, _, _), _, _, 0, 0).
-code_util__count_recursive_calls_2(higher_order_call(_, _,_, _, _, _), _, _,
+code_util__count_recursive_calls_2(generic_call(_, _, _, _), _, _,
 		0, 0).
-code_util__count_recursive_calls_2(pragma_c_code(_,_,_,_, _, _, _, _), _, _,
+code_util__count_recursive_calls_2(pragma_c_code(_,_,_, _, _, _, _), _, _,
 		0, 0).
 code_util__count_recursive_calls_2(call(CallPredId, CallProcId, _, _, _, _),
 		PredId, ProcId, Count, Count) :-
@@ -786,6 +850,8 @@ code_util__count_recursive_calls_2(call(CallPredId, CallProcId, _, _, _, _),
 code_util__count_recursive_calls_2(conj(Goals), PredId, ProcId, Min, Max) :-
 	code_util__count_recursive_calls_conj(Goals, PredId, ProcId, 0, 0,
 		Min, Max).
+code_util__count_recursive_calls_2(par_conj(Goals, _), PredId, ProcId, Min, Max) :-
+	code_util__count_recursive_calls_conj(Goals, PredId, ProcId, 0, 0, Min, Max).
 code_util__count_recursive_calls_2(disj(Goals, _), PredId, ProcId, Min, Max) :-
 	code_util__count_recursive_calls_disj(Goals, PredId, ProcId, Min, Max).
 code_util__count_recursive_calls_2(switch(_, _, Cases, _), PredId, ProcId,
@@ -800,6 +866,10 @@ code_util__count_recursive_calls_2(if_then_else(_, Cond, Then, Else, _),
 	CTMax is CMax + TMax,
 	int__min(CTMin, EMin, Min),
 	int__max(CTMax, EMax, Max).
+code_util__count_recursive_calls_2(bi_implication(_, _),
+		_, _, _, _) :-
+	% these should have been expanded out by now
+	error("code_util__count_recursive_calls_2: unexpected bi_implication").
 
 :- pred code_util__count_recursive_calls_conj(list(hlds_goal),
 	pred_id, proc_id, int, int, int, int).
@@ -853,3 +923,71 @@ code_util__count_recursive_calls_cases([case(_, Goal) | Cases], PredId, ProcId,
 		int__min(Min0, Min1, Min),
 		int__max(Max0, Max1, Max)
 	).
+
+code_util__output_args([], LiveVals) :-
+	set__init(LiveVals).
+code_util__output_args([_V - arg_info(Loc, Mode) | Args], Vs) :-
+	code_util__output_args(Args, Vs0),
+	(
+		Mode = top_out
+	->
+		code_util__arg_loc_to_register(Loc, Reg),
+		set__insert(Vs0, Reg, Vs)
+	;
+		Vs = Vs0
+	).
+
+%-----------------------------------------------------------------------------%
+
+code_util__lvals_in_rval(lval(Lval), [Lval | Lvals]) :-
+	code_util__lvals_in_lval(Lval, Lvals).
+code_util__lvals_in_rval(var(_), []).
+code_util__lvals_in_rval(create(_, _, _, _, _, _), []).
+code_util__lvals_in_rval(mkword(_, Rval), Lvals) :-
+	code_util__lvals_in_rval(Rval, Lvals).
+code_util__lvals_in_rval(const(_), []).
+code_util__lvals_in_rval(unop(_, Rval), Lvals) :-
+	code_util__lvals_in_rval(Rval, Lvals).
+code_util__lvals_in_rval(binop(_, Rval1, Rval2), Lvals) :-
+	code_util__lvals_in_rval(Rval1, Lvals1),
+	code_util__lvals_in_rval(Rval2, Lvals2),
+	list__append(Lvals1, Lvals2, Lvals).
+code_util__lvals_in_rval(mem_addr(MemRef), Lvals) :-
+	code_util__lvals_in_mem_ref(MemRef, Lvals).
+
+code_util__lvals_in_lval(reg(_, _), []).
+code_util__lvals_in_lval(stackvar(_), []).
+code_util__lvals_in_lval(framevar(_), []).
+code_util__lvals_in_lval(succip, []).
+code_util__lvals_in_lval(maxfr, []).
+code_util__lvals_in_lval(curfr, []).
+code_util__lvals_in_lval(succip(Rval), Lvals) :-
+	code_util__lvals_in_rval(Rval, Lvals).
+code_util__lvals_in_lval(redofr(Rval), Lvals) :-
+	code_util__lvals_in_rval(Rval, Lvals).
+code_util__lvals_in_lval(redoip(Rval), Lvals) :-
+	code_util__lvals_in_rval(Rval, Lvals).
+code_util__lvals_in_lval(succfr(Rval), Lvals) :-
+	code_util__lvals_in_rval(Rval, Lvals).
+code_util__lvals_in_lval(prevfr(Rval), Lvals) :-
+	code_util__lvals_in_rval(Rval, Lvals).
+code_util__lvals_in_lval(hp, []).
+code_util__lvals_in_lval(sp, []).
+code_util__lvals_in_lval(field(_, Rval1, Rval2), Lvals) :-
+	code_util__lvals_in_rval(Rval1, Lvals1),
+	code_util__lvals_in_rval(Rval2, Lvals2),
+	list__append(Lvals1, Lvals2, Lvals).
+code_util__lvals_in_lval(lvar(_), []).
+code_util__lvals_in_lval(temp(_, _), []).
+code_util__lvals_in_lval(mem_ref(Rval), Lvals) :-
+	code_util__lvals_in_rval(Rval, Lvals).
+
+:- pred code_util__lvals_in_mem_ref(mem_ref, list(lval)).
+:- mode code_util__lvals_in_mem_ref(in, out) is det.
+
+code_util__lvals_in_mem_ref(stackvar_ref(_), []).
+code_util__lvals_in_mem_ref(framevar_ref(_), []).
+code_util__lvals_in_mem_ref(heap_ref(Rval, _, _), Lvals) :-
+	code_util__lvals_in_rval(Rval, Lvals).
+
+%-----------------------------------------------------------------------------%
