@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -29,7 +29,8 @@
 :- module common.
 :- interface.
 
-:- import_module hlds_pred, simplify.
+:- import_module hlds_pred, hlds_goal, prog_data, simplify.
+:- import_module list.
 
 	% If we find a deconstruction or a construction we cannot optimize,
 	% record the details of the memory cell in CommonInfo.
@@ -38,7 +39,7 @@
 	% we have seen before, replace the construction with an assignment
 	% from the variable unified with that cell.
 
-:- pred common__optimise_unification(unification, var, unify_rhs,
+:- pred common__optimise_unification(unification, prog_var, unify_rhs,
 	unify_mode, unify_context, hlds_goal_expr, hlds_goal_info,
 	hlds_goal_expr, hlds_goal_info, simplify_info, simplify_info).
 :- mode common__optimise_unification(in, in, in, in, in, in, in, 
@@ -50,11 +51,11 @@
 	% A call is replaceable if it has no uniquely moded outputs and no
 	% destructive inputs.
 
-:- pred common__optimise_call(pred_id, proc_id, list(var), hlds_goal_expr,
+:- pred common__optimise_call(pred_id, proc_id, list(prog_var), hlds_goal_expr,
 	hlds_goal_info, hlds_goal_expr, simplify_info, simplify_info).
 :- mode common__optimise_call(in, in, in, in, in, out, in, out) is det.
 
-:- pred common__optimise_higher_order_call(var, list(var), list(mode),
+:- pred common__optimise_higher_order_call(prog_var, list(prog_var), list(mode),
 		determinism, hlds_goal_expr, hlds_goal_info, hlds_goal_expr,
 		simplify_info, simplify_info).
 :- mode common__optimise_higher_order_call(in, in, in, in, in, in, out,
@@ -62,7 +63,7 @@
 
 	% succeeds if the two variables are equivalent
 	% according to the specified equivalence class.
-:- pred common__vars_are_equivalent(var, var, common_info).
+:- pred common__vars_are_equivalent(prog_var, prog_var, common_info).
 :- mode common__vars_are_equivalent(in, in, in) is semidet.
 
 	% Assorted stuff used here that simplify.m doesn't need to know about.
@@ -78,15 +79,16 @@
 
 :- implementation.
 
-:- import_module hlds_goal, hlds_data, quantification, mode_util, type_util.
+:- import_module quantification, mode_util, type_util, prog_util.
 :- import_module det_util, det_report, globals, options, inst_match, instmap.
-:- import_module prog_data, hlds_module, (inst).
-:- import_module bool, term, map, set, list, eqvclass, require, std_util.
+:- import_module hlds_data, hlds_module, (inst), pd_cost, term.
+:- import_module bool, map, set, eqvclass, require, std_util, string.
 
-:- type structure	--->	structure(var, type, cons_id, list(var)).
+:- type structure
+	--->	structure(prog_var, type, cons_id, list(prog_var)).
 
 :- type call_args
-	--->	call_args(term__context, list(var), list(var)).
+	--->	call_args(prog_context, list(prog_var), list(prog_var)).
 			% input, output args. For higher-order calls, 
 			% the closure is the first input argument.
 
@@ -95,7 +97,7 @@
 
 :- type common_info
 	--->	common(
-			eqvclass(var),
+			eqvclass(prog_var),
 			struct_map,	% all structs seen.
 			struct_map,	% structs seen since the last call.
 			seen_calls
@@ -118,55 +120,74 @@ common_info_clear_structs(common(VarEqv, StructMap, _, SeenCalls),
 		
 %---------------------------------------------------------------------------%
 
-common__optimise_unification(Unification0, Left0, _Right0, Mode, Context,
+common__optimise_unification(Unification0, _Left0, _Right0, Mode, _Context,
 		Goal0, GoalInfo0, Goal, GoalInfo, Info0, Info) :-
 	(
-		Unification0 = construct(Var, ConsId, ArgVars, _),
+		Unification0 = construct(Var, ConsId, ArgVars, _, _, _, _),
+		Mode = LVarMode - _,
+		simplify_info_get_module_info(Info0, ModuleInfo),
+		mode_get_insts(ModuleInfo, LVarMode, _, Inst),
 		(
+				% Don't optimise partially instantiated
+				% deconstruction unifications, because it's
+				% tricky to work out how to mode the
+				% replacement asssignment unifications.
+				% In the vast majority of cases, the
+				% variable is ground.
+			\+ inst_is_ground(ModuleInfo, Inst)
+		->
+			Goal = Goal0,
+			GoalInfo = GoalInfo0,
+			Info = Info0
+		;
+			% common__generate_assign assumes that the
+			% output variable is in the instmap_delta, which
+			% will not be true if the variable is a local.
+			% The optimization is pointless in that case.
+			goal_info_get_instmap_delta(GoalInfo0, InstMapDelta),
+			instmap_delta_search_var(InstMapDelta, Var, _),
 			common__find_matching_cell(Var, ConsId, ArgVars,
 				construction, Info0, OldStruct)
 		->
 			OldStruct = structure(OldVar, _, _, _),
-			Unification = assign(Var, OldVar),
-			Right = var(OldVar),
-			Goal = unify(Left0, Right, Mode, Unification, Context), 
-			common__record_equivalence(Var, OldVar, Info0, Info1),
-			simplify_info_set_requantify(Info1, Info),
-			set__list_to_set([OldVar, Var], NonLocals),
-			goal_info_set_nonlocals(GoalInfo0, NonLocals, GoalInfo)
+			UniMode = ((free - Inst) -> (Inst - Inst)),
+			common__generate_assign(Var, OldVar, UniMode,
+				GoalInfo0, Goal - GoalInfo, Info0, Info1),
+			simplify_info_set_requantify(Info1, Info2),
+			pd_cost__goal(Goal0 - GoalInfo0, Cost),
+			simplify_info_incr_cost_delta(Info2, Cost, Info)
 		;
 			Goal = Goal0,
 			GoalInfo = GoalInfo0,
 			common__record_cell(Var, ConsId, ArgVars, Info0, Info)
 		)
 	;
-		Unification0 = deconstruct(Var, ConsId, ArgVars, _, _),
+		Unification0 = deconstruct(Var, ConsId, ArgVars, UniModes, _),
+		simplify_info_get_module_info(Info0, ModuleInfo),
 		(
-			simplify_info_get_module_info(Info0, ModuleInfo),
+				% Don't optimise partially instantiated
+				% deconstruction unifications, because it's
+				% tricky to work out how to mode the
+				% replacement asssignment unifications.
+				% In the vast majority of cases, the
+				% variable is ground.
 			Mode = LVarMode - _,
-			mode_get_insts(ModuleInfo, LVarMode, Inst0, Inst1),
-				% Don't optimise away partially instantiated
-				% deconstruction unifications.
-			inst_matches_binding(Inst0, Inst1, ModuleInfo),
+			mode_get_insts(ModuleInfo, LVarMode, Inst0, _),
+			\+ inst_is_ground(ModuleInfo, Inst0)
+		->
+			Goal = Goal0,
+			Info = Info0
+		;
 			common__find_matching_cell(Var, ConsId, ArgVars,
 				deconstruction, Info0, OldStruct)
 		->
 			OldStruct = structure(_, _, _, OldArgVars),
-			goal_info_get_nonlocals(GoalInfo0, NonLocals),
-			simplify_info_get_instmap(Info0, InstMap),
-			goal_info_get_instmap_delta(GoalInfo0, InstMapDelta),
-			common__create_output_unifications(NonLocals, InstMap,
-				InstMapDelta, ArgVars, OldArgVars, Goals,
-				no, RecomputeAtomic),
-			Goal = conj(Goals),
-			common__record_equivalences(ArgVars,
-				OldArgVars, Info0, Info1),
+			common__create_output_unifications(GoalInfo0, ArgVars,
+				OldArgVars, UniModes, Goals, Info0, Info1),
 			simplify_info_set_requantify(Info1, Info2),
-			( RecomputeAtomic = yes ->
-				simplify_info_set_recompute_atomic(Info2, Info)
-			;
-				Info = Info2
-			)
+			Goal = conj(Goals),
+			pd_cost__goal(Goal0 - GoalInfo0, Cost),
+			simplify_info_incr_cost_delta(Info2, Cost, Info)
 		;
 			Goal = Goal0,
 			common__record_cell(Var, ConsId, ArgVars, Info0, Info)
@@ -183,7 +204,7 @@ common__optimise_unification(Unification0, Left0, _Right0, Mode, Context,
 		common__record_equivalence(Var1, Var2, Info0, Info),
 		GoalInfo = GoalInfo0
 	;
-		Unification0 = complicated_unify(_, _),
+		Unification0 = complicated_unify(_, _, _),
 		Goal = Goal0,
 		Info = Info0,
 		GoalInfo = GoalInfo0
@@ -195,8 +216,8 @@ common__optimise_unification(Unification0, Left0, _Right0, Mode, Context,
 	--->	deconstruction
 	;	construction.
 		
-:- pred common__find_matching_cell(var, cons_id,
-		list(var), unification_type, simplify_info, structure).
+:- pred common__find_matching_cell(prog_var, cons_id,
+		list(prog_var), unification_type, simplify_info, structure).
 :- mode common__find_matching_cell(in, in, in, in, in, out) is semidet.
 
 common__find_matching_cell(Var, ConsId, ArgVars, UniType, Info, OldStruct) :-
@@ -217,8 +238,9 @@ common__find_matching_cell(Var, ConsId, ArgVars, UniType, Info, OldStruct) :-
 	common__find_matching_cell_2(Structs, Var, ConsId, ArgVars, UniType,
 		VarEqv, VarTypes, OldStruct).
 
-:- pred common__find_matching_cell_2(list(structure), var, cons_id, list(var),
-	unification_type, eqvclass(var), map(var, type), structure).
+:- pred common__find_matching_cell_2(list(structure), prog_var, cons_id,
+	list(prog_var),
+	unification_type, eqvclass(prog_var), map(prog_var, type), structure).
 :- mode common__find_matching_cell_2(in, in, in, in, in,
 	in, in, out) is semidet.
 
@@ -276,7 +298,8 @@ common__compatible_types(Type1, Type2) :-
 
 	% succeeds if the two lists of variables are equivalent
 	% according to the specified equivalence class.
-:- pred common__var_lists_are_equiv(list(var), list(var), eqvclass(var)).
+:- pred common__var_lists_are_equiv(list(prog_var), list(prog_var),
+		eqvclass(prog_var)).
 :- mode common__var_lists_are_equiv(in, in, in) is semidet.
 
 common__var_lists_are_equiv([], [], _VarEqv).
@@ -290,7 +313,7 @@ common__vars_are_equivalent(X, Y, CommonInfo) :-
 
 	% succeeds if the two variables are equivalent
 	% according to the specified equivalence class.
-:- pred common__vars_are_equiv(var, var, eqvclass(var)).
+:- pred common__vars_are_equiv(prog_var, prog_var, eqvclass(prog_var)).
 :- mode common__vars_are_equiv(in, in, in) is semidet.
 
 common__vars_are_equiv(X, Y, VarEqv) :-
@@ -311,7 +334,7 @@ common__vars_are_equiv(X, Y, VarEqv) :-
 
 %---------------------------------------------------------------------------%
 
-:- pred common__record_cell(var, cons_id, list(var),
+:- pred common__record_cell(prog_var, cons_id, list(prog_var),
 		simplify_info, simplify_info).
 :- mode common__record_cell(in, in, in, in, out) is det.
 
@@ -355,32 +378,8 @@ common__do_record_cell(StructMap0, ConsId, Struct, StructMap) :-
 
 %---------------------------------------------------------------------------%
 
-:- pred common__record_equivalences(list(var), list(var),
-		simplify_info, simplify_info) is det.
-:- mode common__record_equivalences(in, in, in, out) is det.
-
-common__record_equivalences(Vars0, Vars, Info0, Info) :-
-	simplify_info_get_common_info(Info0, CommonInfo0),
-	CommonInfo0 = common(VarEqv0, StructMap0, StructMap1, SeenCalls),
-	common__record_equivalences_2(Vars0, Vars, VarEqv0, VarEqv),
-	CommonInfo = common(VarEqv, StructMap0, StructMap1, SeenCalls),
-	simplify_info_set_common_info(Info0, CommonInfo, Info).
-
-:- pred common__record_equivalences_2(list(var), list(var),
-		eqvclass(var), eqvclass(var)) is det.
-:- mode common__record_equivalences_2(in, in, in, out) is det.
-
-common__record_equivalences_2([], [_ | _], _, _) :-
-	error("common__record_equivalences").
-common__record_equivalences_2([_ | _], [], _, _) :-
-	error("common__record_equivalences").
-common__record_equivalences_2([], [], VarEqv, VarEqv).
-common__record_equivalences_2([Var1 | Vars1], [Var2 | Vars2],
-		VarEqv0, VarEqv) :-
-	eqvclass__ensure_equivalence(VarEqv0, Var1, Var2, VarEqv1),
-	common__record_equivalences_2(Vars1, Vars2, VarEqv1, VarEqv).
-
-:- pred common__record_equivalence(var, var, simplify_info, simplify_info).
+:- pred common__record_equivalence(prog_var, prog_var,
+		simplify_info, simplify_info).
 :- mode common__record_equivalence(in, in, in, out) is det.
 
 common__record_equivalence(Var1, Var2, Info0, Info) :-
@@ -408,10 +407,11 @@ common__optimise_call(PredId, ProcId, Args, Goal0,
 			ProcId, _, ProcInfo),
 		proc_info_argmodes(ProcInfo, ArgModes),
 	    	common__partition_call_args(ModuleInfo, ArgModes, Args,
-			InputArgs, OutputArgs)
+			InputArgs, OutputArgs, OutputModes)
 	->
 		common__optimise_call_2(seen_call(PredId, ProcId), InputArgs,
-			OutputArgs, Goal0, GoalInfo, Goal, Info0, Info)
+			OutputArgs, OutputModes, Goal0, GoalInfo, Goal,
+			Info0, Info)
 	;
 		Goal = Goal0,
 		Info = Info0
@@ -423,10 +423,10 @@ common__optimise_higher_order_call(Closure, Args, Modes, Det, Goal0,
 		common__check_call_detism(Det),
 		simplify_info_get_module_info(Info0, ModuleInfo),
 	    	common__partition_call_args(ModuleInfo, Modes, Args,
-			InputArgs, OutputArgs)
+			InputArgs, OutputArgs, OutputModes)
 	->
 		common__optimise_call_2(higher_order_call,
-			[Closure | InputArgs], OutputArgs, Goal0,
+			[Closure | InputArgs], OutputArgs, OutputModes, Goal0,
 			GoalInfo, Goal, Info0, Info)
 	;
 		Goal = Goal0,
@@ -443,12 +443,12 @@ common__check_call_detism(Det) :-
 	;	SolnCount = at_most_many_cc
 	).
 
-:- pred common__optimise_call_2(seen_call_id, list(var), list(var),
-		hlds_goal_expr, hlds_goal_info, hlds_goal_expr,
+:- pred common__optimise_call_2(seen_call_id, list(prog_var), list(prog_var),
+		list(mode), hlds_goal_expr, hlds_goal_info, hlds_goal_expr,
 		simplify_info, simplify_info).
-:- mode common__optimise_call_2(in, in, in, in, in, out, in, out) is det.
+:- mode common__optimise_call_2(in, in, in, in, in, in, out, in, out) is det.
 
-common__optimise_call_2(SeenCall, InputArgs, OutputArgs, Goal0,
+common__optimise_call_2(SeenCall, InputArgs, OutputArgs, Modes, Goal0,
 		GoalInfo, Goal, Info0, Info) :-
 	simplify_info_get_common_info(Info0, CommonInfo0),
 	CommonInfo0 = common(Eqv0, Structs0, Structs1, SeenCalls0),
@@ -458,40 +458,39 @@ common__optimise_call_2(SeenCall, InputArgs, OutputArgs, Goal0,
 		( common__find_previous_call(SeenCallsList0, InputArgs,
 			Eqv0, OutputArgs2, PrevContext)
 		->
-			simplify_info_get_instmap(Info0, InstMap0),
-			goal_info_get_nonlocals(GoalInfo, NonLocals),
-			goal_info_get_instmap_delta(GoalInfo, InstMapDelta),
-			common__create_output_unifications(NonLocals, InstMap0,
-			    InstMapDelta, OutputArgs, OutputArgs2, Goals,
-			    no, RecomputeAtomic),
+			simplify_info_get_module_info(Info0, ModuleInfo),
+			mode_util__modes_to_uni_modes(Modes, Modes, ModuleInfo,
+				UniModes), 
+			common__create_output_unifications(GoalInfo,
+			    OutputArgs, OutputArgs2, UniModes,
+			    Goals, Info0, Info1),
 			Goal = conj(Goals),
 			simplify_info_get_var_types(Info0, VarTypes),
 			(
-			    simplify_do_warn_calls(Info0),
+			    simplify_do_warn_calls(Info1),
 				% Don't warn for cases such as:
 				% set__init(Set1 : set(int)),
 				% set__init(Set2 : set(float)).
 			    map__apply_to_list(OutputArgs, VarTypes,
-					OutputArgTypes),
+					OutputArgTypes1),
 			    map__apply_to_list(OutputArgs2, VarTypes,
-					OutputArgTypes)
+					OutputArgTypes2),
+			    common__types_match_exactly_list(OutputArgTypes1,
+			    	OutputArgTypes2)
 			->
 			    goal_info_get_context(GoalInfo, Context),
-			    simplify_info_add_msg(Info0,
+			    simplify_info_do_add_msg(Info1,
 			    	duplicate_call(SeenCall, PrevContext,
 					Context),
-			        Info1)
+			        Info2)
 			;
-			    Info1 = Info0
+			    Info2 = Info1
 			),
 			CommonInfo = common(Eqv0, Structs0,
 				Structs1, SeenCalls0),
-			simplify_info_set_requantify(Info1, Info2),
-			( RecomputeAtomic = yes ->
-				simplify_info_set_recompute_atomic(Info2, Info3)
-			;
-				Info3 = Info2
-			)
+			pd_cost__goal(Goal0 - GoalInfo, Cost),
+			simplify_info_incr_cost_delta(Info2, Cost, Info3),
+			simplify_info_set_requantify(Info3, Info4)
 		;
 			goal_info_get_context(GoalInfo, Context),
 			ThisCall = call_args(Context, InputArgs, OutputArgs),
@@ -500,7 +499,7 @@ common__optimise_call_2(SeenCall, InputArgs, OutputArgs, Goal0,
 			CommonInfo = common(Eqv0, Structs0,
 				Structs1, SeenCalls),
 			Goal = Goal0,
-			Info3 = Info0
+			Info4 = Info0
 		)
 	;
 		goal_info_get_context(GoalInfo, Context),
@@ -508,46 +507,61 @@ common__optimise_call_2(SeenCall, InputArgs, OutputArgs, Goal0,
 		map__det_insert(SeenCalls0, SeenCall, [ThisCall], SeenCalls),
 		CommonInfo = common(Eqv0, Structs0, Structs1, SeenCalls),
 		Goal = Goal0,
-		Info3 = Info0
+		Info4 = Info0
 	),
-	simplify_info_set_common_info(Info3, CommonInfo, Info).
+	simplify_info_set_common_info(Info4, CommonInfo, Info).
 
 %---------------------------------------------------------------------------%
 
-	% Partition the arguments of a call into inputs and outputs, failing
-	% if any of the outputs have a unique component.
+	% Partition the arguments of a call into inputs and outputs,
+	% failing if any of the outputs have a unique component
+	% or if any of the outputs contain any `any' insts.
 :- pred common__partition_call_args(module_info::in, list(mode)::in,
-		list(var)::in, list(var)::out, list(var)::out) is semidet.
+		list(prog_var)::in, list(prog_var)::out,
+		list(prog_var)::out, list(mode)::out) is semidet.
 
-common__partition_call_args(_, [], [_ | _], _, _) :-
+common__partition_call_args(_, [], [_ | _], _, _, _) :-
 	error("common__partition_call_args").
-common__partition_call_args(_, [_ | _], [], _, _) :-
+common__partition_call_args(_, [_ | _], [], _, _, _) :-
 	error("common__partition_call_args").
-common__partition_call_args(_, [], [], [], []).
+common__partition_call_args(_, [], [], [], [], []).
 common__partition_call_args(ModuleInfo, [ArgMode | ArgModes], [Arg | Args],
-		InputArgs, OutputArgs) :-
+		InputArgs, OutputArgs, OutputModes) :-
 	common__partition_call_args(ModuleInfo, ArgModes, Args,
-		InputArgs1, OutputArgs1),
+		InputArgs1, OutputArgs1, OutputModes1),
 	mode_get_insts(ModuleInfo, ArgMode, InitialInst, FinalInst),
 	( inst_matches_binding(InitialInst, FinalInst, ModuleInfo) ->
 		InputArgs = [Arg | InputArgs1],
-		OutputArgs = OutputArgs1
+		OutputArgs = OutputArgs1,
+		OutputModes = OutputModes1
 	; 
 		% Calls with partly unique outputs cannot be replaced,
 		% since a unique copy of the outputs must be produced.
 		inst_is_not_partly_unique(ModuleInfo, FinalInst),
+
+		% Don't optimize calls whose outputs include any
+		% `any' insts, since that would create false aliasing
+		% between the different variables.
+		% (inst_matches_binding applied to identical insts
+		% fails only for `any' insts.)
+		inst_matches_binding(FinalInst, FinalInst, ModuleInfo),
+
 		% Don't optimize calls where a partially instantiated
-		% variable is further instantiated.
+		% variable is further instantiated. That case is difficult
+		% to test properly because mode analysis currently
+		% rejects most potential test cases.
 		inst_is_free(ModuleInfo, InitialInst),
+
 		InputArgs = InputArgs1,
-		OutputArgs = [Arg | OutputArgs1]
+		OutputArgs = [Arg | OutputArgs1],
+		OutputModes = [ArgMode | OutputModes1]
 	).
 
 %---------------------------------------------------------------------------%
 
-:- pred common__find_previous_call(list(call_args)::in, list(var)::in,
-		eqvclass(var)::in, list(var)::out,
-		term__context::out) is semidet.
+:- pred common__find_previous_call(list(call_args)::in, list(prog_var)::in,
+		eqvclass(prog_var)::in, list(prog_var)::out,
+		prog_context::out) is semidet.
 
 common__find_previous_call([SeenCall | SeenCalls], InputArgs,
 		Eqv, OutputArgs2, PrevContext) :-
@@ -561,59 +575,122 @@ common__find_previous_call([SeenCall | SeenCalls], InputArgs,
 
 %---------------------------------------------------------------------------%
 
-:- pred common__create_output_unifications(set(var)::in, instmap::in,
-		instmap_delta::in, list(var)::in, list(var)::in,
-		list(hlds_goal)::out, bool::in, bool::out) is det.
+:- pred common__create_output_unifications(hlds_goal_info::in, 
+		list(prog_var)::in, list(prog_var)::in, list(uni_mode)::in,
+		list(hlds_goal)::out, simplify_info::in,
+		simplify_info::out) is det.
 
-	% Create unifications to assign the non-local vars in OutputArgs from 
+	% Create unifications to assign the vars in OutputArgs from 
 	% the corresponding var in OutputArgs2.
-common__create_output_unifications(_, _, _, [], [], [],
-		RecomputeAtomic, RecomputeAtomic).
-common__create_output_unifications(_, _, _, [_ | _], [], _, _, _) :-
-	error("common__create_output_unifications").
-common__create_output_unifications(_, _, _, [], [_ | _], _, _, _) :-
-	error("common__create_output_unifications").
-common__create_output_unifications(NonLocals, InstMap, InstMapDelta, 
-		[OutputArg | OutputArgs], [OutputArg2 | OutputArgs2],
-		Goals, RecomputeAtomic0, RecomputeAtomic) :-
-	( 
-		set__member(OutputArg, NonLocals),
-		% This can happen if the first cell was created with a
-		% partially instantiated deconstruction.
-		OutputArg \= OutputArg2	
-	->
-		Unification = assign(OutputArg, OutputArg2),
-		instmap__lookup_var(InstMap, OutputArg2, Inst0),
-		% Check if OutputArg2 was local to the call or deconstruction
-		% that produced it. If it was, recompute_instmap_delta must 
-		% recompute the instmap_deltas for atomic goals.
-		( Inst0 = free ->
-			Inst = ground(shared, no),
-			RecomputeAtomic1 = yes
-		;	
-			Inst = Inst0,
-			RecomputeAtomic1 = RecomputeAtomic0
-		),	
-		% This will be fixed up in recompute_instmap_delta
-		UniMode = (free -> Inst) - (Inst -> Inst),
-		UnifyContext = unify_context(explicit, []),
-		Goal = unify(OutputArg, var(OutputArg2),
-			UniMode, Unification, UnifyContext),
-		set__list_to_set([OutputArg, OutputArg2], GoalNonLocals),
-		set__list_to_set([OutputArg], OutputSet),
-		instmap_delta_restrict(InstMapDelta, OutputSet,
-			GoalInstMapDelta),
-		goal_info_init(GoalNonLocals, GoalInstMapDelta, det, GoalInfo),
-		common__create_output_unifications(NonLocals, InstMap,
-			InstMapDelta, OutputArgs, OutputArgs2, Goals1,
-			RecomputeAtomic1, RecomputeAtomic),
-		Goals = [Goal - GoalInfo | Goals1]
+	% This needs to be done even if OutputArg is not a nonlocal in
+	% the original goal because later goals in the conjunction may
+	% match against the cell and need all the output arguments.
+	% The unneeded assignments will be removed later.
+
+common__create_output_unifications(GoalInfo, OutputArgs, OldOutputArgs,
+		UniModes, Goals, Info0, Info) :-
+	(
+		OutputArgs = [OutputArg | OutputArgs1],
+		OldOutputArgs = [OldOutputArg | OldOutputArgs1],
+		UniModes = [UniMode | UniModes1]
+	->	
+		( 
+			% This can happen if the first cell was created
+			% with a partially instantiated deconstruction.
+			OutputArg \= OldOutputArg
+		->
+			common__generate_assign(OutputArg, OldOutputArg,
+				UniMode, GoalInfo, Goal, Info0, Info1),
+			common__create_output_unifications(GoalInfo,
+				OutputArgs1, OldOutputArgs1, UniModes1,
+				Goals1, Info1, Info),
+			Goals = [Goal | Goals1]
+		;
+			common__create_output_unifications(GoalInfo,
+				OutputArgs1, OldOutputArgs1, UniModes1, Goals,
+				Info0, Info)
+		)
 	;
-		common__create_output_unifications(NonLocals, InstMap,
-			InstMapDelta, OutputArgs, OutputArgs2, Goals1,
-			RecomputeAtomic0, RecomputeAtomic),
-		Goals = Goals1
+		OutputArgs = [],
+		OldOutputArgs = [],
+		UniModes = []
+	->
+		Goals = [],
+		Info = Info0
+	;
+		error("comon__create_output_unifications: mode mismatch")
 	).
 
+
+%---------------------------------------------------------------------------%
+
+:- pred common__generate_assign(prog_var, prog_var, uni_mode,
+		hlds_goal_info, hlds_goal, simplify_info, simplify_info).
+:- mode common__generate_assign(in, in, in, in, out, in, out) is det.
+	
+common__generate_assign(ToVar, FromVar, UniMode,
+		GoalInfo0, Goal, Info0, Info) :-
+	goal_info_get_instmap_delta(GoalInfo0, InstMapDelta0),
+	simplify_info_get_var_types(Info0, VarTypes),
+	map__lookup(VarTypes, ToVar, ToVarType),
+	map__lookup(VarTypes, FromVar, FromVarType),
+
+	set__list_to_set([ToVar, FromVar], NonLocals),
+	( common__types_match_exactly(ToVarType, FromVarType) ->
+		UniMode = ((_ - ToVarInst0) -> (_ - ToVarInst)),
+		UnifyContext = unify_context(explicit, []),
+		UnifyMode = (ToVarInst0 -> ToVarInst) -
+				(ToVarInst -> ToVarInst),
+		GoalExpr = unify(ToVar, var(FromVar), UnifyMode,
+			assign(ToVar, FromVar), UnifyContext),
+		instmap_delta_from_assoc_list([ToVar - ToVarInst],
+			InstMapDelta)
+	;	
+		% If the cells we are optimizing don't have exactly the same
+		% type, we insert explicit type casts to ensure type
+		% correctness. This avoids problems with HLDS optimizations
+		% such as inlining which expect the HLDS to be well-typed.
+		% Unfortunately this loses information for other optimizations,
+		% since the call to the type cast hides the equivalence of
+		% the input and output.
+		simplify_info_get_module_info(Info0, ModuleInfo),
+		module_info_get_predicate_table(ModuleInfo, PredTable),
+		mercury_private_builtin_module(MercuryBuiltin),
+		TypeCast = qualified(MercuryBuiltin, "unsafe_type_cast"),
+		(
+			predicate_table_search_pred_sym_arity(
+				PredTable, TypeCast, 2, [PredId])
+		->
+			hlds_pred__initial_proc_id(ProcId),
+			GoalExpr = call(PredId, ProcId, [FromVar, ToVar],
+				inline_builtin, no, TypeCast)
+		;
+			error("common__generate_assign: \
+				can't find unsafe_type_cast")
+		),
+		instmap_delta_restrict(InstMapDelta0, NonLocals, InstMapDelta)
+	),
+	goal_info_init(NonLocals, InstMapDelta, det, GoalInfo),
+	Goal = GoalExpr - GoalInfo,	
+	common__record_equivalence(ToVar, FromVar, Info0, Info).
+
+:- pred common__types_match_exactly((type), (type)).
+:- mode common__types_match_exactly(in, in) is semidet.
+
+common__types_match_exactly(term__variable(Var), term__variable(Var)).
+common__types_match_exactly(Type1, Type2) :-
+	type_to_type_id(Type1, TypeId1, Args1),
+	type_to_type_id(Type2, TypeId2, Args2),
+	TypeId1 = TypeId2,
+	common__types_match_exactly_list(Args1, Args2).
+
+:- pred common__types_match_exactly_list(list(type), list(type)).
+:- mode common__types_match_exactly_list(in, in) is semidet.
+
+common__types_match_exactly_list([], []).
+common__types_match_exactly_list([Type1 | Types1], [Type2 | Types2]) :-
+	common__types_match_exactly(Type1, Type2),
+	common__types_match_exactly_list(Types1, Types2).
+
 %---------------------------------------------------------------------------%
 %---------------------------------------------------------------------------%
