@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -37,7 +37,7 @@
 :- import_module hlds_pred, hlds_goal, hlds_data.
 :- import_module mode_util, passes_aux, code_aux, prog_data, instmap.
 :- import_module delay_info, mode_info, inst_match, modes, mode_debug.
-:- import_module transform, options, globals.
+:- import_module transform, options, globals, varset, term.
 :- import_module mercury_to_mercury, hlds_out, dependency_graph.
 
 :- import_module bool, list, map, set, std_util, assoc_list, string.
@@ -67,7 +67,7 @@ constraint_propagation2([C | Cs], ModuleInfo0, ModuleInfo) -->
 constraint_propagation3([], ModuleInfo, ModuleInfo) --> [].
 constraint_propagation3([proc(Pred, Proc) | Rest], ModuleInfo0, ModuleInfo) -->
 	constraint__propagate_in_proc(Pred, Proc, ModuleInfo0, ModuleInfo1),
-	modecheck_proc(Proc, Pred, ModuleInfo1, ModuleInfo2, Errs),
+	modecheck_proc(Proc, Pred, ModuleInfo1, ModuleInfo2, Errs, _Changed),
 	( { Errs \= 0 } ->
 	    { error("constraint_propagation3") }
 	;
@@ -88,14 +88,14 @@ constraint__propagate_in_proc(PredId, ProcId, ModuleInfo0, ModuleInfo,
 	map__lookup(ProcTable0, ProcId, ProcInfo0),
 
 	proc_info_goal(ProcInfo0, Goal0),
-	proc_info_variables(ProcInfo0, VarSet0),
+	proc_info_varset(ProcInfo0, VarSet0),
 	varset__vars(VarSet0, VarList),
 	set__list_to_set(VarList, VarSet1),
 
 	proc_info_get_initial_instmap(ProcInfo0, ModuleInfo0, InstMap0),
 	proc_info_context(ProcInfo0, Context),
 	mode_info_init(IoState0, ModuleInfo0, PredId, ProcId,
-			Context, VarSet1, InstMap0, ModeInfo0),
+			Context, VarSet1, InstMap0, check_modes, ModeInfo0),
 
 	constraint__propagate_goal(Goal0, Goal, ModeInfo0, ModeInfo),
 
@@ -104,7 +104,7 @@ constraint__propagate_in_proc(PredId, ProcId, ModuleInfo0, ModuleInfo,
 	mode_info_get_var_types(ModeInfo, VarTypes),
 	mode_info_get_module_info(ModeInfo, ModuleInfo1),
 
-	proc_info_set_variables(ProcInfo0, VarSet, ProcInfo1),
+	proc_info_set_varset(ProcInfo0, VarSet, ProcInfo1),
 	proc_info_set_vartypes(ProcInfo1, VarTypes, ProcInfo2),
 	proc_info_set_goal(ProcInfo2, Goal, ProcInfo),
 
@@ -140,6 +140,9 @@ constraint__propagate_goal_2(conj(Goals0), conj(Goals)) -->
 	constraint__propagate_conj(Goals0, Goals),
 	mode_checkpoint(exit, "conj").
 
+constraint__propagate_goal_2(par_conj(_, _), par_conj(_, _)) -->
+	{ error("constraint__propagate_goal_2: par_conj not supported") }.
+
 constraint__propagate_goal_2(disj(Goals0, SM), disj(Goals, SM)) -->
 	mode_checkpoint(enter, "disj"),
 	constraint__propagate_disj(Goals0, Goals),
@@ -173,10 +176,10 @@ constraint__propagate_goal_2(some(Vars, Goal0), some(Vars, Goal)) -->
 	mode_checkpoint(exit, "some").
 
 constraint__propagate_goal_2(
-		higher_order_call(A, B, C, D, E, F),
-		higher_order_call(A, B, C, D, E, F)) -->
-	mode_checkpoint(enter, "higher-order call"),
-	mode_checkpoint(exit, "higher-order call").
+		generic_call(A, B, C, D),
+		generic_call(A, B, C, D)) -->
+	mode_checkpoint(enter, "generic call"),
+	mode_checkpoint(exit, "generic call").
 
 constraint__propagate_goal_2(
 		call(PredId, ProcId, ArgVars, Builtin, Sym, Context),
@@ -189,8 +192,8 @@ constraint__propagate_goal_2(unify(A,B,C,D,E), unify(A,B,C,D,E)) -->
 	mode_checkpoint(exit, "unify").
 
 constraint__propagate_goal_2(
-		pragma_c_code(A, B, C, D, E, F, G, H), 
-		pragma_c_code(A, B, C, D, E, F, G, H)) -->
+		pragma_c_code(A, B, C, D, E, F, G), 
+		pragma_c_code(A, B, C, D, E, F, G)) -->
 	mode_checkpoint(enter, "pragma_c_code"),
 	mode_checkpoint(exit, "pragma_c_code").
 
@@ -343,8 +346,8 @@ mode_info_write_goal(Goal, Indent, ModeInfo0, ModeInfo) :-
 	( semidet_succeed ->
 		mode_info_get_module_info(ModeInfo0, ModuleInfo),
 		mode_info_get_varset(ModeInfo0, VarSet),
-		hlds_out__write_goal(Goal, ModuleInfo, VarSet, no, Indent, "",
-				IOState1, IOState)
+		hlds_out__write_goal(Goal, ModuleInfo, VarSet, no, Indent,
+			"\n", IOState1, IOState)
 	;
 		IOState = IOState1
 	),
