@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -36,8 +36,9 @@
 :- import_module hlds_goal, hlds_data, options, globals, goal_util, hlds_out.
 :- import_module modes, mode_util, make_hlds, quantification, instmap.
 :- import_module prog_data, switch_detection, det_util, inst_match.
+:- import_module switch_detection, term, varset.
 
-:- import_module int, bool, list, map, set, std_util, require, term, varset.
+:- import_module int, bool, list, map, set, std_util, require.
 
 %-----------------------------------------------------------------------------%
 
@@ -91,7 +92,8 @@ detect_cse_in_proc(ProcId, PredId, ModuleInfo0, ModuleInfo) -->
 		;
 			[]
 		),
-		modecheck_proc(ProcId, PredId, ModuleInfo1, ModuleInfo2, Errs),
+		modecheck_proc(ProcId, PredId, ModuleInfo1,
+				ModuleInfo2, Errs, _Changed),
 		{ Errs > 0 ->
 			error("mode check fails when repeated")
 		;
@@ -117,7 +119,8 @@ detect_cse_in_proc(ProcId, PredId, ModuleInfo0, ModuleInfo) -->
 		detect_cse_in_proc(ProcId, PredId, ModuleInfo3, ModuleInfo)
 	).
 
-:- type cse_info	--->	cse_info(varset, map(var, type), module_info).
+:- type cse_info
+	--->	cse_info(prog_varset, map(prog_var, type), module_info).
 
 :- pred detect_cse_in_proc_2(proc_id, pred_id, bool, module_info, module_info).
 % :- mode detect_cse_in_proc_2(in, in, out, di, uo) is det.
@@ -135,7 +138,7 @@ detect_cse_in_proc_2(ProcId, PredId, Redo, ModuleInfo0, ModuleInfo) :-
 
 	proc_info_goal(ProcInfo0, Goal0),
 	proc_info_get_initial_instmap(ProcInfo0, ModuleInfo0, InstMap0),
-	proc_info_variables(ProcInfo0, Varset0),
+	proc_info_varset(ProcInfo0, Varset0),
 	proc_info_vartypes(ProcInfo0, VarTypes0),
 	CseInfo0 = cse_info(Varset0, VarTypes0, ModuleInfo0),
 	detect_cse_in_goal(Goal0, InstMap0, CseInfo0, CseInfo, Redo, Goal1),
@@ -154,7 +157,7 @@ detect_cse_in_proc_2(ProcId, PredId, Redo, ModuleInfo0, ModuleInfo) :-
 			VarTypes1, Goal, Varset, VarTypes, _Warnings),
 
 		proc_info_set_goal(ProcInfo0, Goal, ProcInfo1),
-		proc_info_set_variables(ProcInfo1, Varset, ProcInfo2),
+		proc_info_set_varset(ProcInfo1, Varset, ProcInfo2),
 		proc_info_set_vartypes(ProcInfo2, VarTypes, ProcInfo),
 
 		map__det_update(ProcTable0, ProcId, ProcInfo, ProcTable),
@@ -200,24 +203,28 @@ detect_cse_in_goal_1(Goal0 - GoalInfo, InstMap0, CseInfo0, CseInfo, Redo,
 	cse_info, cse_info, bool, hlds_goal_expr).
 :- mode detect_cse_in_goal_2(in, in, in, in, out, out, out) is det.
 
-detect_cse_in_goal_2(pragma_c_code(A,B,C,D,E,F,G,H), _, _, CseInfo, CseInfo,
-	no, pragma_c_code(A,B,C,D,E,F,G,H)).
+detect_cse_in_goal_2(pragma_c_code(A,B,C,D,E,F,G), _, _, CseInfo, CseInfo,
+	no, pragma_c_code(A,B,C,D,E,F,G)).
 
-detect_cse_in_goal_2(higher_order_call(A,B,C,D,E,F), _, _, CseInfo, CseInfo,
-	no, higher_order_call(A,B,C,D,E,F)).
+detect_cse_in_goal_2(generic_call(A,B,C,D), _, _, CseInfo, CseInfo,
+	no, generic_call(A,B,C,D)).
 
 detect_cse_in_goal_2(call(A,B,C,D,E,F), _, _, CseInfo, CseInfo, no,
 	call(A,B,C,D,E,F)).
 
 detect_cse_in_goal_2(unify(A,B0,C,D,E), _, InstMap0, CseInfo0, CseInfo, Redo,
 		unify(A,B,C,D,E)) :-
-	( B0 = lambda_goal(PredOrFunc, Vars, Modes, Det, Goal0) ->
+	( 
+		B0 = lambda_goal(PredOrFunc, EvalMethod, FixModes,
+			NonLocalVars, Vars, Modes, Det, Goal0)
+	->
 		CseInfo0 = cse_info(_, _, ModuleInfo),
 		instmap__pre_lambda_update(ModuleInfo, 
 			Vars, Modes, InstMap0, InstMap),
 		detect_cse_in_goal(Goal0, InstMap, CseInfo0, CseInfo, Redo,
 			Goal),
-		B = lambda_goal(PredOrFunc, Vars, Modes, Det, Goal)
+		B = lambda_goal(PredOrFunc, EvalMethod, FixModes,
+			NonLocalVars, Vars, Modes, Det, Goal)
 	;
 		B = B0,
 		CseInfo = CseInfo0,
@@ -228,14 +235,18 @@ detect_cse_in_goal_2(not(Goal0), _GoalInfo, InstMap, CseInfo0, CseInfo,
 		Redo, not(Goal)) :-
 	detect_cse_in_goal(Goal0, InstMap, CseInfo0, CseInfo, Redo, Goal).
 
-detect_cse_in_goal_2(some(Vars, Goal0), _GoalInfo, InstMap, CseInfo0, CseInfo,
-		Redo, some(Vars, Goal)) :-
+detect_cse_in_goal_2(some(Vars, CanRemove, Goal0), _GoalInfo, InstMap,
+		CseInfo0, CseInfo, Redo, some(Vars, CanRemove, Goal)) :-
 	detect_cse_in_goal(Goal0, InstMap, CseInfo0, CseInfo, Redo, Goal).
 
 detect_cse_in_goal_2(conj(Goals0), _GoalInfo, InstMap, CseInfo0, CseInfo,
 		Redo, conj(Goals)) :-
 	detect_cse_in_conj(Goals0, InstMap, CseInfo0, CseInfo, Redo, Goals).
 
+detect_cse_in_goal_2(par_conj(Goals0, SM), _, InstMap, CseInfo0, CseInfo, Redo,
+		par_conj(Goals, SM)) :-
+	detect_cse_in_par_conj(Goals0, InstMap, CseInfo0, CseInfo, Redo, Goals).
+
 detect_cse_in_goal_2(disj(Goals0, SM), GoalInfo, InstMap, CseInfo0, CseInfo,
 		Redo, Goal) :-
 	( Goals0 = [] ->
@@ -263,6 +274,10 @@ detect_cse_in_goal_2(if_then_else(Vars, Cond0, Then0, Else0, SM), GoalInfo,
 	detect_cse_in_ite(NonLocalsList, Vars, Cond0, Then0, Else0, GoalInfo,
 		SM, InstMap, CseInfo0, CseInfo, Redo, Goal).
 
+detect_cse_in_goal_2(bi_implication(_, _), _, _, _, _, _, _) :-
+	% these should have been expanded out by now
+	error("detect_cse_in_goal_2: unexpected bi_implication").
+
 %-----------------------------------------------------------------------------%
 
 :- pred detect_cse_in_conj(list(hlds_goal), instmap, cse_info, cse_info,
@@ -282,6 +297,20 @@ detect_cse_in_conj([Goal0 | Goals0], InstMap0, CseInfo0, CseInfo,
 	),
 	bool__or(Redo1, Redo2, Redo).
 
+%-----------------------------------------------------------------------------%
+
+:- pred detect_cse_in_par_conj(list(hlds_goal), instmap, cse_info, cse_info,
+	bool, list(hlds_goal)).
+:- mode detect_cse_in_par_conj(in, in, in, out, out, out) is det.
+
+detect_cse_in_par_conj([], _InstMap, CseInfo, CseInfo, no, []).
+detect_cse_in_par_conj([Goal0 | Goals0], InstMap0, CseInfo0, CseInfo,
+		Redo, [Goal | Goals]) :-
+	detect_cse_in_goal(Goal0, InstMap0, CseInfo0, CseInfo1, Redo1, Goal),
+	detect_cse_in_par_conj(Goals0, InstMap0, CseInfo1, CseInfo,
+		Redo2, Goals),
+	bool__or(Redo1, Redo2, Redo).
+
 %-----------------------------------------------------------------------------%
 
 	% These are the interesting bits - we've found a non-empty branched
@@ -289,7 +318,7 @@ detect_cse_in_conj([Goal0 | Goals0], InstMap0, CseInfo0, CseInfo,
 	% structure. Now for each non-local variable, we check whether each
 	% branch matches that variable against the same functor.
 
-:- pred detect_cse_in_disj(list(var), list(hlds_goal), hlds_goal_info,
+:- pred detect_cse_in_disj(list(prog_var), list(hlds_goal), hlds_goal_info,
 	store_map, instmap, cse_info, cse_info, bool, hlds_goal_expr).
 :- mode detect_cse_in_disj(in, in, in, in, in, in, out, out, out) is det.
 
@@ -327,7 +356,7 @@ detect_cse_in_disj_2([Goal0 | Goals0], InstMap0, CseInfo0, CseInfo, Redo,
 	detect_cse_in_disj_2(Goals0, InstMap0, CseInfo1, CseInfo, Redo2, Goals),
 	bool__or(Redo1, Redo2, Redo).
 
-:- pred detect_cse_in_cases(list(var), var, can_fail, list(case),
+:- pred detect_cse_in_cases(list(prog_var), prog_var, can_fail, list(case),
 	hlds_goal_info, store_map, instmap, cse_info, cse_info, bool,
 	hlds_goal_expr).
 :- mode detect_cse_in_cases(in, in, in, in, in, in, in, in, out, out, out)
@@ -372,7 +401,7 @@ detect_cse_in_cases_2([Case0 | Cases0], InstMap, CseInfo0, CseInfo, Redo,
 	detect_cse_in_cases_2(Cases0, InstMap, CseInfo1, CseInfo, Redo2, Cases),
 	bool__or(Redo1, Redo2, Redo).
 
-:- pred detect_cse_in_ite(list(var), list(var),
+:- pred detect_cse_in_ite(list(prog_var), list(prog_var),
 	hlds_goal, hlds_goal, hlds_goal, hlds_goal_info,
 	store_map, instmap, cse_info, cse_info, bool, hlds_goal_expr).
 :- mode detect_cse_in_ite(in, in, in, in, in, in, in, in, in, out, out, out)
@@ -432,7 +461,7 @@ detect_cse_in_ite_2(Cond0, Then0, Else0, InstMap0, CseInfo0, CseInfo, Redo,
 %	has been hoisted out, with the new variables as the functor arguments.
 %	Unify is the unification that was hoisted out.
 
-:- pred common_deconstruct(list(hlds_goal), var, cse_info, cse_info,
+:- pred common_deconstruct(list(hlds_goal), prog_var, cse_info, cse_info,
 	hlds_goal, list(hlds_goal)).
 :- mode common_deconstruct(in, in, in, out, out, out) is semidet.
 
@@ -441,26 +470,22 @@ common_deconstruct(Goals0, Var, CseInfo0, CseInfo, Unify, Goals) :-
 		Goals, MaybeUnifyGoal),
 	MaybeUnifyGoal = yes(Unify).
 
-:- pred common_deconstruct_2(list(hlds_goal), var, maybe(hlds_goal),
+:- pred common_deconstruct_2(list(hlds_goal), prog_var, maybe(hlds_goal),
 	cse_info, cse_info, list(hlds_goal), maybe(hlds_goal)).
 :- mode common_deconstruct_2(in, in, in, in, out, out, out) is semidet.
 
 common_deconstruct_2([], _Var, MaybeUnify, CseInfo, CseInfo, [], MaybeUnify).
 common_deconstruct_2([Goal0 | Goals0], Var, MaybeUnify0,
 		CseInfo0, CseInfo, [Goal | Goals], MaybeUnify) :-
-	goal_to_conj_list(Goal0, ConjList0),
-	Goal0 = _ - GoalInfo,
-	map__init(Substitution),
-	find_bind_var_for_cse(ConjList0, Substitution, Var, MaybeUnify0,
-		CseInfo0, CseInfo1, ConjList, _NewSubstitution, MaybeUnify1),
+	find_bind_var(Var, find_bind_var_for_cse_in_deconstruct, Goal0, Goal,
+		MaybeUnify0 - no, MaybeUnify1 - yes, CseInfo0, CseInfo1),
 	MaybeUnify1 = yes(_),
-	conj_list_to_goal(ConjList, GoalInfo, Goal),
 	common_deconstruct_2(Goals0, Var, MaybeUnify1, CseInfo1, CseInfo,
 		Goals, MaybeUnify).
 
 %-----------------------------------------------------------------------------%
 
-:- pred common_deconstruct_cases(list(case), var, cse_info, cse_info,
+:- pred common_deconstruct_cases(list(case), prog_var, cse_info, cse_info,
 	hlds_goal, list(case)).
 :- mode common_deconstruct_cases(in, in, in, out, out, out) is semidet.
 
@@ -470,7 +495,7 @@ common_deconstruct_cases(Cases0, Var, CseInfo0, CseInfo,
 		Cases, MaybeUnifyGoal),
 	MaybeUnifyGoal = yes(Unify).
 
-:- pred common_deconstruct_cases_2(list(case), var, maybe(hlds_goal),
+:- pred common_deconstruct_cases_2(list(case), prog_var, maybe(hlds_goal),
 	cse_info, cse_info, list(case), maybe(hlds_goal)).
 :- mode common_deconstruct_cases_2(in, in, in, in, out, out, out) is semidet.
 
@@ -478,128 +503,57 @@ common_deconstruct_cases_2([], _Var, MaybeUnify, CseInfo, CseInfo,
 	[], MaybeUnify).
 common_deconstruct_cases_2([case(ConsId, Goal0) | Cases0], Var, MaybeUnify0,
 		CseInfo0, CseInfo, [case(ConsId, Goal) | Cases], MaybeUnify) :-
-	goal_to_conj_list(Goal0, ConjList0),
-	Goal0 = _ - GoalInfo,
-	map__init(Substitution),
-	find_bind_var_for_cse(ConjList0, Substitution, Var, MaybeUnify0,
-		CseInfo0, CseInfo1, ConjList, _NewSubstitution, MaybeUnify1),
+	find_bind_var(Var, find_bind_var_for_cse_in_deconstruct, Goal0, Goal,
+		MaybeUnify0 - no, MaybeUnify1 - yes, CseInfo0, CseInfo1),
 	MaybeUnify1 = yes(_),
-	conj_list_to_goal(ConjList, GoalInfo, Goal),
 	common_deconstruct_cases_2(Cases0, Var, MaybeUnify1, CseInfo1, CseInfo,
 		Cases, MaybeUnify).
 
 %-----------------------------------------------------------------------------%
 
-	%	Searches through Goals0 looking for a deconstruction
-	%	unification with `Var'.
-	%
-	%	If MaybeUnify0 is no, a unification with any functor
-	%	is acceptable; if it is yes(Unify), only a unification
-	%	involving the same variable and function symbol is OK.
-	%
-	%	If we do find an acceptable deconstruction, we replace it
-	%	in the goal with pairwise equalities between the arguments
-	%	of the functor in that unification and the arguments of the
-	%	functor in Unify, where in Maybeunify = yes(Unify).
-	%	If MaybeUnify0 was no, we have to create the variables in Unify.
-	%
-	%	If we do not find an acceptable deconstruction, we set
-	%	MaybeUnify to no and set `Subst' to the substitution resulting
-	%	from interpreting through the goal.
+	% The hlds_goal is the common unification we are attemping to hoist.
+	% The boolean states whether such a deconstruction has been seen in
+	% this branch.
+:- type cse_result == pair(maybe(hlds_goal), bool).
 
-:- pred find_bind_var_for_cse(list(hlds_goal), substitution, var,
-	maybe(hlds_goal), cse_info, cse_info, list(hlds_goal), 
-	substitution, maybe(hlds_goal)).
-:- mode find_bind_var_for_cse(in, in, in, in, in, out, out, out, out) is det.
+:- pred find_bind_var_for_cse_in_deconstruct(prog_var, hlds_goal,
+		list(hlds_goal), cse_result, cse_result, cse_info, cse_info).
+:- mode find_bind_var_for_cse_in_deconstruct(in, in, out,
+	in, out, in, out) is det.
 
-find_bind_var_for_cse([], Substitution, _Var, _MaybeUnify0, CseInfo, CseInfo,
-	[], Substitution, no).
-find_bind_var_for_cse([GoalPair0 | Goals0], Substitution0, Var, MaybeUnify0,
-		CseInfo0, CseInfo, Goals, Substitution, MaybeUnify) :-
-	GoalPair0 = Goal0 - GoalInfo,
-	( Goal0 = conj(SubGoals0) ->
-		find_bind_var_for_cse(SubGoals0, Substitution0, Var,
-			MaybeUnify0, CseInfo0, CseInfo1,
-			SubGoals, Substitution1, MaybeUnify1),
-		Goal = conj(SubGoals),
-		( MaybeUnify1 = yes(_) ->
-			Goals = [Goal - GoalInfo | Goals0],
-			Substitution = Substitution1,
-			MaybeUnify = MaybeUnify1,
-			CseInfo = CseInfo1
-		;
-			find_bind_var_for_cse(Goals0, Substitution1, Var,
-				MaybeUnify0, CseInfo1, CseInfo,
-				Goals1, Substitution, MaybeUnify),
-			Goals = [Goal0 - GoalInfo | Goals1]
-		)
-	; Goal0 = unify(A, B, _, UnifyInfo0, _) ->
-		term__apply_rec_substitution(term__variable(Var),
-			Substitution0, Term),
-		(
-			Term = term__variable(Var1),
-			UnifyInfo0 = deconstruct(UnifyVar, _, _, _, _),
-			term__apply_rec_substitution(term__variable(UnifyVar),
-				Substitution0, term__variable(UnifyVar1)),
-			Var1 = UnifyVar1,
-			MaybeUnify0 = no
-		->
-			CseInfo0 = cse_info(Varset0, Typemap0, ModuleInfo),
-			construct_common_unify(Var, Goal0 - GoalInfo, Goal,
-				Varset0, Varset, Typemap0, Typemap,
-				Replacements),
-			CseInfo = cse_info(Varset, Typemap, ModuleInfo),
-			MaybeUnify = yes(Goal),
-			list__append(Replacements, Goals0, Goals),
-			Substitution = Substitution0
-		;
-			Term = term__variable(Var1),
-			UnifyInfo0 = deconstruct(UnifyVar, _, _, _, _),
-			term__apply_rec_substitution(term__variable(UnifyVar),
-				Substitution0, term__variable(UnifyVar1)),
-			Var1 = UnifyVar1,
-			UnifyInfo0 = deconstruct(_, _, _, _, _),
-			MaybeUnify0 = yes(OldUnifyGoal),
-			goal_info_get_context(GoalInfo, Context),
-			find_similar_deconstruct(OldUnifyGoal, UnifyInfo0,
-				Context, Replacements)
-		->
-			list__append(Replacements, Goals0, Goals),
-			Substitution = Substitution0,
-			CseInfo = CseInfo0,
-			MaybeUnify = MaybeUnify0
-		;
-		%
-		% if the variable was bound, but the deconstruction wasn't
-		% similar, then stop searching
-		%
-			Term = term__functor(_, _, _)
-		->
-			Goals = [Goal0 - GoalInfo | Goals0],
-			Substitution = Substitution0,
-			CseInfo = CseInfo0,
-			MaybeUnify = no
-		;
-			( interpret_unify(A, B, Substitution0, Substitution1) ->
-				Substitution2 = Substitution1
-			;
-				% the unification must fail - just ignore it
-				Substitution2 = Substitution0
-			),
-			find_bind_var_for_cse(Goals0, Substitution2, Var,
-				MaybeUnify0, CseInfo0, CseInfo,
-				Goals1, Substitution, MaybeUnify),
-			Goals = [Goal0 - GoalInfo | Goals1]
-		)
+find_bind_var_for_cse_in_deconstruct(Var, Goal0, Goals,
+		CseResult0, CseResult, CseInfo0, CseInfo) :-
+	CseResult0 = MaybeUnify0 - _,
+	(
+		MaybeUnify0 = no,
+		CseInfo0 = cse_info(Varset0, Typemap0, ModuleInfo),
+		construct_common_unify(Var, Goal0, Goal,
+			Varset0, Varset, Typemap0, Typemap, Goals),
+		CseInfo = cse_info(Varset, Typemap, ModuleInfo),
+		MaybeUnify = yes(Goal),
+		Seen = yes
 	;
-		Goals = [Goal0 - GoalInfo | Goals0],
-		Substitution = Substitution0,
+		MaybeUnify0 = yes(OldUnifyGoal),
 		CseInfo = CseInfo0,
-		MaybeUnify = no
-	).
+		Goal0 = _ - GoalInfo,
+		goal_info_get_context(GoalInfo, Context),
+		(
+			find_similar_deconstruct(OldUnifyGoal,
+				Goal0, Context, Goals0)
+		->
+			Goals = Goals0,
+			MaybeUnify = MaybeUnify0,
+			Seen = yes
+		;	
+			Goals = [Goal0],
+			MaybeUnify = no,
+			Seen = no
+		)
+	),
+	CseResult = MaybeUnify - Seen.
 
-:- pred construct_common_unify(var, hlds_goal, hlds_goal, varset, varset,
-	map(var, type), map(var, type), list(hlds_goal)).
+:- pred construct_common_unify(prog_var, hlds_goal, hlds_goal, prog_varset,
+	prog_varset, map(prog_var, type), map(prog_var, type), list(hlds_goal)).
 :- mode construct_common_unify(in, in, out, in, out, in, out, out) is det.
 
 construct_common_unify(Var, GoalExpr0 - GoalInfo, Goal, Varset0, Varset,
@@ -622,9 +576,9 @@ construct_common_unify(Var, GoalExpr0 - GoalInfo, Goal, Varset0, Varset,
 		error("unexpected goal in construct_common_unify")
 	).
 
-:- pred create_parallel_subterms(list(var), term__context, unify_context,
-	varset, varset, map(var, type), map(var, type), map(var, var),
-	list(hlds_goal)).
+:- pred create_parallel_subterms(list(prog_var), prog_context, unify_context,
+	prog_varset, prog_varset, map(prog_var, type), map(prog_var, type),
+	map(prog_var, prog_var), list(hlds_goal)).
 :- mode create_parallel_subterms(in, in, in, in, out, in, out, out, out) is det.
 
 create_parallel_subterms([], _, _, Varset, Varset, Typemap, Typemap, Sub, []) :-
@@ -644,14 +598,15 @@ create_parallel_subterms([OFV | OFV0], Context, UnifyContext, Varset0, Varset,
 
 %-----------------------------------------------------------------------------%
 
-:- pred find_similar_deconstruct(hlds_goal, unification, term__context,
-	list(hlds_goal)).
+:- pred find_similar_deconstruct(hlds_goal, hlds_goal, prog_context,
+		list(hlds_goal)).
 :- mode find_similar_deconstruct(in, in, in, out) is semidet.
 
-find_similar_deconstruct(OldUnifyGoal, NewUnifyInfo, Context, Replacements) :-
+find_similar_deconstruct(OldUnifyGoal, NewUnifyGoal, Context, Replacements) :-
 	(
 		OldUnifyGoal = unify(_OT1, _OT2, _OM, OldUnifyInfo, OC) - _,
 		OldUnifyInfo = deconstruct(_OV, OF, OFV, _OUM, _OCF),
+		NewUnifyGoal = unify(_NT1, _NT2, _NM, NewUnifyInfo, _NC) - _,
 		NewUnifyInfo = deconstruct(_NV, NF, NFV, _NUM, _NCF)
 	->
 		OF = NF,
@@ -663,8 +618,8 @@ find_similar_deconstruct(OldUnifyGoal, NewUnifyInfo, Context, Replacements) :-
 		error("find_similar_deconstruct: non-deconstruct unify")
 	).
 
-:- pred pair_subterms(list(var), list(var), term__context, unify_context,
-	list(hlds_goal)).
+:- pred pair_subterms(list(prog_var), list(prog_var), prog_context,
+		unify_context, list(hlds_goal)).
 :- mode pair_subterms(in, in, in, in, out) is det.
 
 pair_subterms(OFV0, NFV0, Context, UnifyContext, Replacements) :-
