@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1997 The University of Melbourne.
+% Copyright (C) 1996-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -18,7 +18,8 @@
 
 :- interface.
 
-:- import_module hlds_module, io.
+:- import_module prog_data, hlds_module, hlds_pred.
+:- import_module map, std_util, io.
 
 :- pred dead_proc_elim(module_info, module_info, io__state, io__state).
 :- mode dead_proc_elim(in, out, di, uo) is det.
@@ -38,18 +39,21 @@
 :- pred dead_pred_elim(module_info, module_info).
 :- mode dead_pred_elim(in, out) is det.
 
-:- type entity		--->	proc(pred_id, proc_id)
-			;	base_gen_info(string, string, int).
+:- type entity	
+	--->	proc(pred_id, proc_id)
+	;	base_gen_info(module_name, string, int).
 
-:- type needed_map ==	map(entity, maybe(int)).
+:- type needed_map == map(entity, maybe(int)).
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
 :- implementation.
-:- import_module hlds_pred, hlds_goal, hlds_data, prog_data, llds.
+
+:- import_module hlds_goal, hlds_data, prog_util, llds.
 :- import_module passes_aux, globals, options, code_util.
-:- import_module int, list, set, queue, map, bool, std_util, require.
+
+:- import_module int, string, list, set, queue, bool, require.
 
 %-----------------------------------------------------------------------------%
 
@@ -94,6 +98,8 @@ dead_proc_elim__analyze(ModuleInfo0, Needed) :-
 		Needed0, Needed).
 
 	% Add all exported entities to the queue and map.
+	% Note: changes here are likely to require changes to
+	% dead_pred_elim as well.
 
 :- pred dead_proc_elim__initialize(module_info, entity_queue, needed_map).
 :- mode dead_proc_elim__initialize(in, out, out) is det.
@@ -110,7 +116,11 @@ dead_proc_elim__initialize(ModuleInfo, Queue, Needed) :-
 		Queue1, Queue2, Needed1, Needed2),
 	module_info_base_gen_infos(ModuleInfo, BaseGenInfos),
 	dead_proc_elim__initialize_base_gen_infos(BaseGenInfos,
-		Queue2, Queue, Needed2, Needed).
+		Queue2, Queue3, Needed2, Needed3),
+	module_info_classes(ModuleInfo, Classes),
+	module_info_instances(ModuleInfo, Instances),
+	dead_proc_elim__initialize_class_methods(Classes, Instances,
+		Queue3, Queue, Needed3, Needed).
 
 	% Add all normally exported procedures within the listed predicates
 	% to the queue and map.
@@ -167,23 +177,22 @@ dead_proc_elim__initialize_base_gen_infos([], Queue, Queue, Needed, Needed).
 dead_proc_elim__initialize_base_gen_infos([BaseGenInfo | BaseGenInfos],
 		Queue0, Queue, Needed0, Needed) :-
 	BaseGenInfo = base_gen_info(_TypeId, ModuleName, TypeName,
-		Arity, _Status, _Elim, _Procs),
+		Arity, _Status, _Elim, _Procs, _HldsDefn),
 	(
 		% XXX: We'd like to do this, but there are problems.
-		% ( Status = exported
-		% ; Status = abstract_exported
-		% )
+		% status_is_exported(Status, yes)
+		%
 		% We need to do more thorough analysis of the
 		% reachability of the special predicates, in general,
-		% because using arg/3 allows us to get at base_type_info
-		% via the base_type_layout. The base_type_infos of
+		% because using arg/3 allows us to get at type_ctor_info
+		% via the type_ctor_layout. The type_ctor_infos of
 		% arguments of functors may have had their special preds
 		% eliminated, but they can still be called. In addition,
 		% it would be nice for pragma C code to have some
 		% support for using compiler generated data structures
 		% and preds, so that they aren't just eliminated.
 		%
-		% So presently, all base_type_infos will be treated
+		% So presently, all type_ctor_infos will be treated
 		% as exported, and hence no special preds will be
 		% eliminated.
 		semidet_succeed
@@ -198,6 +207,67 @@ dead_proc_elim__initialize_base_gen_infos([BaseGenInfo | BaseGenInfos],
 	dead_proc_elim__initialize_base_gen_infos(BaseGenInfos,
 		Queue1, Queue, Needed1, Needed).
 
+:- pred dead_proc_elim__initialize_class_methods(class_table, instance_table, 
+	entity_queue, entity_queue, needed_map, needed_map).
+:- mode dead_proc_elim__initialize_class_methods(in, in,
+	in, out, in, out) is det.
+
+dead_proc_elim__initialize_class_methods(Classes, Instances, Queue0, Queue, 
+		Needed0, Needed) :-
+	map__values(Instances, InstanceDefns0),
+	list__condense(InstanceDefns0, InstanceDefns),
+	list__foldl2(get_instance_pred_procs, InstanceDefns, Queue0, Queue1,
+		Needed0, Needed1),
+	map__values(Classes, ClassDefns),
+	list__foldl2(get_class_pred_procs, ClassDefns, Queue1, Queue,
+		Needed1, Needed).
+
+:- pred get_instance_pred_procs(hlds_instance_defn, entity_queue, entity_queue,
+	needed_map, needed_map).
+:- mode get_instance_pred_procs(in, in, out, in, out) is det.
+
+get_instance_pred_procs(Instance, Queue0, Queue, Needed0, Needed) :-
+	Instance = hlds_instance_defn(_, _, _, _, _, PredProcIds, _, _),
+
+	%
+	% We need to keep the instance methods for all instances
+	% for optimization of method lookups.
+	%
+	(
+			% This should never happen
+		PredProcIds = no,
+		Queue = Queue0,
+		Needed = Needed0
+	;
+		PredProcIds = yes(Ids),
+		get_class_interface_pred_procs(Ids, Queue0, Queue,
+			Needed0, Needed)
+	).
+
+:- pred get_class_pred_procs(hlds_class_defn, entity_queue, entity_queue,
+		needed_map, needed_map).
+:- mode get_class_pred_procs(in, in, out, in, out) is det.
+
+get_class_pred_procs(Class, Queue0, Queue, Needed0, Needed) :-
+	Class = hlds_class_defn(_, _, _, _, Methods, _, _),
+	get_class_interface_pred_procs(Methods,
+		Queue0, Queue, Needed0, Needed).
+
+:- pred get_class_interface_pred_procs(list(hlds_class_proc), 
+	entity_queue, entity_queue, needed_map, needed_map).
+:- mode get_class_interface_pred_procs(in, in, out, in, out) is det.
+
+get_class_interface_pred_procs(Ids, Queue0, Queue, Needed0, Needed) :-
+	AddHldsClassProc = lambda(
+		[PredProc::in, Q0::in, Q::out, N0::in, N::out] is det,
+		(
+			PredProc = hlds_class_proc(PredId, ProcId),
+			queue__put(Q0, proc(PredId, ProcId), Q),
+			map__set(N0, proc(PredId, ProcId), no, N)
+		)),
+	list__foldl2(AddHldsClassProc, Ids, Queue0, Queue, 
+		Needed0, Needed).
+
 %-----------------------------------------------------------------------------%
 
 :- pred dead_proc_elim__examine(entity_queue, examined_set, module_info,
@@ -234,8 +304,8 @@ dead_proc_elim__examine(Queue0, Examined0, ModuleInfo, Needed0, Needed) :-
 
 %-----------------------------------------------------------------------------%
 
-:- pred dead_proc_elim__examine_base_gen_info(string, string, int, module_info,
-	entity_queue, entity_queue, needed_map, needed_map).
+:- pred dead_proc_elim__examine_base_gen_info(module_name, string, arity,
+	module_info, entity_queue, entity_queue, needed_map, needed_map).
 :- mode dead_proc_elim__examine_base_gen_info(in, in, in, in, in, out, in, out)
 	is det.
 
@@ -253,7 +323,7 @@ dead_proc_elim__examine_base_gen_info(ModuleName, TypeName, Arity, ModuleInfo,
 		Needed = Needed0
 	).
 
-:- pred dead_proc_elim__find_base_gen_info(string, string, int,
+:- pred dead_proc_elim__find_base_gen_info(module_name, string, arity,
 	list(base_gen_info), list(pred_proc_id)).
 :- mode dead_proc_elim__find_base_gen_info(in, in, in, in, out) is semidet.
 
@@ -261,7 +331,7 @@ dead_proc_elim__find_base_gen_info(ModuleName, TypeName, TypeArity,
 		[BaseGenInfo | BaseGenInfos], Refs) :-
 	(
 		BaseGenInfo = base_gen_info(_TypeId, ModuleName, TypeName,
-			TypeArity, _Status, _Elim, Refs0)
+			TypeArity, _Status, _Elim, Refs0, _HldsDefn)
 	->
 		Refs = Refs0
 	;
@@ -350,11 +420,15 @@ dead_proc_elim__examine_expr(conj(Goals), CurrProc, Queue0, Queue,
 		Needed0, Needed) :-
 	dead_proc_elim__examine_goals(Goals, CurrProc, Queue0, Queue,
 		Needed0, Needed).
+dead_proc_elim__examine_expr(par_conj(Goals, _SM), CurrProc, Queue0, Queue,
+		Needed0, Needed) :-
+	dead_proc_elim__examine_goals(Goals, CurrProc, Queue0, Queue,
+		Needed0, Needed).
 dead_proc_elim__examine_expr(not(Goal), CurrProc, Queue0, Queue,
 		Needed0, Needed) :-
 	dead_proc_elim__examine_goal(Goal, CurrProc, Queue0, Queue,
 		Needed0, Needed).
-dead_proc_elim__examine_expr(some(_, Goal), CurrProc, Queue0, Queue,
+dead_proc_elim__examine_expr(some(_, _, Goal), CurrProc, Queue0, Queue,
 		Needed0, Needed) :-
 	dead_proc_elim__examine_goal(Goal, CurrProc, Queue0, Queue,
 		Needed0, Needed).
@@ -370,7 +444,7 @@ dead_proc_elim__examine_expr(if_then_else(_, Cond, Then, Else, _),
 		Needed1, Needed2),
 	dead_proc_elim__examine_goal(Else, CurrProc, Queue2, Queue,
 		Needed2, Needed).
-dead_proc_elim__examine_expr(higher_order_call(_,_,_,_,_,_), _,
+dead_proc_elim__examine_expr(generic_call(_,_,_,_), _,
 		Queue, Queue, Needed, Needed).
 dead_proc_elim__examine_expr(call(PredId, ProcId, _,_,_,_),
 		CurrProc, Queue0, Queue, Needed0, Needed) :-
@@ -395,22 +469,22 @@ dead_proc_elim__examine_expr(call(PredId, ProcId, _,_,_,_),
 		NewNotation = yes(1),
 		map__set(Needed0, proc(PredId, ProcId), NewNotation, Needed)
 	).
-dead_proc_elim__examine_expr(pragma_c_code(_, _, PredId, ProcId, _, _, _, _),
+dead_proc_elim__examine_expr(pragma_c_code(_, PredId, ProcId, _, _, _, _),
 		_CurrProc, Queue0, Queue, Needed0, Needed) :-
 	queue__put(Queue0, proc(PredId, ProcId), Queue),
 	map__set(Needed0, proc(PredId, ProcId), no, Needed).
 dead_proc_elim__examine_expr(unify(_,_,_, Uni, _), _CurrProc, Queue0, Queue,
 		Needed0, Needed) :-
 	(
-		Uni = construct(_, ConsId, _, _),
+		Uni = construct(_, ConsId, _, _, _, _, _),
 		(
-			ConsId = pred_const(PredId, ProcId),
+			ConsId = pred_const(PredId, ProcId, _),
 			Entity = proc(PredId, ProcId)
 		;
 			ConsId = code_addr_const(PredId, ProcId),
 			Entity = proc(PredId, ProcId)
 		;
-			ConsId = base_type_info_const(Module, TypeName, Arity),
+			ConsId = type_ctor_info_const(Module, TypeName, Arity),
 			Entity = base_gen_info(Module, TypeName, Arity)
 		)
 	->
@@ -420,6 +494,9 @@ dead_proc_elim__examine_expr(unify(_,_,_, Uni, _), _CurrProc, Queue0, Queue,
 		Queue = Queue0,
 		Needed = Needed0
 	).
+dead_proc_elim__examine_expr(bi_implication(_,_), _, _, _, _, _) :-
+	% these should have been expanded out by now
+	error("detect_cse_in_goal_2: unexpected bi_implication").
 
 %-----------------------------------------------------------------------------%
 
@@ -501,7 +578,8 @@ dead_proc_elim__eliminate_pred(PredId, ElimInfo0, ElimInfo, State0, State) :-
 			)),
 		list__foldl(DestroyGoal, ProcIds, ProcTable0, ProcTable),
 		pred_info_set_procedures(PredInfo0, ProcTable, PredInfo1),
-		pred_info_set_import_status(PredInfo1, imported, PredInfo),
+		pred_info_set_import_status(PredInfo1, imported(interface),
+				PredInfo),
 		map__det_update(PredTable0, PredId, PredInfo, PredTable),
 		globals__io_lookup_bool_option(very_verbose, VeryVerbose,
 			State0, State1),
@@ -559,7 +637,7 @@ dead_proc_elim__eliminate_base_gen_infos([BaseGenInfo0 | BaseGenInfos0], Needed,
 	dead_proc_elim__eliminate_base_gen_infos(BaseGenInfos0, Needed,	
 		BaseGenInfos1),
 	BaseGenInfo0 = base_gen_info(TypeId, ModuleName, TypeName,
-		Arity, Status, Elim0, Procs),
+		Arity, Status, Elim0, Procs, HldsDefn),
 	(
 		Entity = base_gen_info(ModuleName, TypeName, Arity),
 		map__search(Needed, Entity, _)
@@ -578,7 +656,8 @@ dead_proc_elim__eliminate_base_gen_infos([BaseGenInfo0 | BaseGenInfos0], Needed,
 			NumProcs = ProcsLength
 		),
 		NeuteredBaseGenInfo = base_gen_info(TypeId, ModuleName, 
-			TypeName, Arity, Status, yes(NumProcs), []),
+			TypeName, Arity, Status, yes(NumProcs), [],
+			HldsDefn),
 		BaseGenInfos = [NeuteredBaseGenInfo | BaseGenInfos1]
 	).
 
@@ -595,21 +674,75 @@ dead_proc_elim__eliminate_base_gen_infos([BaseGenInfo0 | BaseGenInfos0], Needed,
 		).
 
 dead_pred_elim(ModuleInfo0, ModuleInfo) :-
-	module_info_predids(ModuleInfo0, PredIds),
+
 	queue__init(Queue0),
+	map__init(Needed0),
+	module_info_get_pragma_exported_procs(ModuleInfo0, PragmaExports),
+	dead_proc_elim__initialize_pragma_exports(PragmaExports,
+		Queue0, _, Needed0, Needed1),
+	%
+	% The goals for the class method procs need to be
+	% examined because they contain calls to the actual method
+	% implementations.
+	%
+	module_info_instances(ModuleInfo0, Instances),
+	module_info_classes(ModuleInfo0, Classes),
+	dead_proc_elim__initialize_class_methods(Classes, Instances,
+		Queue0, _, Needed1, Needed),
+	map__keys(Needed, Entities),
+	queue__init(Queue1),
+	set__init(NeededPreds0),
+	list__foldl2(dead_pred_elim_add_entity, Entities,
+		Queue1, Queue, NeededPreds0, NeededPreds1),
+
 	set__init(Preds0),
 	set__init(Names0),
-	DeadInfo0 = dead_pred_info(ModuleInfo0, Queue0, 
-		Preds0, Preds0, Names0),
+	DeadInfo0 = dead_pred_info(ModuleInfo0, Queue, 
+		Preds0, NeededPreds1, Names0),
+
+	module_info_predids(ModuleInfo0, PredIds),
 	list__foldl(dead_pred_elim_initialize, PredIds, 
 		DeadInfo0, DeadInfo1),
 	dead_pred_elim_analyze(DeadInfo1, DeadInfo),
-	DeadInfo = dead_pred_info(ModuleInfo1, _, _, NeededPreds, _),
+	DeadInfo = dead_pred_info(ModuleInfo1, _, _, NeededPreds2, _),
+
+	%
+	% If a predicate is not needed, predicates which were added in
+	% make_hlds.m to force type specialization are also not needed.
+	% Here we add in those which are needed.
+	%
+	module_info_type_spec_info(ModuleInfo1,
+		type_spec_info(TypeSpecProcs0, TypeSpecForcePreds0,
+			SpecMap0, PragmaMap0)),
+	set__to_sorted_list(NeededPreds2, NeededPredList2),
+	list__foldl(
+	    lambda([NeededPred::in, AllPreds0::in, AllPreds::out] is det, (
+		( map__search(SpecMap0, NeededPred, NewNeededPreds) ->
+			set__insert_list(AllPreds0, NewNeededPreds, AllPreds)
+		;
+			AllPreds = AllPreds0
+		)
+	)), NeededPredList2, NeededPreds2, NeededPreds),
+	set__intersect(TypeSpecForcePreds0, NeededPreds, TypeSpecForcePreds),
+
+	module_info_set_type_spec_info(ModuleInfo1,
+		type_spec_info(TypeSpecProcs0, TypeSpecForcePreds, 
+			SpecMap0, PragmaMap0),
+		ModuleInfo2),
+
 	set__list_to_set(PredIds, PredIdSet),
 	set__difference(PredIdSet, NeededPreds, DeadPreds),
 	set__to_sorted_list(DeadPreds, DeadPredList),
 	list__foldl(module_info_remove_predicate, DeadPredList,
-		ModuleInfo1, ModuleInfo).
+		ModuleInfo2, ModuleInfo).
+
+:- pred dead_pred_elim_add_entity(entity::in, queue(pred_id)::in,
+	queue(pred_id)::out, set(pred_id)::in, set(pred_id)::out) is det.
+
+dead_pred_elim_add_entity(base_gen_info(_, _, _), Q, Q, Preds, Preds).
+dead_pred_elim_add_entity(proc(PredId, _), Q0, Q, Preds0, Preds) :-
+	queue__put(Q0, PredId, Q),
+	set__insert(Preds0, PredId, Preds).
 
 :- pred dead_pred_elim_initialize(pred_id::in, dead_pred_info::in,
 		dead_pred_info::out) is det.
@@ -619,24 +752,37 @@ dead_pred_elim_initialize(PredId, DeadInfo0, DeadInfo) :-
 	module_info_pred_info(ModuleInfo, PredId, PredInfo),
 	( 
 		pred_info_module(PredInfo, PredModule),
+		pred_info_name(PredInfo, PredName),
+		pred_info_arity(PredInfo, PredArity),
 		(
 			% Don't eliminate special preds since they won't
 			% be actually called from the HLDS until after 
 			% polymorphism.
 			code_util__compiler_generated(PredInfo)
 		;
-			% Don't eliminate preds from mercury_builtin.m since
-			% polymorphism.m needs unify/2 and friends.
-			PredModule = "mercury_builtin"
+			% Don't eliminate preds from builtin.m or
+			% private_builtin.m, since polymorphism.m
+			% needs unify/2 and friends.
+			mercury_public_builtin_module(PredModule)
+		;
+			mercury_private_builtin_module(PredModule)
 		;
 			% Don't attempt to eliminate local preds here, since we
 			% want to do semantic checking on those even if they 
 			% aren't used.
 			\+ pred_info_is_imported(PredInfo), 
 			\+ pred_info_import_status(PredInfo, opt_imported)
+		;
+			% Don't eliminate <foo>_init_any/1 predicates;
+			% modes.m may insert calls to them to initialize
+			% variables from inst `free' to inst `any'.
+			string__remove_suffix(PredName, "_init_any", _),
+			PredArity = 1
+		;
+			% Don't eliminate the clauses for assertions.
+			pred_info_get_goal_type(PredInfo, assertion)
 		)
 	->
-		pred_info_name(PredInfo, PredName),
 		set__insert(NeededNames0, qualified(PredModule, PredName), 
 			NeededNames),
 		queue__put(Q0, PredId, Q)
@@ -662,7 +808,7 @@ dead_pred_elim_analyze(DeadInfo0, DeadInfo) :-
 				Needed, NeededNames),
 			module_info_pred_info(ModuleInfo, PredId, PredInfo),
 			pred_info_clauses_info(PredInfo, ClausesInfo),
-			ClausesInfo = clauses_info(_,_,_,_, Clauses),
+			clauses_info_clauses(ClausesInfo, Clauses),
 			list__foldl(dead_pred_elim_process_clause, Clauses,
 				DeadInfo1, DeadInfo2)
 		),
@@ -682,6 +828,8 @@ dead_pred_elim_process_clause(clause(_, Goal, _)) -->
 
 pre_modecheck_examine_goal(conj(Goals) - _) -->
 	list__foldl(pre_modecheck_examine_goal, Goals).
+pre_modecheck_examine_goal(par_conj(Goals, _) - _) -->
+	list__foldl(pre_modecheck_examine_goal, Goals).
 pre_modecheck_examine_goal(disj(Goals, _) - _) -->
 	list__foldl(pre_modecheck_examine_goal, Goals).
 pre_modecheck_examine_goal(if_then_else(_, If, Then, Else, _) - _) -->
@@ -692,16 +840,19 @@ pre_modecheck_examine_goal(switch(_, _, Cases, _) - _) -->
 		pre_modecheck_examine_goal(Goal, Info0, Info)
 	)) },
 	list__foldl(ExamineCase, Cases).
-pre_modecheck_examine_goal(higher_order_call(_,_,_,_,_,_) - _) --> [].
+pre_modecheck_examine_goal(generic_call(_,_,_,_) - _) --> [].
 pre_modecheck_examine_goal(not(Goal) - _) -->
 	pre_modecheck_examine_goal(Goal).
-pre_modecheck_examine_goal(some(_, Goal) - _) -->
+pre_modecheck_examine_goal(some(_, _, Goal) - _) -->
 	pre_modecheck_examine_goal(Goal).
 pre_modecheck_examine_goal(call(_, _, _, _, _, PredName) - _) -->
 	dead_pred_info_add_pred_name(PredName).
-pre_modecheck_examine_goal(pragma_c_code(_, _, _, _, _, _, _, _) - _) --> [].
+pre_modecheck_examine_goal(pragma_c_code(_, _, _, _, _, _, _) - _) --> [].
 pre_modecheck_examine_goal(unify(_, Rhs, _, _, _) - _) -->
 	pre_modecheck_examine_unify_rhs(Rhs).
+pre_modecheck_examine_goal(bi_implication(_, _) - _) -->
+	% these should have been expanded out by now
+	{ error("pre_modecheck_examine_goal: unexpected bi_implication") }.
 
 :- pred pre_modecheck_examine_unify_rhs(unify_rhs::in, 
 		dead_pred_info::in, dead_pred_info::out) is det.
@@ -713,7 +864,7 @@ pre_modecheck_examine_unify_rhs(functor(Functor, _)) -->
 	;
 		[]
 	).
-pre_modecheck_examine_unify_rhs(lambda_goal(_, _, _, _, Goal)) -->
+pre_modecheck_examine_unify_rhs(lambda_goal(_, _, _, _, _, _, _, Goal)) -->
 	pre_modecheck_examine_goal(Goal).
 
 :- pred dead_pred_info_add_pred_name(sym_name::in, dead_pred_info::in, 
