@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -52,7 +52,10 @@
 
 :- interface.
 
-:- import_module hlds_module, hlds_pred, hlds_data, det_report, io.
+:- import_module prog_data.
+:- import_module hlds_goal, hlds_module, hlds_pred, hlds_data, instmap.
+:- import_module det_report, det_util, globals.
+:- import_module list, std_util, io.
 
 	% Perform determinism inference for local predicates with no
 	% determinism declarations, and determinism checking for all other
@@ -73,6 +76,22 @@
 	determinism, determinism, list(det_msg)).
 :- mode det_infer_proc(in, in, in, out, in, out, out, out) is det.
 
+	% Infers the determinism of `Goal0' and returns this in `Detism'.
+	% It annotates the goal and all its subgoals with their determinism
+	% and returns the annotated goal in `Goal'.
+
+:- pred det_infer_goal(hlds_goal, instmap, soln_context, det_info,
+	hlds_goal, determinism, list(det_msg)).
+:- mode det_infer_goal(in, in, in, in, out, out, out) is det.
+
+	% Work out how many solutions are needed for a given determinism.
+:- pred det_get_soln_context(determinism, soln_context).
+:- mode det_get_soln_context(in, out) is det.
+
+:- type soln_context
+	--->	all_solns
+	;	first_soln.
+
 	% The tables for computing the determinism of compound goals
 	% from the determinism of their components.
 
@@ -101,29 +120,31 @@
 
 :- implementation.
 
-:- import_module hlds_goal, prog_data, det_report, det_util.
-:- import_module type_util, mode_util, globals, options, passes_aux.
-:- import_module hlds_out, mercury_to_mercury, instmap.
-:- import_module bool, list, map, set, std_util, require, term.
+:- import_module purity.
+:- import_module type_util, modecheck_call, mode_util, options, passes_aux.
+:- import_module hlds_out, mercury_to_mercury.
+:- import_module assoc_list, bool, map, set, require, term.
 
 %-----------------------------------------------------------------------------%
 
 determinism_pass(ModuleInfo0, ModuleInfo) -->
 	{ determinism_declarations(ModuleInfo0, DeclaredProcs,
-		UndeclaredProcs) },
+		UndeclaredProcs, NoInferProcs) },
+	{ list__foldl(set_non_inferred_proc_determinism, NoInferProcs,
+		ModuleInfo0, ModuleInfo1) },
 	globals__io_lookup_bool_option(verbose, Verbose),
 	globals__io_lookup_bool_option(debug_det, Debug),
 	( { UndeclaredProcs = [] } ->
-		{ ModuleInfo1 = ModuleInfo0 }
+		{ ModuleInfo2 = ModuleInfo1 }
 	;
 		maybe_write_string(Verbose,
 			"% Doing determinism inference...\n"),
-		global_inference_pass(ModuleInfo0, UndeclaredProcs, Debug,
-			ModuleInfo1),
+		global_inference_pass(ModuleInfo1, UndeclaredProcs, Debug,
+			ModuleInfo2),
 		maybe_write_string(Verbose, "% done.\n")
 	),
 	maybe_write_string(Verbose, "% Doing determinism checking...\n"),
-	global_final_pass(ModuleInfo1, DeclaredProcs, Debug, ModuleInfo),
+	global_final_pass(ModuleInfo2, DeclaredProcs, Debug, ModuleInfo),
 	maybe_write_string(Verbose, "% done.\n").
 
 determinism_check_proc(ProcId, PredId, ModuleInfo0, ModuleInfo) -->
@@ -210,8 +231,6 @@ global_final_pass(ModuleInfo0, ProcList, Debug, ModuleInfo) -->
 
 %-----------------------------------------------------------------------------%
 
-:- type soln_context	--->	all_solns ; first_soln.
-
 det_infer_proc(PredId, ProcId, ModuleInfo0, ModuleInfo, Globals,
 		Detism0, Detism, Msgs) :-
 
@@ -228,11 +247,9 @@ det_infer_proc(PredId, ProcId, ModuleInfo0, ModuleInfo, Globals,
 		% context or not.  Currently we only assume so if
 		% the predicate has an explicit determinism declaration
 		% that says so.
-	(
-		proc_info_declared_determinism(Proc0, yes(DeclaredDetism)),
-		determinism_components(DeclaredDetism, _, at_most_many_cc)
-	->
-		SolnContext = first_soln
+	proc_info_declared_determinism(Proc0, MaybeDeclaredDetism),
+	( MaybeDeclaredDetism = yes(DeclaredDetism) ->
+		det_get_soln_context(DeclaredDetism, SolnContext)
 	;	
 		SolnContext = all_solns
 	),
@@ -255,8 +272,12 @@ det_infer_proc(PredId, ProcId, ModuleInfo0, ModuleInfo, Globals,
 	determinism_components(Detism1, CanFail1, MaxSoln1),
 	det_switch_canfail(CanFail0, CanFail1, CanFail),
 	det_switch_maxsoln(MaxSoln0, MaxSoln1, MaxSoln),
-	determinism_components(Detism, CanFail, MaxSoln),
+	determinism_components(Detism2, CanFail, MaxSoln),
 
+		% Now see if the evaluation model can change the detism
+	proc_info_eval_method(Proc0, EvalMethod),
+	eval_method_change_determinism(EvalMethod, Detism2, Detism),		
+			
 		% Save the newly inferred information
 	proc_info_set_goal(Proc0, Goal, Proc1),
 	proc_info_set_inferred_determinism(Proc1, Detism, Proc),
@@ -269,23 +290,19 @@ det_infer_proc(PredId, ProcId, ModuleInfo0, ModuleInfo, Globals,
 
 %-----------------------------------------------------------------------------%
 
-	% Infers the determinism of `Goal0' and returns this in `Detism'.
-	% It annotates the goal and all its subgoals with their determinism
-	% and returns the annotated goal in `Goal'.
-
-:- pred det_infer_goal(hlds_goal, instmap, soln_context, det_info,
-	hlds_goal, determinism, list(det_msg)).
-:- mode det_infer_goal(in, in, in, in, out, out, out) is det.
-
 det_infer_goal(Goal0 - GoalInfo0, InstMap0, SolnContext0, DetInfo,
 		Goal - GoalInfo, Detism, Msgs) :-
 	goal_info_get_nonlocals(GoalInfo0, NonLocalVars),
 	goal_info_get_instmap_delta(GoalInfo0, DeltaInstMap),
 
-	% If a goal has no output variables, then the goal is in
-	% single-solution context
+	% If a pure or semipure goal has no output variables,
+	% then the goal is in single-solution context
 
-	( det_no_output_vars(NonLocalVars, InstMap0, DeltaInstMap, DetInfo) ->
+	(
+		det_no_output_vars(NonLocalVars, InstMap0, DeltaInstMap,
+			DetInfo),
+		\+ goal_info_is_impure(GoalInfo0)
+	->
 		OutputVars = no,
 		SolnContext = first_soln
 	;
@@ -311,7 +328,8 @@ det_infer_goal(Goal0 - GoalInfo0, InstMap0, SolnContext0, DetInfo,
 	determinism_components(InternalDetism, InternalCanFail, InternalSolns),
 
 	(
-		% If a goal with multiple solutions has no output variables,
+		% If a pure or semipure goal with multiple solutions
+		% has no output variables,
 		% then it really it has only one solution
 		% (we will need to do pruning)
 
@@ -349,13 +367,13 @@ det_infer_goal(Goal0 - GoalInfo0, InstMap0, SolnContext0, DetInfo,
 		Goal1 \= disj(_, _),	
 
 		% do we already have a commit?
-		Goal1 \= some(_, _)
+		Goal1 \= some(_, _, _)
 	->
 		% a commit needed - we must introduce an explicit `some'
 		% so that the code generator knows to insert the appropriate
 		% code for pruning
 		goal_info_set_determinism(GoalInfo0, InternalDetism, InnerInfo),
-		Goal = some([], Goal1 - InnerInfo),
+		Goal = some([], can_remove, Goal1 - InnerInfo),
 		Msgs = Msgs1
 	;
 		% either no commit needed, or a `some' already present
@@ -366,7 +384,7 @@ det_infer_goal(Goal0 - GoalInfo0, InstMap0, SolnContext0, DetInfo,
 %-----------------------------------------------------------------------------%
 
 :- pred det_infer_goal_2(hlds_goal_expr, hlds_goal_info, instmap,
-	soln_context, det_info, set(var), instmap_delta,
+	soln_context, det_info, set(prog_var), instmap_delta,
 	hlds_goal_expr, determinism, list(det_msg)).
 :- mode det_infer_goal_2(in, in, in, in, in, in, in, out, out, out) is det.
 
@@ -378,6 +396,23 @@ det_infer_goal_2(conj(Goals0), _, InstMap0, SolnContext, DetInfo, _, _,
 	det_infer_conj(Goals0, InstMap0, SolnContext, DetInfo,
 		Goals, Detism, Msgs).
 
+det_infer_goal_2(par_conj(Goals0, SM), GoalInfo, InstMap0, SolnContext,
+		DetInfo, _, _, par_conj(Goals, SM), Detism, Msgs) :-
+	det_infer_par_conj(Goals0, InstMap0, SolnContext, DetInfo,
+		Goals, Detism, Msgs0),
+	(
+		determinism_components(Detism, CanFail, Solns),
+		CanFail = cannot_fail,
+		Solns \= at_most_many
+	->
+		Msgs = Msgs0
+	;
+		det_info_get_pred_id(DetInfo, PredId),
+		det_info_get_proc_id(DetInfo, ProcId),
+		Msg = par_conj_not_det(Detism, PredId, ProcId, GoalInfo, Goals),
+		Msgs = [Msg|Msgs0]
+	).
+
 det_infer_goal_2(disj(Goals0, SM), _, InstMap0, SolnContext, DetInfo, _, _,
 		disj(Goals, SM), Detism, Msgs) :-
 	det_infer_disj(Goals0, InstMap0, SolnContext, DetInfo,
@@ -418,40 +453,64 @@ det_infer_goal_2(switch(Var, SwitchCanFail, Cases0, SM), GoalInfo,
 	% This is the point at which annotations start changing
 	% when we iterate to fixpoint for global determinism inference.
 
-det_infer_goal_2(call(PredId, ModeId, A, B, C, N), GoalInfo, _, SolnContext,
+det_infer_goal_2(call(PredId, ModeId0, A, B, C, N), GoalInfo, _, SolnContext,
 		DetInfo, _, _,
 		call(PredId, ModeId, A, B, C, N), Detism, Msgs) :-
-	det_lookup_detism(DetInfo, PredId, ModeId, Detism),
+	det_lookup_detism(DetInfo, PredId, ModeId0, Detism0),
 	%
 	% Make sure we don't try to call a committed-choice pred
 	% from a non-committed-choice context.
 	%
-	determinism_components(Detism, _, NumSolns),
+	determinism_components(Detism0, CanFail, NumSolns),
 	(
 		NumSolns = at_most_many_cc,
 		SolnContext \= first_soln
 	->
-		Msgs = [cc_pred_in_wrong_context(GoalInfo, Detism,
-				PredId, ModeId)]
+		(
+			det_find_matching_non_cc_mode(DetInfo, PredId, ModeId0,
+				ModeId1)
+		->
+			ModeId = ModeId1,
+			Msgs = [],
+			determinism_components(Detism, CanFail, at_most_many)
+		;
+			Msgs = [cc_pred_in_wrong_context(GoalInfo, Detism0,
+					PredId, ModeId0)],
+			ModeId = ModeId0,
+			% Code elsewhere relies on the assumption that
+			%	SolnContext \= first_soln =>
+			%		NumSolns \= at_most_many_cc,
+			% so we need to enforce that here.
+			determinism_components(Detism, CanFail, at_most_many)
+		)
 	;
-		Msgs = []
+		Msgs = [],
+		ModeId = ModeId0,
+		Detism = Detism0
 	).
 
-det_infer_goal_2(higher_order_call(PredVar, ArgVars, Types, Modes, Det,
-			IsPredOrFunc),
+det_infer_goal_2(generic_call(GenericCall, ArgVars, Modes, Det0),
 		GoalInfo, _InstMap0, SolnContext,
 		_MiscInfo, _NonLocalVars, _DeltaInstMap,
-		higher_order_call(PredVar, ArgVars, Types, Modes, Det,
-			IsPredOrFunc),
+		generic_call(GenericCall, ArgVars, Modes, Det0),
 		Det, Msgs) :-
-	determinism_components(Det, _, NumSolns),
+	determinism_components(Det0, CanFail, NumSolns),
 	(
 		NumSolns = at_most_many_cc,
 		SolnContext \= first_soln
 	->
-		Msgs = [higher_order_cc_pred_in_wrong_context(GoalInfo, Det)]
+		% This error can only occur for higher-order calls.
+		% class_method calls are only introduced by polymorphism,
+		% and the aditi_builtins are all det (for the updates)
+		% or introduced later (for calls).
+		Msgs = [higher_order_cc_pred_in_wrong_context(GoalInfo, Det0)],
+		% Code elsewhere relies on the assumption that
+		% SolnContext \= first_soln => NumSolns \= at_most_many_cc,
+		% so we need to enforce that here.
+		determinism_components(Det, CanFail, at_most_many)
 	;
-		Msgs = []
+		Msgs = [],
+		Det = Det0
 	).
 
 	% unifications are either deterministic or semideterministic.
@@ -459,8 +518,8 @@ det_infer_goal_2(higher_order_call(PredVar, ArgVars, Types, Modes, Det,
 det_infer_goal_2(unify(LT, RT0, M, U, C), GoalInfo, InstMap0, SolnContext,
 		DetInfo, _, _, unify(LT, RT, M, U, C), UnifyDet, Msgs) :-
 	(
-		RT0 = lambda_goal(PredOrFunc, Vars, Modes, LambdaDeclaredDet,
-				Goal0)
+		RT0 = lambda_goal(PredOrFunc, EvalMethod, FixModes,
+			NonLocalVars, Vars, Modes, LambdaDeclaredDet, Goal0)
 	->
 		(
 			determinism_components(LambdaDeclaredDet, _,
@@ -478,8 +537,8 @@ det_infer_goal_2(unify(LT, RT0, M, U, C), GoalInfo, InstMap0, SolnContext,
 		det_check_lambda(LambdaDeclaredDet, LambdaInferredDet,
 				Goal, GoalInfo, DetInfo, Msgs2),
 		list__append(Msgs1, Msgs2, Msgs3),
-		RT = lambda_goal(PredOrFunc, Vars, Modes, LambdaDeclaredDet,
-				Goal)
+		RT = lambda_goal(PredOrFunc, EvalMethod, FixModes,
+			NonLocalVars, Vars, Modes, LambdaDeclaredDet, Goal)
 	;
 		RT = RT0,
 		Msgs3 = []
@@ -576,38 +635,51 @@ det_infer_goal_2(not(Goal0), _, InstMap0, _SolnContext, DetInfo, _, _,
 	% but we cannot rely on explicit quantification to detect this.
 	% Therefore cuts are handled in det_infer_goal.
 
-det_infer_goal_2(some(Vars, Goal0), _, InstMap0, SolnContext, DetInfo, _, _,
-		some(Vars, Goal), Det, Msgs) :-
+det_infer_goal_2(some(Vars, CanRemove, Goal0), _, InstMap0, SolnContext,
+		DetInfo, _, _, some(Vars, CanRemove, Goal), Det, Msgs) :-
 	det_infer_goal(Goal0, InstMap0, SolnContext, DetInfo,
 		Goal, Det, Msgs).
 
 	% pragma c_codes are handled in the same way as predicate calls
-det_infer_goal_2(pragma_c_code(C_Code, IsRecursive, PredId, ProcId, Args,
-			ArgNameMap, OrigArgTypes, Extra), 
+det_infer_goal_2(pragma_c_code(IsRecursive, PredId, ProcId, Args,
+			ArgNameMap, OrigArgTypes, PragmaCode), 
 		GoalInfo, _, SolnContext, DetInfo, _, _,
-		pragma_c_code(C_Code, IsRecursive, PredId, ProcId, Args,
-			ArgNameMap, OrigArgTypes, Extra),
+		pragma_c_code(IsRecursive, PredId, ProcId, Args,
+			ArgNameMap, OrigArgTypes, PragmaCode),
 		Detism, Msgs) :-
-	det_lookup_detism(DetInfo, PredId, ProcId, Detism0),
-	determinism_components(Detism0, CanFail, NumSolns0),
-	( Extra = extra_pragma_info(_, _) ->
-		% pragma C codes that specify saved variables and labels
-		% can have more than one solution
-		NumSolns = at_most_many
+	det_info_get_module_info(DetInfo, ModuleInfo),
+	module_info_pred_proc_info(ModuleInfo, PredId, ProcId, _, ProcInfo),
+	proc_info_declared_determinism(ProcInfo, MaybeDetism),
+	( MaybeDetism = yes(Detism0) ->
+		determinism_components(Detism0, CanFail, NumSolns0),
+		( PragmaCode = nondet(_, _, _, _, _, _, _, _, _) ->
+			% pragma C codes of this form
+			% can have more than one solution
+			NumSolns1 = at_most_many
+		;
+			NumSolns1 = NumSolns0
+		),
+		(
+			NumSolns1 = at_most_many_cc,
+			SolnContext \= first_soln
+		->
+			Msgs = [cc_pred_in_wrong_context(GoalInfo, Detism0,
+					PredId, ProcId)],
+			NumSolns = at_most_many
+		;
+			Msgs = [],
+			NumSolns = NumSolns1
+		),
+		determinism_components(Detism, CanFail, NumSolns)
 	;
-		NumSolns = NumSolns0
-	),
-	determinism_components(Detism, CanFail, NumSolns),
-	(
-		NumSolns = at_most_many_cc,
-		SolnContext \= first_soln
-	->
-		Msgs = [cc_pred_in_wrong_context(GoalInfo, Detism,
-				PredId, ProcId)]
-	;
-		Msgs = []
+		Msgs = [pragma_c_code_without_det_decl(PredId, ProcId)],
+		Detism = erroneous
 	).
 
+det_infer_goal_2(bi_implication(_, _), _, _, _, _, _, _, _, _, _) :-
+	% these should have been expanded out by now
+	error("det_infer_goal_2: unexpected bi_implication").
+
 %-----------------------------------------------------------------------------%
 
 :- pred det_infer_conj(list(hlds_goal), instmap, soln_context, det_info,
@@ -661,6 +733,27 @@ det_infer_conj([Goal0 | Goals0], InstMap0, SolnContext, DetInfo,
 	det_conjunction_detism(DetismA, DetismB, Detism),
 	list__append(MsgsA, MsgsB, Msgs).
 
+:- pred det_infer_par_conj(list(hlds_goal), instmap, soln_context, det_info,
+		list(hlds_goal), determinism, list(det_msg)).
+:- mode det_infer_par_conj(in, in, in, in, out, out, out) is det.
+
+det_infer_par_conj([], _InstMap0, _SolnContext, _DetInfo, [], det, []).
+det_infer_par_conj([Goal0 | Goals0], InstMap0, SolnContext, DetInfo, 
+		[Goal | Goals], Detism, Msgs) :-
+
+	det_infer_goal(Goal0, InstMap0, SolnContext, DetInfo,
+			Goal, DetismA, MsgsA),
+	determinism_components(DetismA, CanFailA, MaxSolnsA),
+	
+	det_infer_par_conj(Goals0, InstMap0, SolnContext, DetInfo,
+			Goals, DetismB, MsgsB),
+	determinism_components(DetismB, CanFailB, MaxSolnsB),
+
+	det_conjunction_maxsoln(MaxSolnsA, MaxSolnsB, MaxSolns),
+	det_conjunction_canfail(CanFailA, CanFailB, CanFail),
+	determinism_components(Detism, CanFail, MaxSolns),
+	list__append(MsgsA, MsgsB, Msgs).
+
 :- pred det_infer_disj(list(hlds_goal), instmap, soln_context, det_info,
 	can_fail, soln_count, list(hlds_goal), determinism, list(det_msg)).
 :- mode det_infer_disj(in, in, in, in, in, in, out, out, out) is det.
@@ -675,8 +768,15 @@ det_infer_disj([Goal0 | Goals0], InstMap0, SolnContext, DetInfo, CanFail0,
 	determinism_components(Detism1, CanFail1, MaxSolns1),
 	det_disjunction_canfail(CanFail0, CanFail1, CanFail2),
 	det_disjunction_maxsoln(MaxSolns0, MaxSolns1, MaxSolns2),
+	% if we're in a single-solution context,
+	% convert `at_most_many' to `at_most_many_cc'
+	( SolnContext = first_soln, MaxSolns2 = at_most_many ->
+		MaxSolns3 = at_most_many_cc
+	;
+		MaxSolns3 = MaxSolns2
+	),
 	det_infer_disj(Goals0, InstMap0, SolnContext, DetInfo, CanFail2,
-		MaxSolns2, Goals1, Detism, Msgs2),
+		MaxSolns3, Goals1, Detism, Msgs2),
 	list__append(Msgs1, Msgs2, Msgs).
 
 :- pred det_infer_switch(list(case), instmap, soln_context, det_info,
@@ -705,7 +805,47 @@ det_infer_switch([Case0 | Cases0], InstMap0, SolnContext, DetInfo, CanFail0,
 
 %-----------------------------------------------------------------------------%
 
-:- pred det_check_for_noncanonical_type(var, bool, can_fail, soln_context,
+	% det_find_matching_non_cc_mode(DetInfo, PredId, ProcId0, ProcId):
+	%	Search for a mode of the given predicate that
+	% 	is identical to the mode ProcId0, except that its
+	%	determinism is non-cc whereas ProcId0's detism is cc.
+	%	Let ProcId be the first such mode.
+
+:- pred det_find_matching_non_cc_mode(det_info, pred_id, proc_id, proc_id).
+:- mode det_find_matching_non_cc_mode(in, in, in, out) is semidet.
+
+det_find_matching_non_cc_mode(DetInfo, PredId, ProcId0, ProcId) :-
+	det_info_get_module_info(DetInfo, ModuleInfo),
+        module_info_preds(ModuleInfo, PredTable),
+	map__lookup(PredTable, PredId, PredInfo),
+	pred_info_procedures(PredInfo, ProcTable),
+	map__to_assoc_list(ProcTable, ProcList),
+	det_find_matching_non_cc_mode_2(ProcList, ModuleInfo, PredInfo,
+		ProcId0, ProcId).
+
+:- pred det_find_matching_non_cc_mode_2(assoc_list(proc_id, proc_info),
+		module_info, pred_info, proc_id, proc_id).
+:- mode det_find_matching_non_cc_mode_2(in, in, in, in, out) is semidet.
+
+det_find_matching_non_cc_mode_2([ProcId1 - ProcInfo | Rest],
+		ModuleInfo, PredInfo, ProcId0, ProcId) :-
+	(
+		ProcId1 \= ProcId0,
+		proc_info_interface_determinism(ProcInfo, Detism),
+		determinism_components(Detism, _CanFail, MaxSoln),
+		MaxSoln = at_most_many,
+		modes_are_identical_bar_cc(ProcId0, ProcId1, PredInfo,
+			ModuleInfo)
+	->
+		ProcId = ProcId1
+	;
+		det_find_matching_non_cc_mode_2(Rest, ModuleInfo, PredInfo,
+				ProcId0, ProcId)
+	).
+
+%-----------------------------------------------------------------------------%
+
+:- pred det_check_for_noncanonical_type(prog_var, bool, can_fail, soln_context,
 		hlds_goal_info, cc_unify_context, det_info, list(det_msg),
 		soln_count, list(det_msg)).
 :- mode det_check_for_noncanonical_type(in, in, in, in,
@@ -724,15 +864,14 @@ det_check_for_noncanonical_type(Var, ExaminesRepresentation, CanFail,
 		det_get_proc_info(DetInfo, ProcInfo),
 		proc_info_vartypes(ProcInfo, VarTypes),
 		map__lookup(VarTypes, Var, Type),
-		det_type_has_user_defined_equality_pred(DetInfo, Type,
-			_TypeContext)
+		det_type_has_user_defined_equality_pred(DetInfo, Type)
 	->
 		( CanFail = can_fail ->
-			proc_info_variables(ProcInfo, VarSet),
+			proc_info_varset(ProcInfo, VarSet),
 			Msgs = [cc_unify_can_fail(GoalInfo, Var, Type,
 				VarSet, GoalContext) | Msgs0]
 		; SolnContext \= first_soln ->
-			proc_info_variables(ProcInfo, VarSet),
+			proc_info_varset(ProcInfo, VarSet),
 			Msgs = [cc_unify_in_wrong_context(GoalInfo, Var,
 				Type, VarSet, GoalContext) | Msgs0]
 		;
@@ -750,25 +889,20 @@ det_check_for_noncanonical_type(Var, ExaminesRepresentation, CanFail,
 
 % return true iff there was a `where equality is <predname>' declaration
 % for the specified type.
-:- pred det_type_has_user_defined_equality_pred(det_info::in, (type)::in,
-		term__context::out) is semidet.
-det_type_has_user_defined_equality_pred(DetInfo, Type, TypeContext) :-
+:- pred det_type_has_user_defined_equality_pred(det_info::in,
+		(type)::in) is semidet.
+det_type_has_user_defined_equality_pred(DetInfo, Type) :-
 	det_info_get_module_info(DetInfo, ModuleInfo),
-	module_info_types(ModuleInfo, TypeTable),
-	type_to_type_id(Type, TypeId, _TypeArgs),
-	map__search(TypeTable, TypeId, TypeDefn),
-	hlds_data__get_type_defn_body(TypeDefn, TypeBody),
-	TypeBody = du_type(_, _, _, yes(_)),
-	hlds_data__get_type_defn_context(TypeDefn, TypeContext).
+	type_has_user_defined_equality_pred(ModuleInfo, Type, _).
 
 % return yes iff the results of the specified unification might depend on
 % the concrete representation of the abstract values involved.
 :- pred det_infer_unify_examines_rep(unification::in, bool::out) is det.
 det_infer_unify_examines_rep(assign(_, _), no).
-det_infer_unify_examines_rep(construct(_, _, _, _), no).
+det_infer_unify_examines_rep(construct(_, _, _, _, _, _, _), no).
 det_infer_unify_examines_rep(deconstruct(_, _, _, _, _), yes).
 det_infer_unify_examines_rep(simple_test(_, _), yes).
-det_infer_unify_examines_rep(complicated_unify(_, _), no).
+det_infer_unify_examines_rep(complicated_unify(_, _, _), no).
 	% Some complicated modes of complicated unifications _do_
 	% examine the representation...
 	% but we will catch those by reporting errors in the
@@ -790,12 +924,21 @@ det_infer_unify_examines_rep(complicated_unify(_, _), no).
 
 det_infer_unify_canfail(deconstruct(_, _, _, _, CanFail), CanFail).
 det_infer_unify_canfail(assign(_, _), cannot_fail).
-det_infer_unify_canfail(construct(_, _, _, _), cannot_fail).
+det_infer_unify_canfail(construct(_, _, _, _, _, _, _), cannot_fail).
 det_infer_unify_canfail(simple_test(_, _), can_fail).
-det_infer_unify_canfail(complicated_unify(_, CanFail), CanFail).
+det_infer_unify_canfail(complicated_unify(_, CanFail, _), CanFail).
 
 %-----------------------------------------------------------------------------%
 
+det_get_soln_context(DeclaredDetism, SolnContext) :-
+	(
+		determinism_components(DeclaredDetism, _, at_most_many_cc)
+	->
+		SolnContext = first_soln
+	;	
+		SolnContext = all_solns
+	).
+
 % When figuring out the determinism of a conjunction,
 % if the second goal is unreachable, then then the
 % determinism of the conjunction is just the determinism
@@ -945,15 +1088,19 @@ det_negation_det(failure,	yes(det)).
 	% returns two lists of procedure ids, the first being those
 	% with determinism declarations, and the second being those without.
 
-:- pred determinism_declarations(module_info, pred_proc_list, pred_proc_list).
-:- mode determinism_declarations(in, out, out) is det.
+:- pred determinism_declarations(module_info, pred_proc_list,
+		pred_proc_list, pred_proc_list).
+:- mode determinism_declarations(in, out, out, out) is det.
 
-determinism_declarations(ModuleInfo, DeclaredProcs, UndeclaredProcs) :-
+determinism_declarations(ModuleInfo, DeclaredProcs,
+		UndeclaredProcs, NoInferProcs) :-
 	get_all_pred_procs(ModuleInfo, PredProcs),
-	segregate_procs(ModuleInfo, PredProcs, DeclaredProcs, UndeclaredProcs).
+	segregate_procs(ModuleInfo, PredProcs, DeclaredProcs,
+		UndeclaredProcs, NoInferProcs).
 
 	% get_all_pred_procs takes a module_info and returns a list
-	% of all the procedures ids for that module.
+	% of all the procedures ids for that module (except class methods,
+	% which do not need to be checked since we generate the code ourselves).
 
 :- pred get_all_pred_procs(module_info, pred_proc_list).
 :- mode get_all_pred_procs(in, out) is det.
@@ -970,7 +1117,7 @@ get_all_pred_procs(ModuleInfo, PredProcs) :-
 get_all_pred_procs_2(_Preds, [], PredProcs, PredProcs).
 get_all_pred_procs_2(Preds, [PredId|PredIds], PredProcs0, PredProcs) :-
 	map__lookup(Preds, PredId, Pred),
-	pred_info_non_imported_procids(Pred, ProcIds),
+	pred_info_procids(Pred, ProcIds),
 	fold_pred_modes(PredId, ProcIds, PredProcs0, PredProcs1),
 	get_all_pred_procs_2(Preds, PredIds, PredProcs1, PredProcs).
 
@@ -987,37 +1134,86 @@ fold_pred_modes(PredId, [ProcId|ProcIds], PredProcs0, PredProcs) :-
 	% UndeclaredProcs.
 
 :- pred segregate_procs(module_info, pred_proc_list, pred_proc_list,
-	pred_proc_list).
-:- mode segregate_procs(in, in, out, out) is det.
+	pred_proc_list, pred_proc_list).
+:- mode segregate_procs(in, in, out, out, out) is det.
 
-segregate_procs(ModuleInfo, PredProcs, DeclaredProcs, UndeclaredProcs) :-
+segregate_procs(ModuleInfo, PredProcs, DeclaredProcs,
+		UndeclaredProcs, NoInferProcs) :-
 	segregate_procs_2(ModuleInfo, PredProcs, [], DeclaredProcs,
-					[], UndeclaredProcs).
+			[], UndeclaredProcs, [], NoInferProcs).
 
 :- pred segregate_procs_2(module_info, pred_proc_list, pred_proc_list,
-			pred_proc_list, pred_proc_list, pred_proc_list).
-:- mode segregate_procs_2(in, in, in, out, in, out) is det.
+			pred_proc_list, pred_proc_list, pred_proc_list,
+			pred_proc_list, pred_proc_list).
+:- mode segregate_procs_2(in, in, in, out, in, out, in, out) is det.
 
 segregate_procs_2(_ModuleInfo, [], DeclaredProcs, DeclaredProcs,
-				UndeclaredProcs, UndeclaredProcs).
+		UndeclaredProcs, UndeclaredProcs, NoInferProcs, NoInferProcs).
 segregate_procs_2(ModuleInfo, [proc(PredId, ProcId) | PredProcs],
 		DeclaredProcs0, DeclaredProcs,
-		UndeclaredProcs0, UndeclaredProcs) :-
+		UndeclaredProcs0, UndeclaredProcs,
+		NoInferProcs0, NoInferProcs) :-
 	module_info_preds(ModuleInfo, Preds),
 	map__lookup(Preds, PredId, Pred),
-	pred_info_procedures(Pred, Procs),
-	map__lookup(Procs, ProcId, Proc),
-	proc_info_declared_determinism(Proc, MaybeDetism),
-	(
-		MaybeDetism = no,
-		UndeclaredProcs1 = [proc(PredId, ProcId) | UndeclaredProcs0],
-		DeclaredProcs1 = DeclaredProcs0
+	( 
+		(
+			pred_info_is_imported(Pred)
+		;
+			pred_info_is_pseudo_imported(Pred),
+			hlds_pred__in_in_unification_proc_id(ProcId)
+		;
+			pred_info_get_markers(Pred, Markers),
+			check_marker(Markers, class_method)
+		)
+	->
+		UndeclaredProcs1 = UndeclaredProcs0,
+		DeclaredProcs1 = DeclaredProcs0,
+		NoInferProcs1 = [proc(PredId, ProcId) | NoInferProcs0]
 	;
-		MaybeDetism = yes(_),
-		DeclaredProcs1 = [proc(PredId, ProcId) | DeclaredProcs0],
-		UndeclaredProcs1 = UndeclaredProcs0
+		pred_info_procedures(Pred, Procs),
+		map__lookup(Procs, ProcId, Proc),
+		proc_info_declared_determinism(Proc, MaybeDetism),
+		(
+			MaybeDetism = no,
+			UndeclaredProcs1 =
+				[proc(PredId, ProcId) | UndeclaredProcs0],
+			DeclaredProcs1 = DeclaredProcs0
+		;
+			MaybeDetism = yes(_),
+			DeclaredProcs1 =
+				[proc(PredId, ProcId) | DeclaredProcs0],
+			UndeclaredProcs1 = UndeclaredProcs0
+		),
+		NoInferProcs1 = NoInferProcs0
 	),
 	segregate_procs_2(ModuleInfo, PredProcs, DeclaredProcs1, DeclaredProcs,
-		UndeclaredProcs1, UndeclaredProcs).
+		UndeclaredProcs1, UndeclaredProcs,
+		NoInferProcs1, NoInferProcs).
+
+	% We can't infer a tighter determinism for imported procedures or
+	% for class methods, so set the inferred determinism to be the
+	% same as the declared determinism. This can't be done easily in 
+	% make_hlds.m since inter-module optimization means that the
+	% import_status of procedures isn't determined until after all
+	% items are processed.
+:- pred set_non_inferred_proc_determinism(pred_proc_id,
+		module_info, module_info).
+:- mode set_non_inferred_proc_determinism(in, in, out) is det.
+
+set_non_inferred_proc_determinism(proc(PredId, ProcId),
+		ModuleInfo0, ModuleInfo) :-
+	module_info_pred_info(ModuleInfo0, PredId, PredInfo0),
+	pred_info_procedures(PredInfo0, Procs0),
+	map__lookup(Procs0, ProcId, ProcInfo0),
+	proc_info_declared_determinism(ProcInfo0, MaybeDet),
+	( MaybeDet = yes(Det) ->
+		proc_info_set_inferred_determinism(ProcInfo0, Det, ProcInfo),
+		map__det_update(Procs0, ProcId, ProcInfo, Procs),
+		pred_info_set_procedures(PredInfo0, Procs, PredInfo),
+		module_info_set_pred_info(ModuleInfo0,
+			PredId, PredInfo, ModuleInfo)
+	;
+		ModuleInfo = ModuleInfo0
+	).
 
 %-----------------------------------------------------------------------------%
