@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -14,42 +14,49 @@
 
 :- interface.
 
-:- import_module hlds_module, hlds_pred, hlds_goal, det_util, prog_data.
-:- import_module io.
+:- import_module hlds_module, hlds_pred, hlds_goal.
+:- import_module det_util, prog_data.
+:- import_module io, list.
 
 :- type det_msg	--->
 			% warnings
-			multidet_disj(term__context, list(term__context))
-		;	det_disj(term__context, list(term__context))
-		;	semidet_disj(term__context, list(term__context))
-		;	zero_soln_disj(term__context, list(term__context))
-		;	zero_soln_disjunct(term__context)
-		;	ite_cond_cannot_fail(term__context)
-		;	ite_cond_cannot_succeed(term__context)
-		;	negated_goal_cannot_fail(term__context)
-		;	negated_goal_cannot_succeed(term__context)
-		;	warn_obsolete(pred_id, term__context)
+			multidet_disj(prog_context, list(prog_context))
+		;	det_disj(prog_context, list(prog_context))
+		;	semidet_disj(prog_context, list(prog_context))
+		;	zero_soln_disj(prog_context, list(prog_context))
+		;	zero_soln_disjunct(prog_context)
+		;	ite_cond_cannot_fail(prog_context)
+		;	ite_cond_cannot_succeed(prog_context)
+		;	negated_goal_cannot_fail(prog_context)
+		;	negated_goal_cannot_succeed(prog_context)
+		;	goal_cannot_succeed(prog_context)
+		;	det_goal_has_no_outputs(prog_context)
+		;	warn_obsolete(pred_id, prog_context)
 				% warning about calls to predicates
 				% for which there is a `:- pragma obsolete'
 				% declaration.
-		;	warn_infinite_recursion(term__context)
+		;	warn_infinite_recursion(prog_context)
 				% warning about recursive calls
 				% which would cause infinite loops.
-		;	duplicate_call(seen_call_id, term__context,
-				term__context)
+		;	duplicate_call(seen_call_id, prog_context,
+				prog_context)
 				% multiple calls with the same input args.
 			% errors
-		;	cc_unify_can_fail(hlds_goal_info, var, type, varset,
-				cc_unify_context)
-		;	cc_unify_in_wrong_context(hlds_goal_info, var, type,
-				varset, cc_unify_context)
+		;	cc_unify_can_fail(hlds_goal_info, prog_var, type,
+				prog_varset, cc_unify_context)
+		;	cc_unify_in_wrong_context(hlds_goal_info, prog_var,
+				type, prog_varset, cc_unify_context)
 		;	cc_pred_in_wrong_context(hlds_goal_info, determinism,
 				pred_id, proc_id)
 		;	higher_order_cc_pred_in_wrong_context(hlds_goal_info,
 				determinism)
 		;	error_in_lambda(
 				determinism, determinism, % declared, inferred
-				hlds_goal, hlds_goal_info, pred_id, proc_id).
+				hlds_goal, hlds_goal_info, pred_id, proc_id)
+		;	par_conj_not_det(determinism, pred_id, proc_id,
+				hlds_goal_info, list(hlds_goal))
+		; 	pragma_c_code_without_det_decl(pred_id, proc_id)
+		.
 
 :- type seen_call_id
 	--->	seen_call(pred_id, proc_id)
@@ -90,6 +97,18 @@
 	io__state, io__state).
 :- mode det_report_msgs(in, in, out, out, di, uo) is det.
 
+
+:- type msg_modes
+	--->    all_modes       % the warning should be reported only
+				% if it occurs in all modes of the predicate
+	;       any_mode	% the warning should be reported 
+				% if it occurs in any mode of the predicate
+	.
+
+	% Return `yes' if the warning should be reported if it occurs in
+	% any mode of the predicate, not only if it occurs in all modes.
+:- pred det_msg_is_any_mode_msg(det_msg::in, msg_modes::out) is det.
+
 %-----------------------------------------------------------------------------%
 
 :- type det_comparison	--->	tighter ; sameas ; looser.
@@ -103,9 +122,9 @@
 
 :- import_module hlds_data, type_util, mode_util, inst_match.
 :- import_module globals, options, prog_out, hlds_out, mercury_to_mercury.
-:- import_module passes_aux.
+:- import_module passes_aux, term, varset.
 
-:- import_module bool, int, list, map, set, varset, std_util, term, require.
+:- import_module bool, int, map, set, std_util, require, string.
 
 %-----------------------------------------------------------------------------%
 
@@ -125,19 +144,19 @@ global_checking_pass([proc(PredId, ProcId) | Rest], ModuleInfo0, ModuleInfo) -->
 		module_info, module_info, io__state, io__state).
 :- mode check_determinism(in, in, in, in, in, out, di, uo) is det.
 
-check_determinism(PredId, ProcId, _PredInfo, ProcInfo,
+check_determinism(PredId, ProcId, PredInfo0, ProcInfo0,
 		ModuleInfo0, ModuleInfo) -->
-	{ proc_info_declared_determinism(ProcInfo, MaybeDetism) },
-	{ proc_info_inferred_determinism(ProcInfo, InferredDetism) },
+	{ proc_info_declared_determinism(ProcInfo0, MaybeDetism) },
+	{ proc_info_inferred_determinism(ProcInfo0, InferredDetism) },
 	(
 		{ MaybeDetism = no },
-		{ ModuleInfo = ModuleInfo0 }
+		{ ModuleInfo1 = ModuleInfo0 }
 	;
 		{ MaybeDetism = yes(DeclaredDetism) },
 		{ compare_determinisms(DeclaredDetism, InferredDetism, Cmp) },
 		(
 			{ Cmp = sameas },
-			{ ModuleInfo = ModuleInfo0 }
+			{ ModuleInfo1 = ModuleInfo0 }
 		;
 			{ Cmp = looser },
 			globals__io_lookup_bool_option(
@@ -151,25 +170,79 @@ check_determinism(PredId, ProcId, _PredInfo, ProcInfo,
 			;
 				[]
 			),
-			{ ModuleInfo = ModuleInfo0 }
+			{ ModuleInfo1 = ModuleInfo0 }
 		;
 			{ Cmp = tighter },
-			{ module_info_incr_errors(ModuleInfo0, ModuleInfo) },
+			{ module_info_incr_errors(ModuleInfo0, ModuleInfo1) },
 			{ Message = "  error: determinism declaration not satisfied.\n" },
 			report_determinism_problem(PredId,
-				ProcId, ModuleInfo, Message,
+				ProcId, ModuleInfo1, Message,
 				DeclaredDetism, InferredDetism),
-			{ proc_info_goal(ProcInfo, Goal) },
+			{ proc_info_goal(ProcInfo0, Goal) },
 			globals__io_get_globals(Globals),
-			{ det_info_init(ModuleInfo, PredId, ProcId, Globals,
+			{ det_info_init(ModuleInfo1, PredId, ProcId, Globals,
 				DetInfo) },
 			det_diagnose_goal(Goal, DeclaredDetism, [], DetInfo, _)
 			% XXX with the right verbosity options, we want to
 			% call report_determinism_problem only if diagnose
 			% returns false, i.e. it didn't print a message.
 		)
+	),
+	
+	% make sure the code model is valid given the eval method
+	{ proc_info_eval_method(ProcInfo0, EvalMethod) },
+	{ determinism_to_code_model(InferredDetism, CodeMod) },
+	( 
+		{ valid_code_model_for_eval_method(EvalMethod, CodeMod) }
+	->
+		{
+		    proc_info_set_eval_method(ProcInfo0, EvalMethod, ProcInfo),
+		    pred_info_procedures(PredInfo0, ProcTable0),
+		    map__det_update(ProcTable0, ProcId, ProcInfo, ProcTable),
+		    pred_info_set_procedures(PredInfo0, ProcTable, PredInfo),
+		    module_info_set_pred_info(ModuleInfo1, PredId, PredInfo, 
+		    	ModuleInfo)
+		}
+	;
+		{ proc_info_context(ProcInfo0, Context) },
+		prog_out__write_context(Context),
+		{ eval_method_to_string(EvalMethod, EvalMethodS) },
+		io__write_string("Error: `pragma "),
+		io__write_string(EvalMethodS),
+		io__write_string("' declaration not allowed for procedure\n"),
+		prog_out__write_context(Context),
+		io__write_string("  with determinism `"),
+		mercury_output_det(InferredDetism),
+		io__write_string("'.\n"), 
+		globals__io_lookup_bool_option(verbose_errors, VerboseErrors),
+		( { VerboseErrors = yes } ->
+			io__write_string(
+"\tThe pragma requested is only valid for the folowing determinism(s):\n"),
+			{ solutions(get_valid_dets(EvalMethod), Sols) },
+			print_dets(Sols)
+		;
+			[]
+		),
+		{ module_info_incr_errors(ModuleInfo1, ModuleInfo) }
 	).
 
+:- pred get_valid_dets(eval_method, determinism).
+:- mode get_valid_dets(in, out) is multidet.
+
+get_valid_dets(EvalMethod, Det) :-
+	valid_code_model_for_eval_method(EvalMethod, CodeModel),
+	determinism_to_code_model(Det, CodeModel).
+
+:- pred print_dets(list(determinism), io__state, io__state).
+:- mode print_dets(in, di, uo) is det.
+
+print_dets([]) --> [].
+print_dets([D|Rest]) -->
+	io__write_string("\t\t"),
+	mercury_output_det(D),
+	io__nl,
+	print_dets(Rest).
+	
 :- pred check_if_main_can_fail(pred_id, proc_id, pred_info, proc_info,
 		module_info, module_info, io__state, io__state).
 :- mode check_if_main_can_fail(in, in, in, in, in, out, di, uo) is det.
@@ -190,8 +263,8 @@ check_if_main_can_fail(_PredId, _ProcId, PredInfo, ProcInfo,
 		  determinism_components(DeclaredDeterminism, can_fail, _)
 		}
 	->
-		{ proc_info_context(ProcInfo, Context) },
-		prog_out__write_context(Context),
+		{ proc_info_context(ProcInfo, Context1) },
+		prog_out__write_context(Context1),
 			% The error message is actually a lie -
 			% main/2 can also be `erroneous'.  But mentioning
 			% that would probably just confuse people.
@@ -233,6 +306,7 @@ check_for_multisoln_func(_PredId, _ProcId, PredInfo, ProcInfo,
 	->
 		% ... then it is an error.
 		{ pred_info_name(PredInfo, PredName) },
+
 		{ proc_info_context(ProcInfo, FuncContext) },
 		prog_out__write_context(FuncContext),
 		io__write_string("Error: invalid determinism for function\n"),
@@ -287,7 +361,7 @@ report_determinism_problem(PredId, ModeId, ModuleInfo, Message,
 		DeclaredDetism, InferredDetism) -->
 	globals__io_lookup_bool_option(halt_at_warn, HaltAtWarn),
 	( { HaltAtWarn = yes } ->
-		 io__set_exit_status(1)
+		io__set_exit_status(1)
 	;
 		[]
 	),
@@ -383,15 +457,19 @@ det_diagnose_goal_2(conj(Goals), _GoalInfo, Desired, _Actual, Context, DetInfo,
 		Diagnosed) -->
 	det_diagnose_conj(Goals, Desired, Context, DetInfo, Diagnosed).
 
+det_diagnose_goal_2(par_conj(Goals, _SM), _GoalInfo, Desired, _Actual,
+		Context, DetInfo, Diagnosed) -->
+	det_diagnose_conj(Goals, Desired, Context, DetInfo, Diagnosed).
+
 det_diagnose_goal_2(disj(Goals, _), GoalInfo, Desired, Actual, SwitchContext,
 		DetInfo, Diagnosed) -->
 	det_diagnose_disj(Goals, Desired, Actual, SwitchContext, DetInfo, 0,
-		Clauses, Diagnosed1),
+		ClausesWithSoln, Diagnosed1),
 	{ determinism_components(Desired, _, DesSolns) },
 	(
 		{ DesSolns \= at_most_many },
 		{ DesSolns \= at_most_many_cc },
-		{ Clauses > 1 }
+		{ ClausesWithSoln > 1 }
 	->
 		{ goal_info_get_context(GoalInfo, Context) },
 		prog_out__write_context(Context),
@@ -417,7 +495,7 @@ det_diagnose_goal_2(switch(Var, SwitchCanFail, Cases, _), GoalInfo,
 			DetInfo),
 		prog_out__write_context(Context),
 		{ det_get_proc_info(DetInfo, ProcInfo) },
-		{ proc_info_variables(ProcInfo, Varset) },
+		{ proc_info_varset(ProcInfo, Varset) },
 		{ det_info_get_module_info(DetInfo, ModuleInfo) },
 		(
 			{ det_lookup_var_type(ModuleInfo, ProcInfo, Var,
@@ -454,12 +532,12 @@ det_diagnose_goal_2(call(PredId, ModeId, _, _, CallContext, _), GoalInfo,
 			PredId, ModeId),
 		Context).
 
-det_diagnose_goal_2(higher_order_call(_, _, _, _, _, _), GoalInfo,
+det_diagnose_goal_2(generic_call(GenericCall, _, _, _), GoalInfo,
 		Desired, Actual, _, _DetInfo, yes) -->
 	{ goal_info_get_context(GoalInfo, Context) },
-	prog_out__write_context(Context),
 	det_diagnose_atomic_goal(Desired, Actual,
-		report_higher_order_call_context(Context), Context).
+		report_generic_call_context(Context, GenericCall),
+		Context).
 
 det_diagnose_goal_2(unify(LT, RT, _, _, UnifyContext), GoalInfo,
 		Desired, Actual, _, DetInfo, yes) -->
@@ -515,7 +593,7 @@ det_diagnose_goal_2(not(_), GoalInfo, Desired, Actual, _, _, Diagnosed) -->
 		{ Diagnosed = no }
 	).
 
-det_diagnose_goal_2(some(_Vars, Goal), _, Desired, Actual,
+det_diagnose_goal_2(some(_Vars, _, Goal), _, Desired, Actual,
 		SwitchContext, DetInfo, Diagnosed) -->
 	{ Goal = _ - GoalInfo },
 	{ goal_info_get_determinism(GoalInfo, Internal) },
@@ -528,7 +606,7 @@ det_diagnose_goal_2(some(_Vars, Goal), _, Desired, Actual,
 	det_diagnose_goal(Goal, InternalDesired, SwitchContext, DetInfo,
 		Diagnosed).
 
-det_diagnose_goal_2(pragma_c_code(_, _, _, _, _, _, _, _), GoalInfo, Desired, 
+det_diagnose_goal_2(pragma_c_code(_, _, _, _, _, _, _), GoalInfo, Desired, 
 		_, _, _, yes) -->
 	{ goal_info_get_context(GoalInfo, Context) },
 	prog_out__write_context(Context),
@@ -536,25 +614,31 @@ det_diagnose_goal_2(pragma_c_code(_, _, _, _, _, _, _, _), GoalInfo, Desired,
 	prog_out__write_context(Context),
 	io__write_string("  determinism is "),
 	hlds_out__write_determinism(Desired),
-	io__write_string(".\n"),
-	prog_out__write_context(Context),
-	io__write_string("  pragma c_code declarations only allowed\n"),
-	prog_out__write_context(Context),
-	io__write_string("  for modes which don't succeed more than once.\n").
-	% XXX
+	io__write_string(".\n").
+	% The "clarification" below is now incorrect.
+	% prog_out__write_context(Context),
+	% io__write_string("  pragma c_code declarations only allowed\n"),
+	% prog_out__write_context(Context),
+	% io__write_string("  for modes which don't succeed more than once.\n").
+
+det_diagnose_goal_2(bi_implication(_, _), _, _, _, _, _, _) -->
+	% these should have been expanded out by now
+	{ error("det_diagnose_goal_2: unexpected bi_implication") }.
 
 %-----------------------------------------------------------------------------%
 
-:- pred report_higher_order_call_context(term__context::in,
-		io__state::di, io__state::uo) is det.
-report_higher_order_call_context(Context) -->
+:- pred report_generic_call_context(prog_context::in,
+		generic_call::in, io__state::di, io__state::uo) is det.
+report_generic_call_context(Context, CallType) -->
 	prog_out__write_context(Context),
-	io__write_string("  Higher-order call").
+	io__write_string("  "),
+	{ hlds_goal__generic_call_id(CallType, CallId) },
+	hlds_out__write_call_id(CallId).
 
 %-----------------------------------------------------------------------------%
 
 :- pred det_diagnose_atomic_goal(determinism, determinism, 
-		pred(io__state, io__state), term__context,
+		pred(io__state, io__state), prog_context,
 		io__state, io__state).
 :- mode det_diagnose_atomic_goal(in, in, pred(di, uo) is det, in,
 		di, uo) is det.
@@ -598,6 +682,9 @@ det_diagnose_atomic_goal(Desired, Actual, WriteContext, Context) -->
 		io__write_string(".\n")
 	).
 
+	% det_diagnose_conj is used for both normal [sequential]
+	% conjunction and parallel conjunction.
+
 :- pred det_diagnose_conj(list(hlds_goal), determinism,
 	list(switch_context), det_info, bool, io__state, io__state).
 :- mode det_diagnose_conj(in, in, in, in, out, di, uo) is det.
@@ -614,9 +701,9 @@ det_diagnose_conj([Goal | Goals], Desired, SwitchContext, DetInfo,
 :- mode det_diagnose_disj(in, in, in, in, in, in, out, out, di, uo) is det.
 
 det_diagnose_disj([], _Desired, _Actual, _SwitchContext, _DetInfo,
-		Clauses, Clauses, no) --> [].
+		ClausesWithSoln, ClausesWithSoln, no) --> [].
 det_diagnose_disj([Goal | Goals], Desired, Actual, SwitchContext, DetInfo,
-		Clauses0, Clauses, Diagnosed) -->
+		ClausesWithSoln0, ClausesWithSoln, Diagnosed) -->
 	{ determinism_components(Actual, ActualCanFail, _) },
 	{ determinism_components(Desired, DesiredCanFail, DesiredSolns) },
 	{ DesiredCanFail = cannot_fail, ActualCanFail = can_fail ->
@@ -635,12 +722,20 @@ det_diagnose_disj([Goal | Goals], Desired, Actual, SwitchContext, DetInfo,
 	{ determinism_components(ClauseDesired, ClauseCanFail, DesiredSolns) },
 	det_diagnose_goal(Goal, ClauseDesired, SwitchContext, DetInfo,
 		Diagnosed1),
-	{ Clauses1 is Clauses0 + 1 },
+	(
+		{ Goal = _ - GoalInfo },
+		{ goal_info_get_determinism(GoalInfo, GoalDetism) },
+		{ determinism_components(GoalDetism, _, at_most_zero) }
+	->
+		{ ClausesWithSoln1 = ClausesWithSoln0 }
+	;
+		{ ClausesWithSoln1 is ClausesWithSoln0 + 1 }
+	),
 	det_diagnose_disj(Goals, Desired, Actual, SwitchContext, DetInfo,
-		Clauses1, Clauses, Diagnosed2),
+		ClausesWithSoln1, ClausesWithSoln, Diagnosed2),
 	{ bool__or(Diagnosed1, Diagnosed2, Diagnosed) }.
 
-:- pred det_diagnose_switch(var, list(case), determinism,
+:- pred det_diagnose_switch(prog_var, list(case), determinism,
 	list(switch_context), det_info, bool, io__state, io__state).
 :- mode det_diagnose_switch(in, in, in, in, in, out, di, uo) is det.
 
@@ -687,9 +782,9 @@ det_output_consid_list([ConsId | ConsIds], First) -->
 
 %-----------------------------------------------------------------------------%
 
-:- type switch_context --->	switch_context(var, cons_id).
+:- type switch_context --->	switch_context(prog_var, cons_id).
 
-:- pred det_diagnose_write_switch_context(term__context, list(switch_context),
+:- pred det_diagnose_write_switch_context(prog_context, list(switch_context),
 	det_info, io__state, io__state).
 :- mode det_diagnose_write_switch_context(in, in, in, di, uo) is det.
 
@@ -698,7 +793,7 @@ det_diagnose_write_switch_context(Context, [SwitchContext | SwitchContexts],
 		DetInfo) -->
 	prog_out__write_context(Context),
 	{ det_get_proc_info(DetInfo, ProcInfo) },
-	{ proc_info_variables(ProcInfo, Varset) },
+	{ proc_info_varset(ProcInfo, Varset) },
 	{ SwitchContext = switch_context(Var, ConsId) },
 	io__write_string("  Inside the case "),
 	hlds_out__write_cons_id(ConsId),
@@ -709,7 +804,7 @@ det_diagnose_write_switch_context(Context, [SwitchContext | SwitchContexts],
 
 %-----------------------------------------------------------------------------%
 
-:- pred det_report_call_context(term__context, maybe(call_unify_context),
+:- pred det_report_call_context(prog_context, maybe(call_unify_context),
 	det_info, pred_id, proc_id, io__state, io__state).
 :- mode det_report_call_context(in, in, in, in, in, di, uo) is det.
 
@@ -769,8 +864,8 @@ det_report_call_context(Context, CallUnifyContext, DetInfo, PredId, ModeId) -->
 % with a capital letter) and whether it is the last part (in which case we
 % omit the word "in" on the final "... in unification ...").
 
-:- pred det_report_unify_context(bool, bool, term__context, unify_context,
-	det_info, var, unify_rhs, io__state, io__state).
+:- pred det_report_unify_context(bool, bool, prog_context, unify_context,
+	det_info, prog_var, unify_rhs, io__state, io__state).
 :- mode det_report_unify_context(in, in, in, in, in, in, in, di, uo) is det.
 
 det_report_unify_context(First0, Last, Context, UnifyContext, DetInfo, LT, RT)
@@ -778,8 +873,11 @@ det_report_unify_context(First0, Last, Context, UnifyContext, DetInfo, LT, RT)
 	hlds_out__write_unify_context(First0, UnifyContext, Context, First),
 	prog_out__write_context(Context),
 	{ det_get_proc_info(DetInfo, ProcInfo) },
-	{ proc_info_variables(ProcInfo, Varset) },
+	{ proc_info_varset(ProcInfo, Varset) },
 	{ det_info_get_module_info(DetInfo, ModuleInfo) },
+		% We don't have the inst varset - it's not in the
+		% proc_info, so we'll just make one up....
+	{ varset__init(InstVarSet) },
 	( { First = yes } ->
 		( { Last = yes } ->
 			io__write_string("  Unification ")
@@ -806,18 +904,20 @@ det_report_unify_context(First0, Last, Context, UnifyContext, DetInfo, LT, RT)
 			io__write_string("of `"),
 			mercury_output_var(LT, Varset, no),
 			io__write_string("' and `"),
-			hlds_out__write_unify_rhs(RT, ModuleInfo, Varset, no, 3)
+			hlds_out__write_unify_rhs(RT, ModuleInfo, Varset,
+				InstVarSet, no, 3)
 		)
 	;
 		io__write_string("with `"),
-		hlds_out__write_unify_rhs(RT, ModuleInfo, Varset, no, 3)
+		hlds_out__write_unify_rhs(RT, ModuleInfo, Varset, InstVarSet,
+			no, 3)
 	),
 	io__write_string("'").
 
 
 %-----------------------------------------------------------------------------%
 
-:- type det_msg_type	--->	warning ; error.
+:- type det_msg_type	--->	simple_code_warning ; call_warning ; error.
 
 det_report_and_handle_msgs(Msgs, ModuleInfo0, ModuleInfo) -->
 	( { Msgs = [] } ->
@@ -842,24 +942,34 @@ det_report_and_handle_msgs(Msgs, ModuleInfo0, ModuleInfo) -->
 	).
 
 det_report_msgs(Msgs, ModuleInfo, WarnCnt, ErrCnt) -->
-	globals__io_lookup_bool_option(warn_simple_code, Warn),
-	det_report_msgs_2(Msgs, Warn, ModuleInfo, 0, WarnCnt, 0, ErrCnt).
+	globals__io_lookup_bool_option(warn_simple_code, WarnSimple),
+	globals__io_lookup_bool_option(warn_duplicate_calls, WarnCalls),
+	det_report_msgs_2(Msgs, WarnSimple, WarnCalls, ModuleInfo,
+		0, WarnCnt, 0, ErrCnt).
 
-:- pred det_report_msgs_2(list(det_msg), bool,  module_info, int, int,
+:- pred det_report_msgs_2(list(det_msg), bool, bool, module_info, int, int,
 	int, int, io__state, io__state).
-:- mode det_report_msgs_2(in, in, in, in, out, in, out, di, uo) is det.
+:- mode det_report_msgs_2(in, in, in, in, in, out, in, out, di, uo) is det.
 
-det_report_msgs_2([], _, _ModuleInfo, WarnCnt, WarnCnt, ErrCnt, ErrCnt) --> [].
-det_report_msgs_2([Msg | Msgs], Warn, ModuleInfo,
+det_report_msgs_2([], _, _, _ModuleInfo,
+		WarnCnt, WarnCnt, ErrCnt, ErrCnt) --> [].
+det_report_msgs_2([Msg | Msgs], WarnSimple, WarnCalls, ModuleInfo,
 		WarnCnt0, WarnCnt, ErrCnt0, ErrCnt) -->
 	{ det_msg_get_type(Msg, MsgType) },
-	( { Warn = no, MsgType = warning } ->
+	( { WarnSimple = no, MsgType = simple_code_warning } ->
+		{ WarnCnt1 = WarnCnt0 },
+		{ ErrCnt1 = ErrCnt0 }
+	; { WarnCalls = no, MsgType = call_warning } ->
 		{ WarnCnt1 = WarnCnt0 },
 		{ ErrCnt1 = ErrCnt0 }
 	;
 		det_report_msg(Msg, ModuleInfo),
 		(
-			{ MsgType = warning },
+			{ MsgType = simple_code_warning },
+			{ WarnCnt1 is WarnCnt0 + 1 },
+			{ ErrCnt1 = ErrCnt0 }
+		;
+			{ MsgType = call_warning },
 			{ WarnCnt1 is WarnCnt0 + 1 },
 			{ ErrCnt1 = ErrCnt0 }
 		;
@@ -868,29 +978,57 @@ det_report_msgs_2([Msg | Msgs], Warn, ModuleInfo,
 			{ WarnCnt1 = WarnCnt0 }
 		)
 	),
-	det_report_msgs_2(Msgs, Warn, ModuleInfo,
+	det_report_msgs_2(Msgs, WarnSimple, WarnCalls, ModuleInfo,
 		WarnCnt1, WarnCnt, ErrCnt1, ErrCnt).
 
 :- pred det_msg_get_type(det_msg, det_msg_type).
 :- mode det_msg_get_type(in, out) is det.
 
-det_msg_get_type(multidet_disj(_, _), warning).
-det_msg_get_type(det_disj(_, _), warning).
-det_msg_get_type(semidet_disj(_, _), warning).
-det_msg_get_type(zero_soln_disj(_, _), warning).
-det_msg_get_type(zero_soln_disjunct(_), warning).
-det_msg_get_type(ite_cond_cannot_fail(_), warning).
-det_msg_get_type(ite_cond_cannot_succeed(_), warning).
-det_msg_get_type(negated_goal_cannot_fail(_), warning).
-det_msg_get_type(negated_goal_cannot_succeed(_), warning).
-det_msg_get_type(warn_obsolete(_, _), warning).
-det_msg_get_type(warn_infinite_recursion(_), warning).
-det_msg_get_type(duplicate_call(_, _, _), warning).
+det_msg_get_type(multidet_disj(_, _), simple_code_warning).
+det_msg_get_type(det_disj(_, _), simple_code_warning).
+det_msg_get_type(semidet_disj(_, _), simple_code_warning).
+det_msg_get_type(zero_soln_disj(_, _), simple_code_warning).
+det_msg_get_type(zero_soln_disjunct(_), simple_code_warning).
+det_msg_get_type(ite_cond_cannot_fail(_), simple_code_warning).
+det_msg_get_type(ite_cond_cannot_succeed(_), simple_code_warning).
+det_msg_get_type(negated_goal_cannot_fail(_), simple_code_warning).
+det_msg_get_type(negated_goal_cannot_succeed(_), simple_code_warning).
+det_msg_get_type(goal_cannot_succeed(_), simple_code_warning).
+det_msg_get_type(det_goal_has_no_outputs(_), simple_code_warning).
+	% XXX warn_obsolete isn't really a simple code warning.
+	% We should add a separate warning type for this.
+det_msg_get_type(warn_obsolete(_, _), simple_code_warning).
+det_msg_get_type(warn_infinite_recursion(_), simple_code_warning).
+det_msg_get_type(duplicate_call(_, _, _), call_warning).
 det_msg_get_type(cc_unify_can_fail(_, _, _, _, _), error).
 det_msg_get_type(cc_unify_in_wrong_context(_, _, _, _, _), error).
 det_msg_get_type(cc_pred_in_wrong_context(_, _, _, _), error).
 det_msg_get_type(higher_order_cc_pred_in_wrong_context(_, _), error).
 det_msg_get_type(error_in_lambda(_, _, _, _, _, _), error).
+det_msg_get_type(par_conj_not_det(_, _, _, _, _), error).
+det_msg_get_type(pragma_c_code_without_det_decl(_, _), error).
+
+det_msg_is_any_mode_msg(multidet_disj(_, _), all_modes).
+det_msg_is_any_mode_msg(det_disj(_, _), all_modes).
+det_msg_is_any_mode_msg(semidet_disj(_, _), all_modes).
+det_msg_is_any_mode_msg(zero_soln_disj(_, _), all_modes).
+det_msg_is_any_mode_msg(zero_soln_disjunct(_), all_modes).
+det_msg_is_any_mode_msg(ite_cond_cannot_fail(_), all_modes).
+det_msg_is_any_mode_msg(ite_cond_cannot_succeed(_), all_modes).
+det_msg_is_any_mode_msg(negated_goal_cannot_fail(_), all_modes).
+det_msg_is_any_mode_msg(negated_goal_cannot_succeed(_), all_modes).
+det_msg_is_any_mode_msg(goal_cannot_succeed(_), all_modes).
+det_msg_is_any_mode_msg(det_goal_has_no_outputs(_), all_modes).
+det_msg_is_any_mode_msg(warn_obsolete(_, _), all_modes).
+det_msg_is_any_mode_msg(warn_infinite_recursion(_), any_mode).
+det_msg_is_any_mode_msg(duplicate_call(_, _, _), any_mode).
+det_msg_is_any_mode_msg(cc_unify_can_fail(_, _, _, _, _), any_mode).
+det_msg_is_any_mode_msg(cc_unify_in_wrong_context(_, _, _, _, _), any_mode).
+det_msg_is_any_mode_msg(cc_pred_in_wrong_context(_, _, _, _), any_mode).
+det_msg_is_any_mode_msg(higher_order_cc_pred_in_wrong_context(_, _), any_mode).
+det_msg_is_any_mode_msg(error_in_lambda(_, _, _, _, _, _), any_mode).
+det_msg_is_any_mode_msg(par_conj_not_det(_, _, _, _, _), any_mode).
+det_msg_is_any_mode_msg(pragma_c_code_without_det_decl(_, _), any_mode).
 
 :- pred det_report_msg(det_msg, module_info, io__state, io__state).
 :- mode det_report_msg(in, in, di, uo) is det.
@@ -938,6 +1076,28 @@ det_report_msg(negated_goal_cannot_fail(Context), _) -->
 det_report_msg(negated_goal_cannot_succeed(Context), _) -->
 	prog_out__write_context(Context),
 	io__write_string("Warning: the negated goal cannot succeed.\n").
+det_report_msg(goal_cannot_succeed(Context), _) -->
+	prog_out__write_context(Context),
+	io__write_string("Warning: this goal cannot succeed.\n"),
+	globals__io_lookup_bool_option(verbose_errors, VerboseErrors),
+	( { VerboseErrors = yes } ->
+		io__write_string(
+"\tThe compiler will optimize away this goal, replacing it with `fail'.
+\tTo disable this optimization, use the `--fully-strict' option.\n")
+	;
+		[]
+	).
+det_report_msg(det_goal_has_no_outputs(Context), _) -->
+	prog_out__write_context(Context),
+	io__write_string("Warning: det goal has no outputs.\n"),
+	globals__io_lookup_bool_option(verbose_errors, VerboseErrors),
+	( { VerboseErrors = yes } ->
+		io__write_string(
+"\tThe compiler will optimize away this goal, replacing it with `true'.
+\tTo disable this optimization, use the `--fully-strict' option.\n")
+	;
+		[]
+	).
 det_report_msg(warn_obsolete(PredId, Context), ModuleInfo) -->
 	prog_out__write_context(Context),
 	io__write_string("Warning: call to obsolete "),
@@ -1105,6 +1265,42 @@ det_report_msg(error_in_lambda(DeclaredDetism, InferredDetism, Goal, GoalInfo,
 	{ det_info_init(ModuleInfo, PredId, ProcId, Globals, DetInfo) },
 	det_diagnose_goal(Goal, DeclaredDetism, [], DetInfo, _),
 	io__set_exit_status(1).
+det_report_msg(par_conj_not_det(InferredDetism, PredId,
+			ProcId, GoalInfo, Goals), ModuleInfo) -->
+	{ goal_info_get_context(GoalInfo, Context) },
+	prog_out__write_context(Context),
+	{ determinism_components(InferredDetism, CanFail, MaxSoln) },
+	(
+		{ CanFail \= cannot_fail }
+	->
+		io__write_string("Error: parallel conjunct may fail.\n")
+	;
+		{ MaxSoln = at_most_many }
+	->
+		prog_out__write_context(Context),
+		io__write_string("Error: parallel conjunct may have multiple solutions.\n")
+	;
+		{ error("strange determinism error for parallel conjunction") }
+	),
+	prog_out__write_context(Context),
+	io__write_string(
+		"  The current implementation supports only single-solution\n"
+	),
+	prog_out__write_context(Context),
+	io__write_string("  non-failing parallel conjunctions.\n"),
+	globals__io_get_globals(Globals),
+	{ det_info_init(ModuleInfo, PredId, ProcId, Globals, DetInfo) },
+	det_diagnose_conj(Goals, det, [], DetInfo, _),
+	io__set_exit_status(1).
+det_report_msg(pragma_c_code_without_det_decl(PredId, ProcId),
+		ModuleInfo) -->
+	report_pred_proc_id(ModuleInfo, PredId, ProcId, no, Context),
+	prog_out__write_context(Context),	
+	io__write_string("  error: `:- pragma c_code(...)' for a procedure"),
+	io__nl,
+	prog_out__write_context(Context),	
+	io__write_string("  without a determinism declaration."),
+	io__nl.
 
 %-----------------------------------------------------------------------------%
 
@@ -1122,7 +1318,7 @@ det_report_seen_call_id(SeenCall, ModuleInfo) -->
 	).
 %-----------------------------------------------------------------------------%
 
-:- pred det_report_context_lines(list(term__context), bool, 
+:- pred det_report_context_lines(list(prog_context), bool, 
 		io__state, io__state).
 :- mode det_report_context_lines(in, in, di, uo) is det.
 
