@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1997 The University of Melbourne.
+% Copyright (C) 1996-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -17,7 +17,7 @@
 :- interface.
 
 :- import_module hlds_module, hlds_pred, hlds_goal, hlds_data, globals.
-:- import_module instmap.
+:- import_module instmap, prog_data.
 :- import_module bool, set, list.
 
 :- type maybe_changed	--->	changed ; unchanged.
@@ -43,7 +43,8 @@
 	% Update the current substitution to account for the effects
 	% of the given unification.
 
-:- pred interpret_unify(var, unify_rhs, substitution, substitution).
+:- pred interpret_unify(prog_var, unify_rhs, prog_substitution,
+		prog_substitution).
 :- mode interpret_unify(in, in, in, out) is semidet.
 
 	% Look up the determinism of a procedure.
@@ -54,10 +55,10 @@
 :- pred det_get_proc_info(det_info, proc_info).
 :- mode det_get_proc_info(in, out) is det.
 
-:- pred det_lookup_var_type(module_info, proc_info, var, hlds_type_defn).
+:- pred det_lookup_var_type(module_info, proc_info, prog_var, hlds_type_defn).
 :- mode det_lookup_var_type(in, in, in, out) is semidet.
 
-:- pred det_no_output_vars(set(var), instmap, instmap_delta, det_info).
+:- pred det_no_output_vars(set(prog_var), instmap, instmap_delta, det_info).
 :- mode det_no_output_vars(in, in, in, in) is semidet.
 
 :- pred det_info_init(module_info, pred_id, proc_id, globals, det_info).
@@ -88,8 +89,8 @@
 
 :- implementation.
 
-:- import_module inst_match, mode_util, type_util, options.
-:- import_module map, term, require, std_util.
+:- import_module inst_match, mode_util, type_util, options, term.
+:- import_module map, require, std_util.
 
 update_instmap(_Goal0 - GoalInfo0, InstMap0, InstMap) :-
 	goal_info_get_instmap_delta(GoalInfo0, DeltaInstMap),
@@ -109,13 +110,14 @@ delete_unreachable_cases([Case | Cases0], [ConsId | ConsIds], Cases) :-
 	).
 
 interpret_unify(X, var(Y), Subst0, Subst) :-
-	term__unify(term__variable(X), term__variable(Y), Subst0, Subst).
+	term__unify(term__variable(X), term__variable(Y),
+		Subst0, Subst).
 interpret_unify(X, functor(ConsId, ArgVars), Subst0, Subst) :-
 	term__var_list_to_term_list(ArgVars, ArgTerms),
 	cons_id_and_args_to_term(ConsId, ArgTerms, RhsTerm),
 	term__unify(term__variable(X), RhsTerm, Subst0, Subst).
-interpret_unify(_X, lambda_goal(_PredOrFunc, _LambdaVars, _Modes, _Det, _Goal),
-		Subst0, Subst) :-
+interpret_unify(_X, lambda_goal(_POrF, _Method, _Fix, _NonLocals,
+			_Vars, _Modes, _Det, _Goal), Subst0, Subst) :-
 		% For ease of implementation we just ignore unifications with
 		% lambda terms.  This is a safe approximation, it just
 		% prevents us from optimizing them as well as we would like.
