@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1997 The University of Melbourne.
+% Copyright (C) 1996-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -11,7 +11,7 @@
 
 :- module equiv_type.
 :- interface.
-:- import_module prog_data, list, io.
+:- import_module bool, prog_data, list, io.
 
 %-----------------------------------------------------------------------------%
 
@@ -36,8 +36,11 @@
 %-----------------------------------------------------------------------------%
 
 :- implementation.
-:- import_module bool, require, std_util, map, term, varset.
-:- import_module hlds_data, type_util, prog_data, prog_util, prog_out.
+:- import_module assoc_list, bool, require, std_util, map, term, varset.
+:- import_module prog_data, prog_util, prog_out.
+
+% XXX we shouldn't import the HLDS here.
+:- import_module hlds_data, type_util.
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
@@ -112,23 +115,63 @@ equiv_type__replace_in_item(type_defn(VarSet0, TypeDefn0, Cond),
 	equiv_type__replace_in_type_defn(TypeDefn0, VarSet0, EqvMap,
 				TypeDefn, VarSet, ContainsCirc).
 
-equiv_type__replace_in_item(pred(VarSet0, PredName, TypesAndModes0, Det, Cond),
-		EqvMap, pred(VarSet, PredName, TypesAndModes, Det, Cond), no) :-
-	equiv_type__replace_in_tms(TypesAndModes0, VarSet0, EqvMap, 
-					TypesAndModes, VarSet).
+equiv_type__replace_in_item(
+		pred(TypeVarSet0, InstVarSet, ExistQVars, PredName,
+			TypesAndModes0, Det, Cond, Purity, ClassContext0),
+		EqvMap,
+		pred(TypeVarSet, InstVarSet, ExistQVars, PredName,
+			TypesAndModes, Det, Cond, Purity, ClassContext),
+		no) :-
+	equiv_type__replace_in_class_constraints(ClassContext0, TypeVarSet0, 
+				EqvMap, ClassContext, TypeVarSet1),
+	equiv_type__replace_in_tms(TypesAndModes0, TypeVarSet1, EqvMap, 
+					TypesAndModes, TypeVarSet).
 
 equiv_type__replace_in_item(
-			func(VarSet0, PredName, TypesAndModes0, 
-				RetTypeAndMode0, Det, Cond),
+			func(TypeVarSet0, InstVarSet, ExistQVars, PredName,
+				TypesAndModes0, RetTypeAndMode0, Det, Cond,
+				Purity, ClassContext0),
 			EqvMap,
-			func(VarSet, PredName, TypesAndModes, RetTypeAndMode,
-				Det, Cond),
+			func(TypeVarSet, InstVarSet, ExistQVars, PredName,
+				TypesAndModes, RetTypeAndMode, Det, Cond,
+				Purity, ClassContext),
 			no) :-
-	equiv_type__replace_in_tms(TypesAndModes0, VarSet0, EqvMap,
-				TypesAndModes, VarSet1),
-	equiv_type__replace_in_tm(RetTypeAndMode0, VarSet1, EqvMap,
-				RetTypeAndMode, VarSet).
+	equiv_type__replace_in_class_constraints(ClassContext0, TypeVarSet0, 
+				EqvMap, ClassContext, TypeVarSet1),
+	equiv_type__replace_in_tms(TypesAndModes0, TypeVarSet1, EqvMap,
+				TypesAndModes, TypeVarSet2),
+	equiv_type__replace_in_tm(RetTypeAndMode0, TypeVarSet2, EqvMap,
+				RetTypeAndMode, TypeVarSet).
 
+equiv_type__replace_in_item(
+			typeclass(Constraints0, ClassName, Vars, 
+				ClassInterface0, VarSet0),
+			EqvMap,
+			typeclass(Constraints, ClassName, Vars, 
+				ClassInterface, VarSet),
+			no) :-
+	equiv_type__replace_in_class_constraint_list(Constraints0, VarSet0, 
+				EqvMap, Constraints, VarSet),
+	equiv_type__replace_in_class_interface(ClassInterface0,
+				EqvMap, ClassInterface).
+
+equiv_type__replace_in_item(
+			instance(Constraints0, ClassName, Ts0, 
+				InstanceBody, VarSet0),
+			EqvMap,
+			instance(Constraints, ClassName, Ts, 
+				InstanceBody, VarSet),
+			no) :-
+	equiv_type__replace_in_class_constraint_list(Constraints0, VarSet0, 
+				EqvMap, Constraints, VarSet1),
+	equiv_type__replace_in_type_list(Ts0, VarSet1, EqvMap, Ts, VarSet, _).
+
+equiv_type__replace_in_item(
+			pragma(type_spec(A, B, C, D, E, Subst0, VarSet0)),
+			EqvMap,
+			pragma(type_spec(A, B, C, D, E, Subst, VarSet)), no) :-
+	equiv_type__replace_in_subst(Subst0, VarSet0, EqvMap, Subst, VarSet).
+	
 :- pred equiv_type__replace_in_type_defn(type_defn, tvarset, eqv_map,
 					type_defn, tvarset, bool).
 :- mode equiv_type__replace_in_type_defn(in, in, in, out, out, out) is semidet.
@@ -150,6 +193,108 @@ equiv_type__replace_in_type_defn(du_type(TName, TArgs, TBody0, EqPred), VarSet0,
 
 %-----------------------------------------------------------------------------%
 
+:- pred equiv_type__replace_in_class_constraints(class_constraints, 
+			tvarset, eqv_map, class_constraints, tvarset).
+:- mode equiv_type__replace_in_class_constraints(in, in, in, out, out) is det.
+
+equiv_type__replace_in_class_constraints(Cs0, VarSet0, EqvMap, Cs, VarSet) :-
+	Cs0 = constraints(UnivCs0, ExistCs0),
+	Cs = constraints(UnivCs, ExistCs),
+	equiv_type__replace_in_class_constraint_list(UnivCs0, VarSet0, EqvMap,
+		UnivCs, VarSet1),
+	equiv_type__replace_in_class_constraint_list(ExistCs0, VarSet1, EqvMap,
+		ExistCs, VarSet).
+
+:- pred equiv_type__replace_in_class_constraint_list(list(class_constraint), 
+			tvarset, eqv_map, list(class_constraint), tvarset).
+:- mode equiv_type__replace_in_class_constraint_list(in, in, in, out, out)
+			is det.
+
+equiv_type__replace_in_class_constraint_list([], VarSet, _, [], VarSet).
+equiv_type__replace_in_class_constraint_list([C0|C0s], VarSet0, EqvMap, 
+				[C|Cs], VarSet) :-
+	equiv_type__replace_in_class_constraint(C0, VarSet0, EqvMap, C,
+		VarSet1),
+	equiv_type__replace_in_class_constraint_list(C0s, VarSet1, EqvMap, Cs, 
+		VarSet).
+
+:- pred equiv_type__replace_in_class_constraint(class_constraint, tvarset, 
+					eqv_map, class_constraint, tvarset).
+:- mode equiv_type__replace_in_class_constraint(in, in, in, out, out) is det.
+
+equiv_type__replace_in_class_constraint(Constraint0, VarSet0, EqvMap,
+				Constraint, VarSet) :-
+	Constraint0 = constraint(ClassName, Ts0),
+	equiv_type__replace_in_type_list(Ts0, VarSet0, EqvMap, Ts1, VarSet, _),
+	% we must maintain the invariant that types in class constraints
+	% do not contain any info in their prog_context fields
+	strip_prog_contexts(Ts1, Ts),
+	Constraint = constraint(ClassName, Ts).
+
+%-----------------------------------------------------------------------------%
+
+:- pred equiv_type__replace_in_class_interface(class_interface,
+					eqv_map, class_interface).
+:- mode equiv_type__replace_in_class_interface(in, in, out) is det.
+
+equiv_type__replace_in_class_interface(ClassInterface0, EqvMap,
+				ClassInterface) :-
+	list__map(equiv_type__replace_in_class_method(EqvMap),
+		ClassInterface0, ClassInterface).
+
+:- pred equiv_type__replace_in_class_method(eqv_map, class_method, 
+					class_method).
+:- mode equiv_type__replace_in_class_method(in, in, out) is det.
+
+equiv_type__replace_in_class_method(EqvMap,
+			pred(TypeVarSet0, InstVarSet, ExistQVars, PredName,
+				TypesAndModes0, Det, Cond, ClassContext0,
+				Context),
+			pred(TypeVarSet, InstVarSet, ExistQVars, PredName,
+				TypesAndModes, Det, Cond, ClassContext, Context)
+			) :-
+	equiv_type__replace_in_class_constraints(ClassContext0, TypeVarSet0, 
+				EqvMap, ClassContext, TypeVarSet1),
+	equiv_type__replace_in_tms(TypesAndModes0, TypeVarSet1, EqvMap, 
+					TypesAndModes, TypeVarSet).
+
+equiv_type__replace_in_class_method(EqvMap,
+			func(TypeVarSet0, InstVarSet, ExistQVars, PredName,
+				TypesAndModes0, RetTypeAndMode0, Det, Cond,
+				ClassContext0, Context),
+			func(TypeVarSet, InstVarSet, ExistQVars, PredName,
+				TypesAndModes, RetTypeAndMode, Det, Cond,
+				ClassContext, Context)
+			) :-
+	equiv_type__replace_in_class_constraints(ClassContext0, TypeVarSet0, 
+				EqvMap, ClassContext, TypeVarSet1),
+	equiv_type__replace_in_tms(TypesAndModes0, TypeVarSet1, EqvMap,
+				TypesAndModes, TypeVarSet2),
+	equiv_type__replace_in_tm(RetTypeAndMode0, TypeVarSet2, EqvMap,
+				RetTypeAndMode, TypeVarSet).
+
+equiv_type__replace_in_class_method(_,
+			pred_mode(A,B,C,D,E,F),
+			pred_mode(A,B,C,D,E,F)).
+
+equiv_type__replace_in_class_method(_, 
+			func_mode(A,B,C,D,E,F,G),
+			func_mode(A,B,C,D,E,F,G)).
+
+%-----------------------------------------------------------------------------%
+
+:- pred equiv_type__replace_in_subst(assoc_list(tvar, type), tvarset,
+		eqv_map, assoc_list(tvar, type), tvarset).
+:- mode equiv_type__replace_in_subst(in, in, in, out, out) is det.
+
+equiv_type__replace_in_subst([], VarSet, _EqvMap, [], VarSet).
+equiv_type__replace_in_subst([Var - Type0 | Subst0], VarSet0,
+		EqvMap, [Var - Type | Subst], VarSet) :-
+	equiv_type__replace_in_type(Type0, VarSet0, EqvMap, Type, VarSet1),
+	equiv_type__replace_in_subst(Subst0, VarSet1, EqvMap, Subst, VarSet).
+
+%-----------------------------------------------------------------------------%
+
 :- pred equiv_type__replace_in_uu(list(type), tvarset, eqv_map,
 					list(type), tvarset).
 :- mode equiv_type__replace_in_uu(in, in, in, out, out) is det.
@@ -174,10 +319,13 @@ equiv_type__replace_in_du([T0|Ts0], VarSet0, EqvMap, [T|Ts], VarSet) :-
 				constructor, tvarset).
 :- mode equiv_type__replace_in_ctor(in, in, in, out, out) is det.
 
-equiv_type__replace_in_ctor(TName - Targs0, VarSet0, EqvMap,
-		TName - Targs, VarSet) :-
+equiv_type__replace_in_ctor(ctor(ExistQVars, Constraints0, TName, Targs0),
+		VarSet0, EqvMap,
+		ctor(ExistQVars, Constraints, TName, Targs), VarSet) :-
 	equiv_type__replace_in_ctor_arg_list(Targs0, VarSet0, EqvMap,
-		Targs, VarSet, _).
+				Targs, VarSet1, _),
+	equiv_type__replace_in_class_constraint_list(Constraints0, VarSet1, 
+				EqvMap, Constraints, VarSet).
 
 %-----------------------------------------------------------------------------%
 
@@ -263,8 +411,20 @@ equiv_type__replace_in_type_2(Type0, VarSet0, EqvMap,
 		(	
 			map__search(EqvMap, EqvTypeId,
 				eqv_type_body(EqvVarSet, Args0, Body0)),
-			varset__merge(VarSet1, EqvVarSet, [Body0 | Args0],
-					VarSet2, [Body | Args]),
+			%
+			% Don't merge in the variable names from the
+			% type declaration to avoid creating multiple
+			% variables with the same name so that
+			% `varset__create_name_var_map' can be used
+			% on the resulting tvarset.
+			% make_hlds.m uses `varset__create_name_var_map' to
+			% match up type variables in `:- pragma type_spec'
+			% declarations and explicit type qualifications
+			% with the type variables in the predicate's
+			% declaration.
+			%
+			varset__merge_without_names(VarSet1, EqvVarSet,
+				[Body0 | Args0], VarSet2, [Body | Args]),
 			Circ0 = no,
 			Circ1 = no
 		->
