@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -42,8 +42,8 @@
 
 :- implementation.
 
-:- import_module hlds_goal, goal_util.
-:- import_module varset, list, bool, map, set, std_util, term.
+:- import_module hlds_goal, goal_util, prog_data, varset, term.
+:- import_module list, bool, map, set, std_util, require.
 
 %-----------------------------------------------------------------------------%
 
@@ -53,7 +53,7 @@ excess_assignments_proc(ProcInfo0, _ModuleInfo, ProcInfo) :-
 	proc_info_set_goal(ProcInfo0, Goal, ProcInfo1),
 
 	% XXX We probably ought to remove these vars from the type map as well.
-	proc_info_variables(ProcInfo1, Varset0),
+	proc_info_varset(ProcInfo1, Varset0),
 	varset__delete_vars(Varset0, ElimVars, Varset),
 	proc_info_set_varset(ProcInfo1, Varset, ProcInfo).
 
@@ -79,8 +79,8 @@ excess_assignments_proc(ProcInfo0, _ModuleInfo, ProcInfo) :-
 % and the notation `<Foo> [X/Y]' means <Foo> with all
 % occurrences of `X' replaced with `Y'.
 
-:- pred excess_assignments_in_goal(hlds_goal, list(var),
-				   hlds_goal, list(var)).
+:- pred excess_assignments_in_goal(hlds_goal, list(prog_var),
+				   hlds_goal, list(prog_var)).
 :- mode excess_assignments_in_goal(in, in, out, out) is det.
 
 excess_assignments_in_goal(GoalExpr0 - GoalInfo0, ElimVars0, Goal, ElimVars) :-
@@ -90,6 +90,12 @@ excess_assignments_in_goal(GoalExpr0 - GoalInfo0, ElimVars0, Goal, ElimVars) :-
 		excess_assignments_in_conj(Goals0, [], ElimVars0, NonLocals,
 					Goals, ElimVars),
 		conj_list_to_goal(Goals, GoalInfo0, Goal)
+	;
+		GoalExpr0 = par_conj(Goals0, _SM),
+		goal_info_get_nonlocals(GoalInfo0, NonLocals),
+		excess_assignments_in_conj(Goals0, [], ElimVars0, NonLocals,
+					Goals, ElimVars),
+		par_conj_list_to_goal(Goals, GoalInfo0, Goal)
 	;
 		GoalExpr0 = disj(Goals0, SM),
 		excess_assignments_in_disj(Goals0, ElimVars0, Goals, ElimVars),
@@ -111,12 +117,12 @@ excess_assignments_in_goal(GoalExpr0 - GoalInfo0, ElimVars0, Goal, ElimVars) :-
 		excess_assignments_in_goal(Else0, ElimVars2, Else, ElimVars),
 		Goal = if_then_else(Vars, Cond, Then, Else, SM) - GoalInfo0
 	;
-		GoalExpr0 = some(Var, SubGoal0),
+		GoalExpr0 = some(Var, CanRemove, SubGoal0),
 		excess_assignments_in_goal(SubGoal0, ElimVars0,
 					   SubGoal, ElimVars),
-		Goal = some(Var, SubGoal) - GoalInfo0
+		Goal = some(Var, CanRemove, SubGoal) - GoalInfo0
 	;
-		GoalExpr0 = higher_order_call(_, _, _, _, _, _),
+		GoalExpr0 = generic_call(_, _, _, _),
 		Goal = GoalExpr0 - GoalInfo0,
 		ElimVars = ElimVars0
 	;
@@ -128,9 +134,13 @@ excess_assignments_in_goal(GoalExpr0 - GoalInfo0, ElimVars0, Goal, ElimVars) :-
 		Goal = GoalExpr0 - GoalInfo0,
 		ElimVars = ElimVars0
 	;
-		GoalExpr0 = pragma_c_code(_, _, _, _, _, _, _, _),
+		GoalExpr0 = pragma_c_code(_, _, _, _, _, _, _),
 		Goal = GoalExpr0 - GoalInfo0,
 		ElimVars = ElimVars0
+	;
+		GoalExpr0 = bi_implication(_, _),
+		% these should have been expanded out by now
+		error("detect_cse_in_goal_2: unexpected bi_implication")
 	),
 	!.
 
@@ -142,9 +152,12 @@ excess_assignments_in_goal(GoalExpr0 - GoalInfo0, ElimVars0, Goal, ElimVars) :-
 	% If (say) V_4 and V_6 are nonlocal, then after the V_5 => V_4
 	% substitution has been made, the second assignment V_4 = V_6
 	% is left alone.
+	%
+	% This code is used for both sequential conjunction (conj/1) and
+	% parallel conjunction (par_conj/2).
 
 :- pred excess_assignments_in_conj(list(hlds_goal), list(hlds_goal),
-	list(var), set(var), list(hlds_goal), list(var)).
+	list(prog_var), set(prog_var), list(hlds_goal), list(prog_var)).
 :- mode excess_assignments_in_conj(in, in, in, in, out, out) is det.
 
 excess_assignments_in_conj([], RevGoals, ElimVars, _, Goals, ElimVars) :-
@@ -177,8 +190,8 @@ excess_assignments_in_conj([Goal0 | Goals0], RevGoals0, ElimVars0, NonLocals,
 
 %-----------------------------------------------------------------------------%
 
-:- pred excess_assignments_in_disj(list(hlds_goal), list(var),
-	list(hlds_goal), list(var)).
+:- pred excess_assignments_in_disj(list(hlds_goal), list(prog_var),
+	list(hlds_goal), list(prog_var)).
 :- mode excess_assignments_in_disj(in, in, out, out) is det.
 
 excess_assignments_in_disj([], ElimVars, [], ElimVars).
@@ -187,8 +200,8 @@ excess_assignments_in_disj([Goal0 | Goals0], ElimVars0,
 	excess_assignments_in_goal(Goal0, ElimVars0, Goal, ElimVars1),
 	excess_assignments_in_disj(Goals0, ElimVars1, Goals, ElimVars).
 
-:- pred excess_assignments_in_switch(list(case), list(var),
-				     list(case), list(var)).
+:- pred excess_assignments_in_switch(list(case), list(prog_var),
+				     list(case), list(prog_var)).
 :- mode excess_assignments_in_switch(in, in, out, out) is det.
 
 excess_assignments_in_switch([], ElimVars, [], ElimVars).
