@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1997 The University of Melbourne.
+% Copyright (C) 1996-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -7,6 +7,9 @@
 % This module defines predicates to produce the functions which are
 % exported to C via a `pragma export' declaration.
 
+% Note: any changes here might also require similar changes to the handling
+% of `pragma import' declarations, which are handled in make_hlds.m.
+
 % Main authors: dgj.
 
 %-----------------------------------------------------------------------------%
@@ -15,23 +18,33 @@
 
 :- interface.
 
-:- import_module hlds_module, prog_data.
-:- import_module io, list, term.
+:- import_module hlds_module, prog_data, llds.
+:- import_module io.
 
-	% From the module_info, get a list of functions, each of which allows
-	% a call to be made to a Mercury procedure from C
-:- pred export__get_pragma_exported_procs(module_info, list(string)).
-:- mode export__get_pragma_exported_procs(in, out) is det.
+	% From the module_info, get a list of c_export_decls,
+	% each of which holds information about the declaration
+	% of a C function named in a `pragma export' declaration,
+	% which is used to allow a call to be made to a Mercury
+	% procedure from C.
+:- pred export__get_c_export_decls(module_info, c_export_decls).
+:- mode export__get_c_export_decls(in, out) is det.
+
+	% From the module_info, get a list of c_export_defns,
+	% each of which is a string containing the C code
+	% for defining a C function named in a `pragma export' decl.
+:- pred export__get_c_export_defns(module_info, c_export_defns).
+:- mode export__get_c_export_defns(in, out) is det.
 
 	% Produce a header file containing prototypes for the exported C
 	% functions
-:- pred export__produce_header_file(module_info, string, io__state, io__state).
+:- pred export__produce_header_file(c_export_decls, module_name,
+					io__state, io__state).
 :- mode export__produce_header_file(in, in, di, uo) is det.
 
-	% Convert the term, which represents a type, to a string corresponding
-	% to its C type. (Defaults to Word).
-:- pred export__term_to_type_string(term, string).
-:- mode export__term_to_type_string(in, out) is det.
+	% Convert the type, to a string corresponding to its C type.
+	% (Defaults to Word).
+:- pred export__type_to_type_string(type, string).
+:- mode export__type_to_type_string(in, out) is det.
 
 	% Generate C code to convert an rval (represented as a string), from
 	% a C type to a mercury C type (ie. convert strings and floats to
@@ -45,13 +58,44 @@
 :- pred convert_type_from_mercury(string, type, string).
 :- mode convert_type_from_mercury(in, in, out) is det.
 
+%-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
+
 :- implementation.
 
-:- import_module code_gen, code_util, hlds_pred, llds, llds_out.
-:- import_module library, map, int, string, std_util, assoc_list, require.
-:- import_module bool.
+:- import_module code_gen, code_util, hlds_pred, llds_out, modules.
+:- import_module type_util.
 
-export__get_pragma_exported_procs(Module, ExportedProcsCode) :-
+:- import_module term, varset.
+:- import_module library, map, int, string, std_util, assoc_list, require.
+:- import_module list, bool.
+
+%-----------------------------------------------------------------------------%
+
+export__get_c_export_decls(HLDS, C_ExportDecls) :-
+	module_info_get_predicate_table(HLDS, PredicateTable),
+	predicate_table_get_preds(PredicateTable, Preds),
+	module_info_get_pragma_exported_procs(HLDS, ExportedProcs),
+	export__get_c_export_decls_2(Preds, ExportedProcs, C_ExportDecls).
+
+:- pred export__get_c_export_decls_2(pred_table, list(pragma_exported_proc),
+	list(c_export_decl)).
+:- mode export__get_c_export_decls_2(in, in, out) is det.
+
+export__get_c_export_decls_2(_Preds, [], []).
+export__get_c_export_decls_2(Preds, [E|ExportedProcs], C_ExportDecls) :-
+	E = pragma_exported_proc(PredId, ProcId, C_Function),
+	get_export_info(Preds, PredId, ProcId, _Exported, C_RetType,
+		_DeclareReturnVal, _FailureAction, _SuccessAction,
+		HeadArgInfoTypes),
+	get_argument_declarations(HeadArgInfoTypes, no, ArgDecls),
+	C_ExportDecl = c_export_decl(C_RetType, C_Function, ArgDecls),
+	export__get_c_export_decls_2(Preds, ExportedProcs, C_ExportDecls0),
+	C_ExportDecls = [C_ExportDecl | C_ExportDecls0].
+
+%-----------------------------------------------------------------------------%
+
+export__get_c_export_defns(Module, ExportedProcsCode) :-
 	module_info_get_pragma_exported_procs(Module, ExportedProcs),
 	module_info_get_predicate_table(Module, PredicateTable),
 	predicate_table_get_preds(PredicateTable, Preds),
@@ -60,6 +104,8 @@ export__get_pragma_exported_procs(Module, ExportedProcsCode) :-
 	% For each exported procedure, produce a C function.
 	% The code we generate is in the form
 	%
+	% Declare_entry(<label of called proc>); /* or Declare_static */
+	%
 	% #if SEMIDET
 	%   bool
 	% #elif FUNCTION
@@ -87,15 +133,14 @@ export__get_pragma_exported_procs(Module, ExportedProcsCode) :-
 	%	restore_registers();
 	%	<copy input arguments from Mercury__Arguments into registers>
 	%		/* save the registers which may be clobbered      */
-	%		/* by the C function call call_engine().          */
+	%		/* by the C function call MR_call_engine().       */
 	%	save_transient_registers();
-	%	{
-	%	Declare_entry(<label of called proc>);
-	%	call_engine(ENTRY(<label of called proc>);
-	%	}
+	%
+	%	(void) MR_call_engine(ENTRY(<label of called proc>), FALSE);
+	%
 	%		/* restore the registers which may have been      */
 	%		/* clobbered by the return from the C function    */
-	%		/* call_engine()				  */
+	%		/* MR_call_engine()				  */
 	%	restore_transient_registers();
 	% #if SEMIDET
 	%	if (!r1) {
@@ -120,7 +165,7 @@ export__get_pragma_exported_procs(Module, ExportedProcsCode) :-
 export__to_c(_Preds, [], _Module, []).
 export__to_c(Preds, [E|ExportedProcs], Module, ExportedProcsCode) :-
 	E = pragma_exported_proc(PredId, ProcId, C_Function),
-	get_export_info(Preds, PredId, ProcId,
+	get_export_info(Preds, PredId, ProcId, Exported,
 		C_RetType, MaybeDeclareRetval, MaybeFail, MaybeSucceed,
 		ArgInfoTypes),
 	get_argument_declarations(ArgInfoTypes, yes, ArgDecls),
@@ -131,9 +176,17 @@ export__to_c(Preds, [E|ExportedProcs], Module, ExportedProcsCode) :-
 	copy_output_args(ArgInfoTypes, 0, OutputArgs),
 	
 	code_util__make_proc_label(Module, PredId, ProcId, ProcLabel),
-	get_proc_label(ProcLabel, ProcLabelString),
+	llds_out__get_proc_label(ProcLabel, yes, ProcLabelString),
+
+	( Exported = yes ->
+		DeclareString = "Declare_entry"
+	;
+		DeclareString = "Declare_static"
+	),
 
 	string__append_list([	"\n",
+				DeclareString, "(", ProcLabelString, ");\n",
+				"\n",
 				C_RetType, "\n", 
 				C_Function, "(", ArgDecls, ")\n{\n",
 				"#if NUM_REAL_REGS > 0\n",
@@ -145,12 +198,8 @@ export__to_c(Preds, [E|ExportedProcs], Module, ExportedProcsCode) :-
 				"\trestore_registers();\n", 
 				InputArgs,
 				"\tsave_transient_registers();\n",
-				"\t{\n\tDeclare_entry(",
-				ProcLabelString,
-				");\n",
-				"\tcall_engine(ENTRY(",
-				ProcLabelString,
-				"));\n\t}\n",
+				"\t(void) MR_call_engine(ENTRY(",
+					ProcLabelString, "), FALSE);\n",
 				"\trestore_transient_registers();\n",
 				MaybeFail,
 				OutputArgs,
@@ -169,19 +218,24 @@ export__to_c(Preds, [E|ExportedProcs], Module, ExportedProcsCode) :-
 	%	Figure out the C return type, the actions on success
 	%	and failure, and the argument locations/modes/types
 	%	for a given procedure.
-:- pred get_export_info(pred_table, pred_id, proc_id,
+:- pred get_export_info(pred_table, pred_id, proc_id, bool,
 			string, string, string, string,
 			assoc_list(arg_info, type)).
-:- mode get_export_info(in, in, in, out, out, out, out, out) is det.
+:- mode get_export_info(in, in, in, out, out, out, out, out, out) is det.
 
-get_export_info(Preds, PredId, ProcId, C_RetType,
+get_export_info(Preds, PredId, ProcId, Exported, C_RetType,
 		MaybeDeclareRetval, MaybeFail, MaybeSucceed, ArgInfoTypes) :-
 	map__lookup(Preds, PredId, PredInfo),
+	( procedure_is_exported(PredInfo, ProcId) ->
+		Exported = yes
+	;
+		Exported = no
+	),
 	pred_info_get_is_pred_or_func(PredInfo, PredOrFunc),
 	pred_info_procedures(PredInfo, ProcTable),
 	map__lookup(ProcTable, ProcId, ProcInfo),
 	proc_info_arg_info(ProcInfo, ArgInfos),
-	pred_info_arg_types(PredInfo, _TVarSet, ArgTypes),
+	pred_info_arg_types(PredInfo, ArgTypes),
 	proc_info_interface_code_model(ProcInfo, CodeModel),
 	assoc_list__from_corresponding_lists(ArgInfos, ArgTypes,
 		ArgInfoTypes0),
@@ -194,9 +248,9 @@ get_export_info(Preds, PredId, ProcId, C_RetType,
 			pred_args_to_func_args(ArgInfoTypes0, ArgInfoTypes1,
 				arg_info(RetArgLoc, RetArgMode) - RetType),
 			RetArgMode = top_out,
-			\+ export__exclude_argument_type(RetType)
+			\+ type_util__is_dummy_argument_type(RetType)
 		->
-			export__term_to_type_string(RetType, C_RetType),
+			export__type_to_type_string(RetType, C_RetType),
 			argloc_to_string(RetArgLoc, RetArgString0),
 			convert_type_from_mercury(RetArgString0, RetType,
 				RetArgString),
@@ -248,7 +302,7 @@ get_export_info(Preds, PredId, ProcId, C_RetType,
 :- pred export__include_arg(pair(arg_info, type)::in) is semidet.
 export__include_arg(arg_info(_Loc, Mode) - Type) :-
 	Mode \= top_unused,
-	\+ export__exclude_argument_type(Type).
+	\+ type_util__is_dummy_argument_type(Type).
 
 	% get_argument_declarations(Args, NameThem, DeclString):
 	% build a string to declare the argument types (and if
@@ -292,7 +346,7 @@ get_argument_declaration(ArgInfo, Type, Num, NameThem, TypeString, ArgName) :-
 	;
 		ArgName = ""
 	),
-	export__term_to_type_string(Type, TypeString0),
+	export__type_to_type_string(Type, TypeString0),
 	(
 		Mode = top_out
 	->
@@ -410,89 +464,63 @@ convert_type_from_mercury(Rval, Type, ConvertedRval) :-
 		ConvertedRval = Rval
 	).
 
-% Certain types, namely io__state and store__store(S),
-% are just dummy types used to ensure logical semantics;
-% there is no need to actually pass them, and so when
-% exporting procedures to C, we don't include arguments with
-% these types.
+%-----------------------------------------------------------------------------%
 
-:- pred export__exclude_argument_type((type)::in) is semidet.
-export__exclude_argument_type(Type) :-
-	Type = term__functor(term__atom(":"), [
-			term__functor(term__atom(ModuleName), [], _),
-			term__functor(term__atom(TypeName), TypeArgs, _)
-		], _),
-	list__length(TypeArgs, TypeArity),
-	export__exclude_argument_type_2(ModuleName, TypeName, TypeArity).
+% Should this predicate go in llds_out.m?
 
-:- pred export__exclude_argument_type_2(string::in, string::in, arity::in)
-	is semidet.
-export__exclude_argument_type_2("io", "state", 0).	% io:state/0
-export__exclude_argument_type_2("store", "store", 1).	% store:store/1.
-
-export__produce_header_file(Module, ModuleName) -->
-	{ module_info_get_pragma_exported_procs(Module, ExportedProcs) },
+export__produce_header_file([], _) --> [].
+export__produce_header_file(C_ExportDecls, ModuleName) -->
+	{ C_ExportDecls = [_|_] },
+	module_name_to_file_name(ModuleName, ".h", yes, FileName),
+	io__tell(FileName, Result),
 	(
-		{ ExportedProcs = [_|_] }
+		{ Result = ok }
 	->
-		{ module_info_get_predicate_table(Module, PredicateTable) },
-		{ predicate_table_get_preds(PredicateTable, Preds) },
-		{ string__append(ModuleName, ".h", FileName) },
-		io__tell(FileName, Result),
-		(
-			{ Result = ok }
-		->
-			{ library__version(Version) },
-			io__write_strings(
-				["/*\n** Automatically generated from `", 
-				ModuleName,
-				".m' by the\n** Mercury compiler, version ", 
-				Version,
-				".  Do not edit.\n*/\n"]),
-			{ string__to_upper(ModuleName, UpperModuleName) },
-			{ string__append(UpperModuleName, "_H", UpperFileName) },
-			io__write_strings([
-					"#ifndef ", UpperFileName, "\n",
-					"#define ", UpperFileName, "\n",
-					"\n",
-					"#ifdef __cplusplus\n",
-					"extern ""C"" {\n",
-					"#endif\n",
-					"\n",
-					"#include ""imp.h""\n",
-					"\n"]),
-			export__produce_header_file_2(Preds, ExportedProcs),
-			io__write_strings([
-					"\n",
-					"#ifdef __cplusplus\n",
-					"}\n",
-					"#endif\n",
-					"\n",
-					"#endif /* ", UpperFileName, " */\n"]),
-			io__told
-		;
-			io__progname_base("export.m", ProgName),
-			io__write_string("\n"),
-			io__write_string(ProgName),
-			io__write_string(": can't open `"),
-			io__write_string(FileName),
-			io__write_string("' for output\n"),
-			io__set_exit_status(1)
-		)
+		module_name_to_file_name(ModuleName, ".m", no, SourceFileName),
+		{ library__version(Version) },
+		io__write_strings(["/*\n** Automatically generated from `", 
+			SourceFileName,
+			".m' by the\n** Mercury compiler, version ", Version,
+			".  Do not edit.\n*/\n"]),
+		{ llds_out__sym_name_mangle(ModuleName, MangledModuleName) },
+		{ string__to_upper(MangledModuleName, UppercaseModuleName) },
+		{ string__append(UppercaseModuleName, "_H", GuardMacroName) },
+		io__write_strings([
+			"#ifndef ", GuardMacroName, "\n",
+			"#define ", GuardMacroName, "\n",
+			"\n",
+			"#ifdef __cplusplus\n",
+			"extern ""C"" {\n",
+			"#endif\n",
+			"\n",
+			"#ifndef MERCURY_HDR_EXCLUDE_IMP_H\n",
+			"#include ""mercury_imp.h""\n",
+			"#endif\n",
+			"\n"]),
+		export__produce_header_file_2(C_ExportDecls),
+		io__write_strings([
+			"\n",
+			"#ifdef __cplusplus\n",
+			"}\n",
+			"#endif\n",
+			"\n",
+			"#endif /* ", GuardMacroName, " */\n"]),
+		io__told
 	;
-		[]
+		io__progname_base("export.m", ProgName),
+		io__write_string("\n"),
+		io__write_string(ProgName),
+		io__write_string(": can't open `"),
+		io__write_string(FileName),
+		io__write_string("' for output\n"),
+		io__set_exit_status(1)
 	).
 
-:- pred export__produce_header_file_2(pred_table, list(pragma_exported_proc),
-	io__state, io__state).
-:- mode export__produce_header_file_2(in, in, di, uo) is det.
-export__produce_header_file_2(_Preds, []) --> [].
-export__produce_header_file_2(Preds, [E|ExportedProcs]) -->
-	{ E = pragma_exported_proc(PredId, ProcId, C_Function) },
-	{ get_export_info(Preds, PredId, ProcId, C_RetType,
-		_DeclareReturnVal, _FailureAction, _SuccessAction,
-		HeadArgInfoTypes) },
-	{ get_argument_declarations(HeadArgInfoTypes, no, ArgDecls) },
+:- pred export__produce_header_file_2(c_export_decls, io__state, io__state).
+:- mode export__produce_header_file_2(in, di, uo) is det.
+export__produce_header_file_2([]) --> [].
+export__produce_header_file_2([E|ExportedProcs]) -->
+	{ E = c_export_decl(C_RetType, C_Function, ArgDecls) },
 
 		% output the function header
 	io__write_string(C_RetType),
@@ -502,12 +530,12 @@ export__produce_header_file_2(Preds, [E|ExportedProcs]) -->
 	io__write_string(ArgDecls),
 	io__write_string(");\n"),
 
-	export__produce_header_file_2(Preds, ExportedProcs).
+	export__produce_header_file_2(ExportedProcs).
 
 	% Convert a term representation of a variable type to a string which
 	% represents the C type of the variable
 	% Apart from special cases, local variables become Words
-export__term_to_type_string(Type, Result) :-
+export__type_to_type_string(Type, Result) :-
 	( Type = term__functor(term__atom("int"), [], _) ->
 		Result = "Integer"
 	; Type = term__functor(term__atom("float"), [], _) ->
@@ -520,3 +548,4 @@ export__term_to_type_string(Type, Result) :-
 		Result = "Word"
 	).
 
+%-----------------------------------------------------------------------------%
