@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %------------------------------------------------------------------------------%
@@ -8,8 +8,8 @@
 
 :- interface.
 
-:- import_module llds, options.
-:- import_module list, std_util, bool, assoc_list, term.
+:- import_module llds, options, prog_data.
+:- import_module list, std_util, bool, assoc_list.
 
 :- type exprn_opts
 	--->	nlg_asm_sgt_ubf(
@@ -51,16 +51,17 @@
 :- pred exprn_aux__substitute_rval_in_rval(rval, rval, rval, rval).
 :- mode exprn_aux__substitute_rval_in_rval(in, in, in, out) is det.
 
-:- pred exprn_aux__substitute_vars_in_rval(assoc_list(var, rval), rval, rval).
+:- pred exprn_aux__substitute_vars_in_rval(assoc_list(prog_var, rval),
+		rval, rval).
 :- mode exprn_aux__substitute_vars_in_rval(in, in, out) is det.
 
 :- pred exprn_aux__substitute_rvals_in_rval(assoc_list(rval, rval), rval, rval).
 :- mode exprn_aux__substitute_rvals_in_rval(in, in, out) is det.
 
-:- pred exprn_aux__vars_in_lval(lval, list(var)).
+:- pred exprn_aux__vars_in_lval(lval, list(prog_var)).
 :- mode exprn_aux__vars_in_lval(in, out) is det.
 
-:- pred exprn_aux__vars_in_rval(rval, list(var)).
+:- pred exprn_aux__vars_in_rval(rval, list(prog_var)).
 :- mode exprn_aux__vars_in_rval(in, out) is det.
 
 :- pred exprn_aux__simplify_rval(rval, rval).
@@ -128,9 +129,12 @@ exprn_aux__const_is_constant(float_const(_), ExprnOpts, IsConst) :-
 		IsConst = StaticGroundTerms
 	).
 exprn_aux__const_is_constant(string_const(_), _, yes).
+exprn_aux__const_is_constant(multi_string_const(_, _), _, yes).
 exprn_aux__const_is_constant(code_addr_const(CodeAddr), ExprnOpts, IsConst) :-
 	exprn_aux__addr_is_constant(CodeAddr, ExprnOpts, IsConst).
 exprn_aux__const_is_constant(data_addr_const(_), _, yes).
+exprn_aux__const_is_constant(label_entry(Label), ExprnOpts, IsConst) :-
+	exprn_aux__addr_is_constant(label(Label), ExprnOpts, IsConst).
 
 :- pred exprn_aux__addr_is_constant(code_addr, exprn_opts, bool).
 :- mode exprn_aux__addr_is_constant(in, in, out) is det.
@@ -145,9 +149,18 @@ exprn_aux__addr_is_constant(succip, _, no).
 exprn_aux__addr_is_constant(do_succeed(_), _, no).
 exprn_aux__addr_is_constant(do_redo, _, no).
 exprn_aux__addr_is_constant(do_fail, _, no).
-exprn_aux__addr_is_constant(do_det_closure, _, no).
-exprn_aux__addr_is_constant(do_semidet_closure, _, no).
-exprn_aux__addr_is_constant(do_nondet_closure, _, no).
+exprn_aux__addr_is_constant(do_trace_redo_fail_shallow, _, no).
+exprn_aux__addr_is_constant(do_trace_redo_fail_deep, _, no).
+exprn_aux__addr_is_constant(do_call_closure, _, no).
+exprn_aux__addr_is_constant(do_call_class_method, _, no).
+exprn_aux__addr_is_constant(do_det_aditi_call, _, no).
+exprn_aux__addr_is_constant(do_semidet_aditi_call, _, no).
+exprn_aux__addr_is_constant(do_nondet_aditi_call, _, no).
+exprn_aux__addr_is_constant(do_aditi_insert, _, no).
+exprn_aux__addr_is_constant(do_aditi_delete, _, no).
+exprn_aux__addr_is_constant(do_aditi_bulk_insert, _, no).
+exprn_aux__addr_is_constant(do_aditi_bulk_delete, _, no).
+exprn_aux__addr_is_constant(do_aditi_modify, _, no).
 exprn_aux__addr_is_constant(do_not_reached, _, no).
 
 :- pred exprn_aux__label_is_constant(label, bool, bool, bool).
@@ -185,7 +198,7 @@ exprn_aux__imported_is_constant(NonLocalGotos, AsmLabels, IsConst) :-
 
 exprn_aux__rval_contains_lval(lval(Lval0), Lval) :-
 	exprn_aux__lval_contains_lval(Lval0, Lval).
-exprn_aux__rval_contains_lval(create(_, Rvals, _, _), Lval) :-
+exprn_aux__rval_contains_lval(create(_, Rvals, _, _, _, _), Lval) :-
 	exprn_aux__args_contain_lval(Rvals, Lval).
 exprn_aux__rval_contains_lval(mkword(_, Rval), Lval) :-
 	exprn_aux__rval_contains_lval(Rval, Lval).
@@ -245,7 +258,7 @@ exprn_aux__rval_contains_rval(Rval0, Rval) :-
 			Rval0 = lval(Lval),
 			exprn_aux__lval_contains_rval(Lval, Rval)
 		;
-			Rval0 = create(_, Rvals, _, _),
+			Rval0 = create(_, Rvals, _, _, _, _),
 			exprn_aux__args_contain_rval(Rvals, Rval)
 		;
 			Rval0 = mkword(_, Rval1),
@@ -287,7 +300,7 @@ exprn_aux__args_contain_rval([M | Ms], Rval) :-
 exprn_aux__vars_in_rval(lval(Lval), Vars) :-
 	exprn_aux__vars_in_lval(Lval, Vars).
 exprn_aux__vars_in_rval(var(Var), [Var]).
-exprn_aux__vars_in_rval(create(_, Rvals, _, _), Vars) :-
+exprn_aux__vars_in_rval(create(_, Rvals, _, _, _, _), Vars) :-
 	exprn_aux__vars_in_args(Rvals, Vars).
 exprn_aux__vars_in_rval(mkword(_, Rval), Vars) :-
 	exprn_aux__vars_in_rval(Rval, Vars).
@@ -314,6 +327,8 @@ exprn_aux__vars_in_lval(succip(Rval), Vars) :-
 	exprn_aux__vars_in_rval(Rval, Vars).
 exprn_aux__vars_in_lval(redoip(Rval), Vars) :-
 	exprn_aux__vars_in_rval(Rval, Vars).
+exprn_aux__vars_in_lval(redofr(Rval), Vars) :-
+	exprn_aux__vars_in_rval(Rval, Vars).
 exprn_aux__vars_in_lval(succfr(Rval), Vars) :-
 	exprn_aux__vars_in_rval(Rval, Vars).
 exprn_aux__vars_in_lval(prevfr(Rval), Vars) :-
@@ -326,7 +341,7 @@ exprn_aux__vars_in_lval(mem_ref(Rval), Vars) :-
 	exprn_aux__vars_in_rval(Rval, Vars).
 exprn_aux__vars_in_lval(lvar(Var), [Var]).
 
-:- pred exprn_aux__vars_in_mem_ref(mem_ref, list(var)).
+:- pred exprn_aux__vars_in_mem_ref(mem_ref, list(prog_var)).
 :- mode exprn_aux__vars_in_mem_ref(in, out) is det.
 
 exprn_aux__vars_in_mem_ref(stackvar_ref(_), []).
@@ -334,7 +349,7 @@ exprn_aux__vars_in_mem_ref(framevar_ref(_), []).
 exprn_aux__vars_in_mem_ref(heap_ref(Rval, _, _), Vars) :-
 	exprn_aux__vars_in_rval(Rval, Vars).
 
-:- pred exprn_aux__vars_in_args(list(maybe(rval)), list(var)).
+:- pred exprn_aux__vars_in_args(list(maybe(rval)), list(prog_var)).
 :- mode exprn_aux__vars_in_args(in, out) is det.
 
 exprn_aux__vars_in_args([], []).
@@ -361,10 +376,10 @@ exprn_aux__substitute_lval_in_rval(OldLval, NewLval, Rval0, Rval) :-
 		Rval0 = var(_Var),
 		Rval = Rval0
 	;
-		Rval0 = create(Tag, Rvals0, Unique, Num),
+		Rval0 = create(Tag, Rvals0, ArgTypes, StatDyn, Num, Msg),
 		exprn_aux__substitute_lval_in_args(OldLval, NewLval,
 			Rvals0, Rvals),
-		Rval = create(Tag, Rvals, Unique, Num)
+		Rval = create(Tag, Rvals, ArgTypes, StatDyn, Num, Msg)
 	;
 		Rval0 = mkword(Tag, Rval1),
 		exprn_aux__substitute_lval_in_rval(OldLval, NewLval,
@@ -463,6 +478,11 @@ exprn_aux__substitute_lval_in_lval_2(OldLval, NewLval, Lval0, Lval) :-
 		exprn_aux__substitute_lval_in_rval(OldLval, NewLval,
 			Rval0, Rval),
 		Lval = redoip(Rval)
+	;
+		Lval0 = redofr(Rval0),
+		exprn_aux__substitute_lval_in_rval(OldLval, NewLval,
+			Rval0, Rval),
+		Lval = redofr(Rval)
 	;
 		Lval0 = succfr(Rval0),
 		exprn_aux__substitute_lval_in_rval(OldLval, NewLval,
@@ -522,10 +542,10 @@ exprn_aux__substitute_rval_in_rval(OldRval, NewRval, Rval0, Rval) :-
 			Rval0 = var(_),
 			Rval = Rval0
 		;
-			Rval0 = create(Tag, Rvals0, Unique, Num),
+			Rval0 = create(Tag, Rvals0, ATs, StatDyn, Num, Msg),
 			exprn_aux__substitute_rval_in_args(OldRval, NewRval,
 				Rvals0, Rvals),
-			Rval = create(Tag, Rvals, Unique, Num)
+			Rval = create(Tag, Rvals, ATs, StatDyn, Num, Msg)
 		;
 			Rval0 = mkword(Tag, Rval1),
 			exprn_aux__substitute_rval_in_rval(OldRval, NewRval,
@@ -612,6 +632,11 @@ exprn_aux__substitute_rval_in_lval(OldRval, NewRval, Lval0, Lval) :-
 		exprn_aux__substitute_rval_in_rval(OldRval, NewRval,
 			Rval0, Rval),
 		Lval = redoip(Rval)
+	;
+		Lval0 = redofr(Rval0),
+		exprn_aux__substitute_rval_in_rval(OldRval, NewRval,
+			Rval0, Rval),
+		Lval = redofr(Rval)
 	;
 		Lval0 = succfr(Rval0),
 		exprn_aux__substitute_rval_in_rval(OldRval, NewRval,
@@ -712,36 +737,42 @@ exprn_aux__simplify_rval(Rval0, Rval) :-
 
 exprn_aux__simplify_rval_2(Rval0, Rval) :-
 	(
-		Rval0 = lval(field(Tag, create(Tag, Args, _, _), Field)),
+		Rval0 = lval(field(MaybeTag, Base, Field)),
+		Base = create(Tag, Args, _, _, _, _),
+		(
+			MaybeTag = yes(Tag)
+		;
+			MaybeTag = no
+		),
 		Field = const(int_const(FieldNum))
 	->
 		list__index0_det(Args, FieldNum, yes(Rval))
 	;
-		Rval0 = lval(field(Tag, Rval1, Num)),
+		Rval0 = lval(field(MaybeTag, Rval1, Num)),
 		exprn_aux__simplify_rval_2(Rval1, Rval2)
 	->
-		Rval = lval(field(Tag, Rval2, Num))
+		Rval = lval(field(MaybeTag, Rval2, Num))
 	;
-		Rval0 = create(Tag, Args0, Unique, CNum),
+		Rval0 = create(Tag, Args0, ArgTypes, StatDyn, CNum, Msg),
 		exprn_aux__simplify_args(Args0, Args),
 		Args \= Args0
 	->
-		Rval = create(Tag, Args, Unique, CNum)
+		Rval = create(Tag, Args, ArgTypes, StatDyn, CNum, Msg)
 	;
-		Rval0 = unop(UOp, Rval1),
+		Rval0 = unop(UnOp, Rval1),
 		exprn_aux__simplify_rval_2(Rval1, Rval2)
 	->
-		Rval = unop(UOp, Rval2)
+		Rval = unop(UnOp, Rval2)
 	;
-		Rval0 = binop(BOp, Rval1, Rval2),
+		Rval0 = binop(BinOp, Rval1, Rval2),
 		exprn_aux__simplify_rval_2(Rval1, Rval3)
 	->
-		Rval = binop(BOp, Rval3, Rval2)
+		Rval = binop(BinOp, Rval3, Rval2)
 	;
-		Rval0 = binop(BOp, Rval1, Rval2),
+		Rval0 = binop(BinOp, Rval1, Rval2),
 		exprn_aux__simplify_rval_2(Rval2, Rval3)
 	->
-		Rval = binop(BOp, Rval1, Rval3)
+		Rval = binop(BinOp, Rval1, Rval3)
 	;
 		fail
 	).
@@ -769,7 +800,7 @@ exprn_aux__simplify_args([MR0 | Ms0], [MR | Ms]) :-
 exprn_aux__rval_addrs(lval(Lval), CodeAddrs, DataAddrs) :-
 	exprn_aux__lval_addrs(Lval, CodeAddrs, DataAddrs).
 exprn_aux__rval_addrs(var(_), [], []).
-exprn_aux__rval_addrs(create(_, MaybeRvals, _, _), CodeAddrs, DataAddrs) :-
+exprn_aux__rval_addrs(create(_, MaybeRvals, _,_,_,_), CodeAddrs, DataAddrs) :-
 	exprn_aux__maybe_rval_list_addrs(MaybeRvals, CodeAddrs, DataAddrs).
 exprn_aux__rval_addrs(mkword(_Tag, Rval), CodeAddrs, DataAddrs) :-
 	exprn_aux__rval_addrs(Rval, CodeAddrs, DataAddrs).
@@ -807,6 +838,8 @@ exprn_aux__lval_addrs(prevfr(Rval), CodeAddrs, DataAddrs) :-
 	exprn_aux__rval_addrs(Rval, CodeAddrs, DataAddrs).
 exprn_aux__lval_addrs(succfr(Rval), CodeAddrs, DataAddrs) :-
 	exprn_aux__rval_addrs(Rval, CodeAddrs, DataAddrs).
+exprn_aux__lval_addrs(redofr(Rval), CodeAddrs, DataAddrs) :-
+	exprn_aux__rval_addrs(Rval, CodeAddrs, DataAddrs).
 exprn_aux__lval_addrs(redoip(Rval), CodeAddrs, DataAddrs) :-
 	exprn_aux__rval_addrs(Rval, CodeAddrs, DataAddrs).
 exprn_aux__lval_addrs(succip(Rval), CodeAddrs, DataAddrs) :-
