@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -13,7 +13,8 @@
 
 :- interface.
 
-:- import_module hlds_module, hlds_pred.
+:- import_module hlds_module, hlds_pred, hlds_goal.
+:- import_module list.
 
 :- pred move_follow_code_in_proc(proc_info, proc_info,
 	module_info, module_info).
@@ -30,10 +31,9 @@
 
 :- implementation.
 
-:- import_module hlds_goal, hlds_data, goal_util, mode_util.
+:- import_module hlds_data, goal_util, mode_util, prog_data.
 :- import_module globals, options, det_analysis, quantification.
-:- import_module bool, list, map, set.
-:- import_module term, std_util, require.
+:- import_module bool, map, set, term, std_util, require.
 
 %-----------------------------------------------------------------------------%
 
@@ -43,7 +43,7 @@ move_follow_code_in_proc(ProcInfo0, ProcInfo, ModuleInfo0, ModuleInfo) :-
 	globals__lookup_bool_option(Globals, prev_code, PrevCode),
 	Flags = FollowCode - PrevCode,
 	proc_info_goal(ProcInfo0, Goal0),
-	proc_info_variables(ProcInfo0, Varset0),
+	proc_info_varset(ProcInfo0, Varset0),
 	proc_info_vartypes(ProcInfo0, VarTypes0),
 	(
 		move_follow_code_in_goal(Goal0, Goal1, Flags, no, Res),
@@ -57,7 +57,7 @@ move_follow_code_in_proc(ProcInfo0, ProcInfo, ModuleInfo0, ModuleInfo) :-
 			Varset0, VarTypes0, Goal2, Varset, VarTypes, _Warnings),
 		proc_info_get_initial_instmap(ProcInfo0,
 			ModuleInfo0, InstMap0),
-		recompute_instmap_delta(no, Goal2, Goal, InstMap0,
+		recompute_instmap_delta(no, Goal2, Goal, VarTypes, InstMap0,
 			ModuleInfo0, ModuleInfo)
 	;
 		Goal = Goal0,
@@ -88,6 +88,13 @@ move_follow_code_in_goal(Goal0 - GoalInfo, Goal - GoalInfo, Flags, R0, R) :-
 move_follow_code_in_goal_2(conj(Goals0), conj(Goals), Flags, R0, R) :-
 	move_follow_code_in_conj(Goals0, Goals, Flags, R0, R).
 
+move_follow_code_in_goal_2(par_conj(Goals0, SM), par_conj(Goals, SM),
+		Flags, R0, R) :-
+		% move_follow_code_in_disj treats its list of goals as
+		% independent goals, so we can use it to process the
+		% independent parallel conjuncts.
+	move_follow_code_in_disj(Goals0, Goals, Flags, R0, R).
+
 move_follow_code_in_goal_2(disj(Goals0, SM), disj(Goals, SM), Flags, R0, R) :-
 	move_follow_code_in_disj(Goals0, Goals, Flags, R0, R).
 
@@ -104,21 +111,29 @@ move_follow_code_in_goal_2(if_then_else(Vars, Cond0, Then0, Else0, SM),
 	move_follow_code_in_goal(Then0, Then, Flags, R1, R2),
 	move_follow_code_in_goal(Else0, Else, Flags, R2, R).
 
-move_follow_code_in_goal_2(some(Vars, Goal0), some(Vars, Goal), Flags, R0, R) :-
+move_follow_code_in_goal_2(some(Vars, CanRemove, Goal0),
+		some(Vars, CanRemove, Goal), Flags, R0, R) :-
 	move_follow_code_in_goal(Goal0, Goal, Flags, R0, R).
 
-move_follow_code_in_goal_2(higher_order_call(A,B,C,D,E,F),
-			higher_order_call(A,B,C,D,E,F), _, R, R).
+move_follow_code_in_goal_2(generic_call(A,B,C,D),
+			generic_call(A,B,C,D), _, R, R).
 
 move_follow_code_in_goal_2(call(A,B,C,D,E,F), call(A,B,C,D,E,F), _, R, R).
 
 move_follow_code_in_goal_2(unify(A,B,C,D,E), unify(A,B,C,D,E), _, R, R).
 
-move_follow_code_in_goal_2(pragma_c_code(A,B,C,D,E,F,G,H), 
-			pragma_c_code(A,B,C,D,E,F,G,H), _, R, R).
+move_follow_code_in_goal_2(pragma_c_code(A,B,C,D,E,F,G), 
+			pragma_c_code(A,B,C,D,E,F,G), _, R, R).
+
+move_follow_code_in_goal_2(bi_implication(_, _), _, _, _, _) :-
+	% these should have been expanded out by now
+	error("move_follow_code_in_goal_2: unexpected bi_implication").
 
 %-----------------------------------------------------------------------------%
 
+	% move_follow_code_in_disj is used both for disjunction and
+	% parallel conjunction.
+
 :- pred move_follow_code_in_disj(list(hlds_goal), list(hlds_goal),
 	pair(bool), bool, bool).
 :- mode move_follow_code_in_disj(in, out, in, in, out) is det.
@@ -299,14 +314,14 @@ check_follow_code_detism([_ - GoalInfo | Goals], Detism0) :-
 :- mode move_follow_code_is_builtin(in) is semidet.
 
 move_follow_code_is_builtin(unify(_, _, _, Unification, _) - _GoalInfo) :-
-	Unification \= complicated_unify(_, _).
+	Unification \= complicated_unify(_, _, _).
 move_follow_code_is_builtin(call(_, _, _, Builtin, _, _) - _GoalInfo) :-
 	Builtin = inline_builtin.
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
-:- pred move_prev_code(hlds_goal, hlds_goal, set(var),
+:- pred move_prev_code(hlds_goal, hlds_goal, set(prog_var),
 	list(hlds_goal), list(hlds_goal), bool, bool).
 % :- mode move_prev_code(di, uo, in, di, uo, in, out) is det.
 :- mode move_prev_code(in, out, in, in, out, in, out) is det.
@@ -344,7 +359,7 @@ move_prev_code(Goal, Goal, _, RevPrevGoals, RevPrevGoals, R, R).
 % 	hlds_goal).
 % :- mode move_prev_code_breakup_branched(in, out, out, out) is semidet.
 
-:- pred move_prev_code_forbidden_vars(list(hlds_goal), set(var)).
+:- pred move_prev_code_forbidden_vars(list(hlds_goal), set(prog_var)).
 :- mode move_prev_code_forbidden_vars(in, out) is det.
 
 move_prev_code_forbidden_vars([], Empty) :-
@@ -356,11 +371,11 @@ move_prev_code_forbidden_vars([_Goal - GoalInfo | Goals], Varset) :-
 
 % 
 % :- pred move_prev_code_new_forbidden_vars(hlds_goal, hlds_goal,
-% 	set(var), set(var)).
+% 	set(prog_var), set(prog_var)).
 % :- mode move_prev_code_new_forbidden_vars(in, in, in, out) is det.
 % 
 % :- pred move_prev_code_vars_difference(hlds_goal, hlds_goal, hlds_goal,
-% 	set(var), set(var)).
+% 	set(prog_var), set(prog_var)).
 % :- mode move_prev_code_vars_difference(in, in, in, in, out) is det.
 % 
 % move_prev_code_vars_difference(Cond, First, Rest, ForbiddenVars, LocalVars) :-
@@ -374,7 +389,7 @@ move_prev_code_forbidden_vars([_Goal - GoalInfo | Goals], Varset) :-
 % 	set__difference(ThisBranchVars, RestVars, LocalVars0),
 % 	set__difference(LocalVars0, ForbiddenVars, LocalVars).
 % 
-% :- pred move_prev_code_can_pull_producer(set(var), list(hlds_goal),
+% :- pred move_prev_code_can_pull_producer(set(prog_var), list(hlds_goal),
 % 	list(hlds_goal), list(hlds_goal)).
 % :- mode move_prev_code_can_pull_producer(in, in, di, uo) is det.
 % 
