@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -19,9 +19,9 @@
 % to reflect where variables should be to make the setting up of the arguments
 % of the call as efficient as possible.
 
-% See notes/ALLOCATION for a description of the framework that this pass
-% operates within, and for a description of which goals have their follow_vars
-% field filled in.
+% See compiler/notes/allocation.html for a description of the framework that
+% this pass operates within, and for a description of which goals have their
+% follow_vars field filled in.
 
 %-----------------------------------------------------------------------------%
 
@@ -29,12 +29,13 @@
 
 :- interface.
 
-:- import_module hlds_module, hlds_pred.
+:- import_module hlds_module, hlds_pred, hlds_goal, prog_data.
+:- import_module map.
 
 :- pred find_final_follow_vars(proc_info, follow_vars).
 :- mode find_final_follow_vars(in, out) is det.
 
-:- pred find_follow_vars_in_goal(hlds_goal, args_method, module_info,
+:- pred find_follow_vars_in_goal(hlds_goal, map(prog_var, type), module_info,
 				follow_vars, hlds_goal, follow_vars).
 :- mode find_follow_vars_in_goal(in, in, in, in, out, out) is det.
 
@@ -43,9 +44,9 @@
 
 :- implementation.
 
-:- import_module hlds_goal, hlds_data, llds, mode_util.
+:- import_module hlds_data, llds, mode_util, prog_data.
 :- import_module code_util, quantification, arg_info, globals.
-:- import_module bool, list, map, set, std_util, term, require.
+:- import_module bool, list, map, set, std_util, require.
 
 %-----------------------------------------------------------------------------%
 
@@ -59,7 +60,7 @@ find_final_follow_vars(ProcInfo, Follow) :-
 		error("find_final_follow_vars: failed")
 	).
 
-:- pred find_final_follow_vars_2(list(arg_info), list(var),
+:- pred find_final_follow_vars_2(list(arg_info), list(prog_var),
 						follow_vars, follow_vars).
 :- mode find_final_follow_vars_2(in, in, in, out) is semidet.
 
@@ -79,44 +80,52 @@ find_final_follow_vars_2([arg_info(Loc, Mode) | Args], [Var | Vars],
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
-find_follow_vars_in_goal(Goal0 - GoalInfo, ArgsMethod, ModuleInfo, FollowVars0,
+find_follow_vars_in_goal(Goal0 - GoalInfo, VarTypes, ModuleInfo, FollowVars0,
 					Goal - GoalInfo, FollowVars) :-
-	find_follow_vars_in_goal_2(Goal0, ArgsMethod, ModuleInfo, FollowVars0,
-							Goal, FollowVars).
+	find_follow_vars_in_goal_2(Goal0, VarTypes, ModuleInfo, FollowVars0,
+					Goal, FollowVars).
 
 %-----------------------------------------------------------------------------%
 
-:- pred find_follow_vars_in_goal_2(hlds_goal_expr, args_method, module_info,
-		follow_vars, hlds_goal_expr, follow_vars).
+:- pred find_follow_vars_in_goal_2(hlds_goal_expr, map(prog_var, type),
+		module_info, follow_vars, hlds_goal_expr, follow_vars).
 :- mode find_follow_vars_in_goal_2(in, in, in, in, out, out) is det.
 
-find_follow_vars_in_goal_2(conj(Goals0), ArgsMethod, ModuleInfo, FollowVars0,
+find_follow_vars_in_goal_2(conj(Goals0), VarTypes, ModuleInfo, FollowVars0,
 		conj(Goals), FollowVars) :-
-	find_follow_vars_in_conj(Goals0, ArgsMethod, ModuleInfo, FollowVars0,
+	find_follow_vars_in_conj(Goals0, VarTypes, ModuleInfo, FollowVars0,
 		no, Goals, FollowVars).
 
+find_follow_vars_in_goal_2(par_conj(Goals0, SM), VarTypes, ModuleInfo,
+		FollowVars0, par_conj(Goals, SM), FollowVars) :-
+		% find_follow_vars_in_disj treats its list of goals as a
+		% series of independent goals, so we can use it to process
+		% independent parallel conjunction.
+	find_follow_vars_in_disj(Goals0, VarTypes, ModuleInfo, FollowVars0,
+		Goals, FollowVars).
+
 	% We record that at the end of each disjunct, live variables should
 	% be in the locations given by the initial follow_vars, which reflects
 	% the requirements of the code following the disjunction.
 
-find_follow_vars_in_goal_2(disj(Goals0, _), ArgsMethod, ModuleInfo, FollowVars0,
+find_follow_vars_in_goal_2(disj(Goals0, _), VarTypes, ModuleInfo, FollowVars0,
 		disj(Goals, FollowVars0), FollowVars) :-
-	find_follow_vars_in_disj(Goals0, ArgsMethod, ModuleInfo, FollowVars0,
+	find_follow_vars_in_disj(Goals0, VarTypes, ModuleInfo, FollowVars0,
 		Goals, FollowVars).
 
-find_follow_vars_in_goal_2(not(Goal0), ArgsMethod, ModuleInfo, FollowVars0,
+find_follow_vars_in_goal_2(not(Goal0), VarTypes, ModuleInfo, FollowVars0,
 		not(Goal), FollowVars) :-
-	find_follow_vars_in_goal(Goal0, ArgsMethod, ModuleInfo, FollowVars0,
+	find_follow_vars_in_goal(Goal0, VarTypes, ModuleInfo, FollowVars0,
 		Goal, FollowVars).
 
 	% We record that at the end of each arm of the switch, live variables
 	% should be in the locations given by the initial follow_vars, which
 	% reflects the requirements of the code following the switch.
 
-find_follow_vars_in_goal_2(switch(Var, Det, Cases0, _), ArgsMethod, ModuleInfo,
-		FollowVars0,
-		switch(Var, Det, Cases, FollowVars0), FollowVars) :-
-	find_follow_vars_in_cases(Cases0, ArgsMethod, ModuleInfo, FollowVars0,
+find_follow_vars_in_goal_2(switch(Var, Det, Cases0, _), VarTypes, ModuleInfo,
+		FollowVars0, switch(Var, Det, Cases, FollowVars0),
+		FollowVars) :-
+	find_follow_vars_in_cases(Cases0, VarTypes, ModuleInfo, FollowVars0,
 		Cases, FollowVars).
 
 	% Set the follow_vars field for the condition, the then-part and the
@@ -137,37 +146,38 @@ find_follow_vars_in_goal_2(switch(Var, Det, Cases0, _), ArgsMethod, ModuleInfo,
 	% following the if-then-else.
 
 find_follow_vars_in_goal_2(if_then_else(Vars, Cond0, Then0, Else0, _),
-		ArgsMethod, ModuleInfo, FollowVars0,
+		VarTypes, ModuleInfo, FollowVars0,
 		if_then_else(Vars, Cond, Then, Else, FollowVars0),
 		FollowVarsCond) :-
-	find_follow_vars_in_goal(Then0, ArgsMethod, ModuleInfo, FollowVars0,
+	find_follow_vars_in_goal(Then0, VarTypes, ModuleInfo, FollowVars0,
 		Then1, FollowVarsThen),
 	goal_set_follow_vars(Then1, yes(FollowVarsThen), Then),
-	find_follow_vars_in_goal(Cond0, ArgsMethod, ModuleInfo, FollowVarsThen,
+	find_follow_vars_in_goal(Cond0, VarTypes, ModuleInfo, FollowVarsThen,
 		Cond1, FollowVarsCond),
 	goal_set_follow_vars(Cond1, yes(FollowVarsCond), Cond),
-	find_follow_vars_in_goal(Else0, ArgsMethod, ModuleInfo, FollowVars0,
+	find_follow_vars_in_goal(Else0, VarTypes, ModuleInfo, FollowVars0,
 		Else1, FollowVarsElse),
 	goal_set_follow_vars(Else1, yes(FollowVarsElse), Else).
 
-find_follow_vars_in_goal_2(some(Vars, Goal0), ArgsMethod, ModuleInfo,
-		FollowVars0, some(Vars, Goal), FollowVars) :-
-	find_follow_vars_in_goal(Goal0, ArgsMethod, ModuleInfo, FollowVars0,
+find_follow_vars_in_goal_2(some(Vars, CanRemove, Goal0), VarTypes, ModuleInfo,
+		FollowVars0, some(Vars, CanRemove, Goal), FollowVars) :-
+	find_follow_vars_in_goal(Goal0, VarTypes, ModuleInfo, FollowVars0,
 		Goal, FollowVars).
 
+	% XXX These follow-vars aren't correct since the desired positions for
+	% XXX the arguments are different from an ordinary call --- they are
+	% XXX as required by the builtin operation.
 find_follow_vars_in_goal_2(
-		higher_order_call(PredVar, Args, Types, Modes, Det,
-			IsPredOrFunc),
-		ArgsMethod, ModuleInfo, _FollowVars0,
-		higher_order_call(PredVar, Args, Types, Modes, Det,
-			IsPredOrFunc),
+		generic_call(GenericCall, Args, Modes, Det),
+		VarTypes, ModuleInfo, _FollowVars0,
+		generic_call(GenericCall, Args, Modes, Det),
 		FollowVars) :-
 	determinism_to_code_model(Det, CodeModel),
-	make_arg_infos(ArgsMethod, Types, Modes, CodeModel, ModuleInfo,
-		ArgInfo),
+	map__apply_to_list(Args, VarTypes, Types),
+	make_arg_infos(Types, Modes, CodeModel, ModuleInfo, ArgInfo),
 	find_follow_vars_from_arginfo(ArgInfo, Args, FollowVars).
 
-find_follow_vars_in_goal_2(call(A,B,C,D,E,F), _ArgsMethod, ModuleInfo,
+find_follow_vars_in_goal_2(call(A,B,C,D,E,F), _, ModuleInfo,
 		FollowVars0, call(A,B,C,D,E,F), FollowVars) :-
 	(
 		D = inline_builtin
@@ -177,17 +187,9 @@ find_follow_vars_in_goal_2(call(A,B,C,D,E,F), _ArgsMethod, ModuleInfo,
 		find_follow_vars_in_call(A, B, C, ModuleInfo, FollowVars)
 	).
 
-find_follow_vars_in_goal_2(unify(A,B,C,D,E), ArgsMethod, _ModuleInfo,
+find_follow_vars_in_goal_2(unify(A,B,C,D,E), _, _ModuleInfo,
 		FollowVars0, unify(A,B,C,D,E), FollowVars) :-
 	(
-		B = var(BVar),
-		D = complicated_unify(_Mode, CanFail)
-	->
-		determinism_components(Det, CanFail, at_most_one),
-		determinism_to_code_model(Det, CodeModel),
-		arg_info__unify_arg_info(ArgsMethod, CodeModel, ArgInfo),
-		find_follow_vars_from_arginfo(ArgInfo, [A, BVar], FollowVars)
-	;
 		D = assign(LVar, RVar),
 		map__search(FollowVars0, LVar, DesiredLoc)
 	->
@@ -196,13 +198,17 @@ find_follow_vars_in_goal_2(unify(A,B,C,D,E), ArgsMethod, _ModuleInfo,
 		FollowVars = FollowVars0
 	).
 
-find_follow_vars_in_goal_2(pragma_c_code(A,B,C,D,E,F,G,H), _ArgInfo,
-		_ModuleInfo, FollowVars,
-		pragma_c_code(A,B,C,D,E,F,G,H), FollowVars).
+find_follow_vars_in_goal_2(pragma_c_code(A,B,C,D,E,F,G), 
+		_, _ModuleInfo, FollowVars,
+		pragma_c_code(A,B,C,D,E,F,G), FollowVars).
+
+find_follow_vars_in_goal_2(bi_implication(_,_), _, _, _, _, _) :-
+	% these should have been expanded out by now
+	error("find_follow_vars_in_goal_2: unexpected bi_implication").
 
 %-----------------------------------------------------------------------------%
 
-:- pred find_follow_vars_in_call(pred_id, proc_id, list(var), module_info,
+:- pred find_follow_vars_in_call(pred_id, proc_id, list(prog_var), module_info,
 						follow_vars).
 :- mode find_follow_vars_in_call(in, in, in, in, out) is det.
 
@@ -214,7 +220,8 @@ find_follow_vars_in_call(PredId, ProcId, Args, ModuleInfo, Follow) :-
 	proc_info_arg_info(ProcInfo, ArgInfo),
 	find_follow_vars_from_arginfo(ArgInfo, Args, Follow).
 
-:- pred find_follow_vars_from_arginfo(list(arg_info), list(var), follow_vars).
+:- pred find_follow_vars_from_arginfo(list(arg_info), list(prog_var),
+		follow_vars).
 :- mode find_follow_vars_from_arginfo(in, in, out) is det.
 
 find_follow_vars_from_arginfo(ArgInfo, Args, Follow) :-
@@ -227,7 +234,7 @@ find_follow_vars_from_arginfo(ArgInfo, Args, Follow) :-
 		error("find_follow_vars_from_arginfo: failed")
 	).
 
-:- pred find_follow_vars_from_arginfo_2(list(arg_info), list(var),
+:- pred find_follow_vars_from_arginfo_2(list(arg_info), list(prog_var),
 						follow_vars, follow_vars).
 :- mode find_follow_vars_from_arginfo_2(in, in, in, out) is semidet.
 
@@ -263,19 +270,21 @@ find_follow_vars_from_arginfo_2([arg_info(Loc, Mode) | Args], [Var | Vars],
 	% they can only be entered with everything in stack slots; for
 	% model_det and model_semi disjunctions, they will never be
 	% entered at all.)
+	%
+	% This code is used both for disjunction and parallel conjunction.
 
-:- pred find_follow_vars_in_disj(list(hlds_goal), args_method, module_info,
-				follow_vars, list(hlds_goal), follow_vars).
+:- pred find_follow_vars_in_disj(list(hlds_goal), map(prog_var, type),
+		module_info, follow_vars, list(hlds_goal), follow_vars).
 :- mode find_follow_vars_in_disj(in, in, in, in, out, out) is det.
 
-find_follow_vars_in_disj([], _ArgsMethod, _ModuleInfo, FollowVars,
+find_follow_vars_in_disj([], _, _ModuleInfo, FollowVars,
 			[], FollowVars).
-find_follow_vars_in_disj([Goal0 | Goals0], ArgsMethod, ModuleInfo, FollowVars0,
+find_follow_vars_in_disj([Goal0 | Goals0], VarTypes, ModuleInfo, FollowVars0,
 						[Goal | Goals], FollowVars) :-
-	find_follow_vars_in_goal(Goal0, ArgsMethod, ModuleInfo, FollowVars0,
+	find_follow_vars_in_goal(Goal0, VarTypes, ModuleInfo, FollowVars0,
 		Goal1, FollowVars),
 	goal_set_follow_vars(Goal1, yes(FollowVars), Goal),
-	find_follow_vars_in_disj(Goals0, ArgsMethod, ModuleInfo, FollowVars0,
+	find_follow_vars_in_disj(Goals0, VarTypes, ModuleInfo, FollowVars0,
 		Goals, _FollowVars1).
 
 %-----------------------------------------------------------------------------%
@@ -291,18 +300,17 @@ find_follow_vars_in_disj([Goal0 | Goals0], ArgsMethod, ModuleInfo, FollowVars0,
 	% its follow_vars) and to let different branches "vote" on
 	% what should be in registers.
 
-:- pred find_follow_vars_in_cases(list(case), args_method, module_info,
+:- pred find_follow_vars_in_cases(list(case), map(prog_var, type), module_info,
 				follow_vars, list(case), follow_vars).
 :- mode find_follow_vars_in_cases(in, in, in, in, out, out) is det.
 
-find_follow_vars_in_cases([], _ArgsMethod, _ModuleInfo, FollowVars,
-			[], FollowVars).
-find_follow_vars_in_cases([case(Cons, Goal0) | Goals0], ArgsMethod, ModuleInfo,
+find_follow_vars_in_cases([], _, _ModuleInfo, FollowVars, [], FollowVars).
+find_follow_vars_in_cases([case(Cons, Goal0) | Goals0], VarTypes, ModuleInfo,
 			FollowVars0, [case(Cons, Goal) | Goals], FollowVars) :-
-	find_follow_vars_in_goal(Goal0, ArgsMethod, ModuleInfo, FollowVars0,
+	find_follow_vars_in_goal(Goal0, VarTypes, ModuleInfo, FollowVars0,
 		Goal1, FollowVars),
 	goal_set_follow_vars(Goal1, yes(FollowVars), Goal),
-	find_follow_vars_in_cases(Goals0, ArgsMethod, ModuleInfo, FollowVars0,
+	find_follow_vars_in_cases(Goals0, VarTypes, ModuleInfo, FollowVars0,
 		Goals, _FollowVars1).
 
 %-----------------------------------------------------------------------------%
@@ -310,13 +318,13 @@ find_follow_vars_in_cases([case(Cons, Goal0) | Goals0], ArgsMethod, ModuleInfo,
 	% We attach the follow_vars to each goal that follows a goal
 	% that is not cachable by the code generator.
 
-:- pred find_follow_vars_in_conj(list(hlds_goal), args_method, module_info,
-			follow_vars, bool, list(hlds_goal), follow_vars).
+:- pred find_follow_vars_in_conj(list(hlds_goal), map(prog_var, type),
+		module_info, follow_vars, bool, list(hlds_goal), follow_vars).
 :- mode find_follow_vars_in_conj(in, in, in, in, in, out, out) is det.
 
-find_follow_vars_in_conj([], _ArgsMethod, _ModuleInfo, FollowVars,
+find_follow_vars_in_conj([], _, _ModuleInfo, FollowVars,
 		_AttachToFirst, [], FollowVars).
-find_follow_vars_in_conj([Goal0 | Goals0], ArgsMethod, ModuleInfo, FollowVars0,
+find_follow_vars_in_conj([Goal0 | Goals0], VarTypes, ModuleInfo, FollowVars0,
 		AttachToFirst, [Goal | Goals], FollowVars) :-
 	(
 		Goal0 = GoalExpr0 - _,
@@ -325,16 +333,16 @@ find_follow_vars_in_conj([Goal0 | Goals0], ArgsMethod, ModuleInfo, FollowVars0,
 			BuiltinState = inline_builtin
 		;
 			GoalExpr0 = unify(_, _, _, Unification, _),
-			Unification \= complicated_unify(_, _)
+			Unification \= complicated_unify(_, _, _)
 		)
 	->
 		AttachToNext = no
 	;
 		AttachToNext = yes
 	),
-	find_follow_vars_in_conj(Goals0, ArgsMethod, ModuleInfo, FollowVars0,
+	find_follow_vars_in_conj(Goals0, VarTypes, ModuleInfo, FollowVars0,
 		AttachToNext, Goals, FollowVars1),
-	find_follow_vars_in_goal(Goal0, ArgsMethod, ModuleInfo, FollowVars1,
+	find_follow_vars_in_goal(Goal0, VarTypes, ModuleInfo, FollowVars1,
 		Goal1, FollowVars),
 	(
 		AttachToFirst = yes,
