@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -31,15 +31,6 @@
 	;	low
 	;	high.
 
-:- type args_method
-	--->	simple
-	;	compact.
-
-	% Once upon a time, there were more type_info options than this.
-
-:- type type_info_method
-	--->	shared_one_or_two_cell.
-
 :- type prolog_dialect
 	--->	default
 	;	nu
@@ -51,50 +42,39 @@
 	;	num_data_elems
 	;	size_data_elems.
 
+:- type trace_level
+	--->	none
+	;	shallow
+	;	deep.
+
 :- pred convert_gc_method(string::in, gc_method::out) is semidet.
-
 :- pred convert_tags_method(string::in, tags_method::out) is semidet.
-
-:- pred convert_args_method(string::in, args_method::out) is semidet.
-
-:- pred convert_type_info_method(string::in, type_info_method::out) is semidet.
-
 :- pred convert_prolog_dialect(string::in, prolog_dialect::out) is semidet.
-
 :- pred convert_termination_norm(string::in, termination_norm::out) is semidet.
+:- pred convert_trace_level(string::in, bool::in, trace_level::out) is semidet.
+	% the bool should be the setting of the `require_tracing' option.
 
 %-----------------------------------------------------------------------------%
 
 	% Access predicates for the `globals' structure.
 
 :- pred globals__init(option_table::di, gc_method::di, tags_method::di,
-	args_method::di, type_info_method::di, prolog_dialect::di, 
-	termination_norm::di, globals::uo) is det.
+	prolog_dialect::di, termination_norm::di, trace_level::di,
+	globals::uo) is det.
 
 :- pred globals__get_options(globals::in, option_table::out) is det.
 :- pred globals__get_gc_method(globals::in, gc_method::out) is det.
 :- pred globals__get_tags_method(globals::in, tags_method::out) is det.
-:- pred globals__get_args_method(globals::in, args_method::out) is det.
-:- pred globals__get_type_info_method(globals::in, type_info_method::out)
-	is det.
 :- pred globals__get_prolog_dialect(globals::in, prolog_dialect::out) is det.
 :- pred globals__get_termination_norm(globals::in, termination_norm::out) 
 	is det.
+:- pred globals__get_trace_level(globals::in, trace_level::out) is det.
 
 :- pred globals__set_options(globals::in, option_table::in, globals::out)
 	is det.
-:- pred globals__set_gc_method(globals::in, gc_method::in, globals::out)
+
+:- pred globals__set_trace_level(globals::in, trace_level::in, globals::out)
 	is det.
-:- pred globals__set_tags_method(globals::in, tags_method::in, globals::out)
-	is det.
-:- pred globals__set_args_method(globals::in, args_method::in, globals::out)
-	is det.
-:- pred globals__set_type_info_method(globals::in, type_info_method::in,
-	globals::out) is det.
-:- pred globals__set_prolog_dialect(globals::in, prolog_dialect::in,
-	globals::out) is det.
-:- pred globals__set_termination_norm(globals::in, termination_norm::in,
-	globals::out) is det.
 
 :- pred globals__lookup_option(globals::in, option::in, option_data::out)
 	is det.
@@ -113,18 +93,23 @@
 	% More complex options
 
 	% Check if static code addresses are available in the
-	% current grade of compilation
+	% current grade of compilation.
 
 :- pred globals__have_static_code_addresses(globals::in, bool::out) is det.
 
+	% Check if we should include variable information in the layout
+	% structures of call return sites.
+
+:- pred globals__want_return_var_layouts(globals::in, bool::out) is det.
+
 %-----------------------------------------------------------------------------%
 
 	% Access predicates for storing a `globals' structure in the
 	% io__state using io__set_globals and io__get_globals.
 
 :- pred globals__io_init(option_table::di, gc_method::in, tags_method::in,
-	args_method::in, type_info_method::in, prolog_dialect::in,
-	termination_norm::in, io__state::di, io__state::uo) is det.
+	prolog_dialect::in, termination_norm::in, trace_level::in,
+	io__state::di, io__state::uo) is det.
 
 :- pred globals__io_get_gc_method(gc_method::out,
 	io__state::di, io__state::uo) is det.
@@ -132,28 +117,30 @@
 :- pred globals__io_get_tags_method(tags_method::out,
 	io__state::di, io__state::uo) is det.
 
-:- pred globals__io_get_args_method(args_method::out,
-	io__state::di, io__state::uo) is det.
-
-:- pred globals__io_get_type_info_method(type_info_method::out,
-	io__state::di, io__state::uo) is det.
-
 :- pred globals__io_get_prolog_dialect(prolog_dialect::out,
 	io__state::di, io__state::uo) is det.
 
 :- pred globals__io_get_termination_norm(termination_norm::out,
 	io__state::di, io__state::uo) is det.
 
+:- pred globals__io_get_trace_level(trace_level::out,
+	io__state::di, io__state::uo) is det.
+
 :- pred globals__io_get_globals(globals::out, io__state::di, io__state::uo)
 	is det.
 
 :- pred globals__io_set_globals(globals::di, io__state::di, io__state::uo)
 	is det.
 
-:- pred globals__io_lookup_option(option::in, option_data::out,
+:- pred globals__io_set_option(option::in, option_data::in,
 	io__state::di, io__state::uo) is det.
 
-:- pred globals__io_set_option(option::in, option_data::in,
+:- pred globals__io_set_trace_level(trace_level::in,
+	io__state::di, io__state::uo) is det.
+
+:- pred globals__io_set_trace_level_none(io__state::di, io__state::uo) is det.
+
+:- pred globals__io_lookup_option(option::in, option_data::out,
 	io__state::di, io__state::uo) is det.
 
 :- pred globals__io_lookup_bool_option(option, bool, io__state, io__state).
@@ -177,8 +164,6 @@
 :- import_module exprn_aux.
 :- import_module map, std_util, io, require.
 
-%-----------------------------------------------------------------------------%
-
 convert_gc_method("none", none).
 convert_gc_method("conservative", conservative).
 convert_gc_method("accurate", accurate).
@@ -187,12 +172,6 @@ convert_tags_method("none", none).
 convert_tags_method("low", low).
 convert_tags_method("high", high).
 
-convert_args_method("simple", simple).
-convert_args_method("compact", compact).
-
-convert_type_info_method("shared-one-or-two-cell", shared_one_or_two_cell).
-convert_type_info_method("default", shared_one_or_two_cell).
-
 convert_prolog_dialect("default", default).
 convert_prolog_dialect("nu", nu).
 convert_prolog_dialect("NU", nu).
@@ -212,6 +191,13 @@ convert_termination_norm("total", total).
 convert_termination_norm("num-data-elems", num_data_elems).
 convert_termination_norm("size-data-elems", size_data_elems).
 
+convert_trace_level("minimum", no, none).
+convert_trace_level("minimum", yes, shallow).
+convert_trace_level("shallow", _, shallow).
+convert_trace_level("deep", _, deep).
+convert_trace_level("default", no, none).
+convert_trace_level("default", yes, deep).
+
 %-----------------------------------------------------------------------------%
 
 :- type globals
@@ -219,42 +205,30 @@ convert_termination_norm("size-data-elems", size_data_elems).
 			option_table,
 			gc_method,
 			tags_method,
-			args_method,
-			type_info_method,
 			prolog_dialect,
-			termination_norm
+			termination_norm,
+			trace_level
 		).
 
-globals__init(Options, GC_Method, TagsMethod, ArgsMethod,
-		TypeInfoMethod, PrologDialect, TerminationNorm, 
-	globals(Options, GC_Method, TagsMethod, ArgsMethod,
-		TypeInfoMethod, PrologDialect, TerminationNorm)).
+globals__init(Options, GC_Method, TagsMethod,
+		PrologDialect, TerminationNorm, TraceLevel,
+	globals(Options, GC_Method, TagsMethod,
+		PrologDialect, TerminationNorm, TraceLevel)).
 
-globals__get_options(globals(Options, _, _, _, _, _, _), Options).
-globals__get_gc_method(globals(_, GC_Method, _, _, _, _, _), GC_Method).
-globals__get_tags_method(globals(_, _, TagsMethod, _, _, _, _), TagsMethod).
-globals__get_args_method(globals(_, _, _, ArgsMethod, _, _, _), ArgsMethod).
-globals__get_type_info_method(globals(_, _, _, _, TypeInfoMethod, _, _),
-	TypeInfoMethod).
-globals__get_prolog_dialect(globals(_, _, _, _, _, PrologDialect, _),
+globals__get_options(globals(Options, _, _, _, _, _), Options).
+globals__get_gc_method(globals(_, GC_Method, _, _, _, _), GC_Method).
+globals__get_tags_method(globals(_, _, TagsMethod, _, _, _), TagsMethod).
+globals__get_prolog_dialect(globals(_, _, _, PrologDialect, _, _),
 	PrologDialect).
-globals__get_termination_norm(globals(_, _, _, _, _, _, TerminationNorm),
+globals__get_termination_norm(globals(_, _, _, _, TerminationNorm, _),
 	TerminationNorm).
+globals__get_trace_level(globals(_, _, _, _, _, TraceLevel), TraceLevel).
 
-globals__set_options(globals(_, B, C, D, E, F, G), Options,
-	globals(Options, B, C, D, E, F, G)).
-globals__set_gc_method(globals(A, _, C, D, E, F, G), GC_Method,
-	globals(A, GC_Method, C, D, E, F, G)).
-globals__set_tags_method(globals(A, B, _, D, E, F, G), TagsMethod,
-	globals(A, B, TagsMethod, D, E, F, G)).
-globals__set_args_method(globals(A, B, C, _, E, F, G), ArgsMethod,
-	globals(A, B, C, ArgsMethod, E, F, G)).
-globals__set_type_info_method(globals(A, B, C, D, _, F, G), TypeInfoMethod,
-	globals(A, B, C, D, TypeInfoMethod, F, G)).
-globals__set_prolog_dialect(globals(A, B, C, D, E, _, G), PrologDialect,
-	globals(A, B, C, D, E, PrologDialect, G)).
-globals__set_termination_norm(globals(A, B, C, D, E, F, _), TerminationNorm,
-	globals(A, B, C, D, E, F, TerminationNorm)).
+globals__set_options(globals(_, B, C, D, E, F), Options,
+	globals(Options, B, C, D, E, F)).
+
+globals__set_trace_level(globals(A, B, C, D, E, _), TraceLevel,
+	globals(A, B, C, D, E, TraceLevel)).
 
 globals__lookup_option(Globals, Option, OptionData) :-
 	globals__get_options(Globals, OptionTable),
@@ -309,19 +283,38 @@ globals__have_static_code_addresses_2(OptionTable, IsConst) :-
 	getopt__lookup_bool_option(OptionTable, asm_labels, AsmLabels),
 	exprn_aux__imported_is_constant(NonLocalGotos, AsmLabels, IsConst).
 
+globals__want_return_var_layouts(Globals, WantReturnLayouts) :-
+	% We need to generate layout info for call return labels
+	% if we are using accurate gc or if the user wants uplevel printing.
+	(
+		(
+			globals__get_gc_method(Globals, GC_Method),
+			GC_Method = accurate
+		;
+			globals__lookup_bool_option(Globals, trace_return,
+				TraceReturn),
+			TraceReturn = yes,
+			globals__get_trace_level(Globals, TraceLevel),
+			TraceLevel \= none
+		)
+	->
+		WantReturnLayouts = yes
+	;
+		WantReturnLayouts = no
+	).
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
-globals__io_init(Options, GC_Method, TagsMethod, ArgsMethod,
-		TypeInfoMethod, PrologDialect, TerminationNorm) -->
+globals__io_init(Options, GC_Method, TagsMethod,
+		PrologDialect, TerminationNorm, TraceLevel) -->
 	{ copy(GC_Method, GC_Method1) },
 	{ copy(TagsMethod, TagsMethod1) },
-	{ copy(ArgsMethod, ArgsMethod1) },
-	{ copy(TypeInfoMethod, TypeInfoMethod1) },
 	{ copy(PrologDialect, PrologDialect1) },
 	{ copy(TerminationNorm, TerminationNorm1) },
-	{ globals__init(Options, GC_Method1, TagsMethod1, ArgsMethod1,
-		TypeInfoMethod1, PrologDialect1, TerminationNorm1, Globals) },
+	{ copy(TraceLevel, TraceLevel1) },
+	{ globals__init(Options, GC_Method1, TagsMethod1,
+		PrologDialect1, TerminationNorm1, TraceLevel1, Globals) },
 	globals__io_set_globals(Globals).
 
 globals__io_get_gc_method(GC_Method) -->
@@ -332,14 +325,6 @@ globals__io_get_tags_method(Tags_Method) -->
 	globals__io_get_globals(Globals),
 	{ globals__get_tags_method(Globals, Tags_Method) }.
 
-globals__io_get_args_method(ArgsMethod) -->
-	globals__io_get_globals(Globals),
-	{ globals__get_args_method(Globals, ArgsMethod) }.
-
-globals__io_get_type_info_method(TypeInfoMethod) -->
-	globals__io_get_globals(Globals),
-	{ globals__get_type_info_method(Globals, TypeInfoMethod) }.
-
 globals__io_get_prolog_dialect(PrologDIalect) -->
 	globals__io_get_globals(Globals),
 	{ globals__get_prolog_dialect(Globals, PrologDIalect) }.
@@ -348,6 +333,10 @@ globals__io_get_termination_norm(TerminationNorm) -->
 	globals__io_get_globals(Globals),
 	{ globals__get_termination_norm(Globals, TerminationNorm) }.
 
+globals__io_get_trace_level(TraceLevel) -->
+	globals__io_get_globals(Globals),
+	{ globals__get_trace_level(Globals, TraceLevel) }.
+
 globals__io_get_globals(Globals) -->
 	io__get_globals(UnivGlobals),
 	{
@@ -379,6 +368,19 @@ globals__io_set_option(Option, OptionData) -->
 	{ unsafe_promise_unique(Globals1, Globals) },
 	globals__io_set_globals(Globals).
 
+globals__io_set_trace_level(TraceLevel) -->
+	globals__io_get_globals(Globals0),
+	{ globals__set_trace_level(Globals0, TraceLevel, Globals1) },
+	{ unsafe_promise_unique(Globals1, Globals) },
+		% XXX there is a bit of a design flaw with regard to
+		% uniqueness and io__set_globals
+	globals__io_set_globals(Globals).
+
+	% This predicate is needed because mercury_compile.m doesn't know
+	% anything about type trace_level.
+globals__io_set_trace_level_none -->
+	globals__io_set_trace_level(none).
+
 %-----------------------------------------------------------------------------%
 
 globals__io_lookup_bool_option(Option, Value) -->
