@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1997 The University of Melbourne.
+% Copyright (C) 1996-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -13,8 +13,8 @@
 
 :- interface.
 
-:- import_module hlds_pred, llds, prog_data, (inst).
-:- import_module bool, list, map, varset.
+:- import_module hlds_pred, llds, prog_data, (inst), term.
+:- import_module bool, list, map, std_util.
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
@@ -29,26 +29,45 @@
 			;	int_const(int)
 			;	string_const(string)
 			;	float_const(float)
-			;	pred_const(pred_id, proc_id)
+			;	pred_const(pred_id, proc_id,
+					lambda_eval_method)
 			;	code_addr_const(pred_id, proc_id)
 				% Used for constructing type_infos.
 				% Note that a pred_const is for a closure
 				% whereas a code_addr_const is just an address.
-			;	base_type_info_const(string, string, int).
+			;	type_ctor_info_const(module_name, string, int)
 				% module name, type name, type arity
+			;	base_typeclass_info_const(module_name,
+					class_id, int, string)
+				% module name of instance declaration
+				% (not filled in so that link errors result
+				% from overlapping instances),
+				% class name and arity,
+				% class instance, a string encoding the type
+				% names and arities of the arguments to the
+				% instance declaration 
+			;	tabling_pointer_const(pred_id, proc_id)
+				% The address of the static variable
+				% that points to the table that implements
+				% memoization, loop checking or the minimal
+				% model semantics for the given procedure.
+			.
 
 	% A cons_defn is the definition of a constructor (i.e. a constant
 	% or a functor) for a particular type.
 
 :- type hlds_cons_defn
 	--->	hlds_cons_defn(
-			% maybe add tvarset?
+			% maybe add tvarset here?
+			% you can get the tvarset from the hlds__type_defn.
+			existq_tvars,		% existential type vars
+			list(class_constraint), % existential class constraints
 			list(type),		% The types of the arguments
 						% of this functor (if any)
 			type_id,		% The result type, i.e. the
 						% type to which this
 						% cons_defn belongs.
-			term__context		% The location of this
+			prog_context		% The location of this
 						% ctor definition in the
 						% original source code
 		).
@@ -59,13 +78,13 @@
 
 	% Given a cons_id and a list of argument terms, convert it into a
 	% term. Fails if the cons_id is a pred_const, code_addr_const or
-	% base_type_info_const.
+	% type_ctor_info_const.
 
-:- pred cons_id_and_args_to_term(cons_id, list(term), term).
+:- pred cons_id_and_args_to_term(cons_id, list(term(T)), term(T)).
 :- mode cons_id_and_args_to_term(in, in, out) is semidet.
 
 	% Get the arity of a cons_id, aborting on pred_const, code_addr_const
-	% and base_type_info_const.
+	% and type_ctor_info_const.
 
 :- pred cons_id_arity(cons_id, arity).
 :- mode cons_id_arity(in, out) is det.
@@ -87,8 +106,8 @@
 
 :- implementation.
 
-:- import_module prog_util.
-:- import_module require, std_util, term.
+:- import_module prog_util, varset.
+:- import_module require.
 
 cons_id_and_args_to_term(int_const(Int), [], Term) :-
 	term__context_init(Context),
@@ -106,20 +125,32 @@ cons_id_arity(cons(_, Arity), Arity).
 cons_id_arity(int_const(_), 0).
 cons_id_arity(string_const(_), 0).
 cons_id_arity(float_const(_), 0).
-cons_id_arity(pred_const(_, _), _) :-
+cons_id_arity(pred_const(_, _, _), _) :-
 	error("cons_id_arity: can't get arity of pred_const").
 cons_id_arity(code_addr_const(_, _), _) :-
 	error("cons_id_arity: can't get arity of code_addr_const").
-cons_id_arity(base_type_info_const(_, _, _), _) :-
-	error("cons_id_arity: can't get arity of base_type_info_const").
+cons_id_arity(type_ctor_info_const(_, _, _), _) :-
+	error("cons_id_arity: can't get arity of type_ctor_info_const").
+cons_id_arity(base_typeclass_info_const(_, _, _, _), _) :-
+	error("cons_id_arity: can't get arity of base_typeclass_info_const").
+cons_id_arity(tabling_pointer_const(_, _), _) :-
+	error("cons_id_arity: can't get arity of tabling_pointer_const").
 
-make_functor_cons_id(term__atom(Name), Arity, cons(unqualified(Name), Arity)).
+make_functor_cons_id(term__atom(Name), Arity,
+		cons(unqualified(Name), Arity)).
 make_functor_cons_id(term__integer(Int), _, int_const(Int)).
 make_functor_cons_id(term__string(String), _, string_const(String)).
 make_functor_cons_id(term__float(Float), _, float_const(Float)).
 
 make_cons_id(SymName0, Args, TypeId, cons(SymName, Arity)) :-
+	% Use the module qualifier on the SymName, if there is one,
+	% otherwise use the module qualifier on the Type, if there is one,
+	% otherwise leave it unqualified.
+	% XXX is that the right thing to do?
 	(
+		SymName0 = qualified(_, _),
+		SymName = SymName0
+	;
 		SymName0 = unqualified(ConsName),
 		(
 			TypeId = unqualified(_) - _,
@@ -128,9 +159,6 @@ make_cons_id(SymName0, Args, TypeId, cons(SymName, Arity)) :-
 			TypeId = qualified(TypeModule, _) - _,
 			SymName = qualified(TypeModule, ConsName)
 		)
-	;
-		SymName0 = qualified(_, _),
-		SymName = SymName0
 	),
 	list__length(Args, Arity).
 
@@ -156,7 +184,7 @@ make_cons_id(SymName0, Args, TypeId, cons(SymName, Arity)) :-
 :- type hlds_type_defn.
 
 :- pred hlds_data__set_type_defn(tvarset, list(type_param),
-	hlds_type_body, import_status, term__context, hlds_type_defn).
+	hlds_type_body, import_status, prog_context, hlds_type_defn).
 :- mode hlds_data__set_type_defn(in, in, in, in, in, out) is det.
 
 :- pred hlds_data__get_type_defn_tvarset(hlds_type_defn, tvarset).
@@ -171,7 +199,7 @@ make_cons_id(SymName0, Args, TypeId, cons(SymName, Arity)) :-
 :- pred hlds_data__get_type_defn_status(hlds_type_defn, import_status).
 :- mode hlds_data__get_type_defn_status(in, out) is det.
 
-:- pred hlds_data__get_type_defn_context(hlds_type_defn, term__context).
+:- pred hlds_data__get_type_defn_context(hlds_type_defn, prog_context).
 :- mode hlds_data__get_type_defn_context(in, out) is det.
 
 :- pred hlds_data__set_type_defn_status(hlds_type_defn, import_status,
@@ -180,7 +208,7 @@ make_cons_id(SymName0, Args, TypeId, cons(SymName, Arity)) :-
 
 	% An `hlds_type_body' holds the body of a type definition:
 	% du = discriminated union, uu = undiscriminated union,
-	% eqv_type = equivalence type (a type defined to be equivalen
+	% eqv_type = equivalence type (a type defined to be equivalent
 	% to some other type)
 
 :- type hlds_type_body
@@ -222,11 +250,12 @@ make_cons_id(SymName0, Args, TypeId, cons(SymName, Arity)) :-
 			% a word containing the specified integer value.
 			% This is used for enumerations and character
 			% constants as well as for int constants.
-	;	pred_closure_tag(pred_id, proc_id)
+	;	pred_closure_tag(pred_id, proc_id, lambda_eval_method)
 			% Higher-order pred closures tags.
 			% These are represented as a pointer to
 			% an argument vector.
-			% The first two words of the argument vector
+			% For closures with lambda_eval_method `normal',
+			% the first two words of the argument vector
 			% hold the number of args and the address of
 			% the procedure respectively.
 			% The remaining words hold the arguments.
@@ -235,29 +264,45 @@ make_cons_id(SymName0, Args, TypeId, cons(SymName, Arity)) :-
 			% (used for constructing type_infos).
 			% The word just contains the address of the
 			% specified procedure.
-	;	base_type_info_constant(string, string, int)
-			% This is how we refer to base_type_info structures
-			% represented as global data. The two strings are
-			% the name of the module the type is defined in
-			% and the name of the type, while the integer is
-			% the arity.
-	;	simple_tag(tag_bits)
-			% This is for constants or functors which only
-			% require a simple tag.  (A "simple" tag is one
-			% which fits on the bottom of a pointer - i.e.
-			% two bits for 32-bit architectures, or three bits
-			% for 64-bit architectures).
+	;	type_ctor_info_constant(module_name, string, arity)
+			% This is how we refer to type_ctor_info structures
+			% represented as global data. The args are
+			% the name of the module the type is defined in,
+			% and the name of the type, and its arity.
+	;	base_typeclass_info_constant(module_name, class_id, string)
+			% This is how we refer to base_typeclass_info structures
+			% represented as global data. The first argument is the
+			% name of the module containing the instance declration,
+			% the second is the class name and arity, while the
+			% third is the string which uniquely identifies the
+			% instance declaration (it is made from the type of
+			% the arguments to the instance decl).
+	;	tabling_pointer_constant(pred_id, proc_id)
+			% This is how we refer to tabling pointer variables
+			% represented as global data. The word just contains
+			% the address of the tabling pointer of the
+			% specified procedure.
+	;	unshared_tag(tag_bits)
+			% This is for constants or functors which can be
+			% distinguished with just a primary tag.
+			% An "unshared" tag is one which fits on the
+			% bottom of a pointer (i.e.  two bits for
+			% 32-bit architectures, or three bits for 64-bit
+			% architectures), and is used for just one
+			% functor.
 			% For constants we store a tagged zero, for functors
 			% we store a tagged pointer to the argument vector.
-	;	complicated_tag(tag_bits, int)
+	;	shared_remote_tag(tag_bits, int)
 			% This is for functors or constants which
 			% require more than just a two-bit tag. In this case,
 			% we use both a primary and a secondary tag.
+			% Several functors share the primary tag and are
+			% distinguished by the secondary tag.
 			% The secondary tag is stored as the first word of
 			% the argument vector. (If it is a constant, then
 			% in this case there is an argument vector of size 1
 			% which just holds the secondary tag.)
-	;	complicated_constant_tag(tag_bits, int)
+	;	shared_local_tag(tag_bits, int)
 			% This is for constants which require more than a
 			% two-bit tag. In this case, we use both a primary
 			% and a secondary tag, but this time the secondary
@@ -268,7 +313,7 @@ make_cons_id(SymName0, Args, TypeId, cons(SymName, Arity)) :-
 			% In this case, we don't need to store the functor,
 			% and instead we store the argument directly.
 
-	% The type `tag_bits' holds a simple tag value.
+	% The type `tag_bits' holds a primary tag value.
 
 :- type tag_bits	==	int.	% actually only 2 (or maybe 3) bits
 
@@ -293,7 +338,7 @@ make_cons_id(SymName0, Args, TypeId, cons(SymName, Arity)) :-
 %				% :- type sorted_list(T) == list(T)
 %				%	where sorted.
 
-			term__context		% The location of this type
+			prog_context		% The location of this type
 						% definition in the original
 						% source code
 		).
@@ -352,7 +397,7 @@ hlds_data__set_type_defn_status(hlds_type_defn(A, B, C, _, E), Status,
 
 :- type hlds_inst_defn
 	--->	hlds_inst_defn(
-			varset,			% The names of the inst
+			inst_varset,		% The names of the inst
 						% parameters (if any).
 			list(inst_param),	% The inst parameters (if any).
 						% ([I] in the above example.)
@@ -360,7 +405,7 @@ hlds_data__set_type_defn_status(hlds_type_defn(A, B, C, _, E), Status,
 			condition,		% Unused (reserved for
 						% holding a user-defined 
 						% invariant).
-			term__context,		% The location in the source
+			prog_context,		% The location in the source
 						% code of this inst definition.
 
 			import_status		% So intermod.m can tell 
@@ -549,7 +594,7 @@ user_inst_table_optimize(user_inst_table(InstDefns0, InstIds0),
 
 :- type hlds_mode_defn
 	--->	hlds_mode_defn(
-			varset,			% The names of the inst
+			inst_varset,		% The names of the inst
 						% parameters (if any).
 			list(inst_param),	% The list of the inst
 						% parameters (if any).
@@ -559,7 +604,7 @@ user_inst_table_optimize(user_inst_table(InstDefns0, InstIds0),
 			condition,		% Unused (reserved for
 						% holding a user-defined
 						% invariant).
-			term__context,		% The location of this mode
+			prog_context,		% The location of this mode
 						% definition in the original
 						% source code.
 			import_status		% So intermod.m can tell 
@@ -625,14 +670,10 @@ mode_table_optimize(mode_table(ModeDefns0, ModeIds0),
 %-----------------------------------------------------------------------------%
 :- interface.
 
-:- type determinism	--->	det
-			;	semidet
-			;	nondet
-			;	multidet
-			;	cc_nondet
-			;	cc_multidet
-			;	erroneous
-			;	failure.
+%
+% Types and procedures for decomposing and analysing determinism.
+% The `determinism' type itself is defined in prog_data.m.
+%
 
 :- type can_fail	--->	can_fail
 			;	cannot_fail.
@@ -654,6 +695,7 @@ mode_table_optimize(mode_table(ModeDefns0, ModeIds0),
 
 :- pred determinism_to_code_model(determinism, code_model).
 :- mode determinism_to_code_model(in, out) is det.
+:- mode determinism_to_code_model(out, in) is multidet.
 
 :- implementation.
 
@@ -676,3 +718,153 @@ determinism_to_code_model(erroneous,   model_det).
 determinism_to_code_model(failure,     model_semi).
 
 %-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
+
+:- interface.
+
+:- type class_table == map(class_id, hlds_class_defn).
+
+:- type class_id 	--->	class_id(sym_name, arity).
+
+	% Information about a single `typeclass' declaration
+:- type hlds_class_defn 
+	--->	hlds_class_defn(
+			import_status,
+			list(class_constraint), % SuperClasses
+			list(tvar),		% ClassVars 
+			class_interface,	% The interface from the
+						% original declaration,
+						% used by intermod.m to
+						% write out the interface
+						% for a local typeclass to
+						% the `.opt' file.
+			hlds_class_interface, 	% Methods
+			tvarset,		% VarNames
+			prog_context		% Location of declaration
+		).
+
+:- type hlds_class_interface	==	list(hlds_class_proc).	
+:- type hlds_class_proc
+	---> 	hlds_class_proc(
+			pred_id,
+			proc_id
+		).
+
+	% For each class, we keep track of a list of its instances, since there
+	% can be more than one instance of each class.
+:- type instance_table == map(class_id, list(hlds_instance_defn)).
+
+	% Information about a single `instance' declaration
+:- type hlds_instance_defn 
+	--->	hlds_instance_defn(
+			import_status,		% import status of the instance
+						% declaration
+			prog_context,		% context of declaration
+			list(class_constraint), % Constraints
+			list(type), 		% ClassTypes 
+			instance_body, 		% Methods
+			maybe(hlds_class_interface),
+						% After check_typeclass, we 
+						% will know the pred_ids and
+						% proc_ids of all the methods
+			tvarset,		% VarNames
+			map(class_constraint, constraint_proof)
+						% "Proofs" of how to build the
+						% typeclass_infos for the
+						% superclasses of this class,
+						% for this instance
+		).
+
+	% `Proof' of why a constraint is redundant
+:- type constraint_proof			
+			% Apply the instance decl with the given number.
+			% Note that we don't store the actual 
+			% hlds_instance_defn for two reasons:
+			% - That would require storing a renamed version of
+			%   the constraint_proofs for *every* use of an
+			%   instance declaration. This would't even get GCed
+			%   for a long time because it would be stored in
+			%   the pred_info.
+			% - The superclass proofs stored in the
+			%   hlds_instance_defn would need to store all the
+			%   constraint_proofs for all its ancestors. This
+			%   would require the class relation to be
+			%   topologically sorted before checking the
+			%   instance declarations.
+	--->	apply_instance(int)
+
+			% The constraint is redundant because of the
+			% following class's superclass declaration
+	;	superclass(class_constraint).
+
+%-----------------------------------------------------------------------------%
+
+:- type subclass_details 
+	--->	subclass_details(
+			list(tvar),		% variables of the superclass
+			class_id,		% name of the subclass
+			list(tvar),		% variables of the subclass
+			tvarset			% the names of these vars
+		).
+
+:- import_module multi_map.
+
+	% I'm sure there's a very clever way of 
+	% doing this with graphs or relations...
+:- type superclass_table == multi_map(class_id, subclass_details).
+
+%-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
+
+:- interface.
+
+	%
+	% A table that records all the assertions in the system.
+	% An assertion is a goal that will always evaluate to true,
+	% subject to the constraints imposed by the quantifiers.
+	%
+	% ie :- promise all [A] some [B] (B > A)
+	% 
+	% The above assertion states that for all possible values of A,
+	% there will exist at least one value, B, such that B is greater
+	% then A.
+	%
+:- type assert_id.
+:- type assertion_table.
+
+:- pred assertion_table_init(assertion_table::out) is det.
+
+:- pred assertion_table_add_assertion(pred_id::in, assertion_table::in,
+		assert_id::out, assertion_table::out) is det.
+
+:- pred assertion_table_lookup(assertion_table::in, assert_id::in,
+		pred_id::out) is det.
+
+:- pred assertion_table_pred_ids(assertion_table::in,
+		list(pred_id)::out) is det.
+
+:- implementation.
+
+:- import_module int.
+
+:- type assert_id == int.
+:- type assertion_table 
+	---> 	assertion_table(assert_id, map(assert_id, pred_id)).
+
+assertion_table_init(assertion_table(0, AssertionMap)) :-
+	map__init(AssertionMap).
+
+assertion_table_add_assertion(Assertion, AssertionTable0, Id, AssertionTable) :-
+	AssertionTable0 = assertion_table(Id, AssertionMap0),
+	map__det_insert(AssertionMap0, Id, Assertion, AssertionMap),
+	AssertionTable = assertion_table(Id + 1, AssertionMap).
+
+assertion_table_lookup(AssertionTable, Id, Assertion) :-
+	AssertionTable = assertion_table(_MaxId, AssertionMap),
+	map__lookup(AssertionMap, Id, Assertion).
+
+assertion_table_pred_ids(assertion_table(_, AssertionMap), PredIds) :-
+	map__values(AssertionMap, PredIds).
+
+%-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
