@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1997 The University of Melbourne.
+% Copyright (C) 1996-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -18,8 +18,11 @@
 :- module instmap.
 
 :- interface.
-:- import_module hlds_module, prog_data, mode_info, (inst).
-:- import_module set, term.
+
+:- import_module hlds_module, prog_data, mode_info, (inst), mode_errors.
+:- import_module hlds_data.
+
+:- import_module map, bool, set, list, assoc_list, std_util.
 
 :- type instmap.
 :- type instmap_delta.
@@ -72,13 +75,14 @@
 :- pred instmap_delta_is_unreachable(instmap_delta).
 :- mode instmap_delta_is_unreachable(in) is semidet.
 
-:- pred instmap__from_assoc_list(assoc_list(var, inst), instmap).
+:- pred instmap__from_assoc_list(assoc_list(prog_var, inst), instmap).
 :- mode instmap__from_assoc_list(in, out) is det.
 
-:- pred instmap_delta_from_assoc_list(assoc_list(var, inst), instmap_delta).
+:- pred instmap_delta_from_assoc_list(assoc_list(prog_var, inst),
+		instmap_delta).
 :- mode instmap_delta_from_assoc_list(in, out) is det.
 
-:- pred instmap_delta_from_mode_list(list(var), list(mode),
+:- pred instmap_delta_from_mode_list(list(prog_var), list(mode),
 		module_info, instmap_delta).
 :- mode instmap_delta_from_mode_list(in, in, in, out) is det.
 
@@ -86,69 +90,93 @@
 
 	% Return the set of variables in an instmap.
 	%
-:- pred instmap__vars(instmap, set(var)).
+:- pred instmap__vars(instmap, set(prog_var)).
 :- mode instmap__vars(in, out) is det.
 
 	% Return the list of variables in an instmap.
 	%
-:- pred instmap__vars_list(instmap, list(var)).
+:- pred instmap__vars_list(instmap, list(prog_var)).
 :- mode instmap__vars_list(in, out) is det.
 
 	% Return the set of variables whose instantiations have
 	% changed (or our knowledge about them has changed) across
 	% an instmap_delta.
 	%
-:- pred instmap_delta_changed_vars(instmap_delta, set(var)).
+	% This predicate shouldn't be used if you want your code to
+	% compile on the alias branch, use instmap_changed_vars instead.
+	%
+:- pred instmap_delta_changed_vars(instmap_delta, set(prog_var)).
 :- mode instmap_delta_changed_vars(in, out) is det.
 
+	%
+	% instmap_changed_vars(IMA, IMB, MI, CV)
+	%
+	% Given an earlier instmap, IMA, and a later instmap, IMB,
+	% determine what variables, CV, have had their instantiatedness
+	% information changed.
+	%
+	% This predicate is meant to be equivalent to
+	% instmap_delta_changed_vars, where the instmap_delta is simply
+	% the one to take IMA to IMB.  However this predicate should
+	% transform more easily to the alias branch.
+	%
+:- pred instmap_changed_vars(instmap::in, instmap::in, module_info::in,
+		set(prog_var)::out) is det.
+
 %-----------------------------------------------------------------------------%
 
 	% Given an instmap and a variable, determine the inst of
 	% that variable.
 	%
-:- pred instmap__lookup_var(instmap, var, inst).
+:- pred instmap__lookup_var(instmap, prog_var, inst).
 :- mode instmap__lookup_var(in, in, out) is det.
 
 	% Given an instmap_delta and a variable, determine the new inst
 	% of that variable (if any).
 	%
-:- pred instmap_delta_search_var(instmap_delta, var, inst).
+:- pred instmap_delta_search_var(instmap_delta, prog_var, inst).
 :- mode instmap_delta_search_var(in, in, out) is semidet.
 
 	% Given an instmap and a list of variables, return a list
 	% containing the insts of those variable.
 	%
-:- pred instmap__lookup_vars(list(var), instmap, list(inst)).
+:- pred instmap__lookup_vars(list(prog_var), instmap, list(inst)).
 :- mode instmap__lookup_vars(in, in, out) is det.
 
 	% Insert an entry into an instmap_delta.  Note that you
 	% cannot call instmap_delta_insert for a variable already
 	% present.
 	%
-:- pred instmap_delta_insert(instmap_delta, var, inst, instmap_delta).
+:- pred instmap_delta_insert(instmap_delta, prog_var, inst, instmap_delta).
 :- mode instmap_delta_insert(in, in, in, out) is det.
 
 	% Set an entry in an instmap.
 	%
-:- pred instmap__set(instmap, var, inst, instmap).
+:- pred instmap__set(instmap, prog_var, inst, instmap).
 :- mode instmap__set(in, in, in, out) is det.
 
-:- pred instmap_delta_set(instmap_delta, var, inst, instmap_delta).
+	% Set multiple entries in an instmap.
+	%
+:- pred instmap__set_vars(instmap, list(prog_var), list(inst), instmap).
+:- mode instmap__set_vars(in, in, in, out) is det.
+
+:- pred instmap_delta_set(instmap_delta, prog_var, inst, instmap_delta).
 :- mode instmap_delta_set(in, in, in, out) is det.
 
 	% Bind a variable in an instmap to a functor at the beginning
 	% of a case in a switch. Aborts on compiler generated cons_ids.
-:- pred instmap_delta_bind_var_to_functor(var, cons_id, instmap,
+:- pred instmap_delta_bind_var_to_functor(prog_var, type, cons_id, instmap,
 		instmap_delta, instmap_delta, module_info, module_info).
-:- mode instmap_delta_bind_var_to_functor(in, in, in, in, out, in, out) is det.
+:- mode instmap_delta_bind_var_to_functor(in, in, in, in, in, out, in, out)
+		is det.
 
-:- pred instmap__bind_var_to_functor(var, cons_id,
+:- pred instmap__bind_var_to_functor(prog_var, type, cons_id,
 		instmap, instmap, module_info, module_info).
-:- mode instmap__bind_var_to_functor(in, in, in, out, in, out) is det.
+:- mode instmap__bind_var_to_functor(in, in, in, in, out, in, out) is det.
 
 	% Update the given instmap to include the initial insts of the
 	% lambda variables.
-:- pred instmap__pre_lambda_update(module_info, list(var), list(mode),
+:- pred instmap__pre_lambda_update(module_info, list(prog_var), list(mode),
 		instmap, instmap).
 :- mode instmap__pre_lambda_update(in, in, in, in, out) is det.
 
@@ -158,7 +186,7 @@
 	% which records the change in the instantiation state of those
 	% variables.
 	%
-:- pred compute_instmap_delta(instmap, instmap, set(var), instmap_delta).
+:- pred compute_instmap_delta(instmap, instmap, set(prog_var), instmap_delta).
 :- mode compute_instmap_delta(in, in, in, out) is det.
 
 	% Given an instmap and an instmap_delta, apply the instmap_delta
@@ -181,43 +209,56 @@
 	%       instantiatedness of all the nonlocal variables,
 	%       checking that it is the same for every branch.
 	%
-:- pred instmap__merge(set(var), list(instmap), merge_context,
+:- pred instmap__merge(set(prog_var), list(instmap), merge_context,
 		mode_info, mode_info).
 :- mode instmap__merge(in, in, in, mode_info_di, mode_info_uo) is det.
 
+	% instmap__unify(NonLocalVars, InstMapNonlocalvarPairss):
+	%       Unify the `InstMaps' in the list of pairs resulting
+	%	from different branches of a parallel conjunction and
+	%	update the instantiatedness of all the nonlocal variables.
+	%	The variable locking that is done when modechecking
+	%	the individual conjuncts ensures that variables have
+	%	at most one producer.
+	%
+:- pred instmap__unify(set(prog_var), list(pair(instmap, set(prog_var))),
+		mode_info, mode_info).
+:- mode instmap__unify(in, in, mode_info_di, mode_info_uo) is det.
+
 	% instmap__restrict takes an instmap and a set of vars and
 	% returns an instmap with its domain restricted to those
 	% vars.
 	%
-:- pred instmap__restrict(instmap, set(var), instmap).
+:- pred instmap__restrict(instmap, set(prog_var), instmap).
 :- mode instmap__restrict(in, in, out) is det.
 
 	% instmap_delta_restrict takes an instmap and a set of vars
 	% and returns an instmap_delta with its domain restricted to
 	% those vars.
 	%
-:- pred instmap_delta_restrict(instmap_delta, set(var), instmap_delta).
+:- pred instmap_delta_restrict(instmap_delta, set(prog_var), instmap_delta).
 :- mode instmap_delta_restrict(in, in, out) is det.
 
 	% instmap_delta_delete_vars takes an instmap_delta and a list of
 	% vars and returns an instmap_delta with those vars removed from
 	% its domain.
 	%
-:- pred instmap_delta_delete_vars(instmap_delta, list(var), instmap_delta).
+:- pred instmap_delta_delete_vars(instmap_delta, list(prog_var), instmap_delta).
 :- mode instmap_delta_delete_vars(in, in, out) is det.
 
 	% `instmap__no_output_vars(Instmap, InstmapDelta, Vars, ModuleInfo)'
 	% is true if none of the vars in the set Vars could have become more
 	% instantiated when InstmapDelta is applied to Instmap.
-:- pred instmap__no_output_vars(instmap, instmap_delta, set(var), module_info).
+:- pred instmap__no_output_vars(instmap, instmap_delta, set(prog_var),
+		module_info).
 :- mode instmap__no_output_vars(in, in, in, in) is semidet.
 
 	% merge_instmap_delta(InitialInstMap, NonLocals,
 	%	InstMapDeltaA, InstMapDeltaB, ModuleInfo0, ModuleInfo)
-	% Merge the instmap_deltas of different branches of an ite, disj
-	% or switch.
-:- pred merge_instmap_delta(instmap, set(var), instmap_delta, instmap_delta,
-		instmap_delta, module_info, module_info).
+	% Merge the instmap_deltas of different branches of an if-then-else,
+	% disj or switch.
+:- pred merge_instmap_delta(instmap, set(prog_var), instmap_delta,
+		instmap_delta, instmap_delta, module_info, module_info).
 :- mode merge_instmap_delta(in, in, in, in, out, in, out) is det.
 
 	% merge_instmap_deltas(Vars, InstMapDeltas,
@@ -225,10 +266,18 @@
 	% takes a list of instmap deltas from the branches of an if-then-else,
 	% switch, or disj and merges them. This is used in situations
 	% where the bindings are known to be compatible.
-:- pred merge_instmap_deltas(instmap, set(var), list(instmap_delta),
+:- pred merge_instmap_deltas(instmap, set(prog_var), list(instmap_delta),
 		instmap_delta, module_info, module_info).
 :- mode merge_instmap_deltas(in, in, in, out, in, out) is det.
 
+	% unify_instmap_delta(InitialInstMap, NonLocals,
+	%	InstMapDeltaA, InstMapDeltaB, ModuleInfo0, ModuleInfo)
+	% Unify the instmap_deltas of different branches of a parallel
+	% conjunction.
+:- pred unify_instmap_delta(instmap, set(prog_var), instmap_delta,
+		instmap_delta, instmap_delta, module_info, module_info).
+:- mode unify_instmap_delta(in, in, in, in, out, in, out) is det.
+
 %-----------------------------------------------------------------------------%
 
 	% `instmap_delta_apply_sub(InstmapDelta0, Must, Sub, InstmapDelta)'
@@ -237,19 +286,19 @@
 	% InstmapDelta0 which does not appear in Sub, it is ignored if
 	% Must is set to no, otherwise it is an error.
 	%
-:- pred instmap_delta_apply_sub(instmap_delta, bool, map(var, var),
+:- pred instmap_delta_apply_sub(instmap_delta, bool, map(prog_var, prog_var),
 		instmap_delta).
 :- mode instmap_delta_apply_sub(in, in, in, out) is det.
 
-:- pred instmap__apply_sub(instmap, bool, map(var, var), instmap).
+:- pred instmap__apply_sub(instmap, bool, map(prog_var, prog_var), instmap).
 :- mode instmap__apply_sub(in, in, in, out) is det.
 
 %-----------------------------------------------------------------------------%
 
-:- pred instmap__to_assoc_list(instmap, assoc_list(var, inst)).
+:- pred instmap__to_assoc_list(instmap, assoc_list(prog_var, inst)).
 :- mode instmap__to_assoc_list(in, out) is det.
 
-:- pred instmap_delta_to_assoc_list(instmap_delta, assoc_list(var, inst)).
+:- pred instmap_delta_to_assoc_list(instmap_delta, assoc_list(prog_var, inst)).
 :- mode instmap_delta_to_assoc_list(in, out) is det.
 
 %-----------------------------------------------------------------------------%
@@ -257,16 +306,17 @@
 
 :- implementation.
 
-:- import_module mode_util, inst_match, prog_data, mode_errors, goal_util.
+:- import_module mode_util, inst_match, prog_data, goal_util, type_util.
 :- import_module hlds_data, inst_util.
-:- import_module list, std_util, bool, map, set, assoc_list, require.
+
+:- import_module std_util, require, string, term.
 
 :- type instmap_delta	==	instmap.
 
 :- type instmap	--->	reachable(instmapping)
-			;	unreachable.
+		;	unreachable.
 
-:- type instmapping	==	map(var, inst).
+:- type instmapping	==	map(prog_var, inst).
 
 %-----------------------------------------------------------------------------%
 
@@ -307,7 +357,7 @@ instmap_delta_from_mode_list(Var, Modes, ModuleInfo, InstMapDelta) :-
 	instmap_delta_from_mode_list_2(Var, Modes, ModuleInfo,
 		InstMapDelta0, InstMapDelta).
 	
-:- pred instmap_delta_from_mode_list_2(list(var), list(mode),
+:- pred instmap_delta_from_mode_list_2(list(prog_var), list(mode),
 		module_info, instmap_delta, instmap_delta).
 :- mode instmap_delta_from_mode_list_2(in, in, in, in, out) is det.
 
@@ -344,6 +394,31 @@ instmap_delta_changed_vars(reachable(InstMapping), ChangedVars) :-
 	map__keys(InstMapping, ChangedVarsList),
 	set__sorted_list_to_set(ChangedVarsList, ChangedVars).
 
+%-----------------------------------------------------------------------------%
+
+instmap_changed_vars(InstMapA, InstMapB, ModuleInfo, ChangedVars) :-
+	instmap__vars_list(InstMapB, VarsB),
+	changed_vars_2(VarsB, InstMapA, InstMapB, ModuleInfo, ChangedVars).
+
+:- pred changed_vars_2(prog_vars::in, instmap::in,
+		instmap::in, module_info::in, set(prog_var)::out) is det.
+
+changed_vars_2([], _InstMapA, _InstMapB, _ModuleInfo, ChangedVars) :-
+	set__init(ChangedVars).
+changed_vars_2([VarB|VarBs], InstMapA, InstMapB, ModuleInfo, ChangedVars) :-
+	changed_vars_2(VarBs, InstMapA, InstMapB, ModuleInfo, ChangedVars0),
+
+	instmap__lookup_var(InstMapA, VarB, InitialInst),
+	instmap__lookup_var(InstMapB, VarB, FinalInst),
+
+	(
+		inst_matches_final(InitialInst, FinalInst, ModuleInfo)
+	->
+		ChangedVars = ChangedVars0
+	;
+		set__insert(ChangedVars0, VarB, ChangedVars)
+	).
+
 %-----------------------------------------------------------------------------%
 
 	% Given an instmap and a variable, determine the inst of
@@ -353,7 +428,7 @@ instmap__lookup_var(unreachable, _Var, not_reached).
 instmap__lookup_var(reachable(InstMap), Var, Inst) :-
 	instmapping_lookup_var(InstMap, Var, Inst).
 
-:- pred instmapping_lookup_var(instmapping, var, inst).
+:- pred instmapping_lookup_var(instmapping, prog_var, inst).
 :- mode instmapping_lookup_var(in, in, out) is det.
 
 instmapping_lookup_var(InstMap, Var, Inst) :-
@@ -377,6 +452,15 @@ instmap__set(reachable(InstMapping0), Var, Inst,
 		reachable(InstMapping)) :-
 	map__set(InstMapping0, Var, Inst, InstMapping).
 
+instmap__set_vars(InstMap, [], [], InstMap).
+instmap__set_vars(InstMap0, [V | Vs], [I | Is], InstMap) :-
+	instmap__set(InstMap0, V, I, InstMap1),
+	instmap__set_vars(InstMap1, Vs, Is, InstMap).
+instmap__set_vars(_, [_ | _], [], _) :-
+	error("instmap__set_vars").
+instmap__set_vars(_, [], [_ | _], _) :-
+	error("instmap__set_vars").
+
 instmap_delta_set(unreachable, _Var, _Inst, unreachable).
 instmap_delta_set(reachable(InstMapping0), Var, Inst, Instmap) :-
 	( Inst = not_reached ->
@@ -397,7 +481,7 @@ instmap_delta_insert(reachable(InstMapping0), Var, Inst, Instmap) :-
 
 %-----------------------------------------------------------------------------%
 
-instmap_delta_bind_var_to_functor(Var, ConsId, InstMap,
+instmap_delta_bind_var_to_functor(Var, Type, ConsId, InstMap,
 		InstmapDelta0, InstmapDelta, ModuleInfo0, ModuleInfo) :-
 	( 
 		InstmapDelta0 = unreachable,
@@ -421,7 +505,7 @@ instmap_delta_bind_var_to_functor(Var, ConsId, InstMap,
 		;
 			NewInst1 = OldInst
 		),
-		bind_inst_to_functor(NewInst1, ConsId, NewInst,
+		bind_inst_to_functor(Type, ConsId, NewInst1, NewInst,
 			ModuleInfo0, ModuleInfo),
 
 		%
@@ -435,18 +519,19 @@ instmap_delta_bind_var_to_functor(Var, ConsId, InstMap,
 		)
 	).
 
-instmap__bind_var_to_functor(Var, ConsId, InstMap0, InstMap,
+instmap__bind_var_to_functor(Var, Type, ConsId, InstMap0, InstMap,
 		ModuleInfo0, ModuleInfo) :-
 	instmap__lookup_var(InstMap0, Var, Inst0),
-	bind_inst_to_functor(Inst0, ConsId, Inst, ModuleInfo0, ModuleInfo),
+	bind_inst_to_functor(Type, ConsId, Inst0, Inst,
+		ModuleInfo0, ModuleInfo),
 	instmap__set(InstMap0, Var, Inst, InstMap).
 
-:- pred bind_inst_to_functor((inst), cons_id, (inst),
+:- pred bind_inst_to_functor(type, cons_id, (inst), (inst),
 		module_info, module_info). 
-:- mode bind_inst_to_functor(in, in, out, in, out) is det.
+:- mode bind_inst_to_functor(in, in, in, out, in, out) is det.
 
-bind_inst_to_functor(Inst0, ConsId, Inst, ModuleInfo0, ModuleInfo) :-
-	cons_id_arity(ConsId, Arity),
+bind_inst_to_functor(Type, ConsId, Inst0, Inst, ModuleInfo0, ModuleInfo) :-
+	Arity = cons_id_adjusted_arity(ModuleInfo0, Type, ConsId),
 	list__duplicate(Arity, dead, ArgLives),
 	list__duplicate(Arity, free, ArgInsts),
 	(
@@ -543,7 +628,7 @@ instmap__merge(NonLocals, InstMapList, MergeContext, ModeInfo0, ModeInfo) :-
 	),
 	mode_info_set_instmap(InstMap, ModeInfo2, ModeInfo).
 
-:- pred get_reachable_instmaps(list(instmap), list(map(var,inst))).
+:- pred get_reachable_instmaps(list(instmap), list(map(prog_var, inst))).
 :- mode get_reachable_instmaps(in, out) is det.
 
 get_reachable_instmaps([], []).
@@ -562,8 +647,9 @@ get_reachable_instmaps([InstMap | InstMaps], Reachables) :-
 	%       there are two instmaps in `InstMaps' for which the inst
 	%       the variable is incompatible.
 
-:- pred instmap__merge_2(list(var), list(instmap), module_info, map(var, inst),
-			module_info, map(var, inst), merge_errors).
+:- pred instmap__merge_2(list(prog_var), list(instmap), module_info,
+		map(prog_var, inst), module_info, map(prog_var, inst),
+		merge_errors).
 :- mode instmap__merge_2(in, in, in, in, out, out, out) is det.
 
 instmap__merge_2([], _, ModuleInfo, InstMap, ModuleInfo, InstMap, []).
@@ -587,7 +673,7 @@ instmap__merge_2([Var|Vars], InstMapList, ModuleInfo0, InstMap0,
 	%       there are two instmaps for which the inst of `Var'
 	%       is incompatible.
 
-:- pred instmap__merge_var(list(instmap), var, module_info,
+:- pred instmap__merge_var(list(instmap), prog_var, module_info,
 				list(inst), inst, module_info, bool).
 :- mode instmap__merge_var(in, in, in, out, out, out, out) is det.
 
@@ -621,7 +707,7 @@ merge_instmap_deltas(InstMap, NonLocals, InstMapDeltaList, MergedDelta,
 			MergedDelta, ModuleInfo0, ModuleInfo)
 	).
 
-:- pred merge_instmap_deltas(instmap, set(var), instmap_delta,
+:- pred merge_instmap_deltas(instmap, set(prog_var), instmap_delta,
 		list(instmap_delta), instmap_delta, module_info, module_info).
 :- mode merge_instmap_deltas(in, in, in, in, out, in, out) is det.
 
@@ -635,6 +721,125 @@ merge_instmap_deltas(InstMap, NonLocals, MergedDelta0, [Delta|Deltas],
 		MergedDelta, ModuleInfo1, ModuleInfo).
 
 %-----------------------------------------------------------------------------%
+
+instmap__unify(NonLocals, InstMapList, ModeInfo0, ModeInfo) :-
+	(
+			% If any of the instmaps is unreachable, then
+			% the final instmap is unreachable.
+		list__member(unreachable - _, InstMapList)
+	->
+		mode_info_set_instmap(unreachable, ModeInfo0, ModeInfo)
+	;
+			% If there is only one instmap, then we just
+			% stick it in the mode_info.
+		InstMapList = [InstMap - _]
+	->
+		mode_info_set_instmap(InstMap, ModeInfo0, ModeInfo)
+	;
+		InstMapList = [InstMap0 - _|InstMapList1],
+		InstMap0 = reachable(InstMapping0)
+	->
+			% having got the first instmapping, to use as
+			% an accumulator, all instmap__unify_2 which
+			% unifies each of the nonlocals from each instmap
+			% with the corresponding inst in the accumulator.
+		mode_info_get_module_info(ModeInfo0, ModuleInfo0),
+		set__to_sorted_list(NonLocals, NonLocalsList),
+		instmap__unify_2(NonLocalsList, InstMap0, InstMapList1,
+			ModuleInfo0, InstMapping0, ModuleInfo,
+			InstMapping, ErrorList),
+		mode_info_set_module_info(ModeInfo0, ModuleInfo, ModeInfo1),
+			
+			% If there were any errors, then add the error
+			% to the list of possible errors in the mode_info.
+		( ErrorList = [FirstError | _] ->
+			FirstError = Var - _,
+			set__singleton_set(WaitingVars, Var),
+			mode_info_error(WaitingVars,
+				mode_error_par_conj(ErrorList),
+				ModeInfo1, ModeInfo2)
+		;
+			ModeInfo2 = ModeInfo1
+		),
+		mode_info_set_instmap(reachable(InstMapping),
+			ModeInfo2, ModeInfo)
+	;
+		ModeInfo = ModeInfo0
+	).
+
+%-----------------------------------------------------------------------------%
+
+	% instmap__unify_2(Vars, InitialInstMap, InstMaps, ModuleInfo,
+	%		ErrorList):
+	%       Let `ErrorList' be the list of variables in `Vars' for
+	%       which there are two instmaps in `InstMaps' for which the insts
+	%       of the variable is incompatible.
+:- pred instmap__unify_2(list(prog_var), instmap, list(pair(instmap,
+		set(prog_var))), module_info, map(prog_var, inst), module_info,
+		map(prog_var, inst), merge_errors).
+:- mode instmap__unify_2(in, in, in, in, in, out, out, out) is det.
+
+instmap__unify_2([], _, _, ModuleInfo, InstMap, ModuleInfo, InstMap, []).
+instmap__unify_2([Var|Vars], InitialInstMap, InstMapList, ModuleInfo0, InstMap0,
+			ModuleInfo, InstMap, ErrorList) :-
+	instmap__unify_2(Vars, InitialInstMap, InstMapList, ModuleInfo0,
+			InstMap0, ModuleInfo1, InstMap1, ErrorList1),
+	instmap__lookup_var(InitialInstMap, Var, InitialVarInst),
+	instmap__unify_var(InstMapList, Var, [], Insts, InitialVarInst, Inst,
+		ModuleInfo1, ModuleInfo, no, Error),
+	( Error = yes ->
+		ErrorList = [Var - Insts | ErrorList1]
+	;
+		ErrorList = ErrorList1
+	),
+	map__set(InstMap1, Var, Inst, InstMap).
+
+	% instmap__unify_var(InstMaps, Var, InitialInstMap, ModuleInfo,
+	%		Insts, Error):
+	%       Let `Insts' be the list of the inst of `Var' in
+	%       each of the corresponding `InstMaps'.  Let `Error' be yes
+	%	iff there are two instmaps for which the inst of `Var'
+	%       is incompatible.
+
+:- pred instmap__unify_var(list(pair(instmap, set(prog_var))), prog_var,
+		list(inst), list(inst), inst, inst, module_info, module_info,
+		bool, bool).
+:- mode instmap__unify_var(in, in, in, out, in, out, in, out, in, out) is det.
+
+instmap__unify_var([], _, Insts, Insts, Inst, Inst, ModuleInfo, ModuleInfo,
+		Error, Error).
+instmap__unify_var([InstMap - Nonlocals| Rest], Var, InstList0, InstList,
+		Inst0, Inst, ModuleInfo0, ModuleInfo, Error0, Error) :-
+	(
+		set__member(Var, Nonlocals)
+	->
+		instmap__lookup_var(InstMap, Var, VarInst),
+		(
+			% We can ignore the determinism of the unification:
+			% if it isn't det, then there will be a mode error
+			% or a determinism error in one of the parallel
+			% conjuncts.
+
+			abstractly_unify_inst(live, Inst0, VarInst, fake_unify,
+				ModuleInfo0, Inst1, _Det, ModuleInfo1)
+		->
+			Inst2 = Inst1,
+			ModuleInfo2 = ModuleInfo1,
+			Error1 = Error0
+		;
+			Error1 = yes,
+			ModuleInfo2 = ModuleInfo0,
+			Inst2 = not_reached
+		)
+	;
+		VarInst = free,
+		Inst2 = Inst0,
+		Error1 = Error0,
+		ModuleInfo2 = ModuleInfo0
+	),
+	instmap__unify_var(Rest, Var, [VarInst | InstList0], InstList,
+		Inst2, Inst, ModuleInfo2, ModuleInfo, Error1, Error).
+
 %-----------------------------------------------------------------------------%
 
 	% Given two instmaps and a set of variables, compute an instmap delta
@@ -649,8 +854,8 @@ compute_instmap_delta(reachable(InstMapA), reachable(InstMapB), NonLocals,
 	compute_instmap_delta_2(NonLocalsList, InstMapA, InstMapB, AssocList),
 	map__from_sorted_assoc_list(AssocList, DeltaInstMap).
 
-:- pred compute_instmap_delta_2(list(var), instmapping, instmapping,
-					assoc_list(var, inst)).
+:- pred compute_instmap_delta_2(list(prog_var), instmapping, instmapping,
+					assoc_list(prog_var, inst)).
 :- mode compute_instmap_delta_2(in, in, in, out) is det.
 
 compute_instmap_delta_2([], _, _, []).
@@ -672,7 +877,8 @@ instmap__no_output_vars(InstMap0, reachable(InstMapDelta), Vars, M) :-
 	set__to_sorted_list(Vars, VarList),
 	instmap__no_output_vars_2(VarList, InstMap0, InstMapDelta, M).
 
-:- pred instmap__no_output_vars_2(list(var), instmap, instmapping, module_info).
+:- pred instmap__no_output_vars_2(list(prog_var), instmap, instmapping,
+		module_info).
 :- mode instmap__no_output_vars_2(in, in, in, in) is semidet.
 
 instmap__no_output_vars_2([], _, _, _).
@@ -707,8 +913,8 @@ merge_instmap_delta(InstMap, NonLocals, reachable(InstMappingA),
 	merge_instmapping_delta(InstMap, NonLocals, InstMappingA,
 		InstMappingB, InstMapping).
 
-:- pred merge_instmapping_delta(instmap, set(var), instmapping, instmapping,
-		instmapping, module_info, module_info).
+:- pred merge_instmapping_delta(instmap, set(prog_var), instmapping,
+		instmapping, instmapping, module_info, module_info).
 :- mode merge_instmapping_delta(in, in, in, in, out, in, out) is det.
 
 merge_instmapping_delta(InstMap, NonLocals, InstMappingA,
@@ -723,8 +929,9 @@ merge_instmapping_delta(InstMap, NonLocals, InstMappingA,
 	merge_instmapping_delta_2(ListofVars, InstMap, InstMappingA,
 		InstMappingB, InstMapping0, InstMapping).
 
-:- pred merge_instmapping_delta_2(list(var), instmap, instmapping, instmapping,
-			instmapping, instmapping, module_info, module_info).
+:- pred merge_instmapping_delta_2(list(prog_var), instmap, instmapping,
+		instmapping, instmapping, instmapping,
+		module_info, module_info).
 :- mode merge_instmapping_delta_2(in, in, in, in, in, out, in, out) is det.
 
 merge_instmapping_delta_2([], _, _, _, InstMapping, InstMapping,
@@ -748,11 +955,88 @@ merge_instmapping_delta_2([Var | Vars], InstMap, InstMappingA, InstMappingB,
 		ModuleInfo1 = ModuleInfoPrime,
 		map__det_insert(InstMapping0, Var, Inst, InstMapping1)
 	;
-		error("merge_instmapping_delta_2: unexpected mode error")
+		term__var_to_int(Var, VarInt),
+		string__format(
+			"merge_instmapping_delta_2: error merging var %i",
+			[i(VarInt)], Msg),
+		error(Msg)
 	),
 	merge_instmapping_delta_2(Vars, InstMap, InstMappingA, InstMappingB,
 		InstMapping1, InstMapping, ModuleInfo1, ModuleInfo).
 
+%-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
+
+	% Given two instmap deltas, unify them to produce a new instmap_delta.
+
+unify_instmap_delta(_, _, unreachable, InstMapDelta, InstMapDelta) --> [].
+unify_instmap_delta(_, _, reachable(InstMapping), unreachable,
+				reachable(InstMapping)) --> [].
+unify_instmap_delta(InstMap, NonLocals, reachable(InstMappingA),
+		reachable(InstMappingB), reachable(InstMapping)) -->
+	unify_instmapping_delta(InstMap, NonLocals, InstMappingA,
+		InstMappingB, InstMapping).
+
+:- pred unify_instmapping_delta(instmap, set(prog_var), instmapping,
+		instmapping, instmapping, module_info, module_info).
+:- mode unify_instmapping_delta(in, in, in, in, out, in, out) is det.
+
+unify_instmapping_delta(InstMap, NonLocals, InstMappingA,
+		InstMappingB, InstMapping) -->
+	{ map__keys(InstMappingA, VarsInA) },
+	{ map__keys(InstMappingB, VarsInB) },
+	{ set__sorted_list_to_set(VarsInA, SetofVarsInA) },
+	{ set__insert_list(SetofVarsInA, VarsInB, SetofVars0) },
+	{ set__intersect(SetofVars0, NonLocals, SetofVars) },
+	{ map__init(InstMapping0) },
+	{ set__to_sorted_list(SetofVars, ListofVars) },
+	unify_instmapping_delta_2(ListofVars, InstMap, InstMappingA,
+		InstMappingB, InstMapping0, InstMapping).
+
+:- pred unify_instmapping_delta_2(list(prog_var), instmap, instmapping,
+		instmapping, instmapping, instmapping,
+		module_info, module_info).
+:- mode unify_instmapping_delta_2(in, in, in, in, in, out, in, out) is det.
+
+unify_instmapping_delta_2([], _, _, _, InstMapping, InstMapping,
+		ModInfo, ModInfo).
+unify_instmapping_delta_2([Var | Vars], InstMap, InstMappingA, InstMappingB,
+			InstMapping0, InstMapping, ModuleInfo0, ModuleInfo) :-
+	( map__search(InstMappingA, Var, InstA) ->
+		( map__search(InstMappingB, Var, InstB) ->
+			(
+				% We can ignore the determinism of the
+				% unification: if it isn't det, then there
+				% will be a mode error or a determinism error
+				% in one of the parallel conjuncts.
+
+				abstractly_unify_inst(live, InstA, InstB,
+					fake_unify, ModuleInfo0, Inst, _Det,
+					ModuleInfoPrime)
+			->
+				ModuleInfo1 = ModuleInfoPrime,
+				map__det_insert(InstMapping0, Var, Inst,
+					InstMapping1)
+			;
+				error(
+				"unify_instmapping_delta_2: unexpected error")
+			)
+		;
+			ModuleInfo1 = ModuleInfo0,
+			map__det_insert(InstMapping0, Var, InstA, InstMapping1)
+		)
+	;
+		( map__search(InstMappingB, Var, InstB) ->
+			ModuleInfo1 = ModuleInfo0,
+			map__det_insert(InstMapping0, Var, InstB, InstMapping1)
+		;
+			ModuleInfo1 = ModuleInfo0,
+			InstMapping1 = InstMapping0
+		)
+	),
+	unify_instmapping_delta_2(Vars, InstMap, InstMappingA, InstMappingB,
+		InstMapping1, InstMapping, ModuleInfo1, ModuleInfo).
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
@@ -767,8 +1051,8 @@ instmap_delta_apply_sub(reachable(OldInstMapping), Must, Sub,
 instmap__apply_sub(InstMap0, Must, Sub, InstMap) :-
 	instmap_delta_apply_sub(InstMap0, Must, Sub, InstMap).
 
-:- pred instmap_delta_apply_sub_2(assoc_list(var, inst), bool, map(var, var),
-		instmapping, instmapping).
+:- pred instmap_delta_apply_sub_2(assoc_list(prog_var, inst), bool,
+		map(prog_var, prog_var), instmapping, instmapping).
 :- mode instmap_delta_apply_sub_2(in, in, in, in, out) is det.
 
 instmap_delta_apply_sub_2([], _Must, _Sub, IM, IM).
