@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -7,14 +7,6 @@
 % jumpopt.m - optimize jumps to jumps.
 %
 % Author: zs.
-%
-% We first build up a bunch of tables giving information about labels.
-% This information includes the first instruction following each label,
-% the block following each label, whether a label is part of a procedure
-% epilog of various types (det, semidet or nondet).
-%
-% We then traverse the instruction list, using the information in the
-% tables to short-circuit jumps.
 
 %-----------------------------------------------------------------------------%
 
@@ -22,19 +14,69 @@
 
 :- interface.
 
-:- import_module list, bool.
-:- import_module llds.
+:- import_module llds, globals.
+:- import_module list, set, bool.
 
-:- pred jumpopt__main(list(instruction), bool, bool, list(instruction), bool).
-:- mode jumpopt__main(in, in, in, out, out) is det.
+	% Take an instruction list and optimize jumps. This includes jumps
+	% implicit in procedure returns.
+	%
+	% The second argument gives the set of labels that have layout
+	% structures. This module will not optimize jumps to labels in this
+	% set, since this may interfere with the RTTI recorded for these
+	% labels. The third argument gives the trace level, which we also
+	% use to avoid optimizations that may interfere with RTTI.
+	%
+	% The three bool inputs should be
+	%
+	% - the value of the --optimize-fulljumps option,
+	%
+	% - an indication of whether this is the final application of this
+	%   optimization, and
+	%
+	% - the value of the --checked-nondet-tailcalls option respectively.
+	%
+	% The bool output says whether the instruction sequence was modified
+	% by the optimization.
+
+:- pred jumpopt_main(list(instruction)::in, set(label)::in, trace_level::in,
+	bool::in, bool::in, bool::in, list(instruction)::out, bool::out)
+	is det.
 
 %-----------------------------------------------------------------------------%
 
 :- implementation.
 
-:- import_module code_util, opt_util, std_util, map, string, require.
+:- import_module builtin_ops, code_util, opt_util.
+:- import_module int, std_util, map, string, require.
 
-jumpopt__main(Instrs0, Blockopt, Recjump, Instrs, Mod) :-
+% We first build up a bunch of tables giving information about labels.
+% We then traverse the instruction list, using the information in the
+% tables to short-circuit jumps.
+%
+% Instrmap:	Maps each label to the next real (non-comment, non-livevals)
+%		instruction after that label.
+% Lvalmap:	Maps each label to yes(Livevals) if the label is followed
+%		by a livevals instruction, and to no otherwise.
+% Blockmap:	Maps each label to the block following that label.
+%		This includes all instructions up to the first one that
+%		cannot fall through.
+% Procmap:	Maps each label that begins a det epilog to the epilog.
+% Succmap:	Maps each label that begins a nondet epilog to the epilog.
+% Sdprocmap:	Maps each label that begins a semidet epilog to the epilog.
+%		This can be the success epilog or the failure epilog.
+% Forkmap:	Maps each label that begins a full semidet epilog (code to
+%		test r1, and to execute the the success or failure epilog
+%		depending on the result) to the epilog.
+%
+% If we are not doing full jump optimization, Blockmap will be empty.
+% Even with full jump optimization, Blockmap will not contain the initial
+% block of the procedure unless Recjump is set. The intention is that
+% Recjump will not be set until optimizations such as frameopt and value
+% numbering, which can do a better job of optimizing this block, have
+% been applied.
+
+jumpopt_main(Instrs0, LayoutLabels, TraceLevel, Blockopt, Recjump,
+		MostlyDetTailCall, Instrs, Mod) :-
 	map__init(Instrmap0),
 	map__init(Lvalmap0),
 	map__init(Procmap0),
@@ -46,21 +88,27 @@ jumpopt__main(Instrs0, Blockopt, Recjump, Instrs, Mod) :-
 		Procmap0, Procmap, Sdprocmap0, Sdprocmap, Succmap0, Succmap),
 	map__init(Forkmap0),
 	jumpopt__build_forkmap(Instrs0, Sdprocmap, Forkmap0, Forkmap),
+	( MostlyDetTailCall = yes ->
+		opt_util__get_prologue(Instrs0, ProcLabel, _, _, _),
+		opt_util__new_label_no(Instrs0, 500, LabelNum),
+		CheckedNondetTailCallInfo = yes(ProcLabel - LabelNum)
+	;
+		CheckedNondetTailCallInfo = no
+	),
 	jumpopt__instr_list(Instrs0, comment(""), Instrmap, Blockmap, Lvalmap,
-		Procmap, Sdprocmap, Forkmap, Succmap, Instrs1, Mod),
-	opt_util__filter_out_bad_livevals(Instrs1, Instrs).
+		Procmap, Sdprocmap, Forkmap, Succmap, LayoutLabels,
+		TraceLevel, CheckedNondetTailCallInfo, _, Instrs1),
+	opt_util__filter_out_bad_livevals(Instrs1, Instrs),
+	( Instrs = Instrs0 ->
+		Mod = no
+	;
+		Mod = yes
+	).
 
 %-----------------------------------------------------------------------------%
 
-	% Build up three tables mapping labels to instruction sequences.
-	% A label has an entry in a table if it is followed by a deterministic,
-	% semideterministic or nondeterministic proceed/succeed; the map target
-	% gives the code sequence between the label and the proceed/succeed.
-	% We also build up a map giving the livevals instruction at the label
-	% if any, and the first real instruction at the label.
-
-:- pred jumpopt__build_maps(list(instruction), bool, bool, instrmap, instrmap,
-	tailmap, tailmap, lvalmap, lvalmap,
+:- pred jumpopt__build_maps(list(instruction), bool, bool,
+	instrmap, instrmap, tailmap, tailmap, lvalmap, lvalmap,
 	tailmap, tailmap, tailmap, tailmap, tailmap, tailmap).
 % :- mode jumpopt__build_maps(in, in, in, di, uo, di, uo, di, uo, di, uo,
 %	di, uo, di, uo) is det.
@@ -167,18 +215,24 @@ jumpopt__build_forkmap([Instr - _Comment|Instrs], Sdprocmap,
 	% between the if-val and the goto.
 
 :- pred jumpopt__instr_list(list(instruction), instr, instrmap, tailmap,
-	lvalmap, tailmap, tailmap, tailmap, tailmap, list(instruction), bool).
-:- mode jumpopt__instr_list(in, in, in, in, in, in, in, in, in, out, out)
-	is det.
+	lvalmap, tailmap, tailmap, tailmap, tailmap, set(label), trace_level,
+	maybe(pair(proc_label, int)), maybe(pair(proc_label, int)),
+	list(instruction)).
+:- mode jumpopt__instr_list(in, in, in, in, in, in, in, in, in, in, in,
+	in, out, out) is det.
 
-jumpopt__instr_list([], _PrevInstr, _Instrmap, _Blockmap,
-		_Lvalmap, _Procmap, _Sdprocmap, _Forkmap, _Succmap, [], no).
+jumpopt__instr_list([], _PrevInstr, _Instrmap, _Blockmap, _Lvalmap,
+		_Procmap, _Sdprocmap, _Forkmap, _Succmap, _LayoutLabels,
+		_, CheckedNondetTailCallInfo, CheckedNondetTailCallInfo, []).
 jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
-		Lvalmap, Procmap, Sdprocmap, Forkmap, Succmap, Instrs, Mod) :-
+		Lvalmap, Procmap, Sdprocmap, Forkmap, Succmap, LayoutLabels,
+		TraceLevel, CheckedNondetTailCallInfo0,
+		CheckedNondetTailCallInfo, Instrs) :-
 	Instr0 = Uinstr0 - Comment0,
 	string__append(Comment0, " (redirected return)", Redirect),
 	(
-		Uinstr0 = call(Proc, label(RetLabel), GC, CallModel)
+		Uinstr0 = call(Proc, label(RetLabel), LiveInfos, Context,
+			CallModel)
 	->
 		(
 			% Look for det style tailcalls. We look for this
@@ -187,29 +241,36 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 			% into a det epilog.
 			( CallModel = det ; CallModel = semidet ),
 			map__search(Procmap, RetLabel, Between0),
-			PrevInstr = livevals(Livevals) 
+			PrevInstr = livevals(Livevals),
+			TraceLevel = none,
+			not set__member(RetLabel, LayoutLabels)
 		->
 			opt_util__filter_out_livevals(Between0, Between1),
 			list__append(Between1, [livevals(Livevals) - "",
 				goto(Proc) - Redirect], NewInstrs),
 			RemainInstrs = Instrs0,
-			Mod0 = yes
+			CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
 		;
 			% Look for semidet style tailcalls.
 			CallModel = semidet,
 			map__search(Forkmap, RetLabel, Between),
-			PrevInstr = livevals(Livevals) 
+			PrevInstr = livevals(Livevals),
+			TraceLevel = none,
+			not set__member(RetLabel, LayoutLabels)
 		->
 			list__append(Between, [livevals(Livevals) - "",
 				goto(Proc) - Redirect], NewInstrs),
 			RemainInstrs = Instrs0,
-			Mod0 = yes
+			CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
 		;
-			% Look for nondet style tailcalls.
-			CallModel = nondet(yes),
+			% Look for nondet style tailcalls which do not need
+			% a runtime check.
+			CallModel = nondet(unchecked_tail_call),
 			map__search(Succmap, RetLabel, BetweenIncl),
 			BetweenIncl = [livevals(_) - _, goto(_) - _],
-			PrevInstr = livevals(Livevals) 
+			PrevInstr = livevals(Livevals),
+			TraceLevel = none,
+			not set__member(RetLabel, LayoutLabels)
 		->
 			NewInstrs = [
 				assign(maxfr, lval(prevfr(lval(curfr))))
@@ -222,27 +283,60 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 				goto(Proc) - Redirect
 			],
 			RemainInstrs = Instrs0,
-			Mod0 = yes
+			CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
+		;
+			% Look for nondet style tailcalls which do need
+			% a runtime check.
+			CallModel = nondet(checked_tail_call),
+			CheckedNondetTailCallInfo0 =
+				yes(ProcLabel - LabelNum0),
+			map__search(Succmap, RetLabel, BetweenIncl),
+			BetweenIncl = [livevals(_) - _, goto(_) - _],
+			PrevInstr = livevals(Livevals),
+			TraceLevel = none,
+			not set__member(RetLabel, LayoutLabels)
+		->
+			NewLabel = local(ProcLabel, LabelNum0),
+			NewInstrs = [
+				if_val(binop(ne, lval(curfr), lval(maxfr)),
+					label(NewLabel))
+					- "branch around if cannot tail call",
+				assign(maxfr, lval(prevfr(lval(curfr))))
+					- "discard this frame",
+				assign(succip, lval(succip(lval(curfr))))
+					- "setup PC on return from tailcall",
+				assign(curfr, lval(succfr(lval(curfr))))
+					- "setup curfr on return from tailcall",
+				livevals(Livevals) - "",
+				goto(Proc) - Redirect,
+				label(NewLabel) - "non tail call",
+				Instr0
+			],
+			RemainInstrs = Instrs0,
+			LabelNum1 is LabelNum0 + 1,
+			CheckedNondetTailCallInfo1 = yes(ProcLabel - LabelNum1)
 		;
 			% Short circuit the return label if possible.
-			map__search(Instrmap, RetLabel, RetInstr)
+			map__search(Instrmap, RetLabel, RetInstr),
+			TraceLevel = none,
+			not set__member(RetLabel, LayoutLabels)
 		->
 			jumpopt__final_dest(RetLabel, RetInstr, Instrmap,
 				DestLabel, _DestInstr),
 			( RetLabel = DestLabel ->
 				NewInstrs = [Instr0],
-				RemainInstrs = Instrs0,
-				Mod0 = no
+				RemainInstrs = Instrs0
 			;
 				NewInstrs = [call(Proc, label(DestLabel),
-					GC, CallModel) - Redirect],
-				RemainInstrs = Instrs0,
-				Mod0 = yes
-			)
+					LiveInfos, Context, CallModel)
+					- Redirect],
+				RemainInstrs = Instrs0
+			),
+			CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
 		;
 			NewInstrs = [Instr0],
 			RemainInstrs = Instrs0,
-			Mod0 = no
+			CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
 		)
 	;
 		Uinstr0 = goto(label(TargetLabel))
@@ -253,7 +347,7 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 		->
 			NewInstrs = [],
 			RemainInstrs = Instrs0,
-			Mod0 = yes
+			CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
 		;
 			PrevInstr = if_val(_, label(IfTargetLabel)),
 			opt_util__is_this_label_next(IfTargetLabel, Instrs0, _)
@@ -266,7 +360,7 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 			% that would require altering the if_val instruction.
 			NewInstrs = [Instr0],
 			RemainInstrs = Instrs0,
-			Mod0 = no
+			CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
 		;
 			% Replace a jump to a det epilog with the epilog.
 			map__search(Procmap, TargetLabel, Between0)
@@ -275,7 +369,7 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 			list__append(Between, [goto(succip) - "shortcircuit"],
 				NewInstrs),
 			RemainInstrs = Instrs0,
-			Mod0 = yes
+			CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
 		;
 			% Replace a jump to a semidet epilog with the epilog.
 			map__search(Sdprocmap, TargetLabel, Between0)
@@ -284,7 +378,7 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 			list__append(Between, [goto(succip) - "shortcircuit"],
 				NewInstrs),
 			RemainInstrs = Instrs0,
-			Mod0 = yes
+			CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
 		;
 			% Replace a jump to a nondet epilog with the epilog.
 			map__search(Succmap, TargetLabel, BetweenIncl0)
@@ -292,7 +386,7 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 			jumpopt__adjust_livevals(PrevInstr, BetweenIncl0,
 				NewInstrs),
 			RemainInstrs = Instrs0,
-			Mod0 = yes
+			CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
 		;
 			% Replace a jump to a non-epilog block with the
 			% block itself. These jumps are treated separately
@@ -324,9 +418,10 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 			map__delete(Blockmap, DestLabel, CrippledBlockmap),
 			jumpopt__instr_list(AdjustedBlock, comment(""),
 				Instrmap, CrippledBlockmap, Lvalmap, Procmap,
-				Sdprocmap, Forkmap, Succmap, NewInstrs, _),
-			RemainInstrs = Instrs0,
-			Mod0 = yes
+				Sdprocmap, Forkmap, Succmap, LayoutLabels,
+				TraceLevel, CheckedNondetTailCallInfo0,
+				CheckedNondetTailCallInfo1, NewInstrs),
+			RemainInstrs = Instrs0
 		;
 			% Short-circuit the goto.
 			map__search(Instrmap, TargetLabel, TargetInstr)
@@ -340,18 +435,15 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 				Canfallthrough),
 			( Canfallthrough = no ->
 				NewInstrs0 = [UdestInstr - Shorted],
-				RemainInstrs = Instrs0,
-				Mod0 = yes
+				RemainInstrs = Instrs0
 			;
 				( TargetLabel = DestLabel ->
 					NewInstrs0 = [Instr0],
-					RemainInstrs = Instrs0,
-					Mod0 = no
+					RemainInstrs = Instrs0
 				;
 					NewInstrs0 = [goto(label(DestLabel))
 						- Shorted],
-					RemainInstrs = Instrs0,
-					Mod0 = yes
+					RemainInstrs = Instrs0
 				)
 			),
 			( map__search(Lvalmap, DestLabel, yes(Lvalinstr)) ->
@@ -359,22 +451,24 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 					[Lvalinstr | NewInstrs0], NewInstrs)
 			;
 				NewInstrs = NewInstrs0
-			)
+			),
+			CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
 		;
 			NewInstrs = [Instr0],
 			RemainInstrs = Instrs0,
-			Mod0 = no
+			CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
 		)
 	; Uinstr0 = computed_goto(Index, LabelList0) ->
 		% Short-circuit all the destination labels.
-		jumpopt__short_labels(LabelList0, Instrmap, LabelList, Mod0),
+		jumpopt__short_labels(LabelList0, Instrmap, LabelList),
 		RemainInstrs = Instrs0,
-		( Mod0 = yes ->
+		CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0,
+		( LabelList = LabelList0 ->
+			NewInstrs = [Instr0]
+		;
 			string__append(Comment0, " (some shortcircuits)",
 				Shorted),
 			NewInstrs = [computed_goto(Index, LabelList) - Shorted]
-		;
-			NewInstrs = [Instr0]
 		)
 	; Uinstr0 = if_val(Cond, label(TargetLabel)) ->
 		(
@@ -424,29 +518,29 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 			% loop because each application of the transformation
 			% strictly reduces the size of the code.
 			RemainInstrs = [NewInstr | AfterGoto],
-			Mod0 = yes
+			CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
 		;
-			% Attempt to transform code such as
-			%
-			%	if (Cond) L2
-			%	r1 = TRUE
-			% 	<epilog>
-			%	...
-			% L2:
-			%	r1 = FALSE
-			%	<epilog>
-			%
-			% into
-			%
-			%	r1 = Cond
-			%	<epilog>
-			%
-
 			map__search(Instrmap, TargetLabel, TargetInstr)
 		->
 			jumpopt__final_dest(TargetLabel, TargetInstr,
 				Instrmap, DestLabel, _DestInstr),
 			(
+				% Attempt to transform code such as
+				%
+				%	if (Cond) L1
+				%	r1 = TRUE
+				% 	<epilog>
+				%	...
+				% L1:
+				%	r1 = FALSE
+				%	<epilog>
+				%
+				% into
+				%
+				%	r1 = Cond
+				%	<epilog>
+				%
+
 				opt_util__is_sdproceed_next(Instrs0, BetweenFT),
 				map__search(Blockmap, DestLabel, Block),
 				opt_util__is_sdproceed_next(Block, BetweenBR),
@@ -474,32 +568,43 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 				Proceed = goto(succip) - "shortcircuit",
 				list__append([NewAssign | Between], [Proceed],
 					NewInstrs),
-				RemainInstrs = Instrs0,
-				Mod0 = yes
+				RemainInstrs = Instrs0
 			;
-				% Short-circuit the destination.
+				% Try to short-circuit the destination.
+
 				TargetLabel \= DestLabel
 			->
 				string__append("shortcircuited jump: ",
 					Comment0, Shorted),
 				NewInstrs = [if_val(Cond, label(DestLabel))
 					- Shorted],
-				RemainInstrs = Instrs0,
-				Mod0 = yes
+				RemainInstrs = Instrs0
 			;
 				NewInstrs = [Instr0],
-				RemainInstrs = Instrs0,
-				Mod0 = no
-			)
+				RemainInstrs = Instrs0
+			),
+			CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
 		;
 			NewInstrs = [Instr0],
 			RemainInstrs = Instrs0,
-			Mod0 = no
+			CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
+		)
+	; Uinstr0 = assign(Lval, Rval0) ->
+		% Any labels mentioned in Rval0 should be short-circuited.
+		jumpopt__short_labels_rval(Rval0, Instrmap, Rval),
+		RemainInstrs = Instrs0,
+		CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0,
+		( Rval = Rval0 ->
+			NewInstrs = [Instr0]
+		;
+			string__append(Comment0, " (some shortcircuits)",
+				Shorted),
+			NewInstrs = [assign(Lval, Rval) - Shorted]
 		)
 	;
 		NewInstrs = [Instr0],
 		RemainInstrs = Instrs0,
-		Mod0 = no
+		CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
 	),
 	( ( Uinstr0 = comment(_) ; NewInstrs = [] ) ->
 		NewPrevInstr = PrevInstr
@@ -507,13 +612,10 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 		NewPrevInstr = Uinstr0
 	),
 	jumpopt__instr_list(RemainInstrs, NewPrevInstr, Instrmap, Blockmap,
-		Lvalmap, Procmap, Sdprocmap, Forkmap, Succmap, Instrs9, Mod1),
-	list__append(NewInstrs, Instrs9, Instrs),
-	( Mod0 = no, Mod1 = no ->
-		Mod = no
-	;
-		Mod = yes
-	).
+		Lvalmap, Procmap, Sdprocmap, Forkmap, Succmap, LayoutLabels,
+		TraceLevel, CheckedNondetTailCallInfo1,
+		CheckedNondetTailCallInfo, Instrs9),
+	list__append(NewInstrs, Instrs9, Instrs).
 
 % We avoid generating statements that redefine the value of a location
 % by comparing its old contents for non-equality with zero.
@@ -575,32 +677,25 @@ jumpopt__adjust_livevals(PrevInstr, Instrs0, Instrs) :-
 	% Short-circuit the given label by following any gotos at the
 	% labelled instruction or by falling through consecutive labels.
 
-:- pred jumpopt__short_label(label, instrmap, label, bool).
-:- mode jumpopt__short_label(in, in, out, out) is det.
+:- pred jumpopt__short_label(label, instrmap, label).
+:- mode jumpopt__short_label(in, in, out) is det.
 
-jumpopt__short_label(Label0, Instrmap, Label, Mod) :-
-	map__lookup(Instrmap, Label0, Instr0),
-	jumpopt__final_dest(Label0, Instr0, Instrmap, Label, _Instr),
-	( Label = Label0 ->
-		Mod = no
+jumpopt__short_label(Label0, Instrmap, Label) :-
+	( map__search(Instrmap, Label0, Instr0) ->
+		jumpopt__final_dest(Label0, Instr0, Instrmap, Label, _Instr)
 	;
-		Mod = yes
+		Label = Label0
 	).
 
-:- pred jumpopt__short_labels(list(label), instrmap, list(label), bool).
-:- mode jumpopt__short_labels(in, in, out, out) is det.
+:- pred jumpopt__short_labels(list(label), instrmap, list(label)).
+:- mode jumpopt__short_labels(in, in, out) is det.
 
 % XXX these uses of the Mod argument should be replaced by accumulator passing
 
-jumpopt__short_labels([], _Instrmap, [], no).
-jumpopt__short_labels([Label0 | Labels0], Instrmap, [Label | Labels], Mod) :-
-	jumpopt__short_label(Label0, Instrmap, Label, Mod1),
-	jumpopt__short_labels(Labels0, Instrmap, Labels, Mod2),
-	( Mod1 = no, Mod2 = no ->
-		Mod = no
-	;
-		Mod = yes
-	).
+jumpopt__short_labels([], _Instrmap, []).
+jumpopt__short_labels([Label0 | Labels0], Instrmap, [Label | Labels]) :-
+	jumpopt__short_label(Label0, Instrmap, Label),
+	jumpopt__short_labels(Labels0, Instrmap, Labels).
 
 %-----------------------------------------------------------------------------%
 
@@ -636,3 +731,101 @@ jumpopt__final_dest_2(SrcLabel, SrcInstr, Instrmap, LabelsSofar,
 		DestLabel = SrcLabel,
 		DestInstr = SrcInstr
 	).
+
+%-----------------------------------------------------------------------------%
+
+:- pred jumpopt__short_labels_rval(rval, instrmap, rval).
+:- mode jumpopt__short_labels_rval(in, in, out) is det.
+
+jumpopt__short_labels_rval(lval(Lval0), Instrmap, lval(Lval)) :-
+	jumpopt__short_labels_lval(Lval0, Instrmap, Lval).
+jumpopt__short_labels_rval(var(_), _, _) :-
+	error("var rval in jumpopt__short_labels_rval").
+jumpopt__short_labels_rval(create(Tag, Rvals0, ArgTypes, StatDyn, Cell, Type),
+		Instrmap, create(Tag, Rvals, ArgTypes, StatDyn, Cell, Type)) :-
+	jumpopt__short_labels_maybe_rvals(Rvals0, Instrmap, Rvals).
+jumpopt__short_labels_rval(mkword(Tag, Rval0), Instrmap,
+		mkword(Tag, Rval)) :-
+	jumpopt__short_labels_rval(Rval0, Instrmap, Rval).
+jumpopt__short_labels_rval(const(Const0), Instrmap, const(Const)) :-
+	jumpopt__short_labels_const(Const0, Instrmap, Const).
+jumpopt__short_labels_rval(unop(Op, Rval0), Instrmap, unop(Op, Rval)) :-
+	jumpopt__short_labels_rval(Rval0, Instrmap, Rval).
+jumpopt__short_labels_rval(binop(Op, LRval0, RRval0), Instrmap,
+		binop(Op, LRval, RRval)) :-
+	jumpopt__short_labels_rval(LRval0, Instrmap, LRval),
+	jumpopt__short_labels_rval(RRval0, Instrmap, RRval).
+jumpopt__short_labels_rval(mem_addr(MemRef), _, mem_addr(MemRef)).
+
+:- pred jumpopt__short_labels_const(rval_const, instrmap, rval_const).
+:- mode jumpopt__short_labels_const(in, in, out) is det.
+
+jumpopt__short_labels_const(true, _, true).
+jumpopt__short_labels_const(false, _, false).
+jumpopt__short_labels_const(int_const(I), _, int_const(I)).
+jumpopt__short_labels_const(float_const(F), _, float_const(F)).
+jumpopt__short_labels_const(string_const(S), _, string_const(S)).
+jumpopt__short_labels_const(multi_string_const(L, S), _,
+		multi_string_const(L, S)).
+jumpopt__short_labels_const(label_entry(Label0), Instrmap,
+		label_entry(Label)) :-
+	jumpopt__short_label(Label0, Instrmap, Label).
+jumpopt__short_labels_const(code_addr_const(CodeAddr0), Instrmap,
+		code_addr_const(CodeAddr)) :-
+	( CodeAddr0 = label(Label0) ->
+		jumpopt__short_label(Label0, Instrmap, Label),
+		CodeAddr = label(Label)
+	;
+		CodeAddr = CodeAddr0
+	).
+jumpopt__short_labels_const(data_addr_const(D), _, data_addr_const(D)).
+
+:- pred jumpopt__short_labels_maybe_rvals(list(maybe(rval)), instrmap,
+	list(maybe(rval))).
+:- mode jumpopt__short_labels_maybe_rvals(in, in, out) is det.
+
+jumpopt__short_labels_maybe_rvals([], _, []).
+jumpopt__short_labels_maybe_rvals([MaybeRval0 | MaybeRvals0], Instrmap,
+		[MaybeRval | MaybeRvals]) :-
+	(
+		MaybeRval0 = no,
+		MaybeRval = no
+	;
+		MaybeRval0 = yes(Rval0),
+		jumpopt__short_labels_rval(Rval0, Instrmap, Rval),
+		MaybeRval = yes(Rval)
+	),
+	jumpopt__short_labels_maybe_rvals(MaybeRvals0, Instrmap, MaybeRvals).
+
+:- pred jumpopt__short_labels_lval(lval, instrmap, lval).
+:- mode jumpopt__short_labels_lval(in, in, out) is det.
+
+jumpopt__short_labels_lval(reg(T, N), _, reg(T, N)).
+jumpopt__short_labels_lval(succip, _, succip).
+jumpopt__short_labels_lval(maxfr, _, maxfr).
+jumpopt__short_labels_lval(curfr, _, curfr).
+jumpopt__short_labels_lval(hp, _, hp).
+jumpopt__short_labels_lval(sp, _, sp).
+jumpopt__short_labels_lval(temp(T, N), _, temp(T, N)).
+jumpopt__short_labels_lval(stackvar(N), _, stackvar(N)).
+jumpopt__short_labels_lval(framevar(N), _, framevar(N)).
+jumpopt__short_labels_lval(succip(Rval0), Instrmap, succip(Rval)) :-
+	jumpopt__short_labels_rval(Rval0, Instrmap, Rval).
+jumpopt__short_labels_lval(redoip(Rval0), Instrmap, redoip(Rval)) :-
+	jumpopt__short_labels_rval(Rval0, Instrmap, Rval).
+jumpopt__short_labels_lval(redofr(Rval0), Instrmap, redofr(Rval)) :-
+	jumpopt__short_labels_rval(Rval0, Instrmap, Rval).
+jumpopt__short_labels_lval(succfr(Rval0), Instrmap, succfr(Rval)) :-
+	jumpopt__short_labels_rval(Rval0, Instrmap, Rval).
+jumpopt__short_labels_lval(prevfr(Rval0), Instrmap, prevfr(Rval)) :-
+	jumpopt__short_labels_rval(Rval0, Instrmap, Rval).
+jumpopt__short_labels_lval(field(Tag, Rval0, Field0), Instrmap,
+		field(Tag, Rval, Field)) :-
+	jumpopt__short_labels_rval(Rval0, Instrmap, Rval),
+	jumpopt__short_labels_rval(Field0, Instrmap, Field).
+jumpopt__short_labels_lval(mem_ref(Rval0), Instrmap, mem_ref(Rval)) :-
+	jumpopt__short_labels_rval(Rval0, Instrmap, Rval).
+jumpopt__short_labels_lval(lvar(_), _, _) :-
+	error("lvar lval in jumpopt__short_labels_lval").
+
+%-----------------------------------------------------------------------------%
