@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1997 The University of Melbourne.
+% Copyright (C) 1996-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -28,7 +28,7 @@
 :- implementation.
 
 :- import_module hlds_goal, passes_aux, hlds_out.
-:- import_module list, std_util.
+:- import_module list, require, std_util.
 
 %-----------------------------------------------------------------------------%
 
@@ -63,6 +63,10 @@ lco_in_goal_2(conj(Goals0), ModuleInfo, conj(Goals)) :-
 	list__reverse(Goals0, RevGoals0),
 	lco_in_conj(RevGoals0, [], ModuleInfo, Goals).
 
+	% XXX Some execution algorithm issues here.
+lco_in_goal_2(par_conj(_Goals0, SM), _ModuleInfo, par_conj(_Goals, SM)) :-
+	error("sorry: lco of parallel conjunction not implemented").
+
 lco_in_goal_2(disj(Goals0, SM), ModuleInfo, disj(Goals, SM)) :-
 	lco_in_disj(Goals0, ModuleInfo, Goals).
 
@@ -75,20 +79,24 @@ lco_in_goal_2(if_then_else(Vars, Cond, Then0, Else0, SM), ModuleInfo,
 	lco_in_goal(Then0, ModuleInfo, Then),
 	lco_in_goal(Else0, ModuleInfo, Else).
 
-lco_in_goal_2(some(Vars, Goal0), ModuleInfo, some(Vars, Goal)) :-
+lco_in_goal_2(some(Vars, CanRemove, Goal0), ModuleInfo,
+		some(Vars, CanRemove, Goal)) :-
 	lco_in_goal(Goal0, ModuleInfo, Goal).
 
 lco_in_goal_2(not(Goal), _ModuleInfo, not(Goal)).
 
-lco_in_goal_2(higher_order_call(A,B,C,D,E,F), _ModuleInfo,
-		higher_order_call(A,B,C,D,E,F)).
+lco_in_goal_2(generic_call(A,B,C,D), _ModuleInfo, generic_call(A,B,C,D)).
 
 lco_in_goal_2(call(A,B,C,D,E,F), _ModuleInfo, call(A,B,C,D,E,F)).
 
 lco_in_goal_2(unify(A,B,C,D,E), _ModuleInfo, unify(A,B,C,D,E)).
 
-lco_in_goal_2(pragma_c_code(A,B,C,D,E,F,G,H), _,
-		pragma_c_code(A,B,C,D,E,F,G,H)).
+lco_in_goal_2(pragma_c_code(A,B,C,D,E,F,G), _,
+		pragma_c_code(A,B,C,D,E,F,G)).
+
+lco_in_goal_2(bi_implication(_, _), _, _) :-
+	% these should have been expanded out by now
+	error("lco_in_goal_2: unexpected bi_implication").
 
 %-----------------------------------------------------------------------------%
 
@@ -140,7 +148,7 @@ lco_in_conj([Goal0 | Goals0], Unifies0, ModuleInfo, Goals) :-
 	Goal0 = GoalExpr0 - _,
 	(
 		GoalExpr0 = unify(_, _, _, Unif, _),
-		Unif = construct(_, _, _, _)
+		Unif = construct(_, _, _, _, _, _, _)
 	->
 		Unifies1 = [Goal0 | Unifies0],
 		lco_in_conj(Goals0, Unifies1, ModuleInfo, Goals)
