@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -16,7 +16,7 @@
 
 :- interface.
 
-:- import_module list, set, map.
+:- import_module list, set, map, std_util.
 :- import_module llds.
 
 :- type livemap		==	map(label, lvalset).
@@ -33,7 +33,7 @@
 :- implementation.
 
 :- import_module opt_util.
-:- import_module std_util, require, string, bool.
+:- import_module require, string, bool.
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
@@ -60,9 +60,9 @@ livemap__build(Instrs, MaybeLivemap) :-
 
 livemap__build_2(Backinstrs, Livemap0, MaybeLivemap) :-
 	set__init(Livevals0),
-	livemap__build_livemap(Backinstrs, Livevals0, no, Ccode1,
+	livemap__build_livemap(Backinstrs, Livevals0, no, DontValueNumber1,
 		Livemap0, Livemap1),
-	( Ccode1 = yes ->
+	( DontValueNumber1 = yes ->
 		MaybeLivemap = no
 	; livemap__equal_livemaps(Livemap0, Livemap1) ->
 		MaybeLivemap = yes(Livemap1)
@@ -105,13 +105,15 @@ livemap__equal_livemaps_keys([Label | Labels], Livemap1, Livemap2) :-
 	livemap, livemap).
 :- mode livemap__build_livemap(in, in, in, out, in, out) is det.
 
-livemap__build_livemap([], _, Ccode, Ccode, Livemap, Livemap).
-livemap__build_livemap([Instr0 | Instrs0], Livevals0, Ccode0, Ccode,
-		Livemap0, Livemap) :-
+livemap__build_livemap([], _, DontValueNumber, DontValueNumber,
+		Livemap, Livemap).
+livemap__build_livemap([Instr0 | Instrs0], Livevals0,
+		DontValueNumber0, DontValueNumber, Livemap0, Livemap) :-
 	livemap__build_livemap_instr(Instr0, Instrs0, Instrs1,
-		Livevals0, Livevals1, Ccode0, Ccode1, Livemap0, Livemap1),
+		Livevals0, Livevals1, DontValueNumber0, DontValueNumber1,
+		Livemap0, Livemap1),
 	livemap__build_livemap(Instrs1, Livevals1,
-		Ccode1, Ccode, Livemap1, Livemap).
+		DontValueNumber1, DontValueNumber, Livemap1, Livemap).
 
 :- pred livemap__build_livemap_instr(instruction, list(instruction),
 	list(instruction), lvalset, lvalset, bool, bool, livemap, livemap).
@@ -119,14 +121,15 @@ livemap__build_livemap([Instr0 | Instrs0], Livevals0, Ccode0, Ccode,
 	is det.
 
 livemap__build_livemap_instr(Instr0, Instrs0, Instrs,
-		Livevals0, Livevals, Ccode0, Ccode, Livemap0, Livemap) :-
+		Livevals0, Livevals, DontValueNumber0, DontValueNumber,
+		Livemap0, Livemap) :-
 	Instr0 = Uinstr0 - _,
 	(
 		Uinstr0 = comment(_),
 		Livemap = Livemap0,
 		Livevals = Livevals0,
 		Instrs = Instrs0,
-		Ccode = Ccode0
+		DontValueNumber = DontValueNumber0
 	;
 		Uinstr0 = livevals(_),
 		error("livevals found in backward scan in build_livemap")
@@ -145,34 +148,29 @@ livemap__build_livemap_instr(Instr0, Instrs0, Instrs,
 
 		set__delete(Livevals0, Lval, Livevals1),
 		opt_util__lval_access_rvals(Lval, Rvals),
-		livemap__make_live_in_rvals([Rval | Rvals], Livevals1, Livevals),
+		livemap__make_live_in_rvals([Rval | Rvals], Livevals1,
+			Livevals),
 		Livemap = Livemap0,
 		Instrs = Instrs0,
-		Ccode = Ccode0
+		DontValueNumber = DontValueNumber0
 	;
-		Uinstr0 = call(_, _, _, _),
+		Uinstr0 = call(_, _, _, _, _),
 		livemap__look_for_livevals(Instrs0, Instrs,
 			Livevals0, Livevals, "call", yes, _),
 		Livemap = Livemap0,
-		Ccode = Ccode0
+		DontValueNumber = DontValueNumber0
 	;
-		Uinstr0 = mkframe(_, _, _),
+		Uinstr0 = mkframe(_, _),
 		Livemap = Livemap0,
 		Livevals = Livevals0,
 		Instrs = Instrs0,
-		Ccode = Ccode0
-	;
-		Uinstr0 = modframe(_),
-		Livemap = Livemap0,
-		Livevals = Livevals0,
-		Instrs = Instrs0,
-		Ccode = Ccode0
+		DontValueNumber = DontValueNumber0
 	;
 		Uinstr0 = label(Label),
 		map__set(Livemap0, Label, Livevals0, Livemap),
 		Livevals = Livevals0,
 		Instrs = Instrs0,
-		Ccode = Ccode0
+		DontValueNumber = DontValueNumber0
 	;
 		Uinstr0 = goto(CodeAddr),
 		opt_util__livevals_addr(CodeAddr, LivevalsNeeded),
@@ -201,7 +199,7 @@ livemap__build_livemap_instr(Instr0, Instrs0, Instrs,
 			Livevals = Livevals3
 		),
 		Livemap = Livemap0,
-		Ccode = Ccode0
+		DontValueNumber = DontValueNumber0
 	;
 		Uinstr0 = computed_goto(Rval, Labels),
 		set__init(Livevals1),
@@ -210,13 +208,13 @@ livemap__build_livemap_instr(Instr0, Instrs0, Instrs,
 			Livevals2, Livevals),
 		Livemap = Livemap0,
 		Instrs = Instrs0,
-		Ccode = Ccode0
+		DontValueNumber = DontValueNumber0
 	;
 		Uinstr0 = c_code(_),
 		Livemap = Livemap0,
 		Livevals = Livevals0,
 		Instrs = Instrs0,
-		Ccode = yes
+		DontValueNumber = yes
 	;
 		Uinstr0 = if_val(Rval, CodeAddr),
 		livemap__look_for_livevals(Instrs0, Instrs,
@@ -224,10 +222,16 @@ livemap__build_livemap_instr(Instr0, Instrs0, Instrs,
 		(
 			Found = yes,
 			% This if_val was put here by middle_rec.
-			Livevals3 = Livevals1
+			% We must make sure that the locations mentioned
+			% in the livevals annotation become live,
+			% since they will be needed at CodeAddr.
+			% The locations in Livevals0 may be needed
+			% in the fall-through continuation.
+			set__union(Livevals0, Livevals1, Livevals3)
 		;
 			Found = no,
-			livemap__make_live_in_rvals([Rval], Livevals1, Livevals2),
+			livemap__make_live_in_rvals([Rval],
+				Livevals1, Livevals2),
 			( CodeAddr = label(Label) ->
 				livemap__insert_label_livevals([Label],
 					Livemap0, Livevals2, Livevals3)
@@ -242,9 +246,9 @@ livemap__build_livemap_instr(Instr0, Instrs0, Instrs,
 			Livevals = Livevals3
 		),
 		Livemap = Livemap0,
-		Ccode = Ccode0
+		DontValueNumber = DontValueNumber0
 	;
-		Uinstr0 = incr_hp(Lval, _Tag, Rval),
+		Uinstr0 = incr_hp(Lval, _, Rval, _),
 
 		% Make dead the variable assigned, but make any variables
 		% needed to access it live. Make the variables in the size
@@ -255,10 +259,11 @@ livemap__build_livemap_instr(Instr0, Instrs0, Instrs,
 
 		set__delete(Livevals0, Lval, Livevals1),
 		opt_util__lval_access_rvals(Lval, Rvals),
-		livemap__make_live_in_rvals([Rval | Rvals], Livevals1, Livevals),
+		livemap__make_live_in_rvals([Rval | Rvals],
+			Livevals1, Livevals),
 		Livemap = Livemap0,
 		Instrs = Instrs0,
-		Ccode = Ccode0
+		DontValueNumber = DontValueNumber0
 	;
 		Uinstr0 = mark_hp(Lval),
 		set__delete(Livevals0, Lval, Livevals1),
@@ -266,13 +271,13 @@ livemap__build_livemap_instr(Instr0, Instrs0, Instrs,
 		livemap__make_live_in_rvals(Rvals, Livevals1, Livevals),
 		Livemap = Livemap0,
 		Instrs = Instrs0,
-		Ccode = Ccode0
+		DontValueNumber = DontValueNumber0
 	;
 		Uinstr0 = restore_hp(Rval),
 		livemap__make_live_in_rvals([Rval], Livevals0, Livevals),
 		Livemap = Livemap0,
 		Instrs = Instrs0,
-		Ccode = Ccode0
+		DontValueNumber = DontValueNumber0
 	;
 		Uinstr0 = store_ticket(Lval),
 		set__delete(Livevals0, Lval, Livevals1),
@@ -280,19 +285,19 @@ livemap__build_livemap_instr(Instr0, Instrs0, Instrs,
 		livemap__make_live_in_rvals(Rvals, Livevals1, Livevals),
 		Livemap = Livemap0,
 		Instrs = Instrs0,
-		Ccode = Ccode0
+		DontValueNumber = DontValueNumber0
 	;
 		Uinstr0 = reset_ticket(Rval, _Reason),
 		livemap__make_live_in_rval(Rval, Livevals0, Livevals),
 		Livemap = Livemap0,
 		Instrs = Instrs0,
-		Ccode = Ccode0
+		DontValueNumber = DontValueNumber0
 	;
 		Uinstr0 = discard_ticket,
 		Livevals = Livevals0,
 		Livemap = Livemap0,
 		Instrs = Instrs0,
-		Ccode = Ccode0
+		DontValueNumber = DontValueNumber0
 	;
 		Uinstr0 = mark_ticket_stack(Lval),
 		set__delete(Livevals0, Lval, Livevals1),
@@ -300,32 +305,62 @@ livemap__build_livemap_instr(Instr0, Instrs0, Instrs,
 		livemap__make_live_in_rvals(Rvals, Livevals1, Livevals),
 		Livemap = Livemap0,
 		Instrs = Instrs0,
-		Ccode = Ccode0
+		DontValueNumber = DontValueNumber0
 	;
 		Uinstr0 = discard_tickets_to(Rval),
 		livemap__make_live_in_rval(Rval, Livevals0, Livevals),
 		Livemap = Livemap0,
 		Instrs = Instrs0,
-		Ccode = Ccode0
+		DontValueNumber = DontValueNumber0
 	;
 		Uinstr0 = incr_sp(_, _),
-		Livevals = Livevals0,
 		Livemap = Livemap0,
+		Livevals = Livevals0,
 		Instrs = Instrs0,
-		Ccode = Ccode0
+		DontValueNumber = DontValueNumber0
 	;
 		Uinstr0 = decr_sp(_),
-		Livevals = Livevals0,
 		Livemap = Livemap0,
+		Livevals = Livevals0,
 		Instrs = Instrs0,
-		Ccode = Ccode0
+		DontValueNumber = DontValueNumber0
+	;
+		Uinstr0 = init_sync_term(_, _),
+		Livemap = Livemap0,
+		Livevals = Livevals0,
+		Instrs = Instrs0,
+		DontValueNumber = DontValueNumber0
+	;
+		% XXX Value numbering doesn't handle fork [yet] so
+		% set DontValueNumber to yes.
+		Uinstr0 = fork(_, _, _),
+		Livemap = Livemap0,
+		Livevals = Livevals0,
+		Instrs = Instrs0,
+		DontValueNumber = yes
+	;
+		% XXX Value numbering doesn't handle join_and_terminate [yet] so
+		% set DontValueNumber to yes.
+		Uinstr0 = join_and_terminate(_),
+		Livemap = Livemap0,
+		Livevals = Livevals0,
+		Instrs = Instrs0,
+		DontValueNumber = yes
+	;
+		% XXX Value numbering doesn't handle join_and_continue [yet] so
+		% set DontValueNumber to yes.
+		Uinstr0 = join_and_continue(_, _),
+		Livemap = Livemap0,
+		Livevals = Livevals0,
+		Instrs = Instrs0,
+		DontValueNumber = yes
 	;
 		% XXX we shouldn't just give up here
-		Uinstr0 = pragma_c(_, _, _, _, _),
+		Uinstr0 = pragma_c(_, _, _, _, _, _),
 		Livemap = Livemap0,
 		Livevals = Livevals0,
 		Instrs = Instrs0,
-		Ccode = yes
+		DontValueNumber = yes
 	).
 
 :- pred livemap__look_for_livevals(list(instruction), list(instruction),
@@ -358,10 +393,19 @@ livemap__special_code_addr(imported(_), no).
 livemap__special_code_addr(succip, yes(succip)).
 livemap__special_code_addr(do_succeed(_), yes(succip(lval(curfr)))).
 livemap__special_code_addr(do_redo, yes(redoip(lval(maxfr)))).
+livemap__special_code_addr(do_trace_redo_fail_shallow, no).
+livemap__special_code_addr(do_trace_redo_fail_deep, no).
 livemap__special_code_addr(do_fail, no).
-livemap__special_code_addr(do_det_closure, no).
-livemap__special_code_addr(do_semidet_closure, no).
-livemap__special_code_addr(do_nondet_closure, no).
+livemap__special_code_addr(do_call_closure, no).
+livemap__special_code_addr(do_call_class_method, no).
+livemap__special_code_addr(do_det_aditi_call, no).
+livemap__special_code_addr(do_semidet_aditi_call, no).
+livemap__special_code_addr(do_nondet_aditi_call, no).
+livemap__special_code_addr(do_aditi_insert, no).
+livemap__special_code_addr(do_aditi_delete, no).
+livemap__special_code_addr(do_aditi_bulk_insert, no).
+livemap__special_code_addr(do_aditi_bulk_delete, no).
+livemap__special_code_addr(do_aditi_modify, no).
 livemap__special_code_addr(do_not_reached, no).
 
 %-----------------------------------------------------------------------------%
@@ -391,7 +435,7 @@ livemap__make_live_in_rval(lval(Lval), Live0, Live) :-
 	),
 	opt_util__lval_access_rvals(Lval, AccessRvals),
 	livemap__make_live_in_rvals(AccessRvals, Live1, Live).
-livemap__make_live_in_rval(create(_, _, _, _), Live, Live).
+livemap__make_live_in_rval(create(_, _, _, _, _, _), Live, Live).
 	% All terms inside creates in the optimizer must be static.
 livemap__make_live_in_rval(mkword(_, Rval), Live0, Live) :-
 	livemap__make_live_in_rval(Rval, Live0, Live).
