@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -44,7 +44,7 @@
 % occurrence of a variable include that variable in their post-death
 % set. In branched structures, branches in which a variable is not
 % used at all include a pre-death set listing the variables that
-% have died in parallel branches. 
+% have died in parallel branches.
 %
 % The third pass, detect_resume_points_in_goal, finds goals that
 % establish resume points and attaches to them a resume_point
@@ -52,15 +52,13 @@
 % code at that resume point as well as the nature of the required
 % entry labels.
 %
-% Accurate garbage collection notes:
+% Typeinfo liveness calculation notes:
 %
-% When using accurate gc, liveness is computed slightly differently.
-% The garbage collector needs access to the typeinfo variables of any
-% variable that could be live at a garbage collection point. In the
-% present design of the garbage collector, garbage collection takes place
-% at procedure returns.
+% When using accurate gc or execution tracing, liveness is computed
+% slightly differently.  The runtime system needs access to the
+% typeinfo variables of any variable that is live at a continuation.
 % 
-% Hence, the invariant needed for accurate GC is:
+% Hence, the invariant needed for alternate liveness calculation:
 % 	a variable holding a typeinfo must be live at any continuation
 % 	where any variable whose type is described (in whole or in part)
 % 	by that typeinfo is live.
@@ -85,7 +83,7 @@
 % 
 % (1) happens without any changes to the liveness computation (it is
 %     the normal condition for variables becoming dead). This more
-%     conservative than what is required for accurate GC, but is
+%     conservative than what is required for the invariant, but is
 %     required for code generation, so we should keep it ;-)
 % (2) is implemented by adding the typeinfo variables for the types of the
 %     nonlocals to the nonlocals for the purposes of computing liveness.
@@ -125,19 +123,20 @@
 
 :- interface.
 
-:- import_module hlds_module, hlds_pred.
+:- import_module hlds_module, hlds_pred, prog_data.
+:- import_module set.
 
 	% Add liveness annotations to the goal of the procedure.
 	% This consists of the {pre,post}{birth,death} sets and
 	% resume point information.
 
-:- pred detect_liveness_proc(proc_info, module_info, proc_info).
-:- mode detect_liveness_proc(in, in, out) is det.
+:- pred detect_liveness_proc(proc_info, pred_id, module_info, proc_info).
+:- mode detect_liveness_proc(in, in, in, out) is det.
 
 	% Return the set of variables live at the start of the procedure.
 
-:- pred initial_liveness(proc_info, module_info, set(var)).
-:- mode initial_liveness(in, in, out) is det.
+:- pred initial_liveness(proc_info, pred_id, module_info, set(prog_var)).
+:- mode initial_liveness(in, in, in, out) is det.
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
@@ -145,26 +144,48 @@
 :- implementation.
 
 :- import_module hlds_goal, hlds_data, llds, quantification, (inst), instmap.
-:- import_module hlds_out, mode_util, code_util, quantification.
-:- import_module prog_data, globals, passes_aux.
-:- import_module bool, list, map, set, std_util, term, assoc_list, require.
-:- import_module varset, string.
+:- import_module hlds_out, mode_util, code_util, quantification, options.
+:- import_module trace, globals, polymorphism, passes_aux.
+:- import_module term, varset.
 
-detect_liveness_proc(ProcInfo0, ModuleInfo, ProcInfo) :-
+:- import_module bool, map, std_util, list, assoc_list, require.
+:- import_module string.
+
+detect_liveness_proc(ProcInfo0, PredId, ModuleInfo, ProcInfo) :-
 	requantify_proc(ProcInfo0, ProcInfo1),
 	proc_info_goal(ProcInfo1, Goal0),
-	proc_info_variables(ProcInfo1, Varset),
+	proc_info_varset(ProcInfo1, Varset),
 	proc_info_vartypes(ProcInfo1, VarTypes),
-	live_info_init(ModuleInfo, ProcInfo1, VarTypes, Varset, LiveInfo),
+	module_info_globals(ModuleInfo, Globals),
 
-	initial_liveness(ProcInfo1, ModuleInfo, Liveness0),
+	module_info_pred_info(ModuleInfo, PredId, PredInfo),
+	pred_info_module(PredInfo, PredModule),
+	pred_info_name(PredInfo, PredName),
+	pred_info_arity(PredInfo, PredArity),
+	(
+		polymorphism__no_type_info_builtin(PredModule,
+			PredName, PredArity)
+	->
+		TypeInfoLiveness = no
+	;
+		body_should_use_typeinfo_liveness(Globals, TypeInfoLiveness)
+	),
+	live_info_init(ModuleInfo, ProcInfo1, TypeInfoLiveness,
+		VarTypes, Varset, LiveInfo),
+
+	initial_liveness(ProcInfo1, PredId, ModuleInfo, Liveness0),
 	detect_liveness_in_goal(Goal0, Liveness0, LiveInfo,
 		_, Goal1),
 
-	initial_deadness(ProcInfo1, ModuleInfo, Deadness0),
+	initial_deadness(ProcInfo1, LiveInfo, ModuleInfo, Deadness0),
 	detect_deadness_in_goal(Goal1, Deadness0, LiveInfo, _, Goal2),
 
-	set__init(ResumeVars0),
+	globals__get_trace_level(Globals, TraceLevel),
+	( TraceLevel \= none ->
+		trace__fail_vars(ModuleInfo, ProcInfo0, ResumeVars0)
+	;
+		set__init(ResumeVars0)
+	),
 	detect_resume_points_in_goal(Goal2, Liveness0, LiveInfo,
 		ResumeVars0, Goal, _),
 
@@ -174,8 +195,8 @@ detect_liveness_proc(ProcInfo0, ModuleInfo, ProcInfo) :-
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
-:- pred detect_liveness_in_goal(hlds_goal, set(var), live_info,
-	set(var), hlds_goal).
+:- pred detect_liveness_in_goal(hlds_goal, set(prog_var), live_info,
+		set(prog_var), hlds_goal).
 :- mode detect_liveness_in_goal(in, in, in, out, out) is det.
 
 detect_liveness_in_goal(Goal0 - GoalInfo0, Liveness0, LiveInfo,
@@ -225,14 +246,21 @@ detect_liveness_in_goal(Goal0 - GoalInfo0, Liveness0, LiveInfo,
 
 	% Here we process each of the different sorts of goals.
 
-:- pred detect_liveness_in_goal_2(hlds_goal_expr, set(var), set(var),
-	live_info, set(var), hlds_goal_expr).
+:- pred detect_liveness_in_goal_2(hlds_goal_expr, set(prog_var), set(prog_var),
+	live_info, set(prog_var), hlds_goal_expr).
 :- mode detect_liveness_in_goal_2(in, in, in, in, out, out) is det.
 
 detect_liveness_in_goal_2(conj(Goals0), Liveness0, _, LiveInfo,
 		Liveness, conj(Goals)) :-
 	detect_liveness_in_conj(Goals0, Liveness0, LiveInfo, Liveness, Goals).
 
+detect_liveness_in_goal_2(par_conj(Goals0, SM), Liveness0, NonLocals, LiveInfo,
+		Liveness, par_conj(Goals, SM)) :-
+	set__init(Union0),
+	detect_liveness_in_par_conj(Goals0, Liveness0, NonLocals, LiveInfo,
+		Union0, Union, Goals),
+	set__union(Liveness0, Union, Liveness).
+
 detect_liveness_in_goal_2(disj(Goals0, SM), Liveness0, NonLocals, LiveInfo,
 		Liveness, disj(Goals, SM)) :-
 	set__init(Union0),
@@ -255,10 +283,9 @@ detect_liveness_in_goal_2(if_then_else(Vars, Cond0, Then0, Else0, SM),
 	detect_liveness_in_goal(Cond0, Liveness0, LiveInfo, LivenessCond, Cond),
 
 	%
-	% If the condition cannot succeed, any variables which become
-	% live in the else part should put in the post-birth set of the
-	% then part by add_liveness_after_goal, and the other sets
-	% should be empty.
+	% If the condition cannot succeed, any variables which become live
+	% in the else part should be put in the post-birth set of the then part
+	% by add_liveness_after_goal, and the other sets should be empty.
 	%
 	Cond = _ - CondInfo,
 	goal_info_get_instmap_delta(CondInfo, CondDelta),
@@ -282,11 +309,11 @@ detect_liveness_in_goal_2(if_then_else(Vars, Cond0, Then0, Else0, SM),
 	add_liveness_after_goal(Then1, ResidueThen, Then),
 	add_liveness_after_goal(Else1, ResidueElse, Else).
 
-detect_liveness_in_goal_2(some(Vars, Goal0), Liveness0, _, LiveInfo,
-		Liveness, some(Vars, Goal)) :-
+detect_liveness_in_goal_2(some(Vars, CanRemove, Goal0), Liveness0, _, LiveInfo,
+		Liveness, some(Vars, CanRemove, Goal)) :-
 	detect_liveness_in_goal(Goal0, Liveness0, LiveInfo, Liveness, Goal).
 
-detect_liveness_in_goal_2(higher_order_call(_,_,_,_,_,_), _, _, _, _, _) :-
+detect_liveness_in_goal_2(generic_call(_,_,_,_), _, _, _, _, _) :-
 	error("higher-order-call in detect_liveness_in_goal_2").
 
 detect_liveness_in_goal_2(call(_,_,_,_,_,_), _, _, _, _, _) :-
@@ -295,13 +322,16 @@ detect_liveness_in_goal_2(call(_,_,_,_,_,_), _, _, _, _, _) :-
 detect_liveness_in_goal_2(unify(_,_,_,_,_), _, _, _, _, _) :-
 	error("unify in detect_liveness_in_goal_2").
 
-detect_liveness_in_goal_2(pragma_c_code(_,_,_,_,_,_,_,_), _, _, _, _, _) :-
+detect_liveness_in_goal_2(pragma_c_code(_,_,_,_,_,_,_), _, _, _, _, _) :-
 	error("pragma_c_code in detect_liveness_in_goal_2").
 
+detect_liveness_in_goal_2(bi_implication(_, _), _, _, _, _, _) :-
+	error("bi_implication in detect_liveness_in_goal_2").
+
 %-----------------------------------------------------------------------------%
 
-:- pred detect_liveness_in_conj(list(hlds_goal), set(var), live_info,
-	set(var), list(hlds_goal)).
+:- pred detect_liveness_in_conj(list(hlds_goal), set(prog_var), live_info,
+	set(prog_var), list(hlds_goal)).
 :- mode detect_liveness_in_conj(in, in, in, out, out) is det.
 
 detect_liveness_in_conj([], Liveness, _LiveInfo, Liveness, []).
@@ -322,8 +352,8 @@ detect_liveness_in_conj([Goal0 | Goals0], Liveness0, LiveInfo, Liveness,
 
 %-----------------------------------------------------------------------------%
 
-:- pred detect_liveness_in_disj(list(hlds_goal), set(var), set(var),
-	live_info, set(var), set(var), list(hlds_goal)).
+:- pred detect_liveness_in_disj(list(hlds_goal), set(prog_var), set(prog_var),
+		live_info, set(prog_var), set(prog_var), list(hlds_goal)).
 :- mode detect_liveness_in_disj(in, in, in, in, in, out, out) is det.
 
 detect_liveness_in_disj([], _Liveness, _NonLocals, _LiveInfo,
@@ -340,8 +370,8 @@ detect_liveness_in_disj([Goal0 | Goals0], Liveness, NonLocals, LiveInfo,
 
 %-----------------------------------------------------------------------------%
 
-:- pred detect_liveness_in_cases(list(case), set(var), set(var),
-	live_info, set(var), set(var), list(case)).
+:- pred detect_liveness_in_cases(list(case), set(prog_var), set(prog_var),
+		live_info, set(prog_var), set(prog_var), list(case)).
 :- mode detect_liveness_in_cases(in, in, in, in, in, out, out) is det.
 
 detect_liveness_in_cases([], _Liveness, _NonLocals, _LiveInfo,
@@ -356,11 +386,30 @@ detect_liveness_in_cases([case(Cons, Goal0) | Goals0], Liveness, NonLocals,
 	set__difference(NonLocalUnion, Liveness1, Residue),
 	add_liveness_after_goal(Goal1, Residue, Goal).
 
+%-----------------------------------------------------------------------------%
+
+:- pred detect_liveness_in_par_conj(list(hlds_goal), set(prog_var),
+		set(prog_var), live_info, set(prog_var), set(prog_var),
+		list(hlds_goal)).
+:- mode detect_liveness_in_par_conj(in, in, in, in, in, out, out) is det.
+
+detect_liveness_in_par_conj([], _Liveness, _NonLocals, _LiveInfo,
+		Union, Union, []).
+detect_liveness_in_par_conj([Goal0 | Goals0], Liveness0, NonLocals, LiveInfo,
+		Union0, Union, [Goal | Goals]) :-
+	detect_liveness_in_goal(Goal0, Liveness0, LiveInfo, Liveness1, Goal1),
+	set__union(Union0, Liveness1, Union1),
+	detect_liveness_in_par_conj(Goals0, Liveness0, NonLocals, LiveInfo,
+		Union1, Union, Goals),
+	set__intersect(Union, NonLocals, NonLocalUnion),
+	set__difference(NonLocalUnion, Liveness1, Residue),
+	add_liveness_after_goal(Goal1, Residue, Goal).
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
-:- pred detect_deadness_in_goal(hlds_goal, set(var), live_info,
-	set(var), hlds_goal).
+:- pred detect_deadness_in_goal(hlds_goal, set(prog_var), live_info,
+	set(prog_var), hlds_goal).
 :- mode detect_deadness_in_goal(in, in, in, out, out) is det.
 
 detect_deadness_in_goal(Goal0 - GoalInfo0, Deadness0, LiveInfo, Deadness,
@@ -412,7 +461,7 @@ detect_deadness_in_goal(Goal0 - GoalInfo0, Deadness0, LiveInfo, Deadness,
 	% Here we process each of the different sorts of goals.
 
 :- pred detect_deadness_in_goal_2(hlds_goal_expr, hlds_goal_info,
-	set(var), live_info, set(var), hlds_goal_expr).
+	set(prog_var), live_info, set(prog_var), hlds_goal_expr).
 :- mode detect_deadness_in_goal_2(in, in, in, in, out, out) is det.
 
 detect_deadness_in_goal_2(conj(Goals0), _, Deadness0, LiveInfo,
@@ -420,6 +469,14 @@ detect_deadness_in_goal_2(conj(Goals0), _, Deadness0, LiveInfo,
 	detect_deadness_in_conj(Goals0, Deadness0, LiveInfo,
 		Goals, Deadness).
 
+detect_deadness_in_goal_2(par_conj(Goals0, SM), GoalInfo, Deadness0, LiveInfo,
+		Deadness, par_conj(Goals, SM)) :-
+	set__init(Union0),
+	goal_info_get_nonlocals(GoalInfo, NonLocals),
+	detect_deadness_in_par_conj(Goals0, Deadness0, NonLocals,
+		LiveInfo, Union0, Union, Goals),
+	set__union(Union, Deadness0, Deadness).
+
 detect_deadness_in_goal_2(disj(Goals0, SM), GoalInfo, Deadness0,
 		LiveInfo, Deadness, disj(Goals, SM)) :-
 	set__init(Union0),
@@ -458,11 +515,11 @@ detect_deadness_in_goal_2(if_then_else(Vars, Cond0, Then0, Else0, SM),
 	add_deadness_before_goal(Cond1, ResidueCond, Cond),
 	add_deadness_before_goal(Else1, ResidueElse, Else).
 
-detect_deadness_in_goal_2(some(Vars, Goal0), _, Deadness0, LiveInfo,
-		Deadness, some(Vars, Goal)) :-
+detect_deadness_in_goal_2(some(Vars, CanRemove, Goal0), _, Deadness0, LiveInfo,
+		Deadness, some(Vars, CanRemove, Goal)) :-
 	detect_deadness_in_goal(Goal0, Deadness0, LiveInfo, Deadness, Goal).
 
-detect_deadness_in_goal_2(higher_order_call(_,_,_,_,_,_), _, _, _, _, _) :-
+detect_deadness_in_goal_2(generic_call(_,_,_,_), _, _, _, _, _) :-
 	error("higher-order-call in detect_deadness_in_goal_2").
 
 detect_deadness_in_goal_2(call(_,_,_,_,_,_), _, _, _, _, _) :-
@@ -471,13 +528,16 @@ detect_deadness_in_goal_2(call(_,_,_,_,_,_), _, _, _, _, _) :-
 detect_deadness_in_goal_2(unify(_,_,_,_,_), _, _, _, _, _) :-
 	error("unify in detect_deadness_in_goal_2").
 
-detect_deadness_in_goal_2(pragma_c_code(_,_,_,_,_,_,_,_), _, _, _, _, _) :-
+detect_deadness_in_goal_2(pragma_c_code(_,_,_,_,_,_,_), _, _, _, _, _) :-
 	error("pragma_c_code in detect_deadness_in_goal_2").
 
+detect_deadness_in_goal_2(bi_implication(_, _), _, _, _, _, _) :-
+	error("bi_implication in detect_deadness_in_goal_2").
+
 %-----------------------------------------------------------------------------%
 
-:- pred detect_deadness_in_conj(list(hlds_goal), set(var), live_info,
-	list(hlds_goal), set(var)).
+:- pred detect_deadness_in_conj(list(hlds_goal), set(prog_var), live_info,
+	list(hlds_goal), set(prog_var)).
 :- mode detect_deadness_in_conj(in, in, in, out, out) is det.
 
 detect_deadness_in_conj([], Deadness, _LiveInfo, [], Deadness).
@@ -500,8 +560,8 @@ detect_deadness_in_conj([Goal0 | Goals0], Deadness0, LiveInfo,
 
 %-----------------------------------------------------------------------------%
 
-:- pred detect_deadness_in_disj(list(hlds_goal), set(var), set(var),
-	live_info, set(var), set(var), list(hlds_goal)).
+:- pred detect_deadness_in_disj(list(hlds_goal), set(prog_var), set(prog_var),
+	live_info, set(prog_var), set(prog_var), list(hlds_goal)).
 :- mode detect_deadness_in_disj(in, in, in, in, in, out, out) is det.
 
 detect_deadness_in_disj([], _Deadness, _NonLocals, _LiveInfo,
@@ -518,8 +578,8 @@ detect_deadness_in_disj([Goal0 | Goals0], Deadness, NonLocals, LiveInfo,
 
 %-----------------------------------------------------------------------------%
 
-:- pred detect_deadness_in_cases(var, list(case), set(var), set(var),
-	live_info, set(var), set(var), list(case)).
+:- pred detect_deadness_in_cases(prog_var, list(case), set(prog_var),
+	set(prog_var), live_info, set(prog_var), set(prog_var), list(case)).
 :- mode detect_deadness_in_cases(in, in, in, in, in, in, out, out) is det.
 
 detect_deadness_in_cases(_Var, [], _Deadness, _NonLocals, _LiveInfo,
@@ -538,11 +598,30 @@ detect_deadness_in_cases(SwitchVar, [case(Cons, Goal0) | Goals0], Deadness0,
 	set__difference(NonLocalUnion, Deadness1, Residue),
 	add_deadness_before_goal(Goal1, Residue, Goal).
 
+%-----------------------------------------------------------------------------%
+
+:- pred detect_deadness_in_par_conj(list(hlds_goal), set(prog_var),
+		set(prog_var), live_info, set(prog_var), set(prog_var),
+		list(hlds_goal)).
+:- mode detect_deadness_in_par_conj(in, in, in, in, in, out, out) is det.
+
+detect_deadness_in_par_conj([], _Deadness, _NonLocals, _LiveInfo,
+		Union, Union, []).
+detect_deadness_in_par_conj([Goal0 | Goals0], Deadness, NonLocals, LiveInfo,
+		Union0, Union, [Goal | Goals]) :-
+	detect_deadness_in_goal(Goal0, Deadness, LiveInfo, Deadness1, Goal1),
+	set__union(Union0, Deadness1, Union1),
+	detect_deadness_in_par_conj(Goals0, Deadness, NonLocals, LiveInfo,
+		Union1, Union, Goals),
+	set__intersect(Union, NonLocals, NonLocalUnion),
+	set__difference(NonLocalUnion, Deadness1, Residue),
+	add_deadness_before_goal(Goal1, Residue, Goal).
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
-:- pred detect_resume_points_in_goal(hlds_goal, set(var), live_info, set(var),
-	hlds_goal, set(var)).
+:- pred detect_resume_points_in_goal(hlds_goal, set(prog_var), live_info,
+		set(prog_var), hlds_goal, set(prog_var)).
 :- mode detect_resume_points_in_goal(in, in, in, in, out, out) is det.
 
 detect_resume_points_in_goal(Goal0 - GoalInfo0, Liveness0, LiveInfo,
@@ -562,7 +641,7 @@ detect_resume_points_in_goal(Goal0 - GoalInfo0, Liveness0, LiveInfo,
 	set__union(Liveness4, PostBirths0, Liveness).
 
 :- pred detect_resume_points_in_goal_2(hlds_goal_expr, hlds_goal_info,
-	set(var), live_info, set(var), hlds_goal_expr, set(var)).
+	set(prog_var), live_info, set(prog_var), hlds_goal_expr, set(prog_var)).
 :- mode detect_resume_points_in_goal_2(in, in, in, in, in, out, out) is det.
 
 detect_resume_points_in_goal_2(conj(Goals0), _, Liveness0, LiveInfo,
@@ -570,6 +649,11 @@ detect_resume_points_in_goal_2(conj(Goals0), _, Liveness0, LiveInfo,
 	detect_resume_points_in_conj(Goals0, Liveness0, LiveInfo, ResumeVars0,
 		Goals, Liveness).
 
+detect_resume_points_in_goal_2(par_conj(Goals0, SM), _, Liveness0, LiveInfo,
+		ResumeVars0, par_conj(Goals, SM), Liveness) :-
+	detect_resume_points_in_par_conj(Goals0, Liveness0, LiveInfo,
+		ResumeVars0, Goals, Liveness).
+
 detect_resume_points_in_goal_2(disj(Goals0, SM), GoalInfo, Liveness0, LiveInfo,
 		ResumeVars0, disj(Goals, SM), Liveness) :-
 	goal_info_get_code_model(GoalInfo, CodeModel),
@@ -628,8 +712,9 @@ detect_resume_points_in_goal_2(if_then_else(Vars, Cond0, Then0, Else0, SM),
 
 	require_equal(LivenessThen, LivenessElse, "if-then-else", LiveInfo).
 
-detect_resume_points_in_goal_2(some(Vars, Goal0), _, Liveness0, LiveInfo,
-		ResumeVars0, some(Vars, Goal), Liveness) :-
+detect_resume_points_in_goal_2(some(Vars, CanRemove, Goal0), _, Liveness0,
+		LiveInfo, ResumeVars0, some(Vars, CanRemove, Goal),
+		Liveness) :-
 	detect_resume_points_in_goal(Goal0, Liveness0, LiveInfo, ResumeVars0,
 					Goal, Liveness).
 
@@ -657,8 +742,8 @@ detect_resume_points_in_goal_2(not(Goal0), _, Liveness0, LiveInfo, ResumeVars0,
 	Resume = resume_point(ResumeVars1, ResumeLocs),
 	goal_set_resume_point(Goal1, Resume, Goal).
 
-detect_resume_points_in_goal_2(higher_order_call(A,B,C,D,E,F), _, Liveness,
-		_, _, higher_order_call(A,B,C,D,E,F), Liveness).
+detect_resume_points_in_goal_2(generic_call(A,B,C,D), _, Liveness,
+		_, _, generic_call(A,B,C,D), Liveness).
 
 detect_resume_points_in_goal_2(call(A,B,C,D,E,F), _, Liveness, _, _,
 		call(A,B,C,D,E,F), Liveness).
@@ -666,12 +751,16 @@ detect_resume_points_in_goal_2(call(A,B,C,D,E,F), _, Liveness, _, _,
 detect_resume_points_in_goal_2(unify(A,B,C,D,E), _, Liveness, _, _,
 		unify(A,B,C,D,E), Liveness).
 
-detect_resume_points_in_goal_2(pragma_c_code(A,B,C,D,E,F,G,H), _, Liveness,
+detect_resume_points_in_goal_2(pragma_c_code(A,B,C,D,E,F,G), _, Liveness,
 		_, _,
-		pragma_c_code(A,B,C,D,E,F,G,H), Liveness).
+		pragma_c_code(A,B,C,D,E,F,G), Liveness).
 
-:- pred detect_resume_points_in_conj(list(hlds_goal), set(var), live_info,
-	set(var), list(hlds_goal), set(var)).
+detect_resume_points_in_goal_2(bi_implication(_, _), _, _, _, _, _, _) :-
+	% these should have been expanded out by now
+	error("detect_resume_points_in_goal_2: unexpected bi_implication").
+
+:- pred detect_resume_points_in_conj(list(hlds_goal), set(prog_var), live_info,
+	set(prog_var), list(hlds_goal), set(prog_var)).
 :- mode detect_resume_points_in_conj(in, in, in, in, out, out) is det.
 
 detect_resume_points_in_conj([], Liveness, _, _, [], Liveness).
@@ -689,8 +778,9 @@ detect_resume_points_in_conj([Goal0 | Goals0], Liveness0, LiveInfo,
 	% Second, an empty pruned disjunction is legal, while an empty
 	% nondet disjunction isn't.
 
-:- pred detect_resume_points_in_non_disj(list(hlds_goal), set(var),
-	live_info, set(var), list(hlds_goal), set(var), set(var)).
+:- pred detect_resume_points_in_non_disj(list(hlds_goal), set(prog_var),
+		live_info, set(prog_var), list(hlds_goal),
+		set(prog_var), set(prog_var)).
 :- mode detect_resume_points_in_non_disj(in, in, in, in, out, out, out) is det.
 
 detect_resume_points_in_non_disj([], _, _, _, _, _, _) :-
@@ -713,8 +803,9 @@ detect_resume_points_in_non_disj([Goal0 | Goals0], Liveness0, LiveInfo,
 		Goals = Goals0
 	).
 
-:- pred detect_resume_points_in_pruned_disj(list(hlds_goal), set(var),
-	live_info, set(var), list(hlds_goal), set(var), set(var)).
+:- pred detect_resume_points_in_pruned_disj(list(hlds_goal), set(prog_var),
+		live_info, set(prog_var), list(hlds_goal), set(prog_var),
+		set(prog_var)).
 :- mode detect_resume_points_in_pruned_disj(in, in, in, in, out, out, out)
 	is det.
 
@@ -746,8 +837,8 @@ detect_resume_points_in_pruned_disj([Goal0 | Goals0], Liveness0, LiveInfo,
 	).
 
 :- pred detect_resume_points_in_non_last_disjunct(hlds_goal, bool,
-	set(var), set(var), live_info, set(var), hlds_goal,
-	set(var), set(var), set(var)).
+	set(prog_var), set(prog_var), live_info, set(prog_var), hlds_goal,
+	set(prog_var), set(prog_var), set(prog_var)).
 :- mode detect_resume_points_in_non_last_disjunct(in, in, in, in, in, in,
 	out, out, in, out) is det.
 
@@ -789,8 +880,8 @@ detect_resume_points_in_non_last_disjunct(Goal0, MayUseOrigOnly,
 
 	require_equal(Liveness, LivenessRest, "disjunction", LiveInfo).
 
-:- pred detect_resume_points_in_last_disjunct(hlds_goal, set(var), live_info,
-	set(var), hlds_goal, set(var), set(var)).
+:- pred detect_resume_points_in_last_disjunct(hlds_goal, set(prog_var),
+	live_info, set(prog_var), hlds_goal, set(prog_var), set(prog_var)).
 :- mode detect_resume_points_in_last_disjunct(in, in, in, in, out, out, out)
 	is det.
 
@@ -802,8 +893,8 @@ detect_resume_points_in_last_disjunct(Goal0, Liveness0, LiveInfo,
 	goal_info_get_pre_deaths(GoalInfo, PreDeaths),
 	set__difference(Liveness0, PreDeaths, Needed).
 
-:- pred detect_resume_points_in_cases(list(case), set(var), live_info,
-	set(var), list(case), set(var)).
+:- pred detect_resume_points_in_cases(list(case), set(prog_var), live_info,
+		set(prog_var), list(case), set(prog_var)).
 :- mode detect_resume_points_in_cases(in, in, in, in, out, out) is det.
 
 detect_resume_points_in_cases([], Liveness, _, _, [], Liveness).
@@ -821,7 +912,19 @@ detect_resume_points_in_cases([case(ConsId, Goal0) | Cases0], Liveness0,
 		Cases = Cases0
 	).
 
-:- pred require_equal(set(var), set(var), string, live_info).
+:- pred detect_resume_points_in_par_conj(list(hlds_goal), set(prog_var),
+		live_info, set(prog_var), list(hlds_goal), set(prog_var)).
+:- mode detect_resume_points_in_par_conj(in, in, in, in, out, out) is det.
+
+detect_resume_points_in_par_conj([], Liveness, _, _, [], Liveness).
+detect_resume_points_in_par_conj([Goal0 | Goals0], Liveness0, LiveInfo,
+		ResumeVars0, [Goal | Goals], LivenessFirst) :-
+	detect_resume_points_in_goal(Goal0, Liveness0, LiveInfo, ResumeVars0,
+		Goal, LivenessFirst),
+	detect_resume_points_in_par_conj(Goals0, Liveness0, LiveInfo,
+		ResumeVars0, Goals, _LivenessRest).
+
+:- pred require_equal(set(prog_var), set(prog_var), string, live_info).
 :- mode require_equal(in, in, in, in) is det.
 
 require_equal(LivenessFirst, LivenessRest, GoalType, LiveInfo) :-
@@ -852,7 +955,7 @@ require_equal(LivenessFirst, LivenessRest, GoalType, LiveInfo) :-
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
-initial_liveness(ProcInfo, ModuleInfo, Liveness) :-
+initial_liveness(ProcInfo, PredId, ModuleInfo, Liveness) :-
 	proc_info_headvars(ProcInfo, Vars),
 	proc_info_argmodes(ProcInfo, Modes),
 	proc_info_vartypes(ProcInfo, VarTypes),
@@ -866,19 +969,26 @@ initial_liveness(ProcInfo, ModuleInfo, Liveness) :-
 	;
 		error("initial_liveness: list length mismatch")
 	),
-	module_info_globals(ModuleInfo, Globals),
-	globals__get_gc_method(Globals, GCmethod),
 
 		% If a variable is unused in the goal, it shouldn't be
 		% in the initial liveness. (If we allowed it to start
 		% live, it wouldn't ever become dead, because it would
 		% have to be used to be killed).
 		% So we intersect the headvars with the non-locals and
-		% their typeinfo vars.
+		% (if doing typeinfo liveness calculation) their
+		% typeinfo vars.
+	module_info_globals(ModuleInfo, Globals),
 	proc_info_goal(ProcInfo, _Goal - GoalInfo),
 	goal_info_get_nonlocals(GoalInfo, NonLocals0),
-	(
-		GCmethod = accurate
+	module_info_pred_info(ModuleInfo, PredId, PredInfo),
+	body_should_use_typeinfo_liveness(Globals, TypeinfoLiveness),
+	pred_info_module(PredInfo, PredModule),
+	pred_info_name(PredInfo, PredName),
+	pred_info_arity(PredInfo, PredArity),
+	( 	
+		TypeinfoLiveness = yes,
+		\+ polymorphism__no_type_info_builtin(PredModule,
+			PredName, PredArity)
 	->
 		proc_info_get_typeinfo_vars_setwise(ProcInfo, NonLocals0,
 			TypeInfoNonLocals),
@@ -888,9 +998,8 @@ initial_liveness(ProcInfo, ModuleInfo, Liveness) :-
 	),
 	set__intersect(Liveness2, NonLocals, Liveness).
 
-
-:- pred initial_liveness_2(list(var), list(mode), list(type), module_info,
-	set(var), set(var)).
+:- pred initial_liveness_2(list(prog_var), list(mode), list(type), module_info,
+		set(prog_var), set(prog_var)).
 :- mode initial_liveness_2(in, in, in, in, in, out) is semidet.
 
 initial_liveness_2([], [], [], _ModuleInfo, Liveness, Liveness).
@@ -907,15 +1016,16 @@ initial_liveness_2([V | Vs], [M | Ms], [T | Ts], ModuleInfo,
 
 %-----------------------------------------------------------------------------%
 
-:- pred initial_deadness(proc_info, module_info, set(var)).
-:- mode initial_deadness(in, in, out) is det.
+:- pred initial_deadness(proc_info, live_info, module_info, set(prog_var)).
+:- mode initial_deadness(in, in, in, out) is det.
 
-initial_deadness(ProcInfo, ModuleInfo, Deadness) :-
+initial_deadness(ProcInfo, LiveInfo, ModuleInfo, Deadness) :-
 	proc_info_headvars(ProcInfo, Vars),
 	proc_info_argmodes(ProcInfo, Modes),
 	proc_info_vartypes(ProcInfo, VarTypes),
 	map__apply_to_list(Vars, VarTypes, Types),
 	set__init(Deadness0),
+		% All output arguments are in the initial deadness.
 	(
 		initial_deadness_2(Vars, Modes, Types, ModuleInfo,
 			Deadness0, Deadness1)
@@ -924,12 +1034,12 @@ initial_deadness(ProcInfo, ModuleInfo, Deadness) :-
 	;
 		error("initial_deadness: list length mis-match")
 	),
-		% If doing accurate garbage collection, the corresponding
+
+		% If doing alternate liveness, the corresponding
 		% typeinfos need to be added to these.
-	module_info_globals(ModuleInfo, Globals),
-	globals__get_gc_method(Globals, GCmethod),
-	(
-		GCmethod = accurate
+	live_info_get_typeinfo_liveness(LiveInfo, TypeinfoLiveness),
+	( 
+		TypeinfoLiveness = yes
 	->
 		proc_info_get_typeinfo_vars_setwise(ProcInfo, Deadness2,
 			TypeInfoVars),
@@ -938,8 +1048,8 @@ initial_deadness(ProcInfo, ModuleInfo, Deadness) :-
 		Deadness = Deadness2
 	).
 
-:- pred initial_deadness_2(list(var), list(mode), list(type),
-				module_info, set(var), set(var)).
+:- pred initial_deadness_2(list(prog_var), list(mode), list(type),
+		module_info, set(prog_var), set(prog_var)).
 :- mode initial_deadness_2(in, in, in, in, in, out) is semidet.
 
 initial_deadness_2([], [], [], _ModuleInfo, Deadness, Deadness).
@@ -957,7 +1067,7 @@ initial_deadness_2([V | Vs], [M | Ms], [T | Ts], ModuleInfo,
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
-:- pred add_liveness_after_goal(hlds_goal, set(var), hlds_goal).
+:- pred add_liveness_after_goal(hlds_goal, set(prog_var), hlds_goal).
 :- mode add_liveness_after_goal(in, in, out) is det.
 
 add_liveness_after_goal(Goal - GoalInfo0, Residue, Goal - GoalInfo) :-
@@ -965,7 +1075,7 @@ add_liveness_after_goal(Goal - GoalInfo0, Residue, Goal - GoalInfo) :-
 	set__union(PostBirths0, Residue, PostBirths),
 	goal_info_set_post_births(GoalInfo0, PostBirths, GoalInfo).
 
-:- pred add_deadness_before_goal(hlds_goal, set(var), hlds_goal).
+:- pred add_deadness_before_goal(hlds_goal, set(prog_var), hlds_goal).
 :- mode add_deadness_before_goal(in, in, out) is det.
 
 add_deadness_before_goal(Goal - GoalInfo0, Residue, Goal - GoalInfo) :-
@@ -984,8 +1094,8 @@ add_deadness_before_goal(Goal - GoalInfo0, Residue, Goal - GoalInfo) :-
 	%
 	% We don't handle the aliasing part yet.
 
-:- pred find_value_giving_occurrences(list(var), live_info,
-	instmap_delta, set(var), set(var)).
+:- pred find_value_giving_occurrences(list(prog_var), live_info,
+	instmap_delta, set(prog_var), set(prog_var)).
 :- mode find_value_giving_occurrences(in, in, in, in, out) is det.
 
 find_value_giving_occurrences([], _, _, ValueVars, ValueVars).
@@ -1011,53 +1121,61 @@ find_value_giving_occurrences([Var | Vars], LiveInfo, InstMapDelta,
 :- type live_info	--->	live_info(
 					module_info,
 					proc_info,
-					map(var, type),
-					varset
+					bool,		% Do we use typeinfo
+							% liveness for this
+							% proc?
+					map(prog_var, type),
+					prog_varset
 				).
 
-:- pred live_info_init(module_info, proc_info, map(var, type),
-	varset, live_info).
-:- mode live_info_init(in, in, in, in, out) is det.
+:- pred live_info_init(module_info, proc_info, bool, map(prog_var, type),
+	prog_varset, live_info).
+:- mode live_info_init(in, in, in, in, in, out) is det.
 
-live_info_init(ModuleInfo, ProcInfo, VarTypes, Varset,
-	live_info(ModuleInfo, ProcInfo, VarTypes, Varset)).
+live_info_init(ModuleInfo, ProcInfo, TypeInfoLiveness, VarTypes, Varset,
+	live_info(ModuleInfo, ProcInfo, TypeInfoLiveness, VarTypes, Varset)).
 
 :- pred live_info_get_module_info(live_info, module_info).
 :- mode live_info_get_module_info(in, out) is det.
 
-live_info_get_module_info(live_info(ModuleInfo, _, _, _), ModuleInfo).
+live_info_get_module_info(live_info(ModuleInfo, _, _, _, _), ModuleInfo).
 
 :- pred live_info_get_proc_info(live_info, proc_info).
 :- mode live_info_get_proc_info(in, out) is det.
 
-live_info_get_proc_info(live_info(_, ProcInfo, _, _), ProcInfo).
+live_info_get_proc_info(live_info(_, ProcInfo, _, _, _), ProcInfo).
 
-:- pred live_info_get_var_types(live_info, map(var, type)).
+:- pred live_info_get_typeinfo_liveness(live_info, bool).
+:- mode live_info_get_typeinfo_liveness(in, out) is det.
+
+live_info_get_typeinfo_liveness(live_info(_, _, TypeInfoLiveness, _, _),
+	TypeInfoLiveness).
+
+:- pred live_info_get_var_types(live_info, map(prog_var, type)).
 :- mode live_info_get_var_types(in, out) is det.
 
-live_info_get_var_types(live_info(_, _, VarTypes, _), VarTypes).
+live_info_get_var_types(live_info(_, _, _, VarTypes, _), VarTypes).
 
-:- pred live_info_get_varset(live_info, varset).
+:- pred live_info_get_varset(live_info, prog_varset).
 :- mode live_info_get_varset(in, out) is det.
 
-live_info_get_varset(live_info(_, _, _, Varset), Varset).
+live_info_get_varset(live_info(_, _, _, _, Varset), Varset).
 
 %-----------------------------------------------------------------------------%
 
-	% Get the nonlocals, and, if doing accurate GC, add the
+	% Get the nonlocals, and, if doing alternate liveness, add the
 	% typeinfo vars for the nonlocals.
 
 :- pred liveness__get_nonlocals_and_typeinfos(live_info, hlds_goal_info,
-		set(var)).
+		set(prog_var)).
 :- mode liveness__get_nonlocals_and_typeinfos(in, in, out) is det.
+
 liveness__get_nonlocals_and_typeinfos(LiveInfo, GoalInfo, 
 		NonLocals) :-
-	live_info_get_module_info(LiveInfo, ModuleInfo),
-	module_info_globals(ModuleInfo, Globals),
-	globals__get_gc_method(Globals, GCmethod),
 	goal_info_get_nonlocals(GoalInfo, NonLocals0),
-	(
-		GCmethod = accurate
+	live_info_get_typeinfo_liveness(LiveInfo, TypeinfoLiveness),
+	( 
+		TypeinfoLiveness = yes
 	->
 		live_info_get_proc_info(LiveInfo, ProcInfo),
 		proc_info_get_typeinfo_vars_setwise(ProcInfo, NonLocals0,
