@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1997 The University of Melbourne.
+% Copyright (C) 1996-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -41,36 +41,39 @@
 
 :- interface.
 
-:- import_module hlds_goal, hlds_data, llds, switch_gen, code_info.
-:- import_module list, term.
+:- import_module hlds_goal, hlds_data, llds, switch_gen, code_info, prog_data.
+:- import_module std_util, map, set, list.
 
 :- type case_consts == list(pair(int, list(rval))).
 
-:- type rval_map == map(var, list(pair(int, rval))).
+:- type rval_map == map(prog_var, list(pair(int, rval))).
 
-:- pred lookup_switch__is_lookup_switch(var, cases_list, hlds_goal_info,
+:- pred lookup_switch__is_lookup_switch(prog_var, cases_list, hlds_goal_info,
 		can_fail, int, code_model, int, int, can_fail, can_fail,
-		list(var), case_consts, maybe(set(var)),
+		list(prog_var), case_consts, maybe(set(prog_var)),
 		code_info, code_info).
 :- mode lookup_switch__is_lookup_switch(in, in, in, in, in, in, out, out,
 		out, out, out, out, out, in, out) is semidet.
 
 	% Generate code for a switch using a lookup table.
 
-:- pred lookup_switch__generate(var, list(var), case_consts,
-		int, int, can_fail, can_fail, maybe(set(var)), store_map,
-		code_tree, code_info, code_info).
+:- pred lookup_switch__generate(prog_var, list(prog_var), case_consts,
+		int, int, can_fail, can_fail, maybe(set(prog_var)),
+		store_map, branch_end, branch_end, code_tree,
+		code_info, code_info).
 :- mode lookup_switch__generate(in, in, in, in, in, in, in, in, in,
-		out, in, out) is det.
+		in, out, out, in, out) is det.
 
 %-----------------------------------------------------------------------------%
 
 :- implementation.
 
-:- import_module set, code_gen, type_util, map, tree, int, std_util, require.
-:- import_module dense_switch, bool, assoc_list, globals, options, mode_util.
+:- import_module builtin_ops, code_gen, type_util, tree.
+:- import_module dense_switch, globals, options, mode_util.
 :- import_module exprn_aux, getopt, prog_data, instmap.
 
+:- import_module int, require, bool, assoc_list.
+
 	% Most of this predicate is taken from dense_switch.m
 
 	% We need the code_info structure to generate code for the cases to
@@ -152,7 +155,7 @@ lookup_switch__is_lookup_switch(CaseVar, TaggedCases, GoalInfo,
 
 %---------------------------------------------------------------------------%
 
-:- pred lookup_switch__figure_out_output_vars(hlds_goal_info, list(var),
+:- pred lookup_switch__figure_out_output_vars(hlds_goal_info, list(prog_var),
 		code_info, code_info).
 :- mode lookup_switch__figure_out_output_vars(in, out, in, out) is det.
 
@@ -186,8 +189,8 @@ lookup_switch__figure_out_output_vars(GoalInfo, OutVars) -->
 
 %---------------------------------------------------------------------------%
 
-:- pred lookup_switch__generate_constants(cases_list, list(var),
-		code_model, case_consts, maybe(set(var)),
+:- pred lookup_switch__generate_constants(cases_list, list(prog_var),
+		code_model, case_consts, maybe(set(prog_var)),
 		code_info, code_info).
 :- mode lookup_switch__generate_constants(in, in, in, out, out,
 		in, out) is semidet.
@@ -199,25 +202,25 @@ lookup_switch__generate_constants([], _Vars, _CodeModel, [], no) --> [].
 lookup_switch__generate_constants([Case|Cases], Vars, CodeModel,
 		[CaseVal|Rest], yes(Liveness)) -->
 	{ Case = case(_, int_constant(CaseTag), _, Goal) },
-	code_info__grab_code_info(CodeInfo),
+	code_info__remember_position(BranchStart),
 	code_gen__generate_goal(CodeModel, Goal, Code),
 	code_info__get_forward_live_vars(Liveness),
-	{ tree__is_empty(Code) },
+	{ tree__tree_of_lists_is_empty(Code) },
 	lookup_switch__get_case_rvals(Vars, CaseRvals),
 	{ CaseVal = CaseTag - CaseRvals },
-	code_info__slap_code_info(CodeInfo),
+	code_info__reset_to_position(BranchStart),
 	lookup_switch__generate_constants(Cases, Vars, CodeModel, Rest, _).
 
 %---------------------------------------------------------------------------%
 
-:- pred lookup_switch__get_case_rvals(list(var), list(rval),
+:- pred lookup_switch__get_case_rvals(list(prog_var), list(rval),
 		code_info, code_info).
 :- mode lookup_switch__get_case_rvals(in, out, in, out) is semidet.
 
 lookup_switch__get_case_rvals([], []) --> [].
 lookup_switch__get_case_rvals([Var|Vars], [Rval|Rvals]) -->
 	code_info__produce_variable(Var, Code, Rval),
-	{ tree__is_empty(Code) },
+	{ tree__tree_of_lists_is_empty(Code) },
 	code_info__get_globals(Globals),
 	{ globals__get_options(Globals, Options) },
 	{ exprn_aux__init_exprn_opts(Options, ExprnOpts) },
@@ -244,10 +247,14 @@ lookup_switch__rval_is_constant(binop(_, Exprn0, Exprn1), ExprnOpts) :-
 	lookup_switch__rval_is_constant(Exprn1, ExprnOpts).
 lookup_switch__rval_is_constant(mkword(_, Exprn0), ExprnOpts) :-
 	lookup_switch__rval_is_constant(Exprn0, ExprnOpts).
-lookup_switch__rval_is_constant(create(_, Args, _, _), ExprnOpts) :-
-	ExprnOpts = nlg_asm_sgt_ubf(_, _, StaticGroundTerms, _),
-	StaticGroundTerms = yes,
-	lookup_switch__rvals_are_constant(Args, ExprnOpts).
+lookup_switch__rval_is_constant(create(_, Args, _, StatDyn, _,_), ExprnOpts) :-
+	(
+		StatDyn = must_be_static
+	;
+		ExprnOpts = nlg_asm_sgt_ubf(_, _, StaticGroundTerms, _),
+		StaticGroundTerms = yes,
+		lookup_switch__rvals_are_constant(Args, ExprnOpts)
+	).
 
 :- pred lookup_switch__rvals_are_constant(list(maybe(rval)), exprn_opts).
 :- mode lookup_switch__rvals_are_constant(in, in) is semidet.
@@ -262,7 +269,7 @@ lookup_switch__rvals_are_constant([MRval|MRvals], ExprnOpts) :-
 
 lookup_switch__generate(Var, OutVars, CaseValues,
 		StartVal, EndVal, NeedRangeCheck, NeedBitVecCheck,
-		MLiveness, StoreMap, Code) -->
+		MLiveness, StoreMap, MaybeEnd0, MaybeEnd, Code) -->
 		% Evaluate the variable which we are going to be switching on
 	code_info__produce_variable(Var, VarCode, Rval),
 		% If the case values start at some number other than 0,
@@ -306,7 +313,8 @@ lookup_switch__generate(Var, OutVars, CaseValues,
 		{ MLiveness = no },
 		{ error("lookup_switch__generate: no liveness!") }
 	),
-	code_info__generate_branch_end(model_det, StoreMap, LookupCode),
+	code_info__generate_branch_end(StoreMap, MaybeEnd0, MaybeEnd,
+		LookupCode),
 		% Assemble to code together
 	{ Comment = node([comment("lookup switch") - ""]) },
 	{ Code =
@@ -343,13 +351,13 @@ lookup_switch__generate_bitvec_test(Index, CaseVals, Start, _End,
 		% low bits specify which bit.
 		%
 	{
-		BitVec = create(_, [yes(SingleWord)], _, _)
+		BitVec = create(_, [yes(SingleWord)], _, _, _, _)
 	->
 		Word = SingleWord,
 		BitNum = UIndex
 	;
 		WordNum = binop(/, UIndex, const(int_const(WordBits))),
-		Word = lval(field(0, BitVec, WordNum)),
+		Word = lval(field(yes(0), BitVec, WordNum)),
 		BitNum = binop(mod, UIndex, const(int_const(WordBits)))
 	},
 	{ HasBit = binop((&),
@@ -387,7 +395,8 @@ generate_bit_vec(CaseVals, Start, WordBits, BitVec) -->
 	{ map__to_assoc_list(BitMap, WordVals) },
 	{ generate_bit_vec_args(WordVals, 0, Args) },
 	code_info__get_next_cell_number(CellNo),
-	{ BitVec = create(0, Args, no, CellNo) }.
+	{ BitVec = create(0, Args, uniform(no), must_be_static,
+		CellNo, "lookup_switch_bit_vector") }.
 
 :- pred generate_bit_vec_2(case_consts, int, int,
 			map(int, int), map(int, int)).
@@ -428,7 +437,7 @@ generate_bit_vec_args([Word - Bits|Rest], Count, [yes(Rval)|Rvals]) :-
 
 %------------------------------------------------------------------------------%
 
-:- pred lookup_switch__generate_terms(rval, list(var),
+:- pred lookup_switch__generate_terms(rval, list(prog_var),
 		case_consts, int, code_info, code_info).
 :- mode lookup_switch__generate_terms(in, in, in, in, in, out) is det.
 
@@ -443,7 +452,7 @@ lookup_switch__generate_terms(Index, OutVars, CaseVals, Start) -->
 	{ rearrange_vals(OutVars, CaseVals, Start, Empty, ValMap) },
 	lookup_switch__generate_terms_2(Index, OutVars, ValMap).
 
-:- pred lookup_switch__generate_terms_2(rval, list(var),
+:- pred lookup_switch__generate_terms_2(rval, list(prog_var),
 		rval_map, code_info, code_info).
 :- mode lookup_switch__generate_terms_2(in, in, in, in, out) is det.
 
@@ -453,8 +462,9 @@ lookup_switch__generate_terms_2(Index, [Var|Vars], Map) -->
 	{ list__sort(Vals0, Vals) },
 	{ construct_args(Vals, 0, Args) },
 	code_info__get_next_cell_number(CellNo),
-	{ ArrayTerm = create(0, Args, no, CellNo) },
-	{ LookupTerm = lval(field(0, ArrayTerm, Index)) },
+	{ ArrayTerm = create(0, Args, uniform(no), must_be_static,
+		CellNo, "lookup_switch_data") },
+	{ LookupTerm = lval(field(yes(0), ArrayTerm, Index)) },
 	code_info__cache_expression(Var, LookupTerm),
 	lookup_switch__generate_terms_2(Index, Vars, Map).
 
@@ -479,7 +489,7 @@ construct_args([Index - Rval|Rest], Count0, [yes(Arg)|Args]) :-
 
 %------------------------------------------------------------------------------%
 
-:- pred rearrange_vals(list(var), case_consts, int, rval_map, rval_map).
+:- pred rearrange_vals(list(prog_var), case_consts, int, rval_map, rval_map).
 :- mode rearrange_vals(in, in, in, in, out) is det.
 
 	% For the purpose of constructing the terms, the case_consts
@@ -492,7 +502,7 @@ rearrange_vals(Vars, [Tag - Rvals|Rest], Start, Map0, Map) :-
 	rearrange_vals_2(Pairs, Index, Map0, Map1),
 	rearrange_vals(Vars, Rest, Start, Map1, Map).
 
-:- pred rearrange_vals_2(list(pair(var, rval)), int, rval_map, rval_map).
+:- pred rearrange_vals_2(list(pair(prog_var, rval)), int, rval_map, rval_map).
 :- mode rearrange_vals_2(in, in, in, out) is det.
 
 rearrange_vals_2([], _, Map, Map).
