@@ -1,5 +1,5 @@
 %----------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -33,15 +33,16 @@
 :- import_module list, io.
 :- import_module prog_data.
 
-:- pred convert_to_goedel(string, list(item_and_context), io__state, io__state).
+:- pred convert_to_goedel(module_name, list(item_and_context),
+			io__state, io__state).
 :- mode convert_to_goedel(in, in, di, uo) is det.
 
 %-----------------------------------------------------------------------------%
 
 :- implementation.
-:- import_module bool, int, char, std_util, varset, term, require, string.
-:- import_module prog_io, prog_out, prog_util, equiv_type.
-:- import_module globals, options.
+:- import_module bool, int, char, std_util, require, string.
+:- import_module prog_io, prog_out, prog_util, equiv_type, purity.
+:- import_module globals, options, term, varset.
 %-----------------------------------------------------------------------------%
 
 	% The following is a hard-coded hack.
@@ -60,7 +61,7 @@ option_handle_functor_overloading("character").
 
 %-----------------------------------------------------------------------------%
 
-convert_to_goedel(ProgName, Items0) -->
+convert_to_goedel(ModuleNameSym, Items0) -->
 	io__stderr_stream(StdErr),
 	io__write_string(StdErr, "% Expanding equivalence types..."),
 	io__flush_output(StdErr),
@@ -69,7 +70,9 @@ convert_to_goedel(ProgName, Items0) -->
 					Items, Error,  _),
 	io__write_string(StdErr, " done\n"),
 	( { Error = no } ->
-		{ convert_functor_name(ProgName, GoedelName) },
+		{ prog_out__sym_name_to_string(ModuleNameSym, "__",
+				ModuleName) },
+		{ convert_functor_name(ModuleName, GoedelName) },
 		{ string__append(GoedelName, ".loc", OutputFileName) },
 		io__tell(OutputFileName, Res),
 		( { Res = ok } ->
@@ -125,7 +128,7 @@ goedel_output_item_list([Item - Context | Items]) -->
 
 %-----------------------------------------------------------------------------%
 
-:- pred goedel_output_item(item, term__context, io__state, io__state).
+:- pred goedel_output_item(item, prog_context, io__state, io__state).
 :- mode goedel_output_item(in, in, di, uo) is det.
 
 	% dispatch on the different types of items
@@ -140,19 +143,39 @@ goedel_output_item(inst_defn(VarSet, InstDefn, _Cond), Context) -->
 goedel_output_item(mode_defn(VarSet, ModeDefn, _Cond), Context) -->
 	goedel_output_mode_defn(VarSet, ModeDefn, Context).
 
-goedel_output_item(pred(VarSet, PredName, TypesAndModes, _Det, _Cond), Context)
-		-->
+	% XXX Should we ignore ClassContext and ExistQVars,
+	% or give an error if they're non-empty?
+goedel_output_item(pred(TypeVarSet, InstVarSet, _ExistQVars, PredName,
+		TypesAndModes, _Det, _Cond, Purity, _ClassContext), Context) -->
 	io__write_string("\n"),
 	maybe_write_line_number(Context),
-	goedel_output_pred(VarSet, PredName, TypesAndModes, Context).
-
-goedel_output_item(func(VarSet, PredName, TypesAndModes, RetTypeAndMode, _Det,
-		_Cond), Context) -->
-	io__write_string("\n"),
-	maybe_write_line_number(Context),
-	goedel_output_func(VarSet, PredName, TypesAndModes, RetTypeAndMode,
+	(   { Purity = pure } ->
+		[]
+	;
+		io__write_string(" /* "),
+		write_purity(Purity),
+		io__write_string(" */ ")
+	),
+	goedel_output_pred(TypeVarSet, InstVarSet, PredName, TypesAndModes,
 		Context).
 
+	% XXX Should we ignore ClassContext and ExistQVars,
+	% or give an error if they're non-empty?
+goedel_output_item(func(TypeVarSet, InstVarSet, _ExistQVars, PredName,
+		TypesAndModes, RetTypeAndMode, _Det,
+		_Cond, Purity, _ClassContext), Context) -->
+	io__write_string("\n"),
+	maybe_write_line_number(Context),
+	(   { Purity = pure } ->
+		[]
+	;
+		io__write_string(" /* "),
+		write_purity(Purity),
+		io__write_string(" */ ")
+	),
+	goedel_output_func(TypeVarSet, InstVarSet, PredName, TypesAndModes,
+		RetTypeAndMode, Context).
+
 goedel_output_item(pred_mode(VarSet, PredName, Modes, _Det, _Cond), Context) -->
 	goedel_output_pred_mode(VarSet, PredName, Modes, Context).
 goedel_output_item(func_mode(VarSet, PredName, ArgModes, RetMode, _Det, _Cond),
@@ -177,13 +200,27 @@ goedel_output_item(func_clause(VarSet, PredName, Args, Result, Body), Context)
 goedel_output_item(pragma(_Pragma), _Context) -->
 	io__stderr_stream(Stderr),
 	io__write_string(Stderr, 
-			"warning: C header declarations not allowed. Ignoring\n").
+			"warning: pragma declarations not allowed. Ignoring\n").
+
+goedel_output_item(assertion(_, _), _Context) -->
+	io__stderr_stream(Stderr),
+	io__write_string(Stderr, 
+			"warning: promise declarations not allowed. Ignoring\n").
 
 goedel_output_item(nothing, _) --> [].
+goedel_output_item(typeclass(_, _, _, _, _), _) -->
+	io__stderr_stream(Stderr),
+	io__write_string(Stderr, 
+			"warning: typeclass declarations not allowed. Ignoring\n").
+
+goedel_output_item(instance(_, _, _, _, _), _) -->
+	io__stderr_stream(Stderr),
+	io__write_string(Stderr, 
+			"warning: instance declarations not allowed. Ignoring\n").
 
 %-----------------------------------------------------------------------------%
 
-:- pred goedel_output_inst_defn(varset, inst_defn, term__context,
+:- pred goedel_output_inst_defn(inst_varset, inst_defn, prog_context,
 			io__state, io__state).
 :- mode goedel_output_inst_defn(in, in, in, di, uo) is det.
 
@@ -191,7 +228,7 @@ goedel_output_inst_defn(_VarSet, _InstDefn, _Context) -->
 	% io__write_string("% inst definitions not supported\n"),
 	[].
 
-:- pred goedel_output_mode_defn(varset, mode_defn, term__context,
+:- pred goedel_output_mode_defn(inst_varset, mode_defn, prog_context,
 			io__state, io__state).
 :- mode goedel_output_mode_defn(in, in, in, di, uo) is det.
 
@@ -201,14 +238,14 @@ goedel_output_mode_defn(_VarSet, _ModeDefn, _Context) -->
 
 %-----------------------------------------------------------------------------%
 
-:- pred goedel_output_type_defn(varset, type_defn, term__context,
+:- pred goedel_output_type_defn(tvarset, type_defn, prog_context,
 			io__state, io__state).
 :- mode goedel_output_type_defn(in, in, in, di, uo) is det.
 
 goedel_output_type_defn(VarSet, TypeDefn, Context) -->
 	goedel_output_type_defn_2(TypeDefn, VarSet, Context).
 
-:- pred goedel_output_type_defn_2(type_defn, varset, term__context,
+:- pred goedel_output_type_defn_2(type_defn, tvarset, prog_context,
 			io__state, io__state).
 :- mode goedel_output_type_defn_2(in, in, in, di, uo) is det.
 
@@ -274,8 +311,8 @@ goedel_output_type_defn_2(du_type(Name, Args, Ctors, EqualityPred),
 				Context)
 	).
 
-:- pred goedel_output_type_defn_3(string, string, list(term), list(constructor),
-			varset, term__context, io__state, io__state).
+:- pred goedel_output_type_defn_3(string, string, list(type), list(constructor),
+			tvarset, prog_context, io__state, io__state).
 :- mode goedel_output_type_defn_3(in, in, in, in, in, in, di, uo) is det.
 
 goedel_output_type_defn_3(Name2, Name3, Args, Ctors, VarSet, Context) -->
@@ -294,16 +331,18 @@ goedel_output_type_defn_3(Name2, Name3, Args, Ctors, VarSet, Context) -->
 	),
 	io__write_string(".\n"),
 	goedel_output_ctors(Ctors,
-		term__functor(term__atom(Name2), Args, Context), VarSet).
+		term__functor(term__atom(Name2), Args, Context),
+		VarSet).
 
-:- pred goedel_output_ctors(list(constructor), type, varset,
+:- pred goedel_output_ctors(list(constructor), type, tvarset,
 				io__state, io__state).
 :- mode goedel_output_ctors(in, in, in, di, uo) is det.
 
 goedel_output_ctors([], _, _) --> [].
-goedel_output_ctors([Name - Args | Ctors], Type, VarSet) -->
-	{ unqualify_name(Name, Name2),
-	  convert_functor_name(Name2, Name3) },
+goedel_output_ctors([Ctor | Ctors], Type, VarSet) -->
+	{ Ctor = ctor(_ExistQVars, _Constraints, Name, Args) },
+	{ unqualify_name(Name, Name2) },
+	{ convert_functor_name(Name2, Name3) },
 	(
 		{ Args = [_ArgName - ArgType | Rest] }
 	->
@@ -338,7 +377,7 @@ goedel_output_ctors([Name - Args | Ctors], Type, VarSet) -->
 	io__write_string(".\n"),
 	goedel_output_ctors(Ctors, Type, VarSet).
 
-:- pred goedel_output_remaining_ctor_args(list(constructor_arg), varset,
+:- pred goedel_output_remaining_ctor_args(list(constructor_arg), tvarset,
 		io__state, io__state).
 :- mode goedel_output_remaining_ctor_args(in, in, di, uo) is det.
 
@@ -350,23 +389,23 @@ goedel_output_remaining_ctor_args([_Name - Type | Args], VarSet) -->
 
 %-----------------------------------------------------------------------------%
 
-:- pred goedel_output_pred(varset, sym_name, list(type_and_mode),
-		term__context, io__state, io__state).
-:- mode goedel_output_pred(in, in, in, in, di, uo) is det.
+:- pred goedel_output_pred(tvarset, inst_varset, sym_name, list(type_and_mode),
+		prog_context, io__state, io__state).
+:- mode goedel_output_pred(in, in, in, in, in, di, uo) is det.
 
-goedel_output_pred(VarSet, PredName, TypesAndModes, Context) -->
+goedel_output_pred(TypeVarSet, InstVarSet, PredName, TypesAndModes, Context) -->
 	{ split_types_and_modes(TypesAndModes, Types, MaybeModes) },
-	goedel_output_pred_type(VarSet, PredName, Types, Context),
+	goedel_output_pred_type(TypeVarSet, PredName, Types, Context),
 	(
 		{ MaybeModes = yes(Modes) }
 	->
-		goedel_output_pred_mode(VarSet, PredName, Modes, Context)
+		goedel_output_pred_mode(InstVarSet, PredName, Modes, Context)
 	;
 		[]
 	).
 
-:- pred goedel_output_pred_type(varset, sym_name, list(type),
-		term__context, io__state, io__state).
+:- pred goedel_output_pred_type(tvarset, sym_name, list(type),
+		prog_context, io__state, io__state).
 :- mode goedel_output_pred_type(in, in, in, in, di, uo) is det.
 
 goedel_output_pred_type(VarSet, PredName, Types, _Context) -->
@@ -394,7 +433,8 @@ goedel_output_pred_type(VarSet, PredName, Types, _Context) -->
 	),
 	io__write_string(".\n").
 
-:- pred goedel_output_remaining_types(list(type), varset, io__state, io__state).
+:- pred goedel_output_remaining_types(list(type), tvarset,
+		io__state, io__state).
 :- mode goedel_output_remaining_types(in, in, di, uo) is det.
 
 goedel_output_remaining_types([], _VarSet) --> [].
@@ -405,39 +445,40 @@ goedel_output_remaining_types([Type | Types], VarSet) -->
 
 %-----------------------------------------------------------------------------%
 
-:- pred goedel_output_func(varset, sym_name, list(type_and_mode), type_and_mode,
-		term__context, io__state, io__state).
-:- mode goedel_output_func(in, in, in, in, in, di, uo) is det.
+:- pred goedel_output_func(tvarset, inst_varset, sym_name, list(type_and_mode),
+		type_and_mode, prog_context, io__state, io__state).
+:- mode goedel_output_func(in, in, in, in, in, in, di, uo) is det.
 
-goedel_output_func(VarSet, PredName, TypesAndModes, RetTypeAndMode, Context) -->
+goedel_output_func(TypeVarSet, InstVarSet, PredName, TypesAndModes,
+		RetTypeAndMode, Context) -->
 	{ split_types_and_modes(TypesAndModes, Types, MaybeModes) },
 	{ split_type_and_mode(RetTypeAndMode, RetType, MaybeRetMode) },
-	goedel_output_func_type(VarSet, PredName, Types, RetType, Context),
+	goedel_output_func_type(TypeVarSet, PredName, Types, RetType, Context),
 	(
 		{ MaybeModes = yes(Modes) },
 		{ MaybeRetMode = yes(RetMode) }
 	->
-		goedel_output_func_mode(VarSet, PredName, Modes, RetMode,
+		goedel_output_func_mode(InstVarSet, PredName, Modes, RetMode,
 			Context)
 	;
 		[]
 	).
 
-:- pred goedel_output_func_type(varset, sym_name, list(type), type,
-		term__context, io__state, io__state).
+:- pred goedel_output_func_type(tvarset, sym_name, list(type), type,
+		prog_context, io__state, io__state).
 :- mode goedel_output_func_type(in, in, in, in, in, di, uo) is det.
 
 goedel_output_func_type(VarSet, FuncName, Types, RetType, _Context) -->
 	{ list__map(lambda([Type::in, Arg::out] is det, (Arg = "" - Type)),
 		Types, Args) },
-	goedel_output_ctors([FuncName - Args], RetType, VarSet).
+	goedel_output_ctors([ctor([], [], FuncName, Args)], RetType, VarSet).
 
 %-----------------------------------------------------------------------------%
 
 	% Output a mode declaration for a predicate.
 
-:- pred goedel_output_pred_mode(varset, sym_name, list(mode), term__context,
-			io__state, io__state).
+:- pred goedel_output_pred_mode(inst_varset, sym_name, list(mode),
+		prog_context, io__state, io__state).
 :- mode goedel_output_pred_mode(in, in, in, in, di, uo) is det.
 
 goedel_output_pred_mode(_VarSet, _PredName, _Modes, _Context) -->
@@ -448,8 +489,8 @@ goedel_output_pred_mode(_VarSet, _PredName, _Modes, _Context) -->
 
 	% Output a mode declaration for a function.
 
-:- pred goedel_output_func_mode(varset, sym_name, list(mode), mode,
-			term__context, io__state, io__state).
+:- pred goedel_output_func_mode(inst_varset, sym_name, list(mode), mode,
+			prog_context, io__state, io__state).
 :- mode goedel_output_func_mode(in, in, in, in, in, di, uo) is det.
 
 goedel_output_func_mode(_VarSet, _PredName, _Modes, _RetMode, _Context) -->
@@ -460,14 +501,14 @@ goedel_output_func_mode(_VarSet, _PredName, _Modes, _RetMode, _Context) -->
 
 	% Output a clause.
 
-:- pred goedel_output_pred_clause(varset, sym_name, list(term), goal,
-			term__context, io__state, io__state).
+:- pred goedel_output_pred_clause(prog_varset, sym_name, list(prog_term), goal,
+			prog_context, io__state, io__state).
 :- mode goedel_output_pred_clause(in, in, in, in, in, di, uo) is det.
 
 goedel_output_pred_clause(VarSet, PredName, Args, Body, Context) -->
 	{ unqualify_name(PredName, PredName2) },
-	goedel_output_term(term__functor(term__atom(PredName2), Args, Context),
-			VarSet),
+	goedel_output_term(term__functor(
+		term__atom(PredName2), Args, Context), VarSet),
 	(
 		{ Body = true - Context }
 	->
@@ -480,16 +521,16 @@ goedel_output_pred_clause(VarSet, PredName, Args, Body, Context) -->
 
 	% Output an equation.
 
-:- pred goedel_output_func_clause(varset, sym_name, list(term), term, goal,
-			term__context, io__state, io__state).
+:- pred goedel_output_func_clause(prog_varset, sym_name, list(prog_term),
+		prog_term, goal, prog_context, io__state, io__state).
 :- mode goedel_output_func_clause(in, in, in, in, in, in, di, uo) is det.
 
 goedel_output_func_clause(VarSet, PredName, Args, Result, Body, Context) -->
 	{ unqualify_name(PredName, PredName2) },
 	goedel_output_term(
 		term__functor(term__atom("="), [
-			term__functor(term__atom(PredName2), Args, Context),
-			Result], Context),
+			term__functor(term__atom(PredName2), Args,
+			Context), Result], Context),
 		VarSet),
 	(
 		{ Body = true - Context }
@@ -501,13 +542,13 @@ goedel_output_func_clause(VarSet, PredName, Args, Result, Body, Context) -->
 	),
 	io__write_string(".\n").
 
-:- pred goedel_output_goal(goal, varset, int, io__state, io__state).
+:- pred goedel_output_goal(goal, prog_varset, int, io__state, io__state).
 :- mode goedel_output_goal(in, in, in, di, uo) is det.
 
 goedel_output_goal(Goal - _Context, VarSet, Indent) -->
 	goedel_output_goal_2(Goal, VarSet, Indent).
 
-:- pred goedel_output_goal_2(goal_expr, varset, int, io__state, io__state).
+:- pred goedel_output_goal_2(goal_expr, prog_varset, int, io__state, io__state).
 :- mode goedel_output_goal_2(in, in, in, di, uo) is det.
 
 goedel_output_goal_2(fail, _, _) -->
@@ -596,6 +637,14 @@ goedel_output_goal_2((A,B), VarSet, Indent) -->
 	goedel_output_newline(Indent),
 	goedel_output_goal(B, VarSet, Indent).
 
+	% Goedel doesn't have parallel conjunction,
+	% but we can use sequential conjunction instead.
+goedel_output_goal_2((A & B), VarSet, Indent) -->
+	goedel_output_goal(A, VarSet, Indent),
+	io__write_string(" &"),
+	goedel_output_newline(Indent),
+	goedel_output_goal(B, VarSet, Indent).
+
 goedel_output_goal_2((A;B), VarSet, Indent) -->
 	io__write_string("("),
 	{ Indent1 is Indent + 1 },
@@ -606,10 +655,18 @@ goedel_output_goal_2((A;B), VarSet, Indent) -->
 	io__write_string(")").
 
 % XXX should preserve some of the qualification information?
-goedel_output_goal_2(call(Name, Term), VarSet, Indent) -->
+goedel_output_goal_2(call(Name, Term, Purity), VarSet, Indent) -->
+	(   { Purity = pure } ->
+		[]
+	;
+		io__write_string("/* "),
+		write_purity(Purity),
+		io__write_string(" */ ")
+	),
 	{ unqualify_name(Name, Name0) },
 	{ term__context_init(Context0) },
-	goedel_output_call(term__functor(term__atom(Name0), Term, Context0), VarSet, Indent).
+	goedel_output_call(term__functor(term__atom(Name0), Term,
+		Context0), VarSet, Indent).
 
 goedel_output_goal_2(unify(A, B), VarSet, _Indent) -->
 	goedel_output_term(A, VarSet),
@@ -617,13 +674,14 @@ goedel_output_goal_2(unify(A, B), VarSet, _Indent) -->
 	goedel_output_term(B, VarSet).
 
 
-:- pred goedel_output_call(term, varset, int, io__state, io__state).
+:- pred goedel_output_call(prog_term, prog_varset, int, io__state, io__state).
 :- mode goedel_output_call(in, in, in, di, uo) is det.
 
 goedel_output_call(term__variable(Var), VarSet, _Indent) -->
 	goedel_output_var(Var, VarSet).
 
-goedel_output_call(term__functor(Functor, Args, _Context), VarSet, _Indent) -->
+goedel_output_call(term__functor(Functor, Args, _Context), VarSet,
+		_Indent) -->
 	(
 		{ Args = [Arg1, Arg2],
 		  Functor = term__atom("is")
@@ -659,7 +717,7 @@ goedel_output_call(term__functor(Functor, Args, _Context), VarSet, _Indent) -->
 		)
 	).
 
-:- pred goedel_output_disj(goal, varset, int, io__state, io__state).
+:- pred goedel_output_disj(goal, prog_varset, int, io__state, io__state).
 :- mode goedel_output_disj(in, in, in, di, uo) is det.
 
 goedel_output_disj(Goal, VarSet, Indent) -->
@@ -676,7 +734,7 @@ goedel_output_disj(Goal, VarSet, Indent) -->
 		goedel_output_goal(Goal, VarSet, Indent1)
 	).
 
-:- pred goedel_output_some(list(var), varset, io__state, io__state).
+:- pred goedel_output_some(list(prog_var), prog_varset, io__state, io__state).
 :- mode goedel_output_some(in, in, di, uo) is det.
 
 goedel_output_some(Vars, VarSet) -->
@@ -690,7 +748,7 @@ goedel_output_some(Vars, VarSet) -->
 		io__write_string("]")
 	).
 
-:- pred goedel_convert_expression(term, term).
+:- pred goedel_convert_expression(term(T), term(T)).
 :- mode goedel_convert_expression(in, out) is det.
 
 goedel_convert_expression(Term0, Term) :-
@@ -732,7 +790,7 @@ goedel_output_tabs(Indent) -->
 
 %-----------------------------------------------------------------------------%
 
-:- pred goedel_output_list_args(term, varset, io__state, io__state).
+:- pred goedel_output_list_args(term(T), varset(T), io__state, io__state).
 :- mode goedel_output_list_args(in, in, di, uo) is det.
 
 goedel_output_list_args(Term, VarSet) -->
@@ -755,7 +813,7 @@ goedel_output_list_args(Term, VarSet) -->
 
 	% write a term to standard output.
 
-:- pred goedel_output_term(term, varset, io__state, io__state).
+:- pred goedel_output_term(term(T), varset(T), io__state, io__state).
 :- mode goedel_output_term(in, in, di, uo) is det.
 
 goedel_output_term(term__variable(Var), VarSet) -->
@@ -821,7 +879,7 @@ goedel_output_term(term__functor(Functor, Args, _), VarSet) -->
 
 	% output the remaining arguments
 
-:- pred goedel_output_term_args(list(term), varset, io__state, io__state).
+:- pred goedel_output_term_args(list(term(T)), varset(T), io__state, io__state).
 :- mode goedel_output_term_args(in, in, di, uo) is det.
 
 goedel_output_term_args([], _VarSet) --> [].
@@ -874,7 +932,7 @@ goedel_quote_char('\b', 'b').
 
 	% output a comma-separated list of variables
 
-:- pred goedel_output_vars(list(var), varset, io__state, io__state).
+:- pred goedel_output_vars(list(var(T)), varset(T), io__state, io__state).
 :- mode goedel_output_vars(in, in, di, uo) is det.
 
 goedel_output_vars([], _VarSet) --> [].
@@ -882,7 +940,7 @@ goedel_output_vars([Var | Vars], VarSet) -->
 	goedel_output_var(Var, VarSet),
 	goedel_output_vars_2(Vars, VarSet).
 
-:- pred goedel_output_vars_2(list(var), varset, io__state, io__state).
+:- pred goedel_output_vars_2(list(var(T)), varset(T), io__state, io__state).
 :- mode goedel_output_vars_2(in, in, di, uo) is det.
 
 goedel_output_vars_2([], _VarSet) --> [].
@@ -897,7 +955,7 @@ goedel_output_vars_2([Var | Vars], VarSet) -->
 	% Variables whose name originally started with `v_' have their
 	% name changed to start with `v__' to avoid name clashes.
 
-:- pred goedel_output_var(var, varset, io__state, io__state).
+:- pred goedel_output_var(var(T), varset(T), io__state, io__state).
 :- mode goedel_output_var(in, in, di, uo) is det.
 
 goedel_output_var(Var, VarSet) -->
@@ -918,7 +976,7 @@ goedel_output_var(Var, VarSet) -->
 
 	% write a type to standard output.
 
-:- pred goedel_output_type(term, varset, io__state, io__state).
+:- pred goedel_output_type((type), tvarset, io__state, io__state).
 :- mode goedel_output_type(in, in, di, uo) is det.
 
 goedel_output_type(term__variable(Var), VarSet) -->
@@ -938,7 +996,7 @@ goedel_output_type(term__functor(Functor, Args, _), VarSet) -->
 
 	% output the remaining arguments
 
-:- pred goedel_output_type_args(list(type), varset, io__state, io__state).
+:- pred goedel_output_type_args(list(type), tvarset, io__state, io__state).
 :- mode goedel_output_type_args(in, in, di, uo) is det.
 
 goedel_output_type_args([], _VarSet) --> [].
@@ -1130,7 +1188,7 @@ convert_var_name(Name, GoedelName) :-
 
 %-----------------------------------------------------------------------------%
 
-:- pred maybe_write_line_number(term__context, io__state, io__state).
+:- pred maybe_write_line_number(prog_context, io__state, io__state).
 :- mode maybe_write_line_number(in, di, uo) is det.
 
 maybe_write_line_number(Context) -->
