@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -25,10 +25,11 @@
 
 :- implementation.
 
-:- import_module hlds_module, hlds_data.
+:- import_module builtin_ops, hlds_module, hlds_data, prog_data, prog_out.
 :- import_module code_gen, unify_gen, code_util, code_aux, opt_util.
+
 :- import_module bool, set, int, std_util, tree, list, assoc_list, require.
-:- import_module string, term.
+:- import_module string.
 
 %---------------------------------------------------------------------------%
 
@@ -86,7 +87,7 @@ middle_rec__match_and_generate(Goal, Instrs, CodeInfo0, CodeInfo) :-
 
 %---------------------------------------------------------------------------%
 
-:- pred middle_rec__generate_ite(list(var), hlds_goal, hlds_goal,
+:- pred middle_rec__generate_ite(list(prog_var), hlds_goal, hlds_goal,
 	hlds_goal, ite_rec, store_map, hlds_goal_info,
 	code_tree, code_info, code_info).
 :- mode middle_rec__generate_ite(in, in, in, in, in, in, in, out, in, out)
@@ -102,7 +103,7 @@ middle_rec__generate_ite(_Vars, _Cond, _Then, _Else, _Rec, _IteGoalInfo, _SM,
 
 %---------------------------------------------------------------------------%
 
-:- pred middle_rec__generate_switch(var, cons_id, hlds_goal, hlds_goal,
+:- pred middle_rec__generate_switch(prog_var, cons_id, hlds_goal, hlds_goal,
 	store_map, hlds_goal_info, code_tree, code_info, code_info).
 :- mode middle_rec__generate_switch(in, in, in, in, in, in, out, in, out)
 	is semidet.
@@ -124,22 +125,24 @@ middle_rec__generate_switch(Var, BaseConsId, Base, Recursive, StoreMap,
 	{ tree__flatten(EntryTestCode, EntryTestListList) },
 	{ list__condense(EntryTestListList, EntryTestList) },
 
-	code_info__grab_code_info(CodeInfo),
+	code_info__remember_position(BranchStart),
 	code_gen__generate_goal(model_det, Base, BaseGoalCode),
-	code_info__generate_branch_end(model_det, StoreMap, BaseSaveCode),
-	code_info__slap_code_info(CodeInfo),
+	code_info__generate_branch_end(StoreMap, no, MaybeEnd1,
+		BaseSaveCode),
+	code_info__reset_to_position(BranchStart),
 	code_gen__generate_goal(model_det, Recursive, RecGoalCode),
-	code_info__generate_branch_end(model_det, StoreMap, RecSaveCode),
+	code_info__generate_branch_end(StoreMap, MaybeEnd1, MaybeEnd,
+		RecSaveCode),
 
 	code_info__post_goal_update(SwitchGoalInfo),
-	code_info__remake_with_store_map(StoreMap),
+	code_info__after_all_branches(StoreMap, MaybeEnd),
 
 	code_info__get_arginfo(ArgModes),
 	code_info__get_headvars(HeadVars),
 	{ assoc_list__from_corresponding_lists(HeadVars, ArgModes, Args) },
 	code_info__setup_call(Args, callee, EpilogCode),
 
-	{ code_gen__output_args(Args, LiveArgs) },
+	{ code_util__output_args(Args, LiveArgs) },
 
 	{ BaseCode = tree(BaseGoalCode, tree(BaseSaveCode, EpilogCode)) },
 	{ RecCode = tree(RecGoalCode, tree(RecSaveCode, EpilogCode)) },
@@ -150,6 +153,12 @@ middle_rec__generate_switch(Var, BaseConsId, Base, Recursive, StoreMap,
 	{ tree__flatten(RecCode, RecListList) },
 	{ list__condense(RecListList, RecList) },
 
+	% In the code we generate, the base instruction sequence is executed
+	% in situations where this procedure has no stack frame. If this
+	% sequence refers to stackvars, it will be to some other procedure's
+	% variables, which is obviously incorrect.
+	{ opt_util__block_refers_stackvars(BaseList, no) },
+
 	{ list__append(BaseList, RecList, AvoidList) },
 	{ middle_rec__find_unused_register(AvoidList, AuxReg) },
 
@@ -157,13 +166,13 @@ middle_rec__generate_switch(Var, BaseConsId, Base, Recursive, StoreMap,
 	{ middle_rec__add_counter_to_livevals(BeforeList0, AuxReg,
 		BeforeList) },
 
-	{ middle_rec__generate_downloop_test(EntryTestList,
-		Loop1Label, Loop1Test) },
-
 	code_info__get_next_label(Loop1Label),
 	code_info__get_next_label(Loop2Label),
 	code_info__get_total_stackslot_count(FrameSize),
 
+	{ middle_rec__generate_downloop_test(EntryTestList,
+		Loop1Label, Loop1Test) },
+
 	{ FrameSize = 0 ->
 		MaybeIncrSp = [],
 		MaybeDecrSp = [],
@@ -183,8 +192,10 @@ middle_rec__generate_switch(Var, BaseConsId, Base, Recursive, StoreMap,
 				- "test on upward loop"]
 	;
 		predicate_module(ModuleInfo, PredId, ModuleName),
+		prog_out__sym_name_to_string(ModuleName, ModuleNameString),
 		predicate_name(ModuleInfo, PredId, PredName),
-		string__append_list([ModuleName, ":", PredName], PushMsg),
+		string__append_list([ModuleNameString, ":", PredName],
+			PushMsg),
 		MaybeIncrSp = [incr_sp(FrameSize, PushMsg) - ""],
 		MaybeDecrSp = [decr_sp(FrameSize) - ""],
 		InitAuxReg =  [assign(AuxReg, lval(sp))
@@ -309,7 +320,7 @@ middle_rec__generate_downloop_test([Instr0 | Instrs0], Target, Instrs) :-
 middle_rec__split_rec_code([], _, _) :-
 	error("did not find call in middle_rec__split_rec_code").
 middle_rec__split_rec_code([Instr0 | Instrs1], Before, After) :-
-	( Instr0 = call(_, _, _, _) - _ ->
+	( Instr0 = call(_, _, _, _, _) - _ ->
 		(
 			opt_util__skip_comments(Instrs1, Instrs2),
 			Instrs2 = [Instr2 | Instrs3],
@@ -386,9 +397,8 @@ middle_rec__find_used_registers_instr(block(_, _, Instrs), Used0, Used) :-
 middle_rec__find_used_registers_instr(assign(Lval, Rval), Used0, Used) :-
 	middle_rec__find_used_registers_lval(Lval, Used0, Used1),
 	middle_rec__find_used_registers_rval(Rval, Used1, Used).
-middle_rec__find_used_registers_instr(call(_, _, _, _), Used, Used).
-middle_rec__find_used_registers_instr(mkframe(_, _, _), Used, Used).
-middle_rec__find_used_registers_instr(modframe(_), Used, Used).
+middle_rec__find_used_registers_instr(call(_, _, _, _, _), Used, Used).
+middle_rec__find_used_registers_instr(mkframe(_, _), Used, Used).
 middle_rec__find_used_registers_instr(label(_), Used, Used).
 middle_rec__find_used_registers_instr(goto(_), Used, Used).
 middle_rec__find_used_registers_instr(computed_goto(Rval, _), Used0, Used) :-
@@ -396,7 +406,7 @@ middle_rec__find_used_registers_instr(computed_goto(Rval, _), Used0, Used) :-
 middle_rec__find_used_registers_instr(c_code(_), Used, Used).
 middle_rec__find_used_registers_instr(if_val(Rval, _), Used0, Used) :-
 	middle_rec__find_used_registers_rval(Rval, Used0, Used).
-middle_rec__find_used_registers_instr(incr_hp(Lval, _, Rval), Used0, Used) :-
+middle_rec__find_used_registers_instr(incr_hp(Lval, _, Rval, _), Used0, Used) :-
 	middle_rec__find_used_registers_lval(Lval, Used0, Used1),
 	middle_rec__find_used_registers_rval(Rval, Used1, Used).
 middle_rec__find_used_registers_instr(mark_hp(Lval), Used0, Used) :-
@@ -414,10 +424,38 @@ middle_rec__find_used_registers_instr(discard_tickets_to(Rval), Used0, Used) :-
 	middle_rec__find_used_registers_rval(Rval, Used0, Used).
 middle_rec__find_used_registers_instr(incr_sp(_, _), Used, Used).
 middle_rec__find_used_registers_instr(decr_sp(_), Used, Used).
-middle_rec__find_used_registers_instr(pragma_c(_, Ins, _, Outs, _),
+middle_rec__find_used_registers_instr(pragma_c(_, Components, _, _, _, _),
 		Used0, Used) :-
-	insert_pragma_c_input_registers(Ins, Used0, Used1),
-	insert_pragma_c_output_registers(Outs, Used1, Used).
+	middle_rec__find_used_registers_components(Components, Used0, Used).
+middle_rec__find_used_registers_instr(init_sync_term(Lval, _), Used0, Used) :-
+	middle_rec__find_used_registers_lval(Lval, Used0, Used).
+middle_rec__find_used_registers_instr(fork(_, _, _), Used, Used).
+middle_rec__find_used_registers_instr(join_and_terminate(Lval), Used0, Used) :-
+	middle_rec__find_used_registers_lval(Lval, Used0, Used).
+middle_rec__find_used_registers_instr(join_and_continue(Lval,_), Used0, Used) :-
+	middle_rec__find_used_registers_lval(Lval, Used0, Used).
+
+:- pred middle_rec__find_used_registers_components(list(pragma_c_component),
+	set(int), set(int)).
+:- mode middle_rec__find_used_registers_components(in, di, uo) is det.
+
+middle_rec__find_used_registers_components([], Used, Used).
+middle_rec__find_used_registers_components([Comp | Comps], Used0, Used) :-
+	middle_rec__find_used_registers_component(Comp, Used0, Used1),
+	middle_rec__find_used_registers_components(Comps, Used1, Used).
+
+:- pred middle_rec__find_used_registers_component(pragma_c_component,
+	set(int), set(int)).
+:- mode middle_rec__find_used_registers_component(in, di, uo) is det.
+
+middle_rec__find_used_registers_component(pragma_c_inputs(In), Used0, Used) :-
+	insert_pragma_c_input_registers(In, Used0, Used).
+middle_rec__find_used_registers_component(pragma_c_outputs(Out), Used0, Used) :-
+	insert_pragma_c_output_registers(Out, Used0, Used).
+middle_rec__find_used_registers_component(pragma_c_user_code(_, _), Used, Used).
+middle_rec__find_used_registers_component(pragma_c_raw_code(_), Used, Used).
+middle_rec__find_used_registers_component(pragma_c_fail_to(_), Used, Used).
+middle_rec__find_used_registers_component(pragma_c_noop, Used, Used).
 
 :- pred middle_rec__find_used_registers_lvals(list(lval), set(int), set(int)).
 :- mode middle_rec__find_used_registers_lvals(in, di, uo) is det.
@@ -454,7 +492,7 @@ middle_rec__find_used_registers_rval(Rval, Used0, Used) :-
 		Rval = var(_),
 		error("var found in middle_rec__find_used_registers_rval")
 	;
-		Rval = create(_, MaybeRvals, _, _),
+		Rval = create(_, MaybeRvals, _, _, _, _),
 		middle_rec__find_used_registers_maybe_rvals(MaybeRvals,
 			Used0, Used)
 	;
