@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -17,8 +17,9 @@
 
 :- interface.
 
-:- import_module hlds_data, prog_data, mode_info, (inst).
-:- import_module set, assoc_list.
+:- import_module hlds_data, hlds_pred, hlds_module, hlds_goal.
+:- import_module prog_data, mode_info, (inst).
+:- import_module bool, set, assoc_list, list.
 
 %-----------------------------------------------------------------------------%
 
@@ -26,11 +27,11 @@
 	---> disj
 	;    if_then_else.
 
-:- type merge_errors == assoc_list(var, list(inst)).
+:- type merge_errors == assoc_list(prog_var, list(inst)).
 
 :- type delayed_goal
 	--->	delayed_goal(
-			set(var),	% The vars it's waiting on
+			set(prog_var),	% The vars it's waiting on
 			mode_error_info,% The reason it can't be scheduled
 			hlds_goal	% The goal itself
 		).
@@ -39,21 +40,30 @@
 	--->	mode_error_disj(merge_context, merge_errors)
 			% different arms of a disjunction result in
 			% different insts for some non-local variables
-	;	mode_error_higher_order_pred_var(pred_or_func, var, inst, arity)
+	;	mode_error_par_conj(merge_errors)
+			% different arms of a parallel conj result in
+			% mutually exclusive bindings - ie the process
+			% of unifying the instmaps from the end of each
+			% branch failed.
+	;	mode_error_higher_order_pred_var(pred_or_func, prog_var, inst,
+				arity)
 			% the predicate variable in a higher-order predicate
 			% or function call didn't have a higher-order
 			% predicate or function inst of the appropriate arity
-	;	mode_error_var_is_live(var)
+	;	mode_error_poly_unify(prog_var, inst)
+			% A variable in a polymorphic unification with unknown
+			% type has inst other than `ground' or `any'.
+	;	mode_error_var_is_live(prog_var)
 			% call to a predicate which will clobber its argument,
 			% but the argument is still live
-	;	mode_error_var_has_inst(var, inst, inst)
+	;	mode_error_var_has_inst(prog_var, inst, inst)
 			% call to a predicate with an insufficiently
 			% instantiated variable (for preds with one mode)
-	;	mode_error_unify_pred(var, mode_error_unify_rhs, type,
+	;	mode_error_unify_pred(prog_var, mode_error_unify_rhs, type,
 				pred_or_func)
 			% an attempt was made to unify two higher-order
 			% predicate or function variables.
-	;	mode_error_implied_mode(var, inst, inst)
+	;	mode_error_implied_mode(prog_var, inst, inst)
 			% a call to a predicate with an overly
 			% instantiated variable would use an implied
 			% mode of the predicate.  XXX This is temporary - 
@@ -62,30 +72,40 @@
 	;	mode_error_no_mode_decl
 			% a call to a predicate for which there are
 			% no mode declarations
-	;	mode_error_no_matching_mode(list(var), list(inst))
+	;	mode_error_no_matching_mode(list(prog_var), list(inst))
 			% call to a predicate with an insufficiently
 			% instantiated variable (for preds with >1 mode)
-	;	mode_error_bind_var(var, inst, inst)
+	;	mode_error_bind_var(var_lock_reason, prog_var, inst, inst)
 			% attempt to bind a non-local variable inside
-			% a negated context
-	;	mode_error_unify_var_var(var, var, inst, inst)
+			% a negated context, or attempt to re-bind a variable
+			% in a parallel conjunct
+	;	mode_error_non_local_lambda_var(prog_var, inst)
+			% attempt to pass a live non-ground var as a
+			% non-local variable to a lambda goal
+	;	mode_error_unify_var_var(prog_var, prog_var, inst, inst)
 			% attempt to unify two free variables
-	;	mode_error_unify_var_functor(var, cons_id, list(var),
+	;	mode_error_unify_var_functor(prog_var, cons_id, list(prog_var),
 							inst, list(inst))
 			% attempt to unify a free var with a functor containing
 			% free arguments
-	;	mode_error_unify_var_lambda(var, inst, inst)
+	;	mode_error_unify_var_lambda(prog_var, inst, inst)
 			% some sort of error in
 			% attempt to unify a variable with lambda expression
-	;	mode_error_conj(list(delayed_goal))
+	;	mode_error_conj(list(delayed_goal), schedule_culprit)
 			% a conjunction contains one or more unscheduleable
-			% goals
-	;	mode_error_final_inst(int, var, inst, inst, final_inst_error)
+			% goals; schedule_culprit gives the reason why
+			% they couldn't be scheduled.
+	;	mode_error_final_inst(int, prog_var, inst, inst,
+				final_inst_error)
 			% one of the head variables did not have the
 			% expected final inst on exit from the proc
-	;	mode_error_undefined_mode_in_lambda.
-			% This is a dummy error - the actual message
-			% is output by module_qual.m.
+	.
+
+:- type schedule_culprit
+	--->	goal_itself_was_impure
+	;	goals_followed_by_impure_goal(hlds_goal)
+	;	conj_floundered. % we've reached the end of a conjunction
+				% and there were still delayed goals
 
 :- type final_inst_error
 	--->	too_instantiated
@@ -94,18 +114,18 @@
 					% fit into the above two categories.
 
 :- type mode_error_unify_rhs
-	--->	error_at_var(var)
-	;	error_at_functor(cons_id, list(var))
-	;	error_at_lambda(list(var), list(mode)).
+	--->	error_at_var(prog_var)
+	;	error_at_functor(cons_id, list(prog_var))
+	;	error_at_lambda(list(prog_var), list(mode)).
 
 :- type mode_error_info
 	---> mode_error_info(
-		set(var),	% the variables which caused the error
+		set(prog_var),	% the variables which caused the error
 				% (we will attempt to reschedule the goal
 				% if the one of these variables becomes
 				% more instantiated)
 		mode_error,	% the nature of the error
-		term__context,	% where the error occurred
+		prog_context,	% where the error occurred
 		mode_context	% where the error occurred
 	).
 
@@ -135,31 +155,45 @@
 :- pred mode_context_init(mode_context).
 :- mode mode_context_init(out) is det.
 
-	% write out the inferred `mode' declarations for a list of pred_ids.
+	% Write out the inferred `mode' declarations for a list of pred_ids.
+	% The bool indicates whether or not to write out determinism
+	% annotations on the modes (it should only be set to `yes' _after_
+	% determinism analysis).
 
-:- pred write_mode_inference_messages(list(pred_id), module_info,
+:- pred write_mode_inference_messages(list(pred_id), bool, module_info,
 				io__state, io__state).
-:- mode write_mode_inference_messages(in, in, di, uo) is det.
+:- mode write_mode_inference_messages(in, in, in, di, uo) is det.
+
+	% report an error for the case when two mode declarations
+	% declare indistinguishable modes
+
+:- pred report_indistinguishable_modes_error(proc_id, proc_id,
+		pred_id, pred_info, module_info, io__state, io__state).
+:- mode report_indistinguishable_modes_error(in, in, in, in, in, di, uo) is det.
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
 :- implementation.
 
-:- import_module hlds_module, hlds_pred, hlds_goal, hlds_out.
+:- import_module hlds_out.
 :- import_module mode_info, mode_util, prog_out, mercury_to_mercury.
-:- import_module options, globals.
-:- import_module bool, int, list, map, io, term, term_io, varset.
+:- import_module options, globals, varset, term.
+:- import_module int, map, io, term_io.
 :- import_module std_util, require.
 
 	% just dispatch on the diffferent sorts of mode errors
 
 report_mode_error(mode_error_disj(MergeContext, ErrorList), ModeInfo) -->
 	report_mode_error_disj(ModeInfo, MergeContext, ErrorList).
+report_mode_error(mode_error_par_conj(ErrorList), ModeInfo) -->
+	report_mode_error_par_conj(ModeInfo, ErrorList).
 report_mode_error(mode_error_higher_order_pred_var(PredOrFunc, Var, Inst,
 		Arity), ModeInfo) -->
 	report_mode_error_higher_order_pred_var(ModeInfo, PredOrFunc, Var,
 		Inst, Arity).
+report_mode_error(mode_error_poly_unify(Var, Inst), ModeInfo) -->
+	report_mode_error_poly_unify(ModeInfo, Var, Inst).
 report_mode_error(mode_error_var_is_live(Var), ModeInfo) -->
 	report_mode_error_var_is_live(ModeInfo, Var).
 report_mode_error(mode_error_var_has_inst(Var, InstA, InstB), ModeInfo) -->
@@ -171,8 +205,10 @@ report_mode_error(mode_error_implied_mode(Var, InstA, InstB), ModeInfo) -->
 	report_mode_error_implied_mode(ModeInfo, Var, InstA, InstB).
 report_mode_error(mode_error_no_mode_decl, ModeInfo) -->
 	report_mode_error_no_mode_decl(ModeInfo).
-report_mode_error(mode_error_bind_var(Var, InstA, InstB), ModeInfo) -->
-	report_mode_error_bind_var(ModeInfo, Var, InstA, InstB).
+report_mode_error(mode_error_bind_var(Reason, Var, InstA, InstB), ModeInfo) -->
+	report_mode_error_bind_var(ModeInfo, Reason, Var, InstA, InstB).
+report_mode_error(mode_error_non_local_lambda_var(Var, Inst), ModeInfo) -->
+	report_mode_error_non_local_lambda_var(ModeInfo, Var, Inst).
 report_mode_error(mode_error_unify_var_var(VarA, VarB, InstA, InstB),
 		ModeInfo) -->
 	report_mode_error_unify_var_var(ModeInfo, VarA, VarB, InstA, InstB).
@@ -183,28 +219,30 @@ report_mode_error(mode_error_unify_var_functor(Var, Name, Args, Inst,
 			ArgInsts), ModeInfo) -->
 	report_mode_error_unify_var_functor(ModeInfo, Var, Name, Args, Inst,
 			ArgInsts).
-report_mode_error(mode_error_conj(Errors), ModeInfo) -->
-	report_mode_error_conj(ModeInfo, Errors).
+report_mode_error(mode_error_conj(Errors, Culprit), ModeInfo) -->
+	report_mode_error_conj(ModeInfo, Errors, Culprit).
 report_mode_error(mode_error_no_matching_mode(Vars, Insts), ModeInfo) -->
 	report_mode_error_no_matching_mode(ModeInfo, Vars, Insts).
 report_mode_error(mode_error_final_inst(ArgNum, Var, VarInst, Inst, Reason),
 		ModeInfo) -->
 	report_mode_error_final_inst(ModeInfo, ArgNum, Var, VarInst, Inst,
 		Reason).
-report_mode_error(mode_error_undefined_mode_in_lambda, _ModeInfo) --> [].
 
 %-----------------------------------------------------------------------------%
 
-:- pred report_mode_error_conj(mode_info, list(delayed_goal),
+:- pred report_mode_error_conj(mode_info, list(delayed_goal), schedule_culprit,
 				io__state, io__state).
-:- mode report_mode_error_conj(mode_info_no_io, in, di, uo) is det.
+:- mode report_mode_error_conj(mode_info_no_io, in, in, di, uo) is det.
 
-report_mode_error_conj(ModeInfo, Errors) -->
+report_mode_error_conj(ModeInfo, Errors, Culprit) -->
 	{ mode_info_get_context(ModeInfo, Context) },
 	{ mode_info_get_varset(ModeInfo, VarSet) },
 	{ find_important_errors(Errors, ImportantErrors, OtherErrors) },
+
+	% if there's more than one error, and we have verbose-errors
+	% enabled, report them all
 	globals__io_lookup_bool_option(verbose_errors, VerboseErrors),
-	( { VerboseErrors = yes } ->
+	( { VerboseErrors = yes, Errors = [_, _ | _] } ->
 		mode_info_write_context(ModeInfo),
 		prog_out__write_context(Context),
 		io__write_string("  mode error in conjunction. The next "),
@@ -230,6 +268,28 @@ report_mode_error_conj(ModeInfo, Errors) -->
 	;
 		% There wasn't any error to report!  This can't happen.
 		{ error("report_mode_error_conj") }
+	),
+
+	% if the goal(s) couldn't be scheduled because we couldn't
+	% reorder things past an impure goal, then report that.
+	( { Culprit = conj_floundered },
+		{ true } % we've already reported everything we can
+	; { Culprit = goal_itself_was_impure },
+		prog_out__write_context(Context),
+		io__write_string(
+		"  The goal could not be reordered, because it was impure.\n")
+	; { Culprit = goals_followed_by_impure_goal(ImpureGoal) },
+		prog_out__write_context(Context),
+		io__write_string(
+			"  The goal could not be reordered, because\n"),
+		prog_out__write_context(Context),
+		io__write_string(
+			"  it was followed by an impure goal.\n"),
+		{ ImpureGoal = _ - ImpureGoalInfo },
+		{ goal_info_get_context(ImpureGoalInfo, ImpureGoalContext) },
+		prog_out__write_context(ImpureGoalContext),
+		io__write_string(
+			"  This is the location of the impure goal.\n")
 	).
 
 :- pred find_important_errors(list(delayed_goal), list(delayed_goal),
@@ -238,12 +298,17 @@ report_mode_error_conj(ModeInfo, Errors) -->
 
 find_important_errors([], [], []).
 find_important_errors([Error | Errors], ImportantErrors, OtherErrors) :-
-	Error = delayed_goal(_, mode_error_info(_, _, _, ModeContext), _),
+	Error = delayed_goal(_, mode_error_info(_, ModeError, _, ModeContext),
+				_),
 	(
 		% an error is important unless it is a non-explicit unification,
 		% i.e. a head unification or a call argument unification
 		ModeContext = unify(unify_context(UnifyContext, _), _),
-		UnifyContext \= explicit
+		UnifyContext \= explicit,
+		% except that errors in lambda goals are important even
+		% if the unification that creates the lambda goal is
+		% an implicit one
+		ModeError \= mode_error_non_local_lambda_var(_, _)
 	->
 		ImportantErrors1 = ImportantErrors,
 		OtherErrors = [Error | OtherErrors1]
@@ -253,7 +318,7 @@ find_important_errors([Error | Errors], ImportantErrors, OtherErrors) :-
 	),
 	find_important_errors(Errors, ImportantErrors1, OtherErrors1).
 
-:- pred report_mode_error_conj_2(list(delayed_goal), varset, term__context,
+:- pred report_mode_error_conj_2(list(delayed_goal), prog_varset, prog_context,
 				mode_info, io__state, io__state).
 :- mode report_mode_error_conj_2(in, in, in, mode_info_no_io, di, uo) is det.
 
@@ -274,7 +339,7 @@ report_mode_error_conj_2([delayed_goal(Vars, Error, Goal) | Rest],
 	( { VeryVerbose = yes } ->
 		io__write_string("\t\t"),
 		{ mode_info_get_module_info(ModeInfo, ModuleInfo) },
-		hlds_out__write_goal(Goal, ModuleInfo, VarSet, no, 2, ".")
+		hlds_out__write_goal(Goal, ModuleInfo, VarSet, no, 2, ".\n")
 	;
 		[]
 	),
@@ -299,6 +364,23 @@ report_mode_error_disj(ModeInfo, MergeContext, ErrorList) -->
 	io__write_string(".\n"),
 	write_merge_error_list(ErrorList, ModeInfo).
 
+:- pred report_mode_error_par_conj(mode_info, merge_errors,
+				io__state, io__state).
+:- mode report_mode_error_par_conj(mode_info_no_io, in, di, uo) is det.
+
+report_mode_error_par_conj(ModeInfo, ErrorList) -->
+	{ mode_info_get_context(ModeInfo, Context) },
+	mode_info_write_context(ModeInfo),
+	prog_out__write_context(Context),
+	io__write_string("  mode error: mutually exclusive bindings in parallel conjunction.\n"),
+	mode_info_write_context(ModeInfo),
+	prog_out__write_context(Context),
+	io__write_string("              (The current implementation does not permit\n"),
+	mode_info_write_context(ModeInfo),
+	prog_out__write_context(Context),
+	io__write_string("              parallel conjunctions to fail.)\n"),
+	write_merge_error_list(ErrorList, ModeInfo).
+
 :- pred write_merge_error_list(merge_errors, mode_info, io__state, io__state).
 :- mode write_merge_error_list(in, mode_info_no_io, di, uo) is det.
 
@@ -325,18 +407,33 @@ write_merge_context(if_then_else) -->
 
 %-----------------------------------------------------------------------------%
 
-:- pred report_mode_error_bind_var(mode_info, var, inst, inst,
-					io__state, io__state).
-:- mode report_mode_error_bind_var(mode_info_ui, in, in, in, di, uo) is det.
+:- pred report_mode_error_bind_var(mode_info, var_lock_reason, prog_var,
+		inst, inst, io__state, io__state).
+:- mode report_mode_error_bind_var(mode_info_ui, in, in, in, in, di, uo) is det.
 
-report_mode_error_bind_var(ModeInfo, Var, VarInst, Inst) -->
+report_mode_error_bind_var(ModeInfo, Reason, Var, VarInst, Inst) -->
 	{ mode_info_get_context(ModeInfo, Context) },
 	{ mode_info_get_varset(ModeInfo, VarSet) },
 	{ mode_info_get_instvarset(ModeInfo, InstVarSet) },
 	mode_info_write_context(ModeInfo),
 	prog_out__write_context(Context),
-	io__write_string(
-		"  scope error: attempt to bind variable inside a negation.\n"),
+	io__write_string("  scope error: "),
+	( { Reason = negation },
+		io__write_string("attempt to bind a variable inside a negation.\n")
+	; { Reason = if_then_else },
+		io__write_string("attempt to bind a non-local variable\n"),
+		prog_out__write_context(Context),
+		io__write_string("  inside the condition of an if-then-else.\n")
+	; { Reason = lambda(PredOrFunc) },
+		{ hlds_out__pred_or_func_to_str(PredOrFunc, PredOrFuncS) },
+		io__write_string("attempt to bind a non-local variable inside\n"),
+		prog_out__write_context(Context),
+		io__write_strings(["  a ", PredOrFuncS, " lambda goal.\n"])
+	; { Reason = par_conj },
+		io__write_string("attempt to bind a non-local variable\n"),
+		prog_out__write_context(Context),
+		io__write_string("  inside more than one parallel conjunct.\n")
+	),
 	prog_out__write_context(Context),
 	io__write_string("  Variable `"),
 	mercury_output_var(Var, VarSet, no),
@@ -349,27 +446,58 @@ report_mode_error_bind_var(ModeInfo, Var, VarInst, Inst) -->
 	io__write_string("'.\n"),
 	globals__io_lookup_bool_option(verbose_errors, VerboseErrors),
 	( { VerboseErrors = yes } ->
-		io__write_string("\tA negation is only allowed to bind variables which are local to the\n"),
-		io__write_string("\tnegation, i.e. those which are implicitly existentially quantified\n"),
-		io__write_string("\tinside the scope of the negation.\n"),
-		io__write_string("\tNote that the condition of an if-then-else is implicitly\n"),
-		io__write_string("\tnegated in the ""else"" part, so the condition can only bind\n"),
-		io__write_string("\tvariables in the ""then"" part.\n")
+		( { Reason = negation },
+			io__write_string("\tA negation is only allowed to bind variables which are local to the\n"),
+			io__write_string("\tnegation, i.e. those which are implicitly existentially quantified\n"),
+			io__write_string("\tinside the scope of the negation.\n")
+		; { Reason = if_then_else },
+			io__write_string("\tThe condition of an if-then-else is only allowed\n"),
+			io__write_string("\tto bind variables which are local to the condition\n"),
+			io__write_string("\tor which occur only in the condition and the `then' part.\n")
+		; { Reason = lambda(_) },
+			io__write_string("\tA lambda goal is only allowed to bind its arguments\n"),
+			io__write_string("\tand variables local to the lambda expression.\n")
+		; { Reason = par_conj },
+			io__write_string("\tA nonlocal variable of a parallel conjunction may be\n"),
+			io__write_string("\tbound in at most one conjunct.\n")
+		)
 	;
 		[]
 	).
 
 %-----------------------------------------------------------------------------%
 
-:- pred report_mode_error_no_matching_mode(mode_info, list(var), list(inst),
+:- pred report_mode_error_non_local_lambda_var(mode_info, prog_var, inst,
 					io__state, io__state).
+:- mode report_mode_error_non_local_lambda_var(mode_info_ui, in, in,
+					di, uo) is det.
+
+report_mode_error_non_local_lambda_var(ModeInfo, Var, VarInst) -->
+	{ mode_info_get_context(ModeInfo, Context) },
+	{ mode_info_get_varset(ModeInfo, VarSet) },
+	{ mode_info_get_instvarset(ModeInfo, InstVarSet) },
+	mode_info_write_context(ModeInfo),
+	prog_out__write_context(Context),
+	io__write_string("  mode error: variable `"),
+	mercury_output_var(Var, VarSet, no),
+	io__write_string("' has instantiatedness `"),
+	output_inst(VarInst, InstVarSet),
+	io__write_string("',\n"),
+	prog_out__write_context(Context),
+	io__write_string("  expected instantiatedness for non-local variables\n"),
+	prog_out__write_context(Context),
+	io__write_string("  of lambda goals is `ground'.\n").
+
+%-----------------------------------------------------------------------------%
+
+:- pred report_mode_error_no_matching_mode(mode_info, list(prog_var),
+		list(inst), io__state, io__state).
 :- mode report_mode_error_no_matching_mode(mode_info_ui, in, in, di, uo) is det.
 
 report_mode_error_no_matching_mode(ModeInfo, Vars, Insts) -->
 	{ mode_info_get_context(ModeInfo, Context) },
 	{ mode_info_get_varset(ModeInfo, VarSet) },
 	{ mode_info_get_instvarset(ModeInfo, InstVarSet) },
-	{ mode_info_get_module_info(ModeInfo, ModuleInfo) },
 	mode_info_write_context(ModeInfo),
 	prog_out__write_context(Context),
 	io__write_string("  mode error: arguments `"),
@@ -382,16 +510,15 @@ report_mode_error_no_matching_mode(ModeInfo, Vars, Insts) -->
 	prog_out__write_context(Context),
 	io__write_string("  which does not match any of the modes for "),
 	{ mode_info_get_mode_context(ModeInfo, ModeContext) },
-	( { ModeContext = call(PredId, _) } ->
-		hlds_out__write_pred_id(ModuleInfo, PredId)
+	( { ModeContext = call(CallId, _) } ->
+		hlds_out__write_call_id(CallId)
 	;
 		{ error("report_mode_error_no_matching_mode: invalid context") }
 	),
 	io__write_string(".\n").
 
-:- pred report_mode_error_higher_order_pred_var(mode_info, pred_or_func, var,
-					inst, arity,
-					io__state, io__state).
+:- pred report_mode_error_higher_order_pred_var(mode_info, pred_or_func,
+		prog_var, inst, arity, io__state, io__state).
 :- mode report_mode_error_higher_order_pred_var(mode_info_ui, in, in, in, in,
 					di, uo) is det.
 
@@ -418,7 +545,39 @@ report_mode_error_higher_order_pred_var(ModeInfo, PredOrFunc, Var, VarInst,
 	),
 	io__write_string(").\n").
 
-:- pred report_mode_error_var_is_live(mode_info, var, io__state, io__state).
+:- pred report_mode_error_poly_unify(mode_info, prog_var, inst,
+					io__state, io__state).
+:- mode report_mode_error_poly_unify(mode_info_ui, in, in, di, uo) is det.
+
+report_mode_error_poly_unify(ModeInfo, Var, VarInst) -->
+	{ mode_info_get_context(ModeInfo, Context) },
+	{ mode_info_get_varset(ModeInfo, VarSet) },
+	{ mode_info_get_instvarset(ModeInfo, InstVarSet) },
+	mode_info_write_context(ModeInfo),
+	prog_out__write_context(Context),
+	io__write_string("  in polymorphically-typed unification:\n"),
+	prog_out__write_context(Context),
+	io__write_string("  mode error: variable `"),
+	mercury_output_var(Var, VarSet, no),
+	io__write_string("' has instantiatedness `"),
+	output_inst(VarInst, InstVarSet),
+	io__write_string("',\n"),
+	prog_out__write_context(Context),
+	io__write_string(
+		"  expected instantiatedness was `ground' or `any'.\n"),
+	globals__io_lookup_bool_option(verbose_errors, VerboseErrors),
+	( { VerboseErrors = yes } ->
+		io__write_string(
+"\tWhen unifying two variables whose type will not be known until
+\truntime, the variables must both be ground (or have inst `any').
+\tUnifications of polymorphically-typed variables with partially
+\tinstantiated modes are not allowed.\n")
+	;
+		[]
+	).
+
+:- pred report_mode_error_var_is_live(mode_info, prog_var,
+		io__state, io__state).
 :- mode report_mode_error_var_is_live(mode_info_ui, in, di, uo) is det.
 
 report_mode_error_var_is_live(ModeInfo, Var) -->
@@ -432,7 +591,7 @@ report_mode_error_var_is_live(ModeInfo, Var) -->
 	mercury_output_var(Var, VarSet, no),
 	io__write_string("' is still live.\n").
 
-:- pred report_mode_error_var_has_inst(mode_info, var, inst, inst,
+:- pred report_mode_error_var_has_inst(mode_info, prog_var, inst, inst,
 					io__state, io__state).
 :- mode report_mode_error_var_has_inst(mode_info_ui, in, in, in, di, uo) is det.
 
@@ -452,7 +611,7 @@ report_mode_error_var_has_inst(ModeInfo, Var, VarInst, Inst) -->
 	output_inst(Inst, InstVarSet),
 	io__write_string("'.\n").
 
-:- pred report_mode_error_implied_mode(mode_info, var, inst, inst,
+:- pred report_mode_error_implied_mode(mode_info, prog_var, inst, inst,
 					io__state, io__state).
 :- mode report_mode_error_implied_mode(mode_info_ui, in, in, in, di, uo) is det.
 
@@ -491,7 +650,7 @@ report_mode_error_no_mode_decl(ModeInfo) -->
 	prog_out__write_context(Context),
 	io__write_string("  no mode declaration for called predicate.\n").
 
-:- pred report_mode_error_unify_pred(mode_info, var, mode_error_unify_rhs,
+:- pred report_mode_error_unify_pred(mode_info, prog_var, mode_error_unify_rhs,
 					type, pred_or_func,
 					io__state, io__state).
 :- mode report_mode_error_unify_pred(mode_info_ui, in, in, in, in,
@@ -500,6 +659,7 @@ report_mode_error_no_mode_decl(ModeInfo) -->
 report_mode_error_unify_pred(ModeInfo, X, RHS, Type, PredOrFunc) -->
 	{ mode_info_get_context(ModeInfo, Context) },
 	{ mode_info_get_varset(ModeInfo, VarSet) },
+	{ mode_info_get_instvarset(ModeInfo, InstVarSet) },
 	mode_info_write_context(ModeInfo),
 	prog_out__write_context(Context),
 	io__write_string("  In unification of `"),
@@ -510,11 +670,14 @@ report_mode_error_unify_pred(ModeInfo, X, RHS, Type, PredOrFunc) -->
 		mercury_output_var(Y, VarSet, no)
 	;
 		{ RHS = error_at_functor(ConsId, ArgVars) },
-		hlds_out__write_functor_cons_id(ConsId, ArgVars, VarSet, no)
+		{ mode_info_get_module_info(ModeInfo, ModuleInfo) },
+		hlds_out__write_functor_cons_id(ConsId, ArgVars, VarSet,
+			ModuleInfo, no)
 	;
 		{ RHS = error_at_lambda(ArgVars, ArgModes) },
 		io__write_string("lambda(["),
-		hlds_out__write_var_modes(ArgVars, ArgModes, VarSet, no),
+		hlds_out__write_var_modes(ArgVars, ArgModes, VarSet,
+			InstVarSet, no),
 		io__write_string("] ... )")
 	),
 	io__write_string("':\n"),
@@ -545,8 +708,8 @@ report_mode_error_unify_pred(ModeInfo, X, RHS, Type, PredOrFunc) -->
 
 %-----------------------------------------------------------------------------%
 
-:- pred report_mode_error_unify_var_var(mode_info, var, var, inst, inst,
-					io__state, io__state).
+:- pred report_mode_error_unify_var_var(mode_info, prog_var, prog_var,
+		inst, inst, io__state, io__state).
 :- mode report_mode_error_unify_var_var(mode_info_ui, in, in, in, in, di, uo)
 	is det.
 
@@ -576,7 +739,7 @@ report_mode_error_unify_var_var(ModeInfo, X, Y, InstX, InstY) -->
 
 %-----------------------------------------------------------------------------%
 
-:- pred report_mode_error_unify_var_lambda(mode_info, var, inst, inst,
+:- pred report_mode_error_unify_var_lambda(mode_info, prog_var, inst, inst,
 					io__state, io__state).
 :- mode report_mode_error_unify_var_lambda(mode_info_ui, in, in, in, di, uo)
 	is det.
@@ -603,8 +766,8 @@ report_mode_error_unify_var_lambda(ModeInfo, X, InstX, InstY) -->
 
 %-----------------------------------------------------------------------------%
 
-:- pred report_mode_error_unify_var_functor(mode_info, var, cons_id, list(var),
-					inst, list(inst), io__state, io__state).
+:- pred report_mode_error_unify_var_functor(mode_info, prog_var, cons_id,
+		list(prog_var), inst, list(inst), io__state, io__state).
 :- mode report_mode_error_unify_var_functor(mode_info_ui, in, in, in, in, in,
 			di, uo) is det.
 
@@ -613,12 +776,13 @@ report_mode_error_unify_var_functor(ModeInfo, X, ConsId, Args, InstX, ArgInsts)
 	{ mode_info_get_context(ModeInfo, Context) },
 	{ mode_info_get_varset(ModeInfo, VarSet) },
 	{ mode_info_get_instvarset(ModeInfo, InstVarSet) },
+	{ mode_info_get_module_info(ModeInfo, ModuleInfo) },
 	mode_info_write_context(ModeInfo),
 	prog_out__write_context(Context),
 	io__write_string("  mode error in unification of `"),
 	mercury_output_var(X, VarSet, no),
 	io__write_string("' and `"),
-	hlds_out__write_functor_cons_id(ConsId, Args, VarSet, no),
+	hlds_out__write_functor_cons_id(ConsId, Args, VarSet, ModuleInfo, no),
 	io__write_string("'.\n"),
 	prog_out__write_context(Context),
 	io__write_string("  Variable `"),
@@ -628,18 +792,18 @@ report_mode_error_unify_var_functor(ModeInfo, X, ConsId, Args, InstX, ArgInsts)
 	io__write_string("',\n"),
 	prog_out__write_context(Context),
 	io__write_string("  term `"),
-	hlds_out__write_functor_cons_id(ConsId, Args, VarSet, no),
+	hlds_out__write_functor_cons_id(ConsId, Args, VarSet, ModuleInfo, no),
 	( { Args \= [] } ->
 		io__write_string("'\n"),
 		prog_out__write_context(Context),
 		io__write_string("  has instantiatedness `"),
-		mercury_output_cons_id(ConsId, no),
+		mercury_output_cons_id(ConsId, does_not_need_brackets),
 		io__write_string("("),
 		output_inst_list(ArgInsts, InstVarSet),
 		io__write_string(")")
 	;
 		io__write_string("' has instantiatedness `"),
-		mercury_output_cons_id(ConsId, no)
+		mercury_output_cons_id(ConsId, does_not_need_brackets)
 	),
 	io__write_string("'.\n").
 
@@ -667,21 +831,15 @@ mode_info_write_context(ModeInfo) -->
 
 	prog_out__write_context(Context),
 	io__write_string("In clause for `"),
-	(	{ PredOrFunc = predicate },
-		mercury_output_pred_mode_subdecl(InstVarSet, Name, Modes,
-				MaybeDet, Context)
-	;	{ PredOrFunc = function },
-		{ pred_args_to_func_args(Modes, ArgModes, RetMode) },
-		mercury_output_func_mode_subdecl(InstVarSet, Name, ArgModes,
-				RetMode, MaybeDet, Context)
-	),
+	mercury_output_mode_subdecl(PredOrFunc, InstVarSet, Name, Modes,
+				MaybeDet, Context),
 	io__write_string("':\n"),
 	{ mode_info_get_mode_context(ModeInfo, ModeContext) },
 	write_mode_context(ModeContext, Context, ModuleInfo).
 
 %-----------------------------------------------------------------------------%
 
-:- pred report_mode_error_final_inst(mode_info, int, var, inst, inst,
+:- pred report_mode_error_final_inst(mode_info, int, prog_var, inst, inst,
 				final_inst_error, io__state, io__state).
 :- mode report_mode_error_final_inst(mode_info_ui, in, in, in, in, in,
 				di, uo) is det.
@@ -717,7 +875,6 @@ report_mode_error_final_inst(ModeInfo, ArgNum, Var, VarInst, Inst, Reason) -->
 	io__write_string("'.\n").
 
 
-%-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
 mode_context_init(uninitialized).
@@ -726,55 +883,17 @@ mode_context_init(uninitialized).
 
 	% XXX some parts of the mode context never get set up
 
-:- pred write_mode_context(mode_context, term__context, module_info,
+:- pred write_mode_context(mode_context, prog_context, module_info,
 				io__state, io__state).
 :- mode write_mode_context(in, in, in, di, uo) is det.
 
 write_mode_context(uninitialized, _Context, _ModuleInfo) -->
 	[].
 
-write_mode_context(higher_order_call(PredOrFunc, ArgNum), Context, _ModuleInfo)
-		-->
+write_mode_context(call(CallId, ArgNum), Context, _ModuleInfo) -->
 	prog_out__write_context(Context),
 	io__write_string("  in "),
-	( { ArgNum = 0 } ->
-		io__write_string("higher-order "),
-		hlds_out__write_pred_or_func(PredOrFunc),
-		io__write_string(" call:\n")
-	;
-		io__write_string("argument "),
-		io__write_int(ArgNum),
-		io__write_string(" of higher-order "),
-		hlds_out__write_pred_or_func(PredOrFunc),
-		io__write_string(" call\n"),
-		prog_out__write_context(Context),
-		io__write_string("  (i.e. in "),
-		( { ArgNum = 1 } ->
-			io__write_string("the "),
-			hlds_out__write_pred_or_func(PredOrFunc),
-			io__write_string(" term")
-		;
-			io__write_string("argument "),
-			{ ArgNum1 is ArgNum - 1 },
-			io__write_int(ArgNum1),
-			io__write_string(" of the called "),
-			hlds_out__write_pred_or_func(PredOrFunc)
-		),
-		io__write_string("):\n")
-	).
-
-write_mode_context(call(PredId, ArgNum), Context, ModuleInfo) -->
-	prog_out__write_context(Context),
-	io__write_string("  in "),
-	( { ArgNum = 0 } ->
-		[]
-	;
-		io__write_string("argument "),
-		io__write_int(ArgNum),
-		io__write_string(" of ")
-	),
-	io__write_string("call to "),
-	hlds_out__write_pred_id(ModuleInfo, PredId),
+	hlds_out__write_call_arg_id(CallId, ArgNum),
 	io__write_string(":\n").
 
 write_mode_context(unify(UnifyContext, _Side), Context, _ModuleInfo) -->
@@ -821,55 +940,75 @@ maybe_report_error_no_modes(PredId, PredInfo, ModuleInfo) -->
 
 	% write out the inferred `mode' declarations for a list of pred_ids.
 
-write_mode_inference_messages([], _) --> [].
-write_mode_inference_messages([PredId | PredIds], ModuleInfo) -->
+write_mode_inference_messages([], _, _) --> [].
+write_mode_inference_messages([PredId | PredIds], OutputDetism, ModuleInfo) -->
 	{ module_info_pred_info(ModuleInfo, PredId, PredInfo) },
-	{ pred_info_get_marker_list(PredInfo, Markers) },
-	( { list__member(request(infer_modes), Markers) } ->
+	{ pred_info_get_markers(PredInfo, Markers) },
+	( { check_marker(Markers, infer_modes) } ->
 		{ pred_info_procedures(PredInfo, Procs) },
 		{ map__keys(Procs, ProcIds) },
-		write_mode_inference_messages_2(ProcIds, Procs, PredInfo)
+		write_mode_inference_messages_2(ProcIds, Procs, PredInfo,
+			OutputDetism)
 	;
 		[]
 	),
-	write_mode_inference_messages(PredIds, ModuleInfo).
+	write_mode_inference_messages(PredIds, OutputDetism, ModuleInfo).
 
 	% write out the inferred `mode' declarations for a list of
 	% proc_ids
 
 :- pred write_mode_inference_messages_2(list(proc_id), proc_table, pred_info,
-				io__state, io__state).
-:- mode write_mode_inference_messages_2(in, in, in, di, uo) is det.
+				bool, io__state, io__state).
+:- mode write_mode_inference_messages_2(in, in, in, in, di, uo) is det.
 
-write_mode_inference_messages_2([], _, _) --> [].
-write_mode_inference_messages_2([ProcId | ProcIds], Procs, PredInfo) -->
+write_mode_inference_messages_2([], _, _, _) --> [].
+write_mode_inference_messages_2([ProcId | ProcIds], Procs, PredInfo,
+		OutputDetism) -->
 	{ map__lookup(Procs, ProcId, ProcInfo) },
-	write_mode_inference_message(PredInfo, ProcInfo),
-	write_mode_inference_messages_2(ProcIds, Procs, PredInfo).
+	write_mode_inference_message(PredInfo, ProcInfo, OutputDetism),
+	write_mode_inference_messages_2(ProcIds, Procs, PredInfo, OutputDetism).
 
 	% write out the inferred `mode' declaration
 	% for a single function or predicate.
 
-:- pred write_mode_inference_message(pred_info, proc_info,
+:- pred write_mode_inference_message(pred_info, proc_info, bool,
 				io__state, io__state).
-:- mode write_mode_inference_message(in, in, di, uo) is det.
+:- mode write_mode_inference_message(in, in, in, di, uo) is det.
 
-write_mode_inference_message(PredInfo, ProcInfo) -->
+write_mode_inference_message(PredInfo, ProcInfo, OutputDetism) -->
 	{ pred_info_name(PredInfo, PredName) },
 	{ Name = unqualified(PredName) },
 	{ pred_info_context(PredInfo, Context) },
-	{ proc_info_argmodes(ProcInfo, Modes0) },
+	{ pred_info_arity(PredInfo, PredArity) },
+	{ proc_info_argmodes(ProcInfo, ArgModes0) },
+
+	% We need to strip off the extra type_info arguments inserted at the
+	% front by polymorphism.m - we only want the last `PredArity' of them.
+	%
+	{ list__length(ArgModes0, NumArgModes) },
+	{ NumToDrop is NumArgModes - PredArity },
+	( { list__drop(NumToDrop, ArgModes0, ArgModes1) } ->
+		{ ArgModes2 = ArgModes1 }
+	;
+		{ error("report_pred_proc_id: list__drop failed") }
+	),
+
 	{ varset__init(VarSet) },
 	{ pred_info_get_is_pred_or_func(PredInfo, PredOrFunc) },
-	{ MaybeDet = no },
+	( { OutputDetism = yes } ->
+		{ proc_info_inferred_determinism(ProcInfo, Detism) },
+		{ MaybeDet = yes(Detism) }
+	;
+		{ MaybeDet = no }
+	),
 	prog_out__write_context(Context),
-	{ strip_builtin_qualifiers_from_mode_list(Modes0, Modes) },
+	{ strip_builtin_qualifiers_from_mode_list(ArgModes2, ArgModes3) },
 	io__write_string("Inferred "),
 	(	{ PredOrFunc = predicate },
-		mercury_output_pred_mode_decl(VarSet, Name, Modes,
+		mercury_output_pred_mode_decl(VarSet, Name, ArgModes3,
 				MaybeDet, Context)
 	;	{ PredOrFunc = function },
-		{ pred_args_to_func_args(Modes, ArgModes, RetMode) },
+		{ pred_args_to_func_args(ArgModes3, ArgModes, RetMode) },
 		mercury_output_func_mode_decl(VarSet, Name, ArgModes, RetMode,
 				MaybeDet, Context)
 	).
@@ -897,19 +1036,79 @@ report_mode_errors(ModeInfo0, ModeInfo) :-
 
 %-----------------------------------------------------------------------------%
 
-:- pred output_inst((inst), varset, io__state, io__state).
+:- pred output_inst((inst), inst_varset, io__state, io__state).
 :- mode output_inst(in, in, di, uo) is det.
 
 output_inst(Inst0, VarSet) -->
 	{ strip_builtin_qualifiers_from_inst(Inst0, Inst) },
 	mercury_output_inst(Inst, VarSet).
 
-:- pred output_inst_list(list(inst), varset, io__state, io__state).
+:- pred output_inst_list(list(inst), inst_varset, io__state, io__state).
 :- mode output_inst_list(in, in, di, uo) is det.
 
 output_inst_list(Insts0, VarSet) -->
 	{ strip_builtin_qualifiers_from_inst_list(Insts0, Insts) },
 	mercury_output_inst_list(Insts, VarSet).
 
+%-----------------------------------------------------------------------------%
+
+report_indistinguishable_modes_error(OldProcId, NewProcId,
+		PredId, PredInfo, ModuleInfo) -->
+
+	io__set_exit_status(1),
+
+	{ pred_info_procedures(PredInfo, Procs) },
+	{ map__lookup(Procs, OldProcId, OldProcInfo) },
+	{ map__lookup(Procs, NewProcId, NewProcInfo) },
+	{ proc_info_context(OldProcInfo, OldContext) },
+	{ proc_info_context(NewProcInfo, NewContext) },
+
+	prog_out__write_context(NewContext),
+	io__write_string("In mode declarations for "),
+	hlds_out__write_pred_id(ModuleInfo, PredId),
+	io__write_string(":\n"),
+
+	prog_out__write_context(NewContext),
+	io__write_string("  error: duplicate mode declaration.\n"),
+
+	globals__io_lookup_bool_option(verbose_errors, VerboseErrors),
+	( { VerboseErrors = yes } ->
+		prog_out__write_context(NewContext),
+		io__write_string("  Modes `"),
+		output_mode_decl(OldProcId, PredInfo),
+		io__write_string("'\n"),
+
+		prog_out__write_context(NewContext),
+		io__write_string("  and `"),
+		output_mode_decl(NewProcId, PredInfo),
+		io__write_string("'\n"),
+
+		prog_out__write_context(NewContext),
+		io__write_string("  are indistinguishable.\n")
+	;
+		[]
+	),
+
+	prog_out__write_context(OldContext),
+	io__write_string(
+		"  Here is the conflicting mode declaration.\n").
+
+:- pred output_mode_decl(proc_id, pred_info, io__state, io__state).
+:- mode output_mode_decl(in, in, di, uo) is det.
+
+output_mode_decl(ProcId, PredInfo) -->
+	{ pred_info_get_is_pred_or_func(PredInfo, PredOrFunc) },
+	{ pred_info_name(PredInfo, Name0) },
+	{ pred_info_procedures(PredInfo, Procs) },
+	{ map__lookup(Procs, ProcId, ProcInfo) },
+	{ proc_info_declared_argmodes(ProcInfo, Modes0) },
+	{ proc_info_declared_determinism(ProcInfo, MaybeDet) },
+	{ proc_info_context(ProcInfo, Context) },
+	{ varset__init(InstVarSet) },
+	{ Name = unqualified(Name0) },
+	{ strip_builtin_qualifiers_from_mode_list(Modes0, Modes) },
+	mercury_output_mode_subdecl(PredOrFunc, InstVarSet, Name, Modes,
+				MaybeDet, Context).
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
