@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -17,8 +17,8 @@
 :- interface.
 
 :- import_module hlds_module, hlds_pred, hlds_goal, hlds_data, instmap.
-:- import_module mode_errors, delay_info, (inst).
-:- import_module map, list, varset, set, bool, term, assoc_list.
+:- import_module prog_data, mode_errors, delay_info, (inst).
+:- import_module map, list, set, bool, assoc_list, std_util.
 
 :- interface.
 
@@ -27,13 +27,11 @@
 	% XXX `side' is not used
 :- type mode_context
 	--->	call(	
-			pred_id,	% pred name / arity
-			int		% argument number
-		)
-	;	higher_order_call(
-			pred_or_func,	% is it call/N (higher-order pred call)
-					% or apply/N (higher-order func call)?
-			int		% argument number
+			call_id,
+			int		% argument number (offset so that
+					% the real arguments start at number 1
+					% whereas the type_info arguments
+					% have numbers <= 0).
 		)
 	;	unify(
 			unify_context,	% original source of the unification
@@ -53,14 +51,42 @@
 
 :- type call_context
 	--->	unify(unify_context)
-	;	call(pred_id)
-	;	higher_order_call(pred_or_func).
+	;	call(call_id).
+
+:- type var_lock_reason
+	--->	negation
+	;	if_then_else
+	;	lambda(pred_or_func)
+	;	par_conj
+	.
+
+	% Specify how to process goals - using either
+	% modes.m or unique_modes.m.
+:- type how_to_check_goal
+	--->    check_modes
+	;       check_unique_modes
+	.
+	
+	% Is mode analysis allowed to change which procedure of a predicate
+	% is called. It may not change the called procedure after deforestation
+	% has performed a generalisation step, since that could result
+	% in selecting a less efficient mode, or one which doesn't even
+	% have the same termination behaviour.
+	% Also, when rechecking a goal after adding extra goals, it is
+	% not necessary to choose again which procedure is to be called.
+:- type may_change_called_proc
+	--->	may_change_called_proc
+	;	may_not_change_called_proc.
+
+:- type locked_vars == assoc_list(var_lock_reason, set(prog_var)).
 
 :- type mode_info.
 
-:- pred mode_info_init(io__state, module_info, pred_id, proc_id,
-			term__context, set(var), instmap, mode_info).
-:- mode mode_info_init(di, in, in, in, in, in, in, mode_info_uo) is det.
+:- pred mode_info_init(io__state, module_info, pred_id, proc_id, prog_context,
+		set(prog_var), instmap, how_to_check_goal,
+		may_change_called_proc, mode_info).
+:- mode mode_info_init(di, in, in, in, in, in, in, in, in,
+		mode_info_uo) is det.
 
 :- pred mode_info_get_io_state(mode_info, io__state).
 :- mode mode_info_get_io_state(mode_info_get_io_state, uo) is det.
@@ -89,10 +115,10 @@
 :- pred mode_info_get_procid(mode_info, proc_id).
 :- mode mode_info_get_procid(mode_info_ui, out) is det.
 
-:- pred mode_info_get_context(mode_info, term__context).
+:- pred mode_info_get_context(mode_info, prog_context).
 :- mode mode_info_get_context(mode_info_ui, out) is det.
 
-:- pred mode_info_set_context(term__context, mode_info, mode_info).
+:- pred mode_info_set_context(prog_context, mode_info, mode_info).
 :- mode mode_info_set_context(in, mode_info_di, mode_info_uo) is det.
 
 :- pred mode_info_get_mode_context(mode_info, mode_context).
@@ -119,10 +145,10 @@
 :- pred mode_info_set_instmap(instmap, mode_info, mode_info).
 :- mode mode_info_set_instmap(in, mode_info_di, mode_info_uo) is det.
 
-:- pred mode_info_get_locked_vars(mode_info, list(set(var))).
+:- pred mode_info_get_locked_vars(mode_info, locked_vars).
 :- mode mode_info_get_locked_vars(mode_info_ui, out) is det.
 
-:- pred mode_info_set_locked_vars(mode_info, list(set(var)), mode_info).
+:- pred mode_info_set_locked_vars(mode_info, locked_vars, mode_info).
 :- mode mode_info_set_locked_vars(mode_info_di, in, mode_info_uo) is det.
 
 :- pred mode_info_get_errors(mode_info, list(mode_error_info)).
@@ -134,56 +160,59 @@
 :- pred mode_info_set_errors(list(mode_error_info), mode_info, mode_info).
 :- mode mode_info_set_errors(in, mode_info_di, mode_info_uo) is det.
 
-:- pred mode_info_add_live_vars(set(var), mode_info, mode_info).
+:- pred mode_info_add_live_vars(set(prog_var), mode_info, mode_info).
 :- mode mode_info_add_live_vars(in, mode_info_di, mode_info_uo) is det.
 
-:- pred mode_info_remove_live_vars(set(var), mode_info, mode_info).
+:- pred mode_info_remove_live_vars(set(prog_var), mode_info, mode_info).
 :- mode mode_info_remove_live_vars(in, mode_info_di, mode_info_uo) is det.
 
-:- pred mode_info_var_list_is_live(list(var), mode_info, list(is_live)).
+:- pred mode_info_var_list_is_live(list(prog_var), mode_info, list(is_live)).
 :- mode mode_info_var_list_is_live(in, mode_info_ui, out) is det.
 
-:- pred mode_info_var_is_live(mode_info, var, is_live).
+:- pred mode_info_var_is_live(mode_info, prog_var, is_live).
 :- mode mode_info_var_is_live(mode_info_ui, in, out) is det.
 
-:- pred mode_info_var_is_nondet_live(mode_info, var, is_live).
+:- pred mode_info_var_is_nondet_live(mode_info, prog_var, is_live).
 :- mode mode_info_var_is_nondet_live(mode_info_ui, in, out) is det.
 
-:- pred mode_info_get_liveness(mode_info, set(var)).
+:- pred mode_info_get_liveness(mode_info, set(prog_var)).
 :- mode mode_info_get_liveness(mode_info_ui, out) is det.
 
-:- pred mode_info_get_liveness_2(list(set(var)), set(var), set(var)).
+:- pred mode_info_get_liveness_2(list(set(prog_var)), set(prog_var),
+		set(prog_var)).
 :- mode mode_info_get_liveness_2(in, in, out) is det.
 
-:- pred mode_info_get_varset(mode_info, varset).
+:- pred mode_info_get_varset(mode_info, prog_varset).
 :- mode mode_info_get_varset(mode_info_ui, out) is det.
 
-:- pred mode_info_set_varset(varset, mode_info, mode_info).
+:- pred mode_info_set_varset(prog_varset, mode_info, mode_info).
 :- mode mode_info_set_varset(in, mode_info_di, mode_info_uo) is det.
 
-:- pred mode_info_get_instvarset(mode_info, varset).
+:- pred mode_info_get_instvarset(mode_info, inst_varset).
 :- mode mode_info_get_instvarset(mode_info_ui, out) is det.
 
-:- pred mode_info_get_var_types(mode_info, map(var,type)).
+:- pred mode_info_get_var_types(mode_info, map(prog_var, type)).
 :- mode mode_info_get_var_types(mode_info_ui, out) is det.
 
-:- pred mode_info_set_var_types(map(var, type), mode_info, mode_info).
+:- pred mode_info_set_var_types(map(prog_var, type), mode_info, mode_info).
 :- mode mode_info_set_var_types(in, mode_info_di, mode_info_uo) is det.
 
-:- pred mode_info_get_types_of_vars(mode_info, list(var), list(type)).
+:- pred mode_info_get_types_of_vars(mode_info, list(prog_var), list(type)).
 :- mode mode_info_get_types_of_vars(mode_info_ui, in, out) is det.
 
-:- pred mode_info_lock_vars(set(var), mode_info, mode_info).
-:- mode mode_info_lock_vars(in, mode_info_di, mode_info_uo) is det.
+:- pred mode_info_lock_vars(var_lock_reason, set(prog_var),
+		mode_info, mode_info).
+:- mode mode_info_lock_vars(in, in, mode_info_di, mode_info_uo) is det.
 
-:- pred mode_info_unlock_vars(set(var), mode_info, mode_info).
-:- mode mode_info_unlock_vars(in, mode_info_di, mode_info_uo) is det.
+:- pred mode_info_unlock_vars(var_lock_reason, set(prog_var),
+		mode_info, mode_info).
+:- mode mode_info_unlock_vars(in, in, mode_info_di, mode_info_uo) is det.
 
-:- pred mode_info_var_is_locked(mode_info, var).
-:- mode mode_info_var_is_locked(mode_info_ui, in) is semidet.
+:- pred mode_info_var_is_locked(mode_info, prog_var, var_lock_reason).
+:- mode mode_info_var_is_locked(mode_info_ui, in, out) is semidet.
 
-:- pred mode_info_var_is_locked_2(list(set(var)), var).
-:- mode mode_info_var_is_locked_2(in, in) is semidet.
+:- pred mode_info_var_is_locked_2(locked_vars, prog_var, var_lock_reason).
+:- mode mode_info_var_is_locked_2(in, in, out) is semidet.
 
 :- pred mode_info_get_delay_info(mode_info, delay_info).
 :- mode mode_info_get_delay_info(mode_info_no_io, out) is det.
@@ -191,26 +220,66 @@
 :- pred mode_info_set_delay_info(delay_info, mode_info, mode_info).
 :- mode mode_info_set_delay_info(in, mode_info_di, mode_info_uo) is det.
 
-:- pred mode_info_get_nondet_live_vars(mode_info, list(set(var))).
+:- pred mode_info_get_live_vars(mode_info, list(set(prog_var))).
+:- mode mode_info_get_live_vars(mode_info_ui, out) is det.
+
+:- pred mode_info_set_live_vars(list(set(prog_var)), mode_info, mode_info).
+:- mode mode_info_set_live_vars(in, mode_info_di, mode_info_uo) is det.
+
+:- pred mode_info_get_nondet_live_vars(mode_info, list(set(prog_var))).
 :- mode mode_info_get_nondet_live_vars(mode_info_no_io, out) is det.
 
-:- pred mode_info_set_nondet_live_vars(list(set(var)), mode_info, mode_info).
+:- pred mode_info_set_nondet_live_vars(list(set(prog_var)),
+		mode_info, mode_info).
 :- mode mode_info_set_nondet_live_vars(in, mode_info_di, mode_info_uo) is det.
 
-:- pred mode_info_get_last_checkpoint_insts(mode_info, assoc_list(var, inst)).
+:- pred mode_info_get_last_checkpoint_insts(mode_info,
+		assoc_list(prog_var, inst)).
 :- mode mode_info_get_last_checkpoint_insts(mode_info_no_io, out) is det.
 
-:- pred mode_info_set_last_checkpoint_insts(assoc_list(var, inst),
+:- pred mode_info_set_last_checkpoint_insts(assoc_list(prog_var, inst),
 	mode_info, mode_info).
 :- mode mode_info_set_last_checkpoint_insts(in, mode_info_di, mode_info_uo)
 	is det.
 
+:- pred mode_info_get_parallel_vars(list(pair(set(prog_var))), mode_info,
+		mode_info).
+:- mode mode_info_get_parallel_vars(out, mode_info_di, mode_info_uo) is det.
+
+:- pred mode_info_set_parallel_vars(list(pair(set(prog_var))), mode_info,
+		mode_info).
+:- mode mode_info_set_parallel_vars(in, mode_info_di, mode_info_uo) is det.
+
 :- pred mode_info_get_changed_flag(mode_info, bool).
 :- mode mode_info_get_changed_flag(mode_info_no_io, out) is det.
 
 :- pred mode_info_set_changed_flag(bool, mode_info, mode_info).
 :- mode mode_info_set_changed_flag(in, mode_info_di, mode_info_uo) is det.
 
+:- pred mode_info_get_how_to_check(mode_info, how_to_check_goal).
+:- mode mode_info_get_how_to_check(mode_info_ui, out) is det.
+
+:- pred mode_info_set_how_to_check(how_to_check_goal, mode_info, mode_info).
+:- mode mode_info_set_how_to_check(in, mode_info_di, mode_info_uo) is det.
+
+:- pred mode_info_get_may_change_called_proc(mode_info,
+		may_change_called_proc).
+:- mode mode_info_get_may_change_called_proc(mode_info_ui, out) is det.
+
+:- pred mode_info_set_may_change_called_proc(may_change_called_proc,
+		mode_info, mode_info).
+:- mode mode_info_set_may_change_called_proc(in,
+		mode_info_di, mode_info_uo) is det.
+
+:- pred mode_info_set_checking_extra_goals(bool, mode_info, mode_info).
+:- mode mode_info_set_checking_extra_goals(in,
+		mode_info_di, mode_info_uo) is det.
+
+	% Find the simple_call_id to use in error messages
+	% for the given pred_id.
+:- pred mode_info_get_call_id(mode_info, pred_id, simple_call_id).
+:- mode mode_info_get_call_id(mode_info_ui, in, out) is det.
+
 /*
 :- inst uniq_mode_info	=	bound_unique(
 					mode_info(
@@ -218,7 +287,8 @@
 						ground, ground, ground,
 						ground, ground, ground, ground,
 						ground, ground, ground, ground,
-						ground, ground, ground
+						ground, ground, ground, ground,
+						ground
 					)
 				).
 */
@@ -237,7 +307,8 @@
 						dead, ground, ground, ground,
 						ground, ground, ground, ground,
 						ground, ground, ground, ground,
-						ground, ground, ground
+						ground, ground, ground, ground,
+						ground
 					)
 				).
 */
@@ -251,7 +322,7 @@
 
         % record a mode error (and associated context info) in the mode_info.
 
-:- pred mode_info_error(set(var), mode_error, mode_info, mode_info).
+:- pred mode_info_error(set(prog_var), mode_error, mode_info, mode_info).
 :- mode mode_info_error(in, in, mode_info_di, mode_info_uo) is det.
 
 :- pred mode_info_add_error(mode_error_info, mode_info, mode_info).
@@ -262,7 +333,8 @@
 
 :- implementation.
 
-:- import_module delay_info, mode_errors, prog_data, mode_util.
+:- import_module delay_info, mode_errors, mode_util.
+:- import_module term, varset.
 :- import_module require, std_util, queue.
 
 :- type mode_info 
@@ -271,15 +343,16 @@
 			module_info,
 			pred_id,	% The pred we are checking
 			proc_id,	% The mode which we are checking
-			varset,		% The variables in the current proc
-			map(var, type),	% The types of the variables
-			term__context,	% The line number of the subgoal we
+			prog_varset,	% The variables in the current proc
+			map(prog_var, type),
+					% The types of the variables
+			prog_context,	% The line number of the subgoal we
 					% are currently checking
 			mode_context,	% A description of where in the
 					% goal the error occurred
 			instmap,	% The current instantiatedness
 					% of the variables
-			list(set(var)),	% The "locked" variables,
+			locked_vars,	% The "locked" variables,
 					% i.e. variables which cannot be
 					% further instantiated inside a
 					% negated context
@@ -287,14 +360,15 @@
 			list(mode_error_info),
 					% The mode errors found
 
-			list(set(var)),	% The live variables,
+			list(set(prog_var)),	% The live variables,
 	% i.e. those variables which may be referenced again on forward
 	% execution or after shallow backtracking.  (By shallow
 	% backtracking, I mean semidet backtracking in a negation,
 	% if-then-else, or semidet disjunction within the current
 	% predicate.)
 
-			list(set(var)),	% The nondet-live variables,
+			list(set(prog_var)),
+					% The nondet-live variables,
 	% i.e. those variables which may be referenced again after deep
 	% backtracking TO THE CURRENT EXECUTION POINT.  These are the
 	% variables which need to be made mostly_unique rather than
@@ -304,7 +378,7 @@
 	% execution point, since those variables will *already* have
 	% been marked as mostly_unique rather than unique.)
 
-			assoc_list(var, inst),
+			assoc_list(prog_var, inst),
 	% This field is used by the checkpoint code when debug_modes is on.
 	% It has the instmap that was current at the last mode checkpoint,
 	% so that checkpoints do not print out the insts of variables
@@ -312,9 +386,32 @@
 	% This field will always contain an empty list if debug_modes is off,
 	% since its information is not needed then.
 
-			bool		% Changed flag
+			list(pair(set(prog_var), set(prog_var))),
+	% A stack of pairs of sets of variables used to mode-check
+	% parallel conjunctions. The first set is the nonlocals of
+	% the parallel conjunction. The second set is a subset of the
+	% first, and is the set of variables that have been [further]
+	% bound inside the current parallel conjunct - the stack is for
+	% the correct handling of nested parallel conjunctions.
+
+			bool,		% Changed flag
 					% If `yes', then we may need
 					% to repeat mode inference.
+
+			how_to_check_goal,
+
+			may_change_called_proc,
+					% Is mode analysis allowed
+					% to change which procedure
+					% is called?
+
+			bool		% Are we rechecking a goal after
+					% introducing unifications for
+					% complicated sub-unifications
+					% or an implied mode?
+					% If so, redoing the mode check
+					% should not introduce more
+					% extra unifications.
 		).
 
 	% The normal inst of a mode_info struct: ground, with
@@ -326,7 +423,7 @@
 	% Initialize the mode_info
 
 mode_info_init(IOState, ModuleInfo, PredId, ProcId, Context,
-		LiveVars, InstMapping0, ModeInfo) :-
+		LiveVars, InstMapping0, HowToCheck, MayChangeProc, ModeInfo) :-
 	mode_context_init(ModeContext),
 	LockedVars = [],
 	delay_info__init(DelayInfo),
@@ -336,7 +433,7 @@ mode_info_init(IOState, ModuleInfo, PredId, ProcId, Context,
 	map__lookup(Preds, PredId, PredInfo),
 	pred_info_procedures(PredInfo, Procs),
 	map__lookup(Procs, ProcId, ProcInfo),
-	proc_info_variables(ProcInfo, VarSet),
+	proc_info_varset(ProcInfo, VarSet),
 	proc_info_vartypes(ProcInfo, VarTypes),
 
 	LiveVarsList = [LiveVars],
@@ -344,119 +441,139 @@ mode_info_init(IOState, ModuleInfo, PredId, ProcId, Context,
 
 	Changed = no,
 
+	CheckingExtraGoals = no,
+
 	ModeInfo = mode_info(
 		IOState, ModuleInfo, PredId, ProcId, VarSet, VarTypes,
 		Context, ModeContext, InstMapping0, LockedVars, DelayInfo,
-		ErrorList, LiveVarsList, NondetLiveVarsList, [],
-		Changed
+		ErrorList, LiveVarsList, NondetLiveVarsList, [], [],
+		Changed, HowToCheck, MayChangeProc, CheckingExtraGoals
 	).
 
 %-----------------------------------------------------------------------------%
 
 	% Lots of very boring access predicates.
 
-mode_info_get_io_state(mode_info(IOState0,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_),
+mode_info_get_io_state(
+		mode_info(IOState0,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_),
 		IOState) :-
 	% XXX
 	unsafe_promise_unique(IOState0, IOState).
 
 %-----------------------------------------------------------------------------%
 
-mode_info_set_io_state( mode_info(_,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P), IOState0,
-			mode_info(IOState,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P)) :-
+mode_info_set_io_state(mode_info(_,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T),
+		IOState0,
+		mode_info(IOState,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T)) :-
 	% XXX
 	unsafe_promise_unique(IOState0, IOState).
 
 %-----------------------------------------------------------------------------%
 
-mode_info_get_module_info(mode_info(_,ModuleInfo,_,_,_,_,_,_,_,_,_,_,_,_,_,_),
-				ModuleInfo).
+mode_info_get_module_info(
+		mode_info(_,ModuleInfo,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_),
+		ModuleInfo).
 
 %-----------------------------------------------------------------------------%
 
-mode_info_set_module_info(mode_info(A,_,C,D,E,F,G,H,I,J,K,L,M,N,O,P), ModuleInfo,
-			mode_info(A,ModuleInfo,C,D,E,F,G,H,I,J,K,L,M,N,O,P)).
+mode_info_set_module_info(mode_info(A,_,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T),
+		ModuleInfo,
+		mode_info(A,ModuleInfo,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T)).
 
 %-----------------------------------------------------------------------------%
 
-mode_info_get_preds(mode_info(_,ModuleInfo,_,_,_,_,_,_,_,_,_,_,_,_,_,_), Preds) :-
+mode_info_get_preds(
+		mode_info(_,ModuleInfo,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_),
+		Preds) :-
 	module_info_preds(ModuleInfo, Preds).
 
 %-----------------------------------------------------------------------------%
 
-mode_info_get_modes(mode_info(_,ModuleInfo,_,_,_,_,_,_,_,_,_,_,_,_,_,_), Modes) :-
+mode_info_get_modes(
+		mode_info(_,ModuleInfo,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_),
+		Modes) :-
 	module_info_modes(ModuleInfo, Modes).
 
 %-----------------------------------------------------------------------------%
 
-mode_info_get_insts(mode_info(_,ModuleInfo,_,_,_,_,_,_,_,_,_,_,_,_,_,_), Insts) :-
+mode_info_get_insts(
+		mode_info(_,ModuleInfo,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_),
+		Insts) :-
 	module_info_insts(ModuleInfo, Insts).
 
 %-----------------------------------------------------------------------------%
 
-mode_info_get_predid(mode_info(_,_,PredId,_,_,_,_,_,_,_,_,_,_,_,_,_), PredId).
+mode_info_get_predid(mode_info(_,_,PredId,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_),
+		PredId).
 
 %-----------------------------------------------------------------------------%
 
-mode_info_get_procid(mode_info(_,_,_,ProcId,_,_,_,_,_,_,_,_,_,_,_,_), ProcId).
+mode_info_get_procid(mode_info(_,_,_,ProcId,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_), 
+		ProcId).
 
 %-----------------------------------------------------------------------------%
 
-mode_info_get_varset(mode_info(_,_,_,_,VarSet,_,_,_,_,_,_,_,_,_,_,_), VarSet).
+mode_info_get_varset(mode_info(_,_,_,_,VarSet,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_),
+		VarSet).
 
 %-----------------------------------------------------------------------------%
 
-mode_info_set_varset(VarSet, mode_info(A,B,C,D,_,F,G,H,I,J,K,L,M,N,O,P),
-				mode_info(A,B,C,D,VarSet,F,G,H,I,J,K,L,M,N,O,P)).
+mode_info_set_varset(VarSet,
+		mode_info(A,B,C,D,_,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T),
+		mode_info(A,B,C,D,VarSet,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T)).
 
 %-----------------------------------------------------------------------------%
 
-mode_info_get_var_types(mode_info(_,_,_,_,_,VarTypes,_,_,_,_,_,_,_,_,_,_),
-				VarTypes).
+mode_info_get_var_types(
+		mode_info(_,_,_,_,_,VarTypes,_,_,_,_,_,_,_,_,_,_,_,_,_,_),
+		VarTypes).
 
 %-----------------------------------------------------------------------------%
 
-mode_info_set_var_types(VarTypes, mode_info(A,B,C,D,E,_,G,H,I,J,K,L,M,N,O,P),
-			mode_info(A,B,C,D,E,VarTypes,G,H,I,J,K,L,M,N,O,P)).
+mode_info_set_var_types(VTypes,
+		mode_info(A,B,C,D,E,_,G,H,I,J,K,L,M,N,O,P,Q,R,S,T),
+		mode_info(A,B,C,D,E,VTypes,G,H,I,J,K,L,M,N,O,P,Q,R,S,T)).
 
 %-----------------------------------------------------------------------------%
 
-mode_info_get_context(mode_info(_,_,_,_,_,_,Context,_,_,_,_,_,_,_,_,_), Context).
+mode_info_get_context(mode_info(_,_,_,_,_,_,Context,_,_,_,_,_,_,_,_,_,_,_,_,_),
+		Context).
 
 %-----------------------------------------------------------------------------%
 
-mode_info_set_context(Context, mode_info(A,B,C,D,E,F,_,H,I,J,K,L,M,N,O,P),
-			mode_info(A,B,C,D,E,F,Context,H,I,J,K,L,M,N,O,P)).
+mode_info_set_context(Context,
+		mode_info(A,B,C,D,E,F,_,H,I,J,K,L,M,N,O,P,Q,R,S,T),
+		mode_info(A,B,C,D,E,F,Context,H,I,J,K,L,M,N,O,P,Q,R,S,T)).
 
 %-----------------------------------------------------------------------------%
 
-mode_info_get_mode_context(mode_info(_,_,_,_,_,_,_,ModeContext,_,_,_,_,_,_,_,_),
-				ModeContext).
+mode_info_get_mode_context(
+		mode_info(_,_,_,_,_,_,_,ModeContext,_,_,_,_,_,_,_,_,_,_,_,_),
+		ModeContext).
 
 %-----------------------------------------------------------------------------%
 
 mode_info_set_mode_context(ModeContext,
-		mode_info(A,B,C,D,E,F,G,_,I,J,K,L,M,N,O,P),
-		mode_info(A,B,C,D,E,F,G,ModeContext,I,J,K,L,M,N,O,P)).
+		mode_info(A,B,C,D,E,F,G,_,I,J,K,L,M,N,O,P,Q,R,S,T),
+		mode_info(A,B,C,D,E,F,G,ModeContext,I,J,K,L,M,N,O,P,Q,R,S,T)).
 
 %-----------------------------------------------------------------------------%
 
 mode_info_set_call_context(unify(UnifyContext)) -->
 	mode_info_set_mode_context(unify(UnifyContext, left)).
-mode_info_set_call_context(call(PredId)) -->
-	mode_info_set_mode_context(call(PredId, 0)).
-mode_info_set_call_context(higher_order_call(PredOrFunc)) -->
-	mode_info_set_mode_context(higher_order_call(PredOrFunc, 0)).
+mode_info_set_call_context(call(CallId)) -->
+	mode_info_set_mode_context(call(CallId, 0)).
 
 mode_info_set_call_arg_context(ArgNum, ModeInfo0, ModeInfo) :-
 	mode_info_get_mode_context(ModeInfo0, ModeContext0),
-	( ModeContext0 = call(PredId, _) ->
-		mode_info_set_mode_context(call(PredId, ArgNum),
-			ModeInfo0, ModeInfo)
-	; ModeContext0 = higher_order_call(PredOrFunc, _) ->
-		mode_info_set_mode_context(
-			higher_order_call(PredOrFunc, ArgNum),
+	( ModeContext0 = call(CallId, _) ->
+		mode_info_set_mode_context(call(CallId, ArgNum),
 			ModeInfo0, ModeInfo)
+	; ModeContext0 = unify(_UnifyContext, _Side) ->
+		% This only happens when checking that the typeinfo variables
+		% for polymorphic complicated unifications are ground.
+		% For that case, we don't care about the ArgNum.
+		ModeInfo = ModeInfo0
 	;
 		error("mode_info_set_call_arg_context")
 	).
@@ -466,7 +583,8 @@ mode_info_unset_call_context -->
 
 %-----------------------------------------------------------------------------%
 
-mode_info_get_instmap(mode_info(_,_,_,_,_,_,_,_,InstMap,_,_,_,_,_,_,_), InstMap).
+mode_info_get_instmap(mode_info(_,_,_,_,_,_,_,_,InstMap,_,_,_,_,_,_,_,_,_,_,_),
+		InstMap).
 
 	% mode_info_dcg_get_instmap/3 is the same as mode_info_get_instmap/2
 	% except that it's easier to use inside a DCG.
@@ -477,8 +595,10 @@ mode_info_dcg_get_instmap(InstMap, ModeInfo, ModeInfo) :-
 %-----------------------------------------------------------------------------%
 
 mode_info_set_instmap( InstMap,
-		mode_info(A,B,C,D,E,F,G,H,InstMap0,J,DelayInfo0,L,M,N,O,P),
-		mode_info(A,B,C,D,E,F,G,H,InstMap,J,DelayInfo,L,M,N,O,P)) :-
+		mode_info(A,B,C,D,E,F,G,H,InstMap0,J,
+			DelayInfo0,L,M,N,O,P,Q,R,S,T),
+		mode_info(A,B,C,D,E,F,G,H,InstMap,J,
+			DelayInfo,L,M,N,O,P,Q,R,S,T)) :-
 	( instmap__is_unreachable(InstMap), instmap__is_reachable(InstMap0) ->
 		delay_info__bind_all_vars(DelayInfo0, DelayInfo)
 	;
@@ -487,28 +607,33 @@ mode_info_set_instmap( InstMap,
 
 %-----------------------------------------------------------------------------%
 
-mode_info_get_locked_vars(mode_info(_,_,_,_,_,_,_,_,_,LockedVars,_,_,_,_,_,_),
+mode_info_get_locked_vars(
+		mode_info(_,_,_,_,_,_,_,_,_,LockedVars,_,_,_,_,_,_,_,_,_,_),
 		LockedVars).
 
 %-----------------------------------------------------------------------------%
 
-mode_info_set_locked_vars( mode_info(A,B,C,D,E,F,G,H,I,_,K,L,M,N,O,P), LockedVars,
-			mode_info(A,B,C,D,E,F,G,H,I,LockedVars,K,L,M,N,O,P)).
+mode_info_set_locked_vars( mode_info(A,B,C,D,E,F,G,H,I,_,K,L,M,N,O,P,Q,R,S,T),
+	LockedVars,
+	mode_info(A,B,C,D,E,F,G,H,I,LockedVars,K,L,M,N,O,P,Q,R,S,T)).
 
 %-----------------------------------------------------------------------------%
 
-mode_info_get_errors(mode_info(_,_,_,_,_,_,_,_,_,_,_,Errors,_,_,_,_), Errors).
+mode_info_get_errors(mode_info(_,_,_,_,_,_,_,_,_,_,_,Errors,_,_,_,_,_,_,_,_),
+		Errors).
 
 %-----------------------------------------------------------------------------%
 
-mode_info_get_num_errors(mode_info(_,_,_,_,_,_,_,_,_,_,_,Errors,_,_,_,_),
+mode_info_get_num_errors(
+		mode_info(_,_,_,_,_,_,_,_,_,_,_,Errors,_,_,_,_,_,_,_,_),
 		NumErrors) :-
 	list__length(Errors, NumErrors).
 
 %-----------------------------------------------------------------------------%
 
-mode_info_set_errors( Errors, mode_info(A,B,C,D,E,F,G,H,I,J,K,_,M,N,O,P), 
-			mode_info(A,B,C,D,E,F,G,H,I,J,K,Errors,M,N,O,P)).
+mode_info_set_errors(Errors,
+		mode_info(A,B,C,D,E,F,G,H,I,J,K,_,M,N,O,P,Q,R,S,T), 
+		mode_info(A,B,C,D,E,F,G,H,I,J,K,Errors,M,N,O,P,Q,R,S,T)).
 
 %-----------------------------------------------------------------------------%
 
@@ -522,9 +647,9 @@ mode_info_set_errors( Errors, mode_info(A,B,C,D,E,F,G,H,I,J,K,_,M,N,O,P),
 
 mode_info_add_live_vars(NewLiveVars,
 		mode_info(A,B,C,D,E,F,G,H,I,J,K,L,
-			LiveVars0,NondetLiveVars0,O,P),
+			LiveVars0,NondetLiveVars0,O,P,Q,R,S,T),
 		mode_info(A,B,C,D,E,F,G,H,I,J,K,L,
-			LiveVars,NondetLiveVars,O,P)) :-
+			LiveVars,NondetLiveVars,O,P,Q,R,S,T)) :-
 
 	LiveVars = [NewLiveVars | LiveVars0],
 	NondetLiveVars = [NewLiveVars | NondetLiveVars0].
@@ -534,9 +659,9 @@ mode_info_add_live_vars(NewLiveVars,
 
 mode_info_remove_live_vars(OldLiveVars, ModeInfo0, ModeInfo) :-
 	ModeInfo0 = mode_info(A,B,C,D,E,F,G,H,I,J,K,L,
-				LiveVars0, NondetLiveVars0,O,P),
+				LiveVars0, NondetLiveVars0,O,P,Q,R,S,T),
 	ModeInfo1 = mode_info(A,B,C,D,E,F,G,H,I,J,K,L,
-				LiveVars, NondetLiveVars,O,P),
+				LiveVars, NondetLiveVars,O,P,Q,R,S,T),
 	(
 		list__delete_first(LiveVars0, OldLiveVars, LiveVars1),
 		list__delete_first(NondetLiveVars0, OldLiveVars,
@@ -563,8 +688,9 @@ mode_info_var_list_is_live([Var | Vars], ModeInfo, [Live | Lives]) :-
 
 	% Check whether a variable is live or not
 
-mode_info_var_is_live(mode_info(_,_,_,_,_,_,_,_,_,_,_,_,LiveVarsList,_,_,_), Var,
-		Result) :-
+mode_info_var_is_live(
+		mode_info(_,_,_,_,_,_,_,_,_,_,_,_,LiveVarsList,_,_,_,_,_,_,_),
+		Var, Result) :-
 	(
 		% some [LiveVars] 
 		list__member(LiveVars, LiveVarsList),
@@ -578,7 +704,7 @@ mode_info_var_is_live(mode_info(_,_,_,_,_,_,_,_,_,_,_,_,LiveVarsList,_,_,_), Var
 	% Check whether a variable is nondet_live or not.
 
 mode_info_var_is_nondet_live(mode_info(_,_,_,_,_,_,_,_,_,_,_,_,_,
-		NondetLiveVarsList,_,_), Var, Result) :-
+		NondetLiveVarsList,_,_,_,_,_,_), Var, Result) :-
 	(
 		% some [LiveVars] 
 		list__member(LiveVars, NondetLiveVarsList),
@@ -589,8 +715,8 @@ mode_info_var_is_nondet_live(mode_info(_,_,_,_,_,_,_,_,_,_,_,_,_,
 		Result = dead
 	).
 
-mode_info_get_liveness(mode_info(_,_,_,_,_,_,_,_,_,_,_,_,LiveVarsList,_,_,_),
-		LiveVars) :-
+mode_info_get_liveness(mode_info(_,_,_,_,_,_,_,_,_,_,_,_,LiveVarsList,
+		_,_,_,_,_,_,_), LiveVars) :-
 	set__init(LiveVars0),
 	mode_info_get_liveness_2(LiveVarsList, LiveVars0, LiveVars).
 
@@ -599,6 +725,13 @@ mode_info_get_liveness_2([LiveVarsSet | LiveVarsList], LiveVars0, LiveVars) :-
 	set__union(LiveVars0, LiveVarsSet, LiveVars1),
 	mode_info_get_liveness_2(LiveVarsList, LiveVars1, LiveVars).
 
+mode_info_get_live_vars(mode_info(_,_,_,_,_,_,_,_,_,_,_,_,LiveVarsList,
+		_,_,_,_,_,_,_), LiveVarsList).
+
+mode_info_set_live_vars(LiveVarsList,
+		mode_info(A,B,C,D,E,F,G,H,I,J,K,L,_,N,O,P,Q,R,S,T),
+		mode_info(A,B,C,D,E,F,G,H,I,J,K,L,LiveVarsList,N,O,P,Q,R,S,T)).
+
 %-----------------------------------------------------------------------------%
 
 	% Since we don't yet handle polymorphic modes, the inst varset
@@ -620,59 +753,109 @@ mode_info_get_types_of_vars(ModeInfo, Vars, TypesOfVars) :-
 	% push them on the stack, and to unlock a set of vars, we just
 	% pop them off the stack.  The stack is implemented as a list.
 
-mode_info_lock_vars(Vars, ModeInfo0, ModeInfo) :-
+mode_info_lock_vars(Reason, Vars, ModeInfo0, ModeInfo) :-
 	mode_info_get_locked_vars(ModeInfo0, LockedVars),
-	mode_info_set_locked_vars(ModeInfo0, [Vars | LockedVars], ModeInfo).
+	mode_info_set_locked_vars(ModeInfo0, [Reason - Vars | LockedVars],
+			ModeInfo).
 
-mode_info_unlock_vars(_, ModeInfo0, ModeInfo) :-
+mode_info_unlock_vars(Reason, Vars, ModeInfo0, ModeInfo) :-
 	mode_info_get_locked_vars(ModeInfo0, LockedVars0),
-	( LockedVars0 = [_ | LockedVars1] ->
+	(
+		LockedVars0 = [Reason - TheseVars | LockedVars1],
+		set__equal(TheseVars, Vars)
+	->
 		LockedVars = LockedVars1
 	;
-		error("mode_info_unlock_vars: stack is empty")
+		error("mode_info_unlock_vars: some kind of nesting error")
 	),
 	mode_info_set_locked_vars(ModeInfo0, LockedVars, ModeInfo).
 
-mode_info_var_is_locked(ModeInfo, Var) :-
+mode_info_var_is_locked(ModeInfo, Var, Reason) :-
 	mode_info_get_locked_vars(ModeInfo, LockedVarsList),
-	mode_info_var_is_locked_2(LockedVarsList, Var).
+	mode_info_var_is_locked_2(LockedVarsList, Var, Reason).
 
-mode_info_var_is_locked_2([Set | Sets], Var) :-
+mode_info_var_is_locked_2([ThisReason - Set | Sets], Var, Reason) :-
 	(
 		set__member(Var, Set)
 	->
-		true
+		Reason = ThisReason
 	;
-		mode_info_var_is_locked_2(Sets, Var)
+		mode_info_var_is_locked_2(Sets, Var, Reason)
 	).
 
-mode_info_get_delay_info(mode_info(_,_,_,_,_,_,_,_,_,_,DelayInfo,_,_,_,_,_),
-	DelayInfo).
+mode_info_get_delay_info(
+		mode_info(_,_,_,_,_,_,_,_,_,_,DelayInfo,_,_,_,_,_,_,_,_,_),
+		DelayInfo).
 
-mode_info_set_delay_info(DelayInfo, mode_info(A,B,C,D,E,F,G,H,I,J,_,L,M,N,O,P),
-			mode_info(A,B,C,D,E,F,G,H,I,J,DelayInfo,L,M,N,O,P)).
+mode_info_set_delay_info(DelayInfo,
+		mode_info(A,B,C,D,E,F,G,H,I,J,_,L,M,N,O,P,Q,R,S,T),
+		mode_info(A,B,C,D,E,F,G,H,I,J,DelayInfo,L,M,N,O,P,Q,R,S,T)).
 
 mode_info_get_nondet_live_vars(mode_info(_,_,_,_,_,_,_,_,_,_,_,_,_,
-			NondetLiveVars,_,_), NondetLiveVars).
+			NondetLiveVars,_,_,_,_,_,_), NondetLiveVars).
 
 mode_info_set_nondet_live_vars(NondetLiveVars,
-		mode_info(A,B,C,D,E,F,G,H,I,J,K,L,M,_,O,P),
-		mode_info(A,B,C,D,E,F,G,H,I,J,K,L,M,NondetLiveVars,O,P)).
+	mode_info(A,B,C,D,E,F,G,H,I,J,K,L,M,_,O,P,Q,R,S,T),
+	mode_info(A,B,C,D,E,F,G,H,I,J,K,L,M,NondetLiveVars,O,P,Q,R,S,T)).
 
 mode_info_get_last_checkpoint_insts(mode_info(_,_,_,_,_,_,_,_,_,_,_,_,_,_,
-		LastCheckpointInsts,_), LastCheckpointInsts).
+		LastCheckpointInsts,_,_,_,_,_), LastCheckpointInsts).
 
 mode_info_set_last_checkpoint_insts(LastCheckpointInsts,
-			mode_info(A,B,C,D,E,F,G,H,I,J,K,L,M,N,_,P),
+			mode_info(A,B,C,D,E,F,G,H,I,J,K,L,M,N,_,P,Q,R,S,T),
 			mode_info(A,B,C,D,E,F,G,H,I,J,K,L,M,N,
-				LastCheckpointInsts,P)).
+				LastCheckpointInsts,P,Q,R,S,T)).
 
-mode_info_get_changed_flag(mode_info(_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,Changed),
-				Changed).
+mode_info_get_parallel_vars(PVars, ModeInfo, ModeInfo) :-
+	ModeInfo = mode_info(_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,PVars,_,_,_,_).
+
+mode_info_set_parallel_vars(PVars,
+		mode_info(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,_,Q,R,S,T),
+		mode_info(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,PVars,Q,R,S,T)).
+
+mode_info_get_changed_flag(
+		mode_info(_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,Changed,_,_,_),
+		Changed).
 
 mode_info_set_changed_flag(Changed,
-			mode_info(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,_),
-			mode_info(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,Changed)).
+		mode_info(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,_,R,S,T),
+		mode_info(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Changed,R,S,T)).
+
+mode_info_get_how_to_check(
+		mode_info(_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,How,_,_),
+		How).
+
+mode_info_set_how_to_check(How,
+		mode_info(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,_,S,T),
+		mode_info(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,How,S,T)).
+
+mode_info_get_may_change_called_proc(
+		mode_info(_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,MayChange,_),
+		MayChange).
+
+mode_info_set_may_change_called_proc(MayChange,
+		mode_info(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,_,T),
+		mode_info(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,MayChange,T)).
+
+mode_info_set_checking_extra_goals(Checking,
+		mode_info(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,Checking0),
+		mode_info(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,Checking)) :-
+	( Checking0 = yes, Checking = yes ->
+		% This should never happen - once the extra goals are
+		% introduced, rechecking the goal should not introduce
+		% more extra goals.
+		error(
+		"mode analysis: rechecking extra goals adds more extra goals")
+	;
+		true
+	).
+
+%-----------------------------------------------------------------------------%
+
+mode_info_get_call_id(ModeInfo, PredId, CallId) :-
+	mode_info_get_module_info(ModeInfo, ModuleInfo),
+	module_info_pred_info(ModuleInfo, PredId, PredInfo),
+	pred_info_get_call_id(PredInfo, CallId).
 
 %-----------------------------------------------------------------------------%
 
