@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -108,9 +108,9 @@
 	% may need to insert new merge_insts into the merge_inst table.
 	% If the first argument is yes, the instmap_deltas for calls
 	% and deconstruction unifications are also recomputed.
-:- pred recompute_instmap_delta(bool, hlds_goal, hlds_goal, instmap,
+:- pred recompute_instmap_delta(bool, hlds_goal, hlds_goal, vartypes, instmap,
 				module_info, module_info).
-:- mode recompute_instmap_delta(in, in, out, in, in, out) is det.
+:- mode recompute_instmap_delta(in, in, out, in, in, in, out) is det.
 
 	% Given corresponding lists of types and modes, produce a new
 	% list of modes which includes the information provided by the
@@ -135,7 +135,7 @@
 :- mode get_arg_lives(in, in, out) is det.
 
 	% Predicates to make error messages more readable by stripping
-	% "mercury_builtin" module qualifiers from modes.
+	% "builtin:" module qualifiers from modes.
 
 :- pred strip_builtin_qualifier_from_cons_id(cons_id, cons_id).
 :- mode strip_builtin_qualifier_from_cons_id(in, out) is det.
@@ -152,24 +152,46 @@
 	% Given the switched on variable and the instmaps before the switch
 	% and after a branch make sure that any information added by the
 	% functor test gets added to the instmap for the case.
-:- pred fixup_switch_var(var, instmap, instmap, hlds_goal, hlds_goal). 
+:- pred fixup_switch_var(prog_var, instmap, instmap, hlds_goal, hlds_goal). 
 :- mode fixup_switch_var(in, in, in, in, out) is det.
 
 %-----------------------------------------------------------------------------%
 
-:- pred normalise_insts(list(inst), module_info, list(inst)).
-:- mode normalise_insts(in, in, out) is det.
+:- pred normalise_insts(list(inst), list(type), module_info, list(inst)).
+:- mode normalise_insts(in, in, in, out) is det.
 
-:- pred normalise_inst(inst, module_info, inst).
-:- mode normalise_inst(in, in, out) is det.
+:- pred normalise_inst(inst, (type), module_info, inst).
+:- mode normalise_inst(in, in, in, out) is det.
+
+%-----------------------------------------------------------------------------%
+
+	% Partition a list of arguments into inputs and others.
+:- pred partition_args(module_info, list(mode), list(T), list(T), list(T)).
+:- mode partition_args(in, in, in, out, out) is det.
+
+%-----------------------------------------------------------------------------%
+
+	% Construct a mode corresponding to the standard `in',
+	% `out', `uo' or `unused' mode.
+:- pred in_mode((mode)::out) is det.
+:- pred out_mode((mode)::out) is det.
+:- pred uo_mode((mode)::out) is det.
+:- pred unused_mode((mode)::out) is det.
+
+	% Construct the modes used for `aditi__state' arguments.
+	% XXX These should be unique, but are not yet because that
+	% would require alias tracking.
+:- func aditi_ui_mode = (mode).
+:- func aditi_di_mode = (mode).
+:- func aditi_uo_mode = (mode).
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
 :- implementation.
-:- import_module require, int, map, set, term, std_util, assoc_list.
-:- import_module prog_util, type_util.
-:- import_module inst_match, inst_util.
+:- import_module require, int, map, set, std_util, assoc_list.
+:- import_module prog_util, prog_io, type_util.
+:- import_module inst_match, inst_util, term.
 
 %-----------------------------------------------------------------------------%
 
@@ -202,25 +224,29 @@ insts_to_mode(Initial, Final, Mode) :-
 	% more readable.
 	%
 	( Initial = free, Final = ground(shared, no) ->
-		Mode = user_defined_mode(
-				qualified("mercury_builtin", "out"), [])
+		make_std_mode("out", [], Mode)
 	; Initial = free, Final = ground(unique, no) ->
-		Mode = user_defined_mode(qualified("mercury_builtin", "uo"), [])
+		make_std_mode("uo", [], Mode)
+	; Initial = free, Final = ground(mostly_unique, no) ->
+		make_std_mode("muo", [], Mode)
 	; Initial = ground(shared, no), Final = ground(shared, no) ->
-		Mode = user_defined_mode(qualified("mercury_builtin", "in"), [])
+		make_std_mode("in", [], Mode)
 	; Initial = ground(unique, no), Final = ground(clobbered, no) ->
-		Mode = user_defined_mode(qualified("mercury_builtin", "di"), [])
+		make_std_mode("di", [], Mode)
+	; Initial = ground(mostly_unique, no),
+	  Final = ground(mostly_clobbered, no) ->
+		make_std_mode("mdi", [], Mode)
 	; Initial = ground(unique, no), Final = ground(unique, no) ->
-		Mode = user_defined_mode(qualified("mercury_builtin", "ui"), [])
+		make_std_mode("ui", [], Mode)
+	; Initial = ground(mostly_unique, no),
+	  Final = ground(mostly_unique, no) ->
+		make_std_mode("mdi", [], Mode)
 	; Initial = free ->
-		Mode = user_defined_mode(qualified("mercury_builtin", "out"),
-								[Final])
+		make_std_mode("out", [Final], Mode)
 	; Final = ground(clobbered, no) ->
-		Mode = user_defined_mode(qualified("mercury_builtin", "di"),
-								[Initial])
+		make_std_mode("di", [Initial], Mode)
 	; Initial = Final ->
-		Mode = user_defined_mode(qualified("mercury_builtin", "in"),
-								[Initial])
+		make_std_mode("in", [Initial], Mode)
 	;
 		Mode = (Initial -> Final)
 	).
@@ -607,7 +633,7 @@ propagate_ctor_info(bound(Uniq, BoundInsts0), Type, _Constructors, ModuleInfo,
 		Inst = bound(Uniq, BoundInsts)
 	).
 propagate_ctor_info(ground(Uniq, no), Type, Constructors, ModuleInfo, Inst) :-
-	( type_is_higher_order(Type, function, ArgTypes) ->
+	( type_is_higher_order(Type, function, _, ArgTypes) ->
 		default_higher_order_func_inst(ArgTypes, ModuleInfo,
 			HigherOrderInstInfo),
 		Inst = ground(Uniq, yes(HigherOrderInstInfo))
@@ -622,7 +648,7 @@ propagate_ctor_info(ground(Uniq, yes(PredInstInfo0)), Type, _Ctors, ModuleInfo,
 	PredInstInfo0 = pred_inst_info(PredOrFunc, Modes0, Det),
 	PredInstInfo = pred_inst_info(PredOrFunc, Modes, Det),
 	(
-		type_is_higher_order(Type, PredOrFunc, ArgTypes),
+		type_is_higher_order(Type, PredOrFunc, _, ArgTypes),
 		list__same_length(ArgTypes, Modes0)
 	->
 		propagate_types_into_mode_list(ArgTypes, ModuleInfo,
@@ -668,7 +694,7 @@ propagate_ctor_info_lazily(bound(Uniq, BoundInsts0), Type0, Subst,
 	).
 propagate_ctor_info_lazily(ground(Uniq, no), Type0, Subst, ModuleInfo, Inst) :-
 	apply_type_subst(Type0, Subst, Type),
-	( type_is_higher_order(Type, function, ArgTypes) ->
+	( type_is_higher_order(Type, function, _, ArgTypes) ->
 		default_higher_order_func_inst(ArgTypes, ModuleInfo,
 			HigherOrderInstInfo),
 		Inst = ground(Uniq, yes(HigherOrderInstInfo))
@@ -688,7 +714,7 @@ propagate_ctor_info_lazily(ground(Uniq, yes(PredInstInfo0)), Type0, Subst,
 	PredInstInfo = pred_inst_info(PredOrFunc, Modes, Det),
 	apply_type_subst(Type0, Subst, Type),
 	(
-		type_is_higher_order(Type, PredOrFunc, ArgTypes),
+		type_is_higher_order(Type, PredOrFunc, _, ArgTypes),
 		list__same_length(ArgTypes, Modes0)
 	->
 		propagate_types_into_mode_list(ArgTypes, ModuleInfo,
@@ -749,7 +775,7 @@ default_higher_order_func_inst(PredArgTypes, ModuleInfo, PredInstInfo) :-
 constructors_to_bound_insts([], _, _, []).
 constructors_to_bound_insts([Ctor | Ctors], Uniq, ModuleInfo,
 		[BoundInst | BoundInsts]) :-
-	Ctor = Name - Args,
+	Ctor = ctor(_ExistQVars, _Constraints, Name, Args),
 	ctor_arg_list_to_inst_list(Args, Uniq, Insts),
 	list__length(Insts, Arity),
 	BoundInst = functor(cons(Name, Arity), Insts),
@@ -791,7 +817,7 @@ propagate_ctor_info_2(BoundInsts0, Type, ModuleInfo, BoundInsts) :-
 		BoundInsts = BoundInsts0
 	).
 
-:- pred propagate_ctor_info_3(list(bound_inst), string, list(constructor),
+:- pred propagate_ctor_info_3(list(bound_inst), module_name, list(constructor),
 		tsubst, module_info, list(bound_inst)).
 :- mode propagate_ctor_info_3(in, in, in, in, in, out) is det.
 
@@ -807,12 +833,12 @@ propagate_ctor_info_3([BoundInst0 | BoundInsts0], TypeModule, Constructors,
 	(
 		ConsId = cons(ConsName, Arity),
 		GetCons = lambda([Ctor::in] is semidet, (
-				Ctor = ConsName - CtorArgs,
+				Ctor = ctor(_, _, ConsName, CtorArgs),
 				list__length(CtorArgs, Arity)
 			)),
 		list__filter(GetCons, Constructors, [Constructor])
 	->
-		Constructor = _ - Args,
+		Constructor = ctor(_ExistQVars, _Constraints, _Name, Args),
 		GetArgTypes = lambda([CtorArg::in, ArgType::out] is det, (
 				CtorArg = _ArgName - ArgType
 			)),
@@ -868,7 +894,7 @@ mode_get_insts_semidet(ModuleInfo, user_defined_mode(Name, Args),
 						_Context, _Status),
 	ModeDefn = eqv_mode(Mode0),
 	mode_substitute_arg_list(Mode0, Params, Args, Mode),
-	mode_get_insts(ModuleInfo, Mode, Initial, Final).
+	mode_get_insts_semidet(ModuleInfo, Mode, Initial, Final).
 
 mode_get_insts(ModuleInfo, Mode, Inst1, Inst2) :-
 	( mode_get_insts_semidet(ModuleInfo, Mode, Inst1a, Inst2a) ->
@@ -1050,163 +1076,239 @@ mode_id_to_int(_ - X, X).
 	% and deconstructions may become non-local (XXX does this require
 	% rerunning mode analysis rather than just recompute_instmap_delta?).
 
-recompute_instmap_delta(RecomputeAtomic, Goal0, Goal, InstMap0) -->
-	recompute_instmap_delta(RecomputeAtomic, Goal0, Goal, InstMap0, _).
+recompute_instmap_delta(RecomputeAtomic, Goal0, Goal, VarTypes, InstMap0) -->
+	recompute_instmap_delta(RecomputeAtomic, Goal0, Goal, VarTypes, InstMap0, _).
 
-:- pred recompute_instmap_delta(bool, hlds_goal, hlds_goal, instmap,
+:- pred recompute_instmap_delta(bool, hlds_goal, hlds_goal, vartypes, instmap,
 		instmap_delta, module_info, module_info).
-:- mode recompute_instmap_delta(in, in, out, in, out, in, out) is det.
+:- mode recompute_instmap_delta(in, in, out, in, in, out, in, out) is det.
 
 recompute_instmap_delta(RecomputeAtomic, Goal0 - GoalInfo0, Goal - GoalInfo,
-		InstMap0, InstMapDelta) -->
+		VarTypes, InstMap0, InstMapDelta) -->
 	( 
 		{ RecomputeAtomic = no },
-		{ goal_is_atomic(Goal0) }
+		( 
+			{ goal_is_atomic(Goal0) }
+		;
+			% Lambda expressions always need to be processed.
+			{ Goal0 = unify(_, Rhs, _, _, _) },
+			{ Rhs \= lambda_goal(_, _, _, _, _, _, _, _) }
+		)
 	->
 		{ Goal = Goal0 },
-		{ GoalInfo = GoalInfo0 },
-		{ goal_info_get_instmap_delta(GoalInfo, InstMapDelta) } 
+		{ GoalInfo1 = GoalInfo0 }
 	;
 		recompute_instmap_delta_2(RecomputeAtomic, Goal0,
-			 GoalInfo0, Goal, InstMap0, InstMapDelta0),
+			 GoalInfo0, Goal, VarTypes, InstMap0, InstMapDelta0),
 		{ goal_info_get_nonlocals(GoalInfo0, NonLocals) },
 		{ instmap_delta_restrict(InstMapDelta0,
-			NonLocals, InstMapDelta) },
+			NonLocals, InstMapDelta1) },
 		{ goal_info_set_instmap_delta(GoalInfo0,
-			InstMapDelta, GoalInfo) }
-	).
+			InstMapDelta1, GoalInfo1) }
+	),
+
+	% If the initial instmap is unreachable so is the final instmap.
+	( { instmap__is_unreachable(InstMap0) } ->
+		{ instmap_delta_init_unreachable(UnreachableInstMapDelta) },
+		{ goal_info_set_instmap_delta(GoalInfo1,
+			UnreachableInstMapDelta, GoalInfo) }
+	;
+		{ GoalInfo = GoalInfo1 }
+	),
+	{ goal_info_get_instmap_delta(GoalInfo, InstMapDelta) }.
 
 :- pred recompute_instmap_delta_2(bool, hlds_goal_expr, hlds_goal_info,
-		hlds_goal_expr, instmap, instmap_delta,
+		hlds_goal_expr, vartypes, instmap, instmap_delta,
 		module_info, module_info).
-:- mode recompute_instmap_delta_2(in, in, in, out, in, out, in, out) is det.
+:- mode recompute_instmap_delta_2(in, in, in, out, in, in, out, in, out) is det.
 
 recompute_instmap_delta_2(Atomic, switch(Var, Det, Cases0, SM), GoalInfo,
-		switch(Var, Det, Cases, SM), InstMap, InstMapDelta) -->
+		switch(Var, Det, Cases, SM), VarTypes, InstMap,
+		InstMapDelta) -->
 	{ goal_info_get_nonlocals(GoalInfo, NonLocals) },
 	recompute_instmap_delta_cases(Atomic, Var, Cases0, Cases,
-		InstMap, NonLocals, InstMapDelta).
+		VarTypes, InstMap, NonLocals, InstMapDelta).
 
 recompute_instmap_delta_2(Atomic, conj(Goals0), _, conj(Goals),
-		InstMap, InstMapDelta) -->
+		VarTypes, InstMap, InstMapDelta) -->
 	recompute_instmap_delta_conj(Atomic, Goals0, Goals,
-		InstMap, InstMapDelta).
+		VarTypes, InstMap, InstMapDelta).
+
+recompute_instmap_delta_2(Atomic, par_conj(Goals0, SM), GoalInfo,
+		par_conj(Goals, SM), VarTypes, InstMap, InstMapDelta) -->
+	{ goal_info_get_nonlocals(GoalInfo, NonLocals) },
+	recompute_instmap_delta_par_conj(Atomic, Goals0, Goals,
+		VarTypes, InstMap, NonLocals, InstMapDelta).
 
 recompute_instmap_delta_2(Atomic, disj(Goals0, SM), GoalInfo, disj(Goals, SM),
-		InstMap, InstMapDelta) -->
+		VarTypes, InstMap, InstMapDelta) -->
 	{ goal_info_get_nonlocals(GoalInfo, NonLocals) },
 	recompute_instmap_delta_disj(Atomic, Goals0, Goals,
-		InstMap, NonLocals, InstMapDelta).
+		VarTypes, InstMap, NonLocals, InstMapDelta).
 
 recompute_instmap_delta_2(Atomic, not(Goal0), _, not(Goal),
-		InstMap, InstMapDelta) -->
+		VarTypes, InstMap, InstMapDelta) -->
 	{ instmap_delta_init_reachable(InstMapDelta) },
-	recompute_instmap_delta(Atomic, Goal0, Goal, InstMap).
+	recompute_instmap_delta(Atomic, Goal0, Goal, VarTypes, InstMap).
 
 recompute_instmap_delta_2(Atomic, if_then_else(Vars, A0, B0, C0, SM), GoalInfo,
-		if_then_else(Vars, A, B, C, SM), InstMap0, InstMapDelta) -->
-	recompute_instmap_delta(Atomic, A0, A, InstMap0, InstMapDelta1),
+		if_then_else(Vars, A, B, C, SM), VarTypes, InstMap0, InstMapDelta) -->
+	recompute_instmap_delta(Atomic, A0, A, VarTypes, InstMap0, InstMapDelta1),
 	{ instmap__apply_instmap_delta(InstMap0, InstMapDelta1, InstMap1) },
-	recompute_instmap_delta(Atomic, B0, B, InstMap1, InstMapDelta2),
-	recompute_instmap_delta(Atomic, C0, C, InstMap0, InstMapDelta3),
+	recompute_instmap_delta(Atomic, B0, B, VarTypes, InstMap1, InstMapDelta2),
+	recompute_instmap_delta(Atomic, C0, C, VarTypes, InstMap0, InstMapDelta3),
 	{ instmap_delta_apply_instmap_delta(InstMapDelta1, InstMapDelta2,
 		InstMapDelta4) },
 	{ goal_info_get_nonlocals(GoalInfo, NonLocals) },
 	merge_instmap_delta(InstMap0, NonLocals, InstMapDelta3,
 		InstMapDelta4, InstMapDelta).
 
-recompute_instmap_delta_2(Atomic, some(Vars, Goal0), _, some(Vars, Goal),
-		InstMap, InstMapDelta) -->
-	recompute_instmap_delta(Atomic, Goal0, Goal, InstMap, InstMapDelta).
+recompute_instmap_delta_2(Atomic, some(Vars, CanRemove, Goal0), _,
+		some(Vars, CanRemove, Goal),
+		VarTypes, InstMap, InstMapDelta) -->
+	recompute_instmap_delta(Atomic, Goal0, Goal, VarTypes, InstMap, InstMapDelta).
 
-recompute_instmap_delta_2(_, higher_order_call(A, Vars, B, Modes, C, D), _,
-		higher_order_call(A, Vars, B, Modes, C, D),
-		_InstMap, InstMapDelta) -->
+recompute_instmap_delta_2(_, generic_call(A, Vars, Modes, D), _,
+		generic_call(A, Vars, Modes, D),
+		_VarTypes, _InstMap, InstMapDelta) -->
 	=(ModuleInfo),
 	{ instmap_delta_from_mode_list(Vars, Modes,
 		ModuleInfo, InstMapDelta) }.
 
 recompute_instmap_delta_2(_, call(PredId, ProcId, Args, D, E, F), _,
-		call(PredId, ProcId, Args, D, E, F), InstMap, InstMapDelta) -->
-	recompute_instmap_delta_call(PredId, ProcId,
-		Args, InstMap, InstMapDelta).
-
-recompute_instmap_delta_2(_, unify(A, B, UniMode0, Uni, E), GoalInfo, 
-		unify(A, B, UniMode, Uni, E), InstMap, InstMapDelta) -->
-	recompute_instmap_delta_unify(Uni, UniMode0, UniMode,
-		GoalInfo, InstMap, InstMapDelta).
-
-recompute_instmap_delta_2(_, pragma_c_code(A, B, PredId, ProcId, Args, F, G,
-		H), _, pragma_c_code(A, B, PredId, ProcId, Args, F, G, H),
+		call(PredId, ProcId, Args, D, E, F), _VarTypes,
 		InstMap, InstMapDelta) -->
 	recompute_instmap_delta_call(PredId, ProcId,
 		Args, InstMap, InstMapDelta).
 
+recompute_instmap_delta_2(Atomic, unify(A, Rhs0, UniMode0, Uni, E), GoalInfo, 
+		unify(A, Rhs, UniMode, Uni, E), VarTypes, InstMap0,
+		InstMapDelta) -->
+	(
+		{ Rhs0 = lambda_goal(PorF, EvalMethod, FixModes, NonLocals,
+			LambdaVars, Modes, Det, Goal0) }
+	->
+		=(ModuleInfo0),
+		{ instmap__pre_lambda_update(ModuleInfo0, LambdaVars, Modes,
+			InstMap0, InstMap) },
+		recompute_instmap_delta(Atomic, Goal0, Goal, VarTypes, InstMap),
+		{ Rhs = lambda_goal(PorF, EvalMethod, FixModes, NonLocals,
+			LambdaVars, Modes, Det, Goal) }
+	;
+		{ Rhs = Rhs0 }
+	),
+	( { Atomic = yes } ->
+		recompute_instmap_delta_unify(Uni, UniMode0, UniMode,
+			GoalInfo, InstMap0, InstMapDelta)
+	;
+		{ UniMode = UniMode0 },
+		{ goal_info_get_instmap_delta(GoalInfo, InstMapDelta) }
+	).
+
+recompute_instmap_delta_2(_, pragma_c_code(A, PredId, ProcId, Args, E, F,
+		G), _, pragma_c_code(A, PredId, ProcId, Args, E, F, G),
+		_VarTypes, InstMap, InstMapDelta) -->
+	recompute_instmap_delta_call(PredId, ProcId,
+		Args, InstMap, InstMapDelta).
+
+recompute_instmap_delta_2(_, bi_implication(_, _), _, _, _, _, _) -->
+	% these should have been expanded out by now
+	{ error("recompute_instmap_delta_2: unexpected bi_implication") }.
+	
 %-----------------------------------------------------------------------------%
 
 :- pred recompute_instmap_delta_conj(bool, list(hlds_goal), list(hlds_goal),
-		instmap, instmap_delta, module_info, module_info).
-:- mode recompute_instmap_delta_conj(in, in, out, in, out, in, out) is det.
+		vartypes, instmap, instmap_delta, module_info, module_info).
+:- mode recompute_instmap_delta_conj(in, in, out, in, in, out, in, out) is det.
 
-recompute_instmap_delta_conj(_, [], [], _InstMap, InstMapDelta) -->
+recompute_instmap_delta_conj(_, [], [], _VarTypes, _InstMap, InstMapDelta) -->
 	{ instmap_delta_init_reachable(InstMapDelta) }.
 recompute_instmap_delta_conj(Atomic, [Goal0 | Goals0], [Goal | Goals],
-		InstMap0, InstMapDelta) -->
+		VarTypes, InstMap0, InstMapDelta) -->
 	recompute_instmap_delta(Atomic, Goal0, Goal,
-		InstMap0, InstMapDelta0),
+		VarTypes, InstMap0, InstMapDelta0),
 	{ instmap__apply_instmap_delta(InstMap0, InstMapDelta0, InstMap1) },
 	recompute_instmap_delta_conj(Atomic, Goals0, Goals,
-		InstMap1, InstMapDelta1),
+		VarTypes, InstMap1, InstMapDelta1),
 	{ instmap_delta_apply_instmap_delta(InstMapDelta0, InstMapDelta1,
 		InstMapDelta) }.
 
 %-----------------------------------------------------------------------------%
 
 :- pred recompute_instmap_delta_disj(bool, list(hlds_goal), list(hlds_goal),
-		instmap, set(var), instmap_delta, module_info, module_info).
-:- mode recompute_instmap_delta_disj(in, in, out, in, in, out, in, out) is det.
+		vartypes, instmap, set(prog_var), instmap_delta,
+		module_info, module_info).
+:- mode recompute_instmap_delta_disj(in, in, out, in, in, in, out, in, out)
+		is det.
 
-recompute_instmap_delta_disj(_, [], [], _, _, InstMapDelta) -->
+recompute_instmap_delta_disj(_, [], [], _, _, _, InstMapDelta) -->
 	{ instmap_delta_init_unreachable(InstMapDelta) }.
 recompute_instmap_delta_disj(Atomic, [Goal0], [Goal],
-		InstMap, _, InstMapDelta) -->
-	recompute_instmap_delta(Atomic, Goal0, Goal, InstMap, InstMapDelta).
+		VarTypes, InstMap, _, InstMapDelta) -->
+	recompute_instmap_delta(Atomic, Goal0, Goal, VarTypes, InstMap,
+		InstMapDelta).
 recompute_instmap_delta_disj(Atomic, [Goal0 | Goals0], [Goal | Goals],
-		InstMap, NonLocals, InstMapDelta) -->
+		VarTypes, InstMap, NonLocals, InstMapDelta) -->
 	{ Goals0 = [_|_] },
 	recompute_instmap_delta(Atomic, Goal0, Goal,
-		InstMap, InstMapDelta0),
+		VarTypes, InstMap, InstMapDelta0),
 	recompute_instmap_delta_disj(Atomic, Goals0, Goals,
-		InstMap, NonLocals, InstMapDelta1),
+		VarTypes, InstMap, NonLocals, InstMapDelta1),
 	merge_instmap_delta(InstMap, NonLocals, InstMapDelta0,
 		InstMapDelta1, InstMapDelta).
 
+:- pred recompute_instmap_delta_par_conj(bool, list(hlds_goal),
+		list(hlds_goal), vartypes, instmap, set(prog_var),
+		instmap_delta, module_info, module_info).
+:- mode recompute_instmap_delta_par_conj(in, in, out, in, in, in, out,
+		in, out) is det.
+
+recompute_instmap_delta_par_conj(_, [], [], _, _, _, InstMapDelta) -->
+	{ instmap_delta_init_unreachable(InstMapDelta) }.
+recompute_instmap_delta_par_conj(Atomic, [Goal0], [Goal],
+		VarTypes, InstMap, _, InstMapDelta) -->
+	recompute_instmap_delta(Atomic, Goal0, Goal, VarTypes, InstMap,
+		InstMapDelta).
+recompute_instmap_delta_par_conj(Atomic, [Goal0 | Goals0], [Goal | Goals],
+		VarTypes, InstMap, NonLocals, InstMapDelta) -->
+	{ Goals0 = [_|_] },
+	recompute_instmap_delta(Atomic, Goal0, Goal,
+		VarTypes, InstMap, InstMapDelta0),
+	recompute_instmap_delta_par_conj(Atomic, Goals0, Goals,
+		VarTypes, InstMap, NonLocals, InstMapDelta1),
+	unify_instmap_delta(InstMap, NonLocals, InstMapDelta0,
+		InstMapDelta1, InstMapDelta).
+
 %-----------------------------------------------------------------------------%
 
-:- pred recompute_instmap_delta_cases(bool, var, list(case), list(case),
-		instmap, set(var), instmap_delta, module_info, module_info).
+:- pred recompute_instmap_delta_cases(bool, prog_var, list(case), list(case),
+		vartypes, instmap, set(prog_var), instmap_delta,
+		module_info, module_info).
 :- mode recompute_instmap_delta_cases(in, in, in, out,
-		in, in, out, in, out) is det.
+		in, in, in, out, in, out) is det.
 
-recompute_instmap_delta_cases(_, _, [], [], _, _, InstMapDelta) -->
+recompute_instmap_delta_cases(_, _, [], [], _, _, _, InstMapDelta) -->
 	{ instmap_delta_init_unreachable(InstMapDelta) }.
 recompute_instmap_delta_cases(Atomic, Var, [Case0 | Cases0], [Case | Cases],
-		InstMap0, NonLocals, InstMapDelta) -->
+		VarTypes, InstMap0, NonLocals, InstMapDelta) -->
 	{ Case0 = case(Functor, Goal0) },
-	instmap__bind_var_to_functor(Var, Functor, InstMap0, InstMap),
-	recompute_instmap_delta(Atomic, Goal0, Goal, InstMap, InstMapDelta0),
-	instmap_delta_bind_var_to_functor(Var, Functor,
+	{ map__lookup(VarTypes, Var, Type) },
+	instmap__bind_var_to_functor(Var, Type, Functor, InstMap0, InstMap),
+	recompute_instmap_delta(Atomic, Goal0, Goal, VarTypes, InstMap,
+		InstMapDelta0),
+	instmap_delta_bind_var_to_functor(Var, Type, Functor,
 		InstMap0, InstMapDelta0, InstMapDelta1),
 	{ Case = case(Functor, Goal) },
 	recompute_instmap_delta_cases(Atomic, Var, Cases0, Cases,
-		InstMap0, NonLocals, InstMapDelta2),
+		VarTypes, InstMap0, NonLocals, InstMapDelta2),
 	merge_instmap_delta(InstMap0, NonLocals, InstMapDelta1,
 		InstMapDelta2, InstMapDelta).
 
 %-----------------------------------------------------------------------------%
 
 :- pred recompute_instmap_delta_call(pred_id, proc_id,
-		list(var), instmap, instmap_delta, module_info, module_info).
+		list(prog_var), instmap, instmap_delta,
+		module_info, module_info).
 :- mode recompute_instmap_delta_call(in, in, in, in, out, in, out) is det.
 
 recompute_instmap_delta_call(PredId, ProcId, Args, InstMap,
@@ -1224,7 +1326,7 @@ recompute_instmap_delta_call(PredId, ProcId, Args, InstMap,
 			ModuleInfo, InstMapDelta)
 	).
 
-:- pred recompute_instmap_delta_call_2(list(var), instmap, list(mode),
+:- pred recompute_instmap_delta_call_2(list(prog_var), instmap, list(mode),
 		list(mode), module_info, module_info).
 :- mode recompute_instmap_delta_call_2(in, in, in, out, in, out) is det.
 
@@ -1245,17 +1347,16 @@ recompute_instmap_delta_call_2([Arg | Args], InstMap, [Mode0 | Modes0],
 				ModuleInfo1)
 		->
 			ModuleInfo2 = ModuleInfo1,
-			Mode = (ArgInst0 -> UnifyInst),
-			recompute_instmap_delta_call_2(Args, InstMap,
-				Modes0, Modes, ModuleInfo2, ModuleInfo)
+			Mode = (ArgInst0 -> UnifyInst)
 		;
 			error("recompute_instmap_delta_call_2: unify_inst failed")
 		)
 	;
 		Mode = (not_reached -> not_reached),
-		Modes = [],
-		ModuleInfo = ModuleInfo0
-	).
+		ModuleInfo2 = ModuleInfo0
+	),
+	recompute_instmap_delta_call_2(Args, InstMap,
+		Modes0, Modes, ModuleInfo2, ModuleInfo).
 
 :- pred recompute_instmap_delta_unify(unification, unify_mode, unify_mode,
 	hlds_goal_info, instmap, instmap_delta, module_info, module_info).
@@ -1315,7 +1416,7 @@ get_arg_lives([Mode|Modes], ModuleInfo, [IsLive|IsLives]) :-
 
 	% 
 	% Predicates to make error messages more readable by stripping
-	% "mercury_builtin" module qualifiers from modes and insts.
+	% "builtin:" module qualifiers from modes and insts.
 	% The interesting part is strip_builtin_qualifier_from_sym_name;
 	% the rest is basically just recursive traversals.
 	%
@@ -1346,7 +1447,12 @@ strip_builtin_qualifier_from_cons_id(ConsId0, ConsId) :-
 						sym_name::out) is det.
 
 strip_builtin_qualifier_from_sym_name(SymName0, SymName) :-
-	( SymName0 = qualified("mercury_builtin", Name) ->
+	(
+		SymName0 = qualified(Module, Name),
+		( mercury_public_builtin_module(Module)
+		; mercury_private_builtin_module(Module)
+		)
+	->
 		SymName = unqualified(Name)
 	;
 		SymName = SymName0
@@ -1417,27 +1523,39 @@ strip_builtin_qualifiers_from_pred_inst(yes(Pred0), yes(Pred)) :-
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
-normalise_insts([], _, []).
-normalise_insts([Inst0|Insts0], ModuleInfo, [Inst|Insts]) :-
-	normalise_inst(Inst0, ModuleInfo, Inst),
-	normalise_insts(Insts0, ModuleInfo, Insts).
+normalise_insts([], [], _, []).
+normalise_insts([Inst0|Insts0], [Type|Types], ModuleInfo, [Inst|Insts]) :-
+	normalise_inst(Inst0, Type, ModuleInfo, Inst),
+	normalise_insts(Insts0, Types, ModuleInfo, Insts).
+normalise_insts([], [_|_], _, _) :-
+	error("normalise_insts: length mismatch").
+normalise_insts([_|_], [], _, _) :-
+	error("normalise_insts: length mismatch").
 
 	% This is a bit of a hack.
 	% The aim is to avoid non-termination due to the creation
 	% of ever-expanding insts.
 	% XXX should also normalise partially instantiated insts.
 
-normalise_inst(Inst0, ModuleInfo, NormalisedInst) :-
+normalise_inst(Inst0, Type, ModuleInfo, NormalisedInst) :-
 	inst_expand(ModuleInfo, Inst0, Inst),
 	( Inst = bound(_, _) ->
 		(
 			inst_is_ground(ModuleInfo, Inst),
-			inst_is_unique(ModuleInfo, Inst)
+			inst_is_unique(ModuleInfo, Inst),
+			% don't infer unique modes for introduced type_infos
+			% arguments, because that leads to an increase
+			% in the number of inferred modes without any benefit
+			\+ is_introduced_type_info_type(Type)
 		->
 			NormalisedInst = ground(unique, no)
 		;
 			inst_is_ground(ModuleInfo, Inst),
-			inst_is_mostly_unique(ModuleInfo, Inst)
+			inst_is_mostly_unique(ModuleInfo, Inst),
+			% don't infer unique modes for introduced type_infos
+			% arguments, because that leads to an increase
+			% in the number of inferred modes without any benefit
+			\+ is_introduced_type_info_type(Type)
 		->
 			NormalisedInst = ground(mostly_unique, no)
 		;
@@ -1470,5 +1588,47 @@ fixup_switch_var(Var, InstMap0, InstMap, Goal0, Goal) :-
 	),
 	Goal = GoalExpr - GoalInfo.
 
+%-----------------------------------------------------------------------------%
+
+in_mode(Mode) :- make_std_mode("in", [], Mode).
+
+out_mode(Mode) :- make_std_mode("out", [], Mode).
+
+uo_mode(Mode) :- make_std_mode("uo", [], Mode).
+
+unused_mode(Mode) :- make_std_mode("unused", [], Mode).
+
+aditi_ui_mode = Mode :- in_mode(Mode). 
+
+aditi_di_mode = Mode :- in_mode(Mode).
+
+aditi_uo_mode = Mode :- out_mode(Mode).
+
+:- pred make_std_mode(string, list(inst), mode).
+:- mode make_std_mode(in, in, out) is det.
+
+make_std_mode(Name, Args, Mode) :-
+	mercury_public_builtin_module(MercuryBuiltin),
+	QualifiedName = qualified(MercuryBuiltin, Name),
+	Mode = user_defined_mode(QualifiedName, Args).
+
+%-----------------------------------------------------------------------------%
+
+partition_args(_, [], [_|_], _, _) :-
+        error("partition_args").
+partition_args(_, [_|_], [], _, _) :-
+	error("partition_args").
+partition_args(_, [], [], [], []).
+partition_args(ModuleInfo, [ArgMode | ArgModes], [Arg | Args],
+		InputArgs, OutputArgs) :-
+	partition_args(ModuleInfo, ArgModes, Args, InputArgs1, OutputArgs1),
+	( mode_is_input(ModuleInfo, ArgMode) ->
+		InputArgs = [Arg | InputArgs1],
+		OutputArgs = OutputArgs1
+	;
+		InputArgs = InputArgs1,
+		OutputArgs = [Arg | OutputArgs1]
+	).
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
