@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1997 The University of Melbourne.
+% Copyright (C) 1996-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -21,85 +21,62 @@
 :- module modecheck_call.
 :- interface.
 
-:- import_module hlds_goal, mode_info.
-:- import_module term.
+:- import_module hlds_goal, hlds_pred, hlds_module.
+:- import_module prog_data, modes, mode_info.
+:- import_module list, std_util.
 
-:- pred modecheck_call_pred(pred_id, list(var), proc_id, list(var),
-				extra_goals, mode_info, mode_info).
-:- mode modecheck_call_pred(in, in, out, out, out,
-				mode_info_di, mode_info_uo) is det.
-
-:- pred modecheck_higher_order_call(pred_or_func, var, list(var),
-				list(type), list(mode), determinism, list(var),
-				extra_goals, mode_info, mode_info).
-:- mode modecheck_higher_order_call(in, in, in, out, out, out, out, out,
-				mode_info_di, mode_info_uo) is det.
-
-:- pred modecheck_higher_order_pred_call(var, list(var), pred_or_func,
-		hlds_goal_info, hlds_goal_expr, mode_info, mode_info).
-:- mode modecheck_higher_order_pred_call(in, in, in, in, out,
+:- pred modecheck_call_pred(pred_id, proc_id, list(prog_var),
+		maybe(determinism), proc_id, list(prog_var),
+		extra_goals, mode_info, mode_info).
+:- mode modecheck_call_pred(in, in, in, in, out, out, out,
 		mode_info_di, mode_info_uo) is det.
 
-:- pred modecheck_higher_order_func_call(var, list(var), var, hlds_goal_info,
-		hlds_goal_expr, mode_info, mode_info).
-:- mode modecheck_higher_order_func_call(in, in, in, in, out,
+:- pred modecheck_higher_order_call(pred_or_func, prog_var, list(prog_var),
+		list(mode), determinism, list(prog_var),
+		extra_goals, mode_info, mode_info).
+:- mode modecheck_higher_order_call(in, in, in, out, out, out, out,
 		mode_info_di, mode_info_uo) is det.
 
+:- pred modecheck_aditi_builtin(aditi_builtin, simple_call_id,
+		list(prog_var), list(mode), determinism,
+		list(prog_var), extra_goals, mode_info, mode_info).
+:- mode modecheck_aditi_builtin(in, in, in, in, out, out, out,
+		mode_info_di, mode_info_uo) is det.
+
+	%
+	% Given two modes of a predicate, figure out whether
+	% they are indistinguishable; that is, whether any valid call to
+	% one mode would also be a valid call to the other.
+	% (If so, it is a mode error.)
+	% Note that mode declarations which only have different final insts
+	% do not count as distinguishable.
+	%
+:- pred modes_are_indistinguishable(proc_id, proc_id, pred_info, module_info).
+:- mode modes_are_indistinguishable(in, in, in, in) is semidet.
+
+	%
+	% Given two modes of a predicate, figure out whether
+	% they are identical, except that one is cc_nondet/cc_multi
+	% and the other is nondet/multi.
+	% This is used by determinism analysis to substitute
+	% a multi mode for a cc_multi one if the call occurs in a
+	% non-cc context.
+	%
+:- pred modes_are_identical_bar_cc(proc_id, proc_id, pred_info, module_info).
+:- mode modes_are_identical_bar_cc(in, in, in, in) is semidet.
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
 :- implementation.
-:- import_module prog_data, hlds_pred, hlds_data, hlds_module, instmap, (inst).
+:- import_module hlds_data, instmap, prog_data, (inst).
 :- import_module mode_info, mode_debug, modes, mode_util, mode_errors.
 :- import_module clause_to_proc, inst_match, make_hlds.
-:- import_module det_report.
-:- import_module map, list, bool, std_util, set, require.
-
-modecheck_higher_order_pred_call(PredVar, Args0, PredOrFunc, GoalInfo0, Goal)
-		-->
-	mode_checkpoint(enter, "higher-order call"),
-	mode_info_set_call_context(higher_order_call(PredOrFunc)),
-	=(ModeInfo0),
-
-	{ mode_info_get_instmap(ModeInfo0, InstMap0) },
-	modecheck_higher_order_call(PredOrFunc, PredVar, Args0,
-			Types, Modes, Det, Args, ExtraGoals),
-
-	=(ModeInfo),
-	{ Call = higher_order_call(PredVar, Args, Types, Modes, Det,
-			PredOrFunc) },
-	{ handle_extra_goals(Call, ExtraGoals, GoalInfo0,
-			[PredVar | Args0], [PredVar | Args],
-			InstMap0, ModeInfo, Goal) },
-	mode_info_unset_call_context,
-	mode_checkpoint(exit, "higher-order predicate call").
-
-modecheck_higher_order_func_call(FuncVar, Args0, RetVar, GoalInfo0, Goal) -->
-	mode_checkpoint(enter, "higher-order function call"),
-	mode_info_set_call_context(higher_order_call(function)),
-
-	=(ModeInfo0),
-	{ mode_info_get_instmap(ModeInfo0, InstMap0) },
-
-	{ list__append(Args0, [RetVar], Args1) },
-	modecheck_higher_order_call(function, FuncVar, Args1,
-			Types, Modes, Det, Args, ExtraGoals),
-
-	=(ModeInfo),
-	{ Call = higher_order_call(FuncVar, Args, Types, Modes, Det,
-				function) },
-	{ handle_extra_goals(Call, ExtraGoals, GoalInfo0,
-				[FuncVar | Args1], [FuncVar | Args],
-				InstMap0, ModeInfo, Goal) },
-
-	mode_info_unset_call_context,
-	mode_checkpoint(exit, "higher-order function call").
-
-modecheck_higher_order_call(PredOrFunc, PredVar, Args0, Types, Modes, Det, Args,
-		ExtraGoals, ModeInfo0, ModeInfo) :-
-
-	mode_info_get_types_of_vars(ModeInfo0, Args0, Types),
+:- import_module det_report, unify_proc.
+:- import_module int, map, bool, set, require.
 
+modecheck_higher_order_call(PredOrFunc, PredVar, Args0, Modes, Det,
+		Args, ExtraGoals, ModeInfo0, ModeInfo) :-
 	%
 	% First, check that `PredVar' has a higher-order pred inst
 	% (of the appropriate arity)
@@ -116,31 +93,15 @@ modecheck_higher_order_call(PredOrFunc, PredVar, Args0, Types, Modes, Det, Args,
 	->
 		Det = Det0,
 		Modes = Modes0,
-
-		%
-		% Check that `Args0' have livenesses which match the
-		% expected livenesses.
-		%
-		get_arg_lives(Modes, ModuleInfo0, ExpectedArgLives),
-		modecheck_var_list_is_live(Args0, ExpectedArgLives, 1,
+		ArgOffset = 1,
+		modecheck_arg_list(ArgOffset, Modes, ExtraGoals, Args0, Args,
 			ModeInfo0, ModeInfo1),
 
-		%
-		% Check that `Args0' have insts which match the expected
-		% initial insts, and set their new final insts (introducing
-		% extra unifications for implied modes, if necessary).
-		%
-		mode_list_get_initial_insts(Modes, ModuleInfo0, InitialInsts),
-		modecheck_var_has_inst_list(Args0, InitialInsts, 1,
-					ModeInfo1, ModeInfo2),
-		mode_list_get_final_insts(Modes, ModuleInfo0, FinalInsts),
-		modecheck_set_var_inst_list(Args0, InitialInsts, FinalInsts,
-			Args, ExtraGoals, ModeInfo2, ModeInfo3),
 		( determinism_components(Det, _, at_most_zero) ->
 			instmap__init_unreachable(Instmap),
-			mode_info_set_instmap(Instmap, ModeInfo3, ModeInfo)
+			mode_info_set_instmap(Instmap, ModeInfo1, ModeInfo)
 		;
-			ModeInfo = ModeInfo3
+			ModeInfo = ModeInfo1
 		)
 	;
 		% the error occurred in argument 1, i.e. the pred term
@@ -155,25 +116,95 @@ modecheck_higher_order_call(PredOrFunc, PredVar, Args0, Types, Modes, Det, Args,
 		ExtraGoals = no_extra_goals
 	).
 
-modecheck_call_pred(PredId, ArgVars0, TheProcId, ArgVars, ExtraGoals,
+modecheck_aditi_builtin(AditiBuiltin, CallId,
+		Args0, Modes, Det, Args, ExtraGoals) -->
+	{ aditi_builtin_determinism(AditiBuiltin, Det) },
+
+	% `aditi_insert' goals have type_info arguments for each
+	% of the arguments of the tuple to insert added to the
+	% start of the argument list by polymorphism.m.
+	( { AditiBuiltin = aditi_insert(_) } ->
+		{ CallId = _ - _/Arity },
+		{ ArgOffset = -Arity }
+	;
+		{ ArgOffset = 0 }
+	),
+
+	% The argument modes are set by post_typecheck.m, so all
+	% that needs to be done here is to check that they match.
+	modecheck_arg_list(ArgOffset, Modes, ExtraGoals, Args0, Args).
+
+:- pred aditi_builtin_determinism(aditi_builtin, determinism).
+:- mode aditi_builtin_determinism(in, out) is det.
+
+aditi_builtin_determinism(aditi_call(_, _, _, _), _) :-
+	error(
+	"modecheck_call__aditi_builtin_determinism: unexpected Aditi call"). 
+aditi_builtin_determinism(aditi_insert(_), det).
+aditi_builtin_determinism(aditi_delete(_, _), det).
+aditi_builtin_determinism(aditi_bulk_operation(_, _), det).
+aditi_builtin_determinism(aditi_modify(_, _), det).
+
+:- pred modecheck_arg_list(int, list(mode), extra_goals,
+		list(prog_var), list(prog_var), mode_info, mode_info).
+:- mode modecheck_arg_list(in, in, out, in, out,
+		mode_info_di, mode_info_uo) is det.
+
+modecheck_arg_list(ArgOffset, Modes, ExtraGoals, Args0, Args,
 		ModeInfo0, ModeInfo) :-
 
-		% Get the list of different possible modes for the called
-		% predicate
+	%
+	% Check that `Args0' have livenesses which match the
+	% expected livenesses.
+	%
+	mode_info_get_module_info(ModeInfo0, ModuleInfo0),
+	get_arg_lives(Modes, ModuleInfo0, ExpectedArgLives),
+	modecheck_var_list_is_live(Args0, ExpectedArgLives, ArgOffset,
+		ModeInfo0, ModeInfo1),
+
+	%
+	% Check that `Args0' have insts which match the expected
+	% initial insts, and set their new final insts (introducing
+	% extra unifications for implied modes, if necessary).
+	%
+	mode_list_get_initial_insts(Modes, ModuleInfo0, InitialInsts),
+	modecheck_var_has_inst_list(Args0, InitialInsts, ArgOffset,
+				ModeInfo1, ModeInfo2),
+	mode_list_get_final_insts(Modes, ModuleInfo0, FinalInsts),
+	modecheck_set_var_inst_list(Args0, InitialInsts, FinalInsts,
+		ArgOffset, Args, ExtraGoals, ModeInfo2, ModeInfo).
+
+modecheck_call_pred(PredId, ProcId0, ArgVars0, DeterminismKnown,
+		TheProcId, ArgVars, ExtraGoals, ModeInfo0, ModeInfo) :-
+
+	mode_info_get_may_change_called_proc(ModeInfo0, MayChangeCalledProc),
+
 	mode_info_get_preds(ModeInfo0, Preds),
 	mode_info_get_module_info(ModeInfo0, ModuleInfo),
-	map__lookup(Preds, PredId, PredInfo0),
-	maybe_add_default_mode(PredInfo0, PredInfo),
+	map__lookup(Preds, PredId, PredInfo),
 	pred_info_procedures(PredInfo, Procs),
-	map__keys(Procs, ProcIds),
-	pred_info_get_marker_list(PredInfo, Markers),
+
+	( MayChangeCalledProc = may_not_change_called_proc ->
+		( invalid_proc_id(ProcId0) ->
+			error("modecheck_call_pred: invalid proc_id")
+		;
+			ProcIds = [ProcId0]
+		)
+	;
+			% Get the list of different possible
+			% modes for the called predicate
+		map__keys(Procs, ProcIds)
+	),
+
+	compute_arg_offset(PredInfo, ArgOffset),
+	pred_info_get_markers(PredInfo, Markers),
 
 		% In order to give better diagnostics, we handle the
 		% cases where there are zero or one modes for the called
 		% predicate specially.
 	(
 		ProcIds = [],
-		\+ list__member(request(infer_modes), Markers)
+		\+ check_marker(Markers, infer_modes)
 	->
 		set__init(WaitingVars),
 		mode_info_error(WaitingVars, mode_error_no_mode_decl,
@@ -183,16 +214,19 @@ modecheck_call_pred(PredId, ArgVars0, TheProcId, ArgVars, ExtraGoals,
 		ExtraGoals = no_extra_goals
 	;
 		ProcIds = [ProcId],
-		\+ list__member(request(infer_modes), Markers)
+		( \+ check_marker(Markers, infer_modes)
+		; MayChangeCalledProc = may_not_change_called_proc
+		)
 	->
 		TheProcId = ProcId,
 		map__lookup(Procs, ProcId, ProcInfo),
+
 		%
 		% Check that `ArgsVars0' have livenesses which match the
 		% expected livenesses.
 		%
 		proc_info_arglives(ProcInfo, ModuleInfo, ProcArgLives0),
-		modecheck_var_list_is_live(ArgVars0, ProcArgLives0, 0,
+		modecheck_var_list_is_live(ArgVars0, ProcArgLives0, ArgOffset,
 					ModeInfo0, ModeInfo1),
 
 		%
@@ -203,10 +237,10 @@ modecheck_call_pred(PredId, ArgVars0, TheProcId, ArgVars, ExtraGoals,
 		proc_info_argmodes(ProcInfo, ProcArgModes),
 		mode_list_get_initial_insts(ProcArgModes, ModuleInfo,
 					InitialInsts),
-		modecheck_var_has_inst_list(ArgVars0, InitialInsts, 0,
+		modecheck_var_has_inst_list(ArgVars0, InitialInsts, ArgOffset,
 					ModeInfo1, ModeInfo2),
 
-		modecheck_end_of_call(ProcInfo, ArgVars0, ArgVars,
+		modecheck_end_of_call(ProcInfo, ArgVars0, ArgOffset, ArgVars,
 					ExtraGoals, ModeInfo2, ModeInfo)
 	;
 			% set the current error list to empty (and
@@ -222,7 +256,8 @@ modecheck_call_pred(PredId, ArgVars0, TheProcId, ArgVars, ExtraGoals,
 			ModeInfo1, ModeInfo2),
 
 		(	RevMatchingProcIds = [],
-			no_matching_modes(PredId, ArgVars0, WaitingVars,
+			no_matching_modes(PredId, ArgVars0,
+				DeterminismKnown, WaitingVars,
 				TheProcId, ModeInfo2, ModeInfo3),
 			ArgVars = ArgVars0,
 			ExtraGoals = no_extra_goals
@@ -230,10 +265,10 @@ modecheck_call_pred(PredId, ArgVars0, TheProcId, ArgVars, ExtraGoals,
 			RevMatchingProcIds = [_|_],
 			list__reverse(RevMatchingProcIds, MatchingProcIds),
 			choose_best_match(MatchingProcIds, PredId, Procs,
-				TheProcId, ModeInfo2),
+				ArgVars0, TheProcId, ModeInfo2),
 			map__lookup(Procs, TheProcId, ProcInfo),
-			modecheck_end_of_call(ProcInfo, ArgVars0, ArgVars,
-				ExtraGoals, ModeInfo2, ModeInfo3)
+			modecheck_end_of_call(ProcInfo, ArgVars0, ArgOffset,
+				ArgVars, ExtraGoals, ModeInfo2, ModeInfo3)
 		),
 
 			% restore the error list, appending any new error(s)
@@ -242,11 +277,12 @@ modecheck_call_pred(PredId, ArgVars0, TheProcId, ArgVars, ExtraGoals,
 		mode_info_set_errors(Errors, ModeInfo3, ModeInfo)
 	).
 
-:- pred no_matching_modes(pred_id, list(var), set(var), proc_id, 	
-				mode_info, mode_info).
-:- mode no_matching_modes(in, in, in, out, mode_info_di, mode_info_uo) is det.
+:- pred no_matching_modes(pred_id, list(prog_var), maybe(determinism),
+		set(prog_var), proc_id, mode_info, mode_info).
+:- mode no_matching_modes(in, in, in, in, out, mode_info_di, mode_info_uo)
+	is det.
 
-no_matching_modes(PredId, ArgVars, WaitingVars, TheProcId,
+no_matching_modes(PredId, ArgVars, DeterminismKnown, WaitingVars, TheProcId,
 		ModeInfo0, ModeInfo) :-
 	%
 	% There were no matching modes.
@@ -256,9 +292,9 @@ no_matching_modes(PredId, ArgVars, WaitingVars, TheProcId,
 	%
 	mode_info_get_preds(ModeInfo0, Preds),
 	map__lookup(Preds, PredId, PredInfo),
-	pred_info_get_marker_list(PredInfo, Markers),
-	( list__member(request(infer_modes), Markers) ->
-		insert_new_mode(PredId, ArgVars, TheProcId,
+	pred_info_get_markers(PredInfo, Markers),
+	( check_marker(Markers, infer_modes) ->
+		insert_new_mode(PredId, ArgVars, DeterminismKnown, TheProcId,
 			ModeInfo0, ModeInfo1),
 		% we don't yet know the final insts for the newly created mode
 		% of the called predicate, so we set the instmap to unreachable,
@@ -277,9 +313,9 @@ no_matching_modes(PredId, ArgVars, WaitingVars, TheProcId,
 	).
 
 :- pred modecheck_find_matching_modes(
-			list(proc_id), pred_id, proc_table, list(var),
-			list(proc_id), list(proc_id), set(var), set(var),
-			mode_info, mode_info).
+		list(proc_id), pred_id, proc_table, list(prog_var),
+		list(proc_id), list(proc_id), set(prog_var), set(prog_var),
+		mode_info, mode_info).
 :- mode modecheck_find_matching_modes(in, in, in, in,
 			in, out, in, out, mode_info_di, mode_info_uo) is det.
 
@@ -332,19 +368,19 @@ modecheck_find_matching_modes([ProcId | ProcIds], PredId, Procs, ArgVars0,
 			MatchingProcIds1, MatchingProcIds,
 			WaitingVars1, WaitingVars, ModeInfo3, ModeInfo).
 
-:- pred modecheck_end_of_call(proc_info, list(var), list(var), extra_goals,
-				mode_info, mode_info).
-:- mode modecheck_end_of_call(in, in, out, out,
+:- pred modecheck_end_of_call(proc_info, list(prog_var), int,
+		list(prog_var), extra_goals, mode_info, mode_info).
+:- mode modecheck_end_of_call(in, in, in, out, out,
 				mode_info_di, mode_info_uo) is det.
 
-modecheck_end_of_call(ProcInfo, ArgVars0, ArgVars, ExtraGoals,
-			ModeInfo0, ModeInfo) :-
+modecheck_end_of_call(ProcInfo, ArgVars0, ArgOffset,
+			ArgVars, ExtraGoals, ModeInfo0, ModeInfo) :-
 	proc_info_argmodes(ProcInfo, ProcArgModes),
 	mode_info_get_module_info(ModeInfo0, ModuleInfo),
 	mode_list_get_initial_insts(ProcArgModes, ModuleInfo, InitialInsts),
 	mode_list_get_final_insts(ProcArgModes, ModuleInfo, FinalInsts),
 	modecheck_set_var_inst_list(ArgVars0, InitialInsts, FinalInsts,
-			ArgVars, ExtraGoals, ModeInfo0, ModeInfo1),
+			ArgOffset, ArgVars, ExtraGoals, ModeInfo0, ModeInfo1),
 	proc_info_never_succeeds(ProcInfo, NeverSucceeds),
 	( NeverSucceeds = yes ->
 		instmap__init_unreachable(Instmap),
@@ -353,8 +389,9 @@ modecheck_end_of_call(ProcInfo, ArgVars0, ArgVars, ExtraGoals,
 		ModeInfo = ModeInfo1
 	).
 
-:- pred insert_new_mode(pred_id, list(var), proc_id, mode_info, mode_info).
-:- mode insert_new_mode(in, in, out, mode_info_di, mode_info_uo) is det.
+:- pred insert_new_mode(pred_id, list(prog_var), maybe(determinism), proc_id,
+			mode_info, mode_info).
+:- mode insert_new_mode(in, in, in, out, mode_info_di, mode_info_uo) is det.
 
 	% Insert a new inferred mode for a predicate.
 	% The initial insts are determined by using a normalised
@@ -362,7 +399,7 @@ modecheck_end_of_call(ProcInfo, ArgVars0, ArgVars, ExtraGoals,
 	% The final insts are initially just assumed to be all `not_reached'.
 	% The determinism for this mode will be inferred.
 
-insert_new_mode(PredId, ArgVars, ProcId, ModeInfo0, ModeInfo) :-
+insert_new_mode(PredId, ArgVars, MaybeDet, ProcId, ModeInfo0, ModeInfo) :-
 	% figure out the values of all the variables we need to
 	% create a new mode for this predicate
 	get_var_insts_and_lives(ArgVars, ModeInfo0, InitialInsts, ArgLives),
@@ -373,26 +410,15 @@ insert_new_mode(PredId, ArgVars, ProcId, ModeInfo0, ModeInfo) :-
 	list__length(ArgVars, Arity),
 	list__duplicate(Arity, not_reached, FinalInsts),
 	inst_lists_to_mode_list(InitialInsts, FinalInsts, Modes),
-	MaybeDeterminism = no,
 
-	% create the new mode
-	add_new_proc(PredInfo0, Arity, Modes, no, yes(ArgLives),
-		MaybeDeterminism, Context, PredInfo1, ProcId),
+	%
+	% call unify_proc__request_proc, which will
+	% create the new procedure, set its "can-process" flag to `no',
+	% and insert it into the queue of requested procedures.
+	%
+	unify_proc__request_proc(PredId, Modes, yes(ArgLives), MaybeDet,
+		Context, ModuleInfo0, ProcId, ModuleInfo),
 
-	% copy the clauses for the predicate to this procedure,
-	% and then store the new proc_info and pred_info
-	% back in the module_info.
-
-	pred_info_procedures(PredInfo1, Procs1),
-	map__lookup(Procs1, ProcId, ProcInfo1),
-	pred_info_clauses_info(PredInfo1, ClausesInfo),
-
-	copy_clauses_to_proc(ProcId, ClausesInfo, ProcInfo1, ProcInfo),
-
-	map__det_update(Procs1, ProcId, ProcInfo, Procs),
-	pred_info_set_procedures(PredInfo1, Procs, PredInfo),
-	map__det_update(Preds0, PredId, PredInfo, Preds),
-	module_info_set_preds(ModuleInfo0, Preds, ModuleInfo),
 	mode_info_set_module_info(ModeInfo0, ModuleInfo, ModeInfo1),
 
 	% Since we've created a new inferred mode for this predicate,
@@ -400,7 +426,7 @@ insert_new_mode(PredId, ArgVars, ProcId, ModeInfo0, ModeInfo) :-
 	% pass of the fixpoint analysis.
 	mode_info_set_changed_flag(yes, ModeInfo1, ModeInfo).
 
-:- pred get_var_insts_and_lives(list(var), mode_info,
+:- pred get_var_insts_and_lives(list(prog_var), mode_info,
 				list(inst), list(is_live)).
 :- mode get_var_insts_and_lives(in, mode_info_ui, out, out) is det.
 
@@ -409,8 +435,10 @@ get_var_insts_and_lives([Var | Vars], ModeInfo,
 		[Inst | Insts], [IsLive | IsLives]) :-
 	mode_info_get_module_info(ModeInfo, ModuleInfo),
 	mode_info_get_instmap(ModeInfo, InstMap),
+	mode_info_get_var_types(ModeInfo, VarTypes),
 	instmap__lookup_var(InstMap, Var, Inst0),
-	normalise_inst(Inst0, ModuleInfo, Inst),
+	map__lookup(VarTypes, Var, Type),
+	normalise_inst(Inst0, Type, ModuleInfo, Inst),
 
 	mode_info_var_is_live(ModeInfo, Var, IsLive0),
 
@@ -424,7 +452,7 @@ get_var_insts_and_lives([Var | Vars], ModeInfo,
 		% a good chance of being able to do destructive update.
 		(
 			inst_is_ground(ModuleInfo, Inst),
-			inst_is_unique(ModuleInfo, Inst)
+			inst_is_mostly_unique(ModuleInfo, Inst)
 		->
 			IsLive = dead
 		;
@@ -434,6 +462,115 @@ get_var_insts_and_lives([Var | Vars], ModeInfo,
 
 	get_var_insts_and_lives(Vars, ModeInfo, Insts, IsLives).
 
+%-----------------------------------------------------------------------------%
+
+	%
+	% Given two modes of a predicate, figure out whether
+	% they are indistinguishable; that is, whether any valid call to
+	% one mode would also be a valid call to the other.
+	% (If so, it is a mode error.)
+	% Note that mode declarations which only have different final insts
+	% do not count as distinguishable.
+	%
+	% The code for this is similar to the code for
+	% modes_are_indentical/4 and compare_proc/5 below.
+	%
+modes_are_indistinguishable(ProcId, OtherProcId, PredInfo, ModuleInfo) :-
+	pred_info_procedures(PredInfo, Procs),
+	map__lookup(Procs, ProcId, ProcInfo),
+	map__lookup(Procs, OtherProcId, OtherProcInfo),
+
+	%
+	% Compare the initial insts of the arguments
+	%
+	proc_info_argmodes(ProcInfo, ProcArgModes),
+	proc_info_argmodes(OtherProcInfo, OtherProcArgModes),
+	mode_list_get_initial_insts(ProcArgModes, ModuleInfo, InitialInsts),
+	mode_list_get_initial_insts(OtherProcArgModes, ModuleInfo,
+							OtherInitialInsts),
+	compare_inst_list(InitialInsts, OtherInitialInsts, no,
+		CompareInsts, ModuleInfo),
+	CompareInsts = same,
+
+	%
+	% Compare the expected livenesses of the arguments
+	%
+	get_arg_lives(ProcArgModes, ModuleInfo, ProcArgLives),
+	get_arg_lives(OtherProcArgModes, ModuleInfo, OtherProcArgLives),
+	compare_liveness_list(ProcArgLives, OtherProcArgLives, CompareLives),
+	CompareLives = same,
+
+	%
+	% Compare the determinisms --
+	%	If both are cc_, or if both are not cc_,
+	%	then they are indistinguishable.
+	%
+	proc_info_interface_determinism(ProcInfo, Detism),
+	proc_info_interface_determinism(OtherProcInfo, OtherDetism),
+	determinism_components(Detism, _CanFail, Solns),
+	determinism_components(OtherDetism, _OtherCanFail, OtherSolns),
+	( Solns = at_most_many_cc, OtherSolns = at_most_many_cc
+	; Solns \= at_most_many_cc, OtherSolns \= at_most_many_cc
+	).
+
+%-----------------------------------------------------------------------------%
+
+	%
+	% Given two modes of a predicate, figure out whether
+	% they are identical, except that one is cc_nondet/cc_multi
+	% and the other is nondet/multi.
+	%
+	% The code for this is similar to the code for compare_proc/5 below
+	% and modes_are_indistinguishable/4 above.
+	%
+modes_are_identical_bar_cc(ProcId, OtherProcId, PredInfo, ModuleInfo) :-
+	pred_info_procedures(PredInfo, Procs),
+	map__lookup(Procs, ProcId, ProcInfo),
+	map__lookup(Procs, OtherProcId, OtherProcInfo),
+
+	%
+	% Compare the initial insts of the arguments
+	%
+	proc_info_argmodes(ProcInfo, ProcArgModes),
+	proc_info_argmodes(OtherProcInfo, OtherProcArgModes),
+	mode_list_get_initial_insts(ProcArgModes, ModuleInfo, InitialInsts),
+	mode_list_get_initial_insts(OtherProcArgModes, ModuleInfo,
+							OtherInitialInsts),
+	compare_inst_list(InitialInsts, OtherInitialInsts, no,
+		CompareInitialInsts, ModuleInfo),
+	CompareInitialInsts = same,
+
+	%
+	% Compare the final insts of the arguments
+	%
+	mode_list_get_final_insts(ProcArgModes, ModuleInfo, FinalInsts),
+	mode_list_get_final_insts(OtherProcArgModes, ModuleInfo,
+							OtherFinalInsts),
+	compare_inst_list(FinalInsts, OtherFinalInsts, no,
+		CompareFinalInsts, ModuleInfo),
+	CompareFinalInsts = same,
+
+	%
+	% Compare the expected livenesses of the arguments
+	%
+	get_arg_lives(ProcArgModes, ModuleInfo, ProcArgLives),
+	get_arg_lives(OtherProcArgModes, ModuleInfo, OtherProcArgLives),
+	compare_liveness_list(ProcArgLives, OtherProcArgLives, CompareLives),
+	CompareLives = same,
+
+	%
+	% Compare the determinisms, ignoring the cc part.
+	%
+	proc_info_interface_determinism(ProcInfo, Detism),
+	proc_info_interface_determinism(OtherProcInfo, OtherDetism),
+	determinism_components(Detism, CanFail, Solns),
+	determinism_components(OtherDetism, OtherCanFail, OtherSolns),
+	CanFail = OtherCanFail,
+	( Solns = OtherSolns
+	; Solns = at_most_many_cc, OtherSolns = at_most_many
+	; Solns = at_most_many, OtherSolns = at_most_many_cc
+	).
+
 %-----------------------------------------------------------------------------%
 
 /*
@@ -444,13 +581,16 @@ to the following specification:
 	    less informative on input than other valid modes; eg,
 	    prefer an (in, in, out) mode over an (out, in, out) mode,
 	    but not necessarily over an (out, out, in) mode,
-	    and prefer a (free -> ...) mode over a (any -> ...) mode,
+	    and prefer a (ground -> ...) mode over a (any -> ...) mode,
 	    and prefer a (bound(f) -> ...) mode over a (ground -> ...) mode,
 	    and prefer a (... -> dead) mode over a (... -> not dead) mode.
 
-	    This is a partial order.
+	    Also prefer a (any -> ...) mode over a (free -> ...) mode,
+	    unless the actual argument is free, in which case prefer
+	    the (free -> ...) mode.
 
- 	2.  Prioritize them by determinism, according to the standard
+ 	2.  If neither is prefered over the other by step 1, then
+	    prioritize them by determinism, according to the standard
 	    partial order (best first):
 
  				erroneous
@@ -471,14 +611,14 @@ to the following specification:
 	;	same
 	;	incomparable.
 
-:- pred choose_best_match(list(proc_id), pred_id, proc_table, proc_id,
-				mode_info).
-:- mode choose_best_match(in, in, in, out,
+:- pred choose_best_match(list(proc_id), pred_id, proc_table, list(prog_var),
+				proc_id, mode_info).
+:- mode choose_best_match(in, in, in, in, out,
 				mode_info_ui) is det.
 
-choose_best_match([], _, _, _, _) :-
+choose_best_match([], _, _, _, _, _) :-
 	error("choose_best_match: no best match").
-choose_best_match([ProcId | ProcIds], PredId, Procs, TheProcId,
+choose_best_match([ProcId | ProcIds], PredId, Procs, ArgVars, TheProcId,
 			ModeInfo) :-
 	%
 	% This ProcId is best iff there is no other proc_id which is better.
@@ -486,24 +626,30 @@ choose_best_match([ProcId | ProcIds], PredId, Procs, TheProcId,
 	(
 		\+ (
 			list__member(OtherProcId, ProcIds),
-			compare_proc(OtherProcId, ProcId, better,
+			compare_proc(OtherProcId, ProcId, ArgVars, better,
 					Procs, ModeInfo)
 		)
 	->
 		TheProcId = ProcId
 	;
-		choose_best_match(ProcIds, PredId, Procs, TheProcId, ModeInfo)
+		choose_best_match(ProcIds, PredId, Procs, ArgVars, TheProcId,
+				ModeInfo)
 	).
 
 	%
 	% Given two modes of a predicate, figure out whether
-	% one of the is a better match than the other,
+	% one of them is a better match than the other,
 	% for calls which could match either mode.
 	%
-:- pred compare_proc(proc_id, proc_id, match, proc_table, mode_info).
-:- mode compare_proc(in, in, out, in, mode_info_ui) is det.
+	% The code for this is similar to the code for
+	% modes_are_indistinguishable/4 and
+	% modes_are_identical_bar_cc/4 above.
+	%
+:- pred compare_proc(proc_id, proc_id, list(prog_var), match, proc_table,
+		mode_info).
+:- mode compare_proc(in, in, in, out, in, mode_info_ui) is det.
 
-compare_proc(ProcId, OtherProcId, Compare, Procs, ModeInfo) :-
+compare_proc(ProcId, OtherProcId, ArgVars, Compare, Procs, ModeInfo) :-
 	map__lookup(Procs, ProcId, ProcInfo),
 	map__lookup(Procs, OtherProcId, OtherProcInfo),
 	%
@@ -515,8 +661,9 @@ compare_proc(ProcId, OtherProcId, Compare, Procs, ModeInfo) :-
 	mode_list_get_initial_insts(ProcArgModes, ModuleInfo, InitialInsts),
 	mode_list_get_initial_insts(OtherProcArgModes, ModuleInfo,
 							OtherInitialInsts),
-	compare_inst_list(InitialInsts, OtherInitialInsts, CompareInsts,
-		ModuleInfo),
+	get_var_insts_and_lives(ArgVars, ModeInfo, ArgInitialInsts, _ArgLives),
+	compare_inst_list(InitialInsts, OtherInitialInsts, yes(ArgInitialInsts),
+		CompareInsts, ModuleInfo),
 	%
 	% Compare the expected livenesses of the arguments
 	%
@@ -540,17 +687,31 @@ compare_proc(ProcId, OtherProcId, Compare, Procs, ModeInfo) :-
 	combine_results(CompareInsts, CompareLives, Compare0),
 	prioritized_combine_results(Compare0, CompareDet, Compare).
 
-:- pred compare_inst_list(list(inst), list(inst), match, module_info).
-:- mode compare_inst_list(in, in, out, in) is det.
+:- pred compare_inst_list(list(inst), list(inst), maybe(list(inst)), match,
+				module_info).
+:- mode compare_inst_list(in, in, in, out, in) is det.
 
-compare_inst_list([], [], same, _).
-compare_inst_list([_|_], [], _, _) :-
-	error("compare_inst_list: length mis-match").
-compare_inst_list([], [_|_], _, _) :-
-	error("compare_inst_list: length mis-match").
-compare_inst_list([InstA | InstsA], [InstB | InstsB], Result, ModuleInfo) :-
-	compare_inst(InstA, InstB, Result0, ModuleInfo),
-	compare_inst_list(InstsA, InstsB, Result1, ModuleInfo),
+compare_inst_list(InstsA, InstsB, ArgInsts, Result, ModuleInfo) :-
+	( compare_inst_list_2(InstsA, InstsB, ArgInsts, Result0, ModuleInfo) ->
+		Result = Result0
+	;
+		error("compare_inst_list: length mis-match")
+	).
+
+:- pred compare_inst_list_2(list(inst), list(inst), maybe(list(inst)), match,
+				module_info).
+:- mode compare_inst_list_2(in, in, in, out, in) is semidet.
+
+compare_inst_list_2([], [], _, same, _).
+compare_inst_list_2([InstA | InstsA], [InstB | InstsB],
+		no, Result, ModuleInfo) :-
+	compare_inst(InstA, InstB, no, Result0, ModuleInfo),
+	compare_inst_list_2(InstsA, InstsB, no, Result1, ModuleInfo),
+	combine_results(Result0, Result1, Result).
+compare_inst_list_2([InstA | InstsA], [InstB | InstsB],
+		yes([ArgInst|ArgInsts]), Result, ModuleInfo) :-
+	compare_inst(InstA, InstB, yes(ArgInst), Result0, ModuleInfo),
+	compare_inst_list_2(InstsA, InstsB, yes(ArgInsts), Result1, ModuleInfo),
 	combine_results(Result0, Result1, Result).
 
 :- pred compare_liveness_list(list(is_live), list(is_live), match).
@@ -625,10 +786,10 @@ combine_results(incomparable, _, incomparable).
 	% 	prefer ground to any	(e.g. prefer in to in(any))
 	% 	prefer any to free	(e.g. prefer any->ground to out)
 
-:- pred compare_inst(inst, inst, match, module_info).
-:- mode compare_inst(in, in, out, in) is det.
+:- pred compare_inst(inst, inst, maybe(inst), match, module_info).
+:- mode compare_inst(in, in, in, out, in) is det.
 
-compare_inst(InstA, InstB, Result, ModuleInfo) :-
+compare_inst(InstA, InstB, MaybeArgInst, Result, ModuleInfo) :-
 	% inst_matches_initial(A,B) succeeds iff
 	%	A specifies at least as much information
 	%	and at least as much binding as B --
@@ -651,24 +812,59 @@ compare_inst(InstA, InstB, Result, ModuleInfo) :-
 		%
 		% We need to further disambiguate the cases involving
 		% `any' and `free', since `any' matches_initial `free'
-		% and vice versa, but we want to prefer `any'.
-		% We use matches_final, because `free' may match_final `any',
-		% but `any' does not match_final `free'.
+		% and vice versa.  For these cases, we want to take
+		% the actual inst of the argument into account:
+		% if the argument is `free', we should prefer `free',
+		% but otherwise, we should prefer `any'.
 		%
-		( inst_matches_final(InstA, InstB, ModuleInfo) ->
-			A_mf_B = yes
+		(
+			MaybeArgInst = no,
+			Result0 = same
 		;
-			A_mf_B = no
+			MaybeArgInst = yes(ArgInst),
+			(
+				inst_matches_final(ArgInst, InstA, ModuleInfo)
+			->
+				Arg_mf_A = yes
+			;
+				Arg_mf_A = no
+			),
+			(
+				inst_matches_final(ArgInst, InstB, ModuleInfo)
+			->
+				Arg_mf_B = yes
+			;
+				Arg_mf_B = no
+			),
+			( Arg_mf_A = yes, Arg_mf_B = no,  Result0 = better
+			; Arg_mf_A = no,  Arg_mf_B = yes, Result0 = worse
+			; Arg_mf_A = yes, Arg_mf_B = yes, Result0 = same
+			; Arg_mf_A = no,  Arg_mf_B = no,  Result0 = same
+			)
 		),
-		( inst_matches_final(InstB, InstA, ModuleInfo) ->
-			B_mf_A = yes
+		( Result0 = same ->
+			%
+			% if the actual arg inst is not available,
+			% or comparing with the arg inst doesn't help,
+			% then compare the two proc insts
+			%
+			( inst_matches_final(InstA, InstB, ModuleInfo) ->
+				A_mf_B = yes
+			;
+				A_mf_B = no
+			),
+			( inst_matches_final(InstB, InstA, ModuleInfo) ->
+				B_mf_A = yes
+			;
+				B_mf_A = no
+			),
+			( A_mf_B = yes, B_mf_A = no,  Result = better
+			; A_mf_B = no,  B_mf_A = yes, Result = worse
+			; A_mf_B = no,  B_mf_A = no,  Result = incomparable
+			; A_mf_B = yes, B_mf_A = yes, Result = same
+			)
 		;
-			B_mf_A = no
-		),
-		( A_mf_B = yes, B_mf_A = no,  Result = worse
-		; A_mf_B = no,  B_mf_A = yes, Result = better
-		; A_mf_B = yes, B_mf_A = yes, Result = same
-		; A_mf_B = no,  B_mf_A = no,  Result = incomparable
+			Result = Result0
 		)
 	).
 
