@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -15,7 +15,7 @@
 :- interface.
 
 :- import_module vn_type, vn_table, livemap.
-:- import_module llds, atsort.
+:- import_module llds, builtin_ops, atsort.
 
 :- import_module io, bool, list, assoc_list, std_util.
 
@@ -185,7 +185,9 @@
 
 :- implementation.
 
-:- import_module llds_out, opt_util, vn_util, hlds_pred.
+:- import_module llds_out, opt_util, vn_util, hlds_pred, globals, options.
+:- import_module prog_out.
+
 :- import_module int, set, map, string.
 
 opt_debug__msg(OptDebug, Msg) -->
@@ -201,18 +203,19 @@ opt_debug__msg(OptDebug, Msg) -->
 opt_debug__dump_instrs(OptDebug, Instrs) -->
 	(
 		{ OptDebug = yes },
-		opt_debug__dump_instrs_2(Instrs)
+		globals__io_lookup_bool_option(auto_comments, PrintComments),
+		opt_debug__dump_instrs_2(Instrs, PrintComments)
 	;
 		{ OptDebug = no }
 	).
 
-:- pred opt_debug__dump_instrs_2(list(instruction), io__state, io__state).
-:- mode opt_debug__dump_instrs_2(in, di, uo) is det.
+:- pred opt_debug__dump_instrs_2(list(instruction), bool, io__state, io__state).
+:- mode opt_debug__dump_instrs_2(in, in, di, uo) is det.
 
-opt_debug__dump_instrs_2([]) --> [].
-opt_debug__dump_instrs_2([Uinstr - _ | Instrs]) -->
-	output_instruction(Uinstr),
-	opt_debug__dump_instrs_2(Instrs).
+opt_debug__dump_instrs_2([], _PrintComments) --> [].
+opt_debug__dump_instrs_2([Uinstr - Comment | Instrs], PrintComments) -->
+	output_instruction_and_comment(Uinstr, Comment, PrintComments),
+	opt_debug__dump_instrs_2(Instrs, PrintComments).
 
 opt_debug__dump_node_relmap(Relmap, Str) :-
 	map__to_assoc_list(Relmap, Nodemap),
@@ -291,11 +294,11 @@ opt_debug__dump_ctrl_list([N - VnInstr | Ctrllist], Str) :-
 
 opt_debug__dump_vninstr(vn_livevals(_), Str) :-
 	string__append_list(["livevals(...)"], Str).
-opt_debug__dump_vninstr(vn_call(Proc, Ret, _, _), Str) :-
+opt_debug__dump_vninstr(vn_call(Proc, Ret, _, _, _), Str) :-
 	opt_debug__dump_code_addr(Proc, P_str),
 	opt_debug__dump_code_addr(Ret, R_str),
 	string__append_list(["call(", P_str, ", ", R_str, ")"], Str).
-opt_debug__dump_vninstr(vn_mkframe(_, _, _), "mkframe").
+opt_debug__dump_vninstr(vn_mkframe(_, _), "mkframe").
 opt_debug__dump_vninstr(vn_label(Label), Str) :-
 	opt_debug__dump_label(Label, L_str),
 	string__append_list(["label(", L_str, ")"], Str).
@@ -500,6 +503,9 @@ opt_debug__dump_vnlval(vn_prevfr(V), Str) :-
 opt_debug__dump_vnlval(vn_succfr(V), Str) :-
 	string__int_to_string(V, V_str),
 	string__append_list(["vn_succfr(", V_str, ")"], Str).
+opt_debug__dump_vnlval(vn_redofr(V), Str) :-
+	string__int_to_string(V, V_str),
+	string__append_list(["vn_redofr(", V_str, ")"], Str).
 opt_debug__dump_vnlval(vn_redoip(V), Str) :-
 	string__int_to_string(V, V_str),
 	string__append_list(["vn_redoip(", V_str, ")"], Str).
@@ -510,8 +516,12 @@ opt_debug__dump_vnlval(vn_hp, Str) :-
 	string__append_list(["vn_hp"], Str).
 opt_debug__dump_vnlval(vn_sp, Str) :-
 	string__append_list(["vn_sp"], Str).
-opt_debug__dump_vnlval(vn_field(T, N, F), Str) :-
-	string__int_to_string(T, T_str),
+opt_debug__dump_vnlval(vn_field(MT, N, F), Str) :-
+	( MT = yes(T) ->
+		string__int_to_string(T, T_str)
+	;
+		T_str = "no"
+	),
 	string__int_to_string(N, N_str),
 	string__int_to_string(F, F_str),
 	string__append_list(["vn_field(", T_str, ", ", N_str, ", ",
@@ -530,7 +540,7 @@ opt_debug__dump_vnrval(vn_mkword(T, N), Str) :-
 opt_debug__dump_vnrval(vn_const(C), Str) :-
 	opt_debug__dump_const(C, C_str),
 	string__append_list(["vn_const(", C_str, ")"], Str).
-opt_debug__dump_vnrval(vn_create(T, MA, _U, L), Str) :-
+opt_debug__dump_vnrval(vn_create(T, MA, _TA, _U, L, _M), Str) :-
 	string__int_to_string(T, T_str),
 	opt_debug__dump_maybe_rvals(MA, 3, MA_str),
 	string__int_to_string(L, L_str),
@@ -580,6 +590,9 @@ opt_debug__dump_lval(succfr(R), Str) :-
 opt_debug__dump_lval(prevfr(R), Str) :-
 	opt_debug__dump_rval(R, R_str),
 	string__append_list(["prevfr(", R_str, ")"], Str).
+opt_debug__dump_lval(redofr(R), Str) :-
+	opt_debug__dump_rval(R, R_str),
+	string__append_list(["redofr(", R_str, ")"], Str).
 opt_debug__dump_lval(redoip(R), Str) :-
 	opt_debug__dump_rval(R, R_str),
 	string__append_list(["redoip(", R_str, ")"], Str).
@@ -590,8 +603,12 @@ opt_debug__dump_lval(hp, Str) :-
 	string__append_list(["hp"], Str).
 opt_debug__dump_lval(sp, Str) :-
 	string__append_list(["sp"], Str).
-opt_debug__dump_lval(field(T, N, F), Str) :-
-	string__int_to_string(T, T_str),
+opt_debug__dump_lval(field(MT, N, F), Str) :-
+	( MT = yes(T) ->
+		string__int_to_string(T, T_str)
+	;
+		T_str = "no"
+	),
 	opt_debug__dump_rval(N, N_str),
 	opt_debug__dump_rval(F, F_str),
 	string__append_list(["field(", T_str, ", ", N_str, ", ",
@@ -617,15 +634,18 @@ opt_debug__dump_rval(mkword(T, N), Str) :-
 opt_debug__dump_rval(const(C), Str) :-
 	opt_debug__dump_const(C, C_str),
 	string__append_list(["const(", C_str, ")"], Str).
-opt_debug__dump_rval(create(T, MA, U, L), Str) :-
+opt_debug__dump_rval(create(T, MA, _, U, L, _), Str) :-
 	string__int_to_string(T, T_str),
 	opt_debug__dump_maybe_rvals(MA, 3, MA_str),
 	(
-		U = yes,
-		U_str = "yes"
+		U = must_be_static,
+		U_str = "static"
 	;
-		U = no,
-		U_str = "no"
+		U = can_be_either,
+		U_str = "either"
+	;
+		U = must_be_dynamic,
+		U_str = "dynamic"
 	),
 	string__int_to_string(L, L_str),
 	string__append_list(["create(", T_str, ", ", MA_str, ", ",
@@ -669,21 +689,39 @@ opt_debug__dump_const(int_const(I), Str) :-
 	string__int_to_string(I, Str).
 opt_debug__dump_const(float_const(F), Str) :-
 	string__float_to_string(F, Str).
-opt_debug__dump_const(string_const(I), Str) :-
-	string__append_list(["""", I, """"], Str).
+opt_debug__dump_const(string_const(S), Str) :-
+	string__append_list(["""", S, """"], Str).
+opt_debug__dump_const(multi_string_const(L, _S), Str) :-
+	string__int_to_string(L, L_str),
+	string__append_list(["multi_string(", L_str, ")"], Str).
 opt_debug__dump_const(code_addr_const(CodeAddr), Str) :-
 	opt_debug__dump_code_addr(CodeAddr, C_str),
 	string__append_list(["code_addr_const(", C_str, ")"], Str).
 opt_debug__dump_const(data_addr_const(data_addr(BaseName, VarName)), Str) :-
 	opt_debug__dump_data_name(VarName, N_str),
-	string__append_list(["data_addr_const(", BaseName, ", ", N_str, ")"],
-		Str).
-
+	prog_out__sym_name_to_string(BaseName, BaseName_str),
+	string__append_list(
+		["data_addr_const(", BaseName_str, ", ", N_str, ")"], Str).
+opt_debug__dump_const(label_entry(Label), Str) :-
+	opt_debug__dump_label(Label, LabelStr),
+	string__append_list(["label_entry(", LabelStr, ")"], Str).
 opt_debug__dump_data_name(common(N), Str) :-
 	string__int_to_string(N, N_str),
 	string__append("common", N_str, Str).
-opt_debug__dump_data_name(base_type(BaseData, TypeName, TypeArity), Str) :-
-	llds_out__make_base_type_name(BaseData, TypeName, TypeArity, Str).
+opt_debug__dump_data_name(type_ctor(BaseData, TypeName, TypeArity), Str) :-
+	llds_out__make_type_ctor_name(BaseData, TypeName, TypeArity, Str).
+opt_debug__dump_data_name(base_typeclass_info(ClassId, InstanceNum), Str) :-
+	llds_out__make_base_typeclass_info_name(ClassId, InstanceNum, Str).
+opt_debug__dump_data_name(module_layout, "module_layout").
+opt_debug__dump_data_name(proc_layout(Label), Str) :-
+	opt_debug__dump_label(Label, LabelStr),
+	string__append_list(["proc_layout(", LabelStr, ")"], Str).
+opt_debug__dump_data_name(internal_layout(Label), Str) :-
+	opt_debug__dump_label(Label, LabelStr),
+	string__append_list(["internal_layout(", LabelStr, ")"], Str).
+opt_debug__dump_data_name(tabling_pointer(ProcLabel), Str) :-
+	opt_debug__dump_proclabel(ProcLabel, ProcLabelStr),
+	string__append_list(["tabling_pointer(", ProcLabelStr, ")"], Str).
 
 opt_debug__dump_unop(mktag, "mktag").
 opt_debug__dump_unop(tag, "tag").
@@ -729,9 +767,19 @@ opt_debug__dump_code_addr(do_succeed(Last), Str) :-
 	).
 opt_debug__dump_code_addr(do_redo, "do_redo").
 opt_debug__dump_code_addr(do_fail, "do_fail").
-opt_debug__dump_code_addr(do_det_closure, "do_det_closure").
-opt_debug__dump_code_addr(do_semidet_closure, "do_semidet_closure").
-opt_debug__dump_code_addr(do_nondet_closure, "do_nondet_closure").
+opt_debug__dump_code_addr(do_trace_redo_fail_shallow,
+	"do_trace_redo_fail_shallow").
+opt_debug__dump_code_addr(do_trace_redo_fail_deep, "do_trace_redo_fail_deep").
+opt_debug__dump_code_addr(do_call_closure, "do_nondet_closure").
+opt_debug__dump_code_addr(do_call_class_method, "do_nondet_class_method").
+opt_debug__dump_code_addr(do_det_aditi_call, "do_det_aditi_call").
+opt_debug__dump_code_addr(do_semidet_aditi_call, "do_semidet_aditi_call").
+opt_debug__dump_code_addr(do_nondet_aditi_call, "do_nondet_aditi_call").
+opt_debug__dump_code_addr(do_aditi_insert, "do_aditi_insert").
+opt_debug__dump_code_addr(do_aditi_delete, "do_aditi_delete").
+opt_debug__dump_code_addr(do_aditi_bulk_insert, "do_aditi_bulk_insert").
+opt_debug__dump_code_addr(do_aditi_bulk_delete, "do_aditi_bulk_delete").
+opt_debug__dump_code_addr(do_aditi_modify, "do_aditi_modify").
 opt_debug__dump_code_addr(do_not_reached, "do_not_reached").
 
 opt_debug__dump_code_addrs([], "").
@@ -766,23 +814,27 @@ opt_debug__dump_label_pairs([L1 - L2 | Labels], Str) :-
 
 opt_debug__dump_proclabel(proc(Module, _PredOrFunc, PredModule,
 		PredName, Arity, ProcId), Str) :-
-	string__int_to_string(Arity, A_str),
-	proc_id_to_int(ProcId, Mode),
-	string__int_to_string(Mode, M_str),
 	( Module = PredModule ->
 		ExtraModule = ""
 	;
-		string__append(Module, "_", ExtraModule)
+		llds_out__sym_name_mangle(PredModule, PredModuleName),
+		string__append(PredModuleName, "_", ExtraModule)
 	),
-	string__append_list([ExtraModule, Module, "_", PredName,
-		"_", A_str, "_", M_str], Str).
-opt_debug__dump_proclabel(special_proc(Module, Pred, TypeModule,
-		Type, Arity, ProcId), Str) :-
+	llds_out__sym_name_mangle(Module, ModuleName),
 	string__int_to_string(Arity, A_str),
 	proc_id_to_int(ProcId, Mode),
 	string__int_to_string(Mode, M_str),
-	llds_out__qualify_name(TypeModule, Type, TypeName),
-	string__append_list([Module, "_", Pred, "_",
+	string__append_list([ExtraModule, ModuleName, "_", PredName,
+		"_", A_str, "_", M_str], Str).
+opt_debug__dump_proclabel(special_proc(Module, Pred, TypeModule,
+		Type, Arity, ProcId), Str) :-
+	llds_out__sym_name_mangle(Module, ModuleName),
+	llds_out__sym_name_mangle(TypeModule, TypeModuleName),
+	llds_out__qualify_name(TypeModuleName, Type, TypeName),
+	string__int_to_string(Arity, A_str),
+	proc_id_to_int(ProcId, Mode),
+	string__int_to_string(Mode, M_str),
+	string__append_list([ModuleName, "_", Pred, "_",
 		TypeName, "_", A_str, "_", M_str], Str).
 
 opt_debug__dump_bool(yes, "yes").
@@ -805,18 +857,34 @@ opt_debug__dump_instr(assign(Lval, Rval), Str) :-
 	opt_debug__dump_lval(Lval, L_str),
 	opt_debug__dump_rval(Rval, R_str),
 	string__append_list(["assign(", L_str, ", ", R_str, ")"], Str).
-opt_debug__dump_instr(call(Proc, Ret, _, _), Str) :-
+opt_debug__dump_instr(call(Proc, Ret, _, _, _), Str) :-
 	opt_debug__dump_code_addr(Proc, P_str),
 	opt_debug__dump_code_addr(Ret, R_str),
 	string__append_list(["call(", P_str, ", ", R_str, ", ...)"], Str).
-opt_debug__dump_instr(mkframe(Name, Size, Redoip), Str) :-
-	string__int_to_string(Size, S_str),
+opt_debug__dump_instr(mkframe(FrameInfo, Redoip), Str) :-
 	opt_debug__dump_code_addr(Redoip, R_str),
-	string__append_list(["mkframe(", Name, ", ", S_str, ", ", R_str, ")"],
-		Str).
-opt_debug__dump_instr(modframe(Redoip), Str) :-
-	opt_debug__dump_code_addr(Redoip, R_str),
-	string__append_list(["modframe(", R_str, ")"], Str).
+	(
+		FrameInfo = ordinary_frame(Name, Size, MaybePragma),
+		string__int_to_string(Size, S_str),
+		( MaybePragma = yes(pragma_c_struct(StructName, Fields, _)) ->
+			string__append_list(["yes(", StructName, ", ",
+				Fields, ")"], P_str)
+		;
+			P_str = "no"
+		),
+		string__append_list(["mkframe(", Name, ", ", S_str, ", ",
+			P_str, ", ", R_str, ")"], Str)
+	;
+		FrameInfo = temp_frame(Kind),
+		(
+			Kind = nondet_stack_proc,
+			string__append_list(["mktempframe(", R_str, ")"], Str)
+		;
+			Kind = det_stack_proc,
+			string__append_list(["mkdettempframe(", R_str, ")"],
+				Str)
+		)
+	).
 opt_debug__dump_instr(label(Label), Str) :-
 	opt_debug__dump_label(Label, L_str),
 	string__append_list(["label(", L_str, ")"], Str).
@@ -833,7 +901,7 @@ opt_debug__dump_instr(if_val(Rval, CodeAddr), Str) :-
 	opt_debug__dump_rval(Rval, R_str),
 	opt_debug__dump_code_addr(CodeAddr, C_str),
 	string__append_list(["if_val(", R_str, ", ", C_str, ")"], Str).
-opt_debug__dump_instr(incr_hp(Lval, MaybeTag, Size), Str) :-
+opt_debug__dump_instr(incr_hp(Lval, MaybeTag, Size, _), Str) :-
 	opt_debug__dump_lval(Lval, L_str),
 	(
 		MaybeTag = no,
@@ -870,9 +938,48 @@ opt_debug__dump_instr(incr_sp(Size, _), Str) :-
 opt_debug__dump_instr(decr_sp(Size), Str) :-
 	string__int_to_string(Size, S_str),
 	string__append_list(["decr_sp(", S_str, ")"], Str).
+opt_debug__dump_instr(init_sync_term(Lval, N), Str) :-
+	opt_debug__dump_lval(Lval, L_str),
+	string__int_to_string(N, N_str),
+	string__append_list(["init_sync_term(", L_str, ", ", N_str, ")"], Str).
+opt_debug__dump_instr(fork(Child, Parent, Lval), Str) :-
+	opt_debug__dump_label(Child, ChildStr),
+	opt_debug__dump_label(Parent, ParentStr),
+	string__int_to_string(Lval, LvalStr),
+	string__append_list(["fork(", ChildStr, ", ", ParentStr, ", ",
+		LvalStr, ")"], Str).
+opt_debug__dump_instr(join_and_terminate(Lval), Str) :-
+	opt_debug__dump_lval(Lval, LvalStr),
+	string__append_list(["join_and_terminate(", LvalStr, ")"], Str).
+opt_debug__dump_instr(join_and_continue(Lval, Label), Str) :-
+	opt_debug__dump_lval(Lval, LvalStr),
+	opt_debug__dump_label(Label, LabelStr),
+	string__append_list(["join(", LvalStr, ", ", LabelStr, ")"], Str).
 % XXX  should probably give more info than this
-opt_debug__dump_instr(pragma_c(_, _, Code, _, _), Str) :-
-	string__append_list(["pragma_c(", Code, ")"], Str).
+opt_debug__dump_instr(pragma_c(_, Comps, _, _, _, _), Str) :-
+	opt_debug__dump_components(Comps, C_str),
+	string__append_list(["pragma_c(", C_str, ")"], Str).
+
+:- pred opt_debug__dump_components(list(pragma_c_component), string).
+:- mode opt_debug__dump_components(in, out) is det.
+
+opt_debug__dump_components([], "").
+opt_debug__dump_components([Comp | Comps], Str) :-
+	opt_debug__dump_component(Comp, Str1),
+	opt_debug__dump_components(Comps, Str2),
+	string__append(Str1, Str2, Str).
+
+:- pred opt_debug__dump_component(pragma_c_component, string).
+:- mode opt_debug__dump_component(in, out) is det.
+
+opt_debug__dump_component(pragma_c_inputs(_), "").
+opt_debug__dump_component(pragma_c_outputs(_), "").
+opt_debug__dump_component(pragma_c_user_code(_, Code), Code).
+opt_debug__dump_component(pragma_c_raw_code(Code), Code).
+opt_debug__dump_component(pragma_c_fail_to(Label), Code) :-
+	opt_debug__dump_label(Label, LabelStr),
+	string__append_list(["fail to ", LabelStr], Code).
+opt_debug__dump_component(pragma_c_noop, "").
 
 opt_debug__dump_fullinstr(Uinstr - Comment, Str) :-
 	opt_debug__dump_instr(Uinstr, U_str),
