@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -14,12 +14,17 @@
 :- interface.
 
 :- import_module hlds_module, hlds_pred, prog_data.
-:- import_module io.
+:- import_module io, std_util, list, bool.
 
 %-----------------------------------------------------------------------------%
 
 :- type task	--->	update_proc(pred(
 				proc_info, module_info, proc_info))
+		;	update_proc_predid(pred(
+				proc_info, pred_id, module_info, proc_info))
+		;	update_proc_predprocid(pred(
+				proc_info, pred_id, proc_id,
+				module_info, proc_info))
 		;	update_proc_io(pred(
 				pred_id, proc_id, module_info,
 				proc_info, proc_info, io__state, io__state))
@@ -27,6 +32,7 @@
 				pred_id, proc_id, module_info, module_info,
 				proc_info, proc_info, int, int,
 				io__state, io__state))
+		;	update_pred_error(pred_error_task)
 		;	update_module(pred(
 				proc_info, proc_info,
 				module_info, module_info))
@@ -34,15 +40,20 @@
 				pred_id, proc_id, proc_info, proc_info,
 				module_info, module_info,
 				io__state, io__state))
-		% It would be better to use an existentiallly-quantified type
-		% rather than `univ' here, but Mercury 0.6 doesn't support
-		% existentially-quantified types.
+		% It would be better to use an existentially-quantified type
+		% rather than `univ' here, but the current version of Mercury 
+		% doesn't support existentially-quantified types.
 		;	update_module_cookie(pred(
 				pred_id, proc_id, proc_info, proc_info,
 				univ, univ, module_info, module_info),
 				univ)
 		.
 
+
+:- type pred_error_task ==
+		pred(pred_id, module_info, module_info, pred_info, pred_info,
+			int, int, io__state, io__state).
+
 /****************
 
 Note that update_module_cookie causes some difficulties.
@@ -73,9 +84,15 @@ about unbound type variables.
 ****************/
 
 :- inst task =	bound(( update_proc(pred(in, in, out) is det)
-		;	update_proc_io(pred(in, in, in, in, out, di, uo) is det)
+		;	update_proc_predid(pred(in, in, in, out) is det)
+		;	update_proc_predprocid(pred(in, in, in, in, out)
+				is det)
+		;	update_proc_io(pred(in, in, in, in, out, di, uo)
+				is det)
 		;	update_proc_error(pred(in, in, in, out, in, out,
 				out, out, di, uo) is det)
+		;	update_pred_error(pred(in, in, out, in, out,
+				out, out, di, uo) is det)
 		;	update_module(pred(in, out, in, out) is det)
 		;	update_module_io(pred(in, in, in, out,
 				in, out, di, uo) is det)
@@ -83,12 +100,35 @@ about unbound type variables.
 				in, out) is det, ground)
 		)).
 
+:- inst pred_error_task =
+		(pred(in, in, out, in, out, out, out, di, uo) is det).
+
 :- mode task ::	task -> task.
 
 :- pred process_all_nonimported_procs(task, module_info, module_info,
 	io__state, io__state).
 :- mode process_all_nonimported_procs(task, in, out, di, uo) is det.
 
+	% Process procedures for which a given test succeeds.
+:- pred process_matching_nonimported_procs(task, pred(pred_info),
+	module_info, module_info, io__state, io__state).
+:- mode process_matching_nonimported_procs(task, pred(in) is semidet,
+	in, out, di, uo) is det.
+
+:- pred process_matching_nonimported_procs(task, task, pred(pred_info),
+	module_info, module_info, io__state, io__state).
+:- mode process_matching_nonimported_procs(task, out(task),
+	pred(in) is semidet, in, out, di, uo) is det.
+
+:- pred process_all_nonimported_nonaditi_procs(task, module_info, module_info,
+	io__state, io__state).
+:- mode process_all_nonimported_nonaditi_procs(task, in, out, di, uo) is det.
+
+:- pred process_all_nonimported_nonaditi_procs(task, task,
+	module_info, module_info, io__state, io__state).
+:- mode process_all_nonimported_nonaditi_procs(task, out(task),
+	in, out, di, uo) is det.
+
 :- pred process_all_nonimported_procs(task, task,
 	module_info, module_info, io__state, io__state).
 :- mode process_all_nonimported_procs(task, out(task), in, out, di, uo) is det.
@@ -114,48 +154,107 @@ about unbound type variables.
 
 
 :- pred report_pred_proc_id(module_info, pred_id, proc_id, 
-		maybe(term__context), term__context, io__state, io__state).
+		maybe(prog_context), prog_context, io__state, io__state).
 :- mode report_pred_proc_id(in, in, in, in, out, di, uo) is det.
 
 :- pred report_pred_name_mode(pred_or_func, string, list((mode)),
 				io__state, io__state).
 :- mode report_pred_name_mode(in, in, in, di, uo) is det.
 	
-
 %-----------------------------------------------------------------------------%
 
 :- implementation.
 
 :- import_module options, globals, hlds_out, prog_out, mode_util.
 :- import_module mercury_to_mercury.
-:- import_module bool, int, map, tree234, std_util, require, list.
 :- import_module varset.
+:- import_module int, map, tree234, require.
 
 process_all_nonimported_procs(Task, ModuleInfo0, ModuleInfo) -->
-	{ module_info_predids(ModuleInfo0, PredIds) },
-	process_nonimported_procs_in_preds(PredIds, Task, _,
+	{ True = lambda([_PredInfo::in] is semidet, true) },
+	process_matching_nonimported_procs(Task, True, 
+		ModuleInfo0, ModuleInfo).
+
+process_all_nonimported_nonaditi_procs(Task, ModuleInfo0, ModuleInfo) -->
+	{ NotAditi = lambda([PredInfo::in] is semidet, (
+			\+ hlds_pred__pred_info_is_aditi_relation(PredInfo)
+		)) }, 
+	process_matching_nonimported_procs(Task, NotAditi, 
+		ModuleInfo0, ModuleInfo).
+
+process_all_nonimported_nonaditi_procs(Task0, Task,
+		ModuleInfo0, ModuleInfo) -->
+	{ NotAditi = lambda([PredInfo::in] is semidet, (
+			\+ hlds_pred__pred_info_is_aditi_relation(PredInfo)
+		)) }, 
+	process_matching_nonimported_procs(Task0, Task, NotAditi, 
 		ModuleInfo0, ModuleInfo).
 
 process_all_nonimported_procs(Task0, Task, ModuleInfo0, ModuleInfo) -->
-	{ module_info_predids(ModuleInfo0, PredIds) },
-	process_nonimported_procs_in_preds(PredIds, Task0, Task,
+	{ True = lambda([_PredInfo::in] is semidet, true) },
+	process_matching_nonimported_procs(Task0, Task, True, 
 		ModuleInfo0, ModuleInfo).
 
-:- pred process_nonimported_procs_in_preds(list(pred_id), task, task,
-	module_info, module_info, io__state, io__state).
-:- mode process_nonimported_procs_in_preds(in, task, out(task), in, out,
-	di, uo) is det.
+process_matching_nonimported_procs(Task, Filter, ModuleInfo0, ModuleInfo) -->
+	{ module_info_predids(ModuleInfo0, PredIds) },
+	( { Task = update_pred_error(Pred) } ->
+		list__foldl2(process_nonimported_pred(Pred, Filter), PredIds,
+			ModuleInfo0, ModuleInfo)
+	;
+		process_nonimported_procs_in_preds(PredIds, Task, _, Filter,
+			ModuleInfo0, ModuleInfo)
+	).
 
-process_nonimported_procs_in_preds([], Task, Task, ModuleInfo, ModuleInfo)
+process_matching_nonimported_procs(Task0, Task, Filter, 
+		ModuleInfo0, ModuleInfo) -->
+	{ module_info_predids(ModuleInfo0, PredIds) },
+	process_nonimported_procs_in_preds(PredIds, Task0, Task, Filter,
+		ModuleInfo0, ModuleInfo).
+
+:- pred process_nonimported_pred(pred_error_task, pred(pred_info), pred_id, 
+	module_info, module_info, io__state, io__state).
+:- mode process_nonimported_pred(in(pred_error_task), pred(in) is semidet, in,
+	in, out, di, uo) is det.
+
+process_nonimported_pred(Task, Filter, PredId, ModuleInfo0, ModuleInfo,
+		IO0, IO) :-
+	module_info_pred_info(ModuleInfo0, PredId, PredInfo0),
+	(
+		( pred_info_is_imported(PredInfo0)
+		; \+ call(Filter, PredInfo0)
+		)
+	->
+		ModuleInfo = ModuleInfo0,
+		IO = IO0
+	;
+		call(Task, PredId, ModuleInfo0, ModuleInfo1,
+			PredInfo0, PredInfo, WarnCnt, ErrCnt, IO0, IO1),
+		module_info_set_pred_info(ModuleInfo1,
+			PredId, PredInfo, ModuleInfo2),
+		passes_aux__handle_errors(WarnCnt, ErrCnt,
+			ModuleInfo2, ModuleInfo, IO1, IO)
+	).
+
+:- pred process_nonimported_procs_in_preds(list(pred_id), task, task,
+	pred(pred_info), module_info, module_info, io__state, io__state).
+:- mode process_nonimported_procs_in_preds(in, task, out(task), 
+	pred(in) is semidet, in, out, di, uo) is det.
+
+process_nonimported_procs_in_preds([], Task, Task, _, ModuleInfo, ModuleInfo)
 		--> [].
-process_nonimported_procs_in_preds([PredId | PredIds], Task0, Task,
+process_nonimported_procs_in_preds([PredId | PredIds], Task0, Task, Filter,
 		ModuleInfo0, ModuleInfo) -->
 	{ module_info_preds(ModuleInfo0, PredTable) },
 	{ map__lookup(PredTable, PredId, PredInfo) },
-	{ pred_info_non_imported_procids(PredInfo, ProcIds) },
-	process_nonimported_procs(ProcIds, PredId, Task0, Task1,
-		ModuleInfo0, ModuleInfo1),
-	process_nonimported_procs_in_preds(PredIds, Task1, Task,
+	( { call(Filter, PredInfo) } ->
+		{ pred_info_non_imported_procids(PredInfo, ProcIds) },
+		process_nonimported_procs(ProcIds, PredId, Task0, Task1,
+			ModuleInfo0, ModuleInfo1)
+	;
+		{ ModuleInfo1 = ModuleInfo0 },
+		{ Task1 = Task0 }
+	),
+	process_nonimported_procs_in_preds(PredIds, Task1, Task, Filter,
 		ModuleInfo1, ModuleInfo).
 
 :- pred process_nonimported_procs(list(proc_id), pred_id, task, task,
@@ -189,6 +288,18 @@ process_nonimported_procs([ProcId | ProcIds], PredId, Task0, Task,
 		ModuleInfo8 = ModuleInfo0,
 		Task1 = Task0,
 		State9 = State0
+	;
+		Task0 = update_proc_predid(Closure),
+		call(Closure, Proc0, PredId, ModuleInfo0, Proc),
+		ModuleInfo8 = ModuleInfo0,
+		Task1 = Task0,
+		State9 = State0
+	;
+		Task0 = update_proc_predprocid(Closure),
+		call(Closure, Proc0, PredId, ProcId, ModuleInfo0, Proc),
+		ModuleInfo8 = ModuleInfo0,
+		Task1 = Task0,
+		State9 = State0
 	;
 		Task0 = update_proc_io(Closure),
 		call(Closure, PredId, ProcId, ModuleInfo0,
@@ -199,23 +310,12 @@ process_nonimported_procs([ProcId | ProcIds], PredId, Task0, Task,
 		Task0 = update_proc_error(Closure),
 		call(Closure, PredId, ProcId, ModuleInfo0, ModuleInfo1,
 			Proc0, Proc, WarnCnt, ErrCnt, State0, State1),
-		globals__io_lookup_bool_option(halt_at_warn, HaltAtWarn,
-			State1, State2),
 		Task1 = Task0,
-		(
-			(
-				ErrCnt > 0
-			;
-				WarnCnt > 0,
-				HaltAtWarn = yes
-			)
-		->
-			io__set_exit_status(1, State2, State9),
-			module_info_incr_errors(ModuleInfo1, ModuleInfo8)
-		;
-			ModuleInfo8 = ModuleInfo1,
-			State9 = State2
-		)
+		passes_aux__handle_errors(WarnCnt, ErrCnt,
+			ModuleInfo1, ModuleInfo8, State1, State9)
+	;
+		Task0 = update_pred_error(_),
+		error("passes_aux:process_non_imported_procs")
 	;
 		Task0 = update_module_cookie(Closure, Cookie0),
 		call(Closure, PredId, ProcId, Proc0, Proc,
@@ -275,11 +375,39 @@ report_error(ErrorMessage) -->
 	io__write_string("\n"),
 	io__set_exit_status(1).
 
+:- pred passes_aux__handle_errors(int, int, module_info, module_info,
+		io__state, io__state).
+:- mode passes_aux__handle_errors(in, in, in, out, di, uo) is det.
+
+passes_aux__handle_errors(WarnCnt, ErrCnt, ModuleInfo1, ModuleInfo8,
+		State1, State9) :-
+	globals__io_lookup_bool_option(halt_at_warn, HaltAtWarn,
+		State1, State2),
+	(
+		(
+			ErrCnt > 0
+		;
+			WarnCnt > 0,
+			HaltAtWarn = yes
+		)
+	->
+		io__set_exit_status(1, State2, State9),
+		module_info_incr_errors(ModuleInfo1, ModuleInfo8)
+	;
+		ModuleInfo8 = ModuleInfo1,
+		State9 = State2
+	).
+
 invoke_system_command(Command, Succeeded) -->
 	globals__io_lookup_bool_option(verbose, Verbose),
-	maybe_write_string(Verbose, "% Invoking system command `"),
-	maybe_write_string(Verbose, Command),
-	maybe_write_string(Verbose, "'...\n"),
+	( { Verbose = yes } ->
+		io__write_string("% Invoking system command `"),
+		io__write_string(Command),
+		io__write_string("'...\n"),
+		io__flush_output
+	;
+		[]
+	),
 	io__call_system(Command, Result),
 	( { Result = ok(0) } ->
 		maybe_write_string(Verbose, "% done.\n"),
@@ -315,18 +443,14 @@ report_sizes(ModuleInfo) -->
 :- mode tree_stats(in, in, di, uo) is det.
 
 tree_stats(Description, Tree) -->
-	{ tree234__count(Tree, Count) },
-	% { tree234__depth(Tree, Depth) },
+	{ map__count(Tree, Count) },
 	io__write_string(Description),
 	io__write_string(": count = "),
 	io__write_int(Count),
-	% io__write_string(", depth = "),
-	% io__write_int(Depth),
 	io__write_string("\n").
 
 %-----------------------------------------------------------------------------%
 
-
 report_pred_proc_id(ModuleInfo, PredId, ProcId, MaybeContext, Context) -->
 	{ module_info_pred_proc_info(ModuleInfo, PredId, ProcId,
 		PredInfo, ProcInfo) },
@@ -385,4 +509,5 @@ report_pred_name_mode(function, FuncName, ArgModes) -->
 	),
 	io__write_string(" = "),
 	mercury_output_mode(FuncRetMode, InstVarSet).
+
 %-----------------------------------------------------------------------------%
