@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1993-1997 The University of Melbourne.
+% Copyright (C) 1993-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -18,28 +18,72 @@
 %-----------------------------------------------------------------------------%
 
 :- interface.
-:- import_module prog_data, io, term.
+:- import_module prog_data.
+:- import_module list, io.
 
 :- pred prog_out__write_messages(message_list, io__state, io__state).
 :- mode prog_out__write_messages(in, di, uo) is det.
 
-:- pred prog_out__write_context(term__context, io__state, io__state).
+:- pred prog_out__write_context(prog_context, io__state, io__state).
 :- mode prog_out__write_context(in, di, uo) is det.
 
+:- pred prog_out__context_to_string(prog_context, string).
+:- mode prog_out__context_to_string(in, out) is det.
+
+	% XXX This pred should be deleted, and all uses replaced with
+	% XXX error_util:write_error_pieces, once zs has committed that
+	% XXX error_util.m.
+:- pred prog_out__write_strings_with_context(prog_context, list(string),
+	io__state, io__state).
+:- mode prog_out__write_strings_with_context(in, in, di, uo) is det.
+
+	% Write out a symbol name, with special characters escaped,
+	% but without any quotes.  This is suitable for use in
+	% error messages, where the caller should print out an
+	% enclosing forward/backward-quote pair (`...').
 :- pred prog_out__write_sym_name(sym_name, io__state, io__state).
 :- mode prog_out__write_sym_name(in, di, uo) is det.
 
+:- pred prog_out__write_sym_name_and_arity(sym_name_and_arity,
+		io__state, io__state).
+:- mode prog_out__write_sym_name_and_arity(in, di, uo) is det.
+
+	% Write out a symbol name, enclosed in single forward quotes ('...')
+	% if necessary, and with any special characters escaped.
+	% The output should be a syntactically valid Mercury term.
+:- pred prog_out__write_quoted_sym_name(sym_name, io__state, io__state).
+:- mode prog_out__write_quoted_sym_name(in, di, uo) is det.
+
+	% sym_name_to_string(SymName, String):
+	%	convert a symbol name to a string,
+	%	with module qualifiers separated by
+	%	the standard Mercury module qualifier operator
+	%	(currently ":", but may eventually change to ".")
+:- pred prog_out__sym_name_to_string(sym_name, string).
+:- mode prog_out__sym_name_to_string(in, out) is det.
+
+	% sym_name_to_string(SymName, Separator, String):
+	%	convert a symbol name to a string,
+	%	with module qualifiers separated by Separator.
+:- pred prog_out__sym_name_to_string(sym_name, string, string).
+:- mode prog_out__sym_name_to_string(in, in, out) is det.
+
 :- pred prog_out__write_module_spec(module_specifier, io__state, io__state).
 :- mode prog_out__write_module_spec(in, di, uo) is det.
 
 :- pred prog_out__write_module_list(list(module_name), io__state, io__state).
 :- mode prog_out__write_module_list(in, di, uo) is det.
 
+:- pred prog_out__write_list(list(T), pred(T, io__state, io__state),
+		io__state, io__state).
+:- mode prog_out__write_list(in, pred(in, di, uo) is det, di, uo) is det.
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
 :- implementation.
-:- import_module require, string, list, varset, std_util, term_io.
+:- import_module term, varset, term_io.
+:- import_module require, string, std_util, term, term_io, varset, int.
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
@@ -57,8 +101,10 @@ prog_out__write_messages([Message | Messages]) -->
 
 prog_out__write_message(Msg - Term) -->
 	(
-		{ Term = term__functor(_Functor, _Args, Context) }
+		{ Term = term__functor(_Functor, _Args, Context0) }
 	->
+		{ Context0 = term__context(File, Line) },
+		{ Context = term__context(File, Line) },
 		prog_out__write_context(Context)
 	;
 		[]
@@ -70,7 +116,8 @@ prog_out__write_message(Msg - Term) -->
 		io__write_string(".\n")
 	;
 		io__write_string(": "),
-		{ varset__init(VarSet) }, % XXX variable names in error messages
+		{ varset__init(VarSet) },
+			% XXX variable names in error messages
 		term_io__write_term_nl(VarSet, Term)
 	).
 
@@ -81,16 +128,71 @@ prog_out__write_message(Msg - Term) -->
 	% error message.
 
 prog_out__write_context(Context) -->
-	{ term__context_file(Context, FileName) },
-	{ term__context_line(Context, LineNumber) },
-	( { FileName = "" } ->
-		[]
+	{ prog_out__context_to_string(Context, ContextMessage) },
+	io__write_string(ContextMessage).
+
+%-----------------------------------------------------------------------------%
+
+	% Write to a string the information in term context (at the moment,
+	% just the line number) in a form suitable for the beginning of an
+	% error message.
+
+prog_out__context_to_string(Context, ContextMessage) :-
+	term__context_file(Context, FileName),
+	term__context_line(Context, LineNumber),
+	( FileName = "" ->
+		ContextMessage = ""
 	;
-		{ string__format("%s:%03d: ", [s(FileName), i(LineNumber)],
-			ContextMessage) }, 
-		io__write_string(ContextMessage)
+		string__format("%s:%03d: ", [s(FileName), i(LineNumber)],
+			ContextMessage)
 	).
 
+%-----------------------------------------------------------------------------%
+
+prog_out__write_strings_with_context(Context, Strings) -->
+	{ prog_out__context_to_string(Context, ContextMessage) },
+	{ string__length(ContextMessage, ContextLength) },
+	prog_out__write_strings_with_context_2(ContextMessage,
+			ContextLength, Strings, 0).
+
+:- pred prog_out__write_strings_with_context_2(string, int, list(string), int,
+	io__state, io__state).
+:- mode prog_out__write_strings_with_context_2(in, in, in, in, di, uo) is det.
+
+prog_out__write_strings_with_context_2(_ContextMessage, _ContextLength,
+		[], _) --> [].
+prog_out__write_strings_with_context_2(ContextMessage, ContextLength,
+		[S|Ss], N0) -->
+	{ string__length(S, MessageLength) },
+	(
+		{ N0 = 0 }
+	->
+		io__write_string(ContextMessage),
+		io__write_string("  "),
+		io__write_string(S),
+		{ N is ContextLength + MessageLength },
+		{ Rest = Ss }
+	;
+		{ N1 is MessageLength + N0 },
+		{ num_columns(NumColumns) },
+		{ N1 < NumColumns }
+	->
+		io__write_string(S),
+		{ N = N1 },
+		{ Rest = Ss }
+	;
+		io__write_char('\n'),
+		{ N = 0 },
+		{ Rest = [S|Ss] }
+	),
+	prog_out__write_strings_with_context_2(ContextMessage,
+			ContextLength, Rest, N).
+
+
+:- pred num_columns(int::out) is det.
+
+num_columns(80).
+
 %-----------------------------------------------------------------------------%
 
 	% write out a (possibly qualified) symbol name
@@ -98,25 +200,65 @@ prog_out__write_context(Context) -->
 prog_out__write_sym_name(qualified(ModuleSpec,Name)) -->
 	prog_out__write_module_spec(ModuleSpec),
 	io__write_string(":"),
-	io__write_string(Name).
+	term_io__write_escaped_string(Name).
 prog_out__write_sym_name(unqualified(Name)) -->
-	io__write_string(Name).
+	term_io__write_escaped_string(Name).
+
+prog_out__write_sym_name_and_arity(Name / Arity) -->
+	prog_out__write_sym_name(Name),
+	io__write_string("/"),
+	io__write_int(Arity).
+
+prog_out__write_quoted_sym_name(SymName) -->
+	io__write_string("'"),
+	prog_out__write_sym_name(SymName),
+	io__write_string("'").
+
+prog_out__sym_name_to_string(SymName, String) :-
+	prog_out__sym_name_to_string(SymName, ":", String).
+
+prog_out__sym_name_to_string(SymName, Separator, String) :-
+	prog_out__sym_name_to_string_2(SymName, Separator, Parts, []),
+	string__append_list(Parts, String).
+	
+:- pred prog_out__sym_name_to_string_2(sym_name, string,
+				list(string), list(string)).
+:- mode prog_out__sym_name_to_string_2(in, in, out, in) is det.
+
+prog_out__sym_name_to_string_2(qualified(ModuleSpec,Name), Separator) -->
+	prog_out__sym_name_to_string_2(ModuleSpec, Separator),
+	[Separator, Name].
+prog_out__sym_name_to_string_2(unqualified(Name), _) -->
+	[Name].
 
 	% write out a module specifier
 
 prog_out__write_module_spec(ModuleSpec) -->
-	io__write_string(ModuleSpec).
+	prog_out__write_sym_name(ModuleSpec).
 
 %-----------------------------------------------------------------------------%
 
-prog_out__write_module_list([Import1, Import2, Import3 | Imports]) --> 
-	io__write_strings(["`", Import1, "', "]),
-	write_module_list([Import2, Import3 | Imports]).
-prog_out__write_module_list([Import1, Import2]) -->
-	io__write_strings(["`", Import1, "' and `", Import2 ,"'"]).
-prog_out__write_module_list([Import]) -->
-	io__write_strings(["`", Import, "'"]).
-prog_out__write_module_list([]) -->
+prog_out__write_module_list(Modules) -->
+	prog_out__write_list(Modules, write_module).
+
+:- pred write_module(module_name::in, io__state::di, io__state::uo) is det.
+
+write_module(Module) -->
+	io__write_string("`"),
+	prog_out__write_sym_name(Module),
+	io__write_string("'").
+
+prog_out__write_list([Import1, Import2, Import3 | Imports], Writer) --> 
+	call(Writer, Import1),
+	io__write_string(", "),
+	prog_out__write_list([Import2, Import3 | Imports], Writer).
+prog_out__write_list([Import1, Import2], Writer) -->
+	call(Writer, Import1),
+	io__write_string(" and "),
+	call(Writer, Import2).
+prog_out__write_list([Import], Writer) -->
+	call(Writer, Import).
+prog_out__write_list([], _) -->
 	{ error("prog_out__write_module_list") }.
 
 %-----------------------------------------------------------------------------%
