@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -12,6 +12,10 @@
 	% For the rules on implicit quantification, see the
 	% Mercury language reference manual.
 	%
+	% This pass also expands out bi-implications (that has to be
+	% done after quantification, and preferably as soon as possible,
+	% so we do it here).
+	%
 	% Rather than making implicit quantification explicit by
 	% inserting additional existential quantifiers in the form of
 	% `some/2' goals, we instead record existential quantification
@@ -35,16 +39,18 @@
 :- interface.
 
 :- import_module hlds_goal, hlds_pred, prog_data.
-:- import_module list, set, term, varset.
+:- import_module map, list, set.
 
-:- pred implicitly_quantify_clause_body(list(var),
-		hlds_goal, varset, map(var, type),
-		hlds_goal, varset, map(var, type), list(quant_warning)).
+:- pred implicitly_quantify_clause_body(list(prog_var),
+		hlds_goal, prog_varset, map(prog_var, type),
+		hlds_goal, prog_varset, map(prog_var, type),
+		list(quant_warning)).
 :- mode implicitly_quantify_clause_body(in, in, in, in, out, out, out, out)
 	is det.
 
-:- pred implicitly_quantify_goal(hlds_goal, varset, map(var, type), set(var),
-		hlds_goal, varset, map(var, type), list(quant_warning)).
+:- pred implicitly_quantify_goal(hlds_goal, prog_varset, map(prog_var, type),
+		set(prog_var), hlds_goal, prog_varset, map(prog_var, type),
+		list(quant_warning)).
 :- mode implicitly_quantify_goal(in, in, in, in, out, out, out, out) is det.
 
 :- pred requantify_proc(proc_info, proc_info) is det.
@@ -55,18 +61,22 @@
 	% overlapping scopes.
 
 :- type quant_warning
-	--->	warn_overlap(list(var), term__context).
+	--->	warn_overlap(list(prog_var), prog_context).
 
 	% quantification__goal_vars(Goal, Vars):
 	%	Vars is the set of variables that are free (unquantified)
 	%	in Goal.
-:- pred quantification__goal_vars(hlds_goal, set(var)).
+:- pred quantification__goal_vars(hlds_goal, set(prog_var)).
 :- mode quantification__goal_vars(in, out) is det.
 
 %-----------------------------------------------------------------------------%
 
 :- implementation.
-:- import_module std_util, bool, map, goal_util, require.
+
+:- import_module instmap, goal_util.
+
+:- import_module term, varset.
+:- import_module std_util, bool, require.
 
 	% The `outside vars', `lambda outside vars', and `quant vars'
 	% fields are inputs; the `nonlocals' field is output; and
@@ -76,13 +86,13 @@
 	% and the outputs are caller save.
 :- type quant_info
 	--->	quant_info(
-			set(var), % outside vars
-			set(var), % quant vars
-			set(var), % outside lambda vars
-			set(var), % nonlocals
-			set(var), % seen so far
-			varset,
-			map(var, type),
+			set(prog_var), % outside vars
+			set(prog_var), % quant vars
+			set(prog_var), % outside lambda vars
+			set(prog_var), % nonlocals
+			set(prog_var), % seen so far
+			prog_varset,
+			map(prog_var, type),
 			list(quant_warning)
 		).
 
@@ -122,13 +132,13 @@ implicitly_quantify_clause_body(HeadVars, Goal0, Varset0, VarTypes0,
 			OutsideVars, Goal, Varset, VarTypes, Warnings).
 
 requantify_proc(ProcInfo0, ProcInfo) :-
-	proc_info_variables(ProcInfo0, Varset0),
+	proc_info_varset(ProcInfo0, Varset0),
 	proc_info_vartypes(ProcInfo0, VarTypes0),
 	proc_info_headvars(ProcInfo0, HeadVars),
 	proc_info_goal(ProcInfo0, Goal0),
 	implicitly_quantify_clause_body(HeadVars, Goal0, Varset0, VarTypes0,
 		Goal, Varset, VarTypes, _),
-	proc_info_set_variables(ProcInfo0, Varset, ProcInfo1),
+	proc_info_set_varset(ProcInfo0, Varset, ProcInfo1),
 	proc_info_set_vartypes(ProcInfo1, VarTypes, ProcInfo2),
 	proc_info_set_goal(ProcInfo2, Goal, ProcInfo).
 
@@ -165,18 +175,33 @@ implicitly_quantify_goal(Goal0 - GoalInfo0, Goal - GoalInfo) -->
 		{ Goal = Goal1 },
 		{ GoalInfo1 = GoalInfo0 }
 	),
-	{ goal_info_set_nonlocals(GoalInfo1, NonLocalVars, GoalInfo) }.
+	{ goal_info_set_nonlocals(GoalInfo1, NonLocalVars, GoalInfo2) },
+	%
+	% If the non-locals set has shrunk (e.g. because some optimization
+	% optimizes away the other occurrences of a variable, causing it
+	% to become local when previously it was non-local),
+	% then we may need to likewise shrink the instmap delta.
+	%
+	{ goal_info_get_instmap_delta(GoalInfo2, InstMapDelta0) },
+	{ instmap_delta_restrict(InstMapDelta0, NonLocalVars, InstMapDelta) },
+	{ goal_info_set_instmap_delta(GoalInfo2, InstMapDelta, GoalInfo) }.
 
-:- pred implicitly_quantify_goal_2(hlds_goal_expr, term__context,
+:- pred implicitly_quantify_goal_2(hlds_goal_expr, prog_context,
 				hlds_goal_expr, quant_info, quant_info).
 :- mode implicitly_quantify_goal_2(in, in, out, in, out) is det.
 
-	% we retain explicit existential quantifiers in the source code,
-	% even though they are redundant with the goal_info non_locals,
-	% so that we can easily recalculate the goal_info non_locals
-	% if necessary after program transformation.
+	% After this pass, explicit quantifiers are redundant,
+	% since all variables which were explicitly quantified
+	% have been renamed apart.  So we don't keep them.
+	% We need to keep the structure, though, so that mode
+	% analysis doesn't try to reorder through quantifiers.
+	% (Actually it would make sense to allow mode analysis
+	% to do that, but the reference manual says it doesn't,
+	% so we don't.)  Thus we replace `some(Vars, Goal0)' with
+	% an empty quantifier `some([], Goal)'.
 
-implicitly_quantify_goal_2(some(Vars0, Goal0), Context, some(Vars, Goal)) -->
+implicitly_quantify_goal_2(some(Vars0, CanRemove, Goal0), Context,
+		some([], CanRemove, Goal)) -->
 	quantification__get_outside(OutsideVars),
 	quantification__get_lambda_outside(LambdaOutsideVars),
 	quantification__get_quant_vars(QuantVars),
@@ -209,6 +234,9 @@ implicitly_quantify_goal_2(some(Vars0, Goal0), Context, some(Vars, Goal)) -->
 implicitly_quantify_goal_2(conj(List0), _, conj(List)) -->
 	implicitly_quantify_conj(List0, List).
 
+implicitly_quantify_goal_2(par_conj(List0, SM), _, par_conj(List, SM)) -->
+	implicitly_quantify_conj(List0, List).
+
 implicitly_quantify_goal_2(disj(Goals0, SM), _, disj(Goals, SM)) -->
 	implicitly_quantify_disj(Goals0, Goals).
 
@@ -293,9 +321,11 @@ implicitly_quantify_goal_2(call(A, B, HeadVars, D, E, F), _,
 		call(A, B, HeadVars, D, E, F)) -->
 	implicitly_quantify_atomic_goal(HeadVars).
 
-implicitly_quantify_goal_2(higher_order_call(PredVar, ArgVars, C, D, E, F), _,
-		higher_order_call(PredVar, ArgVars, C, D, E, F)) -->
-	implicitly_quantify_atomic_goal([PredVar|ArgVars]).
+implicitly_quantify_goal_2(generic_call(GenericCall, ArgVars1, C, D), _,
+		generic_call(GenericCall, ArgVars1, C, D)) -->
+	{ goal_util__generic_call_vars(GenericCall, ArgVars0) },
+	{ list__append(ArgVars0, ArgVars1, ArgVars) },
+	implicitly_quantify_atomic_goal(ArgVars).
 
 implicitly_quantify_goal_2(
 		unify(Var, UnifyRHS0, Mode, Unification0, UnifyContext),
@@ -303,21 +333,107 @@ implicitly_quantify_goal_2(
 		unify(Var, UnifyRHS, Mode, Unification, UnifyContext)) -->
 	quantification__get_outside(OutsideVars),
 	quantification__get_lambda_outside(LambdaOutsideVars),
+	{ quantification__get_unify_typeinfos(Unification0, TypeInfoVars) },
 	implicitly_quantify_unify_rhs(UnifyRHS0, Unification0, Context,
 		UnifyRHS, Unification),
 	quantification__get_nonlocals(VarsUnifyRHS),
-	{ set__insert(VarsUnifyRHS, Var, GoalVars) },
+	{ set__insert(VarsUnifyRHS, Var, GoalVars0) },
+	{ set__insert_list(GoalVars0, TypeInfoVars, GoalVars1) },
+	{
+		Unification = construct(_, _, _, _, CellToReuse, _, _),
+		CellToReuse = yes(cell_to_reuse(ReuseVar, _, _))
+	->
+		set__insert(GoalVars1, ReuseVar, GoalVars)
+	;
+		GoalVars = GoalVars1
+	},
 	quantification__update_seen_vars(GoalVars),
 	{ set__intersect(GoalVars, OutsideVars, NonLocalVars1) },
 	{ set__intersect(GoalVars, LambdaOutsideVars, NonLocalVars2) },
 	{ set__union(NonLocalVars1, NonLocalVars2, NonLocalVars) },
 	quantification__set_nonlocals(NonLocalVars).
 
-implicitly_quantify_goal_2(pragma_c_code(A,B,C,D,Vars,F,G,H), _,
-		pragma_c_code(A,B,C,D,Vars,F,G,H)) --> 
+implicitly_quantify_goal_2(pragma_c_code(A,B,C,Vars,E,F,G), _,
+		pragma_c_code(A,B,C,Vars,E,F,G)) --> 
 	implicitly_quantify_atomic_goal(Vars).
 
-:- pred implicitly_quantify_atomic_goal(list(var), quant_info, quant_info).
+implicitly_quantify_goal_2(bi_implication(LHS0, RHS0), Context, Goal) -->
+		% get the initial values of various settings
+	quantification__get_quant_vars(QuantVars0),
+	quantification__get_outside(OutsideVars0),
+	quantification__get_lambda_outside(LambdaOutsideVars0),
+
+		% quantified variables cannot be pushed inside a negation,
+		% so we insert the quantified vars into the outside vars set,
+		% and initialize the new quantified vars set to be empty
+		% (the lambda outside vars remain unchanged)
+	{ set__union(OutsideVars0, QuantVars0, OutsideVars1) },
+	{ set__init(QuantVars1) },
+	{ LambdaOutsideVars1 = LambdaOutsideVars0 },
+	quantification__set_quant_vars(QuantVars1),
+
+		% prepare for quantifying the LHS:
+		% add variables from the RHS to the outside vars
+		% and the outside lambda vars sets.
+	{ quantification__goal_vars(RHS0, RHS_Vars, RHS_LambdaVars) },
+	{ set__union(OutsideVars1, RHS_Vars, LHS_OutsideVars) },
+	{ set__union(LambdaOutsideVars1, RHS_LambdaVars,
+			LHS_LambdaOutsideVars) },
+
+		% quantify the LHS
+	quantification__set_outside(LHS_OutsideVars),
+	quantification__set_lambda_outside(LHS_LambdaOutsideVars),
+	implicitly_quantify_goal(LHS0, LHS),
+	quantification__get_nonlocals(LHS_NonLocalVars),
+
+		% prepare for quantifying the RHS:
+		% add nonlocals from the LHS to the outside vars.
+		% (We use the nonlocals rather than the more symmetric
+		% approach of calling quantification__goal_vars on the
+		% LHS goal because it is more efficient.)
+	{ set__union(OutsideVars1, LHS_NonLocalVars, RHS_OutsideVars) },
+	{ RHS_LambdaOutsideVars = LambdaOutsideVars1 },
+
+		% quantify the RHS
+	quantification__set_outside(RHS_OutsideVars),
+	quantification__set_lambda_outside(RHS_LambdaOutsideVars),
+	implicitly_quantify_goal(RHS0, RHS),
+	quantification__get_nonlocals(RHS_NonLocalVars),
+
+		% compute the nonlocals for this goal
+	{ set__union(LHS_NonLocalVars, RHS_NonLocalVars, AllNonLocalVars) },
+	{ set__intersect(AllNonLocalVars, OutsideVars0, NonLocalVarsO) },
+	{ set__intersect(AllNonLocalVars, LambdaOutsideVars0, NonLocalVarsL) },
+	{ set__union(NonLocalVarsO, NonLocalVarsL, NonLocalVars) },
+	quantification__set_nonlocals(NonLocalVars),
+
+		% restore the original values of various settings
+	quantification__set_outside(OutsideVars0),
+	quantification__set_lambda_outside(LambdaOutsideVars0),
+	quantification__set_quant_vars(QuantVars0),
+
+		%
+		% We've figured out the quantification.
+		% Now expand the bi-implication according to the usual
+		% rules:
+		%	LHS <=> RHS
+		% ===>
+		%	(LHS => RHS), (RHS => LHS)
+		% ===>
+		%	(not (LHS, not RHS)), (not (RHS, not LHS))
+		%
+	{ goal_info_init(GoalInfo0) },
+	{ goal_info_set_context(GoalInfo0, Context, GoalInfo1) },
+	{ goal_info_set_nonlocals(GoalInfo1, LHS_NonLocalVars, LHS_GI) },
+	{ goal_info_set_nonlocals(GoalInfo1, RHS_NonLocalVars, RHS_GI) },
+	{ goal_info_set_nonlocals(GoalInfo1, NonLocalVars, GI) },
+	{ NotLHS = not(LHS) - LHS_GI },
+	{ NotRHS = not(RHS) - RHS_GI },
+	{ ForwardsImplication = not(conj([LHS, NotRHS]) - GI) - GI },
+	{ ReverseImplication = not(conj([RHS, NotLHS]) - GI) - GI },
+	{ Goal = conj([ForwardsImplication, ReverseImplication]) }.
+
+:- pred implicitly_quantify_atomic_goal(list(prog_var), quant_info, quant_info).
 :- mode implicitly_quantify_atomic_goal(in, in, out) is det.
 
 implicitly_quantify_atomic_goal(HeadVars) -->
@@ -330,7 +446,7 @@ implicitly_quantify_atomic_goal(HeadVars) -->
 	{ set__union(NonLocals1, NonLocals2, NonLocals) },
 	quantification__set_nonlocals(NonLocals).
 
-:- pred implicitly_quantify_unify_rhs(unify_rhs, unification, term__context,
+:- pred implicitly_quantify_unify_rhs(unify_rhs, unification, prog_context,
 					unify_rhs, unification,
 					quant_info, quant_info).
 :- mode implicitly_quantify_unify_rhs(in, in, in, out, out, in, out) is det.
@@ -343,13 +459,20 @@ implicitly_quantify_unify_rhs(functor(Functor, ArgVars), Unification, _,
 	{ set__list_to_set(ArgVars, Vars) },
 	quantification__set_nonlocals(Vars).
 implicitly_quantify_unify_rhs(
-		lambda_goal(PredOrFunc, LambdaVars0, Modes, Det, Goal0),
+		lambda_goal(PredOrFunc, EvalMethod, FixModes, LambdaNonLocals0,
+			LambdaVars0, Modes, Det, Goal0),
 		Unification0,
 		Context,
-		lambda_goal(PredOrFunc, LambdaVars, Modes, Det, Goal),
+		lambda_goal(PredOrFunc, EvalMethod, FixModes, LambdaNonLocals,
+			LambdaVars, Modes, Det, Goal),
 		Unification
 		) -->
-
+	%
+	% Note: make_hlds.m has already done most of the hard work
+	% for lambda expressions.  At this point, LambdaVars0
+	% should in fact be guaranteed to be fresh distinct
+	% variables.  However, the code below does not assume this.
+	%
 	quantification__get_outside(OutsideVars0),
 	{ set__list_to_set(LambdaVars0, QVars) },
 		% Figure out which variables have overlapping scopes
@@ -391,6 +514,7 @@ implicitly_quantify_unify_rhs(
 	{ set__init(LambdaOutsideVars) },
 	quantification__set_lambda_outside(LambdaOutsideVars),
 	implicitly_quantify_goal(Goal1, Goal),
+
 	quantification__get_nonlocals(NonLocals0),
 		% lambda-quantified variables are local
 	{ set__delete_list(NonLocals0, LambdaVars, NonLocals) },
@@ -399,6 +523,19 @@ implicitly_quantify_unify_rhs(
 	quantification__set_lambda_outside(LambdaOutsideVars0),
 	quantification__set_nonlocals(NonLocals),
 
+	%
+	% Work out the list of non-local curried arguments to the lambda
+	% expression. This set must only ever decrease, since the first
+	% approximation that make_hlds uses includes all variables in the 
+	% lambda expression except the quantified variables.
+	%
+	{ Goal = _ - LambdaGoalInfo },
+	{ goal_info_get_nonlocals(LambdaGoalInfo, LambdaGoalNonLocals) },
+	{ IsNonLocal = lambda([V::in] is semidet, (
+			set__member(V, LambdaGoalNonLocals)
+		)) },
+	{ list__filter(IsNonLocal, LambdaNonLocals0, LambdaNonLocals) },
+
 	%
 	% For a unification that constructs a lambda expression,
 	% the argument variables of the construction are the non-local
@@ -409,12 +546,14 @@ implicitly_quantify_unify_rhs(
 	% so we can just use the old modes.
 	%
 	{
-		Unification0 = construct(ConstructVar, ConsId, Args0, ArgModes0)
+		Unification0 = construct(ConstructVar, ConsId, Args0,
+			ArgModes0, Reuse, Uniq, AditiInfo)
 	->
 		map__from_corresponding_lists(Args0, ArgModes0, ArgModesMap),
 		set__to_sorted_list(NonLocals, Args),
 		map__apply_to_list(Args, ArgModesMap, ArgModes),
-		Unification = construct(ConstructVar, ConsId, Args, ArgModes)
+		Unification = construct(ConstructVar, ConsId, Args,
+			ArgModes, Reuse, Uniq, AditiInfo)
 	;
 		% after mode analysis, unifications with lambda variables
 		% should always be construction unifications, but
@@ -431,7 +570,7 @@ implicitly_quantify_conj(Goals0, Goals) -->
 	{ get_vars(Goals0, FollowingVarsList) },
 	implicitly_quantify_conj_2(Goals0, FollowingVarsList, Goals).
 
-:- pred implicitly_quantify_conj_2(list(hlds_goal), list(pair(set(var))),
+:- pred implicitly_quantify_conj_2(list(hlds_goal), list(pair(set(prog_var))),
 			list(hlds_goal), quant_info, quant_info).
 :- mode implicitly_quantify_conj_2(in, in, out, in, out) is det.
 
@@ -500,7 +639,7 @@ implicitly_quantify_cases([case(Cons, Goal0) | Cases0],
 
 	% insert the given set of variables into the set of `seen' variables.
 
-:- pred quantification__update_seen_vars(set(var), quant_info, quant_info).
+:- pred quantification__update_seen_vars(set(prog_var), quant_info, quant_info).
 :- mode quantification__update_seen_vars(in, in, out) is det.
 
 quantification__update_seen_vars(NewVars) -->
@@ -519,14 +658,15 @@ quantification__update_seen_vars(NewVars) -->
 	% and the second contains following variables that
 	% occur in lambda goals.
 
-:- pred get_vars(list(hlds_goal), list(pair(set(var)))).
+:- pred get_vars(list(hlds_goal), list(pair(set(prog_var)))).
 :- mode get_vars(in, out) is det.
 
 get_vars([], []).
 get_vars([_Goal | Goals], [Set - LambdaSet | SetPairs]) :-
 	get_vars_2(Goals, Set, LambdaSet, SetPairs).
 
-:- pred get_vars_2(list(hlds_goal), set(var), set(var), list(pair(set(var)))).
+:- pred get_vars_2(list(hlds_goal), set(prog_var), set(prog_var),
+		list(pair(set(prog_var)))).
 :- mode get_vars_2(in, out, out, out) is det.
 
 get_vars_2([], Set, LambdaSet, []) :-
@@ -539,8 +679,8 @@ get_vars_2([Goal | Goals], Set, LambdaSet, SetPairList) :-
 	set__union(LambdaSet0, LambdaSet1, LambdaSet),
 	SetPairList = [Set0 - LambdaSet0 | SetPairList0].
 
-:- pred goal_list_vars_2(list(hlds_goal), set(var), set(var),
-			set(var), set(var)).
+:- pred goal_list_vars_2(list(hlds_goal), set(prog_var), set(prog_var),
+			set(prog_var), set(prog_var)).
 :- mode goal_list_vars_2(in, in, in, out, out) is det.
 
 goal_list_vars_2([], Set, LambdaSet, Set, LambdaSet).
@@ -548,7 +688,8 @@ goal_list_vars_2([Goal - _GoalInfo| Goals], Set0, LambdaSet0, Set, LambdaSet) :-
 	quantification__goal_vars_2(Goal, Set0, LambdaSet0, Set1, LambdaSet1),
 	goal_list_vars_2(Goals, Set1, LambdaSet1, Set, LambdaSet).
 
-:- pred case_list_vars_2(list(case), set(var), set(var), set(var), set(var)).
+:- pred case_list_vars_2(list(case), set(prog_var), set(prog_var),
+		set(prog_var), set(prog_var)).
 :- mode case_list_vars_2(in, in, in, out, out) is det.
 
 case_list_vars_2([], Set, LambdaSet, Set, LambdaSet).
@@ -569,7 +710,7 @@ quantification__goal_vars(Goal, BothSet) :-
 	%	in Goal, not counting occurrences in lambda expressions.
 	%	LambdaSet is the set of variables that occur free (unquantified)
 	%	in lambda expressions in Goal.
-:- pred quantification__goal_vars(hlds_goal, set(var), set(var)).
+:- pred quantification__goal_vars(hlds_goal, set(prog_var), set(prog_var)).
 :- mode quantification__goal_vars(in, out, out) is det.
 
 quantification__goal_vars(Goal - _GoalInfo, Set, LambdaSet) :-
@@ -577,18 +718,32 @@ quantification__goal_vars(Goal - _GoalInfo, Set, LambdaSet) :-
 	set__init(LambdaSet0),
 	quantification__goal_vars_2(Goal, Set0, LambdaSet0, Set, LambdaSet).
 
-:- pred quantification__goal_vars_2(hlds_goal_expr, set(var), set(var),
-		set(var), set(var)).
+:- pred quantification__goal_vars_2(hlds_goal_expr, set(prog_var),
+		set(prog_var), set(prog_var), set(prog_var)).
 :- mode quantification__goal_vars_2(in, in, in, out, out) is det.
 
-quantification__goal_vars_2(unify(A, B, _, _, _), Set0, LambdaSet0,
+quantification__goal_vars_2(unify(A, B, _, Unification, _), Set0, LambdaSet0,
 		Set, LambdaSet) :-
 	set__insert(Set0, A, Set1),
-	quantification__unify_rhs_vars(B, Set1, LambdaSet0, Set, LambdaSet).
+	(
+		Unification = construct(_, _, _, _, Reuse, _, _),
+		Reuse = yes(cell_to_reuse(ReuseVar, _, _))
+	->
+		set__insert(Set1, ReuseVar, Set2)
+	;
+		Unification = complicated_unify(_, _, TypeInfoVars)
+	->
+		set__insert_list(Set1, TypeInfoVars, Set2)
+	;
+		Set2 = Set1
+	),
+	quantification__unify_rhs_vars(B, Set2, LambdaSet0, Set, LambdaSet).
 
-quantification__goal_vars_2(higher_order_call(PredVar, ArgVars, _, _, _, _),
+quantification__goal_vars_2(generic_call(GenericCall, ArgVars1, _, _),
 		Set0, LambdaSet, Set, LambdaSet) :-
-	set__insert_list(Set0, [PredVar | ArgVars], Set).
+	goal_util__generic_call_vars(GenericCall, ArgVars0),
+	set__insert_list(Set0, ArgVars0, Set1),
+	set__insert_list(Set1, ArgVars1, Set).
 
 quantification__goal_vars_2(call(_, _, ArgVars, _, _, _), Set0, LambdaSet,
 		Set, LambdaSet) :-
@@ -597,6 +752,9 @@ quantification__goal_vars_2(call(_, _, ArgVars, _, _, _), Set0, LambdaSet,
 quantification__goal_vars_2(conj(Goals), Set0, LambdaSet0, Set, LambdaSet) :-
 	goal_list_vars_2(Goals, Set0, LambdaSet0, Set, LambdaSet).
 
+quantification__goal_vars_2(par_conj(Goals, _SM), Set0, LambdaSet0, Set, LambdaSet) :-
+	goal_list_vars_2(Goals, Set0, LambdaSet0, Set, LambdaSet).
+
 quantification__goal_vars_2(disj(Goals, _), Set0, LambdaSet0, Set, LambdaSet) :-
 	goal_list_vars_2(Goals, Set0, LambdaSet0, Set, LambdaSet).
 
@@ -605,7 +763,7 @@ quantification__goal_vars_2(switch(Var, _Det, Cases, _), Set0, LambdaSet0,
 	set__insert(Set0, Var, Set1),
 	case_list_vars_2(Cases, Set1, LambdaSet0, Set, LambdaSet).
 
-quantification__goal_vars_2(some(Vars, Goal), Set0, LambdaSet0,
+quantification__goal_vars_2(some(Vars, _, Goal), Set0, LambdaSet0,
 		Set, LambdaSet) :-
 	quantification__goal_vars(Goal, Set1, LambdaSet1),
 	set__delete_list(Set1, Vars, Set2),
@@ -634,12 +792,16 @@ quantification__goal_vars_2(if_then_else(Vars, A, B, C, _), Set0, LambdaSet0,
 	set__union(Set5, Set6, Set),
 	set__union(LambdaSet5, LambdaSet6, LambdaSet).
 
-quantification__goal_vars_2(pragma_c_code(_, _, _, _, ArgVars, _, _, _),
+quantification__goal_vars_2(pragma_c_code(_, _, _, ArgVars, _, _, _),
 		Set0, LambdaSet, Set, LambdaSet) :-
 	set__insert_list(Set0, ArgVars, Set).
 
-:- pred quantification__unify_rhs_vars(unify_rhs, set(var), set(var),
-					set(var), set(var)).
+quantification__goal_vars_2(bi_implication(LHS, RHS), Set0, LambdaSet0, Set,
+		LambdaSet) :-
+	goal_list_vars_2([LHS, RHS], Set0, LambdaSet0, Set, LambdaSet).
+
+:- pred quantification__unify_rhs_vars(unify_rhs, set(prog_var), set(prog_var),
+					set(prog_var), set(prog_var)).
 :- mode quantification__unify_rhs_vars(in, in, in, out, out) is det.
 
 quantification__unify_rhs_vars(var(X), Set0, LambdaSet, Set, LambdaSet) :-
@@ -647,15 +809,28 @@ quantification__unify_rhs_vars(var(X), Set0, LambdaSet, Set, LambdaSet) :-
 quantification__unify_rhs_vars(functor(_Functor, ArgVars), Set0, LambdaSet,
 		Set, LambdaSet) :-
 	set__insert_list(Set0, ArgVars, Set).
-quantification__unify_rhs_vars(lambda_goal(_PredOrFunc, LambdaVars, _Modes,
-		_Detism, Goal), Set, LambdaSet0, Set, LambdaSet) :-
+quantification__unify_rhs_vars(
+		lambda_goal(_POrF, _E, _F, _N, LambdaVars, _M, _D, Goal), 
+		Set, LambdaSet0, Set, LambdaSet) :-
+	% Note that the NonLocals list is not counted, since all the 
+	% variables in that list must occur in the goal.
 	quantification__goal_vars(Goal, GoalVars),
 	set__delete_list(GoalVars, LambdaVars, GoalVars1),
 	set__union(LambdaSet0, GoalVars1, LambdaSet).
 
+:- pred quantification__get_unify_typeinfos(unification, list(prog_var)).
+:- mode quantification__get_unify_typeinfos(in, out) is det.
+
+quantification__get_unify_typeinfos(Unification, TypeInfoVars) :-
+	( Unification = complicated_unify(_, _, TypeInfoVars0) ->
+		TypeInfoVars = TypeInfoVars0
+	;
+		TypeInfoVars = []
+	).
+
 %-----------------------------------------------------------------------------%
 
-:- pred quantification__warn_overlapping_scope(set(var), term__context,
+:- pred quantification__warn_overlapping_scope(set(prog_var), prog_context,
 					quant_info, quant_info).
 :- mode quantification__warn_overlapping_scope(in, in, in, out) is det.
 
@@ -672,36 +847,42 @@ quantification__warn_overlapping_scope(OverlapVars, Context) -->
 %	and insert the mapping V->V' into RenameMap.
 %	Apply RenameMap to Goal0 giving Goal.
 
-:- pred quantification__rename_apart(set(var), map(var, var),
+:- pred quantification__rename_apart(set(prog_var), map(prog_var, prog_var),
 				hlds_goal, hlds_goal, quant_info, quant_info).
 :- mode quantification__rename_apart(in, out, in, out, in, out) is det.
 
 quantification__rename_apart(RenameSet, RenameMap, Goal0, Goal) -->
-	{ set__to_sorted_list(RenameSet, RenameList) },
-	quantification__get_varset(Varset0),
-	quantification__get_vartypes(VarTypes0),
-	{ map__init(RenameMap0) },
-	{ goal_util__create_variables(RenameList,
-		Varset0, VarTypes0, RenameMap0, VarTypes0, Varset0,
-			% ^ Accumulator		^ Reference ^Var names
-		Varset, VarTypes, RenameMap) },
-	{ goal_util__rename_vars_in_goal(Goal0, RenameMap, Goal) },
-	quantification__set_varset(Varset),
-	quantification__set_vartypes(VarTypes).
+	( { set__empty(RenameSet) } ->
+		{ map__init(RenameMap) },
+		{ Goal = Goal0 }
+	;
+		{ set__to_sorted_list(RenameSet, RenameList) },
+		quantification__get_varset(Varset0),
+		quantification__get_vartypes(VarTypes0),
+		{ map__init(RenameMap0) },
+		{ goal_util__create_variables(RenameList,
+			Varset0, VarTypes0, RenameMap0, VarTypes0, Varset0,
+				% ^ Accumulator		^ Reference ^Var names
+			Varset, VarTypes, RenameMap) },
+		{ goal_util__rename_vars_in_goal(Goal0, RenameMap, Goal) },
+		quantification__set_varset(Varset),
+		quantification__set_vartypes(VarTypes)
 /****
-	We don't need to add the newly created vars to the seen vars
-	because we won't find them anywhere else in the enclosing goal.
-	This is a performance improvement because it keeps the size of
-	the seen var set down.
-	quantification__get_seen(SeenVars0),
-	{ map__values(RenameMap, NewVarsList) },
-	{ set__insert_list(SeenVars0, NewVarsList, SeenVars) },
-	quantification__set_seen(SeenVars).
-***/
+		We don't need to add the newly created vars to the seen vars
+		because we won't find them anywhere else in the enclosing goal.
+		This is a performance improvement because it keeps the size of
+		the seen var set down.
+		quantification__get_seen(SeenVars0),
+		{ map__values(RenameMap, NewVarsList) },
+		{ set__insert_list(SeenVars0, NewVarsList, SeenVars) },
+		quantification__set_seen(SeenVars).
+****/
+	).
 
 %-----------------------------------------------------------------------------%
 
-:- pred quantification__init(set(var), varset, map(var, type), quant_info).
+:- pred quantification__init(set(prog_var), prog_varset, map(prog_var, type),
+		quant_info).
 :- mode quantification__init(in, in, in, out) is det.
 
 quantification__init(OutsideVars, Varset, VarTypes, QuantInfo) :-
@@ -713,91 +894,95 @@ quantification__init(OutsideVars, Varset, VarTypes, QuantInfo) :-
 	QuantInfo = quant_info(OutsideVars, QuantVars, LambdaOutsideVars,
 		NonLocals, Seen, Varset, VarTypes, OverlapWarnings).
 
-:- pred quantification__get_outside(set(var), quant_info, quant_info).
+:- pred quantification__get_outside(set(prog_var), quant_info, quant_info).
 :- mode quantification__get_outside(out, in, out) is det.
 
 quantification__get_outside(A, Q, Q) :-
 	Q = quant_info(A, _, _, _, _, _, _, _).
 
-:- pred quantification__set_outside(set(var), quant_info, quant_info).
+:- pred quantification__set_outside(set(prog_var), quant_info, quant_info).
 :- mode quantification__set_outside(in, in, out) is det.
 
 quantification__set_outside(A, Q0, Q) :-
 	Q0 = quant_info(_, B, C, D, E, F, G, H),
 	Q  = quant_info(A, B, C, D, E, F, G, H).
 
-:- pred quantification__get_quant_vars(set(var), quant_info, quant_info).
+:- pred quantification__get_quant_vars(set(prog_var), quant_info, quant_info).
 :- mode quantification__get_quant_vars(out, in, out) is det.
 
 quantification__get_quant_vars(B, Q, Q) :-
 	Q = quant_info(_, B, _, _, _, _, _, _).
 
-:- pred quantification__set_quant_vars(set(var), quant_info, quant_info).
+:- pred quantification__set_quant_vars(set(prog_var), quant_info, quant_info).
 :- mode quantification__set_quant_vars(in, in, out) is det.
 
 quantification__set_quant_vars(B, Q0, Q) :-
 	Q0 = quant_info(A, _, C, D, E, F, G, H),
 	Q  = quant_info(A, B, C, D, E, F, G, H).
 
-:- pred quantification__get_lambda_outside(set(var), quant_info, quant_info).
+:- pred quantification__get_lambda_outside(set(prog_var),
+		quant_info, quant_info).
 :- mode quantification__get_lambda_outside(out, in, out) is det.
 
 quantification__get_lambda_outside(C, Q, Q) :-
 	Q  = quant_info(_, _, C, _, _, _, _, _).
 
-:- pred quantification__set_lambda_outside(set(var), quant_info, quant_info).
+:- pred quantification__set_lambda_outside(set(prog_var),
+		quant_info, quant_info).
 :- mode quantification__set_lambda_outside(in, in, out) is det.
 
 quantification__set_lambda_outside(C, Q0, Q) :-
 	Q0 = quant_info(A, B, _, D, E, F, G, H),
 	Q  = quant_info(A, B, C, D, E, F, G, H).
 
-:- pred quantification__get_nonlocals(set(var), quant_info, quant_info).
+:- pred quantification__get_nonlocals(set(prog_var), quant_info, quant_info).
 :- mode quantification__get_nonlocals(out, in, out) is det.
 
 quantification__get_nonlocals(D, Q, Q) :-
 	Q  = quant_info(_, _, _, D, _, _, _, _).
 
-:- pred quantification__set_nonlocals(set(var), quant_info, quant_info).
+:- pred quantification__set_nonlocals(set(prog_var), quant_info, quant_info).
 :- mode quantification__set_nonlocals(in, in, out) is det.
 
 quantification__set_nonlocals(D, Q0, Q) :-
 	Q0 = quant_info(A, B, C, _, E, F, G, H),
 	Q  = quant_info(A, B, C, D, E, F, G, H).
 
-:- pred quantification__get_seen(set(var), quant_info, quant_info).
+:- pred quantification__get_seen(set(prog_var), quant_info, quant_info).
 :- mode quantification__get_seen(out, in, out) is det.
 
 quantification__get_seen(E, Q, Q) :-
 	Q  = quant_info(_, _, _, _, E, _, _, _).
 
-:- pred quantification__set_seen(set(var), quant_info, quant_info).
+:- pred quantification__set_seen(set(prog_var), quant_info, quant_info).
 :- mode quantification__set_seen(in, in, out) is det.
 
 quantification__set_seen(E, Q0, Q) :-
 	Q0 = quant_info(A, B, C, D, _, F, G, H),
 	Q  = quant_info(A, B, C, D, E, F, G, H).
 
-:- pred quantification__get_varset(varset, quant_info, quant_info).
+:- pred quantification__get_varset(prog_varset, quant_info, quant_info).
 :- mode quantification__get_varset(out, in, out) is det.
 
 quantification__get_varset(F, Q, Q) :-
 	Q  = quant_info(_, _, _, _, _, F, _, _).
 
-:- pred quantification__set_varset(varset, quant_info, quant_info).
+:- pred quantification__set_varset(prog_varset, quant_info, quant_info).
 :- mode quantification__set_varset(in, in, out) is det.
 
 quantification__set_varset(F, Q0, Q) :-
 	Q0 = quant_info(A, B, C, D, E, _, G, H),
 	Q  = quant_info(A, B, C, D, E, F, G, H).
 
-:- pred quantification__get_vartypes(map(var, type), quant_info, quant_info).
+:- pred quantification__get_vartypes(map(prog_var, type),
+		quant_info, quant_info).
 :- mode quantification__get_vartypes(out, in, out) is det.
 
 quantification__get_vartypes(G, Q, Q) :-
 	Q  = quant_info(_, _, _, _, _, _, G, _).
 
-:- pred quantification__set_vartypes(map(var, type), quant_info, quant_info).
+:- pred quantification__set_vartypes(map(prog_var, type),
+		quant_info, quant_info).
 :- mode quantification__set_vartypes(in, in, out) is det.
 
 quantification__set_vartypes(G, Q0, Q) :-
