@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1997 The University of Melbourne.
+% Copyright (C) 1996-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -32,13 +32,17 @@
 		module_info, module_info, io__state, io__state).
 :- mode saved_vars_proc(in, in, in, out, in, out, di, uo) is det.
 
+:- pred saved_vars_proc_no_io(pred_id, proc_id, proc_info, proc_info,
+		module_info, module_info).
+:- mode saved_vars_proc_no_io(in, in, in, out, in, out) is det.
+
 %-----------------------------------------------------------------------------%
 
 :- implementation.
 
 :- import_module hlds_goal, hlds_out, goal_util, quantification, passes_aux.
-:- import_module mode_util, prog_data.
-:- import_module bool, list, set, map, std_util, term, varset.
+:- import_module mode_util, prog_data, term, varset.
+:- import_module bool, list, set, map, std_util, require.
 
 %-----------------------------------------------------------------------------%
 
@@ -46,30 +50,35 @@ saved_vars_proc(PredId, ProcId, ProcInfo0, ProcInfo,
 		ModuleInfo0, ModuleInfo) -->
 	write_proc_progress_message("% Minimizing saved vars in ",
 		PredId, ProcId, ModuleInfo0),
-	{ proc_info_goal(ProcInfo0, Goal0) },
-	{ proc_info_variables(ProcInfo0, Varset0) },
-	{ proc_info_vartypes(ProcInfo0, VarTypes0) },
-	{ init_slot_info(Varset0, VarTypes0, SlotInfo0) },
+	{ saved_vars_proc_no_io(PredId, ProcId, ProcInfo0, ProcInfo,
+		ModuleInfo0, ModuleInfo) }.
 
-	{ saved_vars_in_goal(Goal0, SlotInfo0, Goal1, SlotInfo) },
+saved_vars_proc_no_io(_PredId, _ProcId, ProcInfo0, ProcInfo,
+		ModuleInfo0, ModuleInfo) :-
+	proc_info_goal(ProcInfo0, Goal0),
+	proc_info_varset(ProcInfo0, Varset0),
+	proc_info_vartypes(ProcInfo0, VarTypes0),
+	init_slot_info(Varset0, VarTypes0, SlotInfo0),
 
-	{ final_slot_info(Varset1, VarTypes1, SlotInfo) },
-	{ proc_info_headvars(ProcInfo0, HeadVars) },
+	saved_vars_in_goal(Goal0, SlotInfo0, Goal1, SlotInfo),
 
-	% hlds_out__write_goal(Goal1, ModuleInfo, Varset1, 0, ""),
+	final_slot_info(Varset1, VarTypes1, SlotInfo),
+	proc_info_headvars(ProcInfo0, HeadVars),
+
+	% hlds_out__write_goal(Goal1, ModuleInfo, Varset1, 0, "\n"),
 
 	% recompute the nonlocals for each goal
-	{ implicitly_quantify_clause_body(HeadVars, Goal1, Varset1,
-		VarTypes1, Goal2, Varset, VarTypes, _Warnings) },
-	{ proc_info_get_initial_instmap(ProcInfo0, ModuleInfo0, InstMap0) },
-	{ recompute_instmap_delta(no, Goal2, Goal, InstMap0, 
-		ModuleInfo0, ModuleInfo) },
+	implicitly_quantify_clause_body(HeadVars, Goal1, Varset1,
+		VarTypes1, Goal2, Varset, VarTypes, _Warnings),
+	proc_info_get_initial_instmap(ProcInfo0, ModuleInfo0, InstMap0),
+	recompute_instmap_delta(no, Goal2, Goal, VarTypes, InstMap0, 
+		ModuleInfo0, ModuleInfo),
 
-	% hlds_out__write_goal(Goal, ModuleInfo, Varset, 0, ""),
+	% hlds_out__write_goal(Goal, ModuleInfo, Varset, 0, "\n"),
 
-	{ proc_info_set_goal(ProcInfo0, Goal, ProcInfo1) },
-	{ proc_info_set_variables(ProcInfo1, Varset, ProcInfo2) },
-	{ proc_info_set_vartypes(ProcInfo2, VarTypes, ProcInfo) }.
+	proc_info_set_goal(ProcInfo0, Goal, ProcInfo1),
+	proc_info_set_varset(ProcInfo1, Varset, ProcInfo2),
+	proc_info_set_vartypes(ProcInfo2, VarTypes, ProcInfo).
 
 %-----------------------------------------------------------------------------%
 
@@ -83,6 +92,13 @@ saved_vars_in_goal(GoalExpr0 - GoalInfo0, SlotInfo0, Goal, SlotInfo) :-
 		saved_vars_in_conj(Goals0, NonLocals, SlotInfo0,
 			Goals, SlotInfo),
 		conj_list_to_goal(Goals, GoalInfo0, Goal)
+	;
+		GoalExpr0 = par_conj(Goals0, SM),
+			% saved_vars_in_disj treats its goal list as
+			% an independent list of goals, so we can use
+			% it to process the list of parallel conjuncts too.
+		saved_vars_in_disj(Goals0, SlotInfo0, Goals, SlotInfo),
+		Goal = par_conj(Goals, SM) - GoalInfo0
 	;
 		GoalExpr0 = disj(Goals0, SM),
 		saved_vars_in_disj(Goals0, SlotInfo0, Goals, SlotInfo),
@@ -102,11 +118,11 @@ saved_vars_in_goal(GoalExpr0 - GoalInfo0, SlotInfo0, Goal, SlotInfo) :-
 		saved_vars_in_goal(Else0, SlotInfo2, Else, SlotInfo),
 		Goal = if_then_else(Vars, Cond, Then, Else, SM) - GoalInfo0
 	;
-		GoalExpr0 = some(Var, SubGoal0),
+		GoalExpr0 = some(Var, CanRemove, SubGoal0),
 		saved_vars_in_goal(SubGoal0, SlotInfo0, SubGoal, SlotInfo),
-		Goal = some(Var, SubGoal) - GoalInfo0
+		Goal = some(Var, CanRemove, SubGoal) - GoalInfo0
 	;
-		GoalExpr0 = higher_order_call(_, _, _, _, _, _),
+		GoalExpr0 = generic_call(_, _, _, _),
 		Goal = GoalExpr0 - GoalInfo0,
 		SlotInfo = SlotInfo0
 	;
@@ -118,9 +134,13 @@ saved_vars_in_goal(GoalExpr0 - GoalInfo0, SlotInfo0, Goal, SlotInfo) :-
 		Goal = GoalExpr0 - GoalInfo0,
 		SlotInfo = SlotInfo0
 	;
-		GoalExpr0 = pragma_c_code(_, _, _, _, _, _, _, _),
+		GoalExpr0 = pragma_c_code(_, _, _, _, _, _, _),
 		Goal = GoalExpr0 - GoalInfo0,
 		SlotInfo = SlotInfo0
+	;
+		GoalExpr0 = bi_implication(_, _),
+		% these should have been expanded out by now
+		error("saved_vars_in_goal: unexpected bi_implication")
 	),
 	!.
 
@@ -136,7 +156,7 @@ saved_vars_in_goal(GoalExpr0 - GoalInfo0, SlotInfo0, Goal, SlotInfo) :-
 % several such unifications about which one should be closest to a following
 % goal that uses all the variables they define.
 
-:- pred saved_vars_in_conj(list(hlds_goal), set(var), slot_info,
+:- pred saved_vars_in_conj(list(hlds_goal), set(prog_var), slot_info,
 	list(hlds_goal), slot_info).
 :- mode saved_vars_in_conj(in, in, in, out, out) is det.
 
@@ -145,7 +165,7 @@ saved_vars_in_conj([Goal0 | Goals0], NonLocals, SlotInfo0,
 		Goals, SlotInfo) :-
 	(
 		Goal0 = unify(_, _, _, Unif, _) - _,
-		Unif = construct(Var, _, [], _),
+		Unif = construct(Var, _, [], _, _, _, _),
 		skip_constant_constructs(Goals0, Constants, Others),
 		Others = [First | _Rest],
 		can_push(Var, First)
@@ -174,7 +194,7 @@ skip_constant_constructs([], [], []).
 skip_constant_constructs([Goal0 | Goals0], Constants, Others) :-
 	(
 		Goal0 = unify(_, _, _, Unif, _) - _,
-		Unif = construct(_, _, [], _)
+		Unif = construct(_, _, [], _, _, _, _)
 	->
 		skip_constant_constructs(Goals0, Constants1, Others),
 		Constants = [Goal0 | Constants1]
@@ -189,7 +209,7 @@ skip_constant_constructs([Goal0 | Goals0], Constants, Others) :-
 %
 % The logic of this predicate must match the logic of saved_vars_delay_goal.
 
-:- pred can_push(var, hlds_goal).
+:- pred can_push(prog_var, hlds_goal).
 :- mode can_push(in, in) is semidet.
 
 can_push(Var, First) :-
@@ -199,7 +219,7 @@ can_push(Var, First) :-
 		(
 			FirstExpr = conj(_)
 		;
-			FirstExpr = some(_, _)
+			FirstExpr = some(_, _, _)
 		;
 			FirstExpr = not(_)
 		;
@@ -228,7 +248,7 @@ can_push(Var, First) :-
 % If Var is exported from the conjunction, we include Construct at the end
 % of the conjunction to give it its value.
 
-:- pred saved_vars_delay_goal(list(hlds_goal), hlds_goal, var, bool,
+:- pred saved_vars_delay_goal(list(hlds_goal), hlds_goal, prog_var, bool,
 	slot_info, list(hlds_goal), slot_info).
 :- mode saved_vars_delay_goal(in, in, in, in, in, out, out) is det.
 
@@ -265,7 +285,7 @@ saved_vars_delay_goal([Goal0 | Goals0], Construct, Var, IsNonLocal, SlotInfo0,
 				IsNonLocal, SlotInfo1, Goals1, SlotInfo),
 			Goals = [NewConstruct, Goal1 | Goals1]
 		;
-			Goal0Expr = higher_order_call(_, _, _, _, _, _),
+			Goal0Expr = generic_call(_, _, _, _),
 			rename_var(SlotInfo0, Var, _NewVar, Subst, SlotInfo1),
 			goal_util__rename_vars_in_goal(Construct, Subst,
 				NewConstruct),
@@ -274,7 +294,7 @@ saved_vars_delay_goal([Goal0 | Goals0], Construct, Var, IsNonLocal, SlotInfo0,
 				IsNonLocal, SlotInfo1, Goals1, SlotInfo),
 			Goals = [NewConstruct, Goal1 | Goals1]
 		;
-			Goal0Expr = pragma_c_code(_, _, _, _, _, _, _, _),
+			Goal0Expr = pragma_c_code(_, _, _, _, _, _, _),
 			rename_var(SlotInfo0, Var, _NewVar, Subst, SlotInfo1),
 			goal_util__rename_vars_in_goal(Construct, Subst,
 				NewConstruct),
@@ -288,7 +308,12 @@ saved_vars_delay_goal([Goal0 | Goals0], Construct, Var, IsNonLocal, SlotInfo0,
 			saved_vars_delay_goal(Goals1, Construct, Var,
 				IsNonLocal, SlotInfo0, Goals, SlotInfo)
 		;
-			Goal0Expr = some(SomeVars, SomeGoal0),
+			Goal0Expr = par_conj(_ParConj, _SM),
+			saved_vars_delay_goal(Goals0, Construct, Var,
+				IsNonLocal, SlotInfo0, Goals1, SlotInfo),
+			Goals = [Goal0|Goals1]
+		;
+			Goal0Expr = some(SomeVars, CanRemove, SomeGoal0),
 			rename_var(SlotInfo0, Var, NewVar, Subst, SlotInfo1),
 			goal_util__rename_vars_in_goal(Construct, Subst,
 				NewConstruct),
@@ -296,7 +321,8 @@ saved_vars_delay_goal([Goal0 | Goals0], Construct, Var, IsNonLocal, SlotInfo0,
 				SomeGoal1),
 			push_into_goal(SomeGoal1, NewConstruct, NewVar,
 				SlotInfo1, SomeGoal, SlotInfo2),
-			Goal1 = some(SomeVars, SomeGoal) - Goal0Info,
+			Goal1 = some(SomeVars, CanRemove, SomeGoal)
+					- Goal0Info,
 			saved_vars_delay_goal(Goals0, Construct, Var,
 				IsNonLocal, SlotInfo2, Goals1, SlotInfo),
 			Goals = [Goal1 | Goals1]
@@ -351,6 +377,10 @@ saved_vars_delay_goal([Goal0 | Goals0], Construct, Var, IsNonLocal, SlotInfo0,
 			saved_vars_delay_goal(Goals0, Construct, Var,
 				IsNonLocal, SlotInfo3, Goals1, SlotInfo),
 			Goals = [Goal1 | Goals1]
+		;
+			Goal0Expr = bi_implication(_, _),
+			% these should have been expanded out by now
+			error("saved_vars_delay_goal: unexpected bi_implication")
 		)
 	;
 		saved_vars_delay_goal(Goals0, Construct, Var, IsNonLocal,
@@ -361,7 +391,7 @@ saved_vars_delay_goal([Goal0 | Goals0], Construct, Var, IsNonLocal, SlotInfo0,
 % Push a non-renamed version of the given construction into the given goal.
 % Also traverse the goal looking for further opprtunities.
 
-:- pred push_into_goal(hlds_goal, hlds_goal, var, slot_info,
+:- pred push_into_goal(hlds_goal, hlds_goal, prog_var, slot_info,
 	hlds_goal, slot_info).
 :- mode push_into_goal(in, in, in, in, out, out) is det.
 
@@ -378,7 +408,7 @@ push_into_goal(Goal0, Construct, Var, SlotInfo0, Goal, SlotInfo) :-
 % then this would have no effect, so we merely traverse the goal looking
 % for other opportunities.
 
-:- pred push_into_goal_rename(hlds_goal, hlds_goal, var, slot_info,
+:- pred push_into_goal_rename(hlds_goal, hlds_goal, prog_var, slot_info,
 	hlds_goal, slot_info).
 :- mode push_into_goal_rename(in, in, in, in, out, out) is det.
 
@@ -397,7 +427,7 @@ push_into_goal_rename(Goal0, Construct, Var, SlotInfo0, Goal, SlotInfo) :-
 
 % Push renamed versions of the given construction into each of several goals.
 
-:- pred push_into_goals_rename(list(hlds_goal), hlds_goal, var, slot_info,
+:- pred push_into_goals_rename(list(hlds_goal), hlds_goal, prog_var, slot_info,
 	list(hlds_goal), slot_info).
 :- mode push_into_goals_rename(in, in, in, in, out, out) is det.
 
@@ -411,7 +441,7 @@ push_into_goals_rename([Goal0 | Goals0], Construct, Var, SlotInfo0,
 
 % Push renamed versions of the given construction into each of several cases.
 
-:- pred push_into_cases_rename(list(case), hlds_goal, var, slot_info,
+:- pred push_into_cases_rename(list(case), hlds_goal, prog_var, slot_info,
 	list(case), slot_info).
 :- mode push_into_cases_rename(in, in, in, in, out, out) is det.
 
@@ -425,6 +455,9 @@ push_into_cases_rename([case(ConsId, Goal0) | Cases0], Construct, Var,
 
 %-----------------------------------------------------------------------------%
 
+	% saved_vars_in_disj does a saved_vars_in_goal on an list of independent
+	% goals, and is used to process disjunctions and parallel conjunctions.
+
 :- pred saved_vars_in_disj(list(hlds_goal), slot_info,
 	list(hlds_goal), slot_info).
 :- mode saved_vars_in_disj(in, in, out, out) is det.
@@ -447,21 +480,22 @@ saved_vars_in_switch([case(Cons, Goal0) | Cases0], SlotInfo0,
 %-----------------------------------------------------------------------------%
 
 :- type slot_info	--->	slot_info(
-					varset,
-					map(var, type)
+					prog_varset,
+					map(prog_var, type)
 				).
 
-:- pred init_slot_info(varset, map(var, type), slot_info).
+:- pred init_slot_info(prog_varset, map(prog_var, type), slot_info).
 :- mode init_slot_info(in, in, out) is det.
 
 init_slot_info(Varset, VarTypes, slot_info(Varset, VarTypes)).
 
-:- pred final_slot_info(varset, map(var, type), slot_info).
+:- pred final_slot_info(prog_varset, map(prog_var, type), slot_info).
 :- mode final_slot_info(out, out, in) is det.
 
 final_slot_info(Varset, VarTypes, slot_info(Varset, VarTypes)).
 
-:- pred rename_var(slot_info, var, var, map(var, var), slot_info).
+:- pred rename_var(slot_info, prog_var, prog_var, map(prog_var, prog_var),
+		slot_info).
 :- mode rename_var(in, in, out, out, out) is det.
 
 rename_var(SlotInfo0, Var, NewVar, Substitution, SlotInfo) :-
