@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1998 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -15,7 +15,8 @@
 
 :- module special_pred.
 :- interface.
-:- import_module list, map, prog_data, hlds_data, hlds_pred.
+:- import_module prog_data, hlds_data, hlds_pred.
+:- import_module list, map, std_util.
 
 :- type special_pred_map	==	map(special_pred, pred_id).
 
@@ -33,7 +34,7 @@
 	% special_pred_name_arity(SpecialPredType, GenericPredName,
 	%		TypeSpecificVersionPredName, Arity):
 	%	true iff there is a special predicate of category
-	%	SpecialPredType, called mercury_builtin:GenericPredName/Arity,
+	%	SpecialPredType, called builtin:GenericPredName/Arity,
 	%	for which the type-specific versions will be called
 	%	TypeSpecificVersionPredName.
 :- pred special_pred_name_arity(special_pred_id, string, string, int).
@@ -50,10 +51,12 @@
 :- pred special_pred_get_type(string, list(Type), Type).
 :- mode special_pred_get_type(in, in, out) is semidet.
 
+:- pred special_pred_description(special_pred_id, string).
+:- mode special_pred_description(in, out) is det.
+
 :- implementation.
 
-:- import_module type_util.
-:- import_module std_util.
+:- import_module type_util, mode_util, prog_util.
 
 special_pred_list([unify, index, compare]).
 
@@ -61,11 +64,8 @@ special_pred_name_arity(unify, "unify", "__Unify__", 2).
 special_pred_name_arity(index, "index", "__Index__", 2).
 special_pred_name_arity(compare, "compare", "__Compare__", 3).
 
-	% mode num is 0 for semidet, 10000 for det
-	% see make_hlds.m
-special_pred_mode_num(unify, 0).
-special_pred_mode_num(index, 10000).
-special_pred_mode_num(compare, 10000).
+	% mode num for special procs is always 0 (the first mode)
+special_pred_mode_num(_, 0).
 
 special_pred_info(unify, Type, "__Unify__", [Type, Type], [In, In], semidet) :-
 	in_mode(In).
@@ -76,20 +76,12 @@ special_pred_info(index, Type, "__Index__", [Type, IntType], [In, Out], det) :-
 	out_mode(Out).
 
 special_pred_info(compare, Type,
-		 "__Compare__", [ResType, Type, Type], [Out, In, In], det) :-
-	construct_type(qualified("mercury_builtin", "comparison_result") - 0,
+		 "__Compare__", [ResType, Type, Type], [Uo, In, In], det) :-
+	mercury_public_builtin_module(PublicBuiltin),
+	construct_type(qualified(PublicBuiltin, "comparison_result") - 0,
 							[], ResType),
 	in_mode(In),
-	out_mode(Out).
-
-:- pred in_mode((mode)::out) is det.
-
-in_mode(user_defined_mode(qualified("mercury_builtin", "in"), [])).
-
-:- pred out_mode((mode)::out) is det.
-
-out_mode(user_defined_mode(qualified("mercury_builtin", "out"), [])).
-
+	uo_mode(Uo).
 
 	% Given the mangled predicate name and the list of argument types,
 	% work out which type this special predicate is for.
@@ -104,9 +96,19 @@ out_mode(user_defined_mode(qualified("mercury_builtin", "out"), [])).
 
 special_pred_get_type("__Unify__", Types, T) :-
 	list__reverse(Types, [T | _]).
+special_pred_get_type("unify", Types, T) :-
+	list__reverse(Types, [T | _]).
 special_pred_get_type("__Index__", Types, T) :-
 	list__reverse(Types, [_, T | _]).
+special_pred_get_type("index", Types, T) :-
+	list__reverse(Types, [_, T | _]).
 special_pred_get_type("__Compare__", Types, T) :-
 	list__reverse(Types, [T | _]).
+special_pred_get_type("compare", Types, T) :-
+	list__reverse(Types, [T | _]).
+
+special_pred_description(unify, "unification predicate").
+special_pred_description(compare, "comparison predicate").
+special_pred_description(index, "indexing predicate").
 
 %-----------------------------------------------------------------------------%
