@@ -1,16 +1,14 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1997 The University of Melbourne.
+% Copyright (C) 1996-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
-%  Public License - see the file COPYING in the Mercury distribution.
+% Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
 
 % stratify.m - the stratification analysis pass.
 
 % Main authors: ohutch, conway.
 
-% This module performs stratification analysis. It is highly
-% conservative and will emit errors for any calls outside the module or
-% for any calls to higher order predicates
+% This module performs stratification analysis. 
 % It works by processing the call graph 1 scc at a time. It traverses 
 % the goal for each procedure in the scc and reports an error or 
 % warning (depending on the context) for any negated call to another member 
@@ -37,19 +35,15 @@
 
 :- import_module hlds_module, io.
 
-
 	% Perform stratification analysis, for the given module.
 	% If the "warn-non-stratification" option is set this 
 	% pred will check the entire module for stratification
 	% otherwise it will only check preds in the stratified_preds 
-	% set of the module_info structure. If possible non 
-	% stratification is detected in loops containing these preds 
-	% an error will be issued otherwise a warning will be issued.
+	% set of the module_info structure. 
 :- pred stratify__check_stratification(module_info, module_info,
 		io__state, io__state).
 :- mode stratify__check_stratification(in, out, di, uo) is det.
 
-
 :- implementation.
 
 :- import_module dependency_graph, hlds_pred, hlds_goal, hlds_data.
@@ -57,8 +51,7 @@
 :- import_module prog_out, globals, options.
 
 :- import_module assoc_list, map, list, set, bool, std_util, relation, require.
-
-
+:- import_module string.
 
 stratify__check_stratification(Module0, Module) -->
 	{ module_info_ensure_dependency_info(Module0, Module1) },
@@ -81,8 +74,6 @@ stratify__check_stratification(Module0, Module) -->
 	%{ dep_sets_to_lists_and_sets(HOSCCs1, [], HOSCCs) },
 	%higher_order_check_sccs(HOSCCs, HOInfo, Module2, Module).
 
-
-
 %-----------------------------------------------------------------------------%
 
 :- pred dep_sets_to_lists_and_sets(list(set(pred_proc_id)), 
@@ -107,27 +98,24 @@ get_proc_id(proc(PredId, _), PredId).
 :- mode first_order_check_sccs(in, in, in, in, out, di, uo) is det.
 
 first_order_check_sccs([], _, _, Module, Module) --> [].
-first_order_check_sccs([SCCl - SCCs|Rest], StratifiedPreds, Warn, 
+first_order_check_sccs([SCCl - SCCs|Rest], StratifiedPreds, Warn0, 
 		Module0, Module) -->
-	(	% if there is a pred that must be stratified in the 
-		% current scc then emit errors instead of warnings
-		% if we encounter non stratification
+	(
 		{ set__intersect(SCCs, StratifiedPreds, I) },
 		{ set__empty(I) }
 	->
-		{ Error = no }
+		{ Warn = Warn0 }
 	;
-		{ Error = yes }
+		{ Warn = yes }
 	),
 	(
-		{ Error = yes 
-		; Warn = yes }
+		{ Warn = yes }
 	->
-		first_order_check_scc(SCCl, Error, Module0, Module1)
+		first_order_check_scc(SCCl, no, Module0, Module1)
 	;
 		{ Module1 = Module0 }
 	),
-	first_order_check_sccs(Rest, StratifiedPreds, Warn, Module1, Module).
+	first_order_check_sccs(Rest, StratifiedPreds, Warn0, Module1, Module).
 
 :- pred first_order_check_scc(list(pred_proc_id), bool, module_info, 
 	module_info, io__state, io__state).
@@ -161,6 +149,10 @@ first_order_check_goal(conj(Goals), _GoalInfo, Negated, WholeScc,
 		ThisPredProcId, Error, Module0, Module) -->
 	first_order_check_goal_list(Goals, Negated, WholeScc, ThisPredProcId,
 		Error, Module0, Module).
+first_order_check_goal(par_conj(Goals, _SM), _GoalInfo, Negated, WholeScc, 
+		ThisPredProcId, Error, Module0, Module) -->
+	first_order_check_goal_list(Goals, Negated, WholeScc, ThisPredProcId,
+		Error, Module0, Module).
 first_order_check_goal(disj(Goals, _Follow), _GoalInfo, Negated, 
 		WholeScc, ThisPredProcId, Error, Module0, Module) -->
 	first_order_check_goal_list(Goals, Negated, WholeScc, ThisPredProcId,
@@ -178,7 +170,7 @@ first_order_check_goal(if_then_else(_Vars, Cond - CInfo, Then - TInfo,
 		Error, Module1, Module2),
 	first_order_check_goal(Else, EInfo, Negated, WholeScc, ThisPredProcId,
 		Error, Module2, Module).
-first_order_check_goal(some(_Vars, Goal - GoalInfo), _GoalInfo,
+first_order_check_goal(some(_Vars, _, Goal - GoalInfo), _GoalInfo,
 		Negated, WholeScc, ThisPredProcId, Error, Module0, Module) -->
 	first_order_check_goal(Goal, GoalInfo, Negated, WholeScc, 
 		ThisPredProcId, Error, Module0, Module).
@@ -186,7 +178,7 @@ first_order_check_goal(not(Goal - GoalInfo), _GoalInfo, _Negated,
 		WholeScc, ThisPredProcId, Error, Module0, Module) -->
 	first_order_check_goal(Goal, GoalInfo, yes, WholeScc, ThisPredProcId,
 		Error, Module0, Module).
-first_order_check_goal(pragma_c_code(_, _IsRec, CPred, CProc, _, _, _, _), 
+first_order_check_goal(pragma_c_code(_IsRec, CPred, CProc, _, _, _, _), 
 		GoalInfo, Negated, WholeScc, ThisPredProcId, 
 		Error, Module0, Module) -->
 	(
@@ -201,8 +193,8 @@ first_order_check_goal(pragma_c_code(_, _IsRec, CPred, CProc, _, _, _, _),
 		{ Module = Module0 }
 	).
 first_order_check_goal(unify(_Var, _RHS, _Mode, _Uni, _Context), _GoalInfo,
-		_Negated, _WholeScc, _ThisPredProcId, _, Module, Module) --> [].
-first_order_check_goal(call(CPred, CProc, _Args, BuiltinState, _Contex, _Sym), 
+	_Negated, _WholeScc, _ThisPredProcId, _, Module, Module) --> [].
+first_order_check_goal(call(CPred, CProc, _Args, _BuiltinState, _Contex, _Sym), 
 		GInfo, Negated, WholeScc, ThisPredProcId, 
 		Error, Module0, Module) -->
 	{ Callee = proc(CPred, CProc) },
@@ -214,26 +206,15 @@ first_order_check_goal(call(CPred, CProc, _Args, BuiltinState, _Contex, _Sym),
 		emit_message(ThisPredProcId, Context, 
 			"call introduces a non-stratified loop", 
 			Error, Module0, Module)	
-	;
-		{ BuiltinState = not_builtin },
-		{ \+ local_proc(Module0, Callee) }
-	->
-		{ goal_info_get_context(GInfo, Context) },
-	        emit_message(ThisPredProcId, Context,
-                        "call to non-local predicate may introduce a non-stratified loop",
-                        Error, Module0, Module)
-										
 	;
 		{ Module = Module0 }
 	).
-first_order_check_goal(higher_order_call(_Var, _Vars, _Types, _Modes, 
-			_Det, _PredOrFunc),
-		GInfo, _Negated, _WholeScc, ThisPredProcId, Error,  
-		Module0, Module) --> 
-	{ goal_info_get_context(GInfo, Context) },
-	emit_message(ThisPredProcId, Context,
-		"higher order call may introduce a non-stratified loop",
-		Error, Module0, Module).
+first_order_check_goal(generic_call(_Var, _Vars, _Modes, _Det),
+	_GInfo, _Negated, _WholeScc, _ThisPredProcId, 
+	_Error,  Module, Module) --> []. 
+first_order_check_goal(bi_implication(_, _), _, _, _, _, _, _, _) -->
+	% these should have been expanded out by now
+	{ error("first_order_check_goal: unexpected bi_implication") }.
 
 :- pred first_order_check_goal_list(list(hlds_goal), bool, 
 	list(pred_proc_id), pred_proc_id, bool, module_info, 
@@ -263,13 +244,6 @@ first_order_check_case_list([Case|Goals], Negated, WholeScc, ThisPredProcId,
 	first_order_check_case_list(Goals, Negated, WholeScc, ThisPredProcId,
 		Error, Module1, Module).
 
-:- pred local_proc(module_info, pred_proc_id).
-:- mode local_proc(in, in) is semidet.
-local_proc(Module, proc(PredId, ProcId)) :-
-	module_info_pred_info(Module, PredId, PredInfo),
-	pred_info_non_imported_procids(PredInfo, ProcIds),
-	list__member(ProcId, ProcIds).
-
 %-----------------------------------------------------------------------------%
 
  % XXX : Currently we don't allow the higher order case so this code
@@ -296,16 +270,8 @@ higher_order_check_scc([PredProcId|Remaining], WholeScc, HOInfo, Module0,
 	{ PredProcId = proc(PredId, ProcId) },
 	{ module_info_pred_info(Module0, PredId, PredInfo) },
 	globals__io_lookup_bool_option(warn_non_stratification, Warn),
-	{ pred_info_get_marker_list(PredInfo, Markers) },
-	( 	
-		{ list__member(request(memo), Markers) }
-	->
-		{ Error = yes }
-	;
-		{ Error = no }
-	),
-	(	( { Error = yes 
-		; Warn = yes } ),
+	{ Error = no },
+	(	( { Error = yes ; Warn = yes } ),
 		{ map__search(HOInfo, PredProcId, HigherOrderInfo) }
 	->
 		{ HigherOrderInfo = info(HOCalls, _) },
@@ -337,6 +303,10 @@ higher_order_check_goal(conj(Goals), _GoalInfo, Negated, WholeScc,
 		ThisPredProcId, HighOrderLoops, Error, Module0, Module) -->
 	higher_order_check_goal_list(Goals, Negated, WholeScc, ThisPredProcId,
 		HighOrderLoops, Error, Module0, Module).
+higher_order_check_goal(par_conj(Goals, _), _GoalInfo, Negated, WholeScc, 
+		ThisPredProcId, HighOrderLoops, Error, Module0, Module) -->
+	higher_order_check_goal_list(Goals, Negated, WholeScc, ThisPredProcId,
+		HighOrderLoops, Error, Module0, Module).
 higher_order_check_goal(disj(Goals, _Follow), _GoalInfo, Negated, WholeScc, 
 		ThisPredProcId, HighOrderLoops, Error, Module0, Module) -->
 	higher_order_check_goal_list(Goals, Negated, WholeScc, ThisPredProcId,
@@ -355,7 +325,7 @@ higher_order_check_goal(if_then_else(_Vars, Cond - CInfo, Then - TInfo,
 		HighOrderLoops, Error, Module1, Module2),
 	higher_order_check_goal(Else, EInfo, Negated, WholeScc, ThisPredProcId,
 		HighOrderLoops, Error, Module2, Module).
-higher_order_check_goal(some(_Vars, Goal - GoalInfo), _GoalInfo, Negated, 
+higher_order_check_goal(some(_Vars, _, Goal - GoalInfo), _GoalInfo, Negated, 
 		WholeScc, ThisPredProcId, HighOrderLoops, 
 		Error, Module0, Module) -->
 	higher_order_check_goal(Goal, GoalInfo, Negated, WholeScc, 
@@ -364,7 +334,7 @@ higher_order_check_goal(not(Goal - GoalInfo), _GoalInfo, _Negated, WholeScc,
 		ThisPredProcId, HighOrderLoops, Error, Module0, Module) -->
 	higher_order_check_goal(Goal, GoalInfo, yes, WholeScc, ThisPredProcId,
 		HighOrderLoops, Error, Module0, Module).
-higher_order_check_goal(pragma_c_code(_, _IsRec, _, _, _, _, _, _), _GoalInfo, 
+higher_order_check_goal(pragma_c_code(_IsRec, _, _, _, _, _, _), _GoalInfo, 
 	_Negated, _WholeScc, _ThisPredProcId, _HighOrderLoops, 
 	_, Module, Module) --> [].
 higher_order_check_goal(unify(_Var, _RHS, _Mode, _Uni, _Context), _GoalInfo,
@@ -391,22 +361,29 @@ higher_order_check_goal((call(_CPred, _CProc, _Args, _Builtin, _Contex, Sym)),
 		{ Module = Module0 }
 	).
 	
-higher_order_check_goal(higher_order_call(_Var, _Vars, _Types, _Modes, _Det,
-			_PredOrFunc), 
+higher_order_check_goal(generic_call(GenericCall, _Vars, _Modes, _Det), 
 		GoalInfo, Negated, _WholeScc, ThisPredProcId, HighOrderLoops, 
 		Error, Module0, Module) -->
 	(
 		{ Negated = yes },
-		{ HighOrderLoops = yes }
+		{ HighOrderLoops = yes },
+		{ GenericCall = higher_order(_, _, _), Msg = "higher order"
+		; GenericCall = class_method(_, _, _, _), Msg = "class method"
+		}
 	->
 		{ goal_info_get_context(GoalInfo, Context) },
-		emit_message(ThisPredProcId, Context, 
-			"higher order call may introduce a non-stratified loop", 
+		{ string__append(Msg, 
+			" call may introduce a non-stratified loop",
+			ErrorMsg) },
+		emit_message(ThisPredProcId, Context, ErrorMsg,
 			Error, Module0, Module)		
 	;
 		{ Module = Module0 }
 	).
-	
+higher_order_check_goal(bi_implication(_, _), _, _, _, _, _, _, _, _) -->
+	% these should have been expanded out by now
+	{ error("higher_order_check_goal: unexpected bi_implication") }.
+
 :- pred higher_order_check_goal_list(list(hlds_goal), bool, set(pred_proc_id),
 	pred_proc_id, bool, bool, module_info, module_info, 
 	io__state, io__state).
@@ -482,7 +459,6 @@ gen_conservative_graph(Module, DepGraph0, DepGraph, HOInfo) :-
 	map__to_assoc_list(HOInfo, HOInfoL),
 	add_new_arcs(HOInfoL, CallsHO, DepGraph0, DepGraph).
 
-
 	% For a given module collects for each non imported proc a set 
 	% of called procs and a higher order info structure. This pred 
 	% also returns a set of all non imported procs that make a 
@@ -498,7 +474,6 @@ get_call_info(Module, ProcCalls, HOInfo, CallsHO) :-
 	expand_predids(PredIds, Module, ProcCalls0, ProcCalls, HOInfo0, 
 		HOInfo, CallsHO0, CallsHO). 
 
-
 	% find the transitive closure of a given list of procs
 	% this pred is used to see how face a higher order address can
 	% reach though proc calls
@@ -531,7 +506,6 @@ tc([P|Ps], ProcCalls, CallsHO, HOInfo0, HOInfo, Changed0, Changed) :-
 		Changed0, Changed1),
 	tc(Ps, ProcCalls, CallsHO, HOInfo1, HOInfo, Changed1, Changed).
 
-
 	% merge any higher order addresses that can pass between the
 	% given caller and callees. This code also merges any possible
 	% addresses that can pass in and out of higher order calls
@@ -627,7 +601,6 @@ merge_calls([C|Cs], P, CallsHO, DoingFirstOrder, HOInfo0, HOInfo, Changed0,
 			Changed0, Changed)
 	).
  
-
 	% given the set of procs that make higher order calls and a
 	% list of procs and higher order call info this pred rebuilds
 	% the given call graph with new arcs for every possible higher
@@ -662,8 +635,6 @@ add_new_arcs2([Callee|Cs], CallerKey, DepGraph0, DepGraph) :-
 	relation__add(DepGraph0, CallerKey, CalleeKey, DepGraph1),
 	add_new_arcs2(Cs, CallerKey, DepGraph1, DepGraph).
 
-
-
 	% for each given pred id pass all non imported procs onto the
 	% process_procs pred
 :- pred expand_predids(list(pred_id), module_info, call_map, call_map, 
@@ -676,13 +647,12 @@ expand_predids([PredId|PredIds], Module, ProcCalls0, ProcCalls, HOInfo0,
 	module_info_pred_info(Module, PredId, PredInfo),
 	pred_info_non_imported_procids(PredInfo, Procs),
 	pred_info_procedures(PredInfo, ProcTable),
-	pred_info_arg_types(PredInfo, _TVarSet, ArgTypes),
+	pred_info_arg_types(PredInfo, ArgTypes),
 	process_procs(Procs, Module, PredId, ArgTypes, ProcTable, ProcCalls0, 
 		ProcCalls1, HOInfo0, HOInfo1, CallsHO0, CallsHO1), 
 	expand_predids(PredIds, Module, ProcCalls1, ProcCalls, HOInfo1, 
 		HOInfo, CallsHO1, CallsHO).
 
-	
 	% for each given proc id generate the set of procs it calls and
 	% its higher order info structure
 :- pred process_procs(list(proc_id), module_info, pred_id, list(type), 
@@ -713,7 +683,6 @@ process_procs([ProcId|Procs], Module, PredId, ArgTypes, ProcTable, ProcCalls0,
 	process_procs(Procs, Module, PredId, ArgTypes, ProcTable, ProcCalls1,
 		ProcCalls, HOInfo1, HOInfo, CallsHO1, CallsHO).
 	
-
 	% determine if a given set of modes and types indicates that
 	% higher order values can be passed into and/or out of a proc
 :- pred higherorder_in_out(list(type), list(mode), module_info, ho_in_out). 
@@ -746,7 +715,7 @@ higherorder_in_out1([Type|Types], [Mode|Modes], Module, HOIn0, HOIn,
 		% XXX : will have to use a more general check for higher
 		% order constants in parameters user could hide higher
 		% order consts in a data structure etc..
-		type_is_higher_order(Type, _, _)
+		type_is_higher_order(Type, _, _, _)
 	->	
 		(
 			mode_is_input(Module, Mode) 
@@ -768,7 +737,6 @@ higherorder_in_out1([Type|Types], [Mode|Modes], Module, HOIn0, HOIn,
 	),
 	higherorder_in_out1(Types, Modes, Module, HOIn1, HOIn, HOOut1, HOOut).
 	
-
 	% return the set of all procs called in and all addresses
 	% taken, in a given goal
 :- pred check_goal(hlds_goal_expr, set(pred_proc_id), set(pred_proc_id), 
@@ -792,8 +760,8 @@ check_goal1(unify(_Var, RHS, _Mode, Unification, _Context), Calls,
 		% lambda goal have addresses taken. this is not
 		% always to case, but should be a suitable approximation for
 		% the stratification analysis
-		RHS = lambda_goal(_PredOrFunc, _Vars, _Modes, _Determinism,
-		                                Goal - _GoalInfo)
+		RHS = lambda_goal(_PredOrFunc, _EvalMethod, _Fix, _NonLocals,
+				_Vars, _Modes, _Determinism, Goal - _GoalInfo)
 	->
 		get_called_procs(Goal, [], CalledProcs),
 		set__insert_list(HasAT0, CalledProcs, HasAT)
@@ -802,11 +770,11 @@ check_goal1(unify(_Var, RHS, _Mode, Unification, _Context), Calls,
 		% currently when this pass is run the construct/4
 		% case will not happen as higher order constants have
 		% been transformed to lambda goals. see above
-		Unification = construct(_Var2, ConsId, _VarList, _ModeList)
+		Unification = construct(_Var2, ConsId, _, _, _, _, _)
 	->
 		(
 			(
-				ConsId = pred_const(PredId, ProcId)
+				ConsId = pred_const(PredId, ProcId, _)
 			;
 				ConsId = code_addr_const(PredId, ProcId)
 			)
@@ -825,11 +793,14 @@ check_goal1(call(CPred, CProc, _Args, _Builtin, _Contex, _Sym), Calls0, Calls,
 	set__insert(Calls0, proc(CPred, CProc), Calls).
 
 	% record that the higher order call was made
-check_goal1(higher_order_call(_Var, _Vars, _Types, _Modes, _Det, _PredOrFUnc),
+check_goal1(generic_call(_Var, _Vars, _Modes, _Det),
 		Calls, Calls, HasAT, HasAT, _, yes).
 
 check_goal1(conj(Goals), Calls0, Calls, HasAT0, HasAT, CallsHO0, CallsHO) :-
 	check_goal_list(Goals, Calls0, Calls, HasAT0, HasAT, CallsHO0, CallsHO).
+check_goal1(par_conj(Goals, _), Calls0, Calls, HasAT0, HasAT,
+		CallsHO0, CallsHO) :-
+	check_goal_list(Goals, Calls0, Calls, HasAT0, HasAT, CallsHO0, CallsHO).
 check_goal1(disj(Goals, _Follow), Calls0, Calls, HasAT0, HasAT, CallsHO0, 
 		CallsHO) :-
 	check_goal_list(Goals, Calls0, Calls, HasAT0, HasAT, CallsHO0, CallsHO).
@@ -842,7 +813,7 @@ check_goal1(if_then_else(_Vars, Cond - _CInfo, Then - _TInfo, Else - _EInfo,
 	check_goal1(Then, Calls1, Calls2, HasAT1, HasAT2, CallsHO1, CallsHO2),
 	check_goal1(Else, Calls2, Calls, HasAT2, HasAT, CallsHO2, CallsHO).
 	
-check_goal1(some(_Vars, Goal - _GoalInfo), Calls0, Calls, HasAT0, HasAT, 
+check_goal1(some(_Vars, _, Goal - _GoalInfo), Calls0, Calls, HasAT0, HasAT, 
 		CallsHO0, CallsHO) :- 
 	check_goal1(Goal, Calls0, Calls, HasAT0, HasAT, CallsHO0, CallsHO).
 
@@ -850,9 +821,12 @@ check_goal1(not(Goal - _GoalInfo), Calls0, Calls, HasAT0, HasAT, CallsHO0,
 		CallsHO) :- 
 	check_goal1(Goal, Calls0, Calls, HasAT0, HasAT, CallsHO0, CallsHO).
 
-check_goal1(pragma_c_code(_, _IsRec, _CPred, _CProc, _, _, _, _), Calls, Calls, 
+check_goal1(pragma_c_code(_IsRec, _CPred, _CProc, _, _, _, _), Calls, Calls, 
 		HasAT, HasAT, CallsHO, CallsHO).
 
+check_goal1(bi_implication(_, _), _, _, _, _, _, _) :-
+	% these should have been expanded out by now
+	error("check_goal1: unexpected bi_implication").
 	
 :- pred check_goal_list(list(hlds_goal), set(pred_proc_id), set(pred_proc_id), 
 	set(pred_proc_id), set(pred_proc_id), bool, bool). 
@@ -875,7 +849,6 @@ check_case_list([Case|Goals], Calls0, Calls, HasAT0, HasAT, CallsHO0,
 	check_goal1(Goal, Calls0, Calls1, HasAT0, HasAT1, CallsHO0, CallsHO1),
 	check_case_list(Goals, Calls1, Calls, HasAT1, HasAT, CallsHO1, CallsHO).
 
-
 	% This pred returns a list of all the calls in a given set of
 	% goals including calls in unification lambda functions and
 	% pred_proc_id's in constructs 
@@ -890,19 +863,19 @@ get_called_procs(unify(_Var, RHS, _Mode, Unification, _Context), Calls0,
 		% lambda goal have addresses taken. this is not
 		% always to case, but should be a suitable approximation for
 		% the stratification analysis
-		RHS = lambda_goal(_PredOrFunc, _Vars, _Modes, _Determinism,
-		                                Goal - _GoalInfo)
+		RHS = lambda_goal(_PredOrFunc, _EvalMethod, _Fix, _NonLocals,
+				_Vars, _Modes, _Determinism, Goal - _GoalInfo)
 	->
 		get_called_procs(Goal, Calls0, Calls)
 	;
 		% currently when this pass is run the construct/4
 		% case will not happen as higher order constants have
 		% been transformed to lambda goals see above
-		Unification = construct(_Var2, ConsId, _VarList, _ModeList)
+		Unification = construct(_Var2, ConsId, _, _, _, _, _)
 	->
 		(
 			(
-				ConsId = pred_const(PredId, ProcId)
+				ConsId = pred_const(PredId, ProcId, _)
 			;
 				ConsId = code_addr_const(PredId, ProcId)
 			)
@@ -915,18 +888,17 @@ get_called_procs(unify(_Var, RHS, _Mode, Unification, _Context), Calls0,
 		Calls = Calls0	
 	).
 	
-
 	% add this call to the call list
 get_called_procs(call(CPred, CProc, _Args, _Builtin, _Contex, _Sym), Calls0, 
 		Calls) :- 
 	Calls = [proc(CPred, CProc) | Calls0].
 
-get_called_procs(higher_order_call(_Var, _Vars, _Types, _Modes, _Det,
-		_PredOrFunc), Calls, Calls).
-
+get_called_procs(generic_call(_Var, _Vars, _Modes, _Det), Calls, Calls).
 
 get_called_procs(conj(Goals), Calls0, Calls) :-
 	check_goal_list(Goals, Calls0, Calls).
+get_called_procs(par_conj(Goals, _), Calls0, Calls) :-
+	check_goal_list(Goals, Calls0, Calls).
 get_called_procs(disj(Goals, _Follow), Calls0, Calls) :-
 	check_goal_list(Goals, Calls0, Calls).
 get_called_procs(switch(_Var, _Fail, Cases, _Follow), Calls0, Calls) :-
@@ -936,13 +908,15 @@ get_called_procs(if_then_else(_Vars, Cond - _CInfo, Then - _TInfo,
 	get_called_procs(Cond, Calls0, Calls1),
 	get_called_procs(Then, Calls1, Calls2),
 	get_called_procs(Else, Calls2, Calls). 
-get_called_procs(some(_Vars, Goal - _GoalInfo), Calls0, Calls) :-
+get_called_procs(some(_Vars, _, Goal - _GoalInfo), Calls0, Calls) :-
 	get_called_procs(Goal, Calls0, Calls).
 get_called_procs(not(Goal - _GoalInfo), Calls0, Calls) :-
 	get_called_procs(Goal, Calls0, Calls).
-get_called_procs(pragma_c_code(_, _IsRec, _CPred, _CProc, _, _, _, _),
+get_called_procs(pragma_c_code(_IsRec, _CPred, _CProc, _, _, _, _),
 	Calls, Calls).
-
+get_called_procs(bi_implication(_, _), _, _) :-
+	% these should have been expanded out by now
+	error("get_called_procs: unexpected bi_implication").
 
 :- pred check_goal_list(list(hlds_goal), list(pred_proc_id), 
 	list(pred_proc_id)).
@@ -965,9 +939,7 @@ check_case_list([Case|Goals], Calls0, Calls) :-
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
-
-
-:- pred emit_message(pred_proc_id, term__context, string, bool, 
+:- pred emit_message(pred_proc_id, prog_context, string, bool, 
 		module_info, module_info, io__state, io__state).
 :- mode emit_message(in, in, in, in, in, out, di, uo) is det.
 
