@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -17,23 +17,23 @@
 
 :- interface.
 
-:- import_module hlds_goal, llds, switch_gen, code_info.
-:- import_module term.
+:- import_module hlds_data, hlds_goal, llds, switch_gen, code_info, prog_data.
 
-:- pred string_switch__generate(cases_list, var, code_model,
-	can_fail, store_map, label, code_tree, code_info, code_info).
-:- mode string_switch__generate(in, in, in, in, in, in, out, in, out)
+:- pred string_switch__generate(cases_list, prog_var, code_model,
+	can_fail, store_map, label, branch_end, branch_end, code_tree,
+	code_info, code_info).
+:- mode string_switch__generate(in, in, in, in, in, in, in, out, out, in, out)
 	is det.
 
 %-----------------------------------------------------------------------------%
 
 :- implementation.
 
-:- import_module hlds_data, code_gen, trace, tree.
+:- import_module builtin_ops, code_gen, trace, tree.
 :- import_module bool, int, string, list, map, std_util, assoc_list, require.
 
 string_switch__generate(Cases, Var, CodeModel, _CanFail, StoreMap,
-		EndLabel, Code) -->
+		EndLabel, MaybeEnd0, MaybeEnd, Code) -->
 	code_info__produce_variable(Var, VarCode, VarRval),
 	code_info__acquire_reg(r, SlotReg),
 	code_info__acquire_reg(r, StringReg),
@@ -67,7 +67,7 @@ string_switch__generate(Cases, Var, CodeModel, _CanFail, StoreMap,
 		% before the code for the cases (which might reuse those
 		% registers), and because that code is generated manually
 		% (below) so we don't need the reg info to be valid when
-		% we generated it.
+		% we generate it.
 	code_info__release_reg(SlotReg),
 	code_info__release_reg(StringReg),
 
@@ -80,13 +80,17 @@ string_switch__generate(Cases, Var, CodeModel, _CanFail, StoreMap,
 		% Generate the code etc. for the hash table
 		%
 	string_switch__gen_hash_slots(0, TableSize, HashSlotsMap, CodeModel,
-		StoreMap, FailLabel, EndLabel, Strings, Labels,
-		NextSlots, SlotsCode),
+		StoreMap, FailLabel, EndLabel, MaybeEnd0, MaybeEnd,
+		Strings, Labels, NextSlots, SlotsCode),
 
 		% Generate code which does the hash table lookup
 	{
-		NextSlotsTable = create(0, NextSlots, no, NextSlotsTableNo),
-		StringTable = create(0, Strings, no, StringTableNo),
+		NextSlotsTable = create(0, NextSlots, uniform(no),
+			must_be_static, NextSlotsTableNo,
+			"string_switch_next_slots_table"),
+		StringTable = create(0, Strings, uniform(no),
+			must_be_static, StringTableNo,
+			"string_switch_string_table"),
 		HashLookupCode = node([
 			comment("hashed string switch") -
 			  "",
@@ -243,16 +247,18 @@ string_switch__next_free_hash_slot(Map, H_Map, LastUsed, FreeSlot) :-
 	).
 
 :- pred string_switch__gen_hash_slots(int, int, map(int, hash_slot),
-	code_model, store_map, label, label, list(maybe(rval)), list(label),
-	list(maybe(rval)), code_tree, code_info, code_info).
+	code_model, store_map, label, label, branch_end, branch_end,
+	list(maybe(rval)), list(label), list(maybe(rval)), code_tree,
+	code_info, code_info).
 :- mode string_switch__gen_hash_slots(in, in, in, in, in, in, in,
-	out, out, out, out, in, out) is det.
+	in, out, out, out, out, out, in, out) is det.
 
 string_switch__gen_hash_slots(Slot, TableSize, HashSlotMap, CodeModel,
-		StoreMap, FailLabel, EndLabel, Strings, Labels,
-		NextSlots, Code) -->
+		StoreMap, FailLabel, EndLabel, MaybeEnd0, MaybeEnd,
+		Strings, Labels, NextSlots, Code) -->
 	( { Slot = TableSize } ->
 		{
+			MaybeEnd = MaybeEnd0,
 			Strings = [],
 			Labels = [],
 			NextSlots = [],
@@ -262,8 +268,9 @@ string_switch__gen_hash_slots(Slot, TableSize, HashSlotMap, CodeModel,
 		}
 	;
 		string_switch__gen_hash_slot(Slot, TableSize, HashSlotMap,
-				CodeModel, StoreMap, FailLabel, EndLabel,
-				String, Label, NextSlot, SlotCode),
+			CodeModel, StoreMap, FailLabel, EndLabel,
+			MaybeEnd0, MaybeEnd1,
+			String, Label, NextSlot, SlotCode),
 		{ Slot1 is Slot + 1 },
 		{ 
 			Strings = [String | Strings0],
@@ -272,19 +279,21 @@ string_switch__gen_hash_slots(Slot, TableSize, HashSlotMap, CodeModel,
 			Code = tree(SlotCode, Code0)
 		},
 		string_switch__gen_hash_slots(Slot1, TableSize, HashSlotMap,
-				CodeModel, StoreMap, FailLabel, EndLabel,
-				Strings0, Labels0, NextSlots0, Code0)
+			CodeModel, StoreMap, FailLabel, EndLabel,
+			MaybeEnd1, MaybeEnd,
+			Strings0, Labels0, NextSlots0, Code0)
 	).
 
 :- pred string_switch__gen_hash_slot(int, int, map(int, hash_slot),
-	code_model, store_map, label, label, maybe(rval), label,
-	maybe(rval), code_tree, code_info, code_info).
+	code_model, store_map, label, label, branch_end, branch_end,
+	maybe(rval), label, maybe(rval), code_tree,
+	code_info, code_info).
 :- mode string_switch__gen_hash_slot(in, in, in, in, in, in, in,
-	out, out, out, out, in, out) is det.
+	in, out, out, out, out, out, in, out) is det.
 
 string_switch__gen_hash_slot(Slot, TblSize, HashSlotMap, CodeModel, StoreMap,
-		FailLabel, EndLabel, yes(StringRval), Label,
-		yes(NextSlotRval), Code) -->
+		FailLabel, EndLabel, MaybeEnd0, MaybeEnd,
+		yes(StringRval), Label, yes(NextSlotRval), Code) -->
 	(
 		{ map__search(HashSlotMap, Slot, hash_slot(Case, Next)) }
 	->
@@ -301,28 +310,18 @@ string_switch__gen_hash_slot(Slot, TblSize, HashSlotMap, CodeModel, StoreMap,
 		{ LabelCode = node([
 			label(Label) - Comment
 		]) },
-		code_info__get_maybe_trace_info(MaybeTraceInfo),
-		( { MaybeTraceInfo = yes(TraceInfo) } ->
-			{ Goal = _ - GoalInfo },
-			{ goal_info_get_goal_path(GoalInfo, Path) },
-			trace__generate_event_code(switch(Path), TraceInfo,
-				TraceCode)
-		;
-			{ TraceCode = empty }
-		),
+		code_info__remember_position(BranchStart),
+		trace__maybe_generate_internal_event_code(Goal, TraceCode),
+		code_gen__generate_goal(CodeModel, Goal, GoalCode),
+		code_info__generate_branch_end(StoreMap, MaybeEnd0, MaybeEnd,
+			SaveCode),
 		(
 			{ string_switch__this_is_last_case(Slot, TblSize,
 				HashSlotMap) }
 		->
-			code_gen__generate_goal(CodeModel, Goal, GoalCode),
-			code_info__generate_branch_end(CodeModel, StoreMap,
-				SaveCode)
+			[]
 		;
-			code_info__grab_code_info(CodeInfo),
-			code_gen__generate_goal(CodeModel, Goal, GoalCode),
-			code_info__generate_branch_end(CodeModel, StoreMap,
-				SaveCode),
-			code_info__slap_code_info(CodeInfo)
+			code_info__reset_to_position(BranchStart)
 		),
 		{ FinishCode = node([
 			goto(label(EndLabel)) - "jump to end of switch"
@@ -335,6 +334,7 @@ string_switch__gen_hash_slot(Slot, TblSize, HashSlotMap, CodeModel, StoreMap,
 			     FinishCode))))
 		}
 	;
+		{ MaybeEnd = MaybeEnd0 },
 		{ StringRval = const(int_const(0)) },
 		{ Label = FailLabel },
 		{ NextSlotRval = const(int_const(-2)) },
