@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -16,7 +16,8 @@
 
 :- interface.
 
-:- import_module hlds_module, hlds_pred.
+:- import_module hlds_goal, hlds_module, hlds_pred, prog_data.
+:- import_module io, list.
 
 :- pred detect_switches(module_info, module_info, io__state, io__state).
 :- mode detect_switches(in, out, di, uo) is det.
@@ -24,6 +25,25 @@
 :- pred detect_switches_in_proc(proc_id, pred_id, module_info, module_info).
 :- mode detect_switches_in_proc(in, in, in, out) is det.
 
+	% find_bind_var(Var, ProcessUnify, Goal0, Goals, Subst0, Subst,
+	%		Result0, Result, Continue):
+	% 	Used by both switch_detection and cse_detection.
+	%	Searches through `Goal0' looking for the first deconstruction
+	%	unification with `Var' or an alias of `Var'.
+	%	`ProcessUnify' is called if a deconstruction unification with
+	%	`Var' is found, returning either the cons_id for
+	%	switch_detection or the common deconstruction for cse_detection.
+	%	If a deconstruction unification of the variable is found,
+	%	`ProcessUnify' is called to handle it and searching is stopped.
+	%	If not, `Result' is set to `Result0'.
+:- pred find_bind_var(prog_var, process_unify(Result, Info), hlds_goal,
+		hlds_goal, Result, Result, Info, Info).
+:- mode find_bind_var(in, in(process_unify), in, out, in, out, in, out) is det.
+
+:- type process_unify(Result, Info) ==
+	pred(prog_var, hlds_goal, list(hlds_goal), Result, Result, Info, Info).
+:- inst process_unify = (pred(in, in, out, in, out, in, out) is det).
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
@@ -31,8 +51,8 @@
 
 :- import_module hlds_goal, hlds_data, prog_data, instmap, inst_match.
 :- import_module modes, mode_util, type_util, det_util.
-:- import_module passes_aux.
-:- import_module char, int, list, assoc_list, map, set, std_util, term, require.
+:- import_module passes_aux, term.
+:- import_module bool, char, int, assoc_list, map, set, std_util, require.
 
 %-----------------------------------------------------------------------------%
 
@@ -102,7 +122,7 @@ detect_switches_in_proc(ProcId, PredId, ModuleInfo0, ModuleInfo) :-
 	% Given a goal, and the instmap on entry to that goal,
 	% replace disjunctions with switches whereever possible.
 
-:- pred detect_switches_in_goal(hlds_goal, instmap, map(var, type),
+:- pred detect_switches_in_goal(hlds_goal, instmap, map(prog_var, type),
 	module_info, hlds_goal).
 :- mode detect_switches_in_goal(in, in, in, in, out) is det.
 
@@ -115,7 +135,7 @@ detect_switches_in_goal(Goal0, InstMap0, VarTypes, ModuleInfo, Goal) :-
 	% computed by applying the instmap delta specified in the
 	% goal's goalinfo.
 
-:- pred detect_switches_in_goal_1(hlds_goal, instmap, map(var, type),
+:- pred detect_switches_in_goal_1(hlds_goal, instmap, map(prog_var, type),
 	module_info, hlds_goal, instmap).
 :- mode detect_switches_in_goal_1(in, in, in, in, out, out) is det.
 
@@ -128,7 +148,7 @@ detect_switches_in_goal_1(Goal0 - GoalInfo, InstMap0, VarTypes, ModuleInfo,
 	% Here we process each of the different sorts of goals.
 
 :- pred detect_switches_in_goal_2(hlds_goal_expr, hlds_goal_info, instmap,
-		map(var, type), module_info, hlds_goal_expr).
+		map(prog_var, type), module_info, hlds_goal_expr).
 :- mode detect_switches_in_goal_2(in, in, in, in, in, out) is det.
 
 detect_switches_in_goal_2(disj(Goals0, SM), GoalInfo, InstMap0,
@@ -147,6 +167,11 @@ detect_switches_in_goal_2(conj(Goals0), _GoalInfo, InstMap0,
 		VarTypes, ModuleInfo, conj(Goals)) :-
 	detect_switches_in_conj(Goals0, InstMap0, VarTypes, ModuleInfo, Goals).
 
+detect_switches_in_goal_2(par_conj(Goals0, SM), _GoalInfo, InstMap0,
+		VarTypes, ModuleInfo, par_conj(Goals, SM)) :-
+	detect_switches_in_par_conj(Goals0, InstMap0, VarTypes,
+		ModuleInfo, Goals).
+
 detect_switches_in_goal_2(not(Goal0), _GoalInfo, InstMap0,
 		VarTypes, ModuleInfo, not(Goal)) :-
 	detect_switches_in_goal(Goal0, InstMap0, VarTypes, ModuleInfo, Goal).
@@ -159,26 +184,30 @@ detect_switches_in_goal_2(if_then_else(Vars, Cond0, Then0, Else0, SM),
 	detect_switches_in_goal(Then0, InstMap1, VarTypes, ModuleInfo, Then),
 	detect_switches_in_goal(Else0, InstMap0, VarTypes, ModuleInfo, Else).
 
-detect_switches_in_goal_2(some(Vars, Goal0), _GoalInfo, InstMap0,
-		VarTypes, ModuleInfo, some(Vars, Goal)) :-
+detect_switches_in_goal_2(some(Vars, CanRemove, Goal0), _GoalInfo, InstMap0,
+		VarTypes, ModuleInfo, some(Vars, CanRemove, Goal)) :-
 	detect_switches_in_goal(Goal0, InstMap0, VarTypes, ModuleInfo, Goal).
 
-detect_switches_in_goal_2(higher_order_call(A,B,C,D,E,F), _, _, _, _,
-		higher_order_call(A,B,C,D,E,F)).
+detect_switches_in_goal_2(generic_call(A,B,C,D), _, _, _, _,
+		generic_call(A,B,C,D)).
 
 detect_switches_in_goal_2(call(A,B,C,D,E,F), _, _, _, _,
 		call(A,B,C,D,E,F)).
 
 detect_switches_in_goal_2(unify(A,RHS0,C,D,E), __GoalInfo, InstMap0,
 		VarTypes, ModuleInfo, unify(A,RHS,C,D,E)) :-
-	( RHS0 = lambda_goal(PredOrFunc, Vars, Modes, Det, Goal0) ->
+	(
+		RHS0 = lambda_goal(PredOrFunc, EvalMethod, FixModes,
+			NonLocals, Vars, Modes, Det, Goal0)
+	->
 		% we need to insert the initial insts for the lambda
 		% variables in the instmap before processing the lambda goal
 		instmap__pre_lambda_update(ModuleInfo, 
 			Vars, Modes, InstMap0, InstMap1),
 		detect_switches_in_goal(Goal0, InstMap1, VarTypes, ModuleInfo,
 			Goal),
-		RHS = lambda_goal(PredOrFunc, Vars, Modes, Det, Goal)
+		RHS = lambda_goal(PredOrFunc, EvalMethod, FixModes,
+			NonLocals, Vars, Modes, Det, Goal)
 	;
 		RHS = RHS0
 	).
@@ -187,8 +216,11 @@ detect_switches_in_goal_2(switch(Var, CanFail, Cases0, SM), _, InstMap,
 		VarTypes, ModuleInfo, switch(Var, CanFail, Cases, SM)) :-
 	detect_switches_in_cases(Cases0, InstMap, VarTypes, ModuleInfo, Cases).
 
-detect_switches_in_goal_2(pragma_c_code(A,B,C,D,E,F,G,H), _, _, _, _,
-		pragma_c_code(A,B,C,D,E,F,G,H)).
+detect_switches_in_goal_2(pragma_c_code(A,B,C,D,E,F,G), _, _, _, _,
+		pragma_c_code(A,B,C,D,E,F,G)).
+detect_switches_in_goal_2(bi_implication(_, _), _, _, _, _, _) :-
+	% these should have been expanded out by now
+	error("detect_switches_in_goal_2: unexpected bi_implication").
 
 %-----------------------------------------------------------------------------%
 
@@ -205,10 +237,10 @@ detect_switches_in_goal_2(pragma_c_code(A,B,C,D,E,F,G,H), _, _, _, _,
 	% the sorted_case_list should always be sorted on cons_id -
 	% `delete_unreachable_cases' relies on this.
 
-:- type again ---> again(var, list(hlds_goal), sorted_case_list).
+:- type again ---> again(prog_var, list(hlds_goal), sorted_case_list).
 
-:- pred detect_switches_in_disj(list(var), list(hlds_goal), hlds_goal_info,
-	store_map, instmap, map(var, type), list(var), module_info,
+:- pred detect_switches_in_disj(list(prog_var), list(hlds_goal), hlds_goal_info,
+	store_map, instmap, map(prog_var, type), list(prog_var), module_info,
 	list(again), hlds_goal_expr).
 :- mode detect_switches_in_disj(in, in, in, in, in, in, in, in, in, out) is det.
 
@@ -299,8 +331,8 @@ select_best_switch([Again | AgainList], BestAgain0, BestAgain) :-
 	),
 	select_best_switch(AgainList, BestAgain1, BestAgain).
 
-:- pred detect_sub_switches_in_disj(list(hlds_goal), instmap, map(var, type),
-	module_info, list(hlds_goal)).
+:- pred detect_sub_switches_in_disj(list(hlds_goal), instmap,
+		map(prog_var, type), module_info, list(hlds_goal)).
 :- mode detect_sub_switches_in_disj(in, in, in, in, out) is det.
 
 detect_sub_switches_in_disj([], _InstMap, _VarTypes, _ModuleInfo, []).
@@ -310,8 +342,8 @@ detect_sub_switches_in_disj([Goal0 | Goals0], InstMap, VarTypes, ModuleInfo,
 	detect_sub_switches_in_disj(Goals0, InstMap, VarTypes, ModuleInfo,
 		Goals).
 
-:- pred detect_switches_in_cases(list(case), instmap, map(var, type),
-	module_info, list(case)).
+:- pred detect_switches_in_cases(list(case), instmap, map(prog_var, type),
+		module_info, list(case)).
 :- mode detect_switches_in_cases(in, in, in, in, out) is det.
 
 detect_switches_in_cases([], _InstMap, _VarTypes, _ModuleInfo, []).
@@ -322,7 +354,18 @@ detect_switches_in_cases([Case0 | Cases0], InstMap, VarTypes, ModuleInfo,
 	Case = case(Functor, Goal),
 	detect_switches_in_cases(Cases0, InstMap, VarTypes, ModuleInfo, Cases).
 
-:- pred detect_switches_in_conj(list(hlds_goal), instmap, map(var, type),
+:- pred detect_switches_in_par_conj(list(hlds_goal), instmap,
+		map(prog_var, type), module_info, list(hlds_goal)).
+:- mode detect_switches_in_par_conj(in, in, in, in, out) is det.
+
+detect_switches_in_par_conj([], _InstMap, _VarTypes, _ModuleInfo, []).
+detect_switches_in_par_conj([Goal0 | Goals0], InstMap, VarTypes, ModuleInfo,
+		[Goal | Goals]) :-
+	detect_switches_in_goal(Goal0, InstMap, VarTypes, ModuleInfo, Goal),
+	detect_switches_in_par_conj(Goals0, InstMap, VarTypes,
+		ModuleInfo, Goals).
+
+:- pred detect_switches_in_conj(list(hlds_goal), instmap, map(prog_var, type),
 	module_info, list(hlds_goal)).
 :- mode detect_switches_in_conj(in, in, in, in, out) is det.
 
@@ -348,8 +391,8 @@ detect_switches_in_conj([Goal0 | Goals0], InstMap0, VarTypes, ModuleInfo,
 	% We partition the goals by abstractly interpreting the unifications
 	% at the start of each disjunction, to build up a substitution.
 
-:- pred partition_disj(list(hlds_goal), var, hlds_goal_info, list(hlds_goal),
-	sorted_case_list).
+:- pred partition_disj(list(hlds_goal), prog_var, hlds_goal_info,
+		list(hlds_goal), sorted_case_list).
 :- mode partition_disj(in, in, in, out, out) is semidet.
 
 partition_disj(Goals0, Var, GoalInfo, Left, CasesList) :-
@@ -359,21 +402,17 @@ partition_disj(Goals0, Var, GoalInfo, Left, CasesList) :-
 	CasesAssocList \= [], % there must be at least one case
 	fix_case_list(CasesAssocList, GoalInfo, CasesList).
 
-:- pred partition_disj_trial(list(hlds_goal), var,
+:- pred partition_disj_trial(list(hlds_goal), prog_var,
 	list(hlds_goal), list(hlds_goal), cases, cases).
 :- mode partition_disj_trial(in, in, in, out, in, out) is det.
 
 partition_disj_trial([], _Var, Left, Left, Cases, Cases).
 partition_disj_trial([Goal0 | Goals], Var, Left0, Left, Cases0, Cases) :-
-	goal_to_conj_list(Goal0, ConjList0),
-	Goal0 = _ - GoalInfo,
-	map__init(Substitution),
-	find_bind_var_for_switch(ConjList0, Substitution, Var,
-			ConjList, _NewSubstitution, MaybeFunctor),
+	find_bind_var(Var, find_bind_var_for_switch_in_deconstruct,
+		Goal0, Goal, no, MaybeFunctor, unit, _),
 	(
 		MaybeFunctor = yes(Functor),
 		Left1 = Left0,
-		conj_list_to_goal(ConjList, GoalInfo, Goal),
 		( map__search(Cases0, Functor, DisjList0) ->
 			DisjList1 = [Goal | DisjList0],
 			map__det_update(Cases0, Functor, DisjList1, Cases1)
@@ -388,78 +427,124 @@ partition_disj_trial([Goal0 | Goals], Var, Left0, Left, Cases0, Cases) :-
 	),
 	partition_disj_trial(Goals, Var, Left1, Left, Cases1, Cases).
 
-	% find_bind_var_for_switch(Goals0, Subst0, Var, Goals, Subst,
-	%	MaybeFunctor):
-	%	Searches through Goals0 looking for a deconstruction
-	%	unification with `Var'. If found, sets `MaybeFunctor'
-	% 	to `yes(Functor)', where Functor is the
-	%	functor which `Var' gets unified, and
-	%	sets `Goals' to be `Goals0' with that deconstruction
-	%	unification made deterministic. If not found, sets
-	%	`MaybeFunctor' to `no', and computes `Subst' as the
-	%	resulting substitution from interpreting through the goal.
+:- pred find_bind_var_for_switch_in_deconstruct(prog_var, hlds_goal,
+		list(hlds_goal), maybe(cons_id), maybe(cons_id), unit, unit).
+:- mode find_bind_var_for_switch_in_deconstruct(in, in, out,
+		in, out, in, out) is det.
 
-:- pred find_bind_var_for_switch(list(hlds_goal), substitution, var,
-	list(hlds_goal), substitution, maybe(cons_id)).
-:- mode find_bind_var_for_switch(in, in, in, out, out, out) is det.
+find_bind_var_for_switch_in_deconstruct(_UnifyVar, Goal0, Goals, 
+		_Result0, Result, _, unit) :-
+	(
+		Goal0 = unify(A, B, C, UnifyInfo0, E) - GoalInfo,
+		UnifyInfo0 = deconstruct(A, Functor, F, G, _)
+	->
+		Result = yes(Functor),
+			% The deconstruction unification now becomes
+			% deterministic, since the test will get
+			% carried out in the switch.
+		UnifyInfo = deconstruct(A, Functor, F, G,
+			cannot_fail),
+		Goals = [unify(A, B, C, UnifyInfo, E) - GoalInfo]
+	;
+		error("find_bind_var_for_switch_in_deconstruct")
+	).
 
-find_bind_var_for_switch([], Substitution, _Var, [], Substitution, no).
-find_bind_var_for_switch([Goal0 - GoalInfo | Goals0], Substitution0, Var,
-		[Goal - GoalInfo | Goals], Substitution, MaybeFunctor) :-
-	( Goal0 = conj(SubGoals0) ->
-		find_bind_var_for_switch(SubGoals0, Substitution0, Var,
-			SubGoals, Substitution1, MaybeFunctor1),
-		Goal = conj(SubGoals),
-		( MaybeFunctor1 = yes(_) ->
-			Goals = Goals0,
-			Substitution = Substitution1,
-			MaybeFunctor = MaybeFunctor1
-		;
-			find_bind_var_for_switch(Goals0, Substitution1, Var,
-				Goals, Substitution, MaybeFunctor)
-		)
-	; Goal0 = unify(A, B, C, UnifyInfo0, E) ->
+%-----------------------------------------------------------------------------%
+
+find_bind_var(Var, ProcessUnify, Goal0, Goal,
+		Result0, Result, Info0, Info) :-
+	map__init(Substitution),
+	find_bind_var(Var, ProcessUnify, Goal0, Goal, Substitution,
+		_, Result0, Result, Info0, Info, _).
+
+:- pred find_bind_var(prog_var, process_unify(Result, Info), hlds_goal,
+		hlds_goal, prog_substitution, prog_substitution, Result, Result,
+		Info, Info, bool).
+:- mode find_bind_var(in, in(process_unify), in, out, in,
+	out, in, out, in, out, out) is det.
+
+find_bind_var(Var, ProcessUnify, Goal0 - GoalInfo, Goal, Substitution0,
+		Substitution, Result0, Result, Info0, Info, Continue) :-
+	( Goal0 = some(Vars, CanRemove, SubGoal0) ->
+		find_bind_var(Var, ProcessUnify, SubGoal0, SubGoal,
+			Substitution0, Substitution, Result0, Result,
+			Info0, Info, Continue),
+		Goal = some(Vars, CanRemove, SubGoal) - GoalInfo
+	; Goal0 = conj(SubGoals0) ->
+		conj_find_bind_var(Var, ProcessUnify, SubGoals0, SubGoals,
+			Substitution0, Substitution, Result0, Result,
+			Info0, Info, Continue),
+		Goal = conj(SubGoals) - GoalInfo
+	; Goal0 = unify(A, B, _, UnifyInfo0, _) ->
+		(
 			% check whether the unification is a deconstruction
 			% unification on Var or a variable aliased to Var
-		(
-			UnifyInfo0 = deconstruct(UnifyVar, Functor, F, G, _),
-			term__apply_rec_substitution(term__variable(Var),
+			UnifyInfo0 = deconstruct(UnifyVar, _, _, _, _),
+			term__apply_rec_substitution(
+				term__variable(Var),
 				Substitution0, term__variable(Var1)),
-			term__apply_rec_substitution(term__variable(UnifyVar),
+			term__apply_rec_substitution(
+				term__variable(UnifyVar),
 				Substitution0, term__variable(UnifyVar1)),
 			Var1 = UnifyVar1
 		->
-			MaybeFunctor = yes(Functor),
-				% The deconstruction unification now becomes
-				% deterministic, since the test will get
-				% carried out in the switch.
-			UnifyInfo = deconstruct(UnifyVar, Functor, F, G,
-				cannot_fail),
-			Goal = unify(A, B, C, UnifyInfo, E),
-			Goals = Goals0,
+			call(ProcessUnify, Var, Goal0 - GoalInfo, Goals,
+				Result0, Result, Info0, Info),
+			conj_list_to_goal(Goals, GoalInfo, Goal),
+			Continue = no,
 			Substitution = Substitution0
 		;
+			Goal = Goal0 - GoalInfo,
+			Continue = yes,
 			% otherwise abstractly interpret the unification
-			% and continue
-			Goal = Goal0,
+			Result = Result0,
+			Info = Info0,
 			( interpret_unify(A, B, Substitution0, Substitution1) ->
-				Substitution2 = Substitution1
+				Substitution = Substitution1
 			;
 				% the unification must fail - just ignore it
-				Substitution2 = Substitution0
-			),
-			find_bind_var_for_switch(Goals0, Substitution2, Var,
-				Goals, Substitution, MaybeFunctor)
+				Substitution = Substitution0
+			)
 		)
 	;
-		Goal = Goal0,
-		Goals = Goals0,
+		Goal = Goal0 - GoalInfo,
 		Substitution = Substitution0,
-		MaybeFunctor = no
+		Result = Result0,
+		Info = Info0,
+		Continue = no
 	).
 
-:- pred cases_to_switch(sorted_case_list, var, map(var, type), hlds_goal_info,
-	store_map, instmap, module_info, hlds_goal_expr).
+:- pred conj_find_bind_var(prog_var, process_unify(Result, Info), 
+		list(hlds_goal), list(hlds_goal), prog_substitution,
+		prog_substitution, Result, Result, Info, Info, bool).
+:- mode conj_find_bind_var(in, in(process_unify), in, out,
+	in, out, in, out, in, out, out) is det.
+
+conj_find_bind_var(_Var, _, [], [], Substitution, Substitution,
+		Result, Result, Info, Info, yes).
+conj_find_bind_var(Var, ProcessUnify, [Goal0 | Goals0], [Goal | Goals],
+		Substitution0, Substitution, Result0, Result,
+		Info0, Info, Continue) :-
+	find_bind_var(Var, ProcessUnify, Goal0, Goal, Substitution0,
+		Substitution1, Result0, Result1,
+		Info0, Info1, Continue1),
+	( Continue1 = no ->
+		Continue = no,
+		Goals = Goals0,
+		Substitution = Substitution1,
+		Result = Result1,
+		Info = Info1
+	;
+		conj_find_bind_var(Var, ProcessUnify, Goals0, Goals,
+			Substitution1, Substitution, Result1, Result,
+			Info1, Info, Continue)
+	).
+
+%-----------------------------------------------------------------------------%
+
+:- pred cases_to_switch(sorted_case_list, prog_var, map(prog_var, type),
+		hlds_goal_info, store_map, instmap, module_info,
+		hlds_goal_expr).
 :- mode cases_to_switch(in, in, in, in, in, in, in, out) is det.
 
 cases_to_switch(CasesList, Var, VarTypes, _GoalInfo, SM, InstMap, ModuleInfo,
@@ -534,8 +619,10 @@ switch_covers_all_cases(CasesList, Type, ModuleInfo) :-
 :- mode fix_case_list(in, in, out) is det.
 
 fix_case_list([], _, []).
-fix_case_list([Functor - DisjList | Cases0], GoalInfo,
+fix_case_list([Functor - DisjList0 | Cases0], GoalInfo,
 		[case(Functor, Goal) | Cases]) :-
+		% We need to put the list back the right way around.
+	list__reverse(DisjList0, DisjList),
 	disj_list_to_goal(DisjList, GoalInfo, Goal),
 	fix_case_list(Cases0, GoalInfo, Cases).
 
