@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -26,9 +26,12 @@
 %	a matter of doing an array index for each output variable - avoiding
 %	the branch overhead of the jump-table.
 %
-%	For switches on discriminated union types, we generate a chain of
-%	if-then-elses on the primary tags but then use the secondary tag
-%	to index into a jump table if the table is big enough.
+%	For switches on discriminated union types, we generate code that does
+%	indexing first on the primary tag, and then on the secondary tag (if
+%	the primary tag is shared between several function symbols). The
+%	indexing code for switches on both primary and secondary tags can be
+%	in the form of a try-me-else chain, a try chain, a dense jump table
+%	or a binary search.
 %
 %	For switches on strings, we lookup the address to jump to in a
 %	hash table, using open addressing to resolve hash collisions.
@@ -42,10 +45,10 @@
 
 :- interface.
 
-:- import_module hlds_goal, hlds_data, code_info.
-:- import_module term.
+:- import_module prog_data, hlds_goal, hlds_data, code_info, llds.
+:- import_module list.
 
-:- pred switch_gen__generate_switch(code_model, var, can_fail, list(case),
+:- pred switch_gen__generate_switch(code_model, prog_var, can_fail, list(case),
 	store_map, hlds_goal_info, code_tree, code_info, code_info).
 :- mode switch_gen__generate_switch(in, in, in, in, in, in, out, in, out)
 	is det.
@@ -61,9 +64,9 @@
 :- implementation.
 
 :- import_module dense_switch, string_switch, tag_switch, lookup_switch.
-:- import_module llds, code_gen, unify_gen, code_aux, type_util, code_util.
+:- import_module code_gen, unify_gen, code_aux, type_util, code_util.
 :- import_module trace, globals, options.
-:- import_module bool, int, string, list, map, tree, std_util, require.
+:- import_module bool, int, string, map, tree, std_util, require.
 
 :- type switch_category
 	--->	atomic_switch
@@ -102,8 +105,8 @@ switch_gen__generate_switch(CodeModel, CaseVar, CanFail, Cases, StoreMap,
 			OutVars, CaseVals, MLiveness)
 	->
 		lookup_switch__generate(CaseVar, OutVars, CaseVals,
-			FirstVal, LastVal, NeedRangeCheck,
-			NeedBitVecCheck, MLiveness, StoreMap, Code)
+			FirstVal, LastVal, NeedRangeCheck, NeedBitVecCheck,
+			MLiveness, StoreMap, no, MaybeEnd, Code)
 	;
 		{ Indexing = yes },
 		{ SwitchCategory = atomic_switch },
@@ -118,7 +121,7 @@ switch_gen__generate_switch(CodeModel, CaseVar, CanFail, Cases, StoreMap,
 	->
 		dense_switch__generate(TaggedCases,
 			FirstVal, LastVal, CaseVar, CodeModel, CanFail1,
-			StoreMap, EndLabel, Code)
+			StoreMap, EndLabel, no, MaybeEnd, Code)
 	;
 		{ Indexing = yes },
 		{ SwitchCategory = string_switch },
@@ -128,7 +131,7 @@ switch_gen__generate_switch(CodeModel, CaseVar, CanFail, Cases, StoreMap,
 		{ NumCases >= StringSize }
 	->
 		string_switch__generate(TaggedCases, CaseVar, CodeModel,
-			CanFail, StoreMap, EndLabel, Code)
+			CanFail, StoreMap, EndLabel, no, MaybeEnd, Code)
 	;
 		{ Indexing = yes },
 		{ SwitchCategory = tag_switch },
@@ -138,16 +141,17 @@ switch_gen__generate_switch(CodeModel, CaseVar, CanFail, Cases, StoreMap,
 		{ NumCases >= TagSize }
 	->
 		tag_switch__generate(TaggedCases, CaseVar, CodeModel, CanFail,
-			StoreMap, EndLabel, Code)
+			StoreMap, EndLabel, no, MaybeEnd, Code)
 	;
 		% To generate a switch, first we flush the
 		% variable on whose tag we are going to switch, then we
 		% generate the cases for the switch.
 
 		switch_gen__generate_all_cases(TaggedCases, CaseVar,
-			CodeModel, CanFail, StoreMap, EndLabel, Code)
+			CodeModel, CanFail, StoreMap, EndLabel, no, MaybeEnd,
+			Code)
 	),
-	code_info__remake_with_store_map(StoreMap).
+	code_info__after_all_branches(StoreMap, MaybeEnd).
 
 %---------------------------------------------------------------------------%
 
@@ -155,7 +159,7 @@ switch_gen__generate_switch(CodeModel, CaseVar, CanFail, Cases, StoreMap,
 	% being switched on is an atomic type, a string, or
 	% something more complicated.
 
-:- pred switch_gen__determine_category(var, switch_category,
+:- pred switch_gen__determine_category(prog_var, switch_category,
 	code_info, code_info).
 :- mode switch_gen__determine_category(in, out, in, out) is det.
 
@@ -179,7 +183,7 @@ switch_gen__type_cat_to_switch_cat(polymorphic_type, other_switch).
 
 %---------------------------------------------------------------------------%
 
-:- pred switch_gen__lookup_tags(list(case), var, cases_list,
+:- pred switch_gen__lookup_tags(list(case), prog_var, cases_list,
 				code_info, code_info).
 :- mode switch_gen__lookup_tags(in, in, out, in, out) is det.
 
@@ -200,14 +204,16 @@ switch_gen__lookup_tags([Case | Cases], Var, [TaggedCase | TaggedCases]) -->
 
 switch_gen__priority(no_tag, 0).			% should never occur
 switch_gen__priority(int_constant(_), 1).
-switch_gen__priority(complicated_constant_tag(_, _), 1).
-switch_gen__priority(simple_tag(_), 2).
+switch_gen__priority(shared_local_tag(_, _), 1).
+switch_gen__priority(unshared_tag(_), 2).
 switch_gen__priority(float_constant(_), 3).
-switch_gen__priority(complicated_tag(_, _), 4).
+switch_gen__priority(shared_remote_tag(_, _), 4).
 switch_gen__priority(string_constant(_), 5).
-switch_gen__priority(pred_closure_tag(_, _), 6).	% should never occur
+switch_gen__priority(pred_closure_tag(_, _, _), 6).	% should never occur
 switch_gen__priority(code_addr_constant(_, _), 6).	% should never occur
-switch_gen__priority(base_type_info_constant(_, _, _), 6).% should never occur
+switch_gen__priority(type_ctor_info_constant(_, _, _), 6).% should never occur
+switch_gen__priority(base_typeclass_info_constant(_, _, _), 6).% shouldn't occur
+switch_gen__priority(tabling_pointer_constant(_, _), 6). % shouldn't occur
 
 %---------------------------------------------------------------------------%
 %---------------------------------------------------------------------------%
@@ -238,13 +244,14 @@ switch_gen__priority(base_type_info_constant(_, _, _), 6).% should never occur
 	% and put that one first. This minimizes the number of pipeline
 	% breaks caused by taken branches.
 
-:- pred switch_gen__generate_all_cases(list(extended_case), var, code_model,
-	can_fail, store_map, label, code_tree, code_info, code_info).
-:- mode switch_gen__generate_all_cases(in, in, in, in, in, in, out, in, out)
-	is det.
+:- pred switch_gen__generate_all_cases(list(extended_case), prog_var,
+	code_model, can_fail, store_map, label, branch_end, branch_end,
+	code_tree, code_info, code_info).
+:- mode switch_gen__generate_all_cases(in, in, in, in, in, in, in, out, out,
+	in, out) is det.
 
 switch_gen__generate_all_cases(Cases0, Var, CodeModel, CanFail, StoreMap,
-		EndLabel, Code) -->
+		EndLabel, MaybeEnd0, MaybeEnd, Code) -->
 	code_info__produce_variable(Var, VarCode, _Rval),
 	(
 		{ CodeModel = model_det },
@@ -271,25 +278,26 @@ switch_gen__generate_all_cases(Cases0, Var, CodeModel, CanFail, StoreMap,
 			Cases = [Case2, Case1]
 		;
 			Cases = Cases0
-		},
-		switch_gen__generate_cases(Cases, Var, CodeModel, CanFail,
-			StoreMap, EndLabel, CasesCode)
+		}
 	;
-		switch_gen__generate_cases(Cases0, Var, CodeModel, CanFail,
-			StoreMap, EndLabel, CasesCode)
+		{ Cases = Cases0 }
 	),
+	switch_gen__generate_cases(Cases, Var, CodeModel, CanFail,
+		StoreMap, EndLabel, MaybeEnd0, MaybeEnd, CasesCode),
 	{ Code = tree(VarCode, CasesCode) }.
 
-:- pred switch_gen__generate_cases(list(extended_case), var, code_model,
-	can_fail, store_map, label, code_tree, code_info, code_info).
-:- mode switch_gen__generate_cases(in, in, in, in, in, in, out, in, out) is det.
+:- pred switch_gen__generate_cases(list(extended_case), prog_var, code_model,
+	can_fail, store_map, label, branch_end, branch_end, code_tree,
+	code_info, code_info).
+:- mode switch_gen__generate_cases(in, in, in, in, in, in, in, out, out,
+	in, out) is det.
 
 	% At the end of a locally semidet switch, we fail because we
 	% came across a tag which was not covered by one of the cases.
 	% It is followed by the end of switch label to which the cases
 	% branch.
 switch_gen__generate_cases([], _Var, _CodeModel, CanFail, _StoreMap,
-		EndLabel, Code) -->
+		EndLabel, MaybeEnd, MaybeEnd, Code) -->
 	( { CanFail = can_fail } ->
 		code_info__generate_failure(FailCode)
 	;
@@ -302,24 +310,17 @@ switch_gen__generate_cases([], _Var, _CodeModel, CanFail, _StoreMap,
 	{ Code = tree(FailCode, EndCode) }.
 
 switch_gen__generate_cases([case(_, _, Cons, Goal) | Cases], Var, CodeModel,
-		CanFail, StoreMap, EndLabel, CasesCode) -->
-	code_info__get_maybe_trace_info(MaybeTraceInfo),
-	( { MaybeTraceInfo = yes(TraceInfo) } ->
-		{ Goal = _ - GoalInfo },
-		{ goal_info_get_goal_path(GoalInfo, Path) },
-		trace__generate_event_code(switch(Path), TraceInfo,
-			TraceCode)
-	;
-		{ TraceCode = empty }
-	),
-	code_info__grab_code_info(CodeInfo0),
+		CanFail, StoreMap, EndLabel, MaybeEnd0, MaybeEnd, CasesCode) -->
+	code_info__remember_position(BranchStart),
 	(
 		{ Cases = [_|_] ; CanFail = can_fail }
 	->
 		unify_gen__generate_tag_test(Var, Cons, branch_on_failure,
 			NextLabel, TestCode),
+		trace__maybe_generate_internal_event_code(Goal, TraceCode),
 		code_gen__generate_goal(CodeModel, Goal, GoalCode),
-		code_info__generate_branch_end(CodeModel, StoreMap, SaveCode),
+		code_info__generate_branch_end(StoreMap, MaybeEnd0, MaybeEnd1,
+			SaveCode),
 		{ ElseCode = node([
 			goto(label(EndLabel)) -
 				"skip to the end of the switch",
@@ -332,24 +333,22 @@ switch_gen__generate_cases([case(_, _, Cons, Goal) | Cases], Var, CodeModel,
 			tree(GoalCode,
 			tree(SaveCode,
 			     ElseCode))))
-		},
-		code_info__grab_code_info(CodeInfo1),
-		code_info__slap_code_info(CodeInfo0)
+		}
 	;
+		trace__maybe_generate_internal_event_code(Goal, TraceCode),
 		code_gen__generate_goal(CodeModel, Goal, GoalCode),
-		code_info__generate_branch_end(CodeModel, StoreMap, SaveCode),
+		code_info__generate_branch_end(StoreMap, MaybeEnd0, MaybeEnd1,
+			SaveCode),
 		{ ThisCaseCode =
 			tree(TraceCode,
 			tree(GoalCode,
 			     SaveCode))
-		},
-		code_info__grab_code_info(CodeInfo1),
-		code_info__slap_code_info(CodeInfo0)
+		}
 	),
+	code_info__reset_to_position(BranchStart),
 		% generate the rest of the cases.
 	switch_gen__generate_cases(Cases, Var, CodeModel, CanFail, StoreMap,
-		EndLabel, OtherCasesCode),
-	{ CasesCode = tree(ThisCaseCode, OtherCasesCode) },
-	code_info__slap_code_info(CodeInfo1).
+		EndLabel, MaybeEnd1, MaybeEnd, OtherCasesCode),
+	{ CasesCode = tree(ThisCaseCode, OtherCasesCode) }.
 
 %------------------------------------------------------------------------------%
