@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1997 The University of Melbourne.
+% Copyright (C) 1996-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -20,6 +20,11 @@
 %	- used to instantiate an output variable
 %	- involved in a simple test, switch or a semidet deconstruction 
 %	- used as an argument to another predicate in this module which is used.
+%  When using alternate liveness calculation, the following variables are 
+%  also considered used
+%	- a type-info (or part of a type-info) of a type parameter of the 
+%	  type of a variable that is used (for example, if a variable
+%	  of type list(T) is used, then TypeInfo_for_T is used)
 %
 %  The first step is to determine which arguments of which predicates are
 %	used locally to their predicate. For each unused argument, a set of
@@ -50,29 +55,29 @@
 :- import_module hlds_pred, hlds_goal, hlds_data, hlds_out, type_util, instmap.
 :- import_module code_util, globals, make_hlds, mercury_to_mercury, mode_util.
 :- import_module options, prog_data, prog_out, quantification, special_pred.
-:- import_module passes_aux, inst_match.
+:- import_module passes_aux, inst_match, modules, polymorphism, goal_util.
 
-:- import_module bool, char, int, list, map, require.
-:- import_module set, std_util, string, term, varset. 
+:- import_module assoc_list, bool, char, int, list, map, require.
+:- import_module set, std_util, string.
 
 		% Information about the dependencies of a variable
 		% that is not known to be used.
 :- type usage_info --->
-		unused(set(var), set(arg)).
+		unused(set(prog_var), set(arg)).
 
 	% A collection of variable usages for each procedure.
 :- type var_usage == map(pred_proc_id, var_dep).
 
 	% arguments are stored as their variable id, not their index
 	%	in the argument vector
-:- type arg == pair(pred_proc_id, var). 
+:- type arg == pair(pred_proc_id, prog_var). 
 
 		% Contains dependency information for the variables
 		% in a procedure that are not yet known to be used.
-:- type var_dep == map(var, usage_info).
+:- type var_dep == map(prog_var, usage_info).
 
 :- type warning_info --->
-		warning_info(term__context, string, int, list(int)).
+		warning_info(prog_context, string, int, list(int)).
 			% context, pred name, arity, list of args to warn 
 
 
@@ -93,13 +98,15 @@ unused_args__process_module(ModuleInfo0, ModuleInfo) -->
 	globals__io_lookup_bool_option(make_optimization_interface, MakeOpt),
 	( { MakeOpt = yes } ->
 		{ module_info_name(ModuleInfo0, ModuleName) },
-		{ string__append(ModuleName, ".opt.tmp", OptFileName) },
+		module_name_to_file_name(ModuleName, ".opt.tmp", no,
+				OptFileName),
 		io__open_append(OptFileName, OptFileRes),
-		( { OptFileRes = ok(OptFile) } ->
+		( { OptFileRes = ok(OptFile) },
 			{ MaybeOptFile = yes(OptFile) }
-		;
-			io__write_strings(["Cannot open `",
-				OptFileName, "' for output\n"]),
+		; { OptFileRes = error(IOError) },
+			{ io__error_message(IOError, IOErrorMessage) },
+			io__write_strings(["Cannot open `", OptFileName,
+				"' for output: ", IOErrorMessage]),
 			io__set_exit_status(1),
 			{ MaybeOptFile = no }
 		)
@@ -247,8 +254,28 @@ setup_pred_args(ModuleInfo, PredId, [ProcId | Rest], UnusedArgInfo, VarUsage0,
 		initialise_vardep(VarDep0, Vars, VarDep1),
 		setup_output_args(ModuleInfo, HeadVars,
 			ArgModes, VarDep1, VarDep2),
+		
+		module_info_globals(ModuleInfo, Globals),
+		proc_interface_should_use_typeinfo_liveness(PredInfo, ProcId,
+			Globals, TypeInfoLiveness),
+		( 
+			TypeInfoLiveness = yes,
+			pred_info_module(PredInfo, PredModule),
+			pred_info_name(PredInfo, PredName),
+			pred_info_arity(PredInfo, PredArity),
+			\+ polymorphism__no_type_info_builtin(PredModule,
+				PredName, PredArity)
+		->
+			proc_info_typeinfo_varmap(ProcInfo, TVarMap),
+			setup_typeinfo_deps(Vars, VarTypes, 
+				proc(PredId, ProcId), TVarMap, VarDep2,
+				VarDep3)
+		;
+			VarDep2 = VarDep3
+		),
+
 		proc_info_goal(ProcInfo, Goal - _),
-		traverse_goal(ModuleInfo, Goal, VarDep2, VarDep),
+		traverse_goal(ModuleInfo, Goal, VarDep3, VarDep),
 		map__set(VarUsage0, proc(PredId, ProcId), VarDep, VarUsage1),
 		PredProcs1 = [proc(PredId, ProcId) | PredProcs0],
 		OptProcs1 = OptProcs0
@@ -257,7 +284,7 @@ setup_pred_args(ModuleInfo, PredId, [ProcId | Rest], UnusedArgInfo, VarUsage0,
 		VarUsage, PredProcs1, PredProcs, OptProcs1, OptProcs).
 
 
-:- pred initialise_vardep(var_dep::in, list(var)::in, var_dep::out) is det.
+:- pred initialise_vardep(var_dep::in, list(prog_var)::in, var_dep::out) is det.
 
 initialise_vardep(VarDep, [], VarDep).
 initialise_vardep(VarDep0, [Var | Vars], VarDep) :-
@@ -269,9 +296,38 @@ initialise_vardep(VarDep0, [Var | Vars], VarDep) :-
 %-------------------------------------------------------------------------------
 	% Predicates for manipulating the var_usage and var_dep structures.
 
+	% For each variable ensure the typeinfos describing the
+	% type parameters of the type of the variable depend on the
+	% head variable.
+	% For example, if HeadVar1 has type list(T), then TypeInfo_for_T
+	% is used if HeadVar1 is used.
+:- pred setup_typeinfo_deps(list(prog_var)::in, map(prog_var, type)::in,
+			pred_proc_id::in, map(tvar, type_info_locn)::in, 
+			var_dep::in, var_dep::out) is det.
+
+setup_typeinfo_deps([], _, _, _, VarDep, VarDep). 
+setup_typeinfo_deps([Var | Vars], VarTypeMap, PredProcId, TVarMap, VarDep0, 
+		VarDep) :-
+	map__lookup(VarTypeMap, Var, Type),
+	type_util__vars(Type, TVars),
+	list__map(lambda([TVar::in, TypeInfoVar::out] is det, 
+		(
+			map__lookup(TVarMap, TVar, Locn),
+			type_info_locn_var(Locn, TypeInfoVar)
+		)), 
+		TVars, TypeInfoVars),
+	AddArgDependency = 
+		lambda([TVar::in, VarDepA::in, VarDepB::out] is det, (
+			add_arg_dep(VarDepA, TVar, PredProcId, Var, VarDepB)
+		)),
+	list__foldl(AddArgDependency, TypeInfoVars, VarDep0, VarDep1),
+	setup_typeinfo_deps(Vars, VarTypeMap, PredProcId, TVarMap, 
+		VarDep1, VarDep).
+
+
 	% Get output arguments for a procedure given the headvars and the
-	% argument modes.
-:- pred setup_output_args(module_info::in, list(var)::in, list(mode)::in,
+	% argument modes, and set them as used.
+:- pred setup_output_args(module_info::in, list(prog_var)::in, list(mode)::in,
 			var_dep::in, var_dep::out) is det.
 
 setup_output_args(ModuleInfo, HVars, ArgModes, VarDep0, VarDep) :-
@@ -300,7 +356,7 @@ setup_output_args(ModuleInfo, HVars, ArgModes, VarDep0, VarDep) :-
 
 	% searches for the dependencies of a variable, succeeds if the variable
 	%	is definitely used
-:- pred var_is_used(pred_proc_id::in, var::in, var_usage::in) is semidet.
+:- pred var_is_used(pred_proc_id::in, prog_var::in, var_usage::in) is semidet.
 
 var_is_used(PredProc, Var, VarUsage) :-
 	\+ (
@@ -308,13 +364,14 @@ var_is_used(PredProc, Var, VarUsage) :-
 		map__contains(UsageInfos, Var)
 	).
 
-:- pred local_var_is_used(var_dep::in, var::in) is semidet.
+:- pred local_var_is_used(var_dep::in, prog_var::in) is semidet.
 
 local_var_is_used(VarDep, Var) :-
 	\+ map__contains(VarDep, Var).
 
 		% add a list of aliases for a variable
-:- pred add_aliases(var_dep::in, var::in, list(var)::in, var_dep::out) is det.
+:- pred add_aliases(var_dep::in, prog_var::in, list(prog_var)::in,
+		var_dep::out) is det.
 
 add_aliases(UseInf0, Var, Aliases, UseInf) :-
 	(
@@ -328,18 +385,19 @@ add_aliases(UseInf0, Var, Aliases, UseInf) :-
 		UseInf = UseInf0
 	).
 
-:- pred set_list_vars_used(var_dep::in, list(var)::in, var_dep::out) is det.
+:- pred set_list_vars_used(var_dep::in, list(prog_var)::in,
+		var_dep::out) is det.
 
 set_list_vars_used(UseInfo0, Vars, UseInfo) :-
 	map__delete_list(UseInfo0, Vars, UseInfo).
 
-:- pred set_var_used(var_dep::in, var::in, var_dep::out) is det.
+:- pred set_var_used(var_dep::in, prog_var::in, var_dep::out) is det.
 
 set_var_used(UseInfo0, Var, UseInfo) :-
 	map__delete(UseInfo0, Var, UseInfo).
 
 
-:- pred lookup_local_var(var_dep::in, var::in, usage_info::out) is semidet.
+:- pred lookup_local_var(var_dep::in, prog_var::in, usage_info::out) is semidet.
 
 lookup_local_var(VarDep, Var, UsageInfo) :-
 	map__search(VarDep, Var, UsageInfo).
@@ -356,6 +414,10 @@ lookup_local_var(VarDep, Var, UsageInfo) :-
 traverse_goal(ModuleInfo, conj(Goals), UseInf0, UseInf) :-
 	traverse_list_of_goals(ModuleInfo, Goals, UseInf0, UseInf).
 
+% handle parallel conjunction
+traverse_goal(ModuleInfo, par_conj(Goals, _SM), UseInf0, UseInf) :-
+	traverse_list_of_goals(ModuleInfo, Goals, UseInf0, UseInf).
+
 % handle disjunction
 traverse_goal(ModuleInfo, disj(Goals, _), UseInf0, UseInf) :-
 	traverse_list_of_goals(ModuleInfo, Goals, UseInf0, UseInf).
@@ -386,17 +448,25 @@ traverse_goal(ModuleInfo, not(Goal - _), UseInf0, UseInf) :-
 	traverse_goal(ModuleInfo, Goal, UseInf0, UseInf).
 
 % handle quantification
-traverse_goal(ModuleInfo, some(_,  Goal - _), UseInf0, UseInf) :-
+traverse_goal(ModuleInfo, some(_, _, Goal - _), UseInf0, UseInf) :-
 	traverse_goal(ModuleInfo, Goal, UseInf0, UseInf).
 
-
 % we assume that higher-order predicate calls use all variables involved
-traverse_goal(_, higher_order_call(PredVar,Args,_,_,_,_), UseInf0, UseInf) :-
-	set_list_vars_used(UseInf0, [PredVar|Args], UseInf).
+traverse_goal(_, generic_call(GenericCall, Args, _, _), UseInf0, UseInf) :-
+	goal_util__generic_call_vars(GenericCall, CallArgs),
+	set_list_vars_used(UseInf0, CallArgs, UseInf1),
+	set_list_vars_used(UseInf1, Args, UseInf).
 
-% handle pragma(c_code, ...) - pragma_c_code uses all its args
-traverse_goal(_, pragma_c_code(_, _, _, _, Args, _, _, _), UseInf0, UseInf) :-
-	set_list_vars_used(UseInf0, Args, UseInf).
+% handle pragma c_code(...) -
+% only those arguments which have C names can be used in the C code.
+traverse_goal(_, pragma_c_code(_, _, _, Args, Names, _, _), UseInf0, UseInf) :-
+	assoc_list__from_corresponding_lists(Args, Names, ArgsAndNames),
+	ArgIsUsed = lambda([ArgAndName::in, Arg::out] is semidet, (
+				ArgAndName = Arg - MaybeName,
+				MaybeName = yes(_)
+			)),
+	list__filter_map(ArgIsUsed, ArgsAndNames, UsedArgs),
+	set_list_vars_used(UseInf0, UsedArgs, UseInf).
 
 % cases to handle all the different types of unification
 traverse_goal(_, unify(_, _, _, simple_test(Var1, Var2),_), UseInf0, UseInf)
@@ -405,46 +475,43 @@ traverse_goal(_, unify(_, _, _, simple_test(Var1, Var2),_), UseInf0, UseInf)
 	set_var_used(UseInf1, Var2, UseInf).
 		
 traverse_goal(_, unify(_, _, _, assign(Var1, Var2), _), UseInf0, UseInf) :-
-	(
-		map__contains(UseInf0, Var1)
-	->
-		add_aliases(UseInf0, Var2, [Var1], UseInf)
-	;
+	( local_var_is_used(UseInf0, Var1) ->
 		% if Var1 used to instantiate an output argument, Var2 used
 		set_var_used(UseInf0, Var2, UseInf)
+	;
+		add_aliases(UseInf0, Var2, [Var1], UseInf)
 	).
 
 traverse_goal(ModuleInfo,
 		unify(Var1, _, _, deconstruct(_, _, Args, Modes, CanFail), _),
 		UseInf0, UseInf) :-
+	partition_deconstruct_args(ModuleInfo, Args,
+		Modes, InputVars, OutputVars),
+		% The deconstructed variable is used if any of the
+		% variables, that the deconstruction binds are used.
+	add_aliases(UseInf0, Var1, OutputVars, UseInf1),
+		% Treat a deconstruction that further instantiates its
+		% left arg as a partial construction.
+	add_construction_aliases(UseInf1, Var1, InputVars, UseInf2),
 	(
 		CanFail = can_fail	
 	->
 		% a deconstruction that can_fail uses its left arg
-		set_var_used(UseInf0, Var1, UseInf)
+		set_var_used(UseInf2, Var1, UseInf)
 	;
-		get_instantiating_variables(ModuleInfo, Args, Modes, InputVars),
-		list__delete_elems(Args, InputVars, OutputVars),
-			% The deconstructed variable is used if any of the
-			% variables, that the deconstruction binds are used.
-		add_aliases(UseInf0, Var1, OutputVars, UseInf1),
-			% Treat a deconstruction that further instantiates its
-			% left arg as a partial construction.
-		add_construction_aliases(UseInf1, Var1, InputVars, UseInf)	
+		UseInf = UseInf2	
 	).
 
-traverse_goal(_, unify(Var1, _, _, construct(_, _, Args, _), _),
+traverse_goal(_, unify(Var1, _, _, construct(_, _, Args, _, _, _, _), _),
 					UseInf0, UseInf) :-
-	(
-		map__contains(UseInf0, Var1)
-	->
-		add_construction_aliases(UseInf0, Var1, Args, UseInf)
-	;
+	( local_var_is_used(UseInf0, Var1) ->
 		set_list_vars_used(UseInf0, Args, UseInf)
+	;
+		add_construction_aliases(UseInf0, Var1, Args, UseInf)
 	).
 	
 	% These should be transformed into calls by polymorphism.m.
-traverse_goal(_, unify(Var, Rhs, _, complicated_unify(_, _), _),
+traverse_goal(_, unify(Var, Rhs, _, complicated_unify(_, _, _), _),
 		UseInf0, UseInf) :-
     	% This is here to cover the case where unused arguments is called 
 	% with --error-check-only and polymorphism has not been run.
@@ -456,9 +523,13 @@ traverse_goal(_, unify(Var, Rhs, _, complicated_unify(_, _), _),
 		error("complicated unifications should only be var-var")
 	).
 
+traverse_goal(_, bi_implication(_, _), _, _) :-
+	% these should have been expanded out by now
+	error("traverse_goal: unexpected bi_implication").
+
 	% add PredProc - HeadVar as an alias for the same element of Args.
-:- pred add_pred_call_arg_dep(pred_proc_id::in, list(var)::in, list(var)::in,
-					var_dep::in, var_dep::out) is det.
+:- pred add_pred_call_arg_dep(pred_proc_id::in, list(prog_var)::in,
+		list(prog_var)::in, var_dep::in, var_dep::out) is det.
 
 add_pred_call_arg_dep(PredProc, LocalArguments, HeadVarIds,
 					UseInf0, UseInf) :-
@@ -477,8 +548,8 @@ add_pred_call_arg_dep(PredProc, LocalArguments, HeadVarIds,
 	).
 		
 
-:- pred add_arg_dep(var_dep::in, var::in, pred_proc_id::in,
-					var::in, var_dep::out) is det.
+:- pred add_arg_dep(var_dep::in, prog_var::in, pred_proc_id::in,
+					prog_var::in, var_dep::out) is det.
 
 add_arg_dep(UseInf0, Var, PredProc, Arg, UseInf) :-
 	(
@@ -491,33 +562,48 @@ add_arg_dep(UseInf0, Var, PredProc, Arg, UseInf) :-
 		UseInf = UseInf0
 	).
 			
-	% Returns variables which further instantiate a deconstructed variable.
-:- pred get_instantiating_variables(module_info::in, list(var)::in,
-				list(uni_mode)::in, list(var)::out) is det.
+	% Partition the arguments to a deconstruction into inputs
+	% and outputs.
+:- pred partition_deconstruct_args(module_info::in, list(prog_var)::in,
+		list(uni_mode)::in, list(prog_var)::out,
+		list(prog_var)::out) is det.
 
-get_instantiating_variables(ModuleInfo, ArgVars, ArgModes, InstVars) :-
+partition_deconstruct_args(ModuleInfo, ArgVars, ArgModes,
+		InputVars, OutputVars) :-
 	(
 		ArgVars = [Var | Vars], ArgModes = [Mode | Modes]
 	->
-		Mode = ((Inst1 - _) -> (Inst2 - _)),
-		(
-			inst_matches_binding(Inst1, Inst2, ModuleInfo)
-		->
-			InstVars = InstVars1
+		partition_deconstruct_args(ModuleInfo,
+			Vars, Modes, InputVars1, OutputVars1),
+		Mode = ((InitialInst1 - InitialInst2) ->
+			(FinalInst1 - FinalInst2)),
+
+		% If the inst of the argument of the LHS is changed,
+		% the argument is input.
+		( inst_matches_binding(InitialInst1, FinalInst1, ModuleInfo) ->
+			InputVars = InputVars1
 		;
-			InstVars = [Var | InstVars1]
+			InputVars = [Var | InputVars1]
 		),
-		get_instantiating_variables(ModuleInfo, Vars, Modes, InstVars1)
+
+		% If the inst of the argument of the RHS is changed,
+		% the argument is output.
+		( inst_matches_binding(InitialInst2, FinalInst2, ModuleInfo) ->
+			OutputVars = OutputVars1
+		;
+			OutputVars = [Var | OutputVars1]
+		)
 	;
-		( ArgVars = [], ArgModes = [] )
+		ArgVars = [], ArgModes = []
 	->
-		InstVars = []
+		InputVars = [],
+		OutputVars = []
 	;
 		error("get_instantiating_variables - invalid call")
 	).
 
 		% add Alias as an alias for all of Vars
-:- pred add_construction_aliases(var_dep::in, var::in, list(var)::in,
+:- pred add_construction_aliases(var_dep::in, prog_var::in, list(prog_var)::in,
 						var_dep::out) is det.
 
 add_construction_aliases(UseInf, _, [], UseInf).
@@ -604,7 +690,7 @@ unused_args_check_proc(PredProcId, Changed0, Changed, VarUsage0, VarUsage) :-
 
 	% check each var of a procedure in turn 
 :- pred unused_args_check_all_vars(var_usage::in, bool::in, bool::out,
-			list(var)::in, var_dep::in, var_dep::out) is det.
+			list(prog_var)::in, var_dep::in, var_dep::out) is det.
 
 unused_args_check_all_vars(_, Changed, Changed, [], LocalVars, LocalVars). 
 unused_args_check_all_vars(VarUsage, Changed0, Changed, [Var| Vars],
@@ -849,7 +935,7 @@ make_new_pred_info(ModuleInfo, PredInfo0, UnusedArgs, NameSuffix, Status,
 	pred_info_get_is_pred_or_func(PredInfo0, PredOrFunc),
 	proc_id_to_int(ProcId, ProcInt),
 	string__int_to_string(ProcInt, Id),
-	pred_info_arg_types(PredInfo0, Tvars, ArgTypes0),
+	pred_info_arg_types(PredInfo0, Tvars, ExistQVars, ArgTypes0),
 		% create a unique new pred name using the old proc_id
 	(
 		string__prefix(Name0, "__"),
@@ -858,22 +944,24 @@ make_new_pred_info(ModuleInfo, PredInfo0, UnusedArgs, NameSuffix, Status,
 		(
 				% fix up special pred names
 			special_pred_get_type(Name0, ArgTypes0, Type),
-			type_to_type_id(Type, TypeId0, _)
+			type_to_type_id(Type, TypeId, _)
 		->
-			TypeId = TypeId0
+			type_util__type_id_module(ModuleInfo,
+				TypeId, TypeModule),
+			type_util__type_id_name(ModuleInfo, TypeId, TypeName),
+			type_util__type_id_arity(ModuleInfo,
+				TypeId, TypeArity),
+			string__int_to_string(TypeArity, TypeAr),
+			prog_out__sym_name_to_string(TypeModule,
+				TypeModuleString0),
+			string__replace_all(TypeModuleString0, ":", "__",
+				TypeModuleString),
+			string__append_list([Name0, "_", TypeModuleString,
+				"__", TypeName, "_", TypeAr], Name1)
 		;
-			string__append_list(["unused_args:make_new_pred_info\n",
-					"cannot make label for special pred `",
-					Name0, "'."], Message),
-			error(Message)
-		),
-		type_util__type_id_module(ModuleInfo, TypeId, TypeModule),
-		type_util__type_id_name(ModuleInfo, TypeId, TypeName),
-		type_util__type_id_arity(ModuleInfo, TypeId, TypeArity),
-		string__int_to_string(TypeArity, TypeAr),
-		string__append_list(
-			[Name0, "_", TypeModule, "__", TypeName, "_", TypeAr],
-			Name1)
+			% The special predicate has already been specialised.
+			Name1 = Name0
+		)
 	;
 		Name1 = Name0
 	),
@@ -883,20 +971,24 @@ make_new_pred_info(ModuleInfo, PredInfo0, UnusedArgs, NameSuffix, Status,
 	remove_listof_elements(ArgTypes0, 1, UnusedArgs, ArgTypes),
 	pred_info_context(PredInfo0, Context),
 	pred_info_clauses_info(PredInfo0, ClausesInfo),
-	pred_info_get_marker_list(PredInfo0, MarkerList),
+	pred_info_get_markers(PredInfo0, Markers),
 	pred_info_get_goal_type(PredInfo0, GoalType),
+	pred_info_get_class_context(PredInfo0, ClassContext),
+	pred_info_get_aditi_owner(PredInfo0, Owner),
+	map__init(EmptyProofs),
 		% *** This will need to be fixed when the condition
 		%	field of the pred_info becomes used.
 	pred_info_init(PredModule, qualified(PredModule, Name), Arity, Tvars,
-		ArgTypes, true, Context, ClausesInfo, Status, MarkerList,
-		GoalType, PredOrFunc, PredInfo1),
+		ExistQVars, ArgTypes, true, Context, ClausesInfo, Status,
+		Markers, GoalType, PredOrFunc, ClassContext, EmptyProofs,
+		Owner, PredInfo1),
 	pred_info_set_typevarset(PredInfo1, TypeVars, PredInfo).
 
-
 	% Replace the goal in the procedure with one to call the given
 	% pred_id and proc_id.
 :- pred create_call_goal(list(int)::in, pred_id::in, proc_id::in,
-		string::in, string::in, proc_info::in, proc_info::out) is det.
+		module_name::in, string::in, proc_info::in, proc_info::out)
+		is det.
 
 create_call_goal(UnusedArgs, NewPredId, NewProcId, PredModule,
 		PredName, OldProc0, OldProc) :-
@@ -909,7 +1001,7 @@ create_call_goal(UnusedArgs, NewPredId, NewProcId, PredModule,
 	map__from_corresponding_lists(HeadVars, VarTypeList, VarTypes1),
 		% the varset should probably be fixed up, but it
 		% shouldn't make too much difference
-	proc_info_variables(OldProc0, Varset0),
+	proc_info_varset(OldProc0, Varset0),
 	remove_listof_elements(HeadVars, 1, UnusedArgs, NewHeadVars),
 	GoalExpr = call(NewPredId, NewProcId, NewHeadVars,
 		      not_builtin, no, qualified(PredModule, PredName)),
@@ -927,9 +1019,9 @@ create_call_goal(UnusedArgs, NewPredId, NewProcId, PredModule,
 		proc_call_info::in, proc_call_info::out,
 		module_info::in, module_info::out) is det.
 
- make_imported_unused_args_pred_infos([], ProcCallInfo, ProcCallInfo,
+make_imported_unused_args_pred_infos([], ProcCallInfo, ProcCallInfo,
 				ModuleInfo, ModuleInfo).
- make_imported_unused_args_pred_infos([OptProc | OptProcs],
+make_imported_unused_args_pred_infos([OptProc | OptProcs],
  		ProcCallInfo0, ProcCallInfo, ModuleInfo0, ModuleInfo) :-
 	module_info_unused_arg_info(ModuleInfo0, UnusedArgInfo),
 	map__lookup(UnusedArgInfo, OptProc, UnusedArgs),
@@ -937,7 +1029,7 @@ create_call_goal(UnusedArgs, NewPredId, NewProcId, PredModule,
 	module_info_pred_proc_info(ModuleInfo0,
 		PredId, ProcId, PredInfo0, ProcInfo0),
 	make_new_pred_info(ModuleInfo0, PredInfo0, UnusedArgs,
-		"__ua", imported, OptProc, NewPredInfo0),
+		"__ua", imported(interface), OptProc, NewPredInfo0),
 	pred_info_procedures(NewPredInfo0, NewProcs0),
 	next_mode_id(NewProcs0, no, NewProcId),
 
@@ -994,7 +1086,7 @@ remove_listof_elements(List0, ArgNo, ElemsToRemove, List) :-
 	).
 
 	
-:- pred get_unused_arg_nos(var_dep::in, list(var)::in, int::in,
+:- pred get_unused_arg_nos(var_dep::in, list(prog_var)::in, int::in,
 						list(int)::out) is det.
 
 get_unused_arg_nos(_, [], _, []).
@@ -1068,7 +1160,7 @@ do_fixup_unused_args(VarUsage, proc(OldPredId, OldProcId), ProcCallInfo,
 
 	proc_info_headvars(ProcInfo0, HeadVars0),
 	proc_info_argmodes(ProcInfo0, ArgModes0),
-	proc_info_variables(ProcInfo0, Varset0),
+	proc_info_varset(ProcInfo0, Varset0),
 	proc_info_goal(ProcInfo0, Goal0),
 	remove_listof_elements(HeadVars0, 1, UnusedArgs, HeadVars),
 	remove_listof_elements(ArgModes0, 1, UnusedArgs, ArgModes),
@@ -1097,7 +1189,7 @@ do_fixup_unused_args(VarUsage, proc(OldPredId, OldProcId), ProcCallInfo,
 
 
 % 	this is the important bit of the transformation
-:- pred fixup_goal(module_info::in, list(var)::in, proc_call_info::in,
+:- pred fixup_goal(module_info::in, list(prog_var)::in, proc_call_info::in,
 			bool::out, hlds_goal::in, hlds_goal::out) is det.
 
 fixup_goal(ModuleInfo, UnusedVars, ProcCallInfo, Changed, Goal0, Goal) :-
@@ -1114,7 +1206,7 @@ fixup_goal(ModuleInfo, UnusedVars, ProcCallInfo, Changed, Goal0, Goal) :-
 	Goal = GoalExpr - GoalInfo.
 		
 
-:- pred fixup_goal_expr(module_info::in, list(var)::in, proc_call_info::in,
+:- pred fixup_goal_expr(module_info::in, list(prog_var)::in, proc_call_info::in,
 			bool::out, hlds_goal::in, hlds_goal::out) is det.
 
 fixup_goal_expr(ModuleInfo, UnusedVars, ProcCallInfo, Changed,
@@ -1122,6 +1214,12 @@ fixup_goal_expr(ModuleInfo, UnusedVars, ProcCallInfo, Changed,
 	fixup_conjuncts(ModuleInfo, UnusedVars, ProcCallInfo, no,
 						Changed, Goals0, Goals).
 
+fixup_goal_expr(ModuleInfo, UnusedVars, ProcCallInfo, Changed,
+		par_conj(Goals0, SM) - GoalInfo,
+		par_conj(Goals, SM) - GoalInfo) :-
+	fixup_conjuncts(ModuleInfo, UnusedVars, ProcCallInfo, no,
+						Changed, Goals0, Goals).
+
 fixup_goal_expr(ModuleInfo, UnusedVars, ProcCallInfo, Changed,
 		disj(Goals0, SM) - GoalInfo, disj(Goals, SM) - GoalInfo) :-
 	fixup_disjuncts(ModuleInfo, UnusedVars, ProcCallInfo,
@@ -1147,8 +1245,8 @@ fixup_goal_expr(ModuleInfo, UnusedVars, ProcCallInfo, Changed,
 	bool__or_list([Changed1, Changed2, Changed3], Changed).
 
 fixup_goal_expr(ModuleInfo, UnusedVars, ProcCallInfo, Changed,
-		some(Vars, SubGoal0) - GoalInfo,
-		some(Vars, SubGoal) - GoalInfo) :-
+		some(Vars, CanRemove, SubGoal0) - GoalInfo,
+		some(Vars, CanRemove, SubGoal) - GoalInfo) :-
 	fixup_goal(ModuleInfo, UnusedVars, ProcCallInfo,
 				Changed, SubGoal0, SubGoal).
 
@@ -1187,14 +1285,18 @@ fixup_goal_expr(ModuleInfo, UnusedVars, _ProcCallInfo,
 
 fixup_goal_expr(_ModuleInfo, _UnusedVars, _ProcCallInfo, no,
 			GoalExpr - GoalInfo, GoalExpr - GoalInfo) :-
-	GoalExpr = higher_order_call(_, _, _, _, _, _).
+	GoalExpr = generic_call(_, _, _, _).
 
 fixup_goal_expr(_ModuleInfo, _UnusedVars, _ProcCallInfo, no,
 			GoalExpr - GoalInfo, GoalExpr - GoalInfo) :-
-	GoalExpr = pragma_c_code(_, _, _, _, _, _, _, _).
+	GoalExpr = pragma_c_code(_, _, _, _, _, _, _).
+
+fixup_goal_expr(_, _, _, _, bi_implication(_, _) - _, _) :-
+	% these should have been expanded out by now
+	error("fixup_goal_expr: unexpected bi_implication").
 
 	% Remove useless unifications from a list of conjuncts.
-:- pred fixup_conjuncts(module_info::in, list(var)::in, proc_call_info::in,
+:- pred fixup_conjuncts(module_info::in, list(prog_var)::in, proc_call_info::in,
 		bool::in, bool::out, hlds_goals::in, hlds_goals::out) is det. 
 
 fixup_conjuncts(_, _, _, Changed, Changed, [], []).
@@ -1225,7 +1327,7 @@ fixup_conjuncts(ModuleInfo, UnusedVars, ProcCallInfo, Changed0, Changed,
 	% We can't remove unused goals from the list of disjuncts as we do
 	% for conjuncts, since that would change the determinism of
 	% the goal.
-:- pred fixup_disjuncts(module_info::in, list(var)::in, proc_call_info::in,
+:- pred fixup_disjuncts(module_info::in, list(prog_var)::in, proc_call_info::in,
 		bool::in, bool::out, hlds_goals::in, hlds_goals::out) is det. 
 
 fixup_disjuncts(_, _, _, Changed, Changed, [], []).
@@ -1243,7 +1345,7 @@ fixup_disjuncts(ModuleInfo, UnusedVars, ProcCallInfo, Changed0, Changed,
 	fixup_disjuncts(ModuleInfo, UnusedVars, ProcCallInfo,
 			Changed1, Changed, Goals0, Goals).
 
-:- pred fixup_cases(module_info::in, list(var)::in, proc_call_info::in,
+:- pred fixup_cases(module_info::in, list(prog_var)::in, proc_call_info::in,
 		bool::in, bool::out, list(case)::in, list(case)::out) is det.
 		
 fixup_cases(_, _, _, Changed, Changed, [], []).
@@ -1264,7 +1366,7 @@ fixup_cases(ModuleInfo, UnusedVars, ProcCallInfo, Changed0, Changed,
 
 		% fix up a unification, fail if the unification is no
 		%	longer needed
-:- pred fixup_unify(module_info::in, list(var)::in, bool::out,
+:- pred fixup_unify(module_info::in, list(prog_var)::in, bool::out,
 				unification::in, unification::out) is semidet.
 
 	% a simple test doesn't have any unused vars to fixup
@@ -1276,13 +1378,12 @@ fixup_unify(_, UnusedVars, no, assign(Var1, Var2), assign(Var1, Var2)) :-
 	\+ list__member(Var1, UnusedVars).
 
 	% LVar unused => we don't need the unification
-fixup_unify(_, UnusedVars, no, construct(LVar, ConsId, ArgVars, ArgModes),
-				construct(LVar, ConsId, ArgVars, ArgModes)) :-	
+fixup_unify(_, UnusedVars, no, Unify, Unify) :-
+	Unify = construct(LVar, _, _, _, _, _, _),
 	\+ list__member(LVar, UnusedVars).
 	
-fixup_unify(ModuleInfo, UnusedVars, Changed,
-		deconstruct(LVar, ConsId, ArgVars, ArgModes, CanFail),
-		deconstruct(LVar, ConsId, ArgVars, ArgModes, CanFail)) :-
+fixup_unify(ModuleInfo, UnusedVars, Changed, Unify, Unify) :-
+	Unify =	deconstruct(LVar, _, ArgVars, ArgModes, CanFail),
 	\+ list__member(LVar, UnusedVars),
 	(
 			% are any of the args unused, if so we need to 	
@@ -1296,15 +1397,16 @@ fixup_unify(ModuleInfo, UnusedVars, Changed,
 	).
 
 	% These should be transformed into calls by polymorphism.m.
-fixup_unify(_, _, _, complicated_unify(_, _), _) :-
+fixup_unify(_, _, _, complicated_unify(_, _, _), _) :-
 		error("unused_args:fixup_goal : complicated unify").
 
 	% Check if any of the arguments of a deconstruction are unused, if
 	% so Changed will be yes and quantification will be rerun. Fails if
 	% none of the arguments are used. Arguments which further instantiate
 	% the deconstructed variable are ignored in this.
-:- pred check_deconstruct_args(module_info::in, list(var)::in, list(var)::in,
-			list(uni_mode)::in, bool::out, bool::in) is semidet.
+:- pred check_deconstruct_args(module_info::in, list(prog_var)::in,
+		list(prog_var)::in, list(uni_mode)::in,
+		bool::out, bool::in) is semidet.
 						
 check_deconstruct_args(ModuleInfo, UnusedVars, Args, Modes, Changed, Used) :-
 	(
@@ -1333,7 +1435,7 @@ check_deconstruct_args(ModuleInfo, UnusedVars, Args, Modes, Changed, Used) :-
 
 	% Remove unused vars from the instmap_delta, quantification fixes
 	%	up the rest.
-:- pred fixup_goal_info(list(var)::in, hlds_goal_info::in,
+:- pred fixup_goal_info(list(prog_var)::in, hlds_goal_info::in,
 						hlds_goal_info::out) is det.
 
 fixup_goal_info(UnusedVars, GoalInfo0, GoalInfo) :-
@@ -1363,6 +1465,7 @@ output_warnings_and_pragmas(ModuleInfo, UnusedArgInfo, WriteOptPragmas,
 			{
 			pred_info_name(PredInfo, Name),
 			\+ pred_info_is_imported(PredInfo),
+			\+ pred_info_import_status(PredInfo, opt_imported),
 				% Don't warn about builtins
 				% that have unused arguments.
 			\+ code_util__predinfo_is_builtin(PredInfo),
@@ -1394,7 +1497,6 @@ output_warnings_and_pragmas(ModuleInfo, UnusedArgInfo, WriteOptPragmas,
 		)
 	;
 		{ WarnedPredIds1 = WarnedPredIds0 }
-
 	),
 	output_warnings_and_pragmas(ModuleInfo, UnusedArgInfo,
 		WriteOptPragmas, DoWarn, Rest, WarnedPredIds1).
@@ -1407,7 +1509,9 @@ output_warnings_and_pragmas(ModuleInfo, UnusedArgInfo, WriteOptPragmas,
 write_unused_args_to_opt_file(no, _, _, _) --> [].
 write_unused_args_to_opt_file(yes(OptStream), PredInfo, ProcId, UnusedArgs) -->
 	(
-		{ pred_info_is_exported(PredInfo) },
+		( { pred_info_is_exported(PredInfo) }
+		; { pred_info_is_exported_to_submodules(PredInfo) }
+		),
 		{ UnusedArgs \= [] }
 	->
 		{ pred_info_module(PredInfo, Module) },
@@ -1415,8 +1519,9 @@ write_unused_args_to_opt_file(yes(OptStream), PredInfo, ProcId, UnusedArgs) -->
 		{ pred_info_arity(PredInfo, Arity) },
 		{ pred_info_get_is_pred_or_func(PredInfo, PredOrFunc) },
 		io__set_output_stream(OptStream, OldOutput),	
+		{ proc_id_to_int(ProcId, ModeNum) },
 		mercury_output_pragma_unused_args(PredOrFunc,
-			qualified(Module, Name), Arity, ProcId, UnusedArgs),
+			qualified(Module, Name), Arity, ModeNum, UnusedArgs),
 		io__set_output_stream(OldOutput, _)
 	;
 		[]
@@ -1470,14 +1575,14 @@ report_unused_args(PredInfo, UnusedArgs) -->
 	hlds_out__write_pred_or_func(PredOrFunc),
 	io__write_string(" `"),
 	{ pred_info_module(PredInfo, Module) },
-	io__write_string(Module),
+	prog_out__write_sym_name(Module),
 	io__write_string(":"),
 	{ pred_info_name(PredInfo, Name) },
 	io__write_string(Name), 
-	io__write_string(("'/")),
+	io__write_string("/"),
 	{ pred_info_arity(PredInfo, Arity) },
 	io__write_int(Arity),
-	io__write_string(":\n"),
+	io__write_string("':\n"),
 	prog_out__write_context(Context),
 	io__write_string("  warning: "),   
 	(
