@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -14,15 +14,15 @@
 
 :- interface.
 
-:- import_module list, io.
 :- import_module llds.
+:- import_module list, set, io.
 
 	% Find straight-line code sequences and optimize them using
 	% value numbering.
 
-:- pred value_number__main(list(instruction), list(instruction),
+:- pred value_number_main(list(instruction), set(label), list(instruction),
 	io__state, io__state).
-:- mode value_number__main(in, out, di, uo) is det.
+:- mode value_number_main(in, in, out, di, uo) is det.
 
 	% The main value numbering pass introduces references to temporary
 	% variables whose values need be preserved only within an extended
@@ -39,24 +39,25 @@
 
 :- implementation.
 
-:- import_module set, map, bimap, require, int, string, std_util, assoc_list.
-:- import_module bool.
-
+:- import_module builtin_ops.
 :- import_module vn_type, vn_table, vn_block, vn_order, vn_flush, vn_temploc. 
 :- import_module vn_cost, vn_debug, vn_util, vn_verify, vn_filter.
 :- import_module opt_debug, opt_util, peephole, labelopt.
 :- import_module globals, options, livemap, code_util.
 
+:- import_module map, bimap, require, int, string, std_util, assoc_list.
+:- import_module bool.
+
 	% We can't find out what variables are used by C code sequences,
 	% so we don't optimize any predicates containing them.
 
-value_number__main(Instrs0, Instrs) -->
+value_number_main(Instrs0, LayoutLabelSet, Instrs) -->
 	{ opt_util__get_prologue(Instrs0, ProcLabel,
 		LabelInstr, Comments, Instrs1) },
 	{ opt_util__new_label_no(Instrs1, 1000, N0) },
 	{ value_number__prepare_for_vn([LabelInstr | Instrs1], ProcLabel,
 		no, AllocSet, BreakSet, N0, N, Instrs2) },
-	{ labelopt__build_useset(Instrs2, UseSet) },
+	{ labelopt__build_useset(Instrs2, LayoutLabelSet, UseSet) },
 	{ livemap__build(Instrs2, MaybeLiveMap) },
 	(
 		{ MaybeLiveMap = yes(LiveMap) },
@@ -106,10 +107,12 @@ value_number__main(Instrs0, Instrs) -->
 	% Our caller will break the code sequence at these labels.
 	%
 	% Mkframe operations also change curfr, and therefore get the
-	% same treatment. We also apply this treatment to assignments
-	% to the control slots in nondet stack frames, since otherwise
-	% a bug in the rest of value numbering may cause them to be
-	% improperly deleted.
+	% same treatment. We also apply this treatment to (1) assignments to
+	% the control slots in nondet stack frames, since otherwise a bug
+	% in the rest of value numbering may cause them to be improperly
+	% deleted, and (b) uses of those control slots in the form of
+	% branches to do_redo and do_fail, since otherwise assignments
+	% to stack variables may be moved across them.
 
 :- pred value_number__prepare_for_vn(list(instruction), proc_label,
 	bool, set(label), set(label), int, int, list(instruction)).
@@ -122,26 +125,43 @@ value_number__prepare_for_vn([Instr0 | Instrs0], ProcLabel,
 		SeenAlloc, AllocSet, BreakSet, N0, N, Instrs) :-
 	Instr0 = Uinstr0 - _Comment,
 	( Uinstr0 = if_val(Test, TrueAddr) ->
-		( Instrs0 = [label(FalseLabelPrime) - _ | _] ->
-			FalseLabel = FalseLabelPrime,
-			FalseAddr = label(FalseLabel),
-			N1 = N0
-		;
-			FalseLabel = local(ProcLabel, N0),
-			FalseAddr = label(FalseLabel),
+		( ( TrueAddr = do_redo ; TrueAddr = do_fail ) ->
+			MaybeBeforeLabel = yes(local(ProcLabel, N0)),
 			N1 is N0 + 1
-		),
-		value_number__breakup_complex_if(Test, TrueAddr, FalseAddr,
-			FalseAddr, ProcLabel, N1, N2, IfInstrs),
-		value_number__prepare_for_vn(Instrs0, ProcLabel,
-			SeenAlloc, AllocSet, BreakSet, N2, N, Instrs1),
-		( N1 = N0 ->
-			list__append(IfInstrs, Instrs1, Instrs)
 		;
-			LabelInstr = label(FalseLabel) - "vn false label",
-			list__append(IfInstrs, [LabelInstr | Instrs1], Instrs)
+			MaybeBeforeLabel = no,
+			N1 = N0
+		),
+		( Instrs0 = [label(OldFalseLabel) - _ | _] ->
+			FalseLabel = OldFalseLabel,
+			MaybeNewFalseLabel = no,
+			N2 = N1
+		;
+			FalseLabel = local(ProcLabel, N1),
+			MaybeNewFalseLabel = yes(FalseLabel),
+			N2 is N1 + 1
+		),
+		FalseAddr = label(FalseLabel),
+		value_number__breakup_complex_if(Test, TrueAddr, FalseAddr,
+			FalseAddr, ProcLabel, N2, N3, IfInstrs),
+		value_number__prepare_for_vn(Instrs0, ProcLabel,
+			SeenAlloc, AllocSet, BreakSet0, N3, N, Instrs1),
+		( MaybeNewFalseLabel = yes(NewFalseLabel) ->
+			FalseInstr = label(NewFalseLabel) - "vn false label",
+			list__append(IfInstrs, [FalseInstr | Instrs1], Instrs2)
+		;
+			list__append(IfInstrs, Instrs1, Instrs2)
+		),
+		( MaybeBeforeLabel = yes(BeforeLabel) ->
+			set__insert(BreakSet0, BeforeLabel, BreakSet1),
+			set__insert(BreakSet1, FalseLabel, BreakSet),
+			BeforeInstr = label(BeforeLabel) - "vn before label",
+			Instrs = [BeforeInstr | Instrs2]
+		;
+			BreakSet = BreakSet0,
+			Instrs = Instrs2
 		)
-	; Uinstr0 = incr_hp(_, _, _) ->
+	; Uinstr0 = incr_hp(_, _, _, _) ->
 		( SeenAlloc = yes ->
 			N1 is N0 + 1,
 			NewLabel = local(ProcLabel, N0),
@@ -172,7 +192,7 @@ value_number__prepare_for_vn([Instr0 | Instrs0], ProcLabel,
 				Target = succfr(_)
 			)
 		;
-			Uinstr0 = mkframe(_, _, _)
+			Uinstr0 = mkframe(_, _)
 		)
 	->
 		N1 is N0 + 1,
@@ -406,9 +426,9 @@ value_number__optimize_fragment_2(Instrs0, LiveMap, Params, ParEntries,
 		{ value_number__push_decr_sp_back(Instrs1, Instrs2) },
 		{ value_number__push_incr_sp_forw(Instrs2, Instrs3) },
 		{ value_number__push_livevals_back(Instrs3, Instrs4) },
-		{ value_number__convert_back_modframe(Instrs4, Instrs5) },
-		{ vn_filter__block(Instrs5, Instrs6) },
-		{ peephole__optimize(Instrs6, Instrs7, _) },
+		{ vn_filter__block(Instrs4, Instrs5) },
+		globals__io_get_gc_method(GC_Method),
+		{ peephole__optimize(GC_Method, Instrs5, Instrs6, _) },
 
 		vn_debug__cost_header_msg("original code sequence"),
 		vn_cost__block_cost(Instrs0, Params, yes, OrigCost),
@@ -416,11 +436,11 @@ value_number__optimize_fragment_2(Instrs0, LiveMap, Params, ParEntries,
 			[]
 		;
 			vn_debug__cost_header_msg("unfiltered code sequence"),
-			vn_cost__block_cost(Instrs5, Params, yes, _)
+			vn_cost__block_cost(Instrs4, Params, yes, _)
 		),
 
 		vn_debug__cost_header_msg("new code sequence"),
-		vn_cost__block_cost(Instrs7, Params, yes, VnCost),
+		vn_cost__block_cost(Instrs6, Params, yes, VnCost),
 
 		globals__io_lookup_int_option(vn_fudge, VnFudge),
 
@@ -428,21 +448,21 @@ value_number__optimize_fragment_2(Instrs0, LiveMap, Params, ParEntries,
 			{ VnCost < OrigCost },
 			{
 				assoc_list__keys(Instrs0, Uinstrs0),
-				assoc_list__keys(Instrs7, Uinstrs7),
-				list__sublist(Uinstrs7, Uinstrs0)
+				assoc_list__keys(Instrs6, Uinstrs6),
+				list__sublist(Uinstrs6, Uinstrs0)
 			;
 				VnCost * 1000 < OrigCost * VnFudge
 			}
 		->
 			vn_debug__cost_msg(yes, OrigCost, VnCost),
-			{ vn_block__build_block_info(Instrs7, LiveMap, Params,
-				ParEntries, LabelNo0, VnTables7, Liveset7,
-				SeenIncr7, Tuple7) },
-			vn_verify__ok(Instrs7, Uinstr0, SeenIncr0, SeenIncr7,
-				Liveset0, Liveset7, VnTables0, VnTables7, OK),
+			{ vn_block__build_block_info(Instrs6, LiveMap, Params,
+				ParEntries, LabelNo0, VnTables6, Liveset6,
+				SeenIncr6, Tuple6) },
+			vn_verify__ok(Instrs6, Uinstr0, SeenIncr0, SeenIncr6,
+				Liveset0, Liveset6, VnTables0, VnTables6, OK),
 			( { OK = yes } ->
-				{ Instrs = Instrs7 },
-				{ Tuple = Tuple7 }
+				{ Instrs = Instrs6 },
+				{ Tuple = Tuple6 }
 			;
 				{ Instrs = Instrs0 },
 				{ Tuple = Tuple0 }
@@ -873,25 +893,6 @@ value_number__find_block_by_label([Block | Blocks], Label, Before, LabelBlock,
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
-:- pred value_number__convert_back_modframe(list(instruction),
-	list(instruction)).
-:- mode value_number__convert_back_modframe(in, out) is det.
-
-value_number__convert_back_modframe([], []).
-value_number__convert_back_modframe([Instr0 | Instrs0], [Instr | Instrs]) :-
-	value_number__convert_back_modframe(Instrs0, Instrs),
-	(
-		Instr0 = assign(redoip(lval(curfr)),
-			const(code_addr_const(Redoip))) - _
-	->
-		Instr = modframe(Redoip) - "recovered modframe"
-	;
-		Instr = Instr0
-	).
-
-%-----------------------------------------------------------------------------%
-%-----------------------------------------------------------------------------%
-
 :- pred value_number__push_decr_sp_back(list(instruction), list(instruction)).
 % :- mode value_number__push_decr_sp_back(di, uo) is det.
 :- mode value_number__push_decr_sp_back(in, out) is det.
@@ -1073,16 +1074,20 @@ value_number__push_livevals_back_2([Instr0 | Instrs0], Livevals, Instrs) :-
 value_number__boundary_instr(comment(_), no).
 value_number__boundary_instr(livevals(_), no).
 value_number__boundary_instr(block(_, _, _), no).
-value_number__boundary_instr(assign(_,_), no).
-value_number__boundary_instr(call(_, _, _, _), yes).
-value_number__boundary_instr(mkframe(_, _, _), yes).
-value_number__boundary_instr(modframe(_), yes).
+value_number__boundary_instr(assign(Lval,_), Boundary) :-
+	( Lval = redoip(_) ->
+		Boundary = yes
+	;
+		Boundary = no
+	).
+value_number__boundary_instr(call(_, _, _, _, _), yes).
+value_number__boundary_instr(mkframe(_, _), yes).
 value_number__boundary_instr(label(_), yes).
 value_number__boundary_instr(goto(_), yes).
 value_number__boundary_instr(computed_goto(_, _), yes).
 value_number__boundary_instr(c_code(_), yes).
 value_number__boundary_instr(if_val(_, _), yes).
-value_number__boundary_instr(incr_hp(_, _, _), no).
+value_number__boundary_instr(incr_hp(_, _, _, _), no).
 value_number__boundary_instr(mark_hp(_), no).
 value_number__boundary_instr(restore_hp(_), no).
 value_number__boundary_instr(store_ticket(_), no).
@@ -1092,7 +1097,11 @@ value_number__boundary_instr(mark_ticket_stack(_), no).
 value_number__boundary_instr(discard_tickets_to(_), no).
 value_number__boundary_instr(incr_sp(_, _), yes).
 value_number__boundary_instr(decr_sp(_), yes).
-value_number__boundary_instr(pragma_c(_, _, _, _, _), yes).
+value_number__boundary_instr(init_sync_term(_, _), no).
+value_number__boundary_instr(fork(_, _, _), yes).
+value_number__boundary_instr(join_and_terminate(_), yes).
+value_number__boundary_instr(join_and_continue(_, _), yes).
+value_number__boundary_instr(pragma_c(_, _, _, _, _, _), yes).
 
 %-----------------------------------------------------------------------------%
 
