@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -35,7 +35,7 @@
 
 :- implementation.
 
-:- import_module vn_util, vn_cost, opt_util, opt_debug.
+:- import_module builtin_ops, vn_util, vn_cost, opt_util, opt_debug.
 :- import_module map, int, string, require, std_util, assoc_list.
 
 %-----------------------------------------------------------------------------%
@@ -205,47 +205,54 @@ vn_block__handle_instr(assign(Lval, Rval),
 	vn_table__set_desired_value(Vnlval, Vn, VnTables2, VnTables),
 	vn_util__find_specials(Vnlval, LeftSpecials),
 	(
-		% Assignments of this form occur in the setup for commit.
-		% We must record the left hand side (which will be a stackvar)
-		% as a must flush location, because liveness will not force
-		% it to be flushed. The reason for this is that livemap.m
-		% may be confused by an incr_sp and assignment to a similarly
-		% numbered stackvar in later code into thinking that this
-		% assignment is redundant.
-		Rval = lval(SubLval),
-		( SubLval = curfr
-		; SubLval = maxfr
-		; SubLval = redoip(_)
-		)
+		% % Assignments of this form occur in hijacking.
+		% % We must record the left hand side (which will be a stackvar
+		% % or framevar) as a must flush location, because liveness will
+		% % not force it to be flushed. The reason for this is that
+		% % we do not properly include temporary slots (such as those
+		% % used by hijacking) in livevals annotations.
+		% Rval = lval(SubLval),
+		% ( SubLval = curfr
+		% ; SubLval = maxfr
+		% ; SubLval = redoip(_)
+		% ; SubLval = redofr(_)
+		% )
+
+		% Before we leave a disjunction through a non-last disjunct,
+		% the code generator makes sure that the variables needed
+		% on resumption at the next disjunct are in their stack slots.
+		% The resumption point may be reached via a redo() or fail()
+		% operation. Unfortunately, livemap.m does not recognize that
+		% these stack variables are live at points that perform a
+		% redo() or fail() operation. We compensate here by ensuring
+		% that assignments to stack variables are not removed by
+		% value numbering.
+		%
+		% The condition of this test subsumes the condition that used
+		% to be here, which is therefore commented out above.
+		( Lval = stackvar(_) ; Lval = framevar(_) )
 	->
 		Specials = [Vnlval | LeftSpecials]
 	;
 		Specials = LeftSpecials
 	),
 	set__insert_list(Liveset0, Specials, Liveset).
-vn_block__handle_instr(call(Proc, Return, Info, CallModel),
+vn_block__handle_instr(call(Proc, Return, Info, Context, CallModel),
 		Livemap, Params, VnTables0, VnTables, Liveset0, Liveset,
 		SeenIncr, SeenIncr, Tuple0, Tuple) :-
-	vn_block__new_ctrl_node(vn_call(Proc, Return, Info, CallModel), Livemap,
-		Params, VnTables0, VnTables,
+	vn_block__new_ctrl_node(vn_call(Proc, Return, Info, Context,
+		CallModel), Livemap, Params, VnTables0, VnTables,
 		Liveset0, Liveset, Tuple0, Tuple).
-vn_block__handle_instr(mkframe(Name, Size, Redoip), Livemap, Params,
+vn_block__handle_instr(mkframe(NondetFrameInfo, Redoip), Livemap, Params,
 		VnTables0, VnTables, Liveset0, Liveset,
 		SeenIncr0, SeenIncr, Tuple0, Tuple) :-
-	vn_block__new_ctrl_node(vn_mkframe(Name, Size, Redoip),
+	vn_block__new_ctrl_node(vn_mkframe(NondetFrameInfo, Redoip),
 		Livemap, Params, VnTables0, VnTables1,
 		Liveset0, Liveset1, Tuple0, Tuple1),
 	vn_block__handle_instr(assign(redoip(lval(maxfr)),
 		const(code_addr_const(Redoip))),
 		Livemap, Params, VnTables1, VnTables, Liveset1, Liveset,
 		SeenIncr0, SeenIncr, Tuple1, Tuple).
-vn_block__handle_instr(modframe(Redoip), Livemap, Params,
-		VnTables0, VnTables, Liveset0, Liveset,
-		SeenIncr0, SeenIncr, Tuple0, Tuple) :-
-	vn_block__handle_instr(assign(redoip(lval(curfr)),
-		const(code_addr_const(Redoip))),
-		Livemap, Params, VnTables0, VnTables, Liveset0, Liveset,
-		SeenIncr0, SeenIncr, Tuple0, Tuple).
 vn_block__handle_instr(label(Label),
 		Livemap, Params, VnTables0, VnTables, Liveset0, Liveset,
 		SeenIncr, SeenIncr, Tuple0, Tuple) :-
@@ -274,7 +281,7 @@ vn_block__handle_instr(if_val(Rval, Target),
 	vn_block__new_ctrl_node(vn_if_val(Vn, Target), Livemap,
 		Params, VnTables1, VnTables,
 		Liveset0, Liveset, Tuple0, Tuple).
-vn_block__handle_instr(incr_hp(Lval, MaybeTag, Rval),
+vn_block__handle_instr(incr_hp(Lval, MaybeTag, Rval, _),
 		Livemap, Params, VnTables0, VnTables, Liveset0, Liveset,
 		_SeenIncr, SeenIncr, Tuple0, Tuple) :-
 	(
@@ -353,10 +360,26 @@ vn_block__handle_instr(decr_sp(N),
 	vn_block__new_ctrl_node(vn_decr_sp(N), Livemap,
 		Params, VnTables0, VnTables,
 		Liveset0, Liveset, Tuple0, Tuple).
-vn_block__handle_instr(pragma_c(_, _, _, _, _),
+vn_block__handle_instr(pragma_c(_, _, _, _, _, _),
 		_Livemap, _Params, VnTables, VnTables, Liveset, Liveset,
 		SeenIncr, SeenIncr, Tuple, Tuple) :-
 	error("value numbering not supported for pragma_c").
+vn_block__handle_instr(init_sync_term(_, _),
+		_Livemap, _Params, VnTables, VnTables, Liveset, Liveset,
+		SeenIncr, SeenIncr, Tuple, Tuple) :-
+	error("value numbering not supported for init_sync_term").
+vn_block__handle_instr(fork(_, _, _),
+		_Livemap, _Params, VnTables, VnTables, Liveset, Liveset,
+		SeenIncr, SeenIncr, Tuple, Tuple) :-
+	error("value numbering not supported for fork").
+vn_block__handle_instr(join_and_terminate(_),
+		_Livemap, _Params, VnTables, VnTables, Liveset, Liveset,
+		SeenIncr, SeenIncr, Tuple, Tuple) :-
+	error("value numbering not supported for join_and_terminate").
+vn_block__handle_instr(join_and_continue(_, _),
+		_Livemap, _Params, VnTables, VnTables, Liveset, Liveset,
+		SeenIncr, SeenIncr, Tuple, Tuple) :-
+	error("value numbering not supported for join_and_continue").
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
@@ -382,13 +405,13 @@ vn_block__new_ctrl_node(VnInstr, Livemap, Params,
 		LabelNo = LabelNo0,
 		Parallels = []
 	;
-		VnInstr = vn_call(_, _, _, _),
+		VnInstr = vn_call(_, _, _, _, _),
 		vn_block__record_at_call(VnTables0, VnTables, Liveset0, Liveset,
 			FlushEntry0, FlushEntry),
 		LabelNo = LabelNo0,
 		Parallels = []
 	;
-		VnInstr = vn_mkframe(_, _, _),
+		VnInstr = vn_mkframe(_, _),
 		VnTables = VnTables0,
 		Liveset = Liveset0,
 		FlushEntry = FlushEntry0,
@@ -784,6 +807,7 @@ vn_block__record_compulsory_lval_list([Vnlval - Vn | Lval_vn_list],
 	(
 		( Vnlval = vn_field(_, _, _)
 		; Vnlval = vn_redoip(_)
+		; Vnlval = vn_redofr(_)
 		; Vnlval = vn_framevar(_)
 		; Vnlval = vn_curfr
 		; Vnlval = vn_maxfr
@@ -873,15 +897,14 @@ vn_block__is_ctrl_instr(comment(_), no).
 vn_block__is_ctrl_instr(livevals(_), yes).
 vn_block__is_ctrl_instr(block(_, _, _), no).
 vn_block__is_ctrl_instr(assign(_, _), no).
-vn_block__is_ctrl_instr(call(_, _, _, _), yes).
-vn_block__is_ctrl_instr(mkframe(_, _, _), yes).
-vn_block__is_ctrl_instr(modframe(_), no).
+vn_block__is_ctrl_instr(call(_, _, _, _, _), yes).
+vn_block__is_ctrl_instr(mkframe(_, _), yes).
 vn_block__is_ctrl_instr(label(_), yes).
 vn_block__is_ctrl_instr(goto(_), yes).
 vn_block__is_ctrl_instr(computed_goto(_, _), yes).
 vn_block__is_ctrl_instr(c_code(_), no).
 vn_block__is_ctrl_instr(if_val(_, _), yes).
-vn_block__is_ctrl_instr(incr_hp(_, _, _), no).
+vn_block__is_ctrl_instr(incr_hp(_, _, _, _), no).
 vn_block__is_ctrl_instr(mark_hp(_), yes).
 vn_block__is_ctrl_instr(restore_hp(_), yes).
 vn_block__is_ctrl_instr(store_ticket(_), yes).
@@ -891,7 +914,11 @@ vn_block__is_ctrl_instr(mark_ticket_stack(_), yes).
 vn_block__is_ctrl_instr(discard_tickets_to(_), yes).
 vn_block__is_ctrl_instr(incr_sp(_, _), yes).
 vn_block__is_ctrl_instr(decr_sp(_), yes).
-vn_block__is_ctrl_instr(pragma_c(_, _, _, _, _), no).
+vn_block__is_ctrl_instr(init_sync_term(_, _), no).
+vn_block__is_ctrl_instr(fork(_, _, _), yes).
+vn_block__is_ctrl_instr(join_and_terminate(_), yes).
+vn_block__is_ctrl_instr(join_and_continue(_, _), yes).
+vn_block__is_ctrl_instr(pragma_c(_, _, _, _, _, _), no).
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
