@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -29,7 +29,7 @@
 
 :- implementation.
 
-:- import_module vn_debug.
+:- import_module builtin_ops, vn_debug.
 :- import_module require, string, std_util, int.
 
 vn_cost__block_cost(Instr, Params, PrintInstr, Cost) -->
@@ -108,13 +108,10 @@ vn_cost__instr_cost(Uinstr, Params, Cost) :-
 			Cost is RvalCost + LvalCost + AssignCost
 		)
 	;
-		Uinstr = call(_, _, _, _),
+		Uinstr = call(_, _, _, _, _),
 		Cost = 0
 	;
-		Uinstr = mkframe(_, _, _),
-		Cost = 0
-	;
-		Uinstr = modframe(_),
+		Uinstr = mkframe(_, _),
 		Cost = 0
 	;
 		Uinstr = label(_),
@@ -134,7 +131,7 @@ vn_cost__instr_cost(Uinstr, Params, Cost) :-
 		vn_cost__rval_cost(Rval, Params, RvalCost),
 		Cost = RvalCost
 	;
-		Uinstr = incr_hp(Lval, MaybeTag, Rval),
+		Uinstr = incr_hp(Lval, MaybeTag, Rval, _),
 		vn_type__costof_assign(Params, AssignCost),
 		vn_cost__lval_cost(Lval, Params, LvalCost),
 		vn_cost__rval_cost(Rval, Params, RvalCost),
@@ -181,8 +178,20 @@ vn_cost__instr_cost(Uinstr, Params, Cost) :-
 		Uinstr = decr_sp(_),
 		Cost = 0
 	;
-		Uinstr = pragma_c(_, _, _, _, _),
+		Uinstr = pragma_c(_, _, _, _, _, _),
 		error("pragma_c found in vn_block_cost")
+	;
+		Uinstr = init_sync_term(_, _),
+		error("init_sync_term found in vn_block_cost")
+	;
+		Uinstr = fork(_, _, _),
+		error("fork found in vn_block_cost")
+	;
+		Uinstr = join_and_terminate(_),
+		error("join_and_terminate found in vn_block_cost")
+	;
+		Uinstr = join_and_continue(_, _),
+		error("join_and_continue found in vn_block_cost")
 	).
 
 vn_cost__lval_cost(Lval, Params, Cost) :-
@@ -242,6 +251,11 @@ vn_cost__lval_cost(Lval, Params, Cost) :-
 		vn_type__costof_stackref(Params, StackrefCost),
 		vn_cost__rval_cost(Rval, Params, RvalCost),
 		Cost is RvalCost + StackrefCost
+	;
+		Lval = redofr(Rval),
+		vn_type__costof_stackref(Params, StackrefCost),
+		vn_cost__rval_cost(Rval, Params, RvalCost),
+		Cost is RvalCost + StackrefCost
 	;
 		Lval = redoip(Rval),
 		vn_type__costof_stackref(Params, StackrefCost),
@@ -292,7 +306,7 @@ vn_cost__rval_cost(Rval, Params, Cost) :-
 		Rval = var(_),
 		error("var found in rval_cost")
 	;
-		Rval = create(_, _, _, _),
+		Rval = create(_, _, _, _, _, _),
 		Cost = 0
 	;
 		Rval = mkword(_, Rval1),
