@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1997 The University of Melbourne.
+% Copyright (C) 1996-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -25,7 +25,7 @@
 
 :- implementation.
 
-:- import_module opt_util.
+:- import_module code_util, opt_util.
 :- import_module require, std_util.
 
 	% Look for assignments to temp variables. If possible and profitable,
@@ -38,7 +38,7 @@ vn_filter__block([Instr0 | Instrs0], Instrs) :-
 		Instr0 = Uinstr0 - _,
 		Uinstr0 = assign(Temp, Defn),
 		Temp = temp(_, _),
-		opt_util__lvals_in_rval(Defn, Deps),
+		code_util__lvals_in_rval(Defn, Deps),
 		vn_filter__can_substitute(Instrs0, Temp, Defn, Deps,
 			Instrs1)
 	->
@@ -73,7 +73,7 @@ vn_filter__can_substitute([Instr0 | Instrs0], Temp, Defn, Deps, Instrs) :-
 	Instr0 = Uinstr0 - Comment,
 	(
 		vn_filter__user_instr(Uinstr0, yes(Rval)),
-		opt_util__lvals_in_rval(Rval, Lvals),
+		code_util__lvals_in_rval(Rval, Lvals),
 		list__delete_first(Lvals, Temp, OtherLvals)
 	->
 		% We don't want to perform the subsitution
@@ -81,7 +81,7 @@ vn_filter__can_substitute([Instr0 | Instrs0], Temp, Defn, Deps, Instrs) :-
 		\+ list__member(Temp, OtherLvals),
 		\+ (
 			vn_filter__defining_instr(Uinstr0, yes(Lval)),
-			opt_util__lvals_in_lval(Lval, AccessLvals),
+			code_util__lvals_in_lval(Lval, AccessLvals),
 			list__member(Temp, AccessLvals)
 		),
 		vn_filter__replace_in_user_instr(Uinstr0, Temp, Defn, Uinstr1),
@@ -99,7 +99,7 @@ vn_filter__can_substitute([Instr0 | Instrs0], Temp, Defn, Deps, Instrs) :-
 		->
 			fail
 		;
-			opt_util__lvals_in_lval(Lval, AccessLvals),
+			code_util__lvals_in_lval(Lval, AccessLvals),
 			list__delete_first(AccessLvals, Temp, OtherAccessLvals)
 		->
 			\+ list__member(Temp, OtherAccessLvals),
@@ -135,16 +135,15 @@ vn_filter__user_instr(livevals(_), no).
 vn_filter__user_instr(block(_, _, _), _):-
 	error("inappropriate instruction in vn__filter").
 vn_filter__user_instr(assign(_, Rval), yes(Rval)).
-vn_filter__user_instr(call(_, _, _, _), no).
-vn_filter__user_instr(mkframe(_, _, _), no).
-vn_filter__user_instr(modframe(_), no).
+vn_filter__user_instr(call(_, _, _, _, _), no).
+vn_filter__user_instr(mkframe(_, _), no).
 vn_filter__user_instr(label(_), no).
 vn_filter__user_instr(goto(_), no).
 vn_filter__user_instr(computed_goto(Rval, _), yes(Rval)).
 vn_filter__user_instr(if_val(Rval, _), yes(Rval)).
 vn_filter__user_instr(c_code(_), _):-
 	error("inappropriate instruction in vn__filter").
-vn_filter__user_instr(incr_hp(_, _, Rval), yes(Rval)).
+vn_filter__user_instr(incr_hp(_, _, Rval, _), yes(Rval)).
 vn_filter__user_instr(mark_hp(_), no).
 vn_filter__user_instr(restore_hp(Rval), yes(Rval)).
 vn_filter__user_instr(store_ticket(_), no).
@@ -154,8 +153,16 @@ vn_filter__user_instr(mark_ticket_stack(_), no).
 vn_filter__user_instr(discard_tickets_to(Rval), yes(Rval)).
 vn_filter__user_instr(incr_sp(_, _), no).
 vn_filter__user_instr(decr_sp(_), no).
-vn_filter__user_instr(pragma_c(_, _, _, _, _), _):-
+vn_filter__user_instr(pragma_c(_, _, _, _, _, _), _):-
 	error("inappropriate instruction in vn__filter").
+vn_filter__user_instr(init_sync_term(_, _), _):-
+	error("init_sync_term instruction in vn__filter").
+vn_filter__user_instr(fork(_, _, _), _):-
+	error("fork instruction in vn__filter").
+vn_filter__user_instr(join_and_terminate(_), _):-
+	error("join_and_terminate instruction in vn__filter").
+vn_filter__user_instr(join_and_continue(_, _), _):-
+	error("join_and_continue instruction in vn__filter").
 
 	% vn_filter__replace_in_user_instr(Instr0, Old, New, Instr):
 	% Given that Instr0 refers to the values of some locations,
@@ -174,11 +181,9 @@ vn_filter__replace_in_user_instr(block(_, _, _), _, _, _):-
 vn_filter__replace_in_user_instr(assign(Lval, Rval0), Temp, Defn,
 		assign(Lval, Rval)) :-
 	vn_filter__replace_in_rval(Rval0, Temp, Defn, Rval).
-vn_filter__replace_in_user_instr(call(_, _, _, _), _, _, _) :-
+vn_filter__replace_in_user_instr(call(_, _, _, _, _), _, _, _) :-
 	error("non-user instruction in vn_filter__replace_in_user_instr").
-vn_filter__replace_in_user_instr(mkframe(_, _, _), _, _, _) :-
-	error("non-user instruction in vn_filter__replace_in_user_instr").
-vn_filter__replace_in_user_instr(modframe(_), _, _, _) :-
+vn_filter__replace_in_user_instr(mkframe(_, _), _, _, _) :-
 	error("non-user instruction in vn_filter__replace_in_user_instr").
 vn_filter__replace_in_user_instr(label(_), _, _, _) :-
 	error("non-user instruction in vn_filter__replace_in_user_instr").
@@ -192,8 +197,8 @@ vn_filter__replace_in_user_instr(if_val(Rval0, Label), Temp, Defn,
 	vn_filter__replace_in_rval(Rval0, Temp, Defn, Rval).
 vn_filter__replace_in_user_instr(c_code(_), _, _, _):-
 	error("inappropriate instruction in vn__filter").
-vn_filter__replace_in_user_instr(incr_hp(Lval, Tag, Rval0), Temp, Defn,
-		incr_hp(Lval, Tag, Rval)) :-
+vn_filter__replace_in_user_instr(incr_hp(Lval, Tag, Rval0, Msg), Temp, Defn,
+		incr_hp(Lval, Tag, Rval, Msg)) :-
 	vn_filter__replace_in_rval(Rval0, Temp, Defn, Rval).
 vn_filter__replace_in_user_instr(mark_hp(_), _, _, _) :-
 	error("non-user instruction in vn_filter__replace_in_user_instr").
@@ -216,8 +221,16 @@ vn_filter__replace_in_user_instr(incr_sp(_, _), _, _, _) :-
 	error("non-user instruction in vn_filter__replace_in_user_instr").
 vn_filter__replace_in_user_instr(decr_sp(_), _, _, _) :-
 	error("non-user instruction in vn_filter__replace_in_user_instr").
-vn_filter__replace_in_user_instr(pragma_c(_, _, _, _, _), _, _, _):-
+vn_filter__replace_in_user_instr(pragma_c(_, _, _, _, _, _), _, _, _):-
 	error("inappropriate instruction in vn__filter").
+vn_filter__replace_in_user_instr(init_sync_term(_, _), _, _, _):-
+	error("init_sync_term instruction in vn__filter").
+vn_filter__replace_in_user_instr(fork(_, _, _), _, _, _):-
+	error("fork instruction in vn__filter").
+vn_filter__replace_in_user_instr(join_and_terminate(_), _, _, _):-
+	error("join_and_terminate instruction in vn__filter").
+vn_filter__replace_in_user_instr(join_and_continue(_, _), _, _, _):-
+	error("join_and_continue instruction in vn__filter").
 
 	% Check whether this instruction defines the value of any lval.
 
@@ -229,16 +242,15 @@ vn_filter__defining_instr(livevals(_), no).
 vn_filter__defining_instr(block(_, _, _), _):-
 	error("inappropriate instruction in vn__filter").
 vn_filter__defining_instr(assign(Lval, _), yes(Lval)).
-vn_filter__defining_instr(call(_, _, _, _), no).
-vn_filter__defining_instr(mkframe(_, _, _), no).
-vn_filter__defining_instr(modframe(_), no).
+vn_filter__defining_instr(call(_, _, _, _, _), no).
+vn_filter__defining_instr(mkframe(_, _), no).
 vn_filter__defining_instr(label(_), no).
 vn_filter__defining_instr(goto(_), no).
 vn_filter__defining_instr(computed_goto(_, _), no).
 vn_filter__defining_instr(if_val(_, _), no).
 vn_filter__defining_instr(c_code(_), _):-
 	error("inappropriate instruction in vn__filter").
-vn_filter__defining_instr(incr_hp(Lval, _, _), yes(Lval)).
+vn_filter__defining_instr(incr_hp(Lval, _, _, _), yes(Lval)).
 vn_filter__defining_instr(mark_hp(Lval), yes(Lval)).
 vn_filter__defining_instr(restore_hp(_), no).
 vn_filter__defining_instr(store_ticket(Lval), yes(Lval)).
@@ -248,8 +260,16 @@ vn_filter__defining_instr(mark_ticket_stack(Lval), yes(Lval)).
 vn_filter__defining_instr(discard_tickets_to(_), no).
 vn_filter__defining_instr(incr_sp(_, _), no).
 vn_filter__defining_instr(decr_sp(_), no).
-vn_filter__defining_instr(pragma_c(_, _, _, _, _), _):-
+vn_filter__defining_instr(pragma_c(_, _, _, _, _, _), _):-
 	error("inappropriate instruction in vn__filter").
+vn_filter__defining_instr(init_sync_term(_, _), _):-
+	error("init_sync_term instruction in vn__filter").
+vn_filter__defining_instr(fork(_, _, _), _):-
+	error("fork instruction in vn__filter").
+vn_filter__defining_instr(join_and_terminate(_), _):-
+	error("join_and_terminate instruction in vn__filter").
+vn_filter__defining_instr(join_and_continue(_, _), _):-
+	error("join_and_continue instruction in vn__filter").
 
 	% vn_filter__replace_in_defining_instr(Instr0, Old, New, Instr):
 	% Given that Instr0 defines the value of a location,
@@ -268,11 +288,9 @@ vn_filter__replace_in_defining_instr(block(_, _, _), _, _, _):-
 vn_filter__replace_in_defining_instr(assign(Lval0, Rval), Temp, Defn,
 		assign(Lval, Rval)) :-
 	vn_filter__replace_in_lval(Lval0, Temp, Defn, Lval).
-vn_filter__replace_in_defining_instr(call(_, _, _, _), _, _, _) :-
+vn_filter__replace_in_defining_instr(call(_, _, _, _, _), _, _, _) :-
 	error("non-def instruction in vn_filter__replace_in_defining_instr").
-vn_filter__replace_in_defining_instr(mkframe(_, _, _), _, _, _) :-
-	error("non-def instruction in vn_filter__replace_in_defining_instr").
-vn_filter__replace_in_defining_instr(modframe(_), _, _, _) :-
+vn_filter__replace_in_defining_instr(mkframe(_, _), _, _, _) :-
 	error("non-def instruction in vn_filter__replace_in_defining_instr").
 vn_filter__replace_in_defining_instr(label(_), _, _, _) :-
 	error("non-def instruction in vn_filter__replace_in_defining_instr").
@@ -284,8 +302,8 @@ vn_filter__replace_in_defining_instr(if_val(_, _), _, _, _) :-
 	error("non-def instruction in vn_filter__replace_in_defining_instr").
 vn_filter__replace_in_defining_instr(c_code(_), _, _, _):-
 	error("inappropriate instruction in vn__filter").
-vn_filter__replace_in_defining_instr(incr_hp(Lval0, Tag, Rval), Temp, Defn,
-		incr_hp(Lval, Tag, Rval)) :-
+vn_filter__replace_in_defining_instr(incr_hp(Lval0, Tag, Rval, Msg), Temp, Defn,
+		incr_hp(Lval, Tag, Rval, Msg)) :-
 	vn_filter__replace_in_lval(Lval0, Temp, Defn, Lval).
 vn_filter__replace_in_defining_instr(mark_hp(Lval0), Temp, Defn,
 		mark_hp(Lval)) :-
@@ -308,7 +326,15 @@ vn_filter__replace_in_defining_instr(incr_sp(_, _), _, _, _) :-
 	error("non-def instruction in vn_filter__replace_in_defining_instr").
 vn_filter__replace_in_defining_instr(decr_sp(_), _, _, _) :-
 	error("non-def instruction in vn_filter__replace_in_defining_instr").
-vn_filter__replace_in_defining_instr(pragma_c(_, _, _, _, _), _, _, _):-
+vn_filter__replace_in_defining_instr(init_sync_term(_, _), _, _, _):-
+	error("init_sync_term instruction in vn_filter__replace_in_defining_instr").
+vn_filter__replace_in_defining_instr(fork(_, _, _), _, _, _):-
+	error("fork instruction in vn_filter__replace_in_defining_instr").
+vn_filter__replace_in_defining_instr(join_and_terminate(_), _, _, _):-
+	error("join_and_terminate instruction in vn_filter__replace_in_defining_instr").
+vn_filter__replace_in_defining_instr(join_and_continue(_, _), _, _, _):-
+	error("join_and_continue instruction in vn_filter__replace_in_defining_instr").
+vn_filter__replace_in_defining_instr(pragma_c(_, _, _, _, _, _), _, _, _):-
 	error("inappropriate instruction in vn__filter").
 
 	% vn_filter__replace_in_lval(Lval0, Old, New, Lval):
@@ -328,6 +354,8 @@ vn_filter__replace_in_lval(succip(Rval0), Temp, Defn, succip(Rval)) :-
 	vn_filter__replace_in_rval(Rval0, Temp, Defn, Rval).
 vn_filter__replace_in_lval(redoip(Rval0), Temp, Defn, redoip(Rval)) :-
 	vn_filter__replace_in_rval(Rval0, Temp, Defn, Rval).
+vn_filter__replace_in_lval(redofr(Rval0), Temp, Defn, redofr(Rval)) :-
+	vn_filter__replace_in_rval(Rval0, Temp, Defn, Rval).
 vn_filter__replace_in_lval(succfr(Rval0), Temp, Defn, succfr(Rval)) :-
 	vn_filter__replace_in_rval(Rval0, Temp, Defn, Rval).
 vn_filter__replace_in_lval(prevfr(Rval0), Temp, Defn, prevfr(Rval)) :-
@@ -360,8 +388,8 @@ vn_filter__replace_in_rval(lval(Lval0), Temp, Defn, Rval) :-
 	).
 vn_filter__replace_in_rval(var(_), _, _, _) :-
 	error("found var in vn_filter__replace_in_rval").
-vn_filter__replace_in_rval(create(Tag, Args, Unique, Label), _, _,
-		create(Tag, Args, Unique, Label)).
+vn_filter__replace_in_rval(create(Tag, Args, AT, StatDyn, Label, Msg), _, _,
+		create(Tag, Args, AT, StatDyn, Label, Msg)).
 vn_filter__replace_in_rval(mkword(Tag, Rval0), Temp, Defn, mkword(Tag, Rval)) :-
 	vn_filter__replace_in_rval(Rval0, Temp, Defn, Rval).
 vn_filter__replace_in_rval(const(Const), _, _, const(Const)).
