@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -31,7 +31,7 @@
 
 :- import_module bool, map, int, string, require, std_util.
 
-:- import_module vn_table, vn_util, vn_debug, opt_debug.
+:- import_module builtin_ops, vn_table, vn_util, vn_debug, opt_debug.
 
 vn_flush__nodelist([], _, _, _, _, []) --> [].
 vn_flush__nodelist([Node0 | Nodes0], Ctrlmap, VnTables0, Templocs0, Params,
@@ -185,12 +185,14 @@ vn_flush__ctrl_node(Vn_instr, N, VnTables0, VnTables, Templocs0, Templocs,
 		Templocs = Templocs0,
 		Instrs = [livevals(Livevals) - ""]
 	;
-		Vn_instr = vn_call(ProcAddr, RetAddr, LiveInfo, CodeModel),
+		Vn_instr = vn_call(ProcAddr, RetAddr, LiveInfo,
+			Context, CodeModel),
 		VnTables = VnTables0,
 		Templocs = Templocs0,
-		Instrs = [call(ProcAddr, RetAddr, LiveInfo, CodeModel) - ""]
+		Instrs = [call(ProcAddr, RetAddr, LiveInfo,
+			Context, CodeModel) - ""]
 	;
-		Vn_instr = vn_mkframe(Name, Size, Redoip),
+		Vn_instr = vn_mkframe(NondetFrameInfo, Redoip),
 		vn_util__rval_to_vn(const(code_addr_const(Redoip)), AddrVn,
 			VnTables0, VnTables1),
 		vn_util__lval_to_vnlval(redoip(lval(maxfr)), SlotVnlval,
@@ -198,7 +200,7 @@ vn_flush__ctrl_node(Vn_instr, N, VnTables0, VnTables, Templocs0, Templocs,
 		vn_table__set_current_value(SlotVnlval, AddrVn,
 			VnTables2, VnTables),
 		Templocs = Templocs0,
-		Instrs = [mkframe(Name, Size, Redoip) - ""]
+		Instrs = [mkframe(NondetFrameInfo, Redoip) - ""]
 	;
 		Vn_instr = vn_label(Label),
 		VnTables = VnTables0,
@@ -527,7 +529,7 @@ vn_flush__generate_assignment(Vnlval, Vn, Forbidden0, VnTables0, VnTables,
 vn_flush__get_incr_hp([], _, _) :-
 	error("could not find incr_hp").
 vn_flush__get_incr_hp([Instr0 | Instrs0], IncrHp, Instrs) :-
-	( Instr0 = incr_hp(_, _, _) - _ ->
+	( Instr0 = incr_hp(_, _, _, _) - _ ->
 		IncrHp = Instr0,
 		Instrs = Instrs0
 	;
@@ -689,8 +691,9 @@ vn_flush__vn_value(Vn, Srcs, Forbidden, Rval, VnTables0, VnTables,
 		Templocs = Templocs0,
 		Instrs = []
 	;
-		Vnrval = vn_create(Tag, MaybeRvals, Unique, Label),
-		Rval = create(Tag, MaybeRvals, Unique, Label),
+		Vnrval = vn_create(Tag, MaybeRvals, ArgTypes, StatDyn,
+			Label, Msg),
+		Rval = create(Tag, MaybeRvals, ArgTypes, StatDyn, Label, Msg),
 		VnTables = VnTables0,
 		Templocs = Templocs0,
 		Instrs = []
@@ -847,7 +850,8 @@ vn_flush__old_hp(Srcs0, Forbidden0, ReturnRval, VnTables0, VnTables,
 	),
 
 	vn_table__set_current_value(Vnlval, AssignedVn, VnTables4, VnTables),
-	Instr = incr_hp(Lval, MaybeTag, Rval) - "",
+	Instr = incr_hp(Lval, MaybeTag, Rval, "origin_lost_in_value_number")
+		- "",
 	list__condense([IncrInstrs, SaveInstrs, [Instr]], Instrs).
 
 %-----------------------------------------------------------------------------%
@@ -892,7 +896,7 @@ vn_flush__hp_incr(Vn, Srcs, Forbidden, MaybeRval, VnTables0, VnTables,
 			Templocs = Templocs0,
 			Instrs = []
 		;
-			Vnrval = vn_create(_, _, _, _),
+			Vnrval = vn_create(_, _, _, _, _, _),
 			error("create in calculation of new hp")
 		;
 			Vnrval = vn_unop(_, _),
@@ -1041,6 +1045,12 @@ vn_flush__access_path(Vnlval, Srcs, Forbidden, Lval, VnTables0, VnTables,
 			VnTables0, VnTables,
 			Templocs0, Templocs, Params, AccessInstrs),
 		Lval = prevfr(Rval)
+	;
+		Vnlval = vn_redofr(Vn1),
+		vn_flush__vn(Vn1, [src_access(Vnlval) | Srcs], Forbidden, Rval,
+			VnTables0, VnTables,
+			Templocs0, Templocs, Params, AccessInstrs),
+		Lval = redofr(Rval)
 	;
 		Vnlval = vn_redoip(Vn1),
 		vn_flush__vn(Vn1, [src_access(Vnlval) | Srcs], Forbidden, Rval,
