@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -13,8 +13,8 @@
 :- module vn_type.
 
 :- interface.
-:- import_module llds, livemap, options.
-:- import_module bool, getopt, map, set, list, std_util.
+:- import_module llds, builtin_ops, livemap, options.
+:- import_module getopt, map, set, list, std_util.
 
 :- type vn == int.
 
@@ -27,11 +27,12 @@
 			;	vn_curfr
 			;	vn_succfr(vn)
 			;	vn_prevfr(vn)
+			;	vn_redofr(vn)
 			;	vn_redoip(vn)
 			;	vn_succip(vn)
 			;	vn_hp
 			;	vn_sp
-			;	vn_field(tag, vn, vn)		% lval
+			;	vn_field(maybe(tag), vn, vn)
 			;	vn_mem_ref(vn).
 
 			% these lvals do not have vnlval parallels
@@ -40,7 +41,9 @@
 :- type vnrval		--->	vn_origlval(vnlval)
 			;	vn_mkword(tag, vn)
 			;	vn_const(rval_const)
-			;	vn_create(tag, list(maybe(rval)), bool, int)
+			;	vn_create(tag, list(maybe(rval)),
+					create_arg_types, static_or_dynamic,
+					int, string)
 			;	vn_unop(unary_op, vn)
 			;	vn_binop(binary_op, vn, vn)
 			;	vn_stackvar_addr(int)
@@ -67,9 +70,9 @@
 			;	node_ctrl(int).
 
 :- type vn_instr	--->	vn_livevals(lvalset)
-			;	vn_call(code_addr, code_addr,
-					list(liveinfo), call_model)
-			;	vn_mkframe(string, int, code_addr)
+			;	vn_call(code_addr, code_addr, list(liveinfo),
+					term__context, call_model)
+			;	vn_mkframe(nondet_frame_info, code_addr)
 			;	vn_label(label)
 			;	vn_goto(code_addr)
 			;	vn_computed_goto(vn, list(label))
@@ -175,7 +178,7 @@ vn_type__costof_heapref(vn_params(_, _, _, _, _, _, _, _, HeapCost),
 
 vn_type__vnrval_type(vn_origlval(Lval), Type) :-
 	vn_type__vnlval_type(Lval, Type).
-vn_type__vnrval_type(vn_create(_, _, _, _), data_ptr).
+vn_type__vnrval_type(vn_create(_, _, _, _, _, _), data_ptr).
 vn_type__vnrval_type(vn_mkword(_, _), data_ptr). % see comment in llds.m
 vn_type__vnrval_type(vn_const(Const), Type) :-
 	llds__const_type(Const, Type).
@@ -200,6 +203,7 @@ vn_type__vnlval_type(vn_stackvar(_), word).
 vn_type__vnlval_type(vn_framevar(_), word).
 vn_type__vnlval_type(vn_succip(_), code_ptr).
 vn_type__vnlval_type(vn_redoip(_), code_ptr).
+vn_type__vnlval_type(vn_redofr(_), data_ptr).
 vn_type__vnlval_type(vn_succfr(_), data_ptr).
 vn_type__vnlval_type(vn_prevfr(_), data_ptr).
 vn_type__vnlval_type(vn_field(_, _, _), word).
