@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -14,7 +14,7 @@
 
 :- interface.
 
-:- import_module list.
+:- import_module list, bool, std_util.
 :- import_module llds, vn_type, vn_table.
 
 %-----------------------------------------------------------------------------%
@@ -33,9 +33,9 @@
 
 :- implementation.
 
-:- import_module bool, int, float, string, set, map, require, std_util.
+:- import_module int, float, string, set, map, require.
 
-:- import_module opt_util.
+:- import_module builtin_ops, opt_util.
 
 vn_util__find_specials(vn_reg(_, _), []).
 vn_util__find_specials(vn_temp(_, _), []).
@@ -44,10 +44,11 @@ vn_util__find_specials(vn_framevar(_), []).
 vn_util__find_specials(vn_succip, [vn_succip]).
 vn_util__find_specials(vn_maxfr, [vn_maxfr]).
 vn_util__find_specials(vn_curfr, [vn_curfr]).
+vn_util__find_specials(vn_prevfr(Vn), [vn_prevfr(Vn)]).
 vn_util__find_specials(vn_redoip(Vn), [vn_redoip(Vn)]).
+vn_util__find_specials(vn_redofr(Vn), [vn_redofr(Vn)]).
 vn_util__find_specials(vn_succip(Vn), [vn_succip(Vn)]).
 vn_util__find_specials(vn_succfr(Vn), [vn_succfr(Vn)]).
-vn_util__find_specials(vn_prevfr(Vn), [vn_prevfr(Vn)]).
 vn_util__find_specials(vn_hp, [vn_hp]).
 vn_util__find_specials(vn_sp, [vn_sp]).
 vn_util__find_specials(vn_field(_, _, _), []).
@@ -147,9 +148,9 @@ vn_util__rval_to_vn(Rval, Vn, VnTables0, VnTables) :-
 		Rval = var(_),
 		error("value_number should never get rval: var")
 	;
-		Rval = create(Tag, Args, Unique, Label),
-		vn_util__vnrval_to_vn(vn_create(Tag, Args, Unique, Label), Vn,
-			VnTables0, VnTables)
+		Rval = create(Tag, Args, ArgTypes, StatDyn, Label, Msg),
+		vn_util__vnrval_to_vn(vn_create(Tag, Args, ArgTypes,
+			StatDyn, Label, Msg), Vn, VnTables0, VnTables)
 	;
 		Rval = mkword(Tag, Rval1),
 		vn_util__rval_to_vn(Rval1, SubVn, VnTables0, VnTables1),
@@ -243,6 +244,10 @@ vn_util__vnrval_may_share_vn(Vnrval) :-
 
 	% XXX more simplification opportunities exist
 
+	% XXX we should check for overflow
+
+	% XXX some of this code should be shared with const_prop.m
+
 :- pred vn_util__simplify_vnrval(vnrval, vnrval, vn_tables, vn_tables).
 % :- mode vn_util__simplify_vnrval(in, out, di, uo) is det.
 :- mode vn_util__simplify_vnrval(in, out, in, out) is det.
@@ -659,7 +664,7 @@ vn_util__simplify_vnrval_binop(Binop, Vn1, Vn2, Vnrval, VnTables0, VnTables) :-
 			Vnrval1 = vn_const(int_const(I1)),
 			Vnrval2 = vn_const(int_const(I2))
 		->
-			I is I1 >> I2,
+			I = unchecked_right_shift(I1, I2),
 			Vnrval = vn_const(int_const(I)),
 			VnTables = VnTables0
 		;
@@ -678,7 +683,7 @@ vn_util__simplify_vnrval_binop(Binop, Vn1, Vn2, Vnrval, VnTables0, VnTables) :-
 			Vnrval1 = vn_const(int_const(I1)),
 			Vnrval2 = vn_const(int_const(I2))
 		->
-			I is I1 << I2,
+			I = unchecked_left_shift(I1, I2),
 			Vnrval = vn_const(int_const(I)),
 			VnTables = VnTables0
 		;
@@ -939,18 +944,21 @@ vn_util__lval_to_vnlval(Lval, Vnlval, VnTables0, VnTables) :-
 	; Lval = mem_ref(Rval1) ->
 		vn_util__rval_to_vn(Rval1, Vn1, VnTables0, VnTables),
 		Vnlval = vn_mem_ref(Vn1)
-	; Lval = succfr(Rval1) ->
-		vn_util__rval_to_vn(Rval1, Vn1, VnTables0, VnTables),
-		Vnlval = vn_succfr(Vn1)
 	; Lval = prevfr(Rval1) ->
 		vn_util__rval_to_vn(Rval1, Vn1, VnTables0, VnTables),
 		Vnlval = vn_prevfr(Vn1)
 	; Lval = redoip(Rval1) ->
 		vn_util__rval_to_vn(Rval1, Vn1, VnTables0, VnTables),
 		Vnlval = vn_redoip(Vn1)
+	; Lval = redofr(Rval1) ->
+		vn_util__rval_to_vn(Rval1, Vn1, VnTables0, VnTables),
+		Vnlval = vn_redofr(Vn1)
 	; Lval = succip(Rval1) ->
 		vn_util__rval_to_vn(Rval1, Vn1, VnTables0, VnTables),
 		Vnlval = vn_succip(Vn1)
+	; Lval = succfr(Rval1) ->
+		vn_util__rval_to_vn(Rval1, Vn1, VnTables0, VnTables),
+		Vnlval = vn_succfr(Vn1)
 	;
 		error("unexpected lval in vn_util__lval_to_vnlval")
 	).
@@ -967,9 +975,10 @@ vn_util__no_access_lval_to_vnlval(framevar(N),		yes(vn_framevar(N))).
 vn_util__no_access_lval_to_vnlval(succip,		yes(vn_succip)).
 vn_util__no_access_lval_to_vnlval(maxfr,		yes(vn_maxfr)).
 vn_util__no_access_lval_to_vnlval(curfr,		yes(vn_curfr)).
-vn_util__no_access_lval_to_vnlval(redoip(_),		no).
-vn_util__no_access_lval_to_vnlval(succip(_),		no).
 vn_util__no_access_lval_to_vnlval(prevfr(_),		no).
+vn_util__no_access_lval_to_vnlval(redoip(_),		no).
+vn_util__no_access_lval_to_vnlval(redofr(_),		no).
+vn_util__no_access_lval_to_vnlval(succip(_),		no).
 vn_util__no_access_lval_to_vnlval(succfr(_),		no).
 vn_util__no_access_lval_to_vnlval(hp,			yes(vn_hp)).
 vn_util__no_access_lval_to_vnlval(sp,			yes(vn_sp)).
@@ -985,10 +994,11 @@ vn_util__no_access_vnlval_to_lval(vn_framevar(N),	yes(framevar(N))).
 vn_util__no_access_vnlval_to_lval(vn_succip,		yes(succip)).
 vn_util__no_access_vnlval_to_lval(vn_maxfr,		yes(maxfr)).
 vn_util__no_access_vnlval_to_lval(vn_curfr,		yes(curfr)).
-vn_util__no_access_vnlval_to_lval(vn_succfr(_),		no).
-vn_util__no_access_vnlval_to_lval(vn_prevfr(_),		no).
+vn_util__no_access_vnlval_to_lval(vn_redofr(_),		no).
 vn_util__no_access_vnlval_to_lval(vn_redoip(_),		no).
+vn_util__no_access_vnlval_to_lval(vn_prevfr(_),		no).
 vn_util__no_access_vnlval_to_lval(vn_succip(_),		no).
+vn_util__no_access_vnlval_to_lval(vn_succfr(_),		no).
 vn_util__no_access_vnlval_to_lval(vn_hp,		yes(hp)).
 vn_util__no_access_vnlval_to_lval(vn_sp,		yes(sp)).
 vn_util__no_access_vnlval_to_lval(vn_field(_, _, _),	no).
@@ -1002,10 +1012,11 @@ vn_util__vnlval_access_vns(vn_framevar(_), []).
 vn_util__vnlval_access_vns(vn_succip, []).
 vn_util__vnlval_access_vns(vn_maxfr, []).
 vn_util__vnlval_access_vns(vn_curfr, []).
-vn_util__vnlval_access_vns(vn_succfr(Vn), [Vn]).
 vn_util__vnlval_access_vns(vn_prevfr(Vn), [Vn]).
 vn_util__vnlval_access_vns(vn_redoip(Vn), [Vn]).
+vn_util__vnlval_access_vns(vn_redofr(Vn), [Vn]).
 vn_util__vnlval_access_vns(vn_succip(Vn), [Vn]).
+vn_util__vnlval_access_vns(vn_succfr(Vn), [Vn]).
 vn_util__vnlval_access_vns(vn_hp, []).
 vn_util__vnlval_access_vns(vn_sp, []).
 vn_util__vnlval_access_vns(vn_field(_, Vn1, Vn2), [Vn1, Vn2]).
@@ -1015,7 +1026,7 @@ vn_util__find_sub_vns(vn_origlval(Vnlval), SubVns) :-
 	vn_util__vnlval_access_vns(Vnlval, SubVns).
 vn_util__find_sub_vns(vn_mkword(_, SubVn), [SubVn]).
 vn_util__find_sub_vns(vn_const(_), []).
-vn_util__find_sub_vns(vn_create(_, _, _, _), []).
+vn_util__find_sub_vns(vn_create(_, _, _, _, _, _), []).
 vn_util__find_sub_vns(vn_unop(_, SubVn), [SubVn]).
 vn_util__find_sub_vns(vn_binop(_, SubVn1, SubVn2), [SubVn1, SubVn2]).
 vn_util__find_sub_vns(vn_stackvar_addr(_), []).
@@ -1034,7 +1045,7 @@ vn_util__is_const_expr(Vn, IsConst, VnTables) :-
 		Vnrval = vn_const(_),
 		IsConst = yes
 	;
-		Vnrval = vn_create(_, _, _, _),
+		Vnrval = vn_create(_, _, _, _, _, _),
 		IsConst = yes
 	;
 		Vnrval = vn_unop(_, Vn1),
@@ -1065,7 +1076,7 @@ vn_util__find_lvals_in_rval(Rval, Lvals) :-
 		Rval = var(_),
 		error("var found in vn_util__find_lvals_in_rval")
 	;
-		Rval = create(_, _, _, _),
+		Rval = create(_, _, _, _, _, _),
 		Lvals = []
 	;
 		Rval = mkword(_, Rval1),
@@ -1179,10 +1190,11 @@ vn_util__classify_loc_cost(vn_framevar(_), 1).
 vn_util__classify_loc_cost(vn_succip, 0).
 vn_util__classify_loc_cost(vn_maxfr, 0).
 vn_util__classify_loc_cost(vn_curfr, 0).
-vn_util__classify_loc_cost(vn_succfr(_), 1).
 vn_util__classify_loc_cost(vn_prevfr(_), 1).
 vn_util__classify_loc_cost(vn_redoip(_), 1).
+vn_util__classify_loc_cost(vn_redofr(_), 1).
 vn_util__classify_loc_cost(vn_succip(_), 1).
+vn_util__classify_loc_cost(vn_succfr(_), 1).
 vn_util__classify_loc_cost(vn_hp, 0).
 vn_util__classify_loc_cost(vn_sp, 0).
 vn_util__classify_loc_cost(vn_field(_, _, _), 2).
@@ -1216,10 +1228,10 @@ vn_util__build_uses_from_ctrl(Ctrl, Ctrlmap, VnTables0, VnTables) :-
 			VnInstr = vn_livevals(_),
 			VnTables1 = VnTables0
 		;
-			VnInstr = vn_call(_, _, _, _),
+			VnInstr = vn_call(_, _, _, _, _),
 			VnTables1 = VnTables0
 		;
-			VnInstr = vn_mkframe(_, _, _),
+			VnInstr = vn_mkframe(_, _),
 			VnTables1 = VnTables0
 		;
 			VnInstr = vn_label(_),
@@ -1331,7 +1343,7 @@ vn_util__record_use(Vn, Src, VnTables0, VnTables) :-
 			Vnrval = vn_const(_),
 			VnTables = VnTables1
 		;
-			Vnrval = vn_create(_, _, _, _),
+			Vnrval = vn_create(_, _, _, _, _, _),
 			VnTables = VnTables1
 		;
 			Vnrval = vn_unop(_, SubVn),
