@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1997 The University of Melbourne.
+% Copyright (C) 1996-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -26,7 +26,7 @@
 
 :- implementation.
 
-:- import_module opt_debug, vn_debug, vn_util.
+:- import_module builtin_ops, opt_debug, vn_debug, vn_util.
 :- import_module map, set, string, std_util, require.
 
 vn_verify__ok(Instrs, Uinstr0, SeenIncr0, SeenIncr, Liveset0, Liveset,
@@ -211,10 +211,11 @@ vn_verify__subst_access_vns(vn_framevar(N), [], framevar(N)).
 vn_verify__subst_access_vns(vn_succip, [], succip).
 vn_verify__subst_access_vns(vn_maxfr, [], maxfr).
 vn_verify__subst_access_vns(vn_curfr, [], curfr).
-vn_verify__subst_access_vns(vn_succfr(_), [R], succfr(R)).
 vn_verify__subst_access_vns(vn_prevfr(_), [R], prevfr(R)).
 vn_verify__subst_access_vns(vn_redoip(_), [R], redoip(R)).
+vn_verify__subst_access_vns(vn_redofr(_), [R], redofr(R)).
 vn_verify__subst_access_vns(vn_succip(_), [R], succip(R)).
+vn_verify__subst_access_vns(vn_succfr(_), [R], succfr(R)).
 vn_verify__subst_access_vns(vn_hp, [], hp).
 vn_verify__subst_access_vns(vn_sp, [], sp).
 vn_verify__subst_access_vns(vn_field(T, _, _), [R1, R2], field(T, R1, R2)).
@@ -227,7 +228,7 @@ vn_verify__subst_sub_vns(vn_origlval(Vnlval), _, VnTables, lval(Lval)) :-
 	vn_verify__lval(Vnlval, VnTables, Lval).
 vn_verify__subst_sub_vns(vn_mkword(Tag, _), [R], _, mkword(Tag, R)).
 vn_verify__subst_sub_vns(vn_const(Const), [], _, const(Const)).
-vn_verify__subst_sub_vns(vn_create(T, A, U, L), [], _, create(T, A, U, L)).
+vn_verify__subst_sub_vns(vn_create(T,A,AT,U,L,M), [], _, create(T,A,AT,U,L,M)).
 vn_verify__subst_sub_vns(vn_unop(Op, _), [R], _, unop(Op, R)).
 vn_verify__subst_sub_vns(vn_binop(Op, _, _), [R1, R2], _, binop(Op, R1, R2)).
 
@@ -294,15 +295,11 @@ vn_verify__tags_instr(Instr, NoDeref0, NoDeref, Tested0, Tested) :-
 		set__delete(NoDeref2, lval(Lval), NoDeref),
 		vn_verify__tags_rval(Rval, NoDeref)
 	;
-		Instr = call(_, _, _, _),
+		Instr = call(_, _, _, _, _),
 		NoDeref = NoDeref0,
 		Tested = Tested0
 	;
-		Instr = mkframe(_, _, _),
-		NoDeref = NoDeref0,
-		Tested = Tested0
-	;
-		Instr = modframe(_),
+		Instr = mkframe(_, _),
 		NoDeref = NoDeref0,
 		Tested = Tested0
 	;
@@ -325,7 +322,7 @@ vn_verify__tags_instr(Instr, NoDeref0, NoDeref, Tested0, Tested) :-
 		vn_verify__tags_cond(Rval, NoDeref0, NoDeref,
 			Tested0, Tested)
 	;
-		Instr = incr_hp(Lval, _, Rval),
+		Instr = incr_hp(Lval, _, Rval, _),
 		vn_verify__tags_lval(Lval, NoDeref0),
 		vn_verify__tags_rval(Rval, NoDeref0),
 		NoDeref = NoDeref0,
@@ -373,7 +370,7 @@ vn_verify__tags_instr(Instr, NoDeref0, NoDeref, Tested0, Tested) :-
 		NoDeref = NoDeref0,
 		Tested = Tested0
 	;
-		Instr = pragma_c(_, _, _, _, _),
+		Instr = pragma_c(_, _, _, _, _, _),
 		error("found c_code in vn_verify__tags_instr")
 	).
 
@@ -386,13 +383,15 @@ vn_verify__tags_lval(framevar(_), _).
 vn_verify__tags_lval(succip, _).
 vn_verify__tags_lval(maxfr, _).
 vn_verify__tags_lval(curfr, _).
-vn_verify__tags_lval(succip(Rval), NoDeref) :-
+vn_verify__tags_lval(prevfr(Rval), NoDeref) :-
 	vn_verify__tags_rval(Rval, NoDeref).
 vn_verify__tags_lval(redoip(Rval), NoDeref) :-
 	vn_verify__tags_rval(Rval, NoDeref).
-vn_verify__tags_lval(succfr(Rval), NoDeref) :-
+vn_verify__tags_lval(redofr(Rval), NoDeref) :-
 	vn_verify__tags_rval(Rval, NoDeref).
-vn_verify__tags_lval(prevfr(Rval), NoDeref) :-
+vn_verify__tags_lval(succip(Rval), NoDeref) :-
+	vn_verify__tags_rval(Rval, NoDeref).
+vn_verify__tags_lval(succfr(Rval), NoDeref) :-
 	vn_verify__tags_rval(Rval, NoDeref).
 vn_verify__tags_lval(hp, _).
 vn_verify__tags_lval(sp, _).
@@ -411,7 +410,7 @@ vn_verify__tags_rval(lval(Lval), NoDeref) :-
 	vn_verify__tags_lval(Lval, NoDeref).
 vn_verify__tags_rval(var(_), _) :-
 	error("found var in vn_verify__tags_rval").
-vn_verify__tags_rval(create(_, _, _, _), _).
+vn_verify__tags_rval(create(_, _, _, _, _, _), _).
 vn_verify__tags_rval(mkword(_, Rval), NoDeref) :-
 	vn_verify__tags_rval(Rval, NoDeref).
 vn_verify__tags_rval(const(_), _).
