@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1993-1995, 1997 The University of Melbourne.
+% Copyright (C) 1993-1995, 1997-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -11,6 +11,26 @@
 % This module provides dynamically-sized one-dimensional arrays.
 % Array indices start at zero.
 
+% By default, the array__set and array__lookup procedures will check
+% for bounds errors.  But for better performance, it is possible to
+% disable some of the checking by compiling with `--intermodule-optimization'
+% and with the C macro symbol `ML_OMIT_ARRAY_BOUNDS_CHECKS'
+% defined, e.g. by using `MCFLAGS=--intermodule-optimization' and
+% `CFLAGS=-DML_OMIT_ARRAY_BOUNDS_CHECKS' in your Mmakefile,
+% or by compiling with the command
+% `mmc --intermodule-optimization --cflags -DML_OMIT_ARRAY_BOUNDS_CHECKS'.
+%
+% For maximum performance, all bounds checking can be disabled by
+% recompiling this module using `CFLAGS=-DML_OMIT_ARRAY_BOUNDS_CHECKS'
+% or `mmc --cflags -DML_OMIT_ARRAY_BOUNDS_CHECKS' as above. You can
+% either recompile the entire library, or just copy `array.m' to your
+% application's source directory and link with it directly instead of as
+% part of the library.
+%
+
+% Ralph Becket <rwab1@cam.sri.com> 24/04/99
+%	Function forms added.
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
@@ -118,6 +138,7 @@
 :- pred array__set(array(T), int, T, array(T)).
 :- mode array__set(array_di, in, in, array_uo) is det.
 
+
 	% array__semidet_set sets the nth element of an array,
 	% and returns the resulting array.
 	% It fails if the index is out of bounds.
@@ -161,6 +182,7 @@
 :- pred array__shrink(array(T), int, array(T)).
 :- mode array__shrink(array_di, in, array_uo) is det.
 
+
 	% array__from_list takes a list,
 	% and returns an array containing those elements in
 	% the same order that they occured in the list.
@@ -181,6 +203,7 @@
 :- pred array__fetch_items(array(T), int, int, list(T)).
 :- mode array__fetch_items(in, in, in, out) is det.
 
+
 	% array__bsearch takes an array, an element to be found
 	% and a comparison predicate and returns the position of
 	% the element in the array.  Assumes the array is in sorted
@@ -192,6 +215,11 @@
 :- mode array__bsearch(array_ui, in, pred(in, in, out) is det, out) is det.
 :- mode array__bsearch(in, in, pred(in, in, out) is det, out) is det.
 
+	% array__map(Closure, OldArray, NewArray) applys `Closure' to
+	% each of the elements of `OldArray' to create `NewArray'.
+:- pred array__map(pred(T1, T2), array(T1), array(T2)).
+:- mode array__map(pred(in, out) is det, array_di, array_uo) is det.
+
 %-----------------------------------------------------------------------------%
 :- implementation.
 
@@ -237,37 +265,50 @@ lower bounds other than zero are not supported
 % Arrays are implemented using the C interface.
 
 % The C type which defines the representation of arrays is
-% MR_ArrayType; it is defined in runtime/type_info.h.
+% MR_ArrayType; it is defined in runtime/mercury_library_types.h.
 
 %-----------------------------------------------------------------------------%
 
-:- pragma(c_code, "
+:- pragma c_code("
 
 Define_extern_entry(mercury____Unify___array__array_1_0);
 Define_extern_entry(mercury____Index___array__array_1_0);
 Define_extern_entry(mercury____Compare___array__array_1_0);
 
-#ifdef  USE_TYPE_LAYOUT
-
-const struct mercury_data_array__base_type_layout_array_1_struct {
+MR_MODULE_STATIC_OR_EXTERN 
+const struct mercury_data_array__type_ctor_layout_array_1_struct {
 	TYPE_LAYOUT_FIELDS
-} mercury_data_array__base_type_layout_array_1 = {
-	make_typelayout_for_all_tags(TYPELAYOUT_CONST_TAG, 
-		mkbody(TYPELAYOUT_ARRAY_VALUE))
+} mercury_data_array__type_ctor_layout_array_1 = {
+	make_typelayout_for_all_tags(TYPE_CTOR_LAYOUT_CONST_TAG, 
+		MR_mkbody(MR_TYPE_CTOR_LAYOUT_ARRAY_VALUE))
 };
 
-const struct mercury_data_array__base_type_functors_array_1_struct {
+MR_MODULE_STATIC_OR_EXTERN
+const struct mercury_data_array__type_ctor_functors_array_1_struct {
 	Integer f1;
-} mercury_data_array__base_type_functors_array_1 = {
-	MR_TYPEFUNCTORS_SPECIAL
+} mercury_data_array__type_ctor_functors_array_1 = {
+	MR_TYPE_CTOR_FUNCTORS_SPECIAL
+};
+
+MR_STATIC_CODE_CONST struct MR_TypeCtorInfo_struct
+mercury_data_array__type_ctor_info_array_1 = {
+	(Integer) 1,
+	MR_MAYBE_STATIC_CODE(ENTRY(mercury____Unify___array__array_1_0)),
+	MR_MAYBE_STATIC_CODE(ENTRY(mercury____Index___array__array_1_0)),
+	MR_MAYBE_STATIC_CODE(ENTRY(mercury____Compare___array__array_1_0)),
+	MR_TYPECTOR_REP_ARRAY,
+	(Word *) &mercury_data_array__type_ctor_functors_array_1,
+	(Word *) &mercury_data_array__type_ctor_layout_array_1,
+	MR_string_const(""array"", 5),
+	MR_string_const(""array"", 5),
+	MR_RTTI_VERSION
 };
 
-#endif
 
 Declare_entry(mercury__array__array_equal_2_0);
 Declare_entry(mercury__array__array_compare_3_0);
 
-BEGIN_MODULE(array_module)
+BEGIN_MODULE(array_module_builtins)
 	init_entry(mercury____Unify___array__array_1_0);
 	init_entry(mercury____Index___array__array_1_0);
 	init_entry(mercury____Compare___array__array_1_0);
@@ -279,7 +320,7 @@ Define_entry(mercury____Unify___array__array_1_0);
 		ENTRY(mercury____Unify___array__array_1_0));
 
 Define_entry(mercury____Index___array__array_1_0);
-	index_output = -1;
+	r1 = -1;
 	proceed();
 
 Define_entry(mercury____Compare___array__array_1_0);
@@ -291,12 +332,18 @@ END_MODULE
 
 /* Ensure that the initialization code for the above module gets run. */
 /*
-INIT sys_init_array_module
+INIT sys_init_array_module_builtins
 */
-void sys_init_array_module(void); /* suppress gcc -Wmissing-decl warning */
-void sys_init_array_module(void) {
-	extern ModuleFunc array_module;
-	array_module();
+
+MR_MODULE_STATIC_OR_EXTERN ModuleFunc array_module_builtins;
+
+void sys_init_array_module_builtins(void);
+		/* suppress gcc -Wmissing-decl warning */
+void sys_init_array_module_builtins(void) {
+	array_module_builtins();
+	MR_INIT_TYPE_CTOR_INFO(
+		mercury_data_array__type_ctor_info_array_1,
+		array__array_1_0);
 }
 
 ").
@@ -357,11 +404,11 @@ array__compare_elements(N, Size, Array1, Array2, Result) :-
 
 %-----------------------------------------------------------------------------%
 
-:- pragma(c_header_code, "
+:- pragma c_header_code("
 MR_ArrayType *ML_make_array(Integer size, Word item);
 ").
 
-:- pragma(c_code, "
+:- pragma c_code("
 MR_ArrayType *
 ML_make_array(Integer size, Word item)
 {
@@ -377,33 +424,35 @@ ML_make_array(Integer size, Word item)
 }
 ").
 
-:- pragma(c_code,
-	array__init(Size::in, Item::in, Array::array_uo),
-"
+:- pragma c_code(array__init(Size::in, Item::in, Array::array_uo),
+		[will_not_call_mercury, thread_safe], "
 	Array = (Word) ML_make_array(Size, Item);
 ").
 
-:- pragma(c_code,
-	array__make_empty_array(Array::array_uo),
-"
+:- pragma c_code(array__make_empty_array(Array::array_uo),
+		[will_not_call_mercury, thread_safe], "
 	Array = (Word) ML_make_array(0, 0);
 ").
 
 %-----------------------------------------------------------------------------%
 
-:- pragma(c_code, array__min(Array::array_ui, Min::out), "
+:- pragma c_code(array__min(Array::array_ui, Min::out),
+		[will_not_call_mercury, thread_safe], "
 	/* Array not used */
 	Min = 0;
 ").
-:- pragma(c_code, array__min(Array::in, Min::out), "
+:- pragma c_code(array__min(Array::in, Min::out),
+		[will_not_call_mercury, thread_safe], "
 	/* Array not used */
 	Min = 0;
 ").
 
-:- pragma(c_code, array__max(Array::array_ui, Max::out), "
+:- pragma c_code(array__max(Array::array_ui, Max::out), 
+		[will_not_call_mercury, thread_safe], "
 	Max = ((MR_ArrayType *)Array)->size - 1;
 ").
-:- pragma(c_code, array__max(Array::in, Max::out), "
+:- pragma c_code(array__max(Array::in, Max::out), 
+		[will_not_call_mercury, thread_safe], "
 	Max = ((MR_ArrayType *)Array)->size - 1;
 ").
 
@@ -413,10 +462,12 @@ array__bounds(Array, Min, Max) :-
 
 %-----------------------------------------------------------------------------%
 
-:- pragma(c_code, array__size(Array::array_ui, Max::out), "
+:- pragma c_code(array__size(Array::array_ui, Max::out),
+		[will_not_call_mercury, thread_safe], "
 	Max = ((MR_ArrayType *)Array)->size;
 ").
-:- pragma(c_code, array__size(Array::in, Max::out), "
+:- pragma c_code(array__size(Array::in, Max::out),
+		[will_not_call_mercury, thread_safe], "
 	Max = ((MR_ArrayType *)Array)->size;
 ").
 
@@ -444,42 +495,50 @@ array__slow_set(Array0, Index, Item, Array) :-
 
 %-----------------------------------------------------------------------------%
 
-:- pragma(c_code, array__lookup(Array::array_ui, Index::in,
-		Item::out), "{
+:- pragma c_code(array__lookup(Array::array_ui, Index::in, Item::out),
+		[will_not_call_mercury, thread_safe], "{
 	MR_ArrayType *array = (MR_ArrayType *)Array;
+#ifndef ML_OMIT_ARRAY_BOUNDS_CHECKS
 	if ((Unsigned) Index >= (Unsigned) array->size) {
 		fatal_error(""array__lookup: array index out of bounds"");
 	}
+#endif
 	Item = array->elements[Index];
 }").
-:- pragma(c_code, array__lookup(Array::in, Index::in, Item::out), "{
+:- pragma c_code(array__lookup(Array::in, Index::in, Item::out),
+		[will_not_call_mercury, thread_safe], "{
 	MR_ArrayType *array = (MR_ArrayType *)Array;
+#ifndef ML_OMIT_ARRAY_BOUNDS_CHECKS
 	if ((Unsigned) Index >= (Unsigned) array->size) {
 		fatal_error(""array__lookup: array index out of bounds"");
 	}
+#endif
 	Item = array->elements[Index];
 }").
 
 %-----------------------------------------------------------------------------%
 
-:- pragma(c_code, array__set(Array0::array_di, Index::in,
-		Item::in, Array::array_uo), "{
+:- pragma c_code(array__set(Array0::array_di, Index::in,
+		Item::in, Array::array_uo),
+		[will_not_call_mercury, thread_safe], "{
 	MR_ArrayType *array = (MR_ArrayType *)Array0;
+#ifndef ML_OMIT_ARRAY_BOUNDS_CHECKS
 	if ((Unsigned) Index >= (Unsigned) array->size) {
 		fatal_error(""array__set: array index out of bounds"");
 	}
+#endif
 	array->elements[Index] = Item;	/* destructive update! */
 	Array = Array0;
 }").
 
 %-----------------------------------------------------------------------------%
 
-:- pragma(c_header_code, "
+:- pragma c_header_code("
 MR_ArrayType * ML_resize_array(MR_ArrayType *old_array,
 					Integer array_size, Word item);
 ").
 
-:- pragma(c_code, "
+:- pragma c_code("
 MR_ArrayType *
 ML_resize_array(MR_ArrayType *old_array, Integer array_size,
 				Word item)
@@ -494,7 +553,7 @@ ML_resize_array(MR_ArrayType *old_array, Integer array_size,
 		elements_to_copy = array_size;
 	}
 
-	array = (MR_ArrayType *) make_many(Word, array_size + 1);
+	array = (MR_ArrayType *) MR_GC_NEW_ARRAY(Word, array_size + 1);
 	array->size = array_size;
 	for (i = 0; i < elements_to_copy; i++) {
 		array->elements[i] = old_array->elements[i];
@@ -507,28 +566,26 @@ ML_resize_array(MR_ArrayType *old_array, Integer array_size,
 	** since the mode on the old array is `array_di', it is safe to
 	** deallocate the storage for it
 	*/
-	oldmem(old_array);
+	MR_GC_free(old_array);
 
 	return array;
 }
 ").
 
-:- pragma(c_code,
-	array__resize(Array0::array_di, Size::in, Item::in,
-		Array::array_uo),
-"
+:- pragma c_code(array__resize(Array0::array_di, Size::in, Item::in,
+		Array::array_uo), [will_not_call_mercury, thread_safe], "
 	Array = (Word) ML_resize_array(
 				(MR_ArrayType *) Array0, Size, Item);
 ").
 
 %-----------------------------------------------------------------------------%
 
-:- pragma(c_header_code, "
+:- pragma c_header_code("
 MR_ArrayType * ML_shrink_array(MR_ArrayType *old_array,
 					Integer array_size);
 ").
 
-:- pragma(c_code, "
+:- pragma c_code("
 MR_ArrayType *
 ML_shrink_array(MR_ArrayType *old_array, Integer array_size)
 {
@@ -542,7 +599,7 @@ ML_shrink_array(MR_ArrayType *old_array, Integer array_size)
 		fatal_error(""array__shrink: can't shrink to a larger size"");
 	}
 
-	array = (MR_ArrayType *) make_many(Word, array_size + 1);
+	array = (MR_ArrayType *) MR_GC_NEW_ARRAY(Word, array_size + 1);
 	array->size = array_size;
 	for (i = 0; i < array_size; i++) {
 		array->elements[i] = old_array->elements[i];
@@ -552,27 +609,25 @@ ML_shrink_array(MR_ArrayType *old_array, Integer array_size)
 	** since the mode on the old array is `array_di', it is safe to
 	** deallocate the storage for it
 	*/
-	oldmem(old_array);
+	MR_GC_free(old_array);
 
 	return array;
 }
 ").
 
-:- pragma(c_code,
-	array__shrink(Array0::array_di, Size::in,
-		Array::array_uo),
-"
+:- pragma c_code(array__shrink(Array0::array_di, Size::in, Array::array_uo),
+		[will_not_call_mercury, thread_safe], "
 	Array = (Word) ML_shrink_array(
 				(MR_ArrayType *) Array0, Size);
 ").
 
 %-----------------------------------------------------------------------------%
 
-:- pragma(c_header_code, "
+:- pragma c_header_code("
 MR_ArrayType *ML_copy_array(MR_ArrayType *old_array);
 ").
 
-:- pragma(c_code, "
+:- pragma c_code("
 MR_ArrayType *
 ML_copy_array(MR_ArrayType *old_array)
 {
@@ -595,18 +650,14 @@ ML_copy_array(MR_ArrayType *old_array)
 }
 ").
 
-:- pragma(c_code,
-	array__copy(Array0::array_ui, Array::array_uo),
-"
-	Array =
-		(Word) ML_copy_array((MR_ArrayType *) Array0);
+:- pragma c_code(array__copy(Array0::array_ui, Array::array_uo),
+		[will_not_call_mercury, thread_safe], "
+	Array = (Word) ML_copy_array((MR_ArrayType *) Array0);
 ").
 
-:- pragma(c_code,
-	array__copy(Array0::in, Array::array_uo),
-"
-	Array =
-		(Word) ML_copy_array((MR_ArrayType *) Array0);
+:- pragma c_code(array__copy(Array0::in, Array::array_uo),
+		[will_not_call_mercury, thread_safe], "
+	Array = (Word) ML_copy_array((MR_ArrayType *) Array0);
 ").
 
 %-----------------------------------------------------------------------------%
@@ -698,4 +749,147 @@ array__bsearch_2(Array, Lo, Hi, El, Compare, Result) :-
 	).
 
 %-----------------------------------------------------------------------------%
+
+array__map(Closure, OldArray, NewArray) :-
+	( array__semidet_lookup(OldArray, 0, Elem0) ->
+		array__size(OldArray, Size),
+		call(Closure, Elem0, Elem),
+		array__init(Size, Elem, NewArray0),
+		array__map_2(1, Size, Closure, OldArray,
+		NewArray0, NewArray)
+	;
+		array__make_empty_array(NewArray)
+	).
+
+:- pred array__map_2(int, int, pred(T1, T2), array(T1), array(T2), array(T2)).
+:- mode array__map_2(in, in, pred(in, out) is det, in, array_di, array_uo)
+		is det.
+
+array__map_2(N, Size, Closure, OldArray, NewArray0, NewArray) :-
+	( N >= Size ->
+		NewArray = NewArray0
+	;
+		array__lookup(OldArray, N, OldElem),
+		Closure(OldElem, NewElem),
+		array__set(NewArray0, N, NewElem, NewArray1),
+		array__map_2(N + 1, Size, Closure, OldArray,
+		NewArray1, NewArray)
+	).
+
 %-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
+% Ralph Becket <rwab1@cam.sri.com> 24/04/99
+%	Function forms added.
+
+:- interface.
+
+:- func array__make_empty_array = array(T).
+:- mode array__make_empty_array = array_uo is det.
+
+:- func array__init(int, T) = array(T).
+:- mode array__init(in, in) = array_uo is det.
+
+:- func array__min(array(_T)) = int.
+:- mode array__min(array_ui) = out is det.
+
+:- func array__max(array(_T)) = int.
+:- mode array__max(array_ui) = out is det.
+
+:- func array__size(array(_T)) = int.
+:- mode array__size(array_ui) = out is det.
+
+:- func array__lookup(array(T), int) = T.
+:- mode array__lookup(array_ui, in) = out is det.
+
+:- func array__set(array(T), int, T) = array(T).
+:- mode array__set(array_di, in, in) = array_uo is det.
+
+:- func array__slow_set(array(T), int, T) = array(T).
+:- mode array__slow_set(array_ui, in, in) = array_uo is det.
+:- mode array__slow_set(in, in, in) = array_uo is det.
+
+:- func array__copy(array(T)) = array(T).
+:- mode array__copy(array_ui) = array_uo is det.
+
+:- func array__resize(array(T), int, T) = array(T).
+:- mode array__resize(array_di, in, in) = array_uo is det.
+
+:- func array__shrink(array(T), int) = array(T).
+:- mode array__shrink(array_di, in) = array_uo is det.
+
+:- func array__from_list(list(T)) = array(T).
+:- mode array__from_list(in) = array_uo is det.
+
+:- func array__to_list(array(T)) = list(T).
+:- mode array__to_list(array_ui) = out is det.
+
+:- func array__fetch_items(array(T), int, int) = list(T).
+:- mode array__fetch_items(array_ui, in, in) = out is det.
+
+:- func array__bsearch(array(T), T, func(T,T) = comparison_result) = maybe(int).
+:- mode array__bsearch(array_ui, in, func(in,in) = out is det) = out is det.
+
+:- func array__map(func(T1) = T2, array(T1)) = array(T2).
+:- mode array__map(func(in) = out is det, array_di) = array_uo is det.
+
+:- func array_compare(array(T), array(T)) = comparison_result.
+:- mode array_compare(in, in) = out is det.
+
+% ---------------------------------------------------------------------------- %
+% ---------------------------------------------------------------------------- %
+
+:- implementation.
+
+array__make_empty_array = A :-
+	array__make_empty_array(A).
+
+array__init(N, X) = A :-
+	array__init(N, X, A).
+
+array__min(A) = N :-
+	array__min(A, N).
+
+array__max(A) = N :-
+	array__max(A, N).
+
+array__size(A) = N :-
+	array__size(A, N).
+
+array__lookup(A, N) = X :-
+	array__lookup(A, N, X).
+
+array__set(A1, N, X) = A2 :-
+	array__set(A1, N, X, A2).
+
+array__slow_set(A1, N, X) = A2 :-
+	array__slow_set(A1, N, X, A2).
+
+array__copy(A1) = A2 :-
+	array__copy(A1, A2).
+
+array__resize(A1, N, X) = A2 :-
+	array__resize(A1, N, X, A2).
+
+array__shrink(A1, N) = A2 :-
+	array__shrink(A1, N, A2).
+
+array__from_list(Xs) = A :-
+	array__from_list(Xs, A).
+
+array__to_list(A) = Xs :-
+	array__to_list(A, Xs).
+
+array__fetch_items(A, N1, N2) = Xs :-
+	array__fetch_items(A, N1, N2, Xs).
+
+array__bsearch(A, X, F) = MN :-
+	P = ( pred(X1::in, X2::in, C::out) is det :- C = F(X1, X2) ),
+	array__bsearch(A, X, P, MN).
+
+array__map(F, A1) = A2 :-
+	P = ( pred(X::in, Y::out) is det :- Y = F(X) ),
+	array__map(P, A1, A2).
+
+array_compare(A1, A2) = C :-
+	array_compare(C, A1, A2).
+
