@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1993-1995, 1997 The University of Melbourne.
+% Copyright (C) 1993-1995, 1997, 1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -50,6 +50,21 @@
 :- mode bintree__search(in, in, in) is semidet.	% implied
 :- mode bintree__search(in, in, out) is semidet.
 
+:- pred bintree__lookup(bintree(K,V), K, V).
+:- mode bintree__lookup(in, in, out) is det.
+
+:- pred bintree__lower_bound_search(bintree(K,V), K, K, V).
+:- mode bintree__lower_bound_search(in, in, out, out) is semidet.
+
+:- pred bintree__lower_bound_lookup(bintree(K,V), K, K, V).
+:- mode bintree__lower_bound_lookup(in, in, out, out) is det.
+
+:- pred bintree__upper_bound_search(bintree(K,V), K, K, V).
+:- mode bintree__upper_bound_search(in, in, out, out) is semidet.
+
+:- pred bintree__upper_bound_lookup(bintree(K,V), K, K, V).
+:- mode bintree__upper_bound_lookup(in, in, out, out) is det.
+
 :- pred bintree__delete(bintree(K,V), K, bintree(K,V)).
 :- mode bintree__delete(in, in, out) is det.
 
@@ -108,8 +123,6 @@ bintree__init(empty).
 
 %-----------------------------------------------------------------------------%
 
-:- bintree__insert(A, B, _, _) when A and B.
-
 bintree__insert(empty, Key, Value, tree(Key, Value, empty, empty)).
 bintree__insert(tree(Key0, Value0, Left, Right), Key, Value, Tree) :-
 	compare(Result, Key0, Key),
@@ -149,8 +162,6 @@ bintree__update(tree(Key0, Value0, Left, Right), Key, Value, Tree) :-
 
 %-----------------------------------------------------------------------------%
 
-:- bintree__set(A, B, _, _) when A and B.
-
 bintree__set(empty, Key, Value, tree(Key, Value, empty, empty)).
 bintree__set(tree(Key0, Value0, Left, Right), Key, Value, Tree) :-
 	compare(Result, Key0, Key),
@@ -170,14 +181,12 @@ bintree__set(tree(Key0, Value0, Left, Right), Key, Value, Tree) :-
 
 %-----------------------------------------------------------------------------%
 
-:- bintree__search(A, B, _) when A and B.
-
 bintree__search(tree(K0, V0, Left, Right), K, V) :-
 	compare(Result, K0, K),
 	(
 		Result = (=)
 	->
-		V0 = V
+		V = V0
 	;
 		Result = (<)
 	->
@@ -188,7 +197,78 @@ bintree__search(tree(K0, V0, Left, Right), K, V) :-
 
 %-----------------------------------------------------------------------------%
 
-:- bintree__delete(A, B, _) when A and B.
+bintree__lookup(Tree, K, V) :-
+	( bintree__search(Tree, K, V0) ->
+		V = V0
+	;
+		report_lookup_error("bintree__lookup: key not found", K, V)
+	).
+
+%-----------------------------------------------------------------------------%
+
+bintree__lower_bound_search(tree(K0, V0, Left, Right), SearchK, K, V) :-
+	compare(Result, K0, SearchK),
+	(
+		Result = (=)
+	->
+		K = K0,
+		V = V0
+	;
+		Result = (<)
+	->
+		( bintree__lower_bound_search(Right, SearchK, Kp, Vp) ->
+			K = Kp,
+			V = Vp
+		;
+			K = K0,
+			V = V0
+		)
+	;
+		bintree__lower_bound_search(Left, SearchK, K, V)
+	).
+
+bintree__lower_bound_lookup(Tree, SearchK, K, V) :-
+	( bintree__lower_bound_search(Tree, SearchK, K0, V0) ->
+		K = K0,
+		V = V0
+	;
+		report_lookup_error("bintree__lower_bound_lookup: key not found",
+			SearchK, V)
+	).
+
+%-----------------------------------------------------------------------------%
+
+bintree__upper_bound_search(tree(K0, V0, Left, Right), SearchK, K, V) :-
+	compare(Result, K0, SearchK),
+	(
+		Result = (=)
+	->
+		K = K0,
+		V = V0
+	;
+		Result = (<)
+	->
+		bintree__upper_bound_search(Right, SearchK, K, V)
+	;
+		( bintree__upper_bound_search(Left, SearchK, Kp, Vp) ->
+			K = Kp,
+			V = Vp
+		;
+			K = K0,
+			V = V0
+		)
+	).
+
+bintree__upper_bound_lookup(Tree, SearchK, K, V) :-
+	( bintree__upper_bound_search(Tree, SearchK, K0, V0) ->
+		K = K0,
+		V = V0
+	;
+		report_lookup_error("bintree__lower_bound_lookup: key not found",
+			SearchK, V)
+	).
+
+%-----------------------------------------------------------------------------%
 
 bintree__delete(empty, _K, empty).
 bintree__delete(tree(K0, V0, Left, Right), K, Tree) :-
@@ -318,7 +398,8 @@ bintree__from_list_2([K - V | List], Tree0, Tree) :-
 	( bintree__insert(Tree0, K, V, Tree1) ->
 		Tree2 = Tree1
 	;
-		error("bintree__from_list: duplicate key")
+		report_lookup_error("bintree__from_list: key already present",
+			K, V)
 	),
 	bintree__from_list_2(List, Tree2, Tree).
 
@@ -364,9 +445,7 @@ bintree__from_corresponding_lists(Keys, Values, Tree) :-
 	( bintree__from_corresponding_lists_2(Keys, Values, empty, Tree0) ->
 		Tree = Tree0
 	;
-		% Either the list weren't the same length, or the
-		% key list contained duplicates
-		error("bintree__from_corresponding_lists")
+		error("bintree__from_corresponding_lists: lists are of different lengths")
 	).
 
 :- pred bintree__from_corresponding_lists_2(list(K), list(V), bintree(K,V),
@@ -375,8 +454,12 @@ bintree__from_corresponding_lists(Keys, Values, Tree) :-
 
 bintree__from_corresponding_lists_2([], [], Tree, Tree).
 bintree__from_corresponding_lists_2([K | Ks], [V | Vs], Tree0, Tree) :-
-	bintree__insert(Tree0, K, V, Tree1),
-	bintree__from_corresponding_lists_2(Ks, Vs, Tree1, Tree).
+	( bintree__insert(Tree0, K, V, Tree1) ->
+		Tree2 = Tree1
+	;
+		report_lookup_error("bintree__from_corresponding_lists: key already present", K, V)
+	),
+	bintree__from_corresponding_lists_2(Ks, Vs, Tree2, Tree).
 
 %-----------------------------------------------------------------------------%
 
@@ -463,3 +546,74 @@ bintree__branching_factor(tree(_K, _V, L, R), Ones, Twos) :-
 	).
 
 %-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
+% Ralph Becket <rwab1@cl.cam.ac.uk> 29/04/99
+% 	Function forms added.
+
+:- interface.
+
+:- func bintree__set(bintree(K,V), K, V) = bintree(K,V).
+
+:- func bintree__lookup(bintree(K,V), K) = V.
+
+:- func bintree__delete(bintree(K,V), K) = bintree(K,V).
+
+:- func bintree__keys(bintree(K,_V)) = list(K).
+
+:- func bintree__values(bintree(_K,V)) = list(V).
+
+:- func bintree__from_list(assoc_list(K,V)) = bintree(K,V).
+
+:- func bintree__from_sorted_list(assoc_list(K,V)) = bintree(K,V).
+
+:- func bintree__from_corresponding_lists(list(K), list(V)) = bintree(K,V).
+
+:- func bintree__to_list(bintree(K,V)) = assoc_list(K,V).
+
+:- func bintree__count(bintree(_K,_V)) = int.
+
+:- func bintree__depth(bintree(_K,_V)) = int.
+
+:- func bintree__balance(bintree(K, V)) = bintree(K, V).
+
+% ---------------------------------------------------------------------------- %
+% ---------------------------------------------------------------------------- %
+
+:- implementation.
+
+bintree__set(BT1, K, V) = BT2 :-
+	bintree__set(BT1, K, V, BT2).
+
+bintree__lookup(BT, K) = V :-
+	bintree__lookup(BT, K, V).
+
+bintree__delete(BT1, K) = BT2 :-
+	bintree__delete(BT1, K, BT2).
+
+bintree__keys(BT) = Ks :-
+	bintree__keys(BT, Ks).
+
+bintree__values(BT) = Vs :-
+	bintree__values(BT, Vs).
+
+bintree__from_list(AL) = BT :-
+	bintree__from_list(AL, BT).
+
+bintree__from_sorted_list(AL) = BT :-
+	bintree__from_sorted_list(AL, BT).
+
+bintree__from_corresponding_lists(Ks, Vs) = BT :-
+	bintree__from_corresponding_lists(Ks, Vs, BT).
+
+bintree__to_list(BT) = AL :-
+	bintree__to_list(BT, AL).
+
+bintree__count(BT) = N :-
+	bintree__count(BT, N).
+
+bintree__depth(BT) = N :-
+	bintree__depth(BT, N).
+
+bintree__balance(BT1) = BT2 :-
+	bintree__balance(BT1, BT2).
+
