@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1997, 1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %--------------------------------------------------------------------------%
@@ -213,8 +213,6 @@ bintree_set__is_member(E, S) :-
 
 %--------------------------------------------------------------------------%
 
-:- bintree_set__insert_list(_, Xs, _) when Xs.
-
 bintree_set__insert_list(S, [], S).
 bintree_set__insert_list(S0, [E|Es], S) :-
 	bintree_set__insert(S0, E, S1),
@@ -225,8 +223,6 @@ bintree_set__insert(S0, E, S) :-
 
 %--------------------------------------------------------------------------%
 
-:- bintree_set__remove_list(_, Xs, _) when Xs.
-
 bintree_set__remove_list(S, [], S).
 bintree_set__remove_list(S0, [X | Xs], S) :-
 	bintree_set__member(X, S0),
@@ -238,8 +234,6 @@ bintree_set__remove(S0, E, S) :-
 
 %--------------------------------------------------------------------------%
 
-:- bintree_set__delete_list(_, Xs, _) when Xs.
-
 bintree_set__delete_list(S, [], S).
 bintree_set__delete_list(S0, [X | Xs], S) :-
 	bintree_set__delete(S0, X, S1),
@@ -261,3 +255,60 @@ bintree_set__intersect(S0, S1, S) :-
 	bintree_set__delete_list(S0, L1, S).
 
 %--------------------------------------------------------------------------%
+%--------------------------------------------------------------------------%
+% Ralph Becket <rwab1@cl.cam.ac.uk> 29/04/99
+% 	Function forms added.
+
+:- interface.
+
+:- func bintree_set__list_to_set(list(T)) = bintree_set(T).
+
+:- func bintree_set__sorted_list_to_set(list(T)) = bintree_set(T).
+
+:- func bintree_set__to_sorted_list(bintree_set(T)) = list(T).
+
+:- func bintree_set__insert(bintree_set(T), T) = bintree_set(T).
+
+:- func bintree_set__insert_list(bintree_set(T), list(T)) = bintree_set(T).
+
+:- func bintree_set__delete(bintree_set(T), T) = bintree_set(T).
+
+:- func bintree_set__delete_list(bintree_set(T), list(T)) = bintree_set(T).
+
+:- func bintree_set__union(bintree_set(T), bintree_set(T)) = bintree_set(T).
+
+:- func bintree_set__intersect(bintree_set(T), bintree_set(T)) = bintree_set(T).
+
+% ---------------------------------------------------------------------------- %
+% ---------------------------------------------------------------------------- %
+
+:- implementation.
+
+bintree_set__list_to_set(Xs) = BT :-
+	bintree_set__list_to_set(Xs, BT).
+
+bintree_set__sorted_list_to_set(Xs) = BT :-
+	bintree_set__sorted_list_to_set(Xs, BT).
+
+bintree_set__to_sorted_list(BT) = Xs :-
+	bintree_set__to_sorted_list(BT, Xs).
+
+bintree_set__insert(BT1, X) = BT2 :-
+	bintree_set__insert(BT1, X, BT2).
+
+bintree_set__insert_list(BT1, Xs) = BT2 :-
+	bintree_set__insert_list(BT1, Xs, BT2).
+
+bintree_set__delete(BT1, X) = BT2 :-
+	bintree_set__delete(BT1, X, BT2).
+
+bintree_set__delete_list(BT1, Xs) = BT2 :-
+	bintree_set__delete_list(BT1, Xs, BT2).
+
+bintree_set__union(BT1, BT2) = BT3 :-
+	bintree_set__union(BT1, BT2, BT3).
+
+bintree_set__intersect(BT1, BT2) = BT3 :-
+	bintree_set__intersect(BT1, BT2, BT3).
+
+
