@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -14,9 +14,6 @@
 % Originally we used `character' rather than `char' for the type name
 % because `char' was used by NU-Prolog to mean something different.
 % But now we use `char' and the use of `character' is discouraged.
-%
-% NU-Prolog atoms can only include 7-bit ASCII chars, so the current
-% implementation does not support 8-bit characters.
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
@@ -373,8 +370,6 @@ char__digit_to_int(Digit, Int) :-
 
 %-----------------------------------------------------------------------------%
 
-:- char__lower_upper(X, Y) when X or Y.
-
 char__lower_upper('a', 'A').
 char__lower_upper('b', 'B').
 char__lower_upper('c', 'C').
@@ -404,16 +399,18 @@ char__lower_upper('z', 'Z').
 
 %-----------------------------------------------------------------------------%
 
-:- pragma c_code(char__to_int(Character::in, Int::out), will_not_call_mercury, "
+:- pragma c_code(char__to_int(Character::in, Int::out),
+               [will_not_call_mercury, thread_safe] , "
 	Int = (UnsignedChar) Character;
 ").
 
-:- pragma c_code(char__to_int(Character::in, Int::in), will_not_call_mercury, "
+:- pragma c_code(char__to_int(Character::in, Int::in),
+               [will_not_call_mercury, thread_safe] , "
 	SUCCESS_INDICATOR = ((UnsignedChar) Character == Int);
 ").
 
-:- pragma c_code(char__to_int(Character::out, Int::in), will_not_call_mercury,
-"
+:- pragma c_code(char__to_int(Character::out, Int::in),
+               [will_not_call_mercury, thread_safe] , "
 	/*
 	** If the integer doesn't fit into a char, then
 	** the assignment `Character = Int' below will truncate it.
@@ -430,8 +427,52 @@ char__lower_upper('z', 'Z').
 char__min_char_value(0).
 
 :- pragma c_header_code("#include <limits.h>").
-:- pragma c_code(char__max_char_value(Max::out), will_not_call_mercury, "
+:- pragma c_code(char__max_char_value(Max::out),
+               [will_not_call_mercury, thread_safe], "
 	Max = UCHAR_MAX;
 ").
-	
+
 %-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
+% Ralph Becket <rwab1@cl.cam.ac.uk> 27/04/99
+%       Functional forms added.
+
+:- interface.
+
+:- func char__to_int(char) = int.
+
+:- func char__max_char_value = int.
+
+:- func char__min_char_value = int.
+
+:- func char__to_upper(char) = char.
+
+:- func char__to_lower(char) = char.
+
+:- func char__det_int_to_digit(int) = char.
+
+% ---------------------------------------------------------------------------- %
+% ---------------------------------------------------------------------------- %
+
+:- implementation.
+
+char__to_int(C) = N :-
+	char__to_int(C, N).
+
+char__max_char_value = N :-
+	char__max_char_value(N).
+
+char__min_char_value = N :-
+	char__min_char_value(N).
+
+char__to_upper(C1) = C2 :-
+	char__to_upper(C1, C2).
+
+char__to_lower(C1) = C2 :-
+	char__to_lower(C1, C2).
+
+char__det_int_to_digit(N) = C :-
+	char__det_int_to_digit(N, C).
+
+
+
