@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1997, 1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -61,6 +61,11 @@
 :- pred eqvclass__same_eqvclass(eqvclass(T), T, T).
 :- mode eqvclass__same_eqvclass(in, in, in) is semidet.
 
+	% Test if a list of elements are equivalent.
+
+:- pred eqvclass__same_eqvclass_list(eqvclass(T), list(T)).
+:- mode eqvclass__same_eqvclass_list(in, in) is semidet.
+
 	% Return the set of the partitions of the equivalence class.
 
 :- pred eqvclass__partition_set(eqvclass(T), set(set(T))).
@@ -189,6 +194,21 @@ eqvclass__same_eqvclass(EqvClass0, Element1, Element2) :-
 	map__search(ElementMap0, Element2, Id2),
 	Id1 = Id2.
 
+eqvclass__same_eqvclass_list(_, []).
+eqvclass__same_eqvclass_list(EqvClass, [Element | Elements]) :-
+	EqvClass = eqvclass(_NextId, _PartitionMap, ElementMap),
+	map__search(ElementMap, Element, Id),
+	eqvclass__same_eqvclass_list_2(ElementMap, Elements, Id).
+
+:- pred eqvclass__same_eqvclass_list_2(map(T, partition_id), 
+		list(T), partition_id).
+:- mode eqvclass__same_eqvclass_list_2(in, in, in) is semidet.
+
+eqvclass__same_eqvclass_list_2(_, [], _).
+eqvclass__same_eqvclass_list_2(ElementMap, [Element | Elements], Id) :-
+	map__search(ElementMap, Element, Id),
+	eqvclass__same_eqvclass_list_2(ElementMap, Elements, Id).
+
 eqvclass__partition_set(EqvClass0, PartitionSet) :-
 	eqvclass__partition_ids(EqvClass0, Ids),
 	eqvclass__partitions(EqvClass0, Ids, PartitionList),
@@ -278,3 +298,57 @@ eqvclass__make_partition([Element | Elements], Id, ElementMap0, ElementMap) :-
 
 %---------------------------------------------------------------------------%
 %---------------------------------------------------------------------------%
+% Ralph Becket <rwab1@cl.cam.ac.uk> 29/04/99
+% 	Function forms added.
+
+:- interface.
+
+:- func eqvclass__init = eqvclass(T).
+
+:- func eqvclass__ensure_element(eqvclass(T), T) = eqvclass(T).
+
+:- func eqvclass__new_element(eqvclass(T), T) = eqvclass(T).
+
+:- func eqvclass__ensure_equivalence(eqvclass(T), T, T) = eqvclass(T).
+
+:- func eqvclass__new_equivalence(eqvclass(T), T, T) = eqvclass(T).
+
+:- func eqvclass__partition_set(eqvclass(T)) = set(set(T)).
+
+:- func eqvclass__partition_list(eqvclass(T)) = list(set(T)).
+
+:- func eqvclass__partition_set_to_eqvclass(set(set(T))) = eqvclass(T).
+
+:- func eqvclass__partition_list_to_eqvclass(list(set(T))) = eqvclass(T).
+
+% ---------------------------------------------------------------------------- %
+% ---------------------------------------------------------------------------- %
+
+:- implementation.
+
+eqvclass__init = EC :-
+	eqvclass__init(EC).
+
+eqvclass__ensure_element(EC1, X) = EC2 :-
+	eqvclass__ensure_element(EC1, X, EC2).
+
+eqvclass__new_element(EC1, X) = EC2 :-
+	eqvclass__new_element(EC1, X, EC2).
+
+eqvclass__ensure_equivalence(EC1, X, Y) = EC2 :-
+	eqvclass__ensure_equivalence(EC1, X, Y, EC2).
+
+eqvclass__new_equivalence(EC1, X, Y) = EC2 :-
+	eqvclass__new_equivalence(EC1, X, Y, EC2).
+
+eqvclass__partition_set(EC) = S :-
+	eqvclass__partition_set(EC, S).
+
+eqvclass__partition_list(EC) = Xs :-
+	eqvclass__partition_list(EC, Xs).
+
+eqvclass__partition_set_to_eqvclass(S) = EC :-
+	eqvclass__partition_set_to_eqvclass(S, EC).
+
+eqvclass__partition_list_to_eqvclass(Xs) = EC :-
+	eqvclass__partition_list_to_eqvclass(Xs, EC).
