@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -24,17 +24,50 @@
 % (An option can have as many names as you like, long or short.)
 % You must provide a predicate `option_default(Option, OptionData)'
 % which specifies both the type and the default value for every option.
-% We support five different option types: bool, int, string, maybe_string
-% (which have either a value of `no' or `yes(string)'), and
-% "accumulating" (which accumulates a list of strings).
-% For the first four option types, if there are multiple occurrences
+% You may optionally provide a predicate `special_handler(Option,
+% SpecialData, OptionTable, MaybeOptionTable)' for handling special
+% option types.  (See below.)
+%
+% We support the following "simple" option types:
+%
+%	- bool
+%	- int
+%	- maybe_int (which have a value of `no' or `yes(int)')
+%	- string
+%	- maybe_string (which have a value of `no' or `yes(string)')
+%
+% We also support one "accumulating" option type:
+%
+%	- accumulating (which accumulates a list of strings)
+%
+% And the following "special" option types:
+%
+%	- special
+%	- bool_special
+%	- int_special
+%	- string_special
+%
+% For the "simple" option types, if there are multiple occurrences
 % of the same option on the command-line, then the last (right-most)
 % occurrence will take precedence.  For "accumulating" options,
 % multiple occurrences will be appended together into a list.
-% Single-character boolean or maybe-string options can be negated by
-% following them with another `-', e.g. `-x-' will negate the `-x' option.
-% Long boolean or maybe-string options can be negated by preceding them with
-% `--no-', e.g. `--no-foo' will negate the `--foo' option.
+%
+% The "special" option types are handled by a special option handler
+% (see `special_handler' below), which may perform arbitrary
+% modifications to the option_table.  For example, an option which
+% is not yet implemented could be handled by a special handler which
+% produces an error report, or an option which is a synonym for a
+% set of more "primitive" options could be handled by a special
+% handler which sets those "primitive" options.
+%
+% It is an error to use a "special" option for which there is no
+% handler, or for which the handler fails.
+%
+% Single-character boolean (i.e. bool or bool_special) or maybe_*
+% options can be negated by following them with another `-', e.g. `-x-'
+% will negate the `-x' option.  Long boolean or maybe_* options can be
+% negated by preceding them with `--no-', e.g. `--no-foo' will negate
+% the `--foo' option.
 
 :- module getopt.
 :- interface.
@@ -64,7 +97,7 @@
 %	corresponding types and default values.
 %
 % :- pred special_handler(option::in, special_data::in,
-%	option_table::in, maybe_option_table::out) is semidet.
+%	option_table::in, maybe_option_table(_)::out) is semidet.
 %	This predicate is invoked whenever getopt finds an option
 %	(long or short) designated as special, with special_data holding
 %	the argument of the option (if any). The predicate can change the
@@ -114,6 +147,7 @@
 	--->	bool(bool)
 	;	int(int)
 	;	string(string)
+	;	maybe_int(maybe(int))
 	;	maybe_string(maybe(string))
 	;	accumulating(list(string))
 	;	special
@@ -147,6 +181,10 @@
 :- pred getopt__lookup_string_option(option_table(Option), Option, string).
 :- mode getopt__lookup_string_option(in, in, out) is det.
 
+:- pred getopt__lookup_maybe_int_option(option_table(Option), Option,
+		maybe(int)).
+:- mode getopt__lookup_maybe_int_option(in, in, out) is det.
+
 :- pred getopt__lookup_maybe_string_option(option_table(Option), Option,
 		maybe(string)).
 :- mode getopt__lookup_maybe_string_option(in, in, out) is det.
@@ -417,6 +455,19 @@ getopt__process_option(string(_), _Option, Flag, MaybeArg, _OptionOps,
 	;
 		error("string argument expected in getopt__process_option")
 	).
+getopt__process_option(maybe_int(_), Option, Flag, MaybeArg, _OptionOps,
+		OptionTable0, Result) :-
+	( MaybeArg = yes(Arg) ->
+		( string__to_int(Arg, IntArg) ->
+			map__set(OptionTable0, Flag, maybe_int(yes(IntArg)),
+					OptionTable),
+			Result = ok(OptionTable)
+		;
+			getopt__numeric_argument(Option, Arg, Result)
+		)
+	;
+		error("integer argument expected in getopt__process_option")
+	).
 getopt__process_option(maybe_string(_), _Option, Flag, MaybeArg, _OptionOps,
 		OptionTable0, Result) :-
 	( MaybeArg = yes(Arg) ->
@@ -482,6 +533,10 @@ process_negated_option(Option, Flag, OptionOps, OptionTable0, Result) :-
 		( OptionData = bool(_) ->
 			map__set(OptionTable0, Flag, bool(no), OptionTable),
 			Result = ok(OptionTable)
+		; OptionData = maybe_int(_) ->
+			map__set(OptionTable0, Flag, maybe_int(no),
+					OptionTable),
+			Result = ok(OptionTable)
 		; OptionData = maybe_string(_) ->
 			map__set(OptionTable0, Flag, maybe_string(no),
 					OptionTable),
@@ -532,6 +587,7 @@ getopt__process_special(Option, Flag, OptionData, OptionOps,
 getopt__need_arg(bool(_), no).
 getopt__need_arg(int(_), yes).
 getopt__need_arg(string(_), yes).
+getopt__need_arg(maybe_int(_), yes).
 getopt__need_arg(maybe_string(_), yes).
 getopt__need_arg(accumulating(_), yes).
 getopt__need_arg(special, no).
@@ -598,6 +654,13 @@ getopt__lookup_string_option(OptionTable, Opt, Val) :-
 		error("Expected string option and didn't get one.")
 	).
 
+getopt__lookup_maybe_int_option(OptionTable, Opt, Val) :-
+	( map__lookup(OptionTable, Opt, maybe_int(Val0)) ->
+		Val = Val0
+	;
+		error("Expected maybe_int option and didn't get one.")
+	).
+
 getopt__lookup_maybe_string_option(OptionTable, Opt, Val) :-
 	( map__lookup(OptionTable, Opt, maybe_string(Val0)) ->
 		Val = Val0
@@ -613,3 +676,47 @@ getopt__lookup_accumulating_option(OptionTable, Opt, Val) :-
 	).
 
 %-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
+% Ralph Becket <rwab1@cl.cam.ac.uk> 29/04/99
+% 	Functional forms added.
+
+:- interface.
+
+:- func getopt__lookup_bool_option(option_table(Option), Option) = bool.
+
+:- func getopt__lookup_int_option(option_table(Option), Option) = int.
+
+:- func getopt__lookup_string_option(option_table(Option), Option) = string.
+
+:- func getopt__lookup_maybe_int_option(option_table(Option), Option) =
+		maybe(int).
+
+:- func getopt__lookup_maybe_string_option(option_table(Option), Option) =
+		maybe(string).
+
+:- func getopt__lookup_accumulating_option(option_table(Option), Option) =
+		list(string).
+
+% ---------------------------------------------------------------------------- %
+% ---------------------------------------------------------------------------- %
+
+:- implementation.
+
+getopt__lookup_bool_option(OT, Opt) = B :-
+	getopt__lookup_bool_option(OT, Opt, B).
+
+getopt__lookup_int_option(OT, Opt) = N :-
+	getopt__lookup_int_option(OT, Opt, N).
+
+getopt__lookup_string_option(OT, Opt) = S :-
+	getopt__lookup_string_option(OT, Opt, S).
+
+getopt__lookup_maybe_int_option(OT, Opt) = MN :-
+	getopt__lookup_maybe_int_option(OT, Opt, MN).
+
+getopt__lookup_maybe_string_option(OT, Opt) =MS :-
+	getopt__lookup_maybe_string_option(OT, Opt, MS).
+
+getopt__lookup_accumulating_option(OT, Opt) =Ss :-
+	getopt__lookup_accumulating_option(OT, Opt, Ss).
+
