@@ -1,14 +1,21 @@
-%---------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+%-----------------------------------------------------------------------------%
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
-%---------------------------------------------------------------------------%
-%
-% int - some predicates for dealing with machine-size integer numbers.
+%-----------------------------------------------------------------------------%
 %
+% File: int.m.
 % Main authors: conway, fjh.
 % Stability: medium.
 %
+% Predicates and functions for dealing with machine-size integer numbers.
+%
+% The behaviour of a computation for which overflow occurs is undefined.
+% (In the current implementation, the predicates and functions in this
+% module do not check for overflow, and the results you get are those
+% delivered by the C compiler.  However, future implementations
+% might check for overflow.)
+%
 %-----------------------------------------------------------------------------%
 
 :- module int.
@@ -75,7 +82,6 @@
 :- mode uo  * in  = in  is semidet.
 :- mode in  * uo  = in  is semidet.
 */
-
 	% subtraction
 :- func int - int = int.
 :- mode in  - in  = uo  is det.
@@ -106,14 +112,34 @@
 :- func int rem int = int.
 :- mode in rem in = uo is det.
 
-	% left shift
+	% Left shift.
+	% X << Y returns X "left shifted" by Y bits. 
+	% To be precise, if Y is negative, the result is
+	% X div (2^(-Y)), otherwise the result is X * (2^Y).
 :- func int << int = int.
 :- mode in  << in  = uo  is det.
 
-	% (arithmetic) right shift
+	% unchecked_left_shift(X, Y) is the same as X << Y
+	% except that the behaviour is undefined if Y is negative,
+	% or greater than or equal to the result of `int__bits_per_int/1'.
+	% It will typically be implemented more efficiently than X << Y.
+:- func unchecked_left_shift(int, int) = int.
+:- mode unchecked_left_shift(in, in) = uo is det.
+
+	% Right shift.
+	% X >> Y returns X "arithmetic right shifted" by Y bits.
+	% To be precise, if Y is negative, the result is
+	% X * (2^(-Y)), otherwise the result is X div (2^Y).
 :- func int >> int = int.
 :- mode in  >> in  = uo  is det.
 
+	% unchecked_right_shift(X, Y) is the same as X >> Y
+	% except that the behaviour is undefined if Y is negative,
+	% or greater than or equal to the result of `int__bits_per_int/1'.
+	% It will typically be implemented more efficiently than X >> Y.
+:- func unchecked_right_shift(int, int) = int.
+:- mode unchecked_right_shift(in, in) = uo is det.
+
 	% bitwise and
 :- func int /\ int = int.
 :- mode in  /\ in  = uo  is det.
@@ -123,8 +149,15 @@
 :- mode in  \/ in  = uo  is det.
 
 	% bitwise exclusive or (xor)
+:- func int__xor(int, int) = int.
+:- mode int__xor(in, in) = uo is det.
+
+	% bitwise exclusive or (xor)
+	% This version will be removed soon - the operator
+	% is needed for record syntax.
 :- func int ^ int = int.
 :- mode in  ^ in  = uo  is det.
+:- pragma obsolete(('^')/2).
 
 	% bitwise complement
 :- func \ int = int.
@@ -156,51 +189,86 @@
 	% on this machine.
 :- pred int__bits_per_int(int::out) is det.
 
-/* The following routines are builtins that the compiler knows about.
-   Don't use them; use the functions above.
-   These will go away in some future release.
-*/
+%-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
 
+:- implementation.
+:- interface.
+
+	% Everything below here will not appear in the
+	% Mercury Library Reference Manual.
+
+%-----------------------------------------------------------------------------%
+
+	% commutivity of +
+:- promise all [A,B,C] ( C = B + A <=> C = A + B ).
+
+	% commutivity of *
+:- promise all [A,B,C] ( C = B * A <=> C = A * B ).
+
+%-----------------------------------------------------------------------------%
+
+%
+% The following routines are builtins that the compiler knows about.
+% Don't use them; use the functions above.
+% These will go away in some future release.
+%
+
+:- pragma obsolete(builtin_plus/3).
 :- pred builtin_plus(int, int, int).
 :- mode builtin_plus(in, in, uo) is det.
-:- mode builtin_plus(in, in, uo) is det.
 
+:- pragma obsolete(builtin_unary_plus/2).
 :- pred builtin_unary_plus(int, int).
 :- mode builtin_unary_plus(in, uo) is det.
 
+:- pragma obsolete(builtin_minus/3).
 :- pred builtin_minus(int, int, int).
 :- mode builtin_minus(in, in, uo) is det.
 
+:- pragma obsolete(builtin_unary_minus/2).
 :- pred builtin_unary_minus(int, int).
 :- mode builtin_unary_minus(in, uo) is det.
 
+:- pragma obsolete(builtin_times/3).
 :- pred builtin_times(int, int, int).
 :- mode builtin_times(in, in, uo) is det.
 
+:- pragma obsolete(builtin_div/3).
 :- pred builtin_div(int, int, int).
 :- mode builtin_div(in, in, uo) is det.
 
+:- pragma obsolete(builtin_mod/3).
 :- pred builtin_mod(int, int, int).
 :- mode builtin_mod(in, in, uo) is det.
 
+:- pragma obsolete(builtin_left_shift/3).
 :- pred builtin_left_shift(int, int, int).
 :- mode builtin_left_shift(in, in, uo) is det.
 
+:- pragma obsolete(builtin_right_shift/3).
 :- pred builtin_right_shift(int, int, int).
 :- mode builtin_right_shift(in, in, uo) is det.
 
+:- pragma obsolete(builtin_bit_or/3).
 :- pred builtin_bit_or(int, int, int).
 :- mode builtin_bit_or(in, in, uo) is det.
 
+:- pragma obsolete(builtin_bit_and/3).
 :- pred builtin_bit_and(int, int, int).
 :- mode builtin_bit_and(in, in, uo) is det.
 
+:- pragma obsolete(builtin_bit_xor/3).
 :- pred builtin_bit_xor(int, int, int).
 :- mode builtin_bit_xor(in, in, uo) is det.
 
+:- pragma obsolete(builtin_bit_neg/2).
 :- pred builtin_bit_neg(int, int).
 :- mode builtin_bit_neg(in, uo) is det.
 
+%-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
+
 :- implementation.
 :- import_module require.
 
@@ -209,9 +277,12 @@
 
 X div Y = Div :-
 	Trunc = X // Y,
-	( X // Y >= 0 ->
-		Div = Trunc
-	; X rem Y = 0 ->
+	(
+		( X >= 0, Y >= 0
+		; X < 0, Y < 0
+		; X rem Y = 0
+		)
+	->
 		Div = Trunc
 	;
 		Div = Trunc - 1
@@ -219,6 +290,40 @@ X div Y = Div :-
 
 X mod Y = X - (X div Y) * Y.
 
+X << Y = Z :-
+	int__bits_per_int(IntBits),
+	( Y >= 0 ->
+		( Y >= IntBits ->
+			Z = 0
+		;
+			Z = unchecked_left_shift(X, Y)
+		)
+	;
+		( Y =< -IntBits ->
+			Z = (if X >= 0 then 0 else -1)
+		;
+			Z = unchecked_right_shift(X, -Y)
+		)
+	).
+
+	% Note: this assumes two's complement arithmetic.
+	% tests/hard_coded/shift_test.m will fail if this is not the case.
+X >> Y = Z :-
+	int__bits_per_int(IntBits),
+	( Y >= 0 ->
+		( Y >= IntBits ->
+			Z = (if X >= 0 then 0 else -1)
+		;
+			Z = unchecked_right_shift(X, Y)
+		)
+	;
+		( Y =< -IntBits ->
+			Z = 0
+		;
+			Z = unchecked_left_shift(X, -Y)
+		)
+	).
+
 int__abs(Num, Abs) :-
 	(
 		Num < 0
@@ -290,11 +395,7 @@ int__log2_2(X, N0, N) :-
 % is/2 is replaced with `=' in the parser, but the following is useful
 % in case you should take the address of `is' or something weird like that.
 
-% we use pragma(c_code) to avoid complaints about redefinition of is/2
-% from the Prolog compilers.
-
-:- pragma(c_code, is(X::uo, Y::di),  "X = Y;").
-:- pragma(c_code, is(X::out, Y::in), "X = Y;").
+is(X, X).
 
 %-----------------------------------------------------------------------------%
 
@@ -302,7 +403,9 @@ int__log2_2(X, N0, N) :-
 :- pred int__to_float(int, float) is det.
 :- mode int__to_float(in, out) is det.
 */
-:- pragma(c_code, int__to_float(IntVal::in, FloatVal::out), "
+:- pragma c_code(int__to_float(IntVal::in, FloatVal::out),
+		will_not_call_mercury,
+"
 	FloatVal = IntVal;
 ").
 
@@ -313,7 +416,7 @@ int__log2_2(X, N0, N) :-
 ").
 
 
-:- pragma c_code(int__max_int(Max::out), "
+:- pragma c_code(int__max_int(Max::out), will_not_call_mercury, "
 	if (sizeof(Integer) == sizeof(int))
 		Max = INT_MAX;
 	else if (sizeof(Integer) == sizeof(long))
@@ -322,7 +425,7 @@ int__log2_2(X, N0, N) :-
 		fatal_error(""Unable to figure out max integer size"");
 ").
 
-:- pragma c_code(int__min_int(Min::out), "
+:- pragma c_code(int__min_int(Min::out), will_not_call_mercury, "
 	if (sizeof(Integer) == sizeof(int))
 		Min = INT_MIN;
 	else if (sizeof(Integer) == sizeof(long))
@@ -331,7 +434,46 @@ int__log2_2(X, N0, N) :-
 		fatal_error(""Unable to figure out min integer size"");
 ").
 
-:- pragma c_code(int__bits_per_int(Bits::out), "
+:- pragma c_code(int__bits_per_int(Bits::out), will_not_call_mercury, "
 	Bits = sizeof(Integer) * CHAR_BIT;
 ").
 
+%-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
+% Ralph Becket <rwab1@cl.cam.ac.uk> 27/04/99
+% 	Functional forms added.
+
+:- interface.
+
+:- func int__plus(int, int) = int.
+
+:- func int__times(int, int) = int.
+
+:- func int__minus(int, int) = int.
+
+:- func int__max_int = int.
+
+:- func int__min_int = int.
+
+:- func int__bits_per_int = int.
+
+% ---------------------------------------------------------------------------- %
+% ---------------------------------------------------------------------------- %
+
+:- implementation.
+
+int__plus(X, Y) = X + Y.
+
+int__times(X, Y) = X * Y.
+
+int__minus(X, Y) = X - Y.
+
+int__max_int = X :-
+	int__max_int(X).
+
+int__min_int = X :-
+	int__min_int(X).
+
+int__bits_per_int = X :-
+	int__bits_per_int(X).
+
