@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1993-1997 The University of Melbourne.
+% Copyright (C) 1993-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -32,6 +32,9 @@
 :- import_module std_util, string, term, term_io, tree234, varset.
 :- import_module store, rbtree, parser, lexer, ops.
 :- import_module prolog.
+:- import_module integer, rational.
+:- import_module exception, gc.
+:- import_module time.
 
 % library__version must be implemented using pragma c_code,
 % so we can get at the MR_VERSION and MR_FULLARCH configuration
@@ -39,8 +42,15 @@
 % at configuration time, because that would cause bootstrapping problems --
 % might not have a Mercury compiler around to compile library.m with.
 
-:- pragma c_code(library__version(Version::out), will_not_call_mercury,
-	"Version = MR_VERSION "", configured for "" MR_FULLARCH;").
+:- pragma c_code(library__version(Version::out), will_not_call_mercury, "
+	ConstString version_string = 
+		MR_VERSION "", configured for "" MR_FULLARCH;
+	/*
+	** Cast away const needed here, because Mercury declares Version
+	** with type String rather than ConstString.
+	*/
+	Version = (String) (Word) version_string;
+").
 
 %---------------------------------------------------------------------------%
 %---------------------------------------------------------------------------%
