@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1993-1997 The University of Melbourne.
+% Copyright (C) 1993-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -149,11 +149,19 @@
 
 	% list__take(Len, List, Start):
 	%	`Start' is the first `Len' elements of `List'.
+	%	Fails if `List' has less than `Len' elements.
 	%	See also: list__split_list.
 	%
 :- pred list__take(int, list(T), list(T)).
 :- mode list__take(in, in, out) is semidet.
 
+	% list__take_upto(Len, List, Start):
+	%	`Start' is the first `Len' elements of `List'.
+	%	If `List' has less than `Len' elements, return the entire list.
+	%
+:- pred list__take_upto(int, list(T), list(T)).
+:- mode list__take_upto(in, in, out) is det.
+
 	% list__drop(Len, List, End):
 	%	`End' is the remainder of `List' after removing the
 	%	first `Len' elements.
@@ -221,6 +229,22 @@
 :- pred list__replace_all(list(T), T, T, list(T)).
 :- mode list__replace_all(in, in, in, out) is det.
 
+	% list__replace_nth(List0, N, R, List) is true iff List is List0 
+	% with Nth element replaced with R.
+	% Fails if N < 1 or if length of List0 < N.
+	% (Position numbers start from 1.)
+	%
+:- pred list__replace_nth(list(T), int, T, list(T)).
+:- mode list__replace_nth(in, in, in, out) is semidet.
+
+	% list__replace_nth_det(List0, N, R, List) is true iff List is List0 
+	% with Nth element replaced with R.
+	% Aborts if N < 1 or if length of List0 < N.
+	% (Position numbers start from 1.)
+	%
+:- pred list__replace_nth_det(list(T), int, T, list(T)).
+:- mode list__replace_nth_det(in, in, in, out) is det.
+
 	% list__sort_and_remove_dups(List0, List):
 	%	List is List0 sorted with duplicates removed.
 	%
@@ -248,6 +272,7 @@
 
 	% list__nth_member_search(List, Elem, Position):
 	%	Elem is the Position'th member of List.
+	% 	(Position numbers start from 1.)
 	%
 :- pred list__nth_member_search(list(T), T, int).
 :- mode list__nth_member_search(in, in, out) is semidet.
@@ -329,6 +354,7 @@
 :- mode list__map(pred(in, out) is semidet, in, out) is semidet.
 :- mode list__map(pred(in, out) is multi, in, out) is multi.
 :- mode list__map(pred(in, out) is nondet, in, out) is nondet.
+:- mode list__map(pred(in, in) is semidet, in, in) is semidet.
 
 	% list__foldl(Pred, List, Start, End) calls Pred with each
 	% element of List (working left-to-right) and an accumulator
@@ -338,6 +364,7 @@
 :- mode list__foldl(pred(in, di, uo) is det, in, di, uo) is det.
 :- mode list__foldl(pred(in, in, out) is det, in, in, out) is det.
 :- mode list__foldl(pred(in, in, out) is semidet, in, in, out) is semidet.
+:- mode list__foldl(pred(in, in, out) is nondet, in, in, out) is nondet.
 
 	% list__foldr(Pred, List, Start, End) calls Pred with each
 	% element of List (working right-to-left) and an accumulator
@@ -346,6 +373,7 @@
 :- pred list__foldr(pred(X, Y, Y), list(X), Y, Y).
 :- mode list__foldr(pred(in, in, out) is det, in, in, out) is det.
 :- mode list__foldr(pred(in, in, out) is semidet, in, in, out) is semidet.
+:- mode list__foldr(pred(in, in, out) is nondet, in, in, out) is nondet.
 
 	% list__foldl2(Pred, List, Start, End, Start2, End2) 
 	% calls Pred with each element of List (working left-to-right),
@@ -356,6 +384,10 @@
 :- pred list__foldl2(pred(X, Y, Y, Z, Z), list(X), Y, Y, Z, Z).
 :- mode list__foldl2(pred(in, in, out, in, out) is det,
 		in, in, out, in, out) is det.
+:- mode list__foldl2(pred(in, in, out, in, out) is semidet,
+		in, in, out, in, out) is semidet.
+:- mode list__foldl2(pred(in, in, out, in, out) is nondet,
+		in, in, out, in, out) is nondet.
 :- mode list__foldl2(pred(in, in, out, mdi, muo) is det,
 		in, in, out, mdi, muo) is det.
 :- mode list__foldl2(pred(in, in, out, di, uo) is det,
@@ -375,6 +407,8 @@
 								is det.
 :- mode list__map_foldl(pred(in, out, in, out) is semidet, in, out, in, out)
                                                                 is semidet.
+:- mode list__map_foldl(pred(in, out, in, out) is nondet, in, out, in, out)
+                                                                is nondet.
 
 	% list__filter(Pred, List, TrueList) takes a closure with one
 	% input argument and for each member of List `X', calls the closure.
@@ -461,8 +495,6 @@ list__append([], Ys, Ys).
 list__append([X | Xs], Ys, [X | Zs]) :-
 	list__append(Xs, Ys, Zs).
 
-:- list__remove_suffix(_List, Suffix, _Prefix) when Suffix.
-
 list__remove_suffix(List, Suffix, Prefix) :-
 	list__length(List, ListLength),
 	list__length(Suffix, SuffixLength),
@@ -576,6 +608,40 @@ list__replace_all([X | Xs], Y, Z, L) :-
 		list__replace_all(Xs, Y, Z, L0)
 	).
 
+list__replace_nth(Xs, P, R, L) :-
+	P > 0,
+	list__replace_nth_2(Xs, P, R, L).
+
+list__replace_nth_det(Xs, P, R, L) :-
+	(
+		P > 0
+	->
+		(
+			list__replace_nth_2(Xs, P, R, L0)
+		->
+			L = L0
+		;
+			error("list__replace_nth_det: Can't replace element whose index position is past the end of the list")
+		)
+	;
+		error("list__replace_nth_det: Can't replace element whose index position is less than 1.")
+	).
+
+
+:- pred list__replace_nth_2(list(T), int, T, list(T)).
+:- mode list__replace_nth_2(in, in, in, out) is semidet.
+
+list__replace_nth_2([X | Xs], P, R, L) :-
+	(
+		P = 1 
+	->
+		L = [R | Xs]
+	;
+		P1 is P - 1,
+		list__replace_nth(Xs, P1, R, L0),
+		L = [X | L0]
+	).
+
 %-----------------------------------------------------------------------------%
 
 list__member(X, [X | _]).
@@ -753,8 +819,6 @@ list__zip([A|As], Bs, [A|Cs]) :-
 :- pred list__zip2(list(T), list(T), list(T)).
 :- mode list__zip2(in, in, out) is det.
 
-:- list__zip2(_, Bs, _) when Bs. % NU-Prolog indexing
-
 list__zip2(As, [], As).
 list__zip2(As, [B|Bs], [B|Cs]) :-
 	list__zip(As, Bs, Cs).
@@ -797,13 +861,20 @@ list__take(N, As, Bs) :-
 		N > 0
 	->
 		N1 is N - 1,
-		As = [A|As1],
-		Bs = [A|Bs1],
+		As = [A | As1],
+		Bs = [A | Bs1],
 		list__take(N1, As1, Bs1)
 	;
 		Bs = []
 	).
 
+list__take_upto(N, As, Bs) :-
+	( list__take(N, As, Bs0) ->
+		Bs = Bs0
+	;
+		Bs = As
+	).
+
 list__drop(N, As, Bs) :-
 	(
 		N > 0
@@ -1049,16 +1120,186 @@ list__merge_and_remove_dups(P, [H1|T1], [H2|T2], L) :-
 	(
 		C = (<),
 		L = [H1|T],   
-		list__merge(P, T1, [H2|T2], T)
+		list__merge_and_remove_dups(P, T1, [H2|T2], T)
 	;
 		C = (=),
 		L = [H1 | T],
-		list__merge(P, T1, T2, T)
+		list__merge_and_remove_dups(P, T1, T2, T)
 	;
 		C = (>),
 		L = [H2|T],   
-		list__merge(P, [H1|T1], T2, T)
+		list__merge_and_remove_dups(P, [H1|T1], T2, T)
 	).
 
 
 %-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
+% Ralph Becket <rwab1@cl.cam.ac.uk> 27/04/99
+% 	Functional forms added.
+
+:- interface.
+
+:- func list__det_head(list(T)) = T.
+
+:- func list__det_tail(list(T)) = list(T).
+
+:- func list__append(list(T), list(T)) = list(T).
+
+:- func list__merge(list(T), list(T)) = list(T).
+
+:- func list__merge_and_remove_dups(list(T), list(T)) = list(T).
+
+:- func list__remove_adjacent_dups(list(T)) = list(T).
+
+:- func list__remove_dups(list(T)) = list(T).
+
+:- func list__length(list(T)) = int.
+
+:- func list__take_upto(int, list(T)) = list(T).
+
+:- func list__delete_all(list(T), T) = list(T).
+
+:- func list__delete_elems(list(T), list(T)) = list(T).
+
+:- func list__replace_all(list(T), T, T) = list(T).
+
+:- func list__replace_nth_det(list(T), int, T) = list(T).
+
+:- func list__sort_and_remove_dups(list(T)) = list(T).
+
+:- func list__sort(list(T)) = list(T).
+
+:- func list__reverse(list(T)) = list(T).
+
+:- func list__index0_det(list(T), int) = T.
+
+:- func list__index1_det(list(T), int) = T.
+
+:- func list__zip(list(T), list(T)) = list(T).
+
+:- func list__duplicate(int, T) = list(T).
+
+:- func list__condense(list(list(T))) = list(T).
+
+:- func list__chunk(list(T), int) = list(list(T)).
+
+:- func list__map(func(X) = Y, list(X)) = list(Y).
+
+:- func list__foldl(func(X, Y) = Y, list(X), Y) = Y.
+
+:- func list__foldr(func(X, Y) = Y, list(X), Y) = Y.
+
+:- func list__filter(pred(X), list(X)) = list(X).
+:- mode list__filter(pred(in) is semidet, in) = out is det.
+
+:- func list__filter_map(func(X) = Y, list(X)) = list(Y).
+:- mode list__filter_map(func(in) = out is semidet, in) = out is det.
+
+:- func list__sort(func(X, X) = comparison_result, list(X)) = list(X).
+
+:- func list__merge(func(X, X) = comparison_result, list(X), list(X)) = list(X).
+
+:- func list__merge_and_remove_dups(func(X, X) = comparison_result, list(X), list(X)) = list(X).
+
+% ---------------------------------------------------------------------------- %
+% ---------------------------------------------------------------------------- %
+
+:- implementation.
+
+list__det_head([]) = _ :- error("list__det_head/1: empty list as argument").
+list__det_head([X | _]) = X.
+
+list__det_tail([]) = _ :- error("list__det_tail/1: empty list as argument").
+list__det_tail([_ | Xs]) = Xs.
+
+list__append(Xs, Ys) = Zs :-
+	list__append(Xs, Ys, Zs).
+
+list__merge(Xs, Ys) = Zs :-
+	list__merge(Xs, Ys, Zs).
+
+list__merge_and_remove_dups(Xs, Ys) = Zs :-
+	list__merge_and_remove_dups(Xs, Ys, Zs).
+
+list__remove_adjacent_dups(Xs) = Ys :-
+	list__remove_adjacent_dups(Xs, Ys).
+
+list__remove_dups(Xs) = Ys :-
+	list__remove_dups(Xs, Ys).
+
+list__length(Xs) = N :-
+	list__length(Xs, N).
+
+list__take_upto(N, Xs) = Ys :-
+	list__take_upto(N, Xs, Ys).
+
+list__delete_all(Xs, A) = Ys :-
+	list__delete_all(Xs, A, Ys).
+
+list__delete_elems(Xs, Ys) = Zs :-
+	list__delete_elems(Xs, Ys, Zs).
+
+list__replace_all(Xs, A, B) = Ys :-
+	list__replace_all(Xs, A, B, Ys).
+
+list__replace_nth_det(Xs, N, A) = Ys :-
+	list__replace_nth_det(Xs, N, A, Ys).
+
+list__sort_and_remove_dups(Xs) = Ys :-
+	list__sort_and_remove_dups(Xs, Ys).
+
+list__sort(Xs) = Ys :-
+	list__sort(Xs, Ys).
+
+list__reverse(Xs) = Ys :-
+	list__reverse(Xs, Ys).
+
+list__index0_det(Xs, N) = A :-
+	list__index0_det(Xs, N, A).
+
+list__index1_det(Xs, N) = A :-
+	list__index1_det(Xs, N, A).
+
+list__zip(Xs, Ys) = Zs :-
+	list__zip(Xs, Ys, Zs).
+
+list__duplicate(N, A) = Xs :-
+	list__duplicate(N, A, Xs).
+
+list__condense(Xss) = Ys :-
+	list__condense(Xss, Ys).
+
+list__chunk(Xs, N) = Ys :-
+	list__chunk(Xs, N, Ys).
+
+list__map(F, Xs) = Ys :-
+	P = ( pred(X::in, Y::out) is det :- Y = F(X) ),
+	list__map(P, Xs, Ys).
+
+list__foldl(F, Xs, A) = B :-
+	P = ( pred(X::in, Y::in, Z::out) is det :- Z = F(X, Y) ),
+	list__foldl(P, Xs, A, B).
+
+list__foldr(F, Xs, A) = B :-
+	P = ( pred(X::in, Y::in, Z::out) is det :- Z = F(X, Y) ),
+	list__foldr(P, Xs, A, B).
+
+list__filter(P, Xs) = Ys :-
+	list__filter(P, Xs, Ys).
+
+list__filter_map(F, Xs) = Ys :-
+	P = ( pred(X::in, Y::out) is semidet :- Y = F(X) ),
+	list__filter_map(P, Xs, Ys).
+
+list__sort(F, Xs) = Ys :-
+	P = ( pred(X::in, Y::in, Z::out) is det :- Z = F(X, Y) ),
+	list__sort(P, Xs, Ys).
+
+list__merge(F, Xs, Ys) = Zs :-
+	P = ( pred(X::in, Y::in, Z::out) is det :- Z = F(X, Y) ),
+	list__merge(P, Xs, Ys, Zs).
+
+list__merge_and_remove_dups(F, Xs, Ys) = Zs :-
+	P = ( pred(X::in, Y::in, Z::out) is det :- Z = F(X, Y) ),
+	list__merge_and_remove_dups(P, Xs, Ys, Zs).
+
