@@ -1,20 +1,40 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
 %
 % File: math.m
 % Main author: bromage
-% Stability: high (but as yet no Prolog implementation)
+% Stability: high
 %
 % Higher mathematical operations.  (The basics are in float.m.)
-% The predicates in this module are not yet implemented in Prolog.
 %
-% Domain errors are currently handled by a program abort.  This is
-% because Mercury currently does not have exceptions built in.
-% Exception-handling would be nice, but it's kind of low on the
-% priority scale.
+% By default, domain errors are currently handled by a program abort.
+% This is because Mercury originally did not have exceptions built in.
+%
+% For better performance, it is possible to disable the Mercury domain
+% checking by compiling with `--intermodule-optimization' and the C macro
+% symbol `ML_OMIT_MATH_DOMAIN_CHECKS' defined, e.g. by using
+% `MCFLAGS=--intermodule-optimization' and
+% `MGNUCFLAGS=-DML_OMIT_MATH_DOMAIN_CHECKS' in your Mmakefile,
+% or by compiling with the command
+% `mmc --intermodule-optimization --cflags -DML_OMIT_MATH_DOMAIN_CHECKS'.
+%
+% For maximum performance, all Mercury domain checking can be disabled by
+% recompiling this module using `MGNUCFLAGS=-DML_OMIT_MATH_DOMAIN_CHECKS'
+% or `mmc --cflags -DML_OMIT_MATH_DOMAIN_CHECKS' as above. You can
+% either recompile the entire library, or just copy `math.m' to your
+% application's source directory and link with it directly instead of as
+% part of the library.
+%
+% Note that the above performance improvements are semantically safe,
+% since the C math library and/or floating point hardware perform these
+% checks for you.  The benefit of having the Mercury library perform the
+% checks instead is that Mercury will tell you in which function or
+% predicate the error occurred, as well as giving you a stack trace if
+% that is enabled; with the checks disabled you only have the information
+% that the floating-point exception signal handler gives you.
 %
 %---------------------------------------------------------------------------%
 
@@ -170,36 +190,44 @@
 
 % These operations are all implemented using the C interface.
 
-:- pragma(c_header_code, "
+:- pragma c_header_code("
 
 	#include <math.h>
 
 	/*
 	** Mathematical constants.
+	**
+	** The maximum number of significant decimal digits which
+	** can be packed into an IEEE-754 extended precision
+	** floating point number is 18.  Therefore 20 significant
+	** decimal digits for these constants should be plenty.
 	*/
 
-	#define	MERCURY_FLOAT__E		2.7182818284590452354
-	#define	MERCURY_FLOAT__PI		3.1415926535897932384
-	#define	MERCURY_FLOAT__LN2		0.69314718055994530941
+	#define	ML_FLOAT_E		2.7182818284590452354
+	#define	ML_FLOAT_PI		3.1415926535897932384
+	#define	ML_FLOAT_LN2		0.69314718055994530941
 
-	void mercury_domain_error(const char *where);
+	void ML_math_domain_error(const char *where);
 
 "). % end pragma c_header_code
 
-:- pragma(c_code, "
+:- pragma c_code("
 
+	#include ""mercury_trace_base.h""
 	#include <stdio.h>
 
 	/*
 	** Handle domain errors.
 	*/
 	void
-	mercury_domain_error(const char *where)
+	ML_math_domain_error(const char *where)
 	{
 		fflush(stdout);
 		fprintf(stderr,
-			""Software error: Domain error in call to `%s'\n"",
+			""Software error: Domain error in call to `%s'\\n"",
 			where);
+		MR_trace_report(stderr);
+		MR_dump_stack(MR_succip, MR_sp, MR_curfr, FALSE);
 		exit(1);
 	}
 
@@ -209,20 +237,21 @@
 % Mathematical constants from math.m
 %
 	% Pythagoras' number
-:- pragma c_code(math__pi = (Pi::out),will_not_call_mercury,"
-	Pi = MERCURY_FLOAT__PI;
+:- pragma c_code(math__pi = (Pi::out), [will_not_call_mercury, thread_safe],"
+	Pi = ML_FLOAT_PI;
 ").
 
 	% Base of natural logarithms
-:- pragma c_code(math__e = (E::out),will_not_call_mercury,"
-	E = MERCURY_FLOAT__E;
+:- pragma c_code(math__e = (E::out), [will_not_call_mercury, thread_safe],"
+	E = ML_FLOAT_E;
 ").
 
 %
 % math__ceiling(X) = Ceil is true if Ceil is the smallest integer
 % not less than X.
 %
-:- pragma c_code(math__ceiling(Num::in) = (Ceil::out), will_not_call_mercury,"
+:- pragma c_code(math__ceiling(Num::in) = (Ceil::out),
+		[will_not_call_mercury, thread_safe],"
 	Ceil = ceil(Num);
 ").
 
@@ -230,7 +259,8 @@
 % math__floor(X) = Floor is true if Floor is the largest integer
 % not greater than X.
 %
-:- pragma c_code(math__floor(Num::in) = (Floor::out), will_not_call_mercury,"
+:- pragma c_code(math__floor(Num::in) = (Floor::out),
+		[will_not_call_mercury, thread_safe],"
 	Floor = floor(Num);
 ").
 
@@ -239,7 +269,8 @@
 % closest to X.  If X has a fractional component of 0.5,
 % it is rounded up.
 %
-:- pragma c_code(math__round(Num::in) = (Rounded::out), will_not_call_mercury,"
+:- pragma c_code(math__round(Num::in) = (Rounded::out),
+		[will_not_call_mercury, thread_safe],"
 	Rounded = floor(Num+0.5);
 ").
 
@@ -247,11 +278,12 @@
 % math__truncate(X) = Trunc is true if Trunc is the integer
 % closest to X such that |Trunc| =< |X|.
 %
-:- pragma c_code(math__truncate(X::in) = (Trunc::out), will_not_call_mercury,"
+:- pragma c_code(math__truncate(X::in) = (Trunc::out),
+		[will_not_call_mercury, thread_safe],"
 	if (X < 0.0) {
-	    Trunc = ceil(X);
+		Trunc = ceil(X);
 	} else {
-	    Trunc = floor(X);
+		Trunc = floor(X);
 	}
 ").
 
@@ -262,10 +294,13 @@
 % Domain restrictions:
 %		X >= 0
 %
-:- pragma c_code(math__sqrt(X::in) = (SquareRoot::out), will_not_call_mercury,"
+:- pragma c_code(math__sqrt(X::in) = (SquareRoot::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X < 0.0) {
-	    mercury_domain_error(""math__sqrt"");
+		ML_math_domain_error(""math__sqrt"");
 	}
+#endif
 	SquareRoot = sqrt(X);
 ").
 
@@ -277,25 +312,31 @@
 %		X >= 0
 %		X = 0 implies Y > 0
 %
-:- pragma c_code(math__pow(X::in, Y::in) = (Res::out), will_not_call_mercury,"
+:- pragma c_code(math__pow(X::in, Y::in) = (Res::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X < 0.0) {
-	    mercury_domain_error(""math__pow"");
+		ML_math_domain_error(""math__pow"");
 	}
 	if (X == 0.0) {
-	    if (Y <= 0.0) {
-		mercury_domain_error(""math__pow"");
-	    }
-	    Res = 0.0;
+		if (Y <= 0.0) {
+			ML_math_domain_error(""math__pow"");
+		}
+		Res = 0.0;
 	} else {
-	    Res = pow(X, Y);
+		Res = pow(X, Y);
 	}
+#else
+	Res = pow(X, Y);
+#endif
 ").
 
 %
 % math__exp(X) = Exp is true if Exp is X raised to the
 % power of e.
 %
-:- pragma c_code(math__exp(X::in) = (Exp::out), will_not_call_mercury,"
+:- pragma c_code(math__exp(X::in) = (Exp::out),
+		[will_not_call_mercury, thread_safe],"
 	Exp = exp(X);
 ").
 
@@ -306,10 +347,13 @@
 % Domain restrictions:
 %		X > 0
 %
-:- pragma c_code(math__ln(X::in) = (Log::out), will_not_call_mercury,"
+:- pragma c_code(math__ln(X::in) = (Log::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X <= 0.0) {
-	    mercury_domain_error(""math__ln"");
+		ML_math_domain_error(""math__ln"");
 	}
+#endif
 	Log = log(X);
 ").
 
@@ -320,9 +364,13 @@
 % Domain restrictions:
 %		X > 0
 %
-:- pragma c_code(math__log10(X::in) = (Log10::out), will_not_call_mercury,"
-	if (X <= 0.0)
-	    mercury_domain_error(""math__log10"");
+:- pragma c_code(math__log10(X::in) = (Log10::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
+	if (X <= 0.0) {
+		ML_math_domain_error(""math__log10"");
+	}
+#endif
 	Log10 = log10(X);
 ").
 
@@ -333,11 +381,14 @@
 % Domain restrictions:
 %		X > 0
 %
-:- pragma c_code(math__log2(X::in) = (Log2::out), will_not_call_mercury,"
+:- pragma c_code(math__log2(X::in) = (Log2::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X <= 0.0) {
-	    mercury_domain_error(""math__log2"");
+		ML_math_domain_error(""math__log2"");
 	}
-	Log2 = log(X) / MERCURY_FLOAT__LN2;
+#endif
+	Log2 = log(X) / ML_FLOAT_LN2;
 ").
 
 %
@@ -349,34 +400,40 @@
 %		B > 0
 %		B \= 1
 %
-:- pragma c_code(math__log(B::in, X::in) = (Log::out), will_not_call_mercury,"
+:- pragma c_code(math__log(B::in, X::in) = (Log::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X <= 0.0 || B <= 0.0) {
-	    mercury_domain_error(""math__log"");
+		ML_math_domain_error(""math__log"");
 	}
 	if (B == 1.0) {
-	    mercury_domain_error(""math__log"");
+		ML_math_domain_error(""math__log"");
 	}
+#endif
 	Log = log(X)/log(B);
 ").
 
 %
 % math__sin(X) = Sin is true if Sin is the sine of X.
 %
-:- pragma c_code(math__sin(X::in) = (Sin::out), will_not_call_mercury,"
+:- pragma c_code(math__sin(X::in) = (Sin::out),
+		[will_not_call_mercury, thread_safe],"
 	Sin = sin(X);
 ").
 
 %
 % math__cos(X) = Sin is true if Cos is the cosine of X.
 %
-:- pragma c_code(math__cos(X::in) = (Cos::out), will_not_call_mercury,"
+:- pragma c_code(math__cos(X::in) = (Cos::out),
+		[will_not_call_mercury, thread_safe],"
 	Cos = cos(X);
 ").
 
 %
 % math__tan(X) = Tan is true if Tan is the tangent of X.
 %
-:- pragma c_code(math__tan(X::in) = (Tan::out), will_not_call_mercury,"
+:- pragma c_code(math__tan(X::in) = (Tan::out),
+		[will_not_call_mercury, thread_safe],"
 	Tan = tan(X);
 ").
 
@@ -387,10 +444,13 @@
 % Domain restrictions:
 %		X must be in the range [-1,1]
 %
-:- pragma c_code(math__asin(X::in) = (ASin::out), will_not_call_mercury,"
+:- pragma c_code(math__asin(X::in) = (ASin::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X < -1.0 || X > 1.0) {
-	    mercury_domain_error(""math__asin"");
+		ML_math_domain_error(""math__asin"");
 	}
+#endif
 	ASin = asin(X);
 ").
 
@@ -401,18 +461,22 @@
 % Domain restrictions:
 %		X must be in the range [-1,1]
 %
-:- pragma c_code(math__acos(X::in) = (ACos::out), will_not_call_mercury,"
+:- pragma c_code(math__acos(X::in) = (ACos::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X < -1.0 || X > 1.0) {
-	    mercury_domain_error(""math__acos"");
+		ML_math_domain_error(""math__acos"");
 	}
-	ACos = asin(X);
+#endif
+	ACos = acos(X);
 ").
 
 %
 % math__atan(X) = ATan is true if ATan is the inverse
 % tangent of X, where ATan is in the range [-pi/2,pi/2].
 %
-:- pragma c_code(math__atan(X::in) = (ATan::out), will_not_call_mercury,"
+:- pragma c_code(math__atan(X::in) = (ATan::out),
+		[will_not_call_mercury, thread_safe],"
 	ATan = atan(X);
 ").
 
@@ -421,7 +485,7 @@
 % tangent of Y/X, where ATan is in the range [-pi,pi].
 %
 :- pragma c_code(math__atan2(Y::in, X::in) = (ATan2::out), 
-	will_not_call_mercury, "
+		[will_not_call_mercury, thread_safe], "
 	ATan2 = atan2(Y, X);
 ").
 
@@ -429,7 +493,8 @@
 % math__sinh(X) = Sinh is true if Sinh is the hyperbolic
 % sine of X.
 %
-:- pragma c_code(math__sinh(X::in) = (Sinh::out), will_not_call_mercury,"
+:- pragma c_code(math__sinh(X::in) = (Sinh::out),
+		[will_not_call_mercury, thread_safe],"
 	Sinh = sinh(X);
 ").
 
@@ -437,7 +502,8 @@
 % math__cosh(X) = Cosh is true if Cosh is the hyperbolic
 % cosine of X.
 %
-:- pragma c_code(math__cosh(X::in) = (Cosh::out), will_not_call_mercury,"
+:- pragma c_code(math__cosh(X::in) = (Cosh::out),
+		[will_not_call_mercury, thread_safe],"
 	Cosh = cosh(X);
 ").
 
@@ -445,7 +511,8 @@
 % math__tanh(X) = Tanh is true if Tanh is the hyperbolic
 % tangent of X.
 %
-:- pragma c_code(math__tanh(X::in) = (Tanh::out), will_not_call_mercury,"
+:- pragma c_code(math__tanh(X::in) = (Tanh::out),
+		[will_not_call_mercury, thread_safe],"
 	Tanh = tanh(X);
 ").
 
@@ -639,29 +706,36 @@
 % Mathematical constants from math.m
 %
 	% Pythagoras' number
-:- pragma(c_code, math__pi(Pi::out), "Pi = MERCURY_FLOAT__PI;").
+:- pragma c_code(math__pi(Pi::out), [will_not_call_mercury, thread_safe],
+		"Pi = ML_FLOAT_PI;").
 
 	% Base of natural logarithms
-:- pragma(c_code, math__e(E::out), "E = MERCURY_FLOAT__E;").
+:- pragma c_code(math__e(E::out), [will_not_call_mercury, thread_safe],
+		"E = ML_FLOAT_E;").
 
 %
 % math__ceiling(X, Ceil) is true if Ceil is the smallest integer
 % not less than X.
 %
-:- pragma(c_code, math__ceiling(Num::in, Ceil::out), "Ceil = ceil(Num);").
+:- pragma c_code(math__ceiling(Num::in, Ceil::out),
+		[will_not_call_mercury, thread_safe],
+	"Ceil = ceil(Num);").
 
 %
 % math__floor(X, Floor) is true if Floor is the largest integer
 % not greater than X.
 %
-:- pragma(c_code, math__floor(Num::in, Floor::out), "Floor = floor(Num);").
+:- pragma c_code(math__floor(Num::in, Floor::out),
+		[will_not_call_mercury, thread_safe],
+	"Floor = floor(Num);").
 
 %
 % math__round(X, Round) is true if Round is the integer
 % closest to X.  If X has a fractional component of 0.5,
 % it is rounded up.
 %
-:- pragma(c_code, math__round(Num::in, Rounded::out), "
+:- pragma c_code(math__round(Num::in, Rounded::out),
+		[will_not_call_mercury, thread_safe], "
 	Rounded = floor(Num+0.5);
 ").
 
@@ -669,11 +743,12 @@
 % math__truncate(X, Trunc) is true if Trunc is the integer
 % closest to X such that |Trunc| =< |X|.
 %
-:- pragma(c_code, math__truncate(X::in, Trunc::out), "
+:- pragma c_code(math__truncate(X::in, Trunc::out),
+		[will_not_call_mercury, thread_safe], "
 	if (X < 0.0) {
-	    Trunc = ceil(X);
+		Trunc = ceil(X);
 	} else {
-	    Trunc = floor(X);
+		Trunc = floor(X);
 	}
 ").
 
@@ -684,10 +759,13 @@
 % Domain restrictions:
 %		X >= 0
 %
-:- pragma(c_code, math__sqrt(X::in, SquareRoot::out), "
+:- pragma c_code(math__sqrt(X::in, SquareRoot::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X < 0.0) {
-	    mercury_domain_error(""math__sqrt"");
+		ML_math_domain_error(""math__sqrt"");
 	}
+#endif
 	SquareRoot = sqrt(X);
 ").
 
@@ -699,25 +777,33 @@
 %		X >= 0
 %		X = 0 implies Y > 0
 %
-:- pragma(c_code, math__pow(X::in, Y::in, Res::out), "
+:- pragma c_code(math__pow(X::in, Y::in, Res::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X < 0.0) {
-	    mercury_domain_error(""math__pow"");
+		ML_math_domain_error(""math__pow"");
 	}
 	if (X == 0.0) {
-	    if (Y <= 0.0) {
-		mercury_domain_error(""math__pow"");
-	    }
-	    Res = 0.0;
+		if (Y <= 0.0) {
+			ML_math_domain_error(""math__pow"");
+		}
+		Res = 0.0;
 	} else {
-	    Res = pow(X, Y);
+		Res = pow(X, Y);
 	}
+#else
+	Res = pow(X, Y);
+#endif
 ").
 
 %
 % math__exp(X, Exp) is true if Exp is X raised to the
 % power of e.
 %
-:- pragma(c_code, math__exp(X::in, Exp::out), "Exp = exp(X);").
+:- pragma c_code(math__exp(X::in, Exp::out),
+		[will_not_call_mercury, thread_safe], "
+	Exp = exp(X);
+").
 
 %
 % math__ln(X, Log) is true if Log is the natural logarithm
@@ -726,10 +812,13 @@
 % Domain restrictions:
 %		X > 0
 %
-:- pragma(c_code, math__ln(X::in, Log::out), "
+:- pragma c_code(math__ln(X::in, Log::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X <= 0.0) {
-	    mercury_domain_error(""math__ln"");
+		ML_math_domain_error(""math__ln"");
 	}
+#endif
 	Log = log(X);
 ").
 
@@ -740,9 +829,13 @@
 % Domain restrictions:
 %		X > 0
 %
-:- pragma(c_code, math__log10(X::in, Log10::out), "
-	if (X <= 0.0)
-	    mercury_domain_error(""math__log10"");
+:- pragma c_code(math__log10(X::in, Log10::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
+	if (X <= 0.0) {
+		ML_math_domain_error(""math__log10"");
+	}
+#endif
 	Log10 = log10(X);
 ").
 
@@ -753,11 +846,14 @@
 % Domain restrictions:
 %		X > 0
 %
-:- pragma(c_code, math__log2(X::in, Log2::out), "
+:- pragma c_code(math__log2(X::in, Log2::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X <= 0.0) {
-	    mercury_domain_error(""math__log2"");
+		ML_math_domain_error(""math__log2"");
 	}
-	Log2 = log(X) / MERCURY_FLOAT__LN2;
+#endif
+	Log2 = log(X) / ML_FLOAT_LN2;
 ").
 
 %
@@ -769,30 +865,42 @@
 %		B > 0
 %		B \= 1
 %
-:- pragma(c_code, math__log(B::in, X::in, Log::out), "
+:- pragma c_code(math__log(B::in, X::in, Log::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X <= 0.0 || B <= 0.0) {
-	    mercury_domain_error(""math__log"");
+		ML_math_domain_error(""math__log"");
 	}
 	if (B == 1.0) {
-	    mercury_domain_error(""math__log"");
+		ML_math_domain_error(""math__log"");
 	}
+#endif
 	Log = log(X)/log(B);
 ").
 
 %
 % math__sin(X, Sin) is true if Sin is the sine of X.
 %
-:- pragma(c_code, math__sin(X::in, Sin::out), "Sin = sin(X);").
+:- pragma c_code(math__sin(X::in, Sin::out),
+		[will_not_call_mercury, thread_safe], "
+	Sin = sin(X);
+").
 
 %
-% math__cos(X, Sin) is true if Cos is the cosine of X.
+% math__cos(X, Cos) is true if Cos is the cosine of X.
 %
-:- pragma(c_code, math__cos(X::in, Cos::out), "Cos = cos(X);").
+:- pragma c_code(math__cos(X::in, Cos::out),
+		[will_not_call_mercury, thread_safe], "
+	Cos = cos(X);
+").
 
 %
 % math__tan(X, Tan) is true if Tan is the tangent of X.
 %
-:- pragma(c_code, math__tan(X::in, Tan::out), "Tan = tan(X);").
+:- pragma c_code(math__tan(X::in, Tan::out),
+		[will_not_call_mercury, thread_safe], "
+	Tan = tan(X);
+").
 
 %
 % math__asin(X, ASin) is true if ASin is the inverse
@@ -801,10 +909,13 @@
 % Domain restrictions:
 %		X must be in the range [-1,1]
 %
-:- pragma(c_code, math__asin(X::in, ASin::out), "
+:- pragma c_code(math__asin(X::in, ASin::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X < -1.0 || X > 1.0) {
-	    mercury_domain_error(""math__asin"");
+		ML_math_domain_error(""math__asin"");
 	}
+#endif
 	ASin = asin(X);
 ").
 
@@ -815,10 +926,13 @@
 % Domain restrictions:
 %		X must be in the range [-1,1]
 %
-:- pragma(c_code, math__acos(X::in, ACos::out), "
+:- pragma c_code(math__acos(X::in, ACos::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X < -1.0 || X > 1.0) {
-	    mercury_domain_error(""math__acos"");
+		ML_math_domain_error(""math__acos"");
 	}
+#endif
 	ACos = asin(X);
 ").
 
@@ -826,13 +940,17 @@
 % math__atan(X, ATan) is true if ATan is the inverse
 % tangent of X, where ATan is in the range [-pi/2,pi/2].
 %
-:- pragma(c_code, math__atan(X::in, ATan::out), "ATan = atan(X);").
+:- pragma c_code(math__atan(X::in, ATan::out),
+		[will_not_call_mercury, thread_safe], "
+	ATan = atan(X);
+").
 
 %
 % math__atan2(Y, X, ATan) is true if ATan is the inverse
 % tangent of Y/X, where ATan is in the range [-pi,pi].
 %
-:- pragma(c_code, math__atan2(Y::in, X::in, ATan2::out), "
+:- pragma c_code(math__atan2(Y::in, X::in, ATan2::out),
+		[will_not_call_mercury, thread_safe], "
 	ATan2 = atan2(Y, X);
 ").
 
@@ -840,19 +958,28 @@
 % math__sinh(X, Sinh) is true if Sinh is the hyperbolic
 % sine of X.
 %
-:- pragma(c_code, math__sinh(X::in, Sinh::out), "Sinh = sinh(X);").
+:- pragma c_code(math__sinh(X::in, Sinh::out),
+		[will_not_call_mercury, thread_safe], "
+	Sinh = sinh(X);
+").
 
 %
 % math__cosh(X, Cosh) is true if Cosh is the hyperbolic
 % cosine of X.
 %
-:- pragma(c_code, math__cosh(X::in, Cosh::out), "Cosh = cosh(X);").
+:- pragma c_code(math__cosh(X::in, Cosh::out),
+		[will_not_call_mercury, thread_safe], "
+	Cosh = cosh(X);
+").
 
 %
 % math__tanh(X, Tanh) is true if Tanh is the hyperbolic
 % tangent of X.
 %
-:- pragma(c_code, math__tanh(X::in, Tanh::out), "Tanh = tanh(X);").
+:- pragma c_code(math__tanh(X::in, Tanh::out),
+		[will_not_call_mercury, thread_safe], "
+	Tanh = tanh(X);
+").
 
 %---------------------------------------------------------------------------%
 %---------------------------------------------------------------------------%
