@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -116,6 +116,8 @@ ops__lookup_postfix_op(_OpTable, Name, Priority, LeftAssoc) :-
 ops__lookup_op(_OpTable, Name) :-
 	ops__op_table(Name, _, _, _).
 
+	% Changes here may require changes to compiler/mercury_to_mercury.m
+	% and doc/transition_guide.texi.
 :- pred ops__op_table(string, ops__category, ops__specifier, ops__priority).
 :- mode ops__op_table(in, in, out, out) is semidet.
 :- mode ops__op_table(in, out, out, out) is nondet.
@@ -135,7 +137,7 @@ ops__op_table(".", after, xfy, 600).		% traditional Prolog (not ISO)
 ops__op_table("/", after, yfx, 400).		% standard ISO Prolog
 ops__op_table("//", after, yfx, 400).		% standard ISO Prolog
 ops__op_table("/\\", after, yfx, 500).		% standard ISO Prolog
-ops__op_table(":", after, xfy, 600).		% different precedence in ISO?
+ops__op_table(":", after, yfx, 600).		% `xfy' in ISO Prolog
 ops__op_table(":-", after, xfx, 1200).		% standard ISO Prolog
 ops__op_table(":-", before, fx, 1200).		% standard ISO Prolog
 ops__op_table("::", after, xfx, 1175).		% Mercury extension
@@ -149,6 +151,7 @@ ops__op_table("=..", after, xfx, 700).		% standard ISO Prolog
 ops__op_table("=:=", after, xfx, 700).		% standard ISO Prolog (*)
 ops__op_table("=<", after, xfx, 700).		% standard ISO Prolog
 ops__op_table("==", after, xfx, 700).		% standard ISO Prolog (*)
+ops__op_table("==>", after, xfx, 1175).		% Mercury extension
 ops__op_table("=>", after, xfy, 920).		% Mercury/NU-Prolog extension
 ops__op_table("=\\=", after, xfx, 700).		% standard ISO Prolog (*)
 ops__op_table(">", after, xfx, 700).		% standard ISO Prolog
@@ -165,6 +168,8 @@ ops__op_table("\\/", after, yfx, 500).		% standard ISO Prolog
 ops__op_table("\\=", after, xfx, 700).		% standard ISO Prolog
 ops__op_table("\\==", after, xfx, 700).		% standard ISO Prolog (*)
 ops__op_table("^", after, xfy, 200).		% standard ISO Prolog
+ops__op_table("aditi_bottom_up", before, fx, 500). % Mercury extension
+ops__op_table("aditi_top_down", before, fx, 500). % Mercury extension
 ops__op_table("all", before, fxy, 950).		% Mercury/NU-Prolog extension
 ops__op_table("and", after, xfy, 720).		% NU-Prolog extension
 ops__op_table("div", after, yfx, 400).		% standard ISO Prolog
@@ -177,16 +182,19 @@ ops__op_table("export_op", before, fx, 1199).	% Mercury extension (NYI)
 ops__op_table("export_pred", before, fx, 1199).	% Mercury extension (NYI)
 ops__op_table("export_sym", before, fx, 1199).	% Mercury extension (NYI)
 ops__op_table("export_type", before, fx, 1199).	% Mercury extension (NYI)
-ops__op_table("func", before, fx, 1180).	% Mercury extension
+ops__op_table("func", before, fx, 800).		% Mercury extension
 ops__op_table("if", before, fx, 1160).		% Mercury/NU-Prolog extension
 ops__op_table("import_adt", before, fx, 1199).	% Mercury extension (NYI)
 ops__op_table("import_cons", before, fx, 1199).	% Mercury extension (NYI)
 ops__op_table("import_module", before, fx, 1199). % Mercury extension
+ops__op_table("include_module", before, fx, 1199). % Mercury extension
 ops__op_table("import_op", before, fx, 1199).	% Mercury extension (NYI)
 ops__op_table("import_pred", before, fx, 1199).	% Mercury extension (NYI)
 ops__op_table("import_sym", before, fx, 1199).	% Mercury extension (NYI)
 ops__op_table("import_type", before, fx, 1199).	% Mercury extension (NYI)
+ops__op_table("impure", before, fy, 800).	% Mercury extension
 ops__op_table("inst", before, fx, 1199).	% Mercury extension
+ops__op_table("instance", before, fx, 1199).	% Mercury extension
 ops__op_table("is", after, xfx, 701).		% ISO Prolog says prec 700
 ops__op_table("lambda", before, fxy, 950).	% Mercury extension
 ops__op_table("mod", after, xfx, 400).		% Standard ISO Prolog
@@ -195,12 +203,15 @@ ops__op_table("module", before, fx, 1199).	% Mercury extension
 ops__op_table("not", before, fy, 900).		% Mercury/NU-Prolog extension
 ops__op_table("or", after, xfy, 740).		% NU-Prolog extension
 ops__op_table("pragma", before, fx, 1199).	% Mercury extension
-ops__op_table("pred", before, fx, 1180).	% Mercury/NU-Prolog extension
+ops__op_table("pred", before, fx, 800).		% Mercury/NU-Prolog extension
+ops__op_table("promise", before, fx, 1199).	% Mercury extension
 ops__op_table("rem", after, xfx, 400).		% Standard ISO Prolog
 ops__op_table("rule", before, fx, 1199).	% NU-Prolog extension
+ops__op_table("semipure", before, fy, 800).	% Mercury extension
 ops__op_table("some", before, fxy, 950).	% Mercury/NU-Prolog extension
 ops__op_table("then", after, xfx, 1150).	% Mercury/NU-Prolog extension
 ops__op_table("type", before, fx, 1180).	% Mercury extension
+ops__op_table("typeclass", before, fx, 1199).	% Mercury extension
 ops__op_table("use_adt", before, fx, 1199).	% Mercury extension (NYI)
 ops__op_table("use_cons", before, fx, 1199).	% Mercury extension (NYI)
 ops__op_table("use_module", before, fx, 1199).	% Mercury extension (NYI)
