@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -9,9 +9,15 @@
 % stability: high.
 %
 % This file exports the predicate parser__read_term, which reads
-% a term from the current input stream.  The parser and lexer are
-% intended to exactly follow ISO Prolog syntax, but there are some
-% departures from that for three reasons:
+% a term from the current input stream.
+% The parser__read_term_from_string predicates are the same as the
+% read_term predicates, except that the term is read from
+% a string rather than from the current input stream.
+% The parser__parse_token_list predicate is similar,
+% but it takes a list of tokens rather than a string.
+%
+% The parser and lexer are intended to exactly follow ISO Prolog
+% syntax, but there are some departures from that for three reasons:
 %
 %	(1) I wrote some of the code at home when the ISO Prolog draft
 %	    was at uni - so in some places I just guessed.
@@ -28,17 +34,59 @@
 
 :- module parser.
 :- interface.
-:- import_module io, term_io.
+:- import_module io, term_io, lexer.
 
-:- pred parser__read_term(read_term, io__state, io__state).
+%-----------------------------------------------------------------------------%
+%
+% parser__read_term/{3,4}:
+%	Reads in (and parses) terms from the current input stream.
+
+:- pred parser__read_term(read_term(T), io__state, io__state).
 :- mode parser__read_term(out, di, uo) is det.
+% parser__read_term(Result):
+%	Reads a term from the current input stream.
 
-	% The string is the filename to use for the current input stream;
-	% this is used in constructing the term__contexts in the read term.
-	% This interface is used to support the `:- pragma source_file'
-	% directive.
-:- pred parser__read_term(string, read_term, io__state, io__state).
+:- pred parser__read_term(string, read_term(T), io__state, io__state).
 :- mode parser__read_term(in, out, di, uo) is det.
+% parser__read_term(FileName, Result):
+%	Reads a term from the current input stream.
+%	The string is the filename to use for the current input stream;
+%	this is used in constructing the term__contexts in the read term.
+%	This interface is used to support the `:- pragma source_file'
+%	directive.
+
+%-----------------------------------------------------------------------------%
+%
+% parser__read_term_from_string/{4,6}:
+%	Parses terms from a string.
+
+	% The read_term_from_string predicates are the same as the
+	% read_term predicates, except that the term is read from
+	% a string rather than from the current input stream.
+	% The returned value `EndPos' is the position one character
+	% past the end of the term read.
+	% The arguments `MaxOffset' and `StartPos' in the six-argument version
+	% specify the length of the string and the position within the
+	% string at which to start parsing.
+
+:- pred parser__read_term_from_string(string, string, posn, read_term(T)).
+:- mode parser__read_term_from_string(in, in, out, out) is det.
+%	parser__read_term_from_string(FileName, String, EndPos, Term).
+
+:- pred parser__read_term_from_string(string, string, int, posn, posn,
+					read_term(T)).
+:- mode parser__read_term_from_string(in, in, in, in, out, out) is det.
+%	parser__read_term_from_string(FileName, String, MaxOffset, StartPos,
+%				    EndPos, Term).
+
+%-----------------------------------------------------------------------------%
+%
+% parser__parse_tokens/3:
+%	Parses a list of tokens.
+
+:- pred parser__parse_tokens(string, token_list, read_term(T)).
+:- mode parser__parse_tokens(in, in, out) is det.
+	% parser__parse_tokens(FileName, TokenList, Result):
 
 %-----------------------------------------------------------------------------%
 
@@ -59,18 +107,34 @@ parser__read_term(Result) -->
 
 parser__read_term(FileName, Result) -->
 	lexer__get_token_list(Tokens),
-	( { Tokens = token_nil } ->
-		{ Result = eof }
+	{ parser__parse_tokens(FileName, Tokens, Result) }.
+
+parser__read_term_from_string(FileName, String, EndPos, Result) :-
+	string__length(String, Len),
+	StartPos = posn(1, 0, 0),
+	parser__read_term_from_string(FileName, String, Len, StartPos, EndPos,
+			Result).
+
+parser__read_term_from_string(FileName, String, Len, StartPos, EndPos,
+		Result) :-
+	lexer__string_get_token_list(String, Len, Tokens, StartPos, EndPos),
+	parser__parse_tokens(FileName, Tokens, Result).
+
+%-----------------------------------------------------------------------------%
+
+parser__parse_tokens(FileName, Tokens, Result) :-
+	( Tokens = token_nil ->
+		Result = eof
 	;
-		{ parser__init_state(FileName, Tokens, ParserState0) },
-		{ parser__parse_whole_term(Term, ParserState0, ParserState) },
-		{ parser__final_state(ParserState, VarSet, LeftOverTokens) },
-		{ parser__check_for_errors(Term, VarSet,
-			Tokens, LeftOverTokens, Result) }
+		parser__init_state(FileName, Tokens, ParserState0),
+		parser__parse_whole_term(Term, ParserState0, ParserState),
+		parser__final_state(ParserState, VarSet, LeftOverTokens),
+		parser__check_for_errors(Term, VarSet,
+			Tokens, LeftOverTokens, Result)
 	).
 
-:- pred parser__check_for_errors(parse(term), varset, token_list, token_list,
-				 read_term).
+:- pred parser__check_for_errors(parse(term(T)), varset(T),
+		token_list, token_list, read_term(T)).
 :- mode parser__check_for_errors(in, in, in, in, out) is det.
 
 parser__check_for_errors(error(ErrorMessage, ErrorTokens), _VarSet, Tokens,
@@ -140,7 +204,8 @@ parser__check_for_bad_token(token_cons(Token, LineNum, Tokens),
 		parser__check_for_bad_token(Tokens, Message, LineNum)
 	).
 
-:- pred parser__parse_whole_term(parse(term), parser__state, parser__state).
+:- pred parser__parse_whole_term(parse(term(T)),
+		parser__state(T), parser__state(T)).
 :- mode parser__parse_whole_term(out, in, out) is det.
 
 parser__parse_whole_term(Term) -->
@@ -158,13 +223,13 @@ parser__parse_whole_term(Term) -->
 
 
 
-:- pred parser__parse_term(parse(term), parser__state, parser__state).
+:- pred parser__parse_term(parse(term(T)), parser__state(T), parser__state(T)).
 :- mode parser__parse_term(out, in, out) is det.
 
 parser__parse_term(Term) -->
 	parser__parse_term_2(1201, no, Term).
 
-:- pred parser__parse_arg(parse(term), parser__state, parser__state).
+:- pred parser__parse_arg(parse(term(T)), parser__state(T), parser__state(T)).
 :- mode parser__parse_arg(out, in, out) is det.
 
 parser__parse_arg(Term) -->
@@ -176,8 +241,8 @@ parser__parse_arg(Term) -->
 	% The above is because we need bug-for-bug compatibility
 	% with the NU-Prolog parser in order to support e.g. `::' in args.
 
-:- pred parser__parse_term_2(int, bool, parse(term),
-				parser__state, parser__state).
+:- pred parser__parse_term_2(int, bool, parse(term(T)),
+				parser__state(T), parser__state(T)).
 :- mode parser__parse_term_2(in, in, out, in, out) is det.
 
 parser__parse_term_2(MaxPriority, IsArg, Term) -->
@@ -190,8 +255,8 @@ parser__parse_term_2(MaxPriority, IsArg, Term) -->
 		{ Term = LeftTerm0 }
 	).
 
-:- pred parser__parse_left_term(int, bool, int, parse(term),
-				parser__state, parser__state).
+:- pred parser__parse_left_term(int, bool, int, parse(term(T)),
+				parser__state(T), parser__state(T)).
 :- mode parser__parse_left_term(in, in, out, out, in, out) is det.
 
 parser__parse_left_term(MaxPriority, IsArg, OpPriority, Term) -->
@@ -285,8 +350,8 @@ parser__parse_left_term(MaxPriority, IsArg, OpPriority, Term) -->
 		{ OpPriority = 0 }
 	).
 
-:- pred parser__parse_rest(int, bool, int, term, parse(term),
-				parser__state, parser__state).
+:- pred parser__parse_rest(int, bool, int, term(T), parse(term(T)),
+				parser__state(T), parser__state(T)).
 :- mode parser__parse_rest(in, in, in, in, out, in, out) is det.
 
 parser__parse_rest(MaxPriority, IsArg, LeftPriority, LeftTerm, Term) -->
@@ -294,13 +359,28 @@ parser__parse_rest(MaxPriority, IsArg, LeftPriority, LeftTerm, Term) -->
 		% infix op
 		parser__get_token(Token, Context),
 		{ Token = comma, IsArg = no ->
-			Op = ","
+			Op0 = ","
 		;
-			Token = name(Op)
+			Token = name(Op0)
 		},
-		parser__get_ops_table(OpTable),
-		{ ops__lookup_infix_op(OpTable, Op,
-				OpPriority, LeftAssoc, RightAssoc) },
+		(
+				% A token surrounded by backquotes is a
+				% prefix token being using in an
+				% infix manner.
+			{ Op0 = "`" }
+		->
+			parser__get_token(name(Op), _),
+			parser__get_token(name("`"), _),
+
+			{ OpPriority = 100 },
+			{ LeftAssoc = y },
+			{ RightAssoc = x }
+		;
+			{ Op = Op0 },
+			parser__get_ops_table(OpTable),
+			{ ops__lookup_infix_op(OpTable, Op,
+					OpPriority, LeftAssoc, RightAssoc) }
+		),
 		{ OpPriority =< MaxPriority },
 		{ parser__check_priority(LeftAssoc, OpPriority, LeftPriority) }
 	->
@@ -335,8 +415,8 @@ parser__parse_rest(MaxPriority, IsArg, LeftPriority, LeftTerm, Term) -->
 
 %-----------------------------------------------------------------------------%
 
-:- pred parser__parse_simple_term(token, token_context, int, parse(term),
-				parser__state, parser__state).
+:- pred parser__parse_simple_term(token, token_context, int, parse(term(T)),
+				parser__state(T), parser__state(T)).
 :- mode parser__parse_simple_term(in, in, in, out, in, out) is det.
 
 parser__parse_simple_term(Token, Context, Priority, Term) -->
@@ -357,10 +437,17 @@ parser__parse_simple_term(Token, Context, Priority, Term) -->
 	%	atom --> open_list, close_list
 	%	atom --> open_curly, close_curly
 	% term --> variable		% priority 0
-	% term -->
+	% term --> atom, open_ct, arg_list, close
+	%	arg_list --> arg
+	%	arg_list --> arg, comma, arg_list
+	% term --> open, term, close
+	% term --> open_ct, term, close
+	% term --> term, op, term	% with various conditions
+	% term --> op, term		% with various conditions
+	% term --> term, op		% with various conditions
 
-:- pred parser__parse_simple_term_2(token, token_context, int, parse(term),
-				parser__state, parser__state).
+:- pred parser__parse_simple_term_2(token, token_context, int, parse(term(T)),
+				parser__state(T), parser__state(T)).
 :- mode parser__parse_simple_term_2(in, in, in, out, in, out) is semidet.
 
 parser__parse_simple_term_2(name(Atom), Context, Prec, Term) -->
@@ -438,9 +525,9 @@ parser__parse_simple_term_2(open_ct, Context, Prec, Term) -->
 	parser__parse_simple_term_2(open, Context, Prec, Term).
 
 parser__parse_simple_term_2(open_list, Context, _, Term) -->
+	parser__get_term_context(Context, TermContext),
 	( parser__get_token(close_list) ->
-		parser__get_term_context(Context, TermContext),
-		{ Term = ok(term__functor(term__atom("[]"), [], TermContext)) }
+		parser__parse_special_atom("[]", TermContext, Term)
 	;
 		parser__parse_list(Term)
 	).
@@ -448,7 +535,7 @@ parser__parse_simple_term_2(open_list, Context, _, Term) -->
 parser__parse_simple_term_2(open_curly, Context, _, Term) -->
 	parser__get_term_context(Context, TermContext),
 	( parser__get_token(close_curly) ->
-		{ Term = ok(term__functor(term__atom("{}"), [], TermContext)) }
+		parser__parse_special_atom("{}", TermContext, Term)
 	;
 		parser__parse_term(SubTerm0),
 		( { SubTerm0 = ok(SubTerm) } ->
@@ -465,7 +552,26 @@ parser__parse_simple_term_2(open_curly, Context, _, Term) -->
 		)
 	).
 
-:- pred parser__parse_list(parse(term), parser__state, parser__state).
+:- pred parser__parse_special_atom(string, term__context, parse(term(T)),
+		parser__state(T), parser__state(T)).
+:- mode parser__parse_special_atom(in, in, out, in, out) is det.
+
+parser__parse_special_atom(Atom, TermContext, Term) -->
+	( parser__get_token(open_ct) ->
+		parser__parse_args(Args0),
+		(	{ Args0 = ok(Args) },
+			{ Term = ok(term__functor(term__atom(Atom),
+				Args, TermContext)) }
+		;
+			% propagate error upwards
+			{ Args0 = error(Message, Tokens) },
+			{ Term = error(Message, Tokens) }
+		)
+	;
+		{ Term = ok(term__functor(term__atom(Atom), [], TermContext)) }
+	).
+
+:- pred parser__parse_list(parse(term(T)), parser__state(T), parser__state(T)).
 :- mode parser__parse_list(out, in, out) is det.
 
 parser__parse_list(List) -->
@@ -514,7 +620,8 @@ parser__parse_list(List) -->
 	    { List = Arg0 }
 	).
 
-:- pred parser__parse_args(parse(list(term)), parser__state, parser__state).
+:- pred parser__parse_args(parse(list(term(T))),
+		parser__state(T), parser__state(T)).
 :- mode parser__parse_args(out, in, out) is det.
 
 parser__parse_args(List) -->
@@ -549,14 +656,15 @@ parser__parse_args(List) -->
 
 	% Routines that manipulate the parser state.
 
-:- type parser__state
+:- type parser__state(T)
 	--->	parser__state(
 			string,		% the name of the stream being parsed
 			ops__table,	% the current set of operators
-			varset,		% the names of the variables in the
+			varset(T),	% the names of the variables in the
 					% term being parsed
 			token_list,	% the remaining tokens
-			map(string, var)% a map from variable name to variable
+			map(string, var(T))
+					% a map from variable name to variable
 					% so we know when to make a fresh var
 		).
 
@@ -569,7 +677,7 @@ parser__parse_args(List) -->
 	% it was some other sort of error, so issue the usual
 	% error message.
 
-:- pred parser__unexpected(string, parse(T), parser__state, parser__state).
+:- pred parser__unexpected(string, parse(U), parser__state(T), parser__state(T)).
 :- mode parser__unexpected(in, out, in, out) is det.
 
 parser__unexpected(UsualMessage, Error) -->
@@ -579,8 +687,8 @@ parser__unexpected(UsualMessage, Error) -->
 		parser__error(UsualMessage, Error)
 	).
 
-:- pred parser__unexpected_tok(token, token_context, string, parse(T),
-				parser__state, parser__state).
+:- pred parser__unexpected_tok(token, token_context, string, parse(U),
+				parser__state(T), parser__state(T)).
 :- mode parser__unexpected_tok(in, in, in, out, in, out) is det.
 
 parser__unexpected_tok(Token, Context, UsualMessage, Error) -->
@@ -603,7 +711,7 @@ parser__unexpected_tok(Token, Context, UsualMessage, Error) -->
 
 %-----------------------------------------------------------------------------%
 
-:- pred parser__error(string, parse(T), parser__state, parser__state).
+:- pred parser__error(string, parse(U), parser__state(T), parser__state(T)).
 :- mode parser__error(in, out, in, out) is det.
 
 parser__error(Message, error(Message, Tokens), ParserState, ParserState) :-
@@ -636,7 +744,7 @@ parser__could_start_term(eof, no).
 
 %-----------------------------------------------------------------------------%
 
-:- pred parser__init_state(string, token_list, parser__state).
+:- pred parser__init_state(string, token_list, parser__state(T)).
 :- mode parser__init_state(in, in, out) is det.
 
 parser__init_state(FileName, Tokens, ParserState) :-
@@ -645,7 +753,7 @@ parser__init_state(FileName, Tokens, ParserState) :-
 	map__init(Names),
 	ParserState = parser__state(FileName, OpTable, VarSet, Tokens, Names).
 
-:- pred parser__final_state(parser__state, varset, token_list).
+:- pred parser__final_state(parser__state(T), varset(T), token_list).
 :- mode parser__final_state(in, out, out) is det.
 
 parser__final_state(parser__state(_FileName, _OpTable, VarSet, TokenList,
@@ -653,13 +761,14 @@ parser__final_state(parser__state(_FileName, _OpTable, VarSet, TokenList,
 
 %-----------------------------------------------------------------------------%
 
-:- pred parser__get_token(token, parser__state, parser__state).
+:- pred parser__get_token(token, parser__state(T), parser__state(T)).
 :- mode parser__get_token(out, in, out) is semidet.
 
 parser__get_token(Token) -->
 	parser__get_token(Token, _Context).
 
-:- pred parser__get_token(token, token_context, parser__state, parser__state).
+:- pred parser__get_token(token, token_context,
+		parser__state(T), parser__state(T)).
 :- mode parser__get_token(out, out, in, out) is semidet.
 :- mode parser__get_token(in, in, out, in) is det.
 
@@ -668,20 +777,22 @@ parser__get_token(Token, Context,
 		parser__state(FileName, OpTable, VarSet, Tokens, Names)) :-
 	Tokens0 = token_cons(Token, Context, Tokens).
 
-:- pred parser__unget_token(token, token_context, parser__state, parser__state).
+:- pred parser__unget_token(token, token_context,
+		parser__state(T), parser__state(T)).
 :- mode parser__unget_token(in, in, in, out) is det.
 :- mode parser__unget_token(out, out, out, in) is semidet.
 
 parser__unget_token(Token, Context, ParseState0, ParseState) :-
 	parser__get_token(Token, Context, ParseState, ParseState0).
 
-:- pred parser__peek_token(token, parser__state, parser__state).
+:- pred parser__peek_token(token, parser__state(T), parser__state(T)).
 :- mode parser__peek_token(out, in, out) is semidet.
 
 parser__peek_token(Token) -->
 	parser__peek_token(Token, _Context).
 
-:- pred parser__peek_token(token, token_context, parser__state, parser__state).
+:- pred parser__peek_token(token, token_context,
+		parser__state(T), parser__state(T)).
 :- mode parser__peek_token(out, out, in, out) is semidet.
 
 parser__peek_token(Token, Context) -->
@@ -690,7 +801,7 @@ parser__peek_token(Token, Context) -->
 
 %-----------------------------------------------------------------------------%
 
-:- pred parser__add_var(string, var, parser__state, parser__state).
+:- pred parser__add_var(string, var(T), parser__state(T), parser__state(T)).
 :- mode parser__add_var(in, out, in, out) is det.
 
 parser__add_var(VarName, Var,
@@ -708,7 +819,7 @@ parser__add_var(VarName, Var,
 		map__det_insert(Names0, VarName, Var, Names)
 	).
 
-:- pred parser__get_ops_table(ops__table, parser__state, parser__state).
+:- pred parser__get_ops_table(ops__table, parser__state(T), parser__state(T)).
 :- mode parser__get_ops_table(out, in, out) is det.
 
 parser__get_ops_table(OpTable) -->
@@ -733,7 +844,7 @@ parser__check_priority(x, MaxPriority, Priority) :-
 	Priority < MaxPriority.
 
 :- pred parser__get_term_context(token_context, term__context,
-				parser__state, parser__state).
+				parser__state(T), parser__state(T)).
 :- mode parser__get_term_context(in, out, in, out) is det.
 
 parser__get_term_context(TokenContext, TermContext) -->
