@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1994-1995, 1997 The University of Melbourne.
+% Copyright (C) 1994-1995, 1997-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -89,6 +89,12 @@
 :- pred queue__list_to_queue(list(T), queue(T)).
 :- mode queue__list_to_queue(in, out) is det.
 
+	% `queue__delete_all(Queue0, Elem, Queue)' is true iff `Queue' is
+	% the same queue as `Queue0' with all occurences of `Elem' removed
+	% from it.
+:- pred queue__delete_all(queue(T), T, queue(T)).
+:- mode queue__delete_all(in, in, out) is det.
+
 %--------------------------------------------------------------------------%
 
 :- implementation.
@@ -158,4 +164,56 @@ queue__length(On - Off, Length) :-
 
 queue__list_to_queue(List, [] - List).
 
+queue__delete_all(On0 - Off0, Elem, On - Off) :-
+	list__delete_all(On0, Elem, On1),
+	list__delete_all(Off0, Elem, Off1),
+	( Off1 = [] ->
+		list__reverse(On1, Off),
+		On = []
+	;
+		On = On1,
+		Off = Off1
+	).
+
 %--------------------------------------------------------------------------%
+%--------------------------------------------------------------------------%
+% Ralph Becket <rwab1@cl.cam.ac.uk> 29/04/99
+% 	Function forms added.
+
+:- interface.
+
+:- func queue__init = queue(T).
+
+:- func queue__put(queue(T), T) = queue(T).
+
+:- func queue__put_list(queue(T), list(T)) = queue(T).
+
+:- func queue__length(queue(T)) = int.
+
+:- func queue__list_to_queue(list(T)) = queue(T).
+
+:- func queue__delete_all(queue(T), T) = queue(T).
+
+% ---------------------------------------------------------------------------- %
+% ---------------------------------------------------------------------------- %
+
+:- implementation.
+
+queue__init = Q :-
+	queue__init(Q).
+
+queue__put(Q1, T) = Q2 :-
+	queue__put(Q1, T, Q2).
+
+queue__put_list(Q1, Xs) = Q2 :-
+	queue__put_list(Q1, Xs, Q2).
+
+queue__length(Q) = N :-
+	queue__length(Q, N).
+
+queue__list_to_queue(Xs) = Q :-
+	queue__list_to_queue(Xs, Q).
+
+queue__delete_all(Q1, T) = Q2 :-
+	queue__delete_all(Q1, T, Q2).
+
