@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %------------------------------------------------------------------------------%
@@ -58,6 +58,17 @@
 :- pred relation__add(relation(T), relation_key, relation_key, relation(T)).
 :- mode relation__add(in, in, in, out) is det.
 
+	% relation__add_values adds an pair of values to the relation's
+	% domain and adds an element to the relation.
+	%
+	% relation__add_values(R0, X, Y, R) :-
+	%	 relation__add_element(R0, X, XKey, R1),
+	%	 relation__add_element(R1, Y, YKey, R2),
+	%	 relation__add(R1, XKey, YKey, R).
+	%
+:- pred relation__add_values(relation(T), T, T, relation(T)).
+:- mode relation__add_values(in, in, in, out) is det.
+
 	% relation__add_assoc_list adds a list of elements to a
 	% relation.
 :- pred relation__add_assoc_list(relation(T),
@@ -99,14 +110,19 @@
 
 	% relation__to_assoc_list turns a relation into a list of
 	% pairs of elements.
-:- pred relation__to_assoc_list(relation(T),
-	assoc_list(relation_key, relation_key)).
+:- pred relation__to_assoc_list(relation(T), assoc_list(T, T)).
 :- mode relation__to_assoc_list(in, out) is det.
 
+	% relation__to_key_assoc_list turns a relation into a list of
+	% pairs of relation keys.
+:- pred relation__to_key_assoc_list(relation(T),
+	assoc_list(relation_key, relation_key)).
+:- mode relation__to_key_assoc_list(in, out) is det.
+
 	% relation__from_assoc_list turns a list of pairs of
 	% elements into a relation.
-% :- pred relation__from_assoc_list(assoc_list(T, T), relation(T)).
-% :- mode relation__from_assoc_list(in, out) is det.
+:- pred relation__from_assoc_list(assoc_list(T, T), relation(T)).
+:- mode relation__from_assoc_list(in, out) is det.
 
 	% relation__domain finds the set of all elements in the
 	% domain of a relation.
@@ -120,8 +136,8 @@
 
 	% relation__compose(R1, R2, R) is true if R is the
 	% composition of the relations R1 and R2.
-% :- pred relation__compose(relation(T), relation(T), relation(T)).
-% :- mode relation__compose(in, in, out) is det.
+:- pred relation__compose(relation(T), relation(T), relation(T)).
+:- mode relation__compose(in, in, out) is det.
 
 	% relation__dfs(Rel, X, Dfs) is true if Dfs is a
 	% depth-first sorting of Rel starting at X.  The
@@ -208,7 +224,7 @@
 
 	% relation__sc(R, SC) is true if SC is the
 	% symmetric closure of R.  In graph terms,
-	% symmetric closure % is the same as turning
+	% symmetric closure is the same as turning
 	% a directed graph into an undirected graph.
 :- pred relation__sc(relation(T), relation(T)).
 :- mode relation__sc(in, out) is det.
@@ -298,6 +314,11 @@ relation__lookup_key(Rel, Key, Elem) :-
 
 %------------------------------------------------------------------------------%
 
+relation__add_values(R0, X, Y, R) :-
+	relation__add_element(R0, X, XKey, R1),
+	relation__add_element(R1, Y, YKey, R2),
+	relation__add(R2, XKey, YKey, R).
+
 	% relation__add adds an element to the relation.
 relation__add(relation(Key, ElMap, FwdIn, BwdIn), U, V,
 		relation(Key, ElMap, FwdOut, BwdOut)) :-
@@ -396,23 +417,43 @@ relation__lookup_to(relation(_Key, _ElMap, _Fwd, Bwd), V, Us) :-
 
 	% relation__to_assoc_list turns a relation into a list of
 	% pairs of elements.
-relation__to_assoc_list(relation(_Key, _ElMap, Fwd, _Bwd), List) :-
+relation__to_assoc_list(relation(_Key, ElMap, Fwd, _Bwd), List) :-
 	map__keys(Fwd, FwdKeys),
-	relation__to_assoc_list_2(Fwd, FwdKeys, List).
+	relation__to_assoc_list_2(Fwd, FwdKeys, ElMap, List).
 
 :- pred relation__to_assoc_list_2(map(relation_key, set(relation_key)),
-	list(relation_key), assoc_list(relation_key, relation_key)).
-:- mode relation__to_assoc_list_2(in, in, out) is det.
-relation__to_assoc_list_2(_Fwd, [], []).
-relation__to_assoc_list_2(Fwd, [Key | Keys], List) :-
-	relation__to_assoc_list_2(Fwd, Keys, List1),
+	list(relation_key), bimap(T, relation_key), assoc_list(T, T)).
+:- mode relation__to_assoc_list_2(in, in, in, out) is det.
+relation__to_assoc_list_2(_Fwd, [], _, []).
+relation__to_assoc_list_2(Fwd, [Key | Keys], ElementMap, AssocList) :-
+	relation__to_assoc_list_2(Fwd, Keys, ElementMap, AssocList1),
 	map__lookup(Fwd, Key, Set),
-	set__to_sorted_list(Set, List2),
-	relation__append_to(Key, List2, List3),
-	list__append(List1, List3, List).
+	set__to_sorted_list(Set, List),
+	bimap__reverse_lookup(ElementMap, KeyEl, Key),
+	Lookup = lambda([U::in, V::out] is det,
+			bimap__reverse_lookup(ElementMap, V, U)),
+	list__map(Lookup, List, ListEls),
+	relation__append_to(KeyEl, ListEls, AssocList2),
+	list__append(AssocList1, AssocList2, AssocList).
 
-:- pred relation__append_to(relation_key, list(relation_key),
-	assoc_list(relation_key, relation_key)).
+	% relation__to_key_assoc_list turns a relation into a list of
+	% pairs of elements.
+relation__to_key_assoc_list(relation(_Key, _ElMap, Fwd, _Bwd), List) :-
+	map__keys(Fwd, FwdKeys),
+	relation__to_key_assoc_list_2(Fwd, FwdKeys, List).
+
+:- pred relation__to_key_assoc_list_2(map(relation_key, set(relation_key)),
+	list(relation_key), assoc_list(relation_key, relation_key)).
+:- mode relation__to_key_assoc_list_2(in, in, out) is det.
+relation__to_key_assoc_list_2(_Fwd, [], []).
+relation__to_key_assoc_list_2(Fwd, [Key | Keys], AssocList) :-
+	relation__to_key_assoc_list_2(Fwd, Keys, AssocList1),
+	map__lookup(Fwd, Key, Set),
+	set__to_sorted_list(Set, List),
+	relation__append_to(Key, List, AssocList2),
+	list__append(AssocList1, AssocList2, AssocList).
+
+:- pred relation__append_to(T1, list(T2), assoc_list(T1, T2)).
 :- mode relation__append_to(in, in, out) is det.
 relation__append_to(_U, [], []).
 relation__append_to(U, [V | Vs], [U - V | UVs]) :-
@@ -420,13 +461,13 @@ relation__append_to(U, [V | Vs], [U - V | UVs]) :-
 
 %------------------------------------------------------------------------------%
 
-% 	% relation__from_assoc_list turns a list of pairs of
-% 	% elements into a relation.
-% relation__from_assoc_list([], Rel) :-
-% 	relation__init(Rel).
-% relation__from_assoc_list([U - V | List], Rel) :-
-% 	relation__from_assoc_list(List, Rel1),
-% 	relation__add(Rel1, U, V, Rel).
+	% relation__from_assoc_list turns a list of pairs of
+	% elements into a relation.
+relation__from_assoc_list([], Rel) :-
+	relation__init(Rel).
+relation__from_assoc_list([U - V | List], Rel) :-
+	relation__from_assoc_list(List, Rel1),
+	relation__add_values(Rel1, U, V, Rel).
 
 %------------------------------------------------------------------------------%
 
@@ -452,35 +493,44 @@ relation__inverse(relation(Key, ElMap, Fwd, Bwd),
 
 %------------------------------------------------------------------------------%
 
-% 	% relation__compose(R1, R2, R) is true iff R is the
-% 	% composition of the relations R1 and R2.
-% relation__compose(R1, R2, Compose) :-
-% 	relation__domain(R1, Dom),
-% 	set__to_sorted_list(Dom, Dom1),
-% 	relation__init(Comp0),
-% 	relation__compose_2(Dom1, R1, R2, Comp0, Compose).
-% 
-% :- pred relation__compose_2(list(T), relation(T), 
-% 		relation(T), relation(T), relation(T)).
-% :- mode relation__compose_2(in, in, in, in, out) is det.
-% relation__compose_2([], _R1, _R2, Comp, Comp).
-% relation__compose_2([X | Xs], R1, R2, Comp0, Comp) :-
-% 	relation__lookup_from(R1, X, R1XSet),
-% 	set__to_sorted_list(R1XSet, R1XList),
-% 	set__init(CXSet0),
-% 	relation__compose_3(R1XList, R2, CXSet0, CXSet),
-% 	set__to_sorted_list(CXSet, CXList),
-% 	relation__append_to(X, CXList, XCX),
-% 	relation__add_assoc_list(Comp0, XCX, Comp1),
-% 	relation__compose_2(Xs, R1, R2, Comp1, Comp).
-% 
-% :- pred relation__compose_3(list(T), relation(T), set(T), set(T)).
-% :- mode relation__compose_3(in, in, in, out) is det.
-% relation__compose_3([], _R2, CX, CX).
-% relation__compose_3([Y | Ys], R2, CX0, CX) :-
-% 	relation__lookup_from(R2, Y, ZsSet),
-% 	set__union(ZsSet, CX0, CX1),
-% 	relation__compose_3(Ys, R2, CX1, CX).
+	% relation__compose(R1, R2, R) is true iff R is the
+	% composition of the relations R1 and R2.
+relation__compose(R1, R2, Compose) :-
+	relation__domain(R1, DomainSet),
+	set__to_sorted_list(DomainSet, DomainList),
+	relation__init(Compose0),
+	list__foldl(relation__compose_2(R1, R2), DomainList,
+		Compose0, Compose).
+
+	% relation__compose_2(R1, R2, X, Comp0, Comp):
+	%	Comp is the union of Comp0 and CompX,
+	%	where CompX = { (X, Z) : some [Y] (X R1 Y, Y R2 Z) }.
+:- pred relation__compose_2(relation(T), relation(T), T,
+			relation(T), relation(T)).
+:- mode relation__compose_2(in, in, in, in, out) is det.
+relation__compose_2(R1, R2, X, Comp0, Comp) :-
+	relation__lookup_element(R1, X, X_R1Key),
+	relation__lookup_from(R1, X_R1Key, Ys_R1KeysSet),
+	set__to_sorted_list(Ys_R1KeysSet, Ys_R1Keys),
+	list__map(relation__lookup_key(R1), Ys_R1Keys, Ys),
+	list__foldl(relation__compose_3(R2, X), Ys, Comp0, Comp).
+
+	% relation__compose_3(R2, X, Y, Comp0, Comp):
+	%	Comp is the union of Comp0 and CompXY,
+	%	where CompXY = { (X, Z) : Y R2 Z }.
+:- pred relation__compose_3(relation(T), T, T, relation(T), relation(T)).
+:- mode relation__compose_3(in, in, in, in, out) is det.
+relation__compose_3(R2, X, Y, Comp0, Comp) :-
+	( relation__search_element(R2, Y, Y_R2Key) ->
+		relation__lookup_from(R2, Y_R2Key, Zs_R2Keys_Set),
+		set__to_sorted_list(Zs_R2Keys_Set, Zs_R2Keys),
+		list__map(relation__lookup_key(R2), Zs_R2Keys, Zs),
+		AddValue = lambda([Z::in, Rel0::in, Rel::out] is det,
+				relation__add_values(Rel0, X, Z, Rel)),
+		list__foldl(AddValue, Zs, Comp0, Comp)
+	;
+		Comp = Comp0
+	).
 
 %------------------------------------------------------------------------------%
 
@@ -726,7 +776,7 @@ relation__reduced(Rel, Red) :-
 	relation__init(Red0),
 	map__init(CliqMap0),
 	relation__make_clique_map(Rel, CliqList, CliqMap0, CliqMap, Red0, Red1),
-	relation__to_assoc_list(Rel, RelAL),
+	relation__to_key_assoc_list(Rel, RelAL),
 	relation__make_reduced_graph(CliqMap, RelAL, Red1, Red).
 
 :- pred relation__make_clique_map(relation(T), list(set(relation_key)),
@@ -887,7 +937,7 @@ relation__atsort_2([H | T], RelInv, Visit0, ATsort0, ATsort) :-
 	% a relation.
 relation__sc(Rel, Sc) :-
 	relation__inverse(Rel, Inv),
-	relation__to_assoc_list(Inv, InvList),
+	relation__to_key_assoc_list(Inv, InvList),
 	relation__add_assoc_list(Rel, InvList, Sc).
 
 %------------------------------------------------------------------------------%
@@ -932,125 +982,222 @@ relation__detect_fake_reflexives(Rel, Rtc, [X | Xs], FakeRefl) :-
 
 %------------------------------------------------------------------------------%
 
-	% relation__rtc returns the reflexive transitive
-	% closure of a relation.  This uses the algorithm
-	% of Eve and Kurki-Suonio.  See:
+	% relation__rtc returns the reflexive transitive closure
+	% of a relation.
 	%
-	%	Eve and Kurki-Suonio, "On computing the
-	%	transitive closure of a relation", Acta
-	%	Informatica, 8, pp 303--314, 1977
-relation__rtc(Rel, Rtc) :-
-	relation__domain_sorted_list(Rel, DomList),
-	list__length(DomList, DomLen),
-	map__init(Map0),
-	relation__rtc_init_map(Map0, DomList, Map),
+	% Note: This is not the most efficient algorithm (in the sense
+	% of minimal number of arc insertions) possible.  However it
+	% "reasonably" efficient and, more importantly, is much easier
+	% to debug than some others.
+	%
+	% The algorithm is very simple, and is based on the
+	% observation that the RTC of any element in a clique is the
+	% same as the RTC of any other element in that clique.  So
+	% we visit each clique in reverse topological sorted order,
+	% compute the RTC for each element in the clique and then
+	% add the appropriate arcs.
+	%
+relation__rtc(Rel, RTC) :-
+	relation__dfs(Rel, Dfs),
+	set_bbbtree__init(Visit),
 
-	% Make a new relation using the same domain as the old one.
 	Rel   = relation(NextElement, ElMap, _, _),
 	map__init(FwdMap),
 	map__init(BwdMap),
-	RtcIn = relation(NextElement, ElMap, FwdMap, BwdMap),
+	RTC0 = relation(NextElement, ElMap, FwdMap, BwdMap),
 
-	Inf is (DomLen + 2) * 2,	% This is close enough to infinity.
-	relation__rtc_2(Inf, Rel, DomList, Map, RtcIn, Rtc).
+	relation__rtc_2(Dfs, Rel, Visit, RTC0, RTC).
 
-	% relation__rtc_init_map takes a domain Xs and creates a map
-	% Map such that:
-	%
-	%	all [X]
-	%	    (
-	%		list__member(X, Xs) =>
-	%		map__lookup(Map, X, 0)
-	%	    ).
-	%
-	% Question: Would an "array" type which allows an
-	% arbitrary domain (not just integers) be a better
-	% solution?  About the only differences between this
-	% and a "map" would be the ease of initialisation,
-	% and the semi-static nature of the domain.  (Arrays
-	% can be resized.)
-:- pred relation__rtc_init_map(map(relation_key, int), list(relation_key),
-		map(relation_key, int)).
-:- mode relation__rtc_init_map(in, in, out) is det.
-relation__rtc_init_map(Map, [], Map).
-relation__rtc_init_map(MapIn, [ X | Xs ], MapOut) :-
-	map__det_insert(MapIn, X, 0, Map1),
-	relation__rtc_init_map(Map1, Xs, MapOut).
+:- pred relation__rtc_2(list(relation_key), relation(T),
+	set_bbbtree(relation_key), relation(T), relation(T)).
+:- mode relation__rtc_2(in, in, in, in, out) is det.
 
-:- pred relation__rtc_2(int, relation(T), list(relation_key),
-		map(relation_key, int), relation(T), relation(T)).
-:- mode relation__rtc_2(in, in, in, in, in, out) is det.
-relation__rtc_2(_, _, [], _, Rtc, Rtc).
-relation__rtc_2(Inf, Rel, [ X | Xs ], Map, RtcIn, RtcOut) :-
-	( map__lookup(Map, X, 0) ->
-		stack__init(S0),
-		rtc(Inf, Rel, X, 1, S0, _S1, Map, Map1, RtcIn, Rtc1)
+relation__rtc_2([], _, _, RTC, RTC).
+relation__rtc_2([H | T], Rel, Visit0, RTC0, RTC) :-
+	( set_bbbtree__member(H, Visit0) ->
+		relation__rtc_2(T, Rel, Visit0, RTC0, RTC)
 	;
-		Map = Map1, RtcIn = Rtc1
-	),
-	relation__rtc_2(Inf, Rel, Xs, Map1, Rtc1, RtcOut).
-
-	% When the arity gets this big, it's probably worth
-	% considering making a state type...
-:- pred rtc(int, relation(T), relation_key, int, stack(relation_key),
-	stack(relation_key), map(relation_key, int), map(relation_key, int),
-	relation(T), relation(T)).
-:- mode rtc(in, in, in, in, in, out, in, out, in, out) is det.
-rtc(Inf, Rel, A, K, S0, S1, MapIn, MapOut, RtcIn, RtcOut) :-
-	K1 is K + 1,
-	stack__push(S0, A, S2),
-	map__set(MapIn, A, K, Map1),
-	relation__add(RtcIn, A, A, Rtc1),
-	relation__lookup_from(Rel, A, BsSet),
-	set__to_sorted_list(BsSet, BsList),
-	rtc_2(Inf, Rel, A, K1, BsList, S2, S3, Map1, Map2, Rtc1, Rtc2),
-	( map__lookup(Map2, A, K) ->
-	    rtc_3(Inf, A, Map2, Map3, S3, S4, Rtc2, Rtc3)
-	;
-	    Map2=Map3, S3=S4, Rtc2=Rtc3
-	),
-	Map3=MapOut, Rtc3=RtcOut, S1=S4.
-
-:- pred rtc_2(int, relation(T), relation_key, int, list(relation_key),
-	stack(relation_key), stack(relation_key), map(relation_key, int),
-	map(relation_key, int), relation(T), relation(T)).
-:- mode rtc_2(in, in, in, in, in, in, out, in, out, in, out) is det.
-rtc_2(_Inf, _Rel, _A, _K1, [], Stack, Stack, Map, Map, Rtc, Rtc).
-rtc_2(Inf, Rel, A, K1, [B | Bs], S0, S1, Map0, Map1, Rtc0, Rtc1) :-
-	( map__lookup(Map0, B, 0) ->
-	    rtc(Inf, Rel, B, K1, S0, S2, Map0, Map2, Rtc0, Rtc2)
-	;
-	    S0=S2, Map0=Map2, Rtc0=Rtc2
-	),
-	map__lookup(Map2, A, Na),
-	map__lookup(Map2, B, Nb),
-	( Na =< Nb ->
-	    Map2 = Map3
-	;
-	    map__set(Map2, A, Nb, Map3)
-	),
-	relation__lookup_from(Rtc2, B, RtcBSet),
-	set__to_sorted_list(RtcBSet, RtcB),
-	relation__append_to(A, RtcB, RtcA),
-	relation__add_assoc_list(Rtc2, RtcA, Rtc3),
-	rtc_2(Inf, Rel, A, K1, Bs, S2, S1, Map3, Map1, Rtc3, Rtc1).
-
-:- pred rtc_3(int, relation_key, map(relation_key, int), map(relation_key, int),
-		stack(relation_key), stack(relation_key), relation(T),
-		relation(T)).
-:- mode rtc_3(in, in, in, out, in, out, in, out) is det.
-rtc_3(Inf, A, MapIn, MapOut, S0, S1, RtcIn, RtcOut) :-
-	stack__pop_det(S0, B, S2),
-	( A = B ->
-	    MapIn=MapOut, S1=S2, RtcIn=RtcOut
-	;
-	    relation__lookup_from(RtcIn, A, RtcASet),
-	    set__to_sorted_list(RtcASet, RtcA),
-	    relation__append_to(B, RtcA, RtcB),
-	    relation__add_assoc_list(RtcIn, RtcB, Rtc1),
-	    map__set(MapIn, B, Inf, Map1),
-	    rtc_3(Inf, A, Map1, MapOut, S2, S1, Rtc1, RtcOut)
+		relation__dfs_3([H], Rel, Visit0, [], Visit, CliqueL0),
+		list__sort_and_remove_dups(CliqueL0, CliqueL),
+		list__foldl(lambda([K :: in, L0 :: in, L :: out] is det,
+			( relation__lookup_from(Rel, K, Followers0),
+			  set__to_sorted_list(Followers0, Followers),
+			  list__merge_and_remove_dups(Followers, L0, L)
+			)),
+			CliqueL, CliqueL, CliqueFollowers),
+		list__foldl(lambda([K :: in, L0 :: in, L :: out] is det,
+			( relation__lookup_from(RTC0, K, Followers0),
+			  set__to_sorted_list(Followers0, Followers),
+			  list__merge_and_remove_dups(Followers, L0, L)
+			)),
+			CliqueFollowers, CliqueL, NewFollowers),
+		relation__add_cartesian_product(CliqueL, NewFollowers,
+			RTC0, RTC1),
+		relation__rtc_2(T, Rel, Visit, RTC1, RTC)
 	).
 
+:- pred relation__add_cartesian_product(list(relation_key), list(relation_key),
+		relation(T), relation(T)).
+:- mode relation__add_cartesian_product(in, in, in, out) is det.
+
+relation__add_cartesian_product([], _, RTC, RTC).
+relation__add_cartesian_product([K1 | Ks1], Ks2, RTC0, RTC) :-
+	relation__add_cartesian_product_2(K1, Ks2, RTC0, RTC1),
+	relation__add_cartesian_product(Ks1, Ks2, RTC1, RTC).
+
+:- pred relation__add_cartesian_product_2(relation_key, list(relation_key),
+		relation(T), relation(T)).
+:- mode relation__add_cartesian_product_2(in, in, in, out) is det.
+
+relation__add_cartesian_product_2(_, [], RTC, RTC).
+relation__add_cartesian_product_2(K1, [K2 | Ks2], RTC0, RTC) :-
+	relation__add(RTC0, K1, K2, RTC1),
+	relation__add_cartesian_product_2(K1, Ks2, RTC1, RTC).
+
 %------------------------------------------------------------------------------%
 %------------------------------------------------------------------------------%
+% Ralph Becket <rwab1@cl.cam.ac.uk> 30/04/99
+% 	Function forms added.
+
+:- interface.
+
+:- func relation__init = relation(T).
+
+:- func relation__lookup_element(relation(T), T) = relation_key.
+
+:- func relation__lookup_key(relation(T), relation_key) = T.
+
+:- func relation__add(relation(T), relation_key, relation_key) = relation(T).
+
+:- func relation__add_values(relation(T), T, T) = relation(T).
+
+:- func relation__add_assoc_list(relation(T), assoc_list(relation_key, relation_key)) = relation(T).
+
+:- func relation__remove(relation(T), relation_key, relation_key) = relation(T).
+
+:- func relation__remove_assoc_list(relation(T), assoc_list(relation_key, relation_key)) = relation(T).
+
+:- func relation__lookup_from(relation(T), relation_key) = set(relation_key).
+
+:- func relation__lookup_to(relation(T), relation_key) = set(relation_key).
+
+:- func relation__to_assoc_list(relation(T)) = assoc_list(T, T).
+
+:- func relation__to_key_assoc_list(relation(T)) = assoc_list(relation_key, relation_key).
+
+:- func relation__from_assoc_list(assoc_list(T, T)) = relation(T).
+
+:- func relation__domain(relation(T)) = set(T).
+
+:- func relation__inverse(relation(T)) = relation(T).
+
+:- func relation__compose(relation(T), relation(T)) = relation(T).
+
+:- func relation__dfs(relation(T), relation_key) = list(relation_key).
+
+:- func relation__dfsrev(relation(T), relation_key) = list(relation_key).
+
+:- func relation__dfs(relation(T)) = list(relation_key).
+
+:- func relation__dfsrev(relation(T)) = list(relation_key).
+
+:- func relation__components(relation(T)) = set(set(relation_key)).
+
+:- func relation__cliques(relation(T)) = set(set(relation_key)).
+
+:- func relation__reduced(relation(T)) = relation(set(T)).
+
+:- func relation__atsort(relation(T)) = list(set(T)).
+
+:- func relation__sc(relation(T)) = relation(T).
+
+:- func relation__tc(relation(T)) = relation(T).
+
+:- func relation__rtc(relation(T)) = relation(T).
+
+% ---------------------------------------------------------------------------- %
+% ---------------------------------------------------------------------------- %
+
+:- implementation.
+
+relation__init = R :-
+	relation__init(R).
+
+relation__lookup_element(R, X) = K :-
+	relation__lookup_element(R, X, K).
+
+relation__lookup_key(R, K) = X :-
+	relation__lookup_key(R, K, X).
+
+relation__add(R1, K1, K2) = R2 :-
+	relation__add(R1, K1, K2, R2).
+
+relation__add_values(R1, X, Y) = R2 :-
+	relation__add_values(R1, X, Y, R2).
+
+relation__add_assoc_list(R1, AL) = R2 :-
+	relation__add_assoc_list(R1, AL, R2).
+
+relation__remove(R1, K1, K2) = R2 :-
+	relation__remove(R1, K1, K2, R2).
+
+relation__remove_assoc_list(R1, AL) = R2 :-
+	relation__remove_assoc_list(R1, AL, R2).
+
+relation__lookup_from(R, K) = S :-
+	relation__lookup_from(R, K, S).
+
+relation__lookup_to(R, K) = S :-
+	relation__lookup_to(R, K, S).
+
+relation__to_assoc_list(R) = AL :-
+	relation__to_assoc_list(R, AL).
+
+relation__to_key_assoc_list(R) = AL :-
+	relation__to_key_assoc_list(R, AL).
+
+relation__from_assoc_list(AL) = R :-
+	relation__from_assoc_list(AL, R).
+
+relation__domain(R) = S :-
+	relation__domain(R, S).
+
+relation__inverse(R1) = R2 :-
+	relation__inverse(R1, R2).
+
+relation__compose(R1, R2) = R3 :-
+	relation__compose(R1, R2, R3).
+
+relation__dfs(R, K) = Ks :-
+	relation__dfs(R, K, Ks).
+
+relation__dfsrev(R, K) = Ks :-
+	relation__dfsrev(R, K, Ks).
+
+relation__dfs(R) = Ks :-
+	relation__dfs(R, Ks).
+
+relation__dfsrev(R) = Ks :-
+	relation__dfsrev(R, Ks).
+
+relation__components(R) = KSS :-
+	relation__components(R, KSS).
+
+relation__cliques(R) = KSS :-
+	relation__cliques(R, KSS).
+
+relation__reduced(R1) = R2 :-
+	relation__reduced(R1, R2).
+
+relation__atsort(R) = Ss :-
+	relation__atsort(R, Ss).
+
+relation__sc(R1) = R2 :-
+	relation__sc(R1, R2).
+
+relation__tc(R1) = R2 :-
+	relation__tc(R1, R2).
+
+relation__rtc(R1) = R2 :-
+	relation__rtc(R1, R2).
+
