@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1997, 1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -191,8 +191,6 @@ set_unordlist__sorted_list_to_set(List, List).
 set_unordlist__to_sorted_list(Set, List) :-
 	list__sort_and_remove_dups(Set, List).
 
-:- set_unordlist__insert_list(_, Xs, _) when Xs.	% NU-Prolog indexing.
-
 set_unordlist__insert_list(Set0, List, Set) :-
 	list__append(List, Set0, Set).
 
@@ -226,8 +224,6 @@ set_unordlist__is_member(E, S, R) :-
 		R = no
 	).
 
-:- set_unordlist__delete_list(_, Xs, _) when Xs.
-
 set_unordlist__delete_list(S, [], S).
 set_unordlist__delete_list(S0, [X | Xs], S) :-
 	set_unordlist__delete(S0, X, S1),
@@ -236,8 +232,6 @@ set_unordlist__delete_list(S0, [X | Xs], S) :-
 set_unordlist__delete(S0, E, S) :-
 	list__delete_all(S0, E, S).
 
-:- set_unordlist__remove_list(_, Xs, _) when Xs.
-
 set_unordlist__remove_list(S, [], S).
 set_unordlist__remove_list(S0, [X | Xs], S) :-
 	set_unordlist__remove(S0, X, S1),
@@ -312,3 +306,100 @@ set_unordlist__difference_2([E|Es], A, C) :-
 	set_unordlist__difference_2(Es, B, C).
 
 %--------------------------------------------------------------------------%
+%--------------------------------------------------------------------------%
+% Ralph Becket <rwab1@cam.sri.com> 24/04/99
+%	Function forms added.
+
+:- interface.
+
+:- func set_unordlist__list_to_set(list(T)) = set_unordlist(T).
+
+:- func set_unordlist__sorted_list_to_set(list(T)) = set_unordlist(T).
+
+:- func set_unordlist__to_sorted_list(set_unordlist(T)) = list(T).
+
+:- func set_unordlist__init = set_unordlist(T).
+
+:- func set_unordlist__make_singleton_set(T) = set_unordlist(T).
+
+:- func set_unordlist__insert(set_unordlist(T), T) = set_unordlist(T).
+
+:- func set_unordlist__insert_list(set_unordlist(T), list(T)) = set_unordlist(T).
+
+:- func set_unordlist__delete(set_unordlist(T), T) = set_unordlist(T).
+
+:- func set_unordlist__delete_list(set_unordlist(T), list(T)) = set_unordlist(T).
+
+:- func set_unordlist__union(set_unordlist(T), set_unordlist(T)) = set_unordlist(T).
+
+:- func set_unordlist__power_union(set_unordlist(set_unordlist(T))) = set_unordlist(T).
+
+:- func set_unordlist__intersect(set_unordlist(T), set_unordlist(T)) = set_unordlist(T).
+
+:- func set_unordlist__power_intersect(set_unordlist(set_unordlist(T))) = set_unordlist(T).
+
+:- func set_unordlist__difference(set_unordlist(T), set_unordlist(T)) = set_unordlist(T).
+
+:- func set_unordlist__map(func(T1) = T2, set_unordlist(T1)) = set_unordlist(T2).
+
+:- func set_unordlist__filter_map(func(T1) = T2, set_unordlist(T1)) = set_unordlist(T2).
+:- mode set_unordlist__filter_map(func(in) = out is semidet, in) = out is det.
+
+:- func set_unordlist__fold(func(T1, T2) = T2, set_unordlist(T1), T2) = T2.
+
+% ---------------------------------------------------------------------------- %
+% ---------------------------------------------------------------------------- %
+
+:- implementation.
+
+set_unordlist__list_to_set(Xs) = S :-
+	set_unordlist__list_to_set(Xs, S).
+
+set_unordlist__sorted_list_to_set(Xs) = S :-
+	set_unordlist__sorted_list_to_set(Xs, S).
+
+set_unordlist__to_sorted_list(S) = Xs :-
+	set_unordlist__to_sorted_list(S, Xs).
+
+set_unordlist__init = S :-
+	set_unordlist__init(S).
+
+set_unordlist__make_singleton_set(T) = S :-
+	set_unordlist__singleton_set(S, T).
+
+set_unordlist__insert(S1, T) = S2 :-
+	set_unordlist__insert(S1, T, S2).
+
+set_unordlist__insert_list(S1, Xs) = S2 :-
+	set_unordlist__insert_list(S1, Xs, S2).
+
+set_unordlist__delete(S1, T) = S2 :-
+	set_unordlist__delete(S1, T, S2).
+
+set_unordlist__delete_list(S1, Xs) = S2 :-
+	set_unordlist__delete_list(S1, Xs, S2).
+
+set_unordlist__union(S1, S2) = S3 :-
+	set_unordlist__union(S1, S2, S3).
+
+set_unordlist__power_union(SS) = S :-
+	set_unordlist__power_union(SS, S).
+
+set_unordlist__intersect(S1, S2) = S3 :-
+	set_unordlist__intersect(S1, S2, S3).
+
+set_unordlist__power_intersect(SS) = S :-
+	set_unordlist__power_intersect(SS, S).
+
+set_unordlist__difference(S1, S2) = S3 :-
+	set_unordlist__difference(S1, S2, S3).
+
+set_unordlist__map(F, S1) = S2 :-
+	S2 = set_unordlist__list_to_set(list__map(F, set_unordlist__to_sorted_list(S1))).
+
+set_unordlist__filter_map(PF, S1) = S2 :-
+	S2 = set_unordlist__list_to_set(list__filter_map(PF, set_unordlist__to_sorted_list(S1))).
+
+set_unordlist__fold(F, S, A) = B :-
+	B = list__foldl(F, set_unordlist__to_sorted_list(S), A).
+
