@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1994-1995, 1997 The University of Melbourne.
+% Copyright (C) 1994-1995, 1997-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -30,9 +30,8 @@
 :- mode stack__is_empty(in) is semidet.
 
 	% `stack__is_full(Stack)' is intended to be true iff `Stack'
-	% is a stack whose capacity is exhausted.  This
-	% implement allows arbitrary-sized stacks, so stack__is_full
-	% always fails.
+	% is a stack whose capacity is exhausted.  This implementation
+	% allows arbitrary-sized stacks, so stack__is_full always fails.
 
 :- pred stack__is_full(stack(_T)).
 :- mode stack__is_full(in) is semidet.
@@ -57,6 +56,12 @@
 :- pred stack__top(stack(T), T).
 :- mode stack__top(in, out) is semidet.
 
+	% `stack__top_det' is like `stack__top' except that it will
+	% call error/1 rather than failing if given an empty stack.
+
+:- pred stack__top_det(stack(T), T).
+:- mode stack__top_det(in, out) is det.
+
 	% `stack__pop(Stack0, Elem, Stack)' is true iff `Stack0' is
 	% a non-empty stack whose top element is `Elem', and `Stack'
 	% the stack which results from popping `Elem' off `Stack0'.
@@ -101,6 +106,13 @@ stack__push_list(Stack0, [Elem | Elems], Stack1) :-
 
 stack__top([Elem | _], Elem).
 
+stack__top_det(Stack, Elem) :-
+	( Stack = [Elem1 | _] ->
+		Elem = Elem1
+	;
+		error("stack__top_det: top of empty stack")
+	).
+
 stack__pop([Elem | Stack], Elem, Stack).
 
 stack__pop_det(Stack0, Elem, Stack) :-
@@ -115,3 +127,40 @@ stack__depth(Stack, Depth) :-
 	list__length(Stack, Depth).
 
 %--------------------------------------------------------------------------%
+%--------------------------------------------------------------------------%
+% Ralph Becket <rwab1@cl.cam.ac.uk> 29/04/99
+% 	Function forms added.
+
+:- interface.
+
+:- func stack__init = stack(T).
+
+:- func stack__push(stack(T), T) = stack(T).
+
+:- func stack__push_list(stack(T), list(T)) = stack(T).
+
+:- func stack__top_det(stack(T)) = T.
+
+:- func stack__depth(stack(T)) = int.
+
+% ---------------------------------------------------------------------------- %
+% ---------------------------------------------------------------------------- %
+
+:- implementation.
+
+stack__init = S :-
+	stack__init(S).
+
+stack__push(S1, X) = S2 :-
+	stack__push(S1, X, S2).
+
+stack__push_list(S1, Xs) = S2 :-
+	stack__push_list(S1, Xs, S2).
+
+stack__top_det(S) = X :-
+	stack__top_det(S, X).
+
+stack__depth(S) = N :-
+	stack__depth(S, N).
+
+
