@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1997, 1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -75,28 +75,33 @@
 :- pred store__set_mutvar(mutvar(T, S), T, store(S), store(S)).
 :- mode store__set_mutvar(in, in, di, uo) is det.
 
-/* 
-The syntax might be nicer if we used some new operators
-
-	:- op(.., xfx, ('<-')).
-	:- op(.., fy, ('!')).
-	:- op(.., xfx, (':=')).
-
-Then we could do something like this:
-
-	Ptr <- new(Val)	  -->	new_mutvar(Val, Ptr).
-	Val <- !Ptr 	  -->	get_mutvar(Ptr, Val).
-	!Ptr := Val	  -->	set_mutvar(Ptr, Val).
-
-I wonder whether it is worth it?
-*/
+	% new_cyclic_mutvar(Func, Mutvar):
+	% create a new mutable variable, whose value is initialized
+	% with the value returned from the specified function `Func'.
+	% The argument passed to the function is the mutvar itself,
+	% whose value has not yet been initialized (this is safe
+	% because the function does not get passed the store, so
+	% it can't examine the uninitialized value).
+	%
+	% This predicate is useful for creating self-referential values
+	% such as circular linked lists. 
+	% For example:
+	%	:- type clist(T, S) ---> node(T, mutvar(clist(T, S))).
+	%	:- pred init_cl(T::in, clist(T, S)::out,
+	%			store(S)::di, store(S)::uo) is det.
+	%	init_cl(X, CList) -->
+	%	    store__new_cyclic_mutvar(func(CL) = node(X, CL), CList).
+	%
+:- pred store__new_cyclic_mutvar(func(mutvar(T, S)) = T, mutvar(T, S),
+		store(S), store(S)).
+:- mode store__new_cyclic_mutvar(in, out, di, uo) is det.
 
 %-----------------------------------------------------------------------------%
 %
 % references
 %
 
-	% mutvar(T, S):
+	% ref(T, S):
 	% a reference to value of type T in store S
 :- type ref(T, S).
 
@@ -150,14 +155,14 @@ I wonder whether it is worth it?
 	% Given a reference to a term (Ref), and a value (Value),
 	% update the store so that the term referred to by Ref
 	% is replaced with Value.
-	% (Argument numbers start from zero).
-:- pred store__set_ref_value(ref(T, S), ArgT, store(S), store(S)).
+:- pred store__set_ref_value(ref(T, S), T, store(S), store(S)).
 :- mode store__set_ref_value(in, di, di, uo) is det.
 
 	% Given a reference to a term, return that term.
 	% Note that this requires making a copy, so this pred may
 	% be inefficient if used to return large terms; it
 	% is most efficient with atomic terms.
+	% XXX current implementation buggy (does shallow copy)
 :- pred store__copy_ref_value(ref(T, S), T, store(S), store(S)).
 :- mode store__copy_ref_value(in, uo, di, uo) is det.
 
@@ -188,7 +193,8 @@ I wonder whether it is worth it?
 	% or if the argument is a `no_tag' type,
 	% then the behaviour is undefined, and probably harmful.
 
-:- pred store__unsafe_arg_ref(ref(T, S), int, ref(ArgT, S), store(S), store(S)).
+:- pred store__unsafe_arg_ref(ref(T, S), int, ref(ArgT, S),
+				store(S), store(S)).
 :- mode store__unsafe_arg_ref(in, in, out, di, uo) is det.
 
 :- pred store__unsafe_new_arg_ref(T, int, ref(ArgT, S), store(S), store(S)).
@@ -210,10 +216,27 @@ I wonder whether it is worth it?
 
 :- pragma c_code(init(_S0::uo), will_not_call_mercury, "").
 
+/* 
+Note -- the syntax for the operations on stores
+might be nicer if we used some new operators, e.g.
+
+	:- op(.., xfx, ('<-')).
+	:- op(.., fy, ('!')).
+	:- op(.., xfx, (':=')).
+
+Then we could do something like this:
+
+	Ptr <- new(Val)	  -->	new_mutvar(Val, Ptr).
+	Val <- !Ptr 	  -->	get_mutvar(Ptr, Val).
+	!Ptr := Val	  -->	set_mutvar(Ptr, Val).
+
+I wonder whether it is worth it?  Hmm, probably not.
+*/
+
 :- pragma c_code(new_mutvar(Val::in, Mutvar::out, S0::di, S::uo),
 		will_not_call_mercury,
 "
-	incr_hp(Mutvar, 1);
+	incr_hp_msg(Mutvar, 1, MR_PROC_LABEL, ""store:mutvar/2"");
 	*(Word *)Mutvar = Val;
 	S = S0;
 ").
@@ -232,12 +255,28 @@ I wonder whether it is worth it?
 	S = S0;
 ").
 
+:- pred store__unsafe_new_uninitialized_mutvar(mutvar(T, S),
+						store(S), store(S)).
+:- mode store__unsafe_new_uninitialized_mutvar(out, di, uo) is det.
+
+:- pragma c_code(unsafe_new_uninitialized_mutvar(Mutvar::out, S0::di, S::uo),
+		will_not_call_mercury,
+"
+	incr_hp_msg(Mutvar, 1, MR_PROC_LABEL, ""store:mutvar/2"");
+	S = S0;
+").
+
+store__new_cyclic_mutvar(Func, MutVar) -->
+	store__unsafe_new_uninitialized_mutvar(MutVar),
+	{ Value = apply(Func, MutVar) },
+	store__set_mutvar(MutVar, Value).
+
 %-----------------------------------------------------------------------------%
 
 :- pragma c_code(new_ref(Val::di, Ref::out, S0::di, S::uo),
 		will_not_call_mercury,
 "
-	incr_hp(Ref, 1);
+	incr_hp_msg(Ref, 1, MR_PROC_LABEL, ""store:ref/2"");
 	*(Word *)Ref = Val;
 	S = S0;
 ").
@@ -264,12 +303,12 @@ ref_functor(Ref, Functor, Arity) -->
 	{ functor(Val, Functor, Arity) }.
 
 :- pragma c_header_code("
+	#include ""mercury_type_info.h""
+
 	/* ML_arg() is defined in std_util.m */
 	bool ML_arg(Word term_type_info, Word *term, Word argument_index,
 			Word *arg_type_info, Word **argument_ptr);
 
-	/* ML_compare_type_info() is defined in std_util.m */
-	int ML_compare_type_info(Word type_info_1, Word type_info_2);
 ").
 
 :- pragma c_code(arg_ref(Ref::in, ArgNum::in, ArgRef::out, S0::di, S::uo),
@@ -286,8 +325,8 @@ ref_functor(Ref, Functor, Arity) -->
 		fatal_error(""store__arg_ref: argument number out of range"");
 	}
 
-	if (ML_compare_type_info(arg_type_info, TypeInfo_for_ArgT) !=
-		COMPARE_EQUAL)
+	if (MR_compare_type_info(arg_type_info, TypeInfo_for_ArgT) !=
+		MR_COMPARE_EQUAL)
 	{
 		fatal_error(""store__arg_ref: argument has wrong type"");
 	}
@@ -312,8 +351,8 @@ ref_functor(Ref, Functor, Arity) -->
 	      fatal_error(""store__new_arg_ref: argument number out of range"");
 	}
 
-	if (ML_compare_type_info(arg_type_info, TypeInfo_for_ArgT) !=
-		COMPARE_EQUAL)
+	if (MR_compare_type_info(arg_type_info, TypeInfo_for_ArgT) !=
+		MR_COMPARE_EQUAL)
 	{
 	      fatal_error(""store__new_arg_ref: argument has wrong type"");
 	}
@@ -327,7 +366,7 @@ ref_functor(Ref, Functor, Arity) -->
 	** to copy it to the heap before returning.
 	*/
 	if (arg_ref == &Val) {
-		incr_hp(ArgRef, 1);
+		incr_hp_msg(ArgRef, 1, MR_PROC_LABEL, ""store:ref/2"");
 		*(Word *)ArgRef = Val;
 	} else {
 		ArgRef = (Word) arg_ref;
@@ -361,7 +400,7 @@ ref_functor(Ref, Functor, Arity) -->
 		will_not_call_mercury,
 "{
 	/* unsafe - does not check type & arity, won't handle no_tag types */
-	Word *Ptr = (Word *) strip_tag(Ref);
+	Word *Ptr = (Word *) MR_strip_tag(Ref);
 	ArgRef = (Word) &Ptr[Arg];
 	S = S0;
 }").
@@ -370,7 +409,7 @@ ref_functor(Ref, Functor, Arity) -->
 				S0::di, S::uo), will_not_call_mercury,
 "{
 	/* unsafe - does not check type & arity, won't handle no_tag types */
-	Word *Ptr = (Word *) strip_tag(Val);
+	Word *Ptr = (Word *) MR_strip_tag(Val);
 	ArgRef = (Word) &Ptr[Arg];
 	S = S0;
 }").
