@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1993-1997 The University of Melbourne.
+% Copyright (C) 1993-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -11,16 +11,13 @@
 
 % This modules provides basic string handling facilities.
 
-% Beware that char_to_string/2 won't work with NU-Prolog 1.5.33 because
-% of a NU-Prolog bug (fixed in NU-Prolog 1.5.35).
-
 %-----------------------------------------------------------------------------%
 
 :- interface.
 :- import_module list, char.
 
 :- pred string__length(string, int).
-:- mode string__length(in, out) is det.
+:- mode string__length(in, uo) is det.
 	% Determine the length of a string.
 	% An empty string has length zero.
 
@@ -81,12 +78,14 @@
 
 :- pred string__replace(string, string, string, string).
 :- mode string__replace(in, in, in, out) is semidet.
+% string__replace(String0, Search, Replace, String):
 % 	string__replace replaces the first occurence of the second string in 
 % 	the first string with the third string to give the fourth string.
 % 	It fails if the second string does not occur in the first.
 
 :- pred string__replace_all(string, string, string, string).
 :- mode string__replace_all(in, in, in, out) is det.
+% string__replace_all(String0, Search, Replace, String):
 % 	string__replace_all replaces any occurences of the second string in 
 % 	the first string with the third string to give the fourth string.
 
@@ -190,6 +189,16 @@
 %	Calls error/1 if `Index' is out of range (negative, or greater than or
 %	equal to the length of `String').
 
+:- pred string__unsafe_index(string, int, char).
+:- mode string__unsafe_index(in, in, out) is det.
+%	string__unsafe_index(String, Index, Char):
+%	`Char' is the (`Index' + 1)-th character of `String'.
+%	WARNING: behavior is UNDEFINED if `Index' is out of range
+%	(negative, or greater than or equal to the length of `String').
+%	This version is constant time, whereas string__index_det
+%	may be linear in the length of the string.
+%	Use with care!
+
 :- pred string__foldl(pred(char, T, T), string, T, T).
 :- mode string__foldl(pred(in, in, out) is det, in, in, out) is det.
 :- mode string__foldl(pred(in, di, uo) is det, in, di, uo) is det.
@@ -238,6 +247,20 @@
 %	If `Count' is out of the range [0, length of `String' - `Start'], it is
 %	treated as if it were the nearest end-point of that range.)
 
+:- pred string__unsafe_substring(string, int, int, string).
+:- mode string__unsafe_substring(in, in, in, out) is det.
+%	string__unsafe_substring(String, Start, Count, Substring):
+%	`Substring' is first the `Count' characters in what would
+%	remain of `String' after the first `Start' characters were
+%	removed.
+%	WARNING: if `Start' is out of the range [0, length of `String'],
+%	or if `Count' is out of the range [0, length of `String' - `Start'],
+%	then the behaviour is UNDEFINED.
+%	Use with care!
+%	This version takes time proportional to the length of the
+%	substring, whereas string__substring may take time proportional
+%	to the length of the whole string.
+
 :- pred string__append_list(list(string), string).
 :- mode string__append_list(in, out) is det.
 :- mode string__append_list(out, in) is multidet.
@@ -481,10 +504,6 @@ string__right(String, RightCount, RightString) :-
 	LeftCount is Length - RightCount,
 	string__split(String, LeftCount, _LeftString, RightString).
 
-string__substring(String, Start, Count, Substring) :-
-	string__split(String, Start, _Left, Right),
-	string__left(Right, Count, Substring).
-
 string__remove_suffix(A, B, C) :-
 	string__to_int_list(A, LA),
 	string__to_int_list(B, LB),
@@ -494,7 +513,6 @@ string__remove_suffix(A, B, C) :-
 string__prefix(String, Prefix) :-
 	string__append(Prefix, _, String).
 
-:- string__char_to_string(Char, String) when Char or String.
 string__char_to_string(Char, String) :-
 	string__to_int_list(String, [Code]),
 	char__to_int(Char, Code).
@@ -516,31 +534,36 @@ string__int_to_base_string(N, Base, Str) :-
 :- mode string__int_to_base_string_1(in, in, out) is det.
 
 string__int_to_base_string_1(N, Base, Str) :-
+	% Note that in order to handle MININT correctly,
+	% we need to do the conversion of the absolute
+	% number into digits using negative numbers
+	% (we can't use positive numbers, since -MININT overflows)
 	(
 		N < 0
 	->
-		N1 is 0 - N,
-		string__int_to_base_string_2(N1, Base, Str1),
+		string__int_to_base_string_2(N, Base, Str1),
 		string__append("-", Str1, Str)
 	;
-		string__int_to_base_string_2(N, Base, Str)
+		N1 is 0 - N,
+		string__int_to_base_string_2(N1, Base, Str)
 	).
 
 :- pred string__int_to_base_string_2(int, int, string).
 :- mode string__int_to_base_string_2(in, in, out) is det.
 
-string__int_to_base_string_2(N, Base, Str) :-
+string__int_to_base_string_2(NegN, Base, Str) :-
 	(
-		N < Base
+		NegN > -Base
 	->
+		N is -NegN,
 		char__det_int_to_digit(N, DigitChar),
 		string__char_to_string(DigitChar, Str)
 	;
-		N10 is N mod Base,
-		N1 is N // Base,
+		NegN1 is NegN // Base,
+		N10 is (NegN1 * Base) - NegN,
 		char__det_int_to_digit(N10, DigitChar),
 		string__char_to_string(DigitChar, DigitString),
-		string__int_to_base_string_2(N1, Base, Str1),
+		string__int_to_base_string_2(NegN1, Base, Str1),
 		string__append(Str1, DigitString, Str)
 	).
 
@@ -559,7 +582,8 @@ string__from_char_list(CharList, String) :-
 % but the optimized implementation in C below is there for efficiency since
 % it improves the overall speed of parsing by about 7%.
 %
-:- pragma(c_code, string__from_rev_char_list(Chars::in, Str::out), "
+:- pragma c_code(string__from_rev_char_list(Chars::in, Str::out),
+		[will_not_call_mercury, thread_safe], "
 {
 	Word list_ptr;
 	Word size, len;
@@ -571,16 +595,18 @@ string__from_char_list(CharList, String) :-
 	size = sizeof(Word);
 	len = 1;
 	list_ptr = Chars;
-	while (!list_is_empty(list_ptr)) {
+	while (!MR_list_is_empty(list_ptr)) {
 		size++;
 		len++;
-		list_ptr = list_tail(list_ptr);
+		list_ptr = MR_list_tail(list_ptr);
 	}
 /*
 ** allocate (length + 1) bytes of heap space for string
 ** i.e. (length + 1 + sizeof(Word) - 1) / sizeof(Word) words
 */
-	incr_hp_atomic(str_ptr, size / sizeof(Word));
+	incr_hp_atomic_msg(str_ptr, size / sizeof(Word),
+		MR_PROC_LABEL,
+		""string:string/0"");
 	Str = (char *) str_ptr;
 /*
 ** set size to be the offset of the end of the string
@@ -592,9 +618,9 @@ string__from_char_list(CharList, String) :-
 ** in reverse order.
 */
 	list_ptr = Chars;
-	while (!list_is_empty(list_ptr)) {
-		Str[--len] = (char) list_head(list_ptr);
-		list_ptr = list_tail(list_ptr);
+	while (!MR_list_is_empty(list_ptr)) {
+		Str[--len] = (char) MR_list_head(list_ptr);
+		list_ptr = MR_list_tail(list_ptr);
 	}
 }").
 
@@ -726,7 +752,7 @@ string__hash(String, HashVal) :-
 	string__length(String, Length),
 	string__to_int_list(String, CodeList),
 	string__hash_2(CodeList, 0, HashVal0),
-	HashVal is HashVal0 ^ Length.
+	HashVal = HashVal0 `xor` Length.
 
 :- pred string__hash_2(list(int), int, int).
 :- mode string__hash_2(in, in, out) is det.
@@ -740,9 +766,9 @@ string__hash_2([X | Xs], HashVal0, HashVal) :-
 :- mode string__combine_hash(in, in, out) is det.
 
 string__combine_hash(H0, X, H) :-
-	H1 is H0 << 5,
-	H2 is H1 ^ H0,
-	H is H2 ^ X.
+	H1 = H0 << 5,
+	H2 = H1 `xor` H0,
+	H = H2 `xor` X.
 
 %-----------------------------------------------------------------------------%
 
@@ -787,7 +813,13 @@ string__format(Format_string, Poly_list, Out_string ) :-
 %	is not efficient.  Instead, many short strings are added to the front
 %	of a (long) and growing string.
 %
-string__format_2([], _, "").
+string__format_2([], Vars_in, Result) :-
+	( Vars_in = [] ->
+		Result = ""
+	;
+		error(
+	"string__format: argument list has more elements than format string")
+	).
 string__format_2([Achar|As], Vars_in, Ostring) :-
 	(
 		Achar = '%'
@@ -805,7 +837,8 @@ string__format_2([Achar|As], Vars_in, Ostring) :-
 				string__format_2(As_out, Vars_out, String_2),
 				string__append(String_1, String_2, Ostring)
 			;
-				error("string__format: Too few variables.")
+				error(
+	"string__format: argument list has fewer elements than format string")
 			)
 		)
 	;
@@ -880,7 +913,7 @@ string__format_mod_conv_char(Precision0, Poly_var, Conversion_in,
 	; 
 	Conversion_in = 'g' ->			%g is either %e of %f
 		(Poly_var = f(F) ->
-			float__abs(F, Ft),
+			Ft = float__abs(F),
 			int__pow(10, Prec, P),
 			int__to_float(P, Pe),
 			(
@@ -899,7 +932,7 @@ string__format_mod_conv_char(Precision0, Poly_var, Conversion_in,
 	;
 	Conversion_in = 'G' ->		%G is either %E of %f
 		(Poly_var = f(F) ->
-			float__abs(F, Ft),
+			Ft = float__abs(F),
 			int__pow(10, Prec, P),
 			int__to_float(P, Pe),
 			(
@@ -1515,18 +1548,20 @@ string__special_precision_and_width(-1).
 
 % The remaining routines are implemented using the C interface.
 
-:- pragma(c_header_code, "
+:- pragma c_header_code("
 #include <string.h>
 #include <stdio.h>
 ").
 
 %-----------------------------------------------------------------------------%
 
-:- pragma(c_code, string__float_to_string(FloatVal::in, FloatString::out), "{
+:- pragma c_code(string__float_to_string(FloatVal::in, FloatString::out),
+		[will_not_call_mercury, thread_safe], "{
 	char buf[500];
 	Word tmp;
 	sprintf(buf, ""%#.15g"", FloatVal);
-	incr_hp_atomic(tmp, (strlen(buf) + sizeof(Word)) / sizeof(Word));
+	incr_hp_atomic_msg(tmp, (strlen(buf) + sizeof(Word)) / sizeof(Word),
+		MR_PROC_LABEL, ""string:string/0"");
 	FloatString = (char *)tmp;
 	strcpy(FloatString, buf);
 }").
@@ -1536,16 +1571,19 @@ string__special_precision_and_width(-1).
 
 :- pred string__float_to_f_string(float::in, string::out) is det.
 
-:- pragma(c_code, string__float_to_f_string(FloatVal::in, FloatString::out), "{
+:- pragma c_code(string__float_to_f_string(FloatVal::in, FloatString::out),
+		[will_not_call_mercury, thread_safe], "{
 	char buf[500];
 	Word tmp;
 	sprintf(buf, ""%.15f"", FloatVal);
-	incr_hp_atomic(tmp, (strlen(buf) + sizeof(Word)) / sizeof(Word));
+	incr_hp_atomic_msg(tmp, (strlen(buf) + sizeof(Word)) / sizeof(Word),
+		MR_PROC_LABEL, ""string:string/0"");
 	FloatString = (char *)tmp;
 	strcpy(FloatString, buf);
 }").
 
-:- pragma(c_code, string__to_float(FloatString::in, FloatVal::out), "{
+:- pragma c_code(string__to_float(FloatString::in, FloatVal::out),
+		[will_not_call_mercury, thread_safe], "{
 	/* use a temporary, since we can't don't know whether FloatVal
 	   is a double or float */
 	double tmp;
@@ -1562,16 +1600,19 @@ string__special_precision_and_width(-1).
 :- mode string__to_int_list(out, in) is det.
 */
 
-:- pragma(c_code, string__to_int_list(Str::in, IntList::out), "{
+:- pragma c_code(string__to_int_list(Str::in, IntList::out),
+		[will_not_call_mercury, thread_safe], "{
 	const char *p = Str + strlen(Str);
-	IntList = list_empty();
+	IntList = MR_list_empty_msg(MR_PROC_LABEL);
 	while (p > Str) {
 		p--;
-		IntList = list_cons((UnsignedChar) *p, IntList);
+		IntList = MR_list_cons_msg((UnsignedChar) *p, IntList,
+			MR_PROC_LABEL);
 	}
 }").
 
-:- pragma(c_code, string__to_int_list(Str::out, IntList::in), "{
+:- pragma c_code(string__to_int_list(Str::out, IntList::in),
+		[will_not_call_mercury, thread_safe], "{
 		/* mode (out, in) is det */
 	Word int_list_ptr;
 	size_t size;
@@ -1582,24 +1623,25 @@ string__special_precision_and_width(-1).
 */
 	size = sizeof(Word);
 	int_list_ptr = IntList;
-	while (!list_is_empty(int_list_ptr)) {
+	while (! MR_list_is_empty(int_list_ptr)) {
 		size++;
-		int_list_ptr = list_tail(int_list_ptr);
+		int_list_ptr = MR_list_tail(int_list_ptr);
 	}
 /*
 ** allocate (length + 1) bytes of heap space for string
 ** i.e. (length + 1 + sizeof(Word) - 1) / sizeof(Word) words
 */
-	incr_hp_atomic(str_ptr, size / sizeof(Word));
+	incr_hp_atomic_msg(str_ptr, size / sizeof(Word),
+		MR_PROC_LABEL, ""string:string/0"");
 	Str = (char *) str_ptr;
 /*
 ** loop to copy the characters from the int_list to the string
 */
 	size = 0;
 	int_list_ptr = IntList;
-	while (!list_is_empty(int_list_ptr)) {
-		Str[size++] = list_head(int_list_ptr);
-		int_list_ptr = list_tail(int_list_ptr);
+	while (! MR_list_is_empty(int_list_ptr)) {
+		Str[size++] = MR_list_head(int_list_ptr);
+		int_list_ptr = MR_list_tail(int_list_ptr);
 	}
 /*
 ** null terminate the string
@@ -1613,7 +1655,8 @@ string__special_precision_and_width(-1).
 :- pred string__contains_char(string, char).
 :- mode string__contains_char(in, in) is semidet.
 */
-:- pragma(c_code, string__contains_char(Str::in, Ch::in), "
+:- pragma c_code(string__contains_char(Str::in, Ch::in),
+		[will_not_call_mercury, thread_safe], "
 	SUCCESS_INDICATOR = (strchr(Str, Ch) != NULL);
 ").
 
@@ -1623,7 +1666,8 @@ string__special_precision_and_width(-1).
 :- pred string__index(string, int, char).
 :- mode string__index(in, in, out) is semidet.
 */
-:- pragma(c_code, string__index(Str::in, Index::in, Ch::out), "
+:- pragma c_code(string__index(Str::in, Index::in, Ch::out),
+		[will_not_call_mercury, thread_safe], "
 	if ((Word) Index >= strlen(Str)) {
 		SUCCESS_INDICATOR = FALSE;
 	} else {
@@ -1634,10 +1678,8 @@ string__special_precision_and_width(-1).
 
 /*-----------------------------------------------------------------------*/
 
-:- pred string__unsafe_index(string, int, char).
-:- mode string__unsafe_index(in, in, out) is det.
-
-:- pragma(c_code, string__unsafe_index(Str::in, Index::in, Ch::out), "
+:- pragma c_code(string__unsafe_index(Str::in, Index::in, Ch::out),
+		[will_not_call_mercury, thread_safe], "
 	Ch = Str[Index];
 ").
 
@@ -1647,7 +1689,8 @@ string__special_precision_and_width(-1).
 :- pred string__length(string, int).
 :- mode string__length(in, out) is det.
 */
-:- pragma(c_code, string__length(Str::in, Length::out), "
+:- pragma c_code(string__length(Str::in, Length::uo),
+		[will_not_call_mercury, thread_safe], "
 	Length = strlen(Str);
 ").
 
@@ -1664,7 +1707,8 @@ string__special_precision_and_width(-1).
 /*
 :- mode string__append(in, in, in) is semidet.
 */
-:- pragma(c_code, string__append(S1::in, S2::in, S3::in), "{
+:- pragma c_code(string__append(S1::in, S2::in, S3::in),
+		[will_not_call_mercury, thread_safe], "{
 	size_t len_1 = strlen(S1);
 	SUCCESS_INDICATOR = (
 		strncmp(S1, S3, len_1) == 0 &&
@@ -1675,7 +1719,8 @@ string__special_precision_and_width(-1).
 /*
 :- mode string__append(in, out, in) is semidet.
 */
-:- pragma(c_code, string__append(S1::in, S2::out, S3::in), "{
+:- pragma c_code(string__append(S1::in, S2::out, S3::in),
+		[will_not_call_mercury, thread_safe], "{
 	Word tmp;
 	size_t len_1, len_2, len_3;
 
@@ -1689,7 +1734,8 @@ string__special_precision_and_width(-1).
 		** We need to make a copy to ensure that the pointer is
 		** word-aligned.
 		*/
-		incr_hp_atomic(tmp, (len_2 + sizeof(Word)) / sizeof(Word));
+		incr_hp_atomic_msg(tmp, (len_2 + sizeof(Word)) / sizeof(Word),
+			MR_PROC_LABEL, ""string:string/0"");
 		S2 = (char *) tmp;
 		strcpy(S2, S3 + len_1);
 		SUCCESS_INDICATOR = TRUE;
@@ -1699,103 +1745,109 @@ string__special_precision_and_width(-1).
 /*
 :- mode string__append(in, in, out) is det.
 */
-:- pragma(c_code, string__append(S1::in, S2::in, S3::out), "{
+:- pragma c_code(string__append(S1::in, S2::in, S3::out),
+		[will_not_call_mercury, thread_safe], "{
 	size_t len_1, len_2;
 	Word tmp;
 	len_1 = strlen(S1);
 	len_2 = strlen(S2);
-	incr_hp_atomic(tmp, (len_1 + len_2 + sizeof(Word)) / sizeof(Word));
+	incr_hp_atomic_msg(tmp, (len_1 + len_2 + sizeof(Word)) / sizeof(Word),
+		MR_PROC_LABEL, ""string:string/0"");
 	S3 = (char *) tmp;
 	strcpy(S3, S1);
 	strcpy(S3 + len_1, S2);
 }").
 
-:- pragma(c_code, "
+:- pragma c_code(string__append(S1::out, S2::out, S3::in),
+		[will_not_call_mercury, thread_safe],
+	local_vars("
+		String s;
+		size_t len;
+		size_t count;
+	"),
+	first_code("
+		LOCALS->s = S3;
+		LOCALS->len = strlen(S3);
+		LOCALS->count = 0;
+	"),
+	retry_code("
+		LOCALS->count++;
+	"),
+	common_code("
+		Word	temp;
 
-#ifdef	COMPACT_ARGS
-#define	string__append_ooi_input_reg	r1
-#else
-#define	string__append_ooi_input_reg	r3
-#endif
+		incr_hp_atomic_msg(temp,
+			(LOCALS->count + sizeof(Word)) / sizeof(Word),
+			MR_PROC_LABEL, ""string:string/0"");
+		S1 = (String) temp;
+		memcpy(S1, LOCALS->s, LOCALS->count);
+		S1[LOCALS->count] = '\\0';
+		incr_hp_atomic_msg(temp,
+			(LOCALS->len - LOCALS->count + sizeof(Word))
+				/ sizeof(Word),
+			MR_PROC_LABEL, ""string:string/0"");
+		S2 = (String) temp;
+		strcpy(S2, LOCALS->s + LOCALS->count);
 
-Define_extern_entry(mercury__string__append_3_3_xx);
-Declare_label(mercury__string__append_3_3_xx_i1);
-
-BEGIN_MODULE(string_append_module)
-	init_entry(mercury__string__append_3_3_xx);
-	init_label(mercury__string__append_3_3_xx_i1);
-BEGIN_CODE
-Define_entry(mercury__string__append_3_3_xx);
-	mkframe(""string__append/3"", 4,
-		LABEL(mercury__string__append_3_3_xx_i1));
-	mark_hp(framevar(0));
-	framevar(1) = string__append_ooi_input_reg;
-	framevar(2) = strlen((char *) string__append_ooi_input_reg);
-	framevar(3) = 0;
-Define_label(mercury__string__append_3_3_xx_i1);
-{
-	String	s3;
-	size_t	s3_len;
-	size_t	count;
-
-	restore_hp(framevar(0));
-	s3 = (String) framevar(1);
-	s3_len = framevar(2);
-	count = framevar(3);
-	if (count > s3_len) {
-		/* modframe(ENTRY(do_fail)); */
-		fail();
-	}
-	incr_hp_atomic(r1, (count + sizeof(Word)) / sizeof(Word));
-	memcpy((char *) r1, s3, count);
-	((char *) r1)[count] = '\\0';
-	/*
-	** We need to make a copy to ensure that the pointer is
-	** word-aligned.
-	*/
-	incr_hp_atomic(r2, (s3_len - count + sizeof(Word)) / sizeof(Word));
-	strcpy((char *) r2, s3 + count);
-	framevar(3) = count + 1;
-	succeed();
-}
-END_MODULE
-
-#undef	string__append_ooi_input_reg
-
-/* Ensure that the initialization code for the above module gets run. */
-/*
-INIT sys_init_string_append_module
-*/
-/* suppress gcc -Wmissing-decl warning */
-void sys_init_string_append_module(void);
-
-void sys_init_string_append_module(void) {
-	extern ModuleFunc string_append_module;
-	string_append_module();
-}
-
-").
-
-% :- mode string__append(out, out, in) is multidet.
-:- pragma(c_code, string__append(S1::out, S2::out, S3::in), "
-	/*
-	** The pragma_c_code will generate a mkframe();
-	** we need to pop off that frame before jumping to the hand-coded
-	** fragment above.
-	**
-	** We mention S1, S2 and S3 here to shut up a warning.
-	*/
-
-	maxfr = curprevfr;
-	curfr = cursuccfr;
-	{
-		Declare_entry(mercury__string__append_3_3_xx);
-		GOTO(ENTRY(mercury__string__append_3_3_xx));
-	}
-").
+		if (LOCALS->count < LOCALS->len) {
+			SUCCEED;
+		} else {
+			SUCCEED_LAST;
+		}
+	")
+).
 
 /*-----------------------------------------------------------------------*/
 
+/*
+:- pred string__substring(string, int, int, string).
+:- mode string__substring(in, in, in, out) is det.
+%	string__substring(String, Start, Count, Substring):
+*/
+
+:- pragma c_code(string__substring(Str::in, Start::in, Count::in,
+		SubString::out),
+		[will_not_call_mercury, thread_safe],
+"{
+	Integer len;
+	Word tmp;
+	if (Start < 0) Start = 0;
+	if (Count <= 0) {
+		MR_make_aligned_string(LVALUE_CAST(ConstString, SubString),
+			"""");
+	} else {
+		len = strlen(Str);
+		if (Start > len) Start = len;
+		if (Count > len - Start) Count = len - Start;
+		incr_hp_atomic_msg(tmp, (Count + sizeof(Word)) / sizeof(Word),
+			MR_PROC_LABEL, ""string:string/0"");
+		SubString = (char *) tmp;
+		memcpy(SubString, Str + Start, Count);
+		SubString[Count] = '\\0';
+	}
+}").
+
+
+/*
+:- pred string__unsafe_substring(string, int, int, string).
+:- mode string__unsafe_substring(in, in, in, out) is det.
+%	string__unsafe_substring(String, Start, Count, Substring):
+*/
+
+:- pragma c_code(string__unsafe_substring(Str::in, Start::in, Count::in,
+		SubString::out),
+		[will_not_call_mercury, thread_safe],
+"{
+	Integer len;
+	Word tmp;
+	incr_hp_atomic_msg(tmp, (Count + sizeof(Word)) / sizeof(Word),
+		MR_PROC_LABEL, ""string:string/0"");
+	SubString = (char *) tmp;
+	memcpy(SubString, Str + Start, Count);
+	SubString[Count] = '\\0';
+}").
+
+
 /*
 :- pred string__split(string, int, string, string).
 :- mode string__split(in, in, out, out) is det.
@@ -1806,16 +1858,18 @@ void sys_init_string_append_module(void) {
 %	treated as if it were the nearest end-point of that range.)
 */
 
-:- pragma(c_code, string__split(Str::in, Count::in, Left::out, Right::out), "{
+:- pragma c_code(string__split(Str::in, Count::in, Left::out, Right::out),
+		[will_not_call_mercury, thread_safe], "{
 	Integer len;
 	Word tmp;
 	if (Count <= 0) {
-		make_aligned_string(LVALUE_CAST(ConstString, Left), """");
+		MR_make_aligned_string(LVALUE_CAST(ConstString, Left), """");
 		Right = Str;
 	} else {
 		len = strlen(Str);
 		if (Count > len) Count = len;
-		incr_hp_atomic(tmp, (Count + sizeof(Word)) / sizeof(Word));
+		incr_hp_atomic_msg(tmp, (Count + sizeof(Word)) / sizeof(Word),
+			MR_PROC_LABEL, ""string:string/0"");
 		Left = (char *) tmp;
 		memcpy(Left, Str, Count);
 		Left[Count] = '\\0';
@@ -1823,8 +1877,9 @@ void sys_init_string_append_module(void) {
 		** We need to make a copy to ensure that the pointer is
 		** word-aligned.
 		*/
-		incr_hp_atomic(tmp,
-			(len - Count + sizeof(Word)) / sizeof(Word));
+		incr_hp_atomic_msg(tmp,
+			(len - Count + sizeof(Word)) / sizeof(Word),
+			MR_PROC_LABEL, ""string:string/0"");
 		Right = (char *) tmp;
 		strcpy(Right, Str + Count);
 	}
@@ -1847,7 +1902,8 @@ void sys_init_string_append_module(void) {
 /*
 :- mode string__first_char(in, in, in) is semidet.	% implied
 */
-:- pragma(c_code, string__first_char(Str::in, First::in, Rest::in), "
+:- pragma c_code(string__first_char(Str::in, First::in, Rest::in),
+		[will_not_call_mercury, thread_safe], "
 	SUCCESS_INDICATOR = (
 		Str[0] == First &&
 		First != '\\0' &&
@@ -1858,7 +1914,8 @@ void sys_init_string_append_module(void) {
 /*
 :- mode string__first_char(in, out, in) is semidet.	% implied
 */
-:- pragma(c_code, string__first_char(Str::in, First::out, Rest::in), "
+:- pragma c_code(string__first_char(Str::in, First::out, Rest::in),
+		[will_not_call_mercury, thread_safe], "
 	First = Str[0];
 	SUCCESS_INDICATOR = (First != '\\0' && strcmp(Str + 1, Rest) == 0);
 ").
@@ -1866,7 +1923,8 @@ void sys_init_string_append_module(void) {
 /*
 :- mode string__first_char(in, in, out) is semidet.	% implied
 */
-:- pragma(c_code, string__first_char(Str::in, First::in, Rest::out), "{
+:- pragma c_code(string__first_char(Str::in, First::in, Rest::out),
+		[will_not_call_mercury, thread_safe], "{
 	Word tmp;
 	if (Str[0] != First || First == '\\0') {
 		SUCCESS_INDICATOR = FALSE;
@@ -1876,8 +1934,9 @@ void sys_init_string_append_module(void) {
 		** We need to make a copy to ensure that the pointer is
 		** word-aligned.
 		*/
-		incr_hp_atomic(tmp,
-			(strlen(Str) + sizeof(Word)) / sizeof(Word));
+		incr_hp_atomic_msg(tmp,
+			(strlen(Str) + sizeof(Word)) / sizeof(Word),
+			MR_PROC_LABEL, ""string:string/0"");
 		Rest = (char *) tmp;
 		strcpy(Rest, Str);
 		SUCCESS_INDICATOR = TRUE;
@@ -1887,7 +1946,8 @@ void sys_init_string_append_module(void) {
 /*
 :- mode string__first_char(in, out, out) is semidet.
 */
-:- pragma(c_code, string__first_char(Str::in, First::out, Rest::out), "{
+:- pragma c_code(string__first_char(Str::in, First::out, Rest::out),
+		[will_not_call_mercury, thread_safe], "{
 	Word tmp;
 	First = Str[0];
 	if (First == '\\0') {
@@ -1898,8 +1958,9 @@ void sys_init_string_append_module(void) {
 		** We need to make a copy to ensure that the pointer is
 		** word-aligned.
 		*/
-		incr_hp_atomic(tmp,
-			(strlen(Str) + sizeof(Word)) / sizeof(Word));
+		incr_hp_atomic_msg(tmp,
+			(strlen(Str) + sizeof(Word)) / sizeof(Word),
+			MR_PROC_LABEL, ""string:string/0"");
 		Rest = (char *) tmp;
 		strcpy(Rest, Str);
 		SUCCESS_INDICATOR = TRUE;
@@ -1909,16 +1970,164 @@ void sys_init_string_append_module(void) {
 /*
 :- mode string__first_char(out, in, in) is det.
 */
-:- pragma(c_code, string__first_char(Str::out, First::in, Rest::in), "{
+:- pragma c_code(string__first_char(Str::out, First::in, Rest::in),
+		[will_not_call_mercury, thread_safe], "{
 	size_t len = strlen(Rest) + 1;
 	Word tmp;
-	incr_hp_atomic(tmp, (len + sizeof(Word)) / sizeof(Word));
+	incr_hp_atomic_msg(tmp, (len + sizeof(Word)) / sizeof(Word),
+		MR_PROC_LABEL, ""string:string/0"");
 	Str = (char *) tmp;
 	Str[0] = First;
 	strcpy(Str + 1, Rest);
 }").
 
+%-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
+% Ralph Becket <rwab1@cl.cam.ac.uk> 27/04/99
+%       Functional forms added.
+
+:- interface.
+
+:- func string__append(string, string) = string.
+
+:- func string__char_to_string(char) = string.
+
+:- func string__int_to_string(int) = string.
+
+:- func string__int_to_base_string(int, int) = string.
+
+:- func string__float_to_string(float) = string.
+
+:- func string__replace_all(string, string, string) = string.
+
+:- func string__to_lower(string) = string.
+
+:- func string__to_upper(string) = string.
+
+:- func string__capitalize_first(string) = string.
+
+:- func string__uncapitalize_first(string) = string.
+
+:- func string__to_char_list(string) = list(char).
+
+:- func string__from_char_list(list(char)) = string.
+
+:- func string__from_rev_char_list(list(char)) = string.
+
+:- func string__pad_left(string, char, int) = string.
+
+:- func string__pad_right(string, char, int) = string.
+
+:- func string__duplicate_char(char, int) = string.
+
+:- func string__index_det(string, int) = char.
+
+:- func string__unsafe_index(string, int) = char.
+
+:- func string__foldl(func(char, T) = T, string, T) = T.
+
+:- func string__left(string, int) = string.
+
+:- func string__right(string, int) = string.
+
+:- func string__substring(string, int, int) = string.
+
+:- func string__unsafe_substring(string, int, int) = string.
+
+:- func string__append_list(list(string)) = string.
+
+:- func string__hash(string) = int.
+
+:- func string__format(string, list(string__poly_type)) = string.
+
+% ---------------------------------------------------------------------------- %
+% ---------------------------------------------------------------------------- %
+
+:- implementation.
+
+string__append(S1, S2) = S3 :-
+	string__append(S1, S2, S3).
+
+string__char_to_string(C) = S1 :-
+	string__char_to_string(C, S1).
+
+string__int_to_string(N) = S1 :-
+	string__int_to_string(N, S1).
+
+string__int_to_base_string(N1, N2) = S2 :-
+	string__int_to_base_string(N1, N2, S2).
+
+string__float_to_string(R) = S2 :-
+	string__float_to_string(R, S2).
+
+string__replace_all(S1, S2, S3) = S4 :-
+	string__replace_all(S1, S2, S3, S4).
+
+string__to_lower(S1) = S2 :-
+	string__to_lower(S1, S2).
+
+string__to_upper(S1) = S2 :-
+	string__to_upper(S1, S2).
+
+string__capitalize_first(S1) = S2 :-
+	string__capitalize_first(S1, S2).
+
+string__uncapitalize_first(S1) = S2 :-
+	string__uncapitalize_first(S1, S2).
+
+string__to_char_list(S) = Cs :-
+	string__to_char_list(S, Cs).
+
+string__from_char_list(Cs) = S :-
+	string__from_char_list(Cs, S).
+
+string__from_rev_char_list(Cs) = S :-
+	string__from_rev_char_list(Cs, S).
+
+string__pad_left(S1, C, N) = S2 :-
+	string__pad_left(S1, C, N, S2).
+
+string__pad_right(S1, C, N) = S2 :-
+	string__pad_right(S1, C, N, S2).
+
+string__duplicate_char(C, N) = S :-
+	string__duplicate_char(C, N, S).
+
+string__index_det(S, N) = C :-
+	string__index_det(S, N, C).
+
+string__unsafe_index(S, N) = C :-
+	string__unsafe_index(S, N, C).
+
+string__foldl(F, S, A) = B :-
+	P = ( pred(X::in, Y::in, Z::out) is det :- Z = F(X, Y) ),
+	string__foldl(P, S, A, B).
+
+string__left(S1, N) = S2 :-
+	string__left(S1, N, S2).
+
+string__right(S1, N) = S2 :-
+	string__right(S1, N, S2).
+
+string__substring(S1, N1, N2) = S2 :-
+	string__substring(S1, N1, N2, S2).
+
+string__unsafe_substring(S1, N1, N2) = S2 :-
+	string__unsafe_substring(S1, N1, N2, S2).
+
+string__append_list(S1s) = S2 :-
+	string__append_list(S1s, S2).
+
+string__hash(S) = N :-
+	string__hash(S, N).
+
+string__format(S1, PT) = S2 :-
+	string__format(S1, PT, S2).
+
+% ---------------------------------------------------------------------------- %
+% ---------------------------------------------------------------------------- %
+
 :- end_module string.
 
-%-----------------------------------------------------------------------------%
-%-----------------------------------------------------------------------------%
+% ---------------------------------------------------------------------------- %
+% ---------------------------------------------------------------------------- %
