@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1993-1997 The University of Melbourne.
+% Copyright (C) 1993-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -10,17 +10,23 @@
 
 % This file provides a type `term' used to represent Prolog terms,
 % and various predicates to manipulate terms and substitutions.
+% Terms are polymorphic so that terms representing different kinds of
+% thing can be made to be of different types so they don't get mixed up.
 
 %-----------------------------------------------------------------------------%
 
 :- module term.
 :- interface.
-:- import_module list, map.
+:- import_module list, map, std_util.
 
 %-----------------------------------------------------------------------------%
 
-:- type term		--->	term__functor(const, list(term), term__context)
-			;	term__variable(var).
+:- type term(T)		--->	term__functor(
+					const,
+					list(term(T)),
+					term__context
+				)
+			;	term__variable(var(T)).
 :- type const		--->	term__atom(string)
 			;	term__integer(int)
 			;	term__string(string)
@@ -29,19 +35,27 @@
 :- type term__context   --->    term__context(string, int).
 				% file name, line number.
 
-:- type var.
-:- type var_supply.
+:- type var(T).
+:- type var_supply(T).
+
+:- type generic
+	--->	generic.
+
+:- type term	==	term(generic).
+:- type var	==	var(generic).
 
 %-----------------------------------------------------------------------------%
 
 	% The following predicates can convert values of (almost)
 	% any type to the type `term' and back again.
 
-:- type term_to_type_result(T)
+:- type term_to_type_result(T, U)
 	--->	ok(T)
-	;	error(term_to_type_error).
+	;	error(term_to_type_error(U)).
 
-:- pred term__try_term_to_type(term, term_to_type_result(T)).
+:- type term_to_type_result(T) == term_to_type_result(T, generic).
+
+:- pred term__try_term_to_type(term(U), term_to_type_result(T, U)).
 :- mode term__try_term_to_type(in, out) is det.
 	% term__try_term_to_type(Term, Result):
 	% Try to convert the given term to a ground value of type T.
@@ -57,10 +71,10 @@
 	% ArgContexts specifies the path from the root of the term
 	% to the offending subterm.
 
-:- type term_to_type_error
-	--->	type_error(term, type_info, term__context,
+:- type term_to_type_error(T)
+	--->	type_error(term(T), type_info, term__context,
 			term_to_type_context)
-	;	mode_error(var, term_to_type_context).
+	;	mode_error(var(T), term_to_type_context).
 
 :- type term_to_type_context == list(term_to_type_arg_context).
 
@@ -71,75 +85,77 @@
 			term__context	% filename & line number
 		).
 
-:- pred term__term_to_type(term, T).
+:- pred term__term_to_type(term(U), T).
 :- mode term__term_to_type(in, out) is semidet.
 	% term_to_type(Term, Type) :- try_term_to_type(Term, ok(Type)).
 
-:- pred term__det_term_to_type(term, T).
+:- pred term__det_term_to_type(term(_), T).
 :- mode term__det_term_to_type(in, out) is det.
 	% like term_to_type, but calls error/1 rather than failing.
 
-:- pred term__type_to_term(T, term).
+:- pred term__type_to_term(T, term(_)).
 :- mode term__type_to_term(in, out) is det.
 	% converts a value to a term representation of that value
 
-:- pred term__univ_to_term(univ, term).
+:- pred term__univ_to_term(univ, term(_)).
 :- mode term__univ_to_term(in, out) is det.
 	% calls term__type_to_term on the value stored in the univ
 	% (as distinct from the univ itself).
 
 %-----------------------------------------------------------------------------%
 
-:- pred term__vars(term, list(var)).
+:- pred term__vars(term(T), list(var(T))).
 :- mode term__vars(in, out) is det.
 %	term__vars(Term, Vars)
 %		Vars is the list of variables contained in Term, in the order 
 %		obtained by traversing the term depth first, left-to-right.
 
-:- pred term__vars_2(term, list(var), list(var)).
+:- pred term__vars_2(term(T), list(var(T)), list(var(T))).
 :- mode term__vars_2(in, in, out) is det.
 %		As above, but with an accumulator.
 
-:- pred term__vars_list(list(term), list(var)).
+:- pred term__vars_list(list(term(T)), list(var(T))).
 :- mode term__vars_list(in, out) is det.
 %	term__vars_list(TermList, Vars)
 %		Vars is the list of variables contained in TermList, in the
 %		order obtained by traversing the list of terms depth-first,
 %		left-to-right.
 
-:- pred term__contains_var(term, var).
+:- pred term__contains_var(term(T), var(T)).
 :- mode term__contains_var(in, in) is semidet.
 :- mode term__contains_var(in, out) is nondet.
 %	term__contains_var(Term, Var)
 %		True if Term contains Var. (On backtracking returns all the 
 %		variables contained in Term.)
 
-:- pred term__contains_var_list(list(term), var).
+:- pred term__contains_var_list(list(term(T)), var(T)).
 :- mode term__contains_var_list(in, in) is semidet.
 :- mode term__contains_var_list(in, out) is nondet.
 %	term__contains_var_list(TermList, Var)
 %		True if TermList contains Var. (On backtracking returns all the 
 %		variables contained in Term.)
 
-:- type substitution == map(var, term).
+:- type substitution(T) == map(var(T), term(T)).
+:- type substitution	== substitution(generic).
 
-:- pred term__unify(term, term, substitution, substitution).
+:- pred term__unify(term(T), term(T), substitution(T), substitution(T)).
 :- mode term__unify(in, in, in, out) is semidet.
 %	term__unify(Term1, Term2, Bindings0, Bindings)
 %		unify (with occur check) two terms with respect to a set
 %	 	of bindings and possibly update the set of bindings
 
-:- pred term__substitute(term, var, term, term).
+:- pred term__substitute(term(T), var(T), term(T), term(T)).
 :- mode term__substitute(in, in, in, out) is det.
 %	term__substitute(Term0, Var, Replacement, Term) :
 %		replace all occurrences of Var in Term0 with Replacement,
 %		and return the result in Term.
 
-:- pred term__substitute_list(list(term), var, term, list(term)).
+:- pred term__substitute_list(list(term(T)), var(T), term(T), list(term(T))).
 :- mode term__substitute_list(in, in, in, out) is det.
 %		as above, except for a list of terms rather than a single term
 
-:- pred term__substitute_corresponding(list(var), list(term), term, term).
+:- pred term__substitute_corresponding(list(var(T)), list(term(T)),
+		term(T), term(T)).
 :- mode term__substitute_corresponding(in, in, in, out) is det.
 %       term__substitute_corresponding(Vars, Repls, Term0, Term).
 %		replace all occurrences of variables in Vars with
@@ -148,105 +164,96 @@
 %	        length as Repls, the behaviour is undefined and probably
 %		harmful.
 
-:- pred term__substitute_corresponding_list(list(var), list(term), list(term),
-						list(term)).
+:- pred term__substitute_corresponding_list(list(var(T)), list(term(T)),
+		list(term(T)), list(term(T))).
 :- mode term__substitute_corresponding_list(in, in, in, out) is det.
 %       term__substitute_corresponding_list(Vars, Repls, TermList0, TermList).
 %		As above, except applies to a list of terms rather than a
 %		single term.
 
-:- pred term__apply_rec_substitution(term, substitution, term).
+:- pred term__apply_rec_substitution(term(T), substitution(T), term(T)).
 :- mode term__apply_rec_substitution(in, in, out) is det.
 %	term__apply_rec_substitution(Term0, Substitution, Term) :
 %		recursively apply substitution to Term0 until
 %		no more substitions can be applied, and then
 %		return the result in Term.
 
-:- pred term__apply_rec_substitution_to_list(list(term), substitution,
-						list(term)).
+:- pred term__apply_rec_substitution_to_list(list(term(T)), substitution(T),
+						list(term(T))).
 :- mode term__apply_rec_substitution_to_list(in, in, out) is det.
 
-:- pred term__apply_substitution(term, substitution, term).
+:- pred term__apply_substitution(term(T), substitution(T), term(T)).
 :- mode term__apply_substitution(in, in, out) is det.
 %	term__apply_substitution(Term0, Substitution, Term) :
 %		apply substitution to Term0 and return the result in Term.
 
-:- pred term__apply_substitution_to_list(list(term), substitution,
-						list(term)).
+:- pred term__apply_substitution_to_list(list(term(T)), substitution(T),
+						list(term(T))).
 :- mode term__apply_substitution_to_list(in, in, out) is det.
 %	term__apply_substitution_to_list(TermList0, Substitution, TermList) :
 %		as above, except for a list of terms rather than a single term
 
 
-:- pred term__occurs(term, var, substitution).
+:- pred term__occurs(term(T), var(T), substitution(T)).
 :- mode term__occurs(in, in, in) is semidet.
 %	term__occurs(Term0, Var, Substitution) :
 %		true iff Var occurs in the term resulting after
 %		applying Substitution to Term0.
 
-:- pred term__occurs_list(list(term), var, substitution).
+:- pred term__occurs_list(list(term(T)), var(T), substitution(T)).
 :- mode term__occurs_list(in, in, in) is semidet.
 %		as above, except for a list of terms rather than a single term
 
-:- pred term__relabel_variable(term, var, var, term).
+:- pred term__relabel_variable(term(T), var(T), var(T), term(T)).
 :- mode term__relabel_variable(in, in, in, out) is det.
 %	term__relabel_variable(Term0, OldVar, NewVar, Term) :
 %		replace all occurences of OldVar in Term0 with
 %		NewVar and put the result in Term.
 
-:- pred term__relabel_variables(list(term), var, var, list(term)).
+:- pred term__relabel_variables(list(term(T)), var(T), var(T), list(term(T))).
 :- mode term__relabel_variables(in, in, in, out) is det.
 %	term__relabel_variables(Terms0, OldVar, NewVar, Terms) :
 %		same as term__relabel_variable but for a list of terms.
 
-:- pred term__apply_variable_renaming(term, map(var, var), term).
+:- pred term__apply_variable_renaming(term(T), map(var(T), var(T)), term(T)).
 :- mode term__apply_variable_renaming(in, in, out) is det.
 % 		same as term__relabel_variable, except relabels
 % 		multiple variables. If a variable is not in the
 % 		map, it is not replaced.
 
-:- pred term__apply_variable_renaming_to_list(list(term), map(var, var),
-							 list(term)).
+:- pred term__apply_variable_renaming_to_list(list(term(T)),
+		map(var(T), var(T)), list(term(T))).
 :- mode term__apply_variable_renaming_to_list(in, in, out) is det.
 %		applies term__apply_variable_renaming to a list of terms.
 		
 
-:- pred term__is_ground(term, substitution).
+:- pred term__is_ground(term(T), substitution(T)).
 :- mode term__is_ground(in, in) is semidet.
 %	term__is_ground(Term, Bindings) is true iff no variables contained
 %		in Term are non-ground in Bindings.
 
-:- pred term__is_ground(term).
+:- pred term__is_ground(term(T)).
 :- mode term__is_ground(in) is semidet.
 %	term__is_ground(Term) is true iff Term contains no variables.
 
-:- pred term__compare(comparison_result, term, term, substitution).
-:- mode term__compare(out, in, in, in) is semidet.
-%	term__compare(Comparison, Term1, Term2, Bindings) is true iff
-%		there is a binding of Comparison to <, =, or > such
-%		that the binding holds for the two ground terms Term1
-%		and Term2 with respect to the bindings in Bindings.
-%		Fails if Term1 or Term2 is not ground (with respect to
-%		the bindings in Bindings).
-
 %-----------------------------------------------------------------------------%
 
 	% To manage a supply of variables, use the following 2 predicates.
 	% (We might want to give these a unique mode later.)
 
-:- pred term__init_var_supply(var_supply).
+:- pred term__init_var_supply(var_supply(T)).
 :- mode term__init_var_supply(out) is det.
 :- mode term__init_var_supply(in) is semidet. % implied
 %	term__init_var_supply(VarSupply) :
 %		returns a fresh var_supply for producing fresh variables.
 
-:- pred term__create_var(var_supply, var, var_supply).
+:- pred term__create_var(var_supply(T), var(T), var_supply(T)).
 :- mode term__create_var(in, out, out) is det.
 %	term__create_var(VarSupply0, Variable, VarSupply) :
 %		create a fresh variable (var) and return the
 %		updated var_supply.
 
-:- pred term__var_to_int(var, int).
+:- pred term__var_to_int(var(T), int).
 :- mode term__var_to_int(in, out) is det.
 %		Convert a variable to an int.
 %		Different variables map to different ints.
@@ -277,16 +284,63 @@
 	% of vars.  Abort (call error/1) if the list contains
 	% any non-variables.
 
-:- pred term__term_list_to_var_list(list(term), list(var)).
+:- pred term__term_list_to_var_list(list(term(T)), list(var(T))).
 :- mode term__term_list_to_var_list(in, out) is det.
 
 	% Convert a list of terms which are all vars into a list
 	% of vars (or vice versa).
 
-:- pred term__var_list_to_term_list(list(var), list(term)).
+:- pred term__var_list_to_term_list(list(var(T)), list(term(T))).
 :- mode term__var_list_to_term_list(in, out) is det.
 :- mode term__var_list_to_term_list(out, in) is semidet.
 
+%-----------------------------------------------------------------------------%
+	
+	% term__generic_term(Term) is true iff `Term' is a term of type
+	% `term' ie `term(generic)'.
+	% It is useful because in some instances it doesn't matter what
+	% the type of a term is, and passing it to this predicate will
+	% ground the type avoiding unbound type variable warnings.
+:- pred term__generic_term(term).
+:- mode term__generic_term(in) is det.
+
+	% Coerce a term of type `T' into a term of type `U'.
+:- pred term__coerce(term(T), term(U)).
+:- mode term__coerce(in, out) is det.
+
+	% Coerce a var of type `T' into a var of type `U'.
+:- pred term__coerce_var(var(T), var(U)).
+:- mode term__coerce_var(in, out) is det.
+
+	% Coerce a var_supply of type `T' into a var_supply of type `U'.
+:- pred term__coerce_var_supply(var_supply(T), var_supply(U)).
+:- mode term__coerce_var_supply(in, out) is det.
+
+%-----------------------------------------------------------------------------%
+
+:- implementation.
+
+% Everything below here is not intended to be part of the public interface,
+% and will not be included in the Mercury library reference manual.
+
+%-----------------------------------------------------------------------------%
+
+:- interface.
+
+	% This predidicate is being phased out, because of the problem
+	% mentioned in the "BEWARE:" below.
+:- pragma obsolete(term__compare/4).
+:- pred term__compare(comparison_result, term(T), term(T), substitution(T)).
+:- mode term__compare(out, in, in, in) is semidet.
+%	term__compare(Comparison, Term1, Term2, Bindings) is true iff
+%		there is a binding of Comparison to <, =, or > such
+%		that the binding holds for the two ground terms Term1
+%		and Term2 with respect to the bindings in Bindings.
+%		Fails if Term1 or Term2 is not ground (with respect to
+%		the bindings in Bindings).
+%		BEWARE: the comparison does not ignore the term__contexts
+%		in the terms.
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
@@ -295,8 +349,10 @@
 
 %-----------------------------------------------------------------------------%
 
-:- type var_supply	==	int.
-:- type var		==	int.
+:- type var_supply(T)
+	--->	var_supply(int).
+:- type var(T)
+	--->	var(int).
 
 %-----------------------------------------------------------------------------%
 
@@ -315,15 +371,15 @@ term__try_term_to_type(Term, Result) :-
 		Result = error(Error)
 	).
 
-:- pred term__try_term_to_univ(term::in, type_info::in,
-		term_to_type_result(univ)::out) is det.
+:- pred term__try_term_to_univ(term(T)::in, type_info::in,
+		term_to_type_result(univ, T)::out) is det.
 
 term__try_term_to_univ(Term, Type, Result) :-
 	term__try_term_to_univ_2(Term, Type, [], Result).
 	
-:- pred term__try_term_to_univ_2(term::in, type_info::in,
+:- pred term__try_term_to_univ_2(term(T)::in, type_info::in,
 		term_to_type_context::in,
-		term_to_type_result(univ)::out) is det.
+		term_to_type_result(univ, T)::out) is det.
 
 term__try_term_to_univ_2(term__variable(Var), _Type, Context,
 		error(mode_error(Var, Context))).
@@ -364,23 +420,24 @@ term__try_term_to_univ_2(Term, Type, Context, Result) :-
 
 :- pred term__term_to_univ_special_case(string::in, string::in, 
 		list(type_info)::in, 
-		term::in(bound(term__functor(ground, ground, ground))),
+		term(T)::in(bound(term__functor(ground, ground, ground))),
 		type_info::in, term_to_type_context::in,
-		term_to_type_result(univ)::out) is semidet.
-term__term_to_univ_special_case("mercury_builtin", "character", [],
+		term_to_type_result(univ, T)::out) is semidet.
+
+term__term_to_univ_special_case("builtin", "character", [],
 		Term, _, _, ok(Univ)) :-
 	Term = term__functor(term__atom(FunctorName), [], _),
 	string__first_char(FunctorName, Char, ""),
 	type_to_univ(Char, Univ).
-term__term_to_univ_special_case("mercury_builtin", "int", [],
+term__term_to_univ_special_case("builtin", "int", [],
 		Term, _, _, ok(Univ)) :-
 	Term = term__functor(term__integer(Int), [], _),
 	type_to_univ(Int, Univ).
-term__term_to_univ_special_case("mercury_builtin", "string", [],
+term__term_to_univ_special_case("builtin", "string", [],
 		Term, _, _, ok(Univ)) :-
 	Term = term__functor(term__string(String), [], _),
 	type_to_univ(String, Univ).
-term__term_to_univ_special_case("mercury_builtin", "float", [],
+term__term_to_univ_special_case("builtin", "float", [],
 		Term, _, _, ok(Univ)) :-
 	Term = term__functor(term__float(Float), [], _),
 	type_to_univ(Float, Univ).
@@ -396,54 +453,56 @@ term__term_to_univ_special_case("array", "array", [ElemType], Term, _Type,
 	% convert the term representing the list of elements back to a list,
 	% and then (if successful) we just call the array/1 function.
 	%
-	ListTypeCtor = type_ctor(type_of([0])),
-	ListType = det_make_type(ListTypeCtor, [ElemType]),
+	has_type(Elem, ElemType),
+	ListType = type_of([Elem]),
 	ArgContext = arg_context(term__atom("array"), 1, TermContext),
 	NewContext = [ArgContext | PrevContext],
 	term__try_term_to_univ_2(ArgList, ListType, NewContext, ArgResult),
 	(
 		ArgResult = ok(ListUniv),
-/***************
-% XXX existential types not yet implemented...
-		% :- some [T] pred has_type(T::unused, type_info::in) is det.
-		has_type(List, ListType),
+		has_type(Elem2, ElemType),
+		same_type(List, [Elem2]),
 		det_univ_to_type(ListUniv, List),
 		Array = array(List),
 		Result = ok(univ(Array))
-****************/
-		% since we don't have existential types, we have to use
-		% some unsafe casts...
-		require_equal(univ_type(ListUniv), ListType),
-		list_of_any(List),   % explicit type qualification
-				     % to avoid unbound type variables
-		List = unsafe_cast(univ_value_as_type_any(ListUniv)),
-		Array = array(List),
-		ArrayTypeCtor = type_ctor(type_of(Array)),
-		ArrayType = det_make_type(ArrayTypeCtor, [ElemType]),
-		Result = ok(unsafe_any_to_univ(ArrayType, unsafe_cast(Array)))
 	;
 		ArgResult = error(Error),
 		Result = error(Error)
 	).
-term__term_to_univ_special_case("mercury_builtin", "c_pointer", _, _, _, 
+term__term_to_univ_special_case("builtin", "c_pointer", _, _, _, 
 		_, _) :-
 	fail.
-term__term_to_univ_special_case("std_util", "univ", _, _, _, _, _) :-
+term__term_to_univ_special_case("std_util", "univ", [], Term, _, _, Result) :-
 	% Implementing this properly would require keeping a
 	% global table mapping from type names to type_infos
 	% for all of the types in the program...
-	% so for the moment, we don't allow it.
-	fail.
+	% so for the moment, we only allow it for basic types.
+	Term = term__functor(term__atom("univ"), [Arg], _),
+	Arg = term__functor(term__atom(":"), [Value, Type], _),
+	(
+		Type = term__functor(term__atom("int"), [], _),
+		Value = term__functor(term__integer(Int), [], _),
+		Univ = univ(Int)
+	;
+		Type = term__functor(term__atom("string"), [], _),
+		Value = term__functor(term__string(String), [], _),
+		Univ = univ(String)
+	;
+		Type = term__functor(term__atom("float"), [], _),
+		Value = term__functor(term__float(Float), [], _),
+		Univ = univ(Float)
+	),
+	% The result is a `univ', but it is also wrapped in a `univ'
+	% like all the other results returned from this procedure.
+	Result = ok(univ(Univ)).
+
 term__term_to_univ_special_case("std_util", "type_info", _, _, _, _, _) :-
 	% ditto
 	fail.
 
-:- pred same_type(T::unused, T::unused) is det.
-same_type(_, _).
-
-:- pred term__term_list_to_univ_list(list(term)::in, list(type_info)::in,
+:- pred term__term_list_to_univ_list(list(term(T))::in, list(type_info)::in,
 		term__const::in, int::in, term_to_type_context::in,
-		term__context::in, term_to_type_result(list(univ))::out)
+		term__context::in, term_to_type_result(list(univ), T)::out)
 		is semidet.
 term__term_list_to_univ_list([], [], _, _, _, _, ok([])).
 term__term_list_to_univ_list([ArgTerm|ArgTerms], [Type|Types],
@@ -504,58 +563,6 @@ term__det_term_to_type(Term, X) :-
 
 %-----------------------------------------------------------------------------%
 
-:- pred det_univ_to_type(univ::in, T::out) is det.
-det_univ_to_type(Univ, Value) :-
-	( univ_to_type(Univ, Value1) ->
-		Value = Value1
-	;
-		error("det_univ_to_type: univ_to_type failed")
-	).
-
-%-----------------------------------------------------------------------------%
-
-/**********
-XXX existential types not yet implemented
-:- some [T] pred has_type(T::unused, type_info::in) is det.
-:- pragma c_code(has_type(Arg::unused, TypeInfo::in), will_not_call_mercury,
-	"TypeInfo_for_T = TypeInfo;"
-).
-**********/
-
-:- func unsafe_cast(T1::in) = (T2::out) is det.
-:- pragma c_code(unsafe_cast(VarIn::in) = (VarOut::out), will_not_call_mercury,
-	"VarOut = VarIn;").
-
-:- type any == c_pointer.
-
-:- pred array_of_any(array(any)::unused) is det.
-array_of_any(_).
-
-:- pred list_of_any(list(any)::unused) is det.
-list_of_any(_).
-
-:- func univ_value_as_type_any(univ) = any.
-:- pragma c_code(univ_value_as_type_any(Univ::in) = (Val::out),
-		will_not_call_mercury,
-"
-	Val = field(mktag(0), Univ, UNIV_OFFSET_FOR_DATA);
-").
-
-:- func unsafe_any_to_univ(type_info, any) = univ.
-	% Forward mode - convert from type to univ.
-	% Allocate heap space, set the first field to contain the address
-	% of the type_info for this type, and then store the input argument
-	% in the second field.
-:- pragma c_code(unsafe_any_to_univ(TypeInfo::in, Value::in) = (Univ::out),
-		will_not_call_mercury,
-"
-	incr_hp(Univ, 2);
-	field(mktag(0), Univ, UNIV_OFFSET_FOR_TYPEINFO) = (Word) TypeInfo;
-	field(mktag(0), Univ, UNIV_OFFSET_FOR_DATA) = (Word) Value;
-").
-
-%-----------------------------------------------------------------------------%
-
 term__type_to_term(Val, Term) :- type_to_univ(Val, Univ),
 	term__univ_to_term(Univ, Term).
 
@@ -589,19 +596,19 @@ term__univ_to_term(Univ, Term) :-
 
 :- pred term__univ_to_term_special_case(string::in, string::in, 
 		list(type_info)::in, univ::in, term__context::in,
-		term::out) is semidet.
+		term(T)::out) is semidet.
 
-term__univ_to_term_special_case("mercury_builtin", "int", [], Univ, Context,
+term__univ_to_term_special_case("builtin", "int", [], Univ, Context,
 		term__functor(term__integer(Int), [], Context)) :-
 	det_univ_to_type(Univ, Int).
-term__univ_to_term_special_case("mercury_builtin", "float", [], Univ, Context,
+term__univ_to_term_special_case("builtin", "float", [], Univ, Context,
 		term__functor(term__float(Float), [], Context)) :-
 	det_univ_to_type(Univ, Float).
-term__univ_to_term_special_case("mercury_builtin", "character", [], Univ, 
+term__univ_to_term_special_case("builtin", "character", [], Univ, 
 		Context, term__functor(term__atom(CharName), [], Context)) :-
 	det_univ_to_type(Univ, Character),
 	string__char_to_string(Character, CharName).
-term__univ_to_term_special_case("mercury_builtin", "string", [], Univ, Context,
+term__univ_to_term_special_case("builtin", "string", [], Univ, Context,
 		term__functor(term__string(String), [], Context)) :-
 	det_univ_to_type(Univ, String).
 term__univ_to_term_special_case("std_util", "type_info", [], Univ, Context,
@@ -615,39 +622,24 @@ term__univ_to_term_special_case("std_util", "univ", [], Univ, Context, Term) :-
 			[term__functor(term__atom(":"),	 % TYPE_QUAL_OP
 				[ValueTerm, TypeTerm],
 				Context)], Context),
-/****
-XXX existential types not implemented
-	det_univ_to_type(univ_value(Univ), UnivValue),
-****/
-	% instead, we use some unsafe casts...
-	require_equal(univ_type(Univ), type_of(UnivValue)),
-	UnivValue = unsafe_cast(univ_value_as_type_any(Univ)),
-
-	type_info_to_term(Context, univ_type(UnivValue), TypeTerm),
-	term__univ_to_term(UnivValue, ValueTerm).
+	type_info_to_term(Context, univ_type(Univ), TypeTerm),
+	UnivValue = univ_value(Univ),
+	term__type_to_term(UnivValue, ValueTerm).
 
 term__univ_to_term_special_case("array", "array", [ElemType], Univ, Context, 
 		Term) :-
 	Term = term__functor(term__atom("array"), [ArgsTerm], Context),
-	ListTypeCtor = type_ctor(type_of([0])),
-	ListType = det_make_type(ListTypeCtor, [ElemType]),
-/***
-XXX existential types not yet implemented
-	has_type(List, ListType),
-	det__univ_to_type(Univ, Array),	
+	has_type(Elem, ElemType),
+	same_type(List, [Elem]),
+	det_univ_to_type(Univ, Array),	
 	array__to_list(Array, List),
 	term__type_to_term(List, ArgsTerm).
-***/
-	% instead, we need to use some unsafe casts...
-	array_of_any(Array), % explicit type qualification
-			     % to avoid unbound type variables
-	Array = unsafe_cast(univ_value_as_type_any(Univ)),	
-	array__to_list(Array, List),
-	ListUniv = unsafe_any_to_univ(ListType, unsafe_cast(List)),
-	term__univ_to_term(ListUniv, ArgsTerm).
+
+:- pred same_type(T::unused, T::unused) is det.
+same_type(_, _).
 
 :- pred term__univ_list_to_term_list(list(univ)::in,
-				list(term)::out) is det.
+				list(term(T))::out) is det.
 
 term__univ_list_to_term_list([], []).
 term__univ_list_to_term_list([Value|Values], [Term|Terms]) :-
@@ -655,17 +647,18 @@ term__univ_list_to_term_list([Value|Values], [Term|Terms]) :-
 	term__univ_list_to_term_list(Values, Terms).
 
 % given a type_info, return a term that represents the name of that type.
-:- pred type_info_to_term(term__context::in, type_info::in, term::out) is det.
+:- pred type_info_to_term(term__context::in, type_info::in,
+		term(T)::out) is det.
 type_info_to_term(Context, TypeInfo, Term) :-
 	type_ctor_and_args(TypeInfo, TypeCtor, ArgTypes),
 	TypeName = type_ctor_name(TypeCtor),
 	ModuleName = type_ctor_name(TypeCtor),
 	list__map(type_info_to_term(Context), ArgTypes, ArgTerms),
 
-	( ModuleName = "mercury_builtin" ->
+	( ModuleName = "builtin" ->
 		Term = term__functor(term__atom(TypeName), ArgTerms, Context)
 	;
-		Term = term__functor(term__atom(":"), % TYPE_QUAL_OP
+		Term = term__functor(term__atom(":"),
 			[term__functor(term__atom(ModuleName), [], Context),
 			 term__functor(term__atom(TypeName), 
 			 	ArgTerms, Context)], Context)
@@ -696,7 +689,7 @@ term__vars_2(term__variable(V), Vs, [V|Vs]).
 term__vars_2(term__functor(_,Args,_), Vs0, Vs) :-
 	term__vars_2_list(Args, Vs0, Vs).
 
-:- pred term__vars_2_list(list(term), list(var), list(var)).
+:- pred term__vars_2_list(list(term(T)), list(var(T)), list(var(T))).
 :- mode term__vars_2_list(in, in, out) is det.
 
 term__vars_2_list([], Vs, Vs).
@@ -724,7 +717,7 @@ term__contains_var_list([_|Ts], V) :-
 	%
 	% CURRENTLY NOT USED.
 
-:- pred term__contains_functor(term, const, list(term)).
+:- pred term__contains_functor(term(T), const, list(term(T))).
 % :- mode term__contains_functor(in, in, in) is semidet.
 :- mode term__contains_functor(in, out, out) is nondet.
 
@@ -740,7 +733,7 @@ term__contains_functor(term__functor(_, Args, _), SubFunctor, SubArgs) :-
 	%
 	% CURRENTLY NOT USED.
 
-:- pred term__subterm(term, term).
+:- pred term__subterm(term(T), term(T)).
 :- mode term__subterm(in, in) is semidet.
 :- mode term__subterm(in, out) is multidet.
 
@@ -773,8 +766,6 @@ term__context_init(File, LineNumber, term__context(File, LineNumber)).
 	% Unify two terms (with occurs check), updating the bindings of
 	% the variables in the terms.  
 
-:- term__unify(X, Y, _, _) when X and Y.		% NU-Prolog indexing
-
 term__unify(term__variable(X), term__variable(Y), Bindings0, Bindings) :-
 	( %%% if some [BindingOfX]
 		map__search(Bindings0, X, BindingOfX)
@@ -848,7 +839,8 @@ term__unify(term__functor(F, As, C), term__variable(X), Bindings0, Bindings) :-
 term__unify(term__functor(F, AsX, _), term__functor(F, AsY, _)) -->
 	term__unify_list(AsX, AsY).
 
-:- pred term__unify_list(list(term), list(term), substitution, substitution).
+:- pred term__unify_list(list(term(T)), list(term(T)),
+		substitution(T), substitution(T)).
 :- mode term__unify_list(in, in, in, out) is semidet.
 
 term__unify_list([], []) --> [].
@@ -924,8 +916,8 @@ term__substitute_corresponding_list(Ss, Rs, TermList0, TermList) :-
 		)
 	).
 
-:- pred term__substitute_corresponding_2(list(var), list(term),
-					substitution, substitution).
+:- pred term__substitute_corresponding_2(list(var(T)), list(term(T)),
+					substitution(T), substitution(T)).
 :- mode term__substitute_corresponding_2(in, in, in, out) is semidet.
 
 term__substitute_corresponding_2([], [], Subst, Subst).
@@ -979,16 +971,16 @@ term__apply_substitution_to_list([Term0 | Terms0], Substitution,
 %-----------------------------------------------------------------------------%
 
 	% create a new supply of variables
-term__init_var_supply(0).
+term__init_var_supply(var_supply(0)).
 
 	% We number variables using sequential numbers,
 
-term__create_var(VarSupply0, VarSupply, VarSupply) :-
-	VarSupply is VarSupply0 + 1.
+term__create_var(var_supply(V0), var(V), var_supply(V)) :-
+	V is V0 + 1.
 
 %-----------------------------------------------------------------------------%
 
-term__var_to_int(Var, Var).
+term__var_to_int(var(Var), Var).
 
 %-----------------------------------------------------------------------------%
 
@@ -1032,8 +1024,6 @@ term__apply_variable_renaming_to_list([Term0|Terms0], Renaming, [Term|Terms]) :-
 
 %-----------------------------------------------------------------------------%
 
-:- term__term_list_to_var_list(Terms, Vars) when Terms or Vars. % Indexing
-
 term__term_list_to_var_list(Terms, Vars) :-
 	( term__var_list_to_term_list(Vars0, Terms) ->
 		Vars = Vars0
@@ -1041,8 +1031,6 @@ term__term_list_to_var_list(Terms, Vars) :-
 		error("term__term_list_to_var_list")
 	).
 
-:- term__var_list_to_term_list(Terms, Vars) when Terms or Vars. % Indexing
-
 term__var_list_to_term_list([], []).
 term__var_list_to_term_list([Var | Vars], [term__variable(Var) | Terms]) :-
 	term__var_list_to_term_list(Vars, Terms).
@@ -1055,7 +1043,7 @@ term__is_ground(term__variable(V), Bindings) :-
 term__is_ground(term__functor(_, Args, _), Bindings) :-
 	term__is_ground_2(Args, Bindings).
 
-:- pred term__is_ground_2(list(term), substitution).
+:- pred term__is_ground_2(list(term(T)), substitution(T)).
 :- mode term__is_ground_2(in, in) is semidet.
 
 term__is_ground_2([], _Bindings).
@@ -1068,7 +1056,7 @@ term__is_ground_2([Term|Terms], Bindings) :-
 term__is_ground(term__functor(_, Args, _)) :-
 	term__is_ground_2(Args).
 
-:- pred term__is_ground_2(list(term)).
+:- pred term__is_ground_2(list(term(T))).
 :- mode term__is_ground_2(in) is semidet.
 
 term__is_ground_2([]).
@@ -1086,3 +1074,179 @@ term__compare(Cmp, Term1, Term2, Bindings) :-
 	compare(Cmp, TermA, TermB).
 
 %-----------------------------------------------------------------------------%
+
+term__generic_term(_).
+
+%-----------------------------------------------------------------------------%
+
+term__coerce(term__variable(var(V)), term__variable(var(V))).
+term__coerce(term__functor(Cons, Args0, Ctxt),
+		term__functor(Cons, Args, Ctxt)) :-
+	list__map(term__coerce, Args0, Args).
+
+term__coerce_var(var(V), var(V)).
+
+term__coerce_var_supply(var_supply(Supply), var_supply(Supply)).
+
+% ---------------------------------------------------------------------------- %
+% ---------------------------------------------------------------------------- %
+% Ralph Becket <rwab1@cl.cam.ac.uk> 30/04/99
+% 	Function forms added.
+
+:- interface.
+
+:- func term__context_init = term__context.
+
+:- func term__init_var_supply = var_supply(T).
+
+:- func term__try_term_to_type(term(U)) = term_to_type_result(T, U).
+
+:- func term__det_term_to_type(term(_)) = T.
+
+:- func term__type_to_term(T) = term(_).
+
+:- func term__univ_to_term(univ) = term(_).
+
+:- func term__vars(term(T)) = list(var(T)).
+
+:- func term__vars_2(term(T), list(var(T))) = list(var(T)).
+
+:- func term__vars_list(list(term(T))) = list(var(T)).
+
+:- func term__substitute(term(T), var(T), term(T)) = term(T).
+
+:- func term__substitute_list(list(term(T)), var(T), term(T)) = list(term(T)).
+
+:- func term__substitute_corresponding(list(var(T)), list(term(T)), term(T)) = term(T).
+
+:- func term__substitute_corresponding_list(list(var(T)), list(term(T)), list(term(T))) = list(term(T)).
+
+:- func term__apply_rec_substitution(term(T), substitution(T)) = term(T).
+
+:- func term__apply_rec_substitution_to_list(list(term(T)), substitution(T)) = list(term(T)).
+
+:- func term__apply_substitution(term(T), substitution(T)) = term(T).
+
+:- func term__apply_substitution_to_list(list(term(T)), substitution(T)) = list(term(T)).
+
+:- func term__relabel_variable(term(T), var(T), var(T)) = term(T).
+
+:- func term__relabel_variables(list(term(T)), var(T), var(T)) = list(term(T)).
+
+:- func term__apply_variable_renaming(term(T), map(var(T), var(T))) = term(T).
+
+:- func term__apply_variable_renaming_to_list(list(term(T)), map(var(T), var(T))) = list(term(T)).
+
+:- func term__var_to_int(var(T)) = int.
+
+:- func term__context_line(term__context) = int.
+
+:- func term__context_file(term__context) = string.
+
+:- func term__context_init(string, int) = term__context.
+
+:- func term__term_list_to_var_list(list(term(T))) = list(var(T)).
+
+:- func term__var_list_to_term_list(list(var(T))) = list(term(T)).
+
+:- func term__coerce(term(T)) = term(U).
+
+:- func term__coerce_var(var(T)) = var(U).
+
+:- func term__coerce_var_supply(var_supply(T)) = var_supply(U).
+
+% ---------------------------------------------------------------------------- %
+% ---------------------------------------------------------------------------- %
+
+:- implementation.
+
+term__context_init = C :-
+	term__context_init(C).
+
+term__init_var_supply = VS :-
+	term__init_var_supply(VS).
+
+term__try_term_to_type(T) = TTTR :-
+	term__try_term_to_type(T, TTTR).
+
+term__det_term_to_type(T1) = T2 :-
+	term__det_term_to_type(T1, T2).
+
+term__type_to_term(T1) = T2 :-
+	term__type_to_term(T1, T2).
+
+term__univ_to_term(U) = T :-
+	term__univ_to_term(U, T).
+
+term__vars(T) = Vs :-
+	term__vars(T, Vs).
+
+term__vars_2(T, Vs1) = Vs2 :-
+	term__vars_2(T, Vs1, Vs2).
+
+term__vars_list(Ts) = Vs :-
+	term__vars_list(Ts, Vs).
+
+term__substitute(T1, V, T2) = T3 :-
+	term__substitute(T1, V, T2, T3).
+
+term__substitute_list(Ts1, V, T) = Ts2 :-
+	term__substitute_list(Ts1, V, T, Ts2).
+
+term__substitute_corresponding(Vs, T1s, T) = T2 :-
+	term__substitute_corresponding(Vs, T1s, T, T2).
+
+term__substitute_corresponding_list(Vs, Ts1, Ts2) = Ts3 :-
+	term__substitute_corresponding_list(Vs, Ts1, Ts2, Ts3).
+
+term__apply_rec_substitution(T1, S) = T2 :-
+	term__apply_rec_substitution(T1, S, T2).
+
+term__apply_rec_substitution_to_list(Ts1, S) = Ts2 :-
+	term__apply_rec_substitution_to_list(Ts1, S, Ts2).
+
+term__apply_substitution(T1, S) = T2 :-
+	term__apply_substitution(T1, S, T2).
+
+term__apply_substitution_to_list(Ts1, S) = Ts2 :-
+	term__apply_substitution_to_list(Ts1, S, Ts2).
+
+term__relabel_variable(T1, V1, V2) = T2 :-
+	term__relabel_variable(T1, V1, V2, T2).
+
+term__relabel_variables(Ts1, V1, V2) = Ts2 :-
+	term__relabel_variables(Ts1, V1, V2, Ts2).
+
+term__apply_variable_renaming(T1, M) = T2 :-
+	term__apply_variable_renaming(T1, M, T2).
+
+term__apply_variable_renaming_to_list(Ts1, M) = Ts2 :-
+	term__apply_variable_renaming_to_list(Ts1, M, Ts2).
+
+term__var_to_int(V) = N :-
+	term__var_to_int(V, N).
+
+term__context_line(C) = N :-
+	term__context_line(C, N).
+
+term__context_file(C) = S :-
+	term__context_file(C, S).
+
+term__context_init(S, N) = C :-
+	term__context_init(S, N, C).
+
+term__term_list_to_var_list(Ts) = Vs :-
+	term__term_list_to_var_list(Ts, Vs).
+
+term__var_list_to_term_list(Vs) = Ts :-
+	term__var_list_to_term_list(Vs, Ts).
+
+term__coerce(T1) = T2 :-
+	term__coerce(T1, T2).
+
+term__coerce_var(V1) = V2 :-
+	term__coerce_var(V1, V2).
+
+term__coerce_var_supply(VS1) = VS2 :-
+	term__coerce_var_supply(VS1, VS2).
+
