@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -16,7 +16,7 @@
 
 :- module term_io.
 :- interface.
-:- import_module io, varset, term.
+:- import_module char, io, varset, term.
 
 % External interface: exported predicates
 
@@ -35,8 +35,11 @@
 %		Does not modify the io__state.
 *****/
 
-:- type read_term ---> eof ; error(string, int) ; term(varset, term).
-:- pred term_io__read_term(read_term, io__state, io__state).
+:- type read_term(T) ---> eof ; error(string, int) ; term(varset(T), term(T)).
+
+:- type read_term	== read_term(generic).
+
+:- pred term_io__read_term(read_term(T), io__state, io__state).
 :- mode term_io__read_term(out, di, uo) is det.
 
 %	term_io__read_term(Result, IO0, IO1).
@@ -45,19 +48,20 @@
 %		representation. Binds Result to either `eof',
 %		`term(VarSet, Term)', or `error(Message, LineNumber)'.
 
-:- pred term_io__write_term(varset, term, io__state, io__state).
+:- pred term_io__write_term(varset(T), term(T), io__state, io__state).
 :- mode term_io__write_term(in, in, di, uo) is det.
 %		Writes a term to standard output.
 
-:- pred term_io__write_term_nl(varset, term, io__state, io__state).
+:- pred term_io__write_term_nl(varset(T), term(T), io__state, io__state).
 :- mode term_io__write_term_nl(in, in, di, uo) is det.
 %		As above, except it appends a period and new-line.
 
 :- pred term_io__write_constant(const, io__state, io__state).
 :- mode term_io__write_constant(in, di, uo) is det.
-%		Writes a constant (integer, float, or atom) to stdout.
+%		Writes a constant (integer, float, string, or atom)
+%		to stdout.
 
-:- pred term_io__write_variable(var, varset, io__state, io__state).
+:- pred term_io__write_variable(var(T), varset(T), io__state, io__state).
 :- mode term_io__write_variable(in, in, di, uo) is det.
 %		Writes a variable to stdout.
 
@@ -94,6 +98,24 @@
 :- mode term_io__quote_single_char(in, di, uo) is det.
 
 %-----------------------------------------------------------------------------%
+:- implementation.
+
+% Everything below here is not intended to be part of the public interface,
+% and will not be included in the Mercury library reference manual.
+
+%-----------------------------------------------------------------------------%
+:- interface.
+
+	% for use by io.m.
+
+:- type adjacent_to_graphic_token
+	--->	maybe_adjacent_to_graphic_token
+	;	not_adjacent_to_graphic_token.
+
+:- pred term_io__quote_atom(string, adjacent_to_graphic_token,
+		io__state, io__state).
+:- mode term_io__quote_atom(in, in, di, uo) is det.
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
@@ -108,13 +130,29 @@ term_io__read_term(Result) -->
 %-----------------------------------------------------------------------------%
 
 	% write a variable to standard output.
-	% use the variable names specified by varset and write _N
-	% for all unnamed variables with N starting at 0.
+	%
+	% There are two ways we could choose to write unnamed variables
+	% (ie `_'):
+	%	Convert the variable to an integer representation and write
+	%	`_N' where N is that integer representation. This has the
+	%	advantage that such variables get printed in a cannonical
+	%	way, so rearranging terms containing such variables will
+	%	not effect the way they are numbered (this includes breaking
+	%	up a term and printing the pieces separately).
+	% or
+	%	Number the unnamed variables from 0 and write `_N' where
+	%	N is the number in the sequence of such variables. This has
+	%	the advantage that such variables can be visually scanned
+	%	rather more easily (for example in error messages).
+	%
+	% An ideal solution would be to provide both, and a flag to choose
+	% between the two. At the moment we provide only the first, though
+	% the infrastructure for the second is present in the code.
 
 term_io__write_variable(Variable, VarSet) -->
 	term_io__write_variable_2(Variable, VarSet, 0, _, _).
 
-:- pred term_io__write_variable_2(var, varset, int, varset, int,
+:- pred term_io__write_variable_2(var(T), varset(T), int, varset(T), int,
 				io__state, io__state).
 :- mode term_io__write_variable_2(in, in, in, out, out, di, uo) is det.
 
@@ -132,7 +170,8 @@ term_io__write_variable_2(Id, VarSet0, N0, VarSet, N) -->
 	;
 		% XXX problems with name clashes
 
-		{ string__int_to_string(N0, Num) },
+		{ term__var_to_int(Id, VarNum) },
+		{ string__int_to_string(VarNum, Num) },
 		{ string__append("_", Num, VarName) },
 		{ varset__name_var(VarSet0, Id, VarName, VarSet) },
 		{ N is N0 + 1 },
@@ -148,7 +187,7 @@ term_io__write_variable_2(Id, VarSet0, N0, VarSet, N) -->
 term_io__write_term(VarSet, Term) -->
 	term_io__write_term_2(Term, VarSet, 0, _, _).
 
-:- pred term_io__write_term_2(term, varset, int, varset, int,
+:- pred term_io__write_term_2(term(T), varset(T), int, varset(T), int,
 				io__state, io__state).
 :- mode term_io__write_term_2(in, in, in, out, out, di, uo) is det.
 
@@ -156,8 +195,16 @@ term_io__write_term_2(Term, VarSet0, N0, VarSet, N) -->
 	{ ops__max_priority(MaxPriority) },
 	term_io__write_term_3(Term, MaxPriority + 1, VarSet0, N0, VarSet, N).
 
-:- pred term_io__write_term_3(term, ops__priority, varset, int, varset, int,
+:- pred term_io__write_arg_term(term(T), varset(T), int, varset(T), int,
 				io__state, io__state).
+:- mode term_io__write_arg_term(in, in, in, out, out, di, uo) is det.
+
+term_io__write_arg_term(Term, VarSet0, N0, VarSet, N) -->
+	{ ArgPriority = 1000 },
+	term_io__write_term_3(Term, ArgPriority - 1, VarSet0, N0, VarSet, N).
+
+:- pred term_io__write_term_3(term(T), ops__priority, varset(T), int, varset(T),
+		int, io__state, io__state).
 :- mode term_io__write_term_3(in, in, in, in, out, out, di, uo) is det.
 
 term_io__write_term_3(term__variable(Id), _, VarSet0, N0, VarSet, N) -->
@@ -170,7 +217,7 @@ term_io__write_term_3(term__functor(Functor, Args, _), Priority,
 		{ Args = [ListHead, ListTail] }
 	->
 		io__write_char('['),
-		term_io__write_term_2(ListHead, VarSet0, N0, VarSet1, N1),
+		term_io__write_arg_term(ListHead, VarSet0, N0, VarSet1, N1),
 		term_io__write_list_tail(ListTail, VarSet1, N1, VarSet, N),
 		io__write_char(']')
 	;
@@ -196,7 +243,7 @@ term_io__write_term_3(term__functor(Functor, Args, _), Priority,
 	->
 		term_io__write_variable_2(Var, VarSet0, N0, VarSet1, N1),
 		io__write_char('('),
-		term_io__write_term_2(FirstArg, VarSet1, N1, VarSet2, N2),
+		term_io__write_arg_term(FirstArg, VarSet1, N1, VarSet2, N2),
 		term_io__write_term_args(OtherArgs, VarSet2, N2, VarSet, N),
 		io__write_char(')')
 	;
@@ -275,13 +322,14 @@ term_io__write_term_3(term__functor(Functor, Args, _), Priority,
 			term_io__write_constant(Functor),
 			io__write_char(')')
 		;
-			term_io__write_constant(Functor)
+			term_io__write_constant(Functor,
+				maybe_adjacent_to_graphic_token)
 		),
 		(
 			{ Args = [X|Xs] }
 		->
 			io__write_char('('),
-			term_io__write_term_2(X, VarSet0, N0, VarSet1, N1),
+			term_io__write_arg_term(X, VarSet0, N0, VarSet1, N1),
 			term_io__write_term_args(Xs, VarSet1, N1, VarSet, N),
 			io__write_char(')')
 		;
@@ -307,7 +355,7 @@ maybe_write_char(Char, Priority, OpPriority) -->
 adjust_priority(Priority, y, Priority).
 adjust_priority(Priority, x, Priority - 1).
 
-:- pred term_io__write_list_tail(term, varset, int, varset, int,
+:- pred term_io__write_list_tail(term(T), varset(T), int, varset(T), int,
 				io__state, io__state).
 :- mode term_io__write_list_tail(in, in, in, out, out, di, uo) is det.
 
@@ -321,7 +369,7 @@ term_io__write_list_tail(Term, VarSet0, N0, VarSet, N) -->
 		{ Term = term__functor(term__atom("."), [ListHead, ListTail], _) }
 	->
 		io__write_string(", "),
-		term_io__write_term_2(ListHead, VarSet0, N0, VarSet1, N1),
+		term_io__write_arg_term(ListHead, VarSet0, N0, VarSet1, N1),
 		term_io__write_list_tail(ListTail, VarSet1, N1, VarSet, N)
 	;
 		{ Term = term__functor(term__atom("[]"), [], _) }
@@ -335,7 +383,7 @@ term_io__write_list_tail(Term, VarSet0, N0, VarSet, N) -->
 
 %-----------------------------------------------------------------------------%
 
-:- pred term_io__write_term_args(list(term), varset, int, varset, int,
+:- pred term_io__write_term_args(list(term(T)), varset(T), int, varset(T), int,
 				io__state, io__state).
 :- mode term_io__write_term_args(in, in, in, out, out, di, uo) is det.
 
@@ -343,19 +391,25 @@ term_io__write_list_tail(Term, VarSet0, N0, VarSet, N) -->
 term_io__write_term_args([], VarSet, N, VarSet, N) --> [].
 term_io__write_term_args([X|Xs], VarSet0, N0, VarSet, N) -->
 	io__write_string(", "),
-	term_io__write_term_2(X, VarSet0, N0, VarSet1, N1),
+	term_io__write_arg_term(X, VarSet0, N0, VarSet1, N1),
 	term_io__write_term_args(Xs, VarSet1, N1, VarSet, N).
 
 %-----------------------------------------------------------------------------%
 
-	% write the functor
-term_io__write_constant(term__integer(I)) -->
+term_io__write_constant(Const) -->
+	term_io__write_constant(Const, not_adjacent_to_graphic_token).
+
+:- pred term_io__write_constant(const, adjacent_to_graphic_token,
+	io__state, io__state).
+:- mode term_io__write_constant(in, in, di, uo) is det.
+
+term_io__write_constant(term__integer(I), _) -->
 	io__write_int(I).
-term_io__write_constant(term__float(F)) -->
+term_io__write_constant(term__float(F), _) -->
 	io__write_float(F).
-term_io__write_constant(term__atom(A))  -->
-	term_io__quote_atom(A).
-term_io__write_constant(term__string(S)) -->
+term_io__write_constant(term__atom(A), NextToGraphicToken)  -->
+	term_io__quote_atom(A, NextToGraphicToken).
+term_io__write_constant(term__string(S), _) -->
 	term_io__quote_string(S).
 
 %-----------------------------------------------------------------------------%
@@ -366,6 +420,9 @@ term_io__quote_char(C) -->
 	io__write_char('''').
 
 term_io__quote_atom(S) -->
+	term_io__quote_atom(S, not_adjacent_to_graphic_token).
+
+term_io__quote_atom(S, NextToGraphicToken) -->
 	(
 		% I didn't make these rules up: see ISO Prolog 6.3.1.3
 		% and 6.4.2.
@@ -385,7 +442,23 @@ term_io__quote_atom(S) -->
 			{ string__to_char_list(S, Chars) },
 			{ \+ (  list__member(Char, Chars),
 				\+ lexer__graphic_token_char(Char)) },
-			{ Chars \= [] }
+			{ Chars \= [] },
+			%
+			% We need to quote tokens starting with '#',
+			% because Mercury uses '#' to start source line
+			% number indicators.
+			% 
+			{ Chars \= ['#' | _] },
+			%
+			% If the token could be the last token in a term,
+			% and the term could be followed with ".\n",
+			% then we need to quote the token, otherwise
+			% the "." would be considered part of the
+			% same graphic token.  We can only leave it
+			% unquoted if we're sure it won't be adjacent
+			% to any graphic token.
+			%
+			{ NextToGraphicToken = not_adjacent_to_graphic_token }
 		;
 			% 6.3.1.3: atom = open list, close list ;
 			{ S = "[]" }
@@ -402,6 +475,10 @@ term_io__quote_atom(S) -->
 		io__write_char('''')
 	).
 
+	% Note: the code here is similar to code in
+	% compiler/mercury_to_mercury.m; any changes here
+	% may require similar changes there.
+
 term_io__quote_string(S) -->
 	io__write_char('"'),
 	term_io__write_escaped_string(S),
@@ -413,6 +490,10 @@ term_io__write_escaped_string(String) -->
 term_io__quote_single_char(Char) -->
 	term_io__write_escaped_char(Char).
 
+	% Note: the code here is similar to code in
+	% compiler/mercury_to_mercury.m; any changes here
+	% may require similar changes there.
+
 term_io__write_escaped_char(Char) -->
 	( { mercury_escape_special_char(Char, QuoteChar) } ->
 		io__write_char('\\'),
@@ -436,6 +517,10 @@ term_io__write_escaped_char(Char) -->
 	% ok so long as you don't have two escaped characters
 	% in a row :-(
 
+	% Note: the code here is similar to code in
+	% compiler/mercury_to_mercury.m; any changes here
+	% may require similar changes there.
+
 mercury_escape_char(Char, EscapeCode) :-
 	char__to_int(Char, Int),
 	string__int_to_base_string(Int, 8, OctalString0),
@@ -448,6 +533,10 @@ mercury_escape_char(Char, EscapeCode) :-
 	% Succeed if Char is a character which is allowed in
 	% Mercury string and character literals.
 
+	% Note: the code here is similar to code in
+	% compiler/mercury_to_mercury.m; any changes here
+	% may require similar changes there.
+
 is_mercury_source_char(Char) :-
 	( char__is_alnum(Char) ->
 		true
@@ -460,6 +549,10 @@ is_mercury_source_char(Char) :-
 	% Currently we only allow the following characters.
 	% XXX should we just use is_printable(Char) instead?
 
+	% Note: the code here is similar to code in
+	% compiler/mercury_to_mercury.m; any changes here
+	% may require similar changes there.
+
 :- pred is_mercury_punctuation_char(char).
 :- mode is_mercury_punctuation_char(in) is semidet.
 
@@ -504,6 +597,10 @@ is_mercury_punctuation_char('|').
 	% backslash-escape character EscapeChar that can be used
 	% after a backslash in string literals or atoms to represent Char.
 
+	% Note: the code here is similar to code in
+	% compiler/mercury_to_mercury.m; any changes here
+	% may require similar changes there.
+
 :- pred mercury_escape_special_char(char, char).
 :- mode mercury_escape_special_char(in, out) is semidet.
 
