@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1994-1997 The University of Melbourne.
+% Copyright (C) 1994-1997, 1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -32,6 +32,18 @@
 :- pred tree234__lookup(tree234(K, V), K, V).
 :- mode tree234__lookup(in, in, out) is det.
 
+:- pred tree234__lower_bound_search(tree234(K, V), K, K, V).
+:- mode tree234__lower_bound_search(in, in, out, out) is semidet.
+
+:- pred tree234__lower_bound_lookup(tree234(K, V), K, K, V).
+:- mode tree234__lower_bound_lookup(in, in, out, out) is det.
+
+:- pred tree234__upper_bound_search(tree234(K, V), K, K, V).
+:- mode tree234__upper_bound_search(in, in, out, out) is semidet.
+
+:- pred tree234__upper_bound_lookup(tree234(K, V), K, K, V).
+:- mode tree234__upper_bound_lookup(in, in, out, out) is det.
+
 :- pred tree234__insert(tree234(K, V), K, V, tree234(K, V)).
 :- mode tree234__insert(in, in, in, out) is semidet.
 % :- mode tree234__insert(di_tree234, in, in, uo_tree234) is semidet.
@@ -79,6 +91,16 @@
 :- pred tree234__tree234_to_assoc_list(tree234(K, V), assoc_list(K, V)).
 :- mode tree234__tree234_to_assoc_list(in, out) is det.
 
+:- pred tree234__foldl(pred(K, V, T, T), tree234(K, V), T, T).
+:- mode tree234__foldl(pred(in, in, in, out) is det, in, in, out) is det.
+:- mode tree234__foldl(pred(in, in, in, out) is semidet, in, in, out)
+		is semidet.
+:- mode tree234__foldl(pred(in, in, di, uo) is det, in, di, uo) is det.
+
+:- pred tree234__map_values(pred(K, V, W), tree234(K, V), tree234(K, W)).
+:- mode tree234__map_values(pred(in, in, out) is det, in, out) is det.
+:- mode tree234__map_values(pred(in, in, out) is semidet, in, out) is semidet.
+
 %------------------------------------------------------------------------------%
 %------------------------------------------------------------------------------%
 
@@ -268,6 +290,335 @@ tree234__search(T, K, V) :-
 		)
 	).
 
+tree234__lookup(T, K, V) :-
+	( tree234__search(T, K, V0) ->
+		V = V0
+	;
+		report_lookup_error("tree234__lookup: key not found.", K, V)
+	).
+
+%------------------------------------------------------------------------------%
+
+tree234__lower_bound_search(T, SearchK, K, V) :-
+	(
+		T = empty,
+		fail
+	;
+		T = two(K0, _, _, _),
+		compare(Result, SearchK, K0),
+		(
+			Result = (<),
+			T = two(_, _, T0, _),
+			tree234__lower_bound_search(T0, SearchK, K, V)
+		;
+			Result = (=),
+			T = two(_, V0, _, _),
+			K = SearchK,
+			V = V0
+		;
+			Result = (>),
+			T = two(_, _, _, T1),
+			( tree234__lower_bound_search(T1, SearchK, Kp, Vp) ->
+				K = Kp,
+				V = Vp
+			;
+				T = two(_, V0, _, _),
+				K = K0,
+				V = V0
+			)
+		)
+	;
+		T = three(K0, _, _, _, _, _, _),
+		compare(Result0, SearchK, K0),
+		(
+			Result0 = (<),
+			T = three(_, _, _, _, T0, _, _),
+			tree234__lower_bound_search(T0, SearchK, K, V)
+		;
+			Result0 = (=),
+			T = three(_, V0, _, _, _, _, _),
+			K = SearchK,
+			V = V0
+		;
+			Result0 = (>),
+			T = three(_, _, K1, _, _, _, _),
+			compare(Result1, SearchK, K1),
+			(
+				Result1 = (<),
+				T = three(_, _, _, _, _, T1, _),
+				( tree234__lower_bound_search(T1, SearchK,
+					Kp, Vp)
+				-> 
+					K = Kp,
+					V = Vp
+				;
+					T = three(_, V0, _, _, _, _, _),
+					K = K0,
+					V = V0
+				)
+			;
+				Result1 = (=),
+				T = three(_, _, _, V1, _, _, _),
+				K = SearchK,
+				V = V1
+			;
+				Result1 = (>),
+				T = three(_, _, _, _, _, _, T2),
+				( tree234__lower_bound_search(T2, SearchK,
+					Kp, Vp)
+				-> 
+					K = Kp,
+					V = Vp
+				;
+					T = three(_, _, _, V1, _, _, _),
+					K = K1,
+					V = V1
+				)
+			)
+		)
+	;
+		T = four(_, _, K1, _, _, _, _, _, _, _),
+		compare(Result1, SearchK, K1),
+		(
+			Result1 = (<),
+			T = four(K0, _, _, _, _, _, _, _, _, _),
+			compare(Result0, SearchK, K0),
+			(
+				Result0 = (<),
+				T = four(_, _, _, _, _, _, T0, _, _, _),
+				tree234__lower_bound_search(T0, SearchK, K, V)
+			;
+				Result0 = (=),
+				T = four(_, V0, _, _, _, _, _, _, _, _),
+				K = SearchK,
+				V = V0
+			;
+				Result0 = (>),
+				T = four(_, _, _, _, _, _, _, T1, _, _),
+				( tree234__lower_bound_search(T1, SearchK,
+					Kp, Vp)
+				-> 
+					K = Kp,
+					V = Vp
+				;
+					T = four(_, V0, _, _, _, _, _, _, _, _),
+					K = K0,
+					V = V0
+				)
+			)
+		;
+			Result1 = (=),
+			T = four(_, _, _, V1, _, _, _, _, _, _),
+			K = SearchK,
+			V = V1
+		;
+			Result1 = (>),
+			T = four(_, _, _, _, K2, _, _, _, _, _),
+			compare(Result2, SearchK, K2),
+			(
+				Result2 = (<),
+				T = four(_, _, _, _, _, _, _, _, T2, _),
+				( tree234__lower_bound_search(T2, SearchK,
+					Kp, Vp)
+				-> 
+					K = Kp,
+					V = Vp
+				;
+					T = four(_, _, _, V1, _, _, _, _, _, _),
+					K = K1,
+					V = V1
+				)
+			;
+				Result2 = (=),
+				T = four(_, _, _, _, _, V2, _, _, _, _),
+				K = SearchK,
+				V = V2
+			;
+				Result2 = (>),
+				T = four(_, _, _, _, _, _, _, _, _, T3),
+				( tree234__lower_bound_search(T3, SearchK,
+					Kp, Vp)
+				-> 
+					K = Kp,
+					V = Vp
+				;
+					T = four(_, _, _, _, _, V2, _, _, _, _),
+					K = K2,
+					V = V2
+				)
+			)
+		)
+	).
+
+tree234__lower_bound_lookup(T, SearchK, K, V) :-
+	( tree234__lower_bound_search(T, SearchK, K0, V0) ->
+		K = K0,
+		V = V0
+	;
+		report_lookup_error("tree234__lower_bound_lookup: key not found.",
+			SearchK, V)
+	).
+
+%------------------------------------------------------------------------------%
+
+tree234__upper_bound_search(T, SearchK, K, V) :-
+	(
+		T = empty,
+		fail
+	;
+		T = two(K0, _, _, _),
+		compare(Result, SearchK, K0),
+		(
+			Result = (<),
+			T = two(_, _, T0, _),
+			( tree234__upper_bound_search(T0, SearchK, Kp, Vp) -> 
+				K = Kp,
+				V = Vp
+			;
+				T = two(_, V0, _, _),
+				K = K0,
+				V = V0
+			)
+		;
+			Result = (=),
+			T = two(_, V0, _, _),
+			K = SearchK,
+			V = V0
+		;
+			Result = (>),
+			T = two(_, _, _, T1),
+			tree234__upper_bound_search(T1, SearchK, K, V)
+		)
+	;
+		T = three(K0, _, _, _, _, _, _),
+		compare(Result0, SearchK, K0),
+		(
+			Result0 = (<),
+			T = three(_, _, _, _, T0, _, _),
+			( tree234__upper_bound_search(T0, SearchK, Kp, Vp) ->
+				K = Kp,
+				V = Vp
+			;
+				T = three(_, V0, _, _, _, _, _),
+				K = K0,
+				V = V0
+			)
+		;
+			Result0 = (=),
+			T = three(_, V0, _, _, _, _, _),
+			K = SearchK,
+			V = V0
+		;
+			Result0 = (>),
+			T = three(_, _, K1, _, _, _, _),
+			compare(Result1, SearchK, K1),
+			(
+				Result1 = (<),
+				T = three(_, _, _, _, _, T1, _),
+				( tree234__upper_bound_search(T1, SearchK,
+					Kp, Vp)
+				->
+					K = Kp,
+					V = Vp
+				;
+					T = three(_, _, _, V1, _, _, _),
+					K = K1,
+					V = V1
+				)
+			;
+				Result1 = (=),
+				T = three(_, _, _, V1, _, _, _),
+				K = SearchK,
+				V = V1
+			;
+				Result1 = (>),
+				T = three(_, _, _, _, _, _, T2),
+				tree234__upper_bound_search(T2, SearchK, K, V)
+			)
+		)
+	;
+		T = four(_, _, K1, _, _, _, _, _, _, _),
+		compare(Result1, SearchK, K1),
+		(
+			Result1 = (<),
+			T = four(K0, _, _, _, _, _, _, _, _, _),
+			compare(Result0, SearchK, K0),
+			(
+				Result0 = (<),
+				T = four(_, _, _, _, _, _, T0, _, _, _),
+				( tree234__upper_bound_search(T0, SearchK,
+					Kp, Vp)
+				->
+					K = Kp,
+					V = Vp
+				;
+					T = four(_, V0, _, _, _, _, _, _, _, _),
+					K = K0,
+					V = V0
+				)
+			;
+				Result0 = (=),
+				T = four(_, V0, _, _, _, _, _, _, _, _),
+				K = SearchK,
+				V = V0
+			;
+				Result0 = (>),
+				T = four(_, _, _, _, _, _, _, T1, _, _),
+				( tree234__upper_bound_search(T1, SearchK,
+					Kp, Vp)
+				->
+					K = Kp,
+					V = Vp
+				;
+					T = four(_, _, _, V1, _, _, _, _, _, _),
+					K = K1,
+					V = V1
+				)
+			)
+		;
+			Result1 = (=),
+			T = four(_, _, _, V1, _, _, _, _, _, _),
+			K = SearchK,
+			V = V1
+		;
+			Result1 = (>),
+			T = four(_, _, _, _, K2, _, _, _, _, _),
+			compare(Result2, SearchK, K2),
+			(
+				Result2 = (<),
+				T = four(_, _, _, _, _, _, _, _, T2, _),
+				( tree234__upper_bound_search(T2, SearchK,
+					Kp, Vp)
+				->
+					K = Kp,
+					V = Vp
+				;
+					T = four(_, _, _, _, _, V2, _, _, _, _),
+					K = K2,
+					V = V2
+				)
+			;
+				Result2 = (=),
+				T = four(_, _, _, _, _, V2, _, _, _, _),
+				K = SearchK,
+				V = V2
+			;
+				Result2 = (>),
+				T = four(_, _, _, _, _, _, _, _, _, T3),
+				tree234__upper_bound_search(T3, SearchK, K, V)
+			)
+		)
+	).
+
+tree234__upper_bound_lookup(T, SearchK, K, V) :-
+	( tree234__upper_bound_search(T, SearchK, K0, V0) ->
+		K = K0,
+		V = V0
+	;
+		report_lookup_error("tree234__upper_bound_lookup: key not found.",
+			SearchK, V)
+	).
+
 %------------------------------------------------------------------------------%
 
 tree234__update(Tin, K, V, Tout) :-
@@ -387,17 +738,6 @@ tree234__update(Tin, K, V, Tout) :-
 		)
 	).
 
-%------------------------------------------------------------------------------%
-
-tree234__lookup(T, K, V) :-
-	(
-		tree234__search(T, K, V0)
-	->
-		V = V0
-	;
-		error("tree234__lookup: key not found.")
-	).
-
 %------------------------------------------------------------------------------%
 %------------------------------------------------------------------------------%
 
@@ -2060,6 +2400,56 @@ tree234__tree234_to_assoc_list_2(four(K0, V0, K1, V1, K2, V2, T0, T1, T2, T3),
 	tree234__tree234_to_assoc_list_2(T1, [K1 - V1 | L2], L3),
 	tree234__tree234_to_assoc_list_2(T0, [K0 - V0 | L3], L).
 
+%------------------------------------------------------------------------------%
+
+tree234__foldl(_Pred, empty, Acc, Acc).
+tree234__foldl(Pred, two(K, V, T0, T1), Acc0, Acc) :-
+	tree234__foldl(Pred, T0, Acc0, Acc1),
+	call(Pred, K, V, Acc1, Acc2),
+	tree234__foldl(Pred, T1, Acc2, Acc).
+tree234__foldl(Pred, three(K0, V0, K1, V1, T0, T1, T2), Acc0, Acc) :-
+	tree234__foldl(Pred, T0, Acc0, Acc1),
+	call(Pred, K0, V0, Acc1, Acc2),
+	tree234__foldl(Pred, T1, Acc2, Acc3),
+	call(Pred, K1, V1, Acc3, Acc4),
+	tree234__foldl(Pred, T2, Acc4, Acc).
+tree234__foldl(Pred, four(K0, V0, K1, V1, K2, V2, T0, T1, T2, T3), Acc0, Acc) :-
+	tree234__foldl(Pred, T0, Acc0, Acc1),
+	call(Pred, K0, V0, Acc1, Acc2),
+	tree234__foldl(Pred, T1, Acc2, Acc3),
+	call(Pred, K1, V1, Acc3, Acc4),
+	tree234__foldl(Pred, T2, Acc4, Acc5),
+	call(Pred, K2, V2, Acc5, Acc6),
+	tree234__foldl(Pred, T3, Acc6, Acc).
+
+%------------------------------------------------------------------------------%
+
+tree234__map_values(_Pred, empty, empty).
+tree234__map_values(Pred, Tree0, Tree) :-
+	Tree0 = two(K0, V0, Left0, Right0),
+	Tree  = two(K0, W0, Left, Right),
+	call(Pred, K0, V0, W0),
+	tree234__map_values(Pred, Left0, Left),
+	tree234__map_values(Pred, Right0, Right).
+tree234__map_values(Pred, Tree0, Tree) :-
+	Tree0 = three(K0, V0, K1, V1, Left0, Middle0, Right0),
+	Tree  = three(K0, W0, K1, W1, Left, Middle, Right),
+	call(Pred, K0, V0, W0),
+	call(Pred, K1, V1, W1),
+	tree234__map_values(Pred, Left0, Left),
+	tree234__map_values(Pred, Middle0, Middle),
+	tree234__map_values(Pred, Right0, Right).
+tree234__map_values(Pred, Tree0, Tree) :-
+	Tree0 = four(K0, V0, K1, V1, K2, V2, Left0, LMid0, RMid0, Right0),
+	Tree  = four(K0, W0, K1, W1, K2, W2, Left, LMid, RMid, Right),
+	call(Pred, K0, V0, W0),
+	call(Pred, K1, V1, W1),
+	call(Pred, K2, V2, W2),
+	tree234__map_values(Pred, Left0, Left),
+	tree234__map_values(Pred, LMid0, LMid),
+	tree234__map_values(Pred, RMid0, RMid),
+	tree234__map_values(Pred, Right0, Right).
+
 %------------------------------------------------------------------------------%
 
 	% count the number of elements in a tree
@@ -2082,3 +2472,70 @@ tree234__count(four(_, _, _, _, _, _, T0, T1, T2, T3), N) :-
 
 %------------------------------------------------------------------------------%
 %------------------------------------------------------------------------------%
+% Ralph Becket <rwab1@cl.cam.ac.uk> 30/04/99
+%       Function forms added.
+
+:- interface.
+
+:- func tree234__init = tree234(K, V).
+
+:- func tree234__lookup(tree234(K, V), K) = V.
+
+:- func tree234__set(tree234(K, V), K, V) = tree234(K, V).
+
+:- func tree234__delete(tree234(K, V), K) = tree234(K, V).
+
+:- func tree234__keys(tree234(K, V)) = list(K).
+
+:- func tree234__values(tree234(K, V)) = list(V).
+
+:- func tree234__count(tree234(K, V)) = int.
+
+:- func tree234__assoc_list_to_tree234(assoc_list(K, V)) = tree234(K, V).
+
+:- func tree234__tree234_to_assoc_list(tree234(K, V)) = assoc_list(K, V).
+
+:- func tree234__foldl(func(K, V, T) = T, tree234(K, V), T) = T.
+
+:- func tree234__map_values(func(K, V) = W, tree234(K, V)) = tree234(K, W).
+
+% ---------------------------------------------------------------------------- %
+% ---------------------------------------------------------------------------- %
+
+:- implementation.
+
+tree234__init = T :-
+	tree234__init(T).
+
+tree234__lookup(T, K) = V :-
+	tree234__lookup(T, K, V).
+
+tree234__set(T1, K, V) = T2 :-
+	tree234__set(T1, K, V, T2).
+
+tree234__delete(T1, K) = T2 :-
+	tree234__delete(T1, K, T2).
+
+tree234__keys(T) = Ks :-
+	tree234__keys(T, Ks).
+
+tree234__values(T) = Vs :-
+	tree234__values(T, Vs).
+
+tree234__count(T) = N :-
+	tree234__count(T, N).
+
+tree234__assoc_list_to_tree234(AL) = T :-
+	tree234__assoc_list_to_tree234(AL, T).
+
+tree234__tree234_to_assoc_list(T) = AL :-
+	tree234__tree234_to_assoc_list(T, AL).
+
+tree234__foldl(F, T, A) = B :-
+	P = ( pred(W::in, X::in, Y::in, Z::out) is det :- Z = F(W, X, Y) ),
+	tree234__foldl(P, T, A, B).
+
+tree234__map_values(F, T1) = T2 :-
+	P = ( pred(X::in, Y::in, Z::out) is det :- Z = F(X, Y) ),
+	tree234__map_values(P, T1, T2).
+
