@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1993-1997 The University of Melbourne.
+% Copyright (C) 1993-1999 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -28,79 +28,78 @@
 
 :- module varset.
 :- interface.
-:- import_module term.
+:- import_module term, list, map, set, assoc_list.
 
-:- type varset.
+:- type varset(T).
+
+:- type varset	==	varset(generic).
 
 	% construct an empty varset.
-:- pred varset__init(varset).
+:- pred varset__init(varset(T)).
 :- mode varset__init(out) is det.
 
 	% check whether a varset is empty.
-:- pred varset__is_empty(varset).
+:- pred varset__is_empty(varset(T)).
 :- mode varset__is_empty(in) is semidet.
 
 	% create a new variable
-:- pred varset__new_var(varset, var, varset).
+:- pred varset__new_var(varset(T), var(T), varset(T)).
 :- mode varset__new_var(in, out, out) is det.
 
 	% create a new named variable
-:- pred varset__new_named_var(varset, string, var, varset).
+:- pred varset__new_named_var(varset(T), string, var(T), varset(T)).
 :- mode varset__new_named_var(in, in, out, out) is det.
 
 	% create multiple new variables
-:- pred varset__new_vars(varset, int, list(var), varset).
+:- pred varset__new_vars(varset(T), int, list(var(T)), varset(T)).
 :- mode varset__new_vars(in, in, out, out) is det.
 
 	% delete the name and value for a variable
-:- pred varset__delete_var(varset, var, varset).
+:- pred varset__delete_var(varset(T), var(T), varset(T)).
 :- mode varset__delete_var(in, in, out) is det.
 
 	% delete the names and values for a list of variables
-:- pred varset__delete_vars(varset, list(var), varset).
+:- pred varset__delete_vars(varset(T), list(var(T)), varset(T)).
 :- mode varset__delete_vars(in, in, out) is det.
 
 	% return a list of all the variables in a varset
-:- pred varset__vars(varset, list(var)).
+:- pred varset__vars(varset(T), list(var(T))).
 :- mode varset__vars(in, out) is det.
 
 	% set the name of a variable
-	% (if there is already a variable with the same name "Foo",
-	% then try naming it "Foo'", or "Foo''", or "Foo'''", etc. until
-	% an unused name is found.)
-:- pred varset__name_var(varset, var, string, varset).
+:- pred varset__name_var(varset(T), var(T), string, varset(T)).
 :- mode varset__name_var(in, in, in, out) is det.
 
 	% lookup the name of a variable;
 	% create one if it doesn't have one using V_ as a prefix
-:- pred varset__lookup_name(varset, var, string).
+:- pred varset__lookup_name(varset(T), var(T), string).
 :- mode varset__lookup_name(in, in, out) is det.
 
 	% lookup the name of a variable;
 	% create one if it doesn't have one using the specified prefix
-:- pred varset__lookup_name(varset, var, string, string).
+:- pred varset__lookup_name(varset(T), var(T), string, string).
 :- mode varset__lookup_name(in, in, in, out) is det.
 
 	% lookup the name of a variable;
 	% fail if it doesn't have one
-:- pred varset__search_name(varset, var, string).
+:- pred varset__search_name(varset(T), var(T), string).
 :- mode varset__search_name(in, in, out) is semidet.
 
 	% bind a value to a variable
 	% (will overwrite any existing binding).
-:- pred varset__bind_var(varset, var, term, varset).
+:- pred varset__bind_var(varset(T), var(T), term(T), varset(T)).
 :- mode varset__bind_var(in, in, in, out) is det.
 
 	% bind a set of terms to a set of variables.
-:- pred varset__bind_vars(varset, substitution, varset).
+:- pred varset__bind_vars(varset(T), substitution(T), varset(T)).
 :- mode varset__bind_vars(in, in, out) is det.
 
 	% lookup the value of a variable
-:- pred varset__search_var(varset, var, term).
+:- pred varset__search_var(varset(T), var(T), term(T)).
 :- mode varset__search_var(in, in, out) is semidet.
 
 	% get the bindings for all the bound variables.
-:- pred varset__lookup_vars(varset, substitution).
+:- pred varset__lookup_vars(varset(T), substitution(T)).
 :- mode varset__lookup_vars(in, out) is det.
 
 	% Combine two different varsets, renaming apart:
@@ -110,21 +109,38 @@
 	% and Terms is Terms0 renamed accordingly.
 	% (Any bindings in NewVarSet are ignored.)
 
-:- pred varset__merge(varset, varset, list(term), varset, list(term)).
+:- pred varset__merge(varset(T), varset(T), list(term(T)),
+		varset(T), list(term(T))).
 :- mode varset__merge(in, in, in, out, out) is det.
 
 	% As above, except return the substitution directly
 	% rather than applying it to a list of terms.
 
-:- pred varset__merge_subst(varset, varset, varset, substitution).
+:- pred varset__merge_subst(varset(T), varset(T), varset(T), substitution(T)).
 :- mode varset__merge_subst(in, in, out, out) is det.
 
+	% Same as varset__merge, except that the names of variables
+	% in NewVarSet are not included in the final varset.
+	% This is useful if varset__create_name_var_map needs
+	% to be used on the resulting varset.
+
+:- pred varset__merge_without_names(varset(T), varset(T), list(term(T)),
+		varset(T), list(term(T))).
+:- mode varset__merge_without_names(in, in, in, out, out) is det.
+
+	% As above, except return the substitution directly
+	% rather than applying it to a list of terms.
+
+:- pred varset__merge_subst_without_names(varset(T),
+		varset(T), varset(T), substitution(T)).
+:- mode varset__merge_subst_without_names(in, in, out, out) is det.
+
 	% get the bindings for all the bound variables.
-:- pred varset__get_bindings(varset, substitution).
+:- pred varset__get_bindings(varset(T), substitution(T)).
 :- mode varset__get_bindings(in, out) is det.
 
 	% set the bindings for all the bound variables.
-:- pred varset__set_bindings(varset, substitution, varset).
+:- pred varset__set_bindings(varset(T), substitution(T), varset(T)).
 :- mode varset__set_bindings(in, in, out) is det.
 
 	% Create a map from names to variables.
@@ -132,13 +148,13 @@
 	% shared by more than one variable. Therefore this predicate
 	% is only really useful if it is already known that no two
 	% variables share the same name.
-:- pred varset__create_name_var_map(varset, map(string, var)).
+:- pred varset__create_name_var_map(varset(T), map(string, var(T))).
 :- mode varset__create_name_var_map(in, out) is det.
 
 	% Return an association list giving the name of each variable.
 	% Every variable has an entry in the returned association list,
 	% even if it shares its name with another variable.
-:- pred varset__var_name_list(varset, assoc_list(var, string)).
+:- pred varset__var_name_list(varset(T), assoc_list(var(T), string)).
 :- mode varset__var_name_list(in, out) is det.
 
 	% Given a list of variable and varset in which some variables have
@@ -146,18 +162,37 @@
 	% return another varset in which every variable has a unique name.
 	% If necessary, names will have suffixes added on the end;
 	% the second argument gives the suffix to use.
-:- pred varset__ensure_unique_names(list(var), string, varset, varset).
+:- pred varset__ensure_unique_names(list(var(T)), string, varset(T), varset(T)).
 :- mode varset__ensure_unique_names(in, in, in, out) is det.
 
+	% Given a varset and a set of variables, remove the names
+	% and values of any other variables stored in the varset.
+:- pred varset__select(varset(T), set(var(T)), varset(T)).
+:- mode varset__select(in, in, out) is det.
+
+	% Given a varset and a list of variables, construct a new varset
+	% containing one variable for each one in the list (and no others).
+	% Also return a substitution mapping the selected variables in the
+	% original varset into variables in the new varset. The relative
+	% ordering of variables in the original varset is maintained.
+:- pred varset__squash(varset(T), list(var(T)), varset(T), map(var(T), var(T))).
+:- mode varset__squash(in, in, out, out) is det.
+
+	% Coerce the types of the variables in a varset.
+:- pred varset__coerce(varset(T), varset(U)).
+:- mode varset__coerce(in, out) is det.
+
+:- implementation.
 %-----------------------------------------------------------------------------%
 
 :- implementation.
-:- import_module int, list, map, std_util, assoc_list, set, require, string.
+:- import_module bool, int, list, map, std_util, assoc_list.
+:- import_module set, require, string.
 
-:- type varset		--->	varset(
-					var_supply,
-					map(var, string),
-					map(var, term)
+:- type varset(T)	--->	varset(
+					var_supply(T),
+					map(var(T), string),
+					map(var(T), term(T))
 				).
 
 %-----------------------------------------------------------------------------%
@@ -181,12 +216,13 @@ varset__new_var(varset(MaxId0, Names, Vals), Var,
 varset__new_named_var(varset(MaxId0, Names0, Vals), Name, Var,
 		varset(MaxId, Names, Vals)) :-
 	term__create_var(MaxId0, Var, MaxId),
-	varset__name_var_2(Names0, Var, Name, Names).
+	map__set(Names0, Var, Name, Names).
 
 varset__new_vars(Varset0, NumVars, NewVars, Varset) :-
 	varset__new_vars_2(Varset0, NumVars, [], NewVars, Varset).
 
-:- pred varset__new_vars_2(varset, int, list(var), list(var), varset).
+:- pred varset__new_vars_2(varset(T), int, list(var(T)),
+		list(var(T)), varset(T)).
 :- mode varset__new_vars_2(in, in, in, out, out) is det.
 
 varset__new_vars_2(Varset0, NumVars, NewVars0, NewVars, Varset) :-
@@ -227,8 +263,8 @@ varset__vars(varset(MaxId0, _, _), L) :-
 	varset__vars_2(V0, MaxId0, [], L1),
 	list__reverse(L1, L).
 
-:- pred varset__vars_2(var_supply, var_supply, list(var),
-			list(var)).
+:- pred varset__vars_2(var_supply(T), var_supply(T), list(var(T)),
+			list(var(T))).
 :- mode varset__vars_2(in, in, in, out) is det.
 
 varset__vars_2(N, Max, L0, L) :-
@@ -241,27 +277,11 @@ varset__vars_2(N, Max, L0, L) :-
 
 %-----------------------------------------------------------------------------%
 
-	% If you attempt to name a variable a name which is already
-	% in use, we append sufficient primes ("'") to make the name unique.
-	% Potential efficiency problem: this means that giving N variables
-	% the same name costs O(N*N) time and space.
-
 varset__name_var(VarSet0, Id, Name, VarSet) :-
 	VarSet0 = varset(MaxId, Names0, Vals),
-	varset__name_var_2(Names0, Id, Name, Names),
+	map__set(Names0, Id, Name, Names),
 	VarSet = varset(MaxId, Names, Vals).
 
-:- pred varset__name_var_2(map(var, string), var, string, map(var, string)).
-:- mode varset__name_var_2(in, in, in, out) is det.
-
-varset__name_var_2(Names0, Id, Name, Names) :-
-	( string__remove_suffix(Name, "'", _) ->
-		error("varset__name_var: name is already primed")
-	;
-		true
-	),
-	map__set(Names0, Id, Name, Names).
-
 %-----------------------------------------------------------------------------%
 
 varset__lookup_name(Varset, Id, Name) :-
@@ -310,7 +330,7 @@ varset__bind_vars(Varset0, Subst, Varset) :-
 	map__to_assoc_list(Subst, VarTermList),
 	varset__bind_vars_2(VarTermList, Varset0, Varset).
 
-:- pred varset__bind_vars_2(assoc_list(var, term), varset, varset).
+:- pred varset__bind_vars_2(assoc_list(var(T), term(T)), varset(T), varset(T)).
 :- mode varset__bind_vars_2(in, in, out) is det.
 
 varset__bind_vars_2([], Varset, Varset).
@@ -343,21 +363,47 @@ varset__set_bindings(varset(C, N, _), S, varset(C, N, S)).
 	% this substition to the list of terms.
 
 varset__merge(VarSet0, VarSet1, TermList0, VarSet, TermList) :-
-	varset__merge_subst(VarSet0, VarSet1, VarSet, Subst),
+	IncludeNames = yes,
+	varset__merge_subst(IncludeNames, VarSet0, VarSet1, VarSet, Subst),
+	term__apply_substitution_to_list(TermList0, Subst, TermList).
+
+varset__merge_without_names(VarSet0, VarSet1, TermList0, VarSet, TermList) :-
+	IncludeNames = no,
+	varset__merge_subst(IncludeNames,
+		VarSet0, VarSet1, VarSet, Subst),
 	term__apply_substitution_to_list(TermList0, Subst, TermList).
 
 varset__merge_subst(VarSet0, varset(MaxId, Names, Vals),
 			VarSet, Subst) :-
+	IncludeNames = yes,	
+	varset__merge_subst(IncludeNames, VarSet0, varset(MaxId, Names, Vals),
+			VarSet, Subst).
+
+varset__merge_subst_without_names(VarSet0, varset(MaxId, Names, Vals),
+			VarSet, Subst) :-
+	IncludeNames = no,	
+	varset__merge_subst(IncludeNames, VarSet0, varset(MaxId, Names, Vals),
+			VarSet, Subst).
+
+:- pred varset__merge_subst(bool, varset(T), varset(T), varset(T),
+		substitution(T)).
+:- mode varset__merge_subst(in, in, in, out, out) is det.
+
+varset__merge_subst(IncludeNames, VarSet0, varset(MaxId, Names, Vals),
+			VarSet, Subst) :-
 	term__init_var_supply(N),
 	map__init(Subst0),
-	varset__merge_subst_2(N, MaxId, Names, Vals, VarSet0, Subst0,
-				VarSet, Subst).
+	varset__merge_subst_2(IncludeNames, N, MaxId, Names, Vals,
+			VarSet0, Subst0, VarSet, Subst).
 
-:- pred varset__merge_subst_2(var_supply, var_supply, map(var, string),
-	map(var, term), varset, substitution, varset, substitution).
-:- mode varset__merge_subst_2(in, in, in, in, in, in, out, out) is det.
+:- pred varset__merge_subst_2(bool, var_supply(T),
+	var_supply(T), map(var(T), string),
+	map(var(T), term(T)), varset(T), substitution(T),
+	varset(T), substitution(T)).
+:- mode varset__merge_subst_2(in, in, in, in, in, in, in, out, out) is det.
 
-varset__merge_subst_2(N, Max, Names, Vals, VarSet0, Subst0, VarSet, Subst) :-
+varset__merge_subst_2(IncludeNames, N, Max, Names, Vals,
+		VarSet0, Subst0, VarSet, Subst) :-
 	( N = Max ->
 		VarSet = VarSet0,
 		Subst0 = Subst
@@ -365,6 +411,7 @@ varset__merge_subst_2(N, Max, Names, Vals, VarSet0, Subst0, VarSet, Subst) :-
 		varset__new_var(VarSet0, VarId, VarSet1),
 		term__create_var(N, VarN, N1),
 		(
+			IncludeNames = yes,
 			map__search(Names, VarN, Name)
 		->
 			varset__name_var(VarSet1, VarId, Name, VarSet2)
@@ -372,8 +419,8 @@ varset__merge_subst_2(N, Max, Names, Vals, VarSet0, Subst0, VarSet, Subst) :-
 			VarSet2 = VarSet1
 		),
 		map__set(Subst0, VarN, term__variable(VarId), Subst1),
-		varset__merge_subst_2(N1, Max, Names, Vals, VarSet2, Subst1,
-				VarSet, Subst)
+		varset__merge_subst_2(IncludeNames, N1, Max, Names,
+				Vals, VarSet2, Subst1, VarSet, Subst)
 	).
 
 %-----------------------------------------------------------------------------%
@@ -398,8 +445,8 @@ varset__ensure_unique_names(AllVars, Suffix,
 	varset__ensure_unique_names_2(AllVars, Suffix, UsedNames, VarNameMap0,
 		VarNameMap1, VarNameMap).
 
-:- pred varset__ensure_unique_names_2(list(var), string, set(string),
-	map(var, string), map(var, string), map(var, string)).
+:- pred varset__ensure_unique_names_2(list(var(T)), string, set(string),
+	map(var(T), string), map(var(T), string), map(var(T), string)).
 :- mode varset__ensure_unique_names_2(in, in, in, in, in, out) is det.
 
 varset__ensure_unique_names_2([], _, _, _, VarNameMap, VarNameMap).
@@ -437,4 +484,169 @@ varset__ensure_unique_names_3(Trial0, Suffix, UsedNames, Final) :-
 	).
 
 %-----------------------------------------------------------------------------%
+
+varset__select(varset(Supply, VarNameMap0, Values0), Vars,
+		varset(Supply, VarNameMap, Values)) :-
+	map__select(VarNameMap0, Vars, VarNameMap),
+	map__select(Values0, Vars, Values).
+
 %-----------------------------------------------------------------------------%
+
+varset__squash(OldVarSet, KeptVars, NewVarSet, Subst) :-
+	%
+	% Create a new varset with the same number of variables. 
+	%
+	list__length(KeptVars, NumVars),
+	varset__init(NewVarSet0),
+	varset__new_vars(NewVarSet0, NumVars, 
+		NewVars0, NewVarSet1),
+	%
+	% We need to sort the fresh variables, to
+	% ensure that the substitution that we create below
+	% does not alter the relative ordering of the variables
+	%
+	list__sort(NewVars0, NewVars),
+
+	%
+	% Copy the variable names across from the old
+	% varset to the new varset.
+	%
+	varset__var_name_list(OldVarSet, VarNames),
+	map__from_corresponding_lists(KeptVars, NewVars, Subst),
+	copy_var_names(VarNames, Subst, NewVarSet1, NewVarSet).
+
+:- pred copy_var_names(assoc_list(var(T), string), map(var(T), var(T)),
+		varset(T), varset(T)).
+:- mode copy_var_names(in, in, in, out) is det.
+
+copy_var_names([], _Subst, NewVarSet, NewVarSet).
+copy_var_names([OldVar - Name | Rest], Subst, NewVarSet0, NewVarSet) :-
+	( map__search(Subst, OldVar, NewVar) ->
+		varset__name_var(NewVarSet0, NewVar, Name, NewVarSet1)
+	;
+		NewVarSet1 = NewVarSet0
+	),
+	copy_var_names(Rest, Subst, NewVarSet1, NewVarSet).
+
+%-----------------------------------------------------------------------------%
+
+varset__coerce(varset(S0, N0, B0), varset(S, N, B)) :-
+		% Coerce the var_supply
+	term__coerce_var_supply(S0, S),
+
+		% Coerce all the vars in the mapping from vars to names
+	map__to_assoc_list(N0, NList0),
+	list__map(lambda([P0::in, P::out] is det, (
+		P0 = V0 - Name,
+		P = V - Name,
+		term__coerce_var(V0, V)
+	)), NList0, NList),
+	map__from_assoc_list(NList, N),
+
+		% Coerce all the vars and terms in the substitution
+	map__to_assoc_list(B0, BList0),
+	list__map(lambda([Q0::in, Q::out] is det, (
+		Q0 = V0 - Binding0,
+		Q = V - Binding,
+		term__coerce_var(V0, V),
+		term__coerce(Binding0, Binding)
+	)), BList0, BList),
+	map__from_assoc_list(BList, B).
+
+%-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
+% Ralph Becket <rwab1@cl.cam.ac.uk> 30/04/99
+%	Function forms added.
+
+:- interface.
+
+:- func varset__init = varset(T).
+
+:- func varset__delete_var(varset(T), var(T)) = varset(T).
+
+:- func varset__delete_vars(varset(T), list(var(T))) = varset(T).
+
+:- func varset__vars(varset(T)) = list(var(T)).
+
+:- func varset__name_var(varset(T), var(T), string) = varset(T).
+
+:- func varset__lookup_name(varset(T), var(T)) = string.
+
+:- func varset__lookup_name(varset(T), var(T), string) = string.
+
+:- func varset__bind_var(varset(T), var(T), term(T)) = varset(T).
+
+:- func varset__bind_vars(varset(T), substitution(T)) = varset(T).
+
+:- func varset__lookup_vars(varset(T)) = substitution(T).
+
+:- func varset__get_bindings(varset(T)) = substitution(T).
+
+:- func varset__set_bindings(varset(T), substitution(T)) = varset(T).
+
+:- func varset__create_name_var_map(varset(T)) = map(string, var(T)).
+
+:- func varset__var_name_list(varset(T)) = assoc_list(var(T), string).
+
+:- func varset__ensure_unique_names(list(var(T)), string, varset(T)) = varset(T).
+
+:- func varset__select(varset(T), set(var(T))) = varset(T).
+
+:- func varset__coerce(varset(T)) = varset(U).
+
+% ---------------------------------------------------------------------------- %
+% ---------------------------------------------------------------------------- %
+
+:- implementation.
+
+varset__init = VS :-
+	varset__init(VS).
+
+varset__delete_var(VS1, V) = VS2 :-
+	varset__delete_var(VS1, V, VS2).
+
+varset__delete_vars(VS1, Vs) = VS2 :-
+	varset__delete_vars(VS1, Vs, VS2).
+
+varset__vars(VS) = Vs :-
+	varset__vars(VS, Vs).
+
+varset__name_var(VS1, V, S) = VS2 :-
+	varset__name_var(VS1, V, S, VS2).
+
+varset__lookup_name(VS, V) = S :-
+	varset__lookup_name(VS, V, S).
+
+varset__lookup_name(VS1, V, S) = S2 :-
+	varset__lookup_name(VS1, V, S, S2).
+
+varset__bind_var(VS1, V, T) = VS2 :-
+	varset__bind_var(VS1, V, T, VS2).
+
+varset__bind_vars(VS1, S) = VS2 :-
+	varset__bind_vars(VS1, S, VS2).
+
+varset__lookup_vars(VS) = S :-
+	varset__lookup_vars(VS, S).
+
+varset__get_bindings(VS) = S :-
+	varset__get_bindings(VS, S).
+
+varset__set_bindings(VS1, S) = VS2 :-
+	varset__set_bindings(VS1, S, VS2).
+
+varset__create_name_var_map(VS) = M :-
+	varset__create_name_var_map(VS, M).
+
+varset__var_name_list(VS) = AL :-
+	varset__var_name_list(VS, AL).
+
+varset__ensure_unique_names(Vs, S1, VS1) = VS2 :-
+	varset__ensure_unique_names(Vs, S1, VS1, VS2).
+
+varset__select(VS1, S) = VS2 :-
+	varset__select(VS1, S, VS2).
+
+varset__coerce(VS1) = VS2 :-
+	varset__coerce(VS1, VS2).
+
