@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1997 The University of Melbourne.
+% Copyright (C) 1995-1998 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -35,6 +35,11 @@
 :- import_module globals, options, generate_output.
 
 output__main(Output, IndexMap) -->
+	globals__io_get_globals(Globals),
+	{ globals__get_what_to_profile(Globals, WhatToProfile) },
+	{ what_to_profile(WhatToProfileString, WhatToProfile) },
+	io__format("*** profiling %s ***\n\n", [s(WhatToProfileString)]),
+
 	{ Output = output(InfoMap, CallList, FlatList) },
 	globals__io_lookup_bool_option(call_graph, CallGraphOpt),
 	(
@@ -45,87 +50,126 @@ output__main(Output, IndexMap) -->
 	;
 		{ true }
 	),
+
 	output__flat_headers,
 	output__flat_profile(FlatList, 0.0, InfoMap, IndexMap),
+
 	output_alphabet_headers,
 	output_alphabet_listing(IndexMap).
 
+:- type header_category
+	--->	time_headers
+	;	memory_words_headers
+	;	memory_cells_headers.
+
+:- pred classify_profile(what_to_profile::in, header_category::out) is det.
+classify_profile(user_time, time_headers).
+classify_profile(user_plus_system_time, time_headers).
+classify_profile(real_time, time_headers).
+classify_profile(memory_words, memory_words_headers).
+classify_profile(memory_cells, memory_cells_headers).
+
+:- pred units(header_category::in, string::out, string::out, string::out, string::out,
+		string::out, string::out, string::out, string::out) is det.
+units(time_headers,	"time", "time", "running time",
+			"seconds", "seconds", "milliseconds", "ms/call", "spent executing").
+units(memory_words_headers,
+			"mem", "memory", "allocated memory",
+			"k-words", "kilowords", "words", "wds/call", "allocated by").
+units(memory_cells_headers,
+			"cells", "allocations", "number of memory allocations",
+			"k-cells", "kilocells", "cells", "cls/call", "occurring in").
 
 :- pred output__call_graph_headers(io__state, io__state).
 :- mode output__call_graph_headers(di, uo) is det.
 
 output__call_graph_headers -->
+	globals__io_get_globals(Globals),
+	{ globals__get_what_to_profile(Globals, WhatToProfile) },
+	{ classify_profile(WhatToProfile, Category) },
+	{ units(Category, ShortWhat, What, LongWhat,
+			_ShortUnits, Units, _MilliUnits, _MilliUnitsPerCall, SpentIn) },
+
 	io__write_string("call graph profile:\n"),
-	io__write_string("\tSorted on the %time field.\n\n"),
+	io__format("\tSorted on the %%%s field.\n\n", [s(ShortWhat)]),
 
-	io__write_string("\tpredicate entries:\n\n"),
+	io__write_string("\tprocedure entries:\n\n"),
 
-	io__write_string("index\tthe index number of the predicate in the call graph\n"),
-	io__write_string("\tlisting.\n\n"),
+	io__write_string("index\t\tthe index number of the procedure in the call graph\n"),
+	io__write_string("\t\tlisting.\n\n"),
 
-	io__write_string("%time\tthe percentage of the total running time of\n"),
-	io__write_string("\tthe program spent in this predicate and its\n"),
-	io__write_string("\tdescendents.\n\n"),
+	io__format("%%%s\t\tthe percentage of the total %s of\n",
+		[s(ShortWhat), s(LongWhat)]),
+	io__format("\t\tthe program %s this procedure and its\n",
+		[s(SpentIn)]),
+	io__write_string("\t\tdescendents.\n\n"),
 
-	io__write_string("self\tthe number of seconds spent actually executing\n"),
-	io__write_string("\tthe predicate's own code.\n\n"),
+	io__format("self\t\tthe number of %s actually %s\n",
+		[s(Units), s(SpentIn)]),
+	io__write_string("\t\tthe procedure's own code.\n\n"),
 
-	io__write_string("descendents\n"),
-	io__write_string("\tthe number of seconds spent executing the\n"),
-	io__write_string("\tdescendents of the current predicate.\n\n"),
+	io__format("descendents\tthe number of %s %s the\n",
+		[s(Units), s(SpentIn)]),
+	io__write_string("\t\tdescendents of the current procedure.\n\n"),
 
-	io__write_string("called\tthe number of times the current predicate is\n"),
-	io__write_string("\tcalled (not counting self recursive calls).\n\n"),
+	io__write_string("called\t\tthe number of times the current procedure is\n"),
+	io__write_string("\t\tcalled (not counting self recursive calls).\n\n"),
 
-	io__write_string("self\tthe number of self recursive calls.\n\n"), 
+	io__write_string("self\t\tthe number of self recursive calls.\n\n"), 
 
-	io__write_string("name\tthe name of the current predicate.\n\n"),
+	io__write_string("name\t\tthe name of the current procedure.\n\n"),
 
-	io__write_string("index\ta index number to locate the function easily.\n\n\n\n"),
+	io__write_string("index\t\tan index number to locate the function easily.\n\n\n\n"),
 
 	io__write_string("\tparent listings:\n\n"),
 
-	io__write_string("self*\tthe number of seconds of the current predicates self\n"),
-	io__write_string("\ttime due to calls from this parent.\n\n"),
+	io__format("self*\t\tthe number of %s of the current procedure's self\n",
+		[s(Units)]),
+	io__format("\t\t%s due to calls from this parent.\n\n",
+		[s(What)]),
 
-	io__write_string("descendents*\n"),
-	io__write_string("\tthe number of seconds of the current predicate's descendent\n"),
-	io__write_string("\ttime which is due to calls from this parent.\n\n"),
+	io__format("descendents*\tthe number of %s of the current procedure's descendent\n",
+		[s(Units)]),
+	io__format("\t\t%s which is due to calls from this parent.\n\n",
+		[s(What)]),
 
-	io__write_string("called*\tthe number of times the current predicate is called\n"),
-	io__write_string("\tby this parent.\n\n"),
+	io__write_string("called*\t\tthe number of times the current procedure is called\n"),
+	io__write_string("\t\tby this parent.\n\n"),
 
-	io__write_string("total\tthe number of times this predicate is called by its parents.\n\n"),
+	io__write_string("total\t\tthe number of times this procedure is called by its parents.\n\n"),
 
-	io__write_string("parents\tthe name of this parent.\n\n"),
+	io__write_string("parents\t\tthe name of this parent.\n\n"),
 
-	io__write_string("index\tthe index number of the parent predicate\n\n\n\n"),
+	io__write_string("index\t\tthe index number of the parent procedure\n\n\n\n"),
 
 
 
 	io__write_string("children listings:\n\n"),
 
-	io__write_string("self*\tthe number of seconds of this child's self time which is due\n"),
-	io__write_string("\tto being called by the current predicate.\n\n"),
+	io__format("self*\t\tthe number of %s of this child's self %s which is\n",
+		[s(Units), s(What)]),
+	io__write_string("\t\tdue to being called by the current procedure.\n\n"),
 
-	io__write_string("descendent*\n"),
-	io__write_string("\tthe number of seconds of this child's desdendent time which is due\n"),
-	io__write_string("\tto the current predicate.\n\n"),
+	io__format("descendent*\tthe number of %s of this child's descendent %s which\n",
+		[s(Units), s(What)]),
+	io__write_string("\t\tis due to the current procedure.\n\n"),
 
-	io__write_string("called*\tthe number of times this child is called by the current\n"),
-	io__write_string("\tpredicate.\n\n"),
+	io__write_string("called*\t\tthe number of times this child is called by the current\n"),
+	io__write_string("\t\tprocedure.\n\n"),
 
-	io__write_string("total*\tthe number of times this child is called by all predicates.\n\n"),
+	io__write_string("total*\t\tthe number of times this child is called by all procedures.\n\n"),
 
 	io__write_string("children\tthe name of this child.\n\n"),
 
-	io__write_string("index\tthe index number of the child.\n\n"),
+	io__write_string("index\t\tthe index number of the child.\n\n\n\n"),
 
 
 
 	io__write_string("                                  called/total"),
 	io__write_string("       parents\n"),
-	io__write_string("index  %time    self descendents  called+self"),
+	{ string__append("%", ShortWhat, PercentShortWhat) },
+	io__format("index %6s    self descendents  called+self",
+		[s(PercentShortWhat)]),
 	io__write_string("    name           index\n"),
 	io__write_string("                                  called/total"),
 	io__write_string("       children\n\n").
@@ -191,12 +235,16 @@ output_formatted_prof_node(ProfNode, Index, IndexMap) -->
 		{ string__format("%67s", [s("<spontaneous>\n")], String) }, 
 		io__write_string(String)
 	;
-		output_formatted_cycle_parent_list(CycleParentList, IndexMap),
-		output_formatted_parent_list(ParentList, IndexMap, TotalCalls)
+		{ list__sort(CycleParentList, SortedCycleParentList) },
+		output_formatted_cycle_parent_list(SortedCycleParentList,
+			IndexMap),
+		{ list__sort(ParentList, SortedParentList) },
+		output_formatted_parent_list(SortedParentList, IndexMap,
+			TotalCalls)
 	),
 
 
-	% Output the info about the current predicate.
+	% Output the info about the current procedure.
 	io__write_string(InitMiddleStr),
 	(
 		{ SelfCalls = 0 }
@@ -212,8 +260,10 @@ output_formatted_prof_node(ProfNode, Index, IndexMap) -->
 	io__write_string(IndexStr),
 	io__write_string("\n"),
 
-	output_formatted_child_list(ChildList, IndexMap),
-	output_formatted_cycle_child_list(CycleChildList, IndexMap),
+	{ list__sort(ChildList, SortedChildList) },
+	output_formatted_child_list(SortedChildList, IndexMap),
+	{ list__sort(CycleChildList, SortedCycleChildList) },
+	output_formatted_cycle_child_list(SortedCycleChildList, IndexMap),
 
 	(
 		{ SelfCalls \= 0 }
@@ -225,7 +275,7 @@ output_formatted_prof_node(ProfNode, Index, IndexMap) -->
 	
 
 % output_formatted_cycle_parent_list
-%	outputs the parents of a predicate that are in the same cycle.
+%	outputs the parents of a procedure that are in the same cycle.
 %
 :- pred output_formatted_cycle_parent_list(list(parent), map(string, int),
                                                         io__state, io__state).
@@ -247,7 +297,7 @@ output_formatted_cycle_parent_list([Parent | Parents], IndexMap) -->
 	
 
 % output_formatted_parent_list:
-%	Outputs the parent list of the current predicate 
+%	Outputs the parent list of the current procedure 
 
 :- pred output_formatted_parent_list(list(parent), map(string, int), int, 
 							io__state, io__state).
@@ -269,7 +319,7 @@ output_formatted_parent_list([Parent | Parents], IndexMap, TotalCalls) -->
 	
 
 % output_formatted_cycle_child_list
-%	outputs the childs of a predicate that are in the same cycle.
+%	outputs the children of a procedure that are in the same cycle.
 %
 :- pred output_formatted_cycle_child_list(list(child), map(string, int),
                                                         io__state, io__state).
@@ -291,7 +341,7 @@ output_formatted_cycle_child_list([Child | Childs], IndexMap) -->
 	
 
 % output_formatted_child_list:
-%	outputs the child list of the current predicate.
+%	outputs the child list of the current procedure.
 %
 :- pred output_formatted_child_list(list(child), map(string, int),
 							io__state, io__state).
@@ -315,31 +365,53 @@ output_formatted_child_list([Child | Children], IndexMap) -->
 :- mode output__flat_headers(di, uo) is det.
 
 output__flat_headers -->
+	globals__io_get_globals(Globals),
+	{ globals__get_what_to_profile(Globals, WhatToProfile) },
+	{ classify_profile(WhatToProfile, Category) },
+	{ units(Category, ShortWhat, _What, LongWhat,
+			ShortUnits, Units, MilliUnits, MilliUnitsPerCall, SpentIn) },
+
 	io__write_string("\nflat profile:\n\n"),
 
-	io__write_string(" %\tthe percentage of total running time of the program\n"),
-	io__write_string("time\tused by this function.\n\n"),
+	io__format("%%\t\tthe percentage of total %s of the program\n",
+		[s(LongWhat)]),
+	io__format("%s\t\tused by this procedure.\n\n",
+		[s(ShortWhat)]),
 
-	io__write_string(" cum\tthe total time of the current predicate and the ones\n"),
-	io__write_string("time\tlisted above it.\n\n"),
+	io__format(
+		"cumulative\tthe total number of %s for the current procedure and\n",
+		[s(Units)]),
+	io__format("%s\t\tthe ones listed above it.\n\n",
+		[s(ShortUnits)]),
 
-	io__write_string(" self\tthe number of seconds accounted for by this predicate alone.\n"),
-	io__write_string("seconds\tThe listing is sorted on this row.\n\n"),
+	io__format(
+	    "self\t\tthe number of %s accounted for by this procedure alone.\n",
+		[s(Units)]),
+	io__format("%s\t\tThe listing is sorted on this row.\n\n",
+		[s(ShortUnits)]),
 
-	io__write_string("calls\tthe number of times this predicate was called.\n\n"),
+	io__write_string(
+		"calls\t\tthe number of times this procedure was called.\n\n"),
 
-	io__write_string(" self\tthe average number of milliseconds spent in\n"),
-	io__write_string("ms/call\tthis predicate per call.\n\n"),
+	io__format("self\t\tthe average number of %s %s\n",
+		[s(MilliUnits), s(SpentIn)]),
+	io__format("%s  \tthis procedure per call.\n\n",
+		[s(MilliUnitsPerCall)]),
 
-	io__write_string(" total\tthe average number of milliseconds spent in this predicate and its\n"),
-	io__write_string("ms/call\tdescendents per call.\n\n"),
+	io__format(
+	   "total\t\tthe average number of %s %s this procedure and its\n",
+		[s(MilliUnits), s(SpentIn)]),
+	io__format("%s  \tdescendents per call.\n\n",
+		[s(MilliUnitsPerCall)]),
 
-	io__write_string("name\tthe name of the predicate followed by its index number.\n\n"),
+	io__write_string(
+	  "name\t\tthe name of the procedure followed by its index number.\n\n"),
 
 	io__write_string("   %  cumulative    self              self"),
 	io__write_string("    total\n"),
-	io__write_string(" time   seconds   seconds    calls  ms/call"),
-	io__write_string("  ms/call name\n").
+	io__format(" %4s    %7s  %7s    calls %8s %8s name\n",
+		[s(ShortWhat), s(ShortUnits), s(ShortUnits),
+		 s(MilliUnitsPerCall), s(MilliUnitsPerCall)]).
 
 
 :- pred output__flat_profile(list(string), float, map(string, output_prof),
@@ -371,13 +443,11 @@ output__flat_profile([LabelName | LNs], CumTime0, InfoMap, IndexMap) -->
 
 	FloatTotalCalls is FloatTotalCalls0 + FloatSelfCalls,
 	Calls is SelfCalls + TotalCalls,
-
-
-	builtin_float_plus(Self, CumTime0, CumTime),
-	checked_float_divide(Self, FloatTotalCalls, Self1),
-	checked_float_divide(Descendant, FloatTotalCalls, Desc1),
-	builtin_float_times(1000.0, Self1, SelfMs),
-	builtin_float_times(1000.0, Desc1, DescMs),
+	CumTime is CumTime0 + Self,
+	checked_float_divide(Self, FloatTotalCalls, SelfSeconds),
+	checked_float_divide(Descendant, FloatTotalCalls, DescSeconds),
+	SelfMs is 1000.0 * SelfSeconds,
+	DescMs is 1000.0 * DescSeconds,
 
 	output__construct_name(Name, CycleNum, FullName),
 	string__int_to_string(Index, IndexStr0),
