@@ -40,45 +40,50 @@
 
 
 process_file__main(Prof, DynamicCallGraph) -->
-        globals__io_lookup_bool_option(very_verbose, VVerbose),
+	globals__io_lookup_bool_option(very_verbose, VVerbose),
 	globals__io_lookup_string_option(declfile, DeclFile),
 	globals__io_lookup_string_option(countfile, CountFile),
 	globals__io_lookup_string_option(pairfile, PairFile),
 	% globals__io_lookup_string_option(libraryfile, LibFile),
 	globals__io_lookup_bool_option(dynamic_cg, Dynamic),
 
-        maybe_write_string(VVerbose, "\n\t% Processing "),
+	% process the decl file
+	maybe_write_string(VVerbose, "\n\t% Processing "),
 	maybe_write_string(VVerbose, DeclFile),
 	maybe_write_string(VVerbose, "..."),
+	process_addr_decl(AddrDeclMap, ProfNodeMap0),
+	maybe_write_string(VVerbose, " done.\n"),
 
-        process_addr_decl(AddrDeclMap, ProfNodeMap0),
-
-        maybe_write_string(VVerbose, " done.\n\t% Processing "),
+	% process the timing counts file
+	maybe_write_string(VVerbose, "\t% Processing "),
 	maybe_write_string(VVerbose, CountFile),
 	maybe_write_string(VVerbose, "..."),
-        process_addr(ProfNodeMap0, ProfNodeMap1, Hertz, ClockTicks,TotalCounts),
+	process_addr(ProfNodeMap0, ProfNodeMap1, WhatToProfile, Scale, Units,
+		TotalCounts),
+	maybe_write_string(VVerbose, " done.\n"),
 
-        maybe_write_string(VVerbose, " done.\n\t% Processing "),
+	% process the call pair counts file
+	maybe_write_string(VVerbose, "\t% Processing "),
 	maybe_write_string(VVerbose, PairFile),
 	maybe_write_string(VVerbose, "..."),
-
-        process_addr_pair(ProfNodeMap1, DynamicCallGraph, ProfNodeMap),
-
-        maybe_write_string(VVerbose, " done.\n"),
+	process_addr_pair(ProfNodeMap1, DynamicCallGraph, ProfNodeMap),
+	maybe_write_string(VVerbose, " done.\n"),
 
 	{ map__init(CycleMap) },
-        { prof_set_entire(Hertz, ClockTicks, TotalCounts, AddrDeclMap, 
+	{ prof_set_entire(Scale, Units, TotalCounts, AddrDeclMap, 
 						ProfNodeMap, CycleMap, Prof) },
+	globals__io_get_globals(Globals0),
+	{ globals__set_what_to_profile(Globals0, WhatToProfile, Globals) },
+	globals__io_set_globals(Globals),
 	
 	(
 		{ Dynamic = no }
 	->
-		maybe_write_string(VVerbose, " done.\n"),
 		% maybe_write_string(VVerbose, "\t% Processing "),
 		% maybe_write_string(VVerbose, LibFile),
 		% maybe_write_string(VVerbose, "..."),
-
 		% process_library_callgraph(_, _),
+		% maybe_write_string(VVerbose, " done.\n"),
 		{ true }
 
 	;
@@ -90,7 +95,7 @@ process_file__main(Prof, DynamicCallGraph) -->
 
 
 % process_addr_decl:
-%	Reads in the addrdecl.out file.
+%	Reads in the Prof.Decl file.
 %	Builds the addrdecl map which associates label names(key) with 
 %	label addresses.
 %	Also builds the prof_node_map which associates label addresses with
@@ -112,7 +117,7 @@ process_addr_decl(AddrDeclMap, ProfNodeMap) -->
 		io__seen
 	;
 		{ Result = error(Error) },
-                { io__error_message(Error, ErrorMsg) },
+		{ io__error_message(Error, ErrorMsg) },
 
 		{ string__append("error opening declaration file `", DeclFile, 
 					Str0) },
@@ -160,31 +165,42 @@ process_addr_decl_2(AddrDecl0, ProfNodeMap0, AddrDecl, ProfNodeMap) -->
 
 
 % process_addr:
-% 	Reads in the addr.out file and stores all the counts in the 
+% 	Reads in the Prof.Counts file and stores all the counts in the 
 % 	prof_node structure.  Also sums the total counts at the same time.
 %
-:- pred process_addr(prof_node_map, prof_node_map, int, int, int,
-							io__state, io__state).
-:- mode process_addr(in, out, out, out, out, di, uo) is det.
+:- pred process_addr(prof_node_map, prof_node_map,
+		what_to_profile, float, string, int, io__state, io__state).
+:- mode process_addr(in, out, out, out, out, out, di, uo) is det.
 
-process_addr(ProfNodeMap0, ProfNodeMap, Hertz, ClockTicks, TotalCounts) -->
+process_addr(ProfNodeMap0, ProfNodeMap, WhatToProfile, Scale, Units,
+		TotalCounts) -->
 	globals__io_lookup_string_option(countfile, CountFile),
 	io__see(CountFile, Result),
 	(
 		{ Result = ok },
-		read_int(Hertz),
-		read_int(ClockTicks),
+		read_what_to_profile(WhatToProfile),
+		read_float(Scale),
+		read_string(Units),
 		process_addr_2(0, ProfNodeMap0, TotalCounts, ProfNodeMap),
 		io__seen
 	;
 		{ Result = error(Error) },
 		{ io__error_message(Error, ErrorMsg) },
-		{ string__append("error opening count file `", CountFile, 
-					Str0) },
-		{ string__append(Str0, "': ", Str1) },
-		{ string__append(Str1, ErrorMsg, Str2) },
-		{ string__append(Str2, "\n", ErrorStr) },
-		{ error(ErrorStr) }
+		io__write_string("\nWarning: error opening `"),
+		io__write_string(CountFile),
+		io__write_string("': "),
+		io__write_string(ErrorMsg),
+		io__write_string("\n"),
+		io__write_string("The generated profile will only include "),
+		io__write_string("call counts.\n\n"),
+		{ TotalCounts = 0 },
+		{ ProfNodeMap = ProfNodeMap0 },
+		% We can use any arbitrary values for WhatToProfile and Scale;
+		% the values specified here won't be used,
+		% since all the times will be zero.
+		{ WhatToProfile = user_plus_system_time },
+		{ Scale = 1.0 },
+		{ Units = "" }
 	).
 
 :- pred process_addr_2(int, prof_node_map, int, prof_node_map, 
@@ -229,7 +245,7 @@ process_addr_2(TotalCounts0, ProfNodeMap0, TotalCounts, ProfNodeMap) -->
 
 
 % process_addr_pair:
-%	Reads in the addrpair.out file and stores the data in the relevant
+%	Reads in the Prof.CallPair file and stores the data in the relevant
 %	lists of the prof_node structure.  Also calculates the number of times
 %	a predicate is called.
 %
@@ -249,7 +265,7 @@ process_addr_pair(ProfNodeMap0, DynamicCallGraph, ProfNodeMap) -->
 		io__seen
 	;
 		{ Result = error(Error) },
-                { io__error_message(Error, ErrorMsg) },
+		{ io__error_message(Error, ErrorMsg) },
 		{ string__append("error opening pair file `", PairFile, 
 					Str0) },
 		{ string__append(Str0, "': ", Str1) },
@@ -343,10 +359,10 @@ process_library_callgraph(LibraryATSort, LibPredMap) -->
 		io__seen
 	;
 		{ Result = error(Error) },
-                { io__error_message(Error, ErrorMsg) },
-                io__stderr_stream(StdErr),
-                io__write_strings(StdErr, ["mprof: error opening pair file `",
-                        LibFile, "': ", ErrorMsg, "\n"]),
+		{ io__error_message(Error, ErrorMsg) },
+		io__stderr_stream(StdErr),
+		io__write_strings(StdErr, ["mprof: error opening pair file `",
+			LibFile, "': ", ErrorMsg, "\n"]),
 		{ LibraryATSort = [] },
 		{ LibPredMap = LibPredMap0 }
 	).
