@@ -1,5 +1,5 @@
 /*
-** Copyright (C) 1993-1997 The University of Melbourne.
+** Copyright (C) 1993-1998 The University of Melbourne.
 ** This file may only be copied under the terms of the GNU Library General
 ** Public License - see the file COPYING.LIB in the Mercury distribution.
 */
@@ -61,10 +61,16 @@
 register	Word	mr0 __asm__("esi");	/* sp */
 register	Word	mr1 __asm__("edi");	/* succip */
 
+#define MR_real_reg_number_mr0	esi
+#define MR_real_reg_number_mr1	edi
+
 #if PIC_REG
-  #define mr2	fake_reg[2]
+  #define mr2	MR_fake_reg[2]
 #else
   register	Word	mr2 __asm__("ebx");	/* r1 */
+
+  #define MR_real_reg_number_mr2	ebx
+
 #endif
 
 #if PIC_REG
@@ -102,39 +108,40 @@ register	Word	mr1 __asm__("edi");	/* succip */
 #define save_transient_regs_to_mem(save_area)		((void)0)
 #define restore_transient_regs_from_mem(save_area)	((void)0)
 
-#define	mr3	fake_reg[3]
-#define	mr4	fake_reg[4]
-#define	mr5	fake_reg[5]
-#define	mr6	fake_reg[6]
-#define	mr7	fake_reg[7]
-#define	mr8	fake_reg[8]
-#define	mr9	fake_reg[9]
-#define	mr10	fake_reg[10]
-#define	mr11	fake_reg[11]
-#define	mr12	fake_reg[12]
-#define	mr13	fake_reg[13]
-#define	mr14	fake_reg[14]
-#define	mr15	fake_reg[15]
-#define	mr16	fake_reg[16]
-#define	mr17	fake_reg[17]
-#define	mr18	fake_reg[18]
-#define	mr19	fake_reg[19]
-#define	mr20	fake_reg[20]
-#define	mr21	fake_reg[21]
-#define	mr22	fake_reg[22]
-#define	mr23	fake_reg[23]
-#define	mr24	fake_reg[24]
-#define	mr25	fake_reg[25]
-#define	mr26	fake_reg[26]
-#define	mr27	fake_reg[27]
-#define	mr28	fake_reg[28]
-#define	mr29	fake_reg[29]
-#define	mr30	fake_reg[30]
-#define	mr31	fake_reg[31]
-#define	mr32	fake_reg[32]
-#define	mr33	fake_reg[33]
-#define	mr34	fake_reg[34]
-#define	mr35	fake_reg[35]
-#define	mr36	fake_reg[36]
+#define	mr3	MR_fake_reg[3]
+#define	mr4	MR_fake_reg[4]
+#define	mr5	MR_fake_reg[5]
+#define	mr6	MR_fake_reg[6]
+#define	mr7	MR_fake_reg[7]
+#define	mr8	MR_fake_reg[8]
+#define	mr9	MR_fake_reg[9]
+#define	mr10	MR_fake_reg[10]
+#define	mr11	MR_fake_reg[11]
+#define	mr12	MR_fake_reg[12]
+#define	mr13	MR_fake_reg[13]
+#define	mr14	MR_fake_reg[14]
+#define	mr15	MR_fake_reg[15]
+#define	mr16	MR_fake_reg[16]
+#define	mr17	MR_fake_reg[17]
+#define	mr18	MR_fake_reg[18]
+#define	mr19	MR_fake_reg[19]
+#define	mr20	MR_fake_reg[20]
+#define	mr21	MR_fake_reg[21]
+#define	mr22	MR_fake_reg[22]
+#define	mr23	MR_fake_reg[23]
+#define	mr24	MR_fake_reg[24]
+#define	mr25	MR_fake_reg[25]
+#define	mr26	MR_fake_reg[26]
+#define	mr27	MR_fake_reg[27]
+#define	mr28	MR_fake_reg[28]
+#define	mr29	MR_fake_reg[29]
+#define	mr30	MR_fake_reg[30]
+#define	mr31	MR_fake_reg[31]
+#define	mr32	MR_fake_reg[32]
+#define	mr33	MR_fake_reg[33]
+#define	mr34	MR_fake_reg[34]
+#define	mr35	MR_fake_reg[35]
+#define	mr36	MR_fake_reg[36]
+#define	mr37	MR_fake_reg[37]
 
 #endif /* not I386_REGS_H */
