@@ -1,7 +1,7 @@
 /*---------------------------------------------------------------------------*/
 
 /*
-** Copyright (C) 1995-1997 The University of Melbourne.
+** Copyright (C) 1995-1999 The University of Melbourne.
 ** This file may only be copied under the terms of the GNU General
 ** Public License - see the file COPYING in the Mercury distribution.
 */
@@ -25,9 +25,14 @@
 #include <string.h>
 #include <stdlib.h>
 #include <stdio.h>
-#include "std.h"
+#include "mercury_std.h"
 
-static void demangle(char *name);
+/* We used this for the size of fixed-length buffers in a few places <sigh> */
+#define MAX_SYMBOL_LENGTH 1000
+
+static void demangle(const char *name);
+static const char *strip_module_name(char **start_ptr, char *end,
+		const char *trailing_context[]);
 static bool check_for_suffix(char *start, char *position, const char *suffix,
 		int sizeof_suffix, int *mode_num2);
 static char *fix_mangled_ascii(char *str, char **end);
@@ -36,6 +41,7 @@ static bool cut_at_double_underscore(char **str, char *end);
 static bool cut_trailing_integer(char *str, char **end, int *num);
 static bool cut_trailing_underscore_integer(char *str, char **end, int *num);
 static bool strip_prefix(char **str, const char *prefix);
+static bool strip_leading_integer(char **start_ptr, int *num);
 
 /*
 ** Bloody SunOS 4.x doesn't have memmove()...
@@ -46,9 +52,46 @@ static bool strip_prefix(char **str, const char *prefix);
 #define memmove memcpy
 #endif
 
+/*
+** This option indicates whether we should output verbose
+** explanations of linker error messages.
+*/
+bool explain_link_errors = FALSE;
+
+/*
+** This variable gets set if the symbols MR_grade_* or MR_mercury_grade
+** were found.  If it gets set, then we print out the error message below.
+*/
+char *found_grade_symbol = NULL;
+const char probably_grade_error[] =
+  "Mercury Linker:\n"
+  "\tNote: the symbol `%s' was mentioned.\n"
+  "\tAny link errors are most likely due to linking together object\n"
+  "\tfiles compiled with different compilation model options.\n"
+  "\tTry doing `mmake clean' and then rebuilding.\n";
+
 int 
 main(int argc, char **argv)
 {
+	const char *progname = argv[0];
+
+	/* we should use getopt_long(), but for one option, that is overkill */
+	while (argc > 1 && argv[1][0] == '-') {
+		if (strcmp(argv[1], "-e") == 0 ||
+		    strcmp(argv[1], "--explain-link-errors") == 0)
+		{
+			explain_link_errors = TRUE;
+			argc--, argv++;
+		} else if (strcmp(argv[1], "--") == 0) {
+			argc--, argv++;
+			break;
+		} else {
+			fprintf(stderr, "%s: unknown option `%s'\n",
+				progname, argv[1]);
+			exit(1);
+		}
+	}
+
 	if (argc > 1) {
 		/*
 		** invoke demangle() on each command line argument
@@ -64,8 +107,8 @@ main(int argc, char **argv)
 		** every valid C identifier in the input
 		*/
 		for (;;) {
-			char buf[1000];
-			int len = 0;
+			char buf[MAX_SYMBOL_LENGTH];
+			size_t len = 0;
 			int c = getchar();
 			while (c != EOF && (isalnum(c) || c == '_')) {
 				if (len >= sizeof(buf) - 1)
@@ -83,6 +126,12 @@ main(int argc, char **argv)
 			putchar(c);
 		}
 	}
+
+	if (explain_link_errors && found_grade_symbol) {
+		printf(probably_grade_error, found_grade_symbol);
+		free(found_grade_symbol);
+	}
+
 	return 0;
 }
 
@@ -92,7 +141,7 @@ main(int argc, char **argv)
 */
 
 static void 
-demangle(char *name)
+demangle(const char *orig_name)
 {
 	static const char entry[]   = "_entry_";
 	static const char mercury[] = "mercury__";
@@ -104,6 +153,8 @@ demangle(char *name)
 	   avoid a naming conflict with strchr's alter ego index() */
 
 	static const char introduced[]  = "IntroducedFrom__";
+	static const char deforestation[]  = "DeforestationIn__";
+	static const char accumulator[]  = "AccFrom__";
 	static const char pred[]  = "pred__";
 	static const char func[]  = "func__";
 
@@ -112,14 +163,41 @@ demangle(char *name)
 	static const char ho_suffix[] = "__ho"; /* added by higher_order.m */
 
 	static const char mercury_data[] = "mercury_data_";
-	static const char base_type_layout[] = "base_type_layout_";
-	static const char base_type_info[] = "base_type_info_";
-	static const char base_type_functors[] = "base_type_functors_";
+	static const char type_ctor_layout[] = "type_ctor_layout_";
+	static const char type_ctor_info[] = "type_ctor_info_";
+	static const char type_ctor_functors[] = "type_ctor_functors_";
+	static const char base_typeclass_info[] = "__base_typeclass_info_";
 	static const char common[] = "common";
+	static const char arity_string[] = "arity";
+	static const char underscores_arity_string[] = "__arity";
 
+	static const char MR_grade[] = "MR_grade_";
+	static const char MR_runtime_grade[] = "MR_runtime_grade";
+
+	static const char * trailing_context_1[] = {
+		introduced,
+		deforestation,
+		accumulator,
+		NULL
+	};
+
+	static const char * trailing_context_2[] = {
+		type_ctor_layout,
+		type_ctor_info,
+		type_ctor_functors,
+		common,
+		NULL
+	};
+
+	static const char * trailing_context_3[] = {
+		arity_string,
+		NULL
+	};
+
+	char name[MAX_SYMBOL_LENGTH];
 	char *start = name;
 	const char *module = "";	/* module name */
-	char *end = name + strlen(name);
+	char *end = name + strlen(orig_name);
 	char *position;		/* current position in string */
 	int mode_num;
 	int mode_num2;
@@ -132,8 +210,24 @@ demangle(char *name)
 	int lambda_seq_number = 0;
 	char *lambda_pred_name = NULL;
 	const char *lambda_kind = NULL;
-	enum { ORDINARY, UNIFY, COMPARE, INDEX, LAMBDA } category;
+	enum { ORDINARY, UNIFY, COMPARE, INDEX,
+		LAMBDA, DEFORESTATION, ACCUMULATOR }
+		category;
 	enum { COMMON, INFO, LAYOUT, FUNCTORS } data_category;
+	const char * class_name;
+	int class_arity;
+	char class_arg_buf[MAX_SYMBOL_LENGTH];
+	int class_arg_num;
+	const char* class_arg;
+
+	/*
+	** copy orig_name to a local buffer which we can modify,
+	** making sure that we don't overflow the buffer
+	*/
+	if (strlen(orig_name) >= sizeof(name)) {
+		goto wrong_format;
+	}
+	strcpy(name, orig_name);
 
 	/*
 	** skip any leading underscore inserted by the C compiler
@@ -143,6 +237,21 @@ demangle(char *name)
 		start++;
 	}
 
+	/*
+	** check for `MR_grade_*' and `MR_runtime_grade'.
+	*/
+	if (strncmp(start, MR_grade, strlen(MR_grade)) == 0 ||
+	    strcmp(start, MR_runtime_grade) == 0)
+	{
+		if (found_grade_symbol == NULL) {
+			found_grade_symbol = malloc(strlen(start) + 1);
+			if (found_grade_symbol != NULL) {
+				strcpy(found_grade_symbol, start);
+			}
+		}
+		goto wrong_format;
+	}
+
 	/*
 	** skip the `_entry_' prefix, if any
 	*/
@@ -240,7 +349,7 @@ demangle(char *name)
 	do {
 		if (position == start) goto wrong_format;
 		position--;
-	} while (isdigit((unsigned char)*position));
+	} while (MR_isdigit(*position));
 		/* get the mode number */
 	
 	if (check_for_suffix(start, position, ua_suffix,
@@ -266,7 +375,7 @@ demangle(char *name)
 	do {
 		if (position == start) goto wrong_format;
 		position--;
-	} while (isdigit((unsigned char)*position));
+	} while (MR_isdigit(*position));
 	if (check_for_suffix(start, position, ho_suffix,
 			sizeof(ho_suffix), &mode_num2)) {
 		end = position + 1 - (sizeof(ho_suffix) - 1);
@@ -280,14 +389,6 @@ demangle(char *name)
 	*/
 	*end = '\0';
 
-	/*
-	** Strip off the module name
-	*/
-	module = start;
-	if (!cut_at_double_underscore(&start, end)) {
-		module = "";
-	}
-
 	/*
 	** Make sure special predicates with unused_args 
 	** are reported correctly.
@@ -299,13 +400,24 @@ demangle(char *name)
 		}
 	}
 
+	module = strip_module_name(&start, end, trailing_context_1);
+
 	/*
-	** look for "IntroducedFrom"
+	** look for "IntroducedFrom" or "DeforestationIn" or "AccFrom"
 	*/
-	if (category == ORDINARY
-			&& strip_prefix(&start, introduced))
+	if (category == ORDINARY) {
+		if (strip_prefix(&start, introduced)) {
+			category = LAMBDA;
+		} else if (strip_prefix(&start, deforestation)) {
+			category = DEFORESTATION;
+		} else if (strip_prefix(&start, accumulator)) {
+			category = ACCUMULATOR;
+		}
+	}
+
+	if (category == LAMBDA || category == DEFORESTATION || 
+			category == ACCUMULATOR) 
 	{
-		category = LAMBDA;
 		if (strip_prefix(&start, pred)) {
 			lambda_kind = "pred";
 		} else if (strip_prefix(&start, func)) {
@@ -318,7 +430,7 @@ demangle(char *name)
 			goto wrong_format;
 		}
 		lambda_line = 0;
-		while (start < end && isdigit(*start)) {
+		while (start < end && MR_isdigit(*start)) {
 			lambda_line = lambda_line * 10 + (*start - '0');
 			start++;
 		}
@@ -326,7 +438,7 @@ demangle(char *name)
 			goto wrong_format;
 		}
 		lambda_seq_number = 0;
-		while (start < end && isdigit(*start)) {
+		while (start < end && MR_isdigit(*start)) {
 			lambda_seq_number = lambda_seq_number * 10 +
 				(*start - '0');
 			start++;
@@ -355,6 +467,15 @@ demangle(char *name)
 			lambda_kind, lambda_seq_number,
 			lambda_pred_name, module, lambda_line);
 		break;
+	case ACCUMULATOR:
+		printf("accumulator procedure from '%s' in module '%s' line %d",
+			lambda_pred_name, module, lambda_line);
+		break;
+	case DEFORESTATION:
+		printf("deforestation procedure (#%d) from '%s' in module '%s' line %d",
+			lambda_seq_number, lambda_pred_name,
+			module, lambda_line);
+		break;
 	default:
 		if (*module == '\0') {
 			printf("%s '%s'/%d mode %d",
@@ -385,22 +506,24 @@ not_plain_mercury:
 	if (!strip_prefix(&start, mercury_data)) {
 		goto wrong_format;
 	}
-	module = start;
-	if (!cut_at_double_underscore(&start, end)) {
-		module = "";
+
+	if (strip_prefix(&start, base_typeclass_info)) {
+		goto typeclass_info;
 	}
 
-	if (strip_prefix(&start, base_type_info)) {
+	module = strip_module_name(&start, end, trailing_context_2);
+
+	if (strip_prefix(&start, type_ctor_info)) {
 		data_category = INFO;
 		if (!cut_trailing_underscore_integer(start, &end, &arity)) {
 			goto wrong_format;
 		}
-	} else if (strip_prefix(&start, base_type_layout)) {
+	} else if (strip_prefix(&start, type_ctor_layout)) {
 		data_category = LAYOUT;
 		if (!cut_trailing_underscore_integer(start, &end, &arity)) {
 			goto wrong_format;
 		}
-	} else if (strip_prefix(&start, base_type_functors)) {
+	} else if (strip_prefix(&start, type_ctor_functors)) {
 		data_category = FUNCTORS;
 		if (!cut_trailing_underscore_integer(start, &end, &arity)) {
 			goto wrong_format;
@@ -420,28 +543,28 @@ not_plain_mercury:
 
 	case INFO:
 		if (*module == '\0') {
-			printf("<base type_info for type '%s'/%d>",
+			printf("<type_ctor_info for type '%s'/%d>",
 				start, arity);
 		} else {
-			printf("<base type_info for type '%s:%s'/%d>",
+			printf("<type_ctor_info for type '%s:%s'/%d>",
 				module, start, arity);
 		}
 		break;
 	case LAYOUT:
 		if (*module == '\0') {
-			printf("<type layout for type '%s'/%d>",
+			printf("<type_ctor_layout for type '%s'/%d>",
 				start, arity);
 		} else {
-			printf("<type layout for type '%s:%s'/%d>",
+			printf("<type_ctor_layout for type '%s:%s'/%d>",
 				module, start, arity);
 		}
 		break;
 	case FUNCTORS:
 		if (*module == '\0') {
-			printf("<type functors for type '%s'/%d>",
+			printf("<type_ctor_functors for type '%s'/%d>",
 				start, arity);
 		} else {
-			printf("<type functors for type '%s:%s'/%d>",
+			printf("<type_ctor_functors for type '%s:%s'/%d>",
 				module, start, arity);
 		}
 		break;
@@ -456,11 +579,121 @@ not_plain_mercury:
 	}
 	return;
 
+typeclass_info:
+	/*
+	** Parse the class name and class arity, which have the following
+	** layout:
+	**	<module-qualified class name>__arity<arity>__
+	*/
+	class_name = strip_module_name(&start, end, trailing_context_3);
+	/* XXX fix_mangled_ascii() */
+	if (!(strip_prefix(&start, arity_string)
+		&& strip_leading_integer(&start, &class_arity)
+		&& strip_prefix(&start, "__")))
+	{
+		goto wrong_format;
+	}
+
+	/*
+	** Parse the class argument types, which each have the following
+	** layout:
+	**	<module-qualified type name>__arity<arity>__
+	**
+	** We store the human-readable formatted output in
+	** class_arg_buf as we go.
+	*/
+	fix_mangled_ascii(start, &end);
+	strcpy(class_arg_buf, "");
+	for (class_arg_num = 0; class_arg_num < class_arity; class_arg_num++) {
+		if (class_arg_num != 0) {
+			strcat(class_arg_buf, ", ");
+		}
+		class_arg = strip_module_name(&start, end, trailing_context_3);
+		if (!(strip_prefix(&start, arity_string)
+		      && strip_leading_integer(&start, &arity)
+		      && strip_prefix(&start, "__")))
+		{
+			goto wrong_format;
+		}
+		sprintf(class_arg_buf + strlen(class_arg_buf),
+			"%s/%d", class_arg, arity);
+	}
+		
+	/*
+	** now print the results
+	*/
+	printf("<instance declaration for %s(%s)>",
+		class_name, class_arg_buf);
+	return;
+
 wrong_format:
-	printf("%s", name);
+	printf("%s", orig_name);
 	return;
 } /* end demangle() */
 
+	/*
+	** Remove a module name prefix.
+	** Just keep munching up double-underscores until we
+	** get to something that matches the specified trailing context,
+	** at which point we stop, or until there are no double-underscores
+	** left.
+	*/
+static const char *
+strip_module_name(char **start_ptr, char *end, const char *trailing_context[])
+{
+	const char *module;		/* module name */
+	char *module_end;		/* end of the module name */
+	char *next_double_underscore;
+	char *start;
+
+	start = *start_ptr;
+
+	/*
+	** Strip off the module name
+	*/
+	module = start;
+	module_end = start;
+	while ((next_double_underscore = strstr(start, "__")) != NULL) {
+		int len, i;
+
+		/*
+		** Check for special cases
+		*/
+		bool stop = FALSE;
+		for (i = 0; trailing_context[i] != NULL; i++) {
+			if (strncmp(start,
+				trailing_context[i],
+				strlen(trailing_context[i])) == 0)
+			{
+				stop = TRUE;
+			}
+		}
+		if (stop) break;
+
+		len = next_double_underscore - start;
+		if (module != module_end) {
+			/*
+			** append a module qualifier, and
+			** shift the module name into the right place
+			*/
+			*module_end = ':';
+			module_end++;
+			memmove(module_end, start, len);
+		}
+		module_end += len;
+
+		start = next_double_underscore + 2;
+	}
+	if (module == module_end) {
+		module = "";
+	} else {
+		*module_end = '\0';
+	}
+
+	*start_ptr = start;
+	return module;
+}
+
 	/*
 	** Remove the prefix from a string, if it has 
 	** it. 
@@ -483,6 +716,36 @@ strip_prefix(char **str, const char *prefix)
 	return FALSE;
 }
 
+	/*
+	** If the string pointed to by *start_ptr starts with
+	** an integer, then advance *start_ptr past the leading integer,
+	** store the value of the integer in the int pointed to by `num',
+	** and return true; otherwise leave *start_ptr unchanged and
+	** return false.  (The string itself is always left unchanged.)
+	*/
+static bool 
+strip_leading_integer(char **start_ptr, int *num) 
+{
+	char *start = *start_ptr;
+	char save_char;
+	bool got_int;;
+
+	while(MR_isdigit(*start)) {
+		start++;
+	}
+	if (start == *start_ptr) return FALSE;
+	save_char = *start;
+	*start = '\0';
+	got_int = (sscanf(*start_ptr, "%d", num) == 1);
+	*start = save_char;
+	if (got_int) {
+		*start_ptr = start;
+		return TRUE;
+	} else {
+		return FALSE;
+	}
+}
+
 	/*
 	** Remove trailing integer (at the supplied `real_end' of the
 	** string), and return it in the int pointed to by `num'.   True
@@ -498,7 +761,7 @@ cut_trailing_integer(char *str, char **real_end, int *num)
 	do { 
 		if (end == str) return FALSE;
 		end--;
-	} while (isdigit((unsigned char)*end));
+	} while (MR_isdigit(*end));
 
 	if (sscanf(end + 1, "%d", num) != 1) {
 		return FALSE;
@@ -597,12 +860,12 @@ fix_mangled_ascii(char *str, char **real_end)
 	** ASCII codes back into an identifier.
 	*/
 	if (strncmp(str, "f_", 2) == 0) {
-		char buf[1000];
+		char buf[MAX_SYMBOL_LENGTH];
 		char *num = str + 2;
 		int count = 0;
 		while (num < end) {
 			char *next_num = num;
-			while (isdigit((unsigned char)*next_num)) {
+			while (MR_isdigit(*next_num)) {
 				next_num++;
 			}
 			if (*next_num != '_' && *next_num != '\0') 
